import java.awt.*;
import java.sql.*;
import org.garret.consus.*;

/**
 * Main class of the Graph Editor example for Consus. This class implements a GUI editor with a set of primitive tools for drawing:
 * line, point, circle and rectangle. Each drawing is stored in a separate record in the database, using the Consus
 * object-oriented interface extension.
 * @see Drawing
 */
public class GraphEditor extends Frame {
    int x, y;
    Figure figure;
    Graphics graphics;
    Drawing drawing;
    Menu addMenu;
    boolean selecting;
    ConsusConnection db;
    PreparedStatement stmt;

    /** Constructor for Graph Editor, which accepts a ConsusConnection object. */
    public GraphEditor(ConsusConnection conxn) throws Exception {
        super("Graph Editor");
        db = conxn;
        stmt = db.prepareStatement("select * from Drawing where name=?");
        MenuBar mainMenu = new MenuBar();
        Menu fileMenu = new Menu("Record");
        fileMenu.add(new MenuItem("New"));
        fileMenu.add(new MenuItem("Open"));
        fileMenu.addSeparator();
        fileMenu.add(new MenuItem("Exit"));
        mainMenu.add(fileMenu);
        addMenu = new Menu("Tool");
        addMenu.add(new MenuItem("Point"));
        addMenu.add(new MenuItem("Line"));
        addMenu.add(new MenuItem("Circle"));
        addMenu.add(new MenuItem("Rectangle"));
        addMenu.disable();
        mainMenu.add(addMenu);
        setMenuBar(mainMenu);
        setLayout(null);
        addNotify();
        resize(insets().left + insets().right + 500, insets().top + insets().bottom + 300);
        show();
        setCursor(CROSSHAIR_CURSOR);
        graphics = getGraphics();
    }

    /** Event handler procedure. */
    public boolean handleEvent(Event event) {
        switch (event.id) {
            case Event.WINDOW_DESTROY:
                close();
                return true;
            case Event.MOUSE_DOWN:
                x = event.x;
                y = event.y;
                if (y >= insets().top && figure != null) {
                    figure.set(x, y, x, y);
                    if (figure instanceof Point) {
                        figure.paint(graphics);
                        drawing.add((Figure)figure.clone());
                    } else if (figure != null) {
                        if (!selecting) {
                            graphics.setXORMode(Color.red);
                            figure.paint(graphics);
                        }
                        selecting = true;
                    }
                    return true;
                }
                break;
            case Event.MOUSE_DRAG:
                if (selecting && figure != null) {
                    graphics.setXORMode(Color.red);
                    figure.paint(graphics);
                    figure.set(x, y, event.x, event.y);
                    figure.paint(graphics);
                }
                return true;
            case Event.MOUSE_UP:
                if (figure != null && selecting) {
                    figure.paint(graphics);
                    graphics.setPaintMode();
                    figure.set(x, y, event.x, event.y);
                    figure.paint(graphics);
                    drawing.add((Figure)figure.clone());
                    selecting = false;
                }
                return true;
        }
        return super.handleEvent(event);
    }

    /** Draws the current drawing to the screen. */
    public void paint(Graphics g) {
        super.paint(g);
        if (drawing != null) {
            graphics.setPaintMode();
            drawing.paint(g);
        }
    }

    /** The handler for menu selection events. */
    public boolean action(Event event, Object arg) {
        if (event.target instanceof MenuItem) {
            String label = (String)arg;
            if (label.equals("Exit")) {
                handleEvent(new Event(this, Event.WINDOW_DESTROY, null));
            } else if (label.equals("Open")) {
                openGraphic();
            } else if (label.equals("New")) {
                newGraphic();
            } else if (label.equals("Point")) {
                figure = new Point();
            } else if (label.equals("Line")) {
                figure = new Line();
            } else if (label.equals("Circle")) {
                figure = new Circle();
            } else if (label.equals("Rectangle")) {
                figure = new Rectangle();
            } else {
                return super.action(event, arg);
            }
            return true;
        }
        return super.action(event, arg);
    }

    public static void main(String[] args) throws Exception {
        Class.forName("org.garret.consus.ConsusDriver");
        String file = "grapheditor.dbs";
        if (args.length >= 1) {
            file = args[0];
        }
        Connection connection = DriverManager.getConnection("jdbc:consus:" + file);
        if (connection == null) {
            System.err.println("No connection with database " + file);
            System.exit(1);
        }
        new GraphEditor((ConsusConnection)connection);
    }

    /**
     * Displays the "open drawing" dialog. Class <code>org.garret.consus.TableIterator</code> is used to iterate through
     * the drawing records in the database.
     */
    public void openGraphic() {
        List list = new List();
        try {
            TableIterator iterator = db.getTableIterator("Drawing");
            while (iterator.next()) {
                list.addItem(iterator.get().toString());
            }
            iterator.close();
        } catch (Exception x) {
            System.err.println("Caught exception: " + x);
        }
        new OpenDialog(this, list);
    }

    /** Displays "new drawing" dialog. */
    public void newGraphic() {
        new NewDialog(this);
    }

    /** Called by NewDialog in response to Ok selection. Creates a new Drawing and stores it immediately in the database. */
    public synchronized void newDrawing(String name) {
        drawing = new Drawing(name);
        drawing.store(db);
        repaint();
    }

    /**
     * Called by OpenDrawing dialog in response to the selection of an existing drawing. Retrieves the selected Drawing object
     * from the database using the <code>java.sql.ResultSet.getObject</code> method.
     */
    public synchronized void openDrawing(String name) {
        try {
            stmt.setString(1, name);
            ResultSet cursor = stmt.executeQuery();
            if (cursor.next()) {
                drawing = (Drawing)cursor.getObject(0);
                repaint();
            }
        } catch (SQLException x) {
            System.err.println("Caught exception: " + x);
        }
    }

    /** Closes the editors frame. */
    public synchronized void close() {
        hide();
        dispose();
        System.out.println("Close database session");
        try {
            db.close();
        } catch (SQLException x) {
            System.err.println("Caught exception: " + x);
        }
        System.exit(0);
    }
}
