/* 
 * "Hello, world!" example
 */

import java.io.*;
import java.sql.*;
import java.text.*;

/**
 * This example tests the basic components of the JDBC driver, and shows
 * how even the simplest of queries can be implemented.
 * 
 * To  run  this  example you  need  to  specify the  location  of  an existing
 * database. A new database will be created  if it  doesn't exist. This example
 * will create a single table called Hello and remove it after running the test.
 */

public class HelloWorld
{
  Connection db;	// The connection to the database
  Statement  st;	// Our statement to run queries with
  
  
  /**
   * The constructor accepts an array of strings as parameters:
   * <OL>
   * <LI>URL to the database
   * <LI>user name to be used for establishing connection
   * <LI>password for user authentication
   * </OL> 
   */
  public HelloWorld(String[] args) throws ClassNotFoundException, FileNotFoundException, IOException, SQLException
  {
    String url = args[0];
    String usr = args[1];
    String pwd = args[2];
    
    // Load the driver
    Class.forName("org.garret.consus.ConsusDriver");
    
    // Connect to the database
    System.out.println("Connecting to Database URL = " + url);
    db = DriverManager.getConnection(url, usr, pwd);
    
    System.out.println("Connected...Now creating a statement");
    st = db.createStatement();
    
    // Clean up the database (in case we failed earlier) and initialise
    cleanup();
    
    // Now run tests using the JDBC methods
    doexample();
    
    // Clean up the database
    cleanup();
    
    // Finally close the database
    System.out.println("Now closing the connection");
    st.close();
    db.close();
  }
  
  /**
   * Drops (removes) the table from the database (if it existed). No errors are reported.
   */
  public void cleanup()
  {
    try {
      st.executeUpdate("drop table Hello");
    } catch(Exception ex) {
      // We ignore any errors here
    }
  }
  
  /**
   * Main method of the example. The execution is tracked with comments displayed to the standard output.
   */
  public void doexample() throws SQLException
  {
    System.out.println("\nRunning tests:");
    
    // First we need a table to store the data in
    st.executeUpdate("create table Hello (a smallint, b smallint)");
    
    // Now insert some data, using the Statement
    st.executeUpdate("insert into Hello values (1,1)");
    st.executeUpdate("insert into Hello values (2,1)");
    st.executeUpdate("insert into Hello values (3,1)");
    
    // Now change the value of b from 1 to 8
    st.executeUpdate("update Hello set b=8");
    System.out.println("Updated "+st.getUpdateCount()+" rows");
    
    // For large inserts PreparedStatement is more efficient, because it
    // supports the idea of precompiling a SQL statement, and storing
    // directly a Java object into a column.
    //
    PreparedStatement ps = db.prepareStatement("insert into Hello values (?,?)");
    for(int i=2;i<5;i++) {
      ps.setInt(1,4);		// "column a" = 5
      ps.setInt(2,i);		// "column b" = i
      ps.executeUpdate();	// executeUpdate because insert returns no data
    }
    ps.close();			// Always close when we are done
    
    // Finally perform a query on the table
    System.out.println("performing query");
    ResultSet rs = st.executeQuery("select a, b from Hello");
    if(rs!=null) {
      // Now we will run through the result set, printing out the result.
      // Note, that we must call .next() before attempting to read any results
      while(rs.next()) {
		int a = rs.getInt("a");	// This shows how to get the value by name
		int b = rs.getInt(2);	// This shows how to get the value by column
		System.out.println("  a="+a+" b="+b);
      }
      rs.close();	// again, we must close the result when done
    }
    
    // Now run the query again, showing a more efficient way of getting the
    // result if the column number of the value is not known.
    System.out.println("performing another query");
    rs = st.executeQuery("select * from Hello where b>1");
    if(rs!=null) {
      // First, find out the column numbers.
      //
      // It's best to do it here, since calling the methods with the column
      // names actually performs this call each time they are called. This
      // really speeds things up on large queries.
      //
      int col_a = rs.findColumn("a");
      int col_b = rs.findColumn("b");
      
      // Now we will run through the result set, printing out the result.
      // Again, we must call .next() before attempting to read any results
      while(rs.next()) {
		int a = rs.getInt(col_a);
		int b = rs.getInt(col_b);
		System.out.println("  a="+a+" b="+b);
      }
      rs.close();	// again, we must close the result when done
    }
    
    // The last thing to do is to drop the table. This is done in the
    // cleanup() method.
  }
  
  /**
   * Display instructions on how to run the example
   */
  public static void instructions()
  {
    System.out.println("\nThis example tests the basic components of the JDBC driver, demonstrating\nhow to build simple queries in java.\n");
    System.out.println("Useage:\n java HelloWorld jdbc:consus:database user password");
    System.exit(1);
  }
  
  /**
   * This little lot starts the test
   */
  public static void main(String[] args)
  {
    System.out.println("Hello, world!\n");
    
    if(args.length<3)
      instructions();
    
    // Now run the tests
    try {
      HelloWorld test = new HelloWorld(args);
    } catch(Exception ex) {
      System.err.println("Exception caught.\n"+ex);
      ex.printStackTrace();
    }
  }
}
