/* Consus - Pure Java Object Relational Database Management System
 * Generic operations with arrays. This class extends <code>java.util.Arrays</code> class. 
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.consus;

import java.lang.reflect.Array;

/**
 * Modification (insert, remove) and search operastions with Java arrays
 */
public class AnyArray { 
    /** 
     * Extend array by one element 
     * @param arr original array
     * @return new array 
     */
    public static Object insert(Object arr) { 
	return insert(arr, Array.getLength(arr), 1);
    }

    /** 
     * Insert n elemenst at the end of array
     * @param arr original array
     * @param n number of inserted elements 
     * @return new array 
     */
    public static Object insert(Object arr, int n) { 
	return insert(arr, Array.getLength(arr), n);
    }
    /** 
     * Insert <code>n</code> elements at the position <code>pos</code>.
     * @param arr original array
     * @param pos insert position
     * @param n number of elements to be added 
     * @return new array 
     */
    public static Object insert(Object arr, int pos, int n) { 
	int len = Array.getLength(arr);
	Object newArr = Array.newInstance(arr.getClass().getComponentType(), len+n);
	System.arraycopy(arr, 0, newArr, 0, pos);
	System.arraycopy(arr, pos, newArr, pos+n, len-pos);
	return newArr;
    }

    /** 
     * Truncate array by one element 
     * @param arr original array
     * @return new array 
     */
    public static Object remove(Object arr) { 
	int len = Array.getLength(arr);
	return remove(arr, len-1, 1); 
    }
    /** 
     * Remove one element from the specified position
     * @param arr original array
     * @param from position of removed element
     * @return new array 
     */
    public static Object remove(Object arr, int from) { 
	return remove(arr, from, 1);
    }

    /** 
     * Remove <code>n</code> elements starting from the position <code>from</code>
     * @param arr original array
     * @param from position of removed element
     * @param n number of elements to be removed
     * @return new array 
     */
    public static Object remove(Object arr, int from, int n) { 
	int len = Array.getLength(arr);
	Object newArr = Array.newInstance(arr.getClass().getComponentType(), len-n);
	System.arraycopy(arr, 0, newArr, 0, from);	
	System.arraycopy(arr, from+n, newArr, from, len-from-n);
	return newArr;
    }

    /** 
     * Remove object from the array. Method first tries to locate specified object 
     * and remove it if the object was found. Method uses == operator for comparing objects.
     * @param arr original array
     * @param elem object to be remove
     * @return new array (the same if object was not found in array)
     */
    public static Object remove(Object[] arr, Object elem) { 
	for (int i = arr.length; --i >= 0;) { 
	    if (arr[i] == elem) { 
		return remove(arr, i, 1);
	    }
	}
	return arr;
    }
    
    /**
     * Append new element to the array if there is no such element in the array.
     * @param arr original array
     * @param val object to be appended
     * @return new array 
     */
    public static Object add(Object[] arr, Object val) { 
	for (int i = arr.length; --i >= 0;) { 
	    if (arr[i] == val) { 
		return arr;
	    }
	}	
	Object newArr = insert(arr);
	Array.set(newArr, Array.getLength(newArr)-1, val);
	return newArr;
    }

    /**
     * Append one element to the array.
     * @param arr original array
     * @param val object to be appended
     * @return new array 
     */
    public static Object append(Object[] arr, Object val) { 
	Object newArr = insert(arr);
	Array.set(newArr, Array.getLength(newArr)-1, val);
	return newArr;
    }
    /**
     * Insert object at the first position in the array.
     * @param arr original array
     * @param val object to be inserted
     * @return new array 
     */
    public static Object prepend(Object[] arr, Object val) { 
	Object newArr = insert(arr, 0, 1);
	Array.set(newArr, 0, val);
	return newArr;
    }
    
    /** 
     * Find object in the array. The method uses == operator for comparing objects.
     * @param arr original array
     * @param val object to be found
     * @return index of object (-1 if object is not present in the array).
     */
    public static int find(Object[] arr, Object val) { 
	for (int i = arr.length; --i >= 0;) { 
	    if (arr[i] == val) { 
		return i;
	    }
	}
	return -1;
    }

    /** 
     * Get array length.
     * @param arr array
     * @return number of elements in the array.
     */     
    public static int length(Object arr) { 
	return Array.getLength(arr);
    }

    /** 
     * Copy array elements. This method calls <code>System.arraycopy</code> but
     * with the order of parameters as expected by C programmers.
     * @param dst destination array
     * @param dstPos start position in destination array
     * @param src source array
     * @param srcPos start position in source array
     * @param n number of elements to be copied
     */     
    public static void memcpy(Object dst, int dstPos, Object src, int srcPos, int n) { 
	System.arraycopy(src, srcPos, dst, dstPos, n);
    }
}
	



