/* Consus - Pure Java Object Relational Database Management System
 * Connection protocol implementation (most of the database methods are implemented here)
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.consus;

import java.sql.*;
import java.io.*;


/**
 * <P>A connection (session) with a specific
 * database. Within the context of a Connection, SQL statements are
 * executed and results are returned.
 *
 * <P>A Connection's database is able to provide information
 * describing its tables, supported SQL grammar, its stored
 * procedures, the capabilities of this connection and so on. This
 * information is obtained with the <code>getMetaData</code> method.
 *
 * <P><B>Note:</B> By default the Connection automatically commits
 * changes after executing each statement. If auto commit has been
 * disabled, an explicit commit must be done or the database changes will
 * not be saved.
 *
 * @see Statement 
 * @see ResultSet
 * @see DatabaseMetaData
 */
public interface ConsusConnection extends Connection, PersistentObjectStorage {
    /** 
     * Set autocommit mode only for the current thread. Transaction commit modes of other
     * threads are not changed. Transactions will be automatically committed after each
     * statement when either the global auto-commit mode is set or the current thread
     * is in the auto-commit state.
     */
    void setThreadAutoCommit(boolean autoCommit) throws SQLException;


    /**
     * Gets the current auto-commit state.
     *
     * @return the current state of the auto-commit mode
     * @exception SQLException if a database access error occurs
     * @see #setAutoCommit 
     */
    boolean getThreadAutoCommit() throws SQLException;

    /** 
     * Insert the specified object into the database. The table in which object will be inserted
     * is determined by the object class.
     * @param obj Java object to be inserted
     * @return reference to the cerated object
     */
    Ref insert(Object obj) throws SQLException;

    /**
     * Update the object in the table.
     *
     * @param ref reference to the updated object
     * @param obj object with new values 
     */
    void update(Ref ref, Object obj) throws SQLException;

    /**
     * Get object by it's OID
     *
     * @param ref reference to the object
     * @return reference to the fetched object
     */
    Object get(Ref ref) throws SQLException;

    /**
     * Remove object with specifed OID
     *
     * @param ref reference to the removed object
     */
    void remove(Ref ref) throws SQLException;

    /** 
     * Lock the database in exclusive mode. Lock will be released by transaction commit/rollback
     */
    void lock() throws Exception;

    /**
     * Commit nested subtransaction. Exterior transaction will be committed when all nested
     * subtransaction are successfully committed.
     */ 
    void commitSubtransaction() throws SQLException; 

    /**
     * Online backup of the database.
     * @param stream to were backup is performed
     */
    void backup(java.io.OutputStream stream) 
	throws SQLException, java.io.IOException;

    /**
     * Allocate object cache (instance manager) if not yet allocated.
     * With object caching enabled for each OID, there is exactly one instance
     * of the object in memory. The object cache is implicitly created when the
     * PersistentObjectStorage interface is used. 
     */
    void createObjectCache(); 

    /**
     * Get the size used by objects in the database
     * @return total size of all the objects in the database (bytes)
     */
    long getUsedSize() throws SQLException;

    /**
     * Specify root tables for garbage collection. All tables should already exist.
     * All records (objects) for these tables will be marked as accessible. Then all
     * other objects in the database, recursively referenced from these objects, will be marked.
     * At the final stage, all unmarked objects will be removed.
     * @param tableNames array with table names
     */
    void setGCRoots(String tableNames[]) throws SQLException;
    
    /** 
     * Specify the size of allocated objects after which GC should be initiated.
     * @param threshold difference between allocated and deallocated object since last GC
     * after which a new GC is initiated. By default it is set to <code>Long.MAX_VALUE</code>
     * so GC is disabled.
     */
    void setGCThreshold(long threshold) throws SQLException;

    /** 
     * Explicitly initiate new garbage collection (if not yet started)
     */
    void doGarbageCollection() throws SQLException;
    
    /** 
     * Update replica table by transfering all changes from the master table.
     * It is not necessary for master and replica table to be exactly the same. Replica table
     * can contain a subset of master table columns. But they both should have
     * <code>recordTimestamp</code> and optionally <code>masterOID</code> fields. 
     * Then the application can close connection with the server and work normally with
     * replica table, autonomously.
     * 
     * @param tableName name of the table to check-out
     * @param conxn database connection
     */
    void checkOutTable(String tableName, Connection conxn) 
	throws SQLException;
	
    /** 
     * Synchronize the replica and master table.
     * This method will first transfer 
     * to the client all the records that were changed on the server from the moment of the
     * check out. If some of these records was also changed by the client, then a merge
     * conflict is detected and <code>MergeConflictException</code> is thrown. 
     * The exception object is given the value of the key of the record that caused the conflict.
     * Then Consus will transfer all the records which were changed by the client to the master
     * table, assigning them the value of timestamp greater than of any other record in the 
     * master table. It is possible to perform check-ins several times, synchronizing content
     * of the master and replica tables.
     *
     * @param tableName name of the table to check-in
     * @param conxn database connection
     * @exception MergeConflictException
     */
    void checkInTable(String tableName, Connection conxn) 
	throws SQLException;


    /**
     * Get the name of the field containing the referenence to the master object instance
     * on the server (used for table replication)
     * if <code>null</code> then an existing primary key will be used instead.
     * @return name of master field ("masterOID" by default)
     */
    String getMasterFieldName();

    /**
     * Set the name of the field containing the refenence to the master object instance on the
     * server (used for table replication).
     * @param name name of the field, if <code>null</code>
     * then an existing primary key will be used instead
     */

    void setMasterFieldName(String name);


    /**
     * Get the name of the timestamp field used for data replication.
     * @return timestamp field name (recordTimestamp" by default)
     */
    String getTimestampFieldName();

    /**
     * Set the name of the timestamp field used for data replication.
     * @param name name of the timestamp field
     */
    void setTimestampFieldName(String name);
    
    /**
     * Create a new String index in the database.
     * @return persistent object representing string index
     */
    StringIndex createStringIndex() throws SQLException;

    /**
     * Create BLOB
     * @param length size of stored data
     * @param in input stream to provide BLOB data
     * @return <code>ConsusBlob</code>
     */
    Blob createBlob(long length, InputStream in);
    
    /**
     * Create CLOB
     * @param length number of characters
     * @param in reader to provide CLOB data
     * @return <code>ConsusClob</code>
     */
    Clob createClob(long length, Reader in);
    
    /**
     * Create CLOB
     * @param str - string which will be stored as CLOB body
     * @return <code>ConsusClob</code>
     */
    Clob createClob(String str);

    /**
     * Switch tracing on/off.
     * @param enabled - switch on tracing when <code>true</code> 
     */
    void enableTracing(boolean enabled);
    
}