/* Consus - Pure Java Object Relational Database Management System
 * Reference to the Consus database object
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.consus;

import java.sql.*;
 
/**
 * JDBC 2.0
 *
 * A reference to an SQL structured type value in the database.  A
 * Ref can be saved to a persistent storage.  A Ref is dereferenced by
 * passing it as a parameter to an SQL statement and executing the 
 * statement.
 */
public class ConsusRef implements Ref, Comparable, java.io.Serializable
{
    long                    oid;
    PersistentObjectStorage db;
    
    /**
     * Gets the fully-qualified SQL structured type name of the 
     * referenced item.
     * 
     * @return fully-qualified SQL structured type name of the referenced item.
     * @exception SQLException if a database access error occurs
     */
    public String getBaseTypeName() throws SQLException { 
	Object obj = getObject();
	return obj == null ? null : obj.getClass().getName();
    }

    /**
     * Retrieves the referenced object and maps it to a Java type
     * using the given type map.
     *
     * @param map a <code>java.util.Map</code> object that contains 
     *        the mapping to use (the fully-qualified name of the SQL
     *        structured type being referenced and the class object for
     *        <code>SQLData</code> implementation to which the SQL
     *        structured type will be mapped)
     * @return  a Java <code>Object</code> that is the custom mapping for 
     *          the SQL structured type to which this <code>Ref</code>
     *          object refers
     * @exception SQLException if a database access error occurs
     * @since 1.4
     * @see #setObject
     */
    public Object getObject(java.util.Map map) throws SQLException
    {
	return getObject();
    }

    /**
     * Retrieves the SQL structured type instance referenced by
     * this <code>Ref</code> object.  If the connection's type map has an entry
     * for the structured type, the instance will be custom mapped to
     * the Java class indicated in the type map.  Otherwise, the
     * structured type instance will be mapped to a <code>Struct</code> object.
     *
     * @return  a Java <code>Object</code> that is the mapping for 
     *          the SQL structured type to which this <code>Ref</code>
     *          object refers
     * @exception SQLException if a database access error occurs
     * @since 1.4
     * @see #setObject
     */   
    public Object getObject() throws SQLException
    {
	if (oid == 0) { 
	    return null;
	} else if (db != null) { 
	    try { 
		return db.getObjectByOid(oid);
	    } catch (Exception x) { 
		throw new SQLException("getObjectByOid(" + oid + ") failed: " + x);
	    }
	} else { 
	    throw new ConsusRuntimeError("Reference is not assiciated with any database");
	}
    }
	
    /**
     * Sets the structured type value that this <code>Ref</code>
     * object references to the given instance of <code>Object</code>.
     * The driver converts this to an SQL structured type when it
     * sends it to the database.
     *
     * @param value an <code>Object</code> representing the SQL 
     *        structured type instance that this
     *        <code>Ref</code> object will reference
     * @exception SQLException if a database access error occurs
     * @since 1.4
     * @see #getObject()
     * @see #getObject(Map)
     * @see PreparedStatement#setObject(int, Object)
     */
    public void setObject(Object value) throws SQLException {
	oid = ((Persistent)value).oid;
    }

    /**
     * Get the OID.
     * @return object identifier 
     */
    final public long getOid() { 
	return oid;
    }

    /**
     * Compare two references for equality
     * @return <code>true</code> if there are references to the same object or
     * <code>null</code> references, <code>false</code> otherwise.
     */
    public boolean equals(Object ref) { 
	if (ref == null) { 
	    return oid == 0;
	}
	return oid == ((ConsusRef)ref).oid;
    }

    /**
     * Compare two OID as longs
     * @return <code>-1</code> if less than, 0 if equal, and 1 if greater
     */
    public int compareTo(Object obj) { 
	if (obj instanceof ConsusRef) { 
	    long diff = oid - ((ConsusRef)obj).oid;
	    return diff < 0 ? -1 : diff == 0 ? 0 : 1;
	} 
	throw new ClassCastException();
    }

    /**
     * Convert reference to a String.
     * @return integer representing OID with prepended '@' symbol
     */
    public String toString() {
	return "@" + oid;
    }

    /** 
     * Rerference the hash code.
     * @return OID
     */
    public int hashCode() { 
	return (int)oid;
    }

    /**
     * Construct a reference to the object with the specified OID
     * @param oid OID of the object or 0
     */
    public ConsusRef(long oid) { 
	this.oid = oid;
    }

    /**
     * Construct a reference to the object with the specified OID
     * @param oid OID of the object or 0
     */
    public ConsusRef(PersistentObjectStorage db, long oid) { 
	this.db = db;
	this.oid = oid;
    }
}

