/* Consus - Pure Java Object Relational Database Management System
 * Implementation of ResultSet protocol. Bidirectional, random accessible database cursor.
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.consus;

import java.sql.*;

/**
 * <P>A ResultSet provides access to a table of data.  A ResultSet
 * object is usually generated by executing a Statement. 
 * 
 * <P>A ResultSet maintains a cursor pointing to its current row of
 * data.  Initially the cursor is positioned before the first row.
 * The 'next' method moves the cursor to the next row.
 *
 * <P>The getXXX methods retrieve column values for the current
 * row.  You can retrieve values using either the index number of the
 * column or the name of the column.  In general, using the 
 * column index will be more efficient.  Columns are numbered from 1.
 *
 * <P>For maximum portability, ResultSet columns within each row should be
 * read in left-to-right order and each column should be read only once.
 *
 * <P>For the getXXX methods, the JDBC driver attempts to convert the
 * underlying data to the specified Java type and returns a suitable
 * Java value.  See the JDBC specification for allowable mappings
 * from SQL types to Java types with the ResultSet.getXXX methods.
 *
 * <P>Column names used as input to the getXXX methods are case
 * insensitive.  When performing a getXXX using a column name, if
 * several columns have the same name, then the value of the first
 * matching column will be returned. The column name option is
 * designed to be used when column names are used in the SQL
 * query. For columns that are NOT explicitly named in the query, it
 * is best to use the column numbers. If column names are used, there is
 * no way for the programmer to guarantee that they actually refer to
 * the intended columns.
 *
 * <P>A ResultSet is automatically closed by the Statement that
 * is generated when that Statement is closed, re-executed, or used
 * to retrieve the next result from a sequence of multiple results.
 * 
 * <P>The number, types and properties of a ResultSet's columns are
 * provided by the ResultSetMetaData object returned by the getMetaData
 * method.
 *
 * @see Statement#executeQuery 
 * @see Statement#getResultSet 
 * @see ResultSetMetaData 
 */

public interface ConsusResultSet extends ResultSet {
    /**
     * Deletes all selected records.<BR>
     * 
     * @exception SQLException if a database access error occurs or if
     * called when on the insert row.
     */
    void deleteAllSelectedRows() throws SQLException;

    /**
     * Extract the current table row as a Java object.
     * @return current object
     */
    Object getSelfObject() throws SQLException; 

    /**
     * Insert the new object into the table.
     * @param obj inserted object
     * @return OID of the object in the database
     */
    Ref insertObject(Object obj) throws SQLException;

    /**
     * Update the current row in the result set.
     * @param obj new version of the object. Should be of the same type as
     * the original object
     */
    void updateSelfObject(Object obj) throws SQLException;

    /**
     * Get the OID of the current record (object) in the result set.
     * @return <code>ConsusRef</code> with the OID of the current object
     */
    Ref getSelfRef(); 

    /**
     * Get the number of records (object) in the result set.
     * @return selection size
     */
    int getSelectionSize(); 
}