/* Consus - Pure Java Object Relational Database Management System
 * Database file inspection utility
 * Created by Konstantin Knizhnik, Mar 2000.
 */
package org.garret.consus;

import java.io.*;

public class Dump { 
    static int unpack4(byte[] arr, int offs) { 
	return (arr[offs] << 24) | ((arr[offs+1] & 0xFF) << 16)
	    | ((arr[offs+2] & 0xFF) << 8) | (arr[offs+3] & 0xFF);
    }
    static long unpack8(byte[] arr, int offs) { 
	return ((long)unpack4(arr, offs) << 32)
	    | (unpack4(arr, offs+4) & 0xFFFFFFFFL);
    }
    static final int dbDatabaseOffsetBits = 37;  // up to 1 terabyte database
    static final int dbAllocationQuantumBits = 6;
    static final int dbAllocationQuantum = 1 << dbAllocationQuantumBits;
    static final int pageBits = 12;
    static final int pageSize = 1 << pageBits;
    static final int dbBitmapSegmentBits = pageBits + 3 + dbAllocationQuantumBits;
    static final int dbBitmapPages = 1 << (dbDatabaseOffsetBits-dbBitmapSegmentBits);
    static final int dbBitmapId    = 2;
    static final int dbFirstUserId = dbBitmapId + dbBitmapPages;

    public static void main(String args[]) throws Exception {
	if (args.length < 1) { 
	    System.out.println("Usage: java Dump <database-file>");
	    return;
	}
	RandomAccessFile in = new RandomAccessFile(args[0], "r");
	byte page[] = new byte[pageSize];
	in.read(page);
	int offs = 0;
	int curr = page[offs++];
	if (page[offs++] != 0) { 
	    System.out.println("Database was not normally closed");
	} 
	if (page[offs++] == 0) { 
	    System.out.println("Database was not initialized");
	    return;
	} 
	offs += curr*64;
	System.out.println("Database size: " + unpack8(page, offs));
	offs += 8;
	long indexPos = unpack8(page, offs);
	offs += 16;
	System.out.println("Size used by objects: " + unpack8(page, offs));
	offs += 8;
	int indexSize = unpack4(page, offs);
	offs += 8;
	System.out.println("Index size: " + indexSize);
	int indexUsed = unpack4(page, offs);
	offs += 8;
	System.out.println("Used index size: " + indexUsed);
	int bitmapEnd = unpack4(page, offs);
	System.out.println("Bitmap pages: " + (bitmapEnd - dbBitmapId));
	byte index[] = new byte[indexUsed*8];
	in.seek(indexPos);
	in.read(index);
	long usedSize = 0;
	long freeSize = 0;
	int nHoles = 0;
	int minHoleSize = Integer.MAX_VALUE;  
	int maxHoleSize = 0; 
	int lastHoleSize = 0;
	int i, j;
	int[] histogram = new int[1000000];
	int nUsedObjects = 0;
	int nPages = 0;
	int nFreeObjects = 0;
	for (i = dbFirstUserId; i < bitmapEnd; i++) { 
	    long pos = unpack8(index, i*8);
	    if ((pos & 1) != 0) nPages += 1;
	    else if ((pos & 4) != 0) nFreeObjects += 1;
	    else nUsedObjects += 1;
	}
	System.out.println("Number of used objects: " + nUsedObjects);
	System.out.println("Number of free objects: " + nFreeObjects);
	System.out.println("Number of used pages: " + nPages);
	for (i = dbBitmapId; i < bitmapEnd; i++) { 
	    long pos = unpack8(index, i*8) - 1;
	    in.seek(pos);
	    in.read(page);
	    j = 0;
	    while (true) { 
		while (j < pageSize*8 && (page[j>>3] & (1 << (j&7))) == 0) { 
		    j += 1;
		    lastHoleSize += 1;
		}
		if (j < pageSize*8) { 
		    if (lastHoleSize != 0) { 
			if (lastHoleSize > maxHoleSize) { 
			    maxHoleSize = lastHoleSize;
			} 
			if (lastHoleSize < minHoleSize) { 
			    minHoleSize = lastHoleSize;
			}
			freeSize += lastHoleSize;
			nHoles += 1;
			if (lastHoleSize < histogram.length) { 
			    histogram[lastHoleSize] += 1;
			}
			lastHoleSize = 0;
		    }
		} else { 
		    break;
		}
		while (j < pageSize*8 && (page[j>>3] & (1 << (j&7))) != 0) { 
		    j += 1;
		    usedSize += 1;
		}
	    }
	}
	System.out.println("Free size: " + (freeSize*dbAllocationQuantum));
	System.out.println("Used size: " + (usedSize*dbAllocationQuantum));
	System.out.println("Number of holes: " + nHoles);
	if (nHoles > 0) { 
	    System.out.println("Min. hole size: " + (minHoleSize*dbAllocationQuantum));
	    System.out.println("Max. hole size: " + (maxHoleSize*dbAllocationQuantum));
	    if (maxHoleSize >= histogram.length) { 
		maxHoleSize = histogram.length-1;
	    }
	    for (i = 1; i <= maxHoleSize; i++) { 
		if (histogram[i] != 0) { 
		    System.out.println("Holes with size " + (i*dbAllocationQuantum) 
				       + " bytes: " + histogram[i]);
		}
	    }
	}
    }
}








