/* Consus - Pure Java Object Relational Database Management System
 * Implementation of the object persistent storage interface on top of the standard relational DBMS
 * Created by Konstantin Knizhnik, Apr 2001.
 */
package org.garret.consus;

import java.sql.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.Hashtable;
import java.util.Comparator;
import java.util.Arrays;
import org.garret.internal.consus.WeakHashTable;
import org.garret.internal.consus.ByteBuffer;

/**
 * This class implements the Consus OO API on top of standard relational database systems.
 * The tables used by applications should be explicitly created by the programmer using the standard JDBC
 * protocol. This is the main difference with implementation of PersistentObjectStorage by
 * ConsusConnection. In the last case, tables are automatically generated from class definitions.
 * Below is the initialization sequence for a database to be used with <code>JDBCRerlationalAdapter</code>
 * <OL>
 * <LI>Create <code>TypeInfo</code> table with <code>TID, TableName, ClassName</code> fields.
 * <LI>Insert in this table the information about all the tables/classes used by the application. This table
 * is assigned a unique id to the type and establishes mapping between the application class and database table.</LI>
 * <LI>Create <code>OidTable</code>. This table contains a single column <code>ID</code> with
 * a single row and is used to generate new object identifiers.</LI>
 * <LI>Insert initial value of identifier into the <code>OidTable</code> table.</LI>
 * <LI>Create tables for all classes used by the application. For reference fields you should
 * choose the database type which will be able to hold a Java <code>long</code> value.</OL>
 * All arrays should  be stored as a <code>VARBINARY</code> type. Non-persistent object fields
 * should be represented by their components concatenated with the
 * <code>JDBCObjectRelationalAdapter.STRUCTURE_FIELDS_SEP</code> character (by default <code>$</code>).
 * Mapping of other Java types to native RDBMS types is up to the programmer. The only requirement
 * is that JDBC driver will be able to extract/store fields of that object, using
 * <code>getXXX</code> and <code>putXXX</code> methods of <code>java.sql.ResultSet</code> and
 * <code>java.sql.PreparedStatament</code>, where <code>XXX</code> is the Java type of the field.
 * Each table should have <code>RID</code> and <code>TID</code> fields which are used
 * to store record and type identifiers. Both identifiers are of the Java <code>int</code> type.
 * The Object identifier (OID) is constructed by concatenation of type and record identifiers.
 * </LI><P>
 * To check whether the database was initialized or not, the programmer should use the
 * <code>PersistentObjectStorage.isInitialized()</code> method.
 * It should be called <B>before</B> opening and so the database should be initialized prior
 * to invocation of the <code>PersistetnObjectStorage.open()</code> method.<P>
 * 
 * These rules are illustrated by the following example:<P>
 * Application classes:
 * <PRE>
 * class Address { 
 *     String  country;
 *     String  city;
 *     String  street;
 * };
 * 
 * class Person extends Persistent { 
 *     String   name;
 *     Address  address;
 *     Person   mother;
 *     Person   father;
 *     Person[] childs; 
 * };
 * 
 * Connection con = DriverManager.getConnection(databaseURL);	
 * PersistentObjectStorage storage = (con instanceof PersistentObjectStorage) 
 *     ? (PersistentObjectStorage)con : new JDBCObjectRelationalAdapter(con);
 * if (!storage.isInitialized()) {
 *     Statement stmt = con.createStatement();
 *     stmt.executeUpdate("create table TypeInfo (TID int, TableName varchar, ClassName varchar)");
 *     stmt.executeUpdate("create table OidTable (ID int)");
 *     stmt.executeUpdate("insert into OidTable values (1)");
 *     stmt.executeUpdate("create table Person (name varchar primary key," + 
 *                                             "address$country varchar," + 
 *                                             "address$city varchar," + 
 *                                             "address$street varchar," + 
 *                                             "mother bigint," + 
 *                                             "father bigint," + 
 *                                             "childs varbinary)");
 *     stmt.executeUpdate("insert into TypeInfo values (1, 'Person', 'Person')");
 * }
 * storage.open();
 * </PRE>
 * Underlying RDBMS should have
 * <OL>
 * <LI>TypeInfo table with TID, TableName and ClassName columns, 
 * which specifies the type mapping between Java classes and RDBMS table.
 * <LI>OidTable with a single row in the single OID column,
 * which is used for assigning OIDs to the inserted object.
 * <LI>OID and TID columns in each table, which store object and table identifiers
 * </OL>
 */
public class JDBCObjectRelationalAdapter implements PersistentObjectStorage
{
    //
    // Underlying RDBMS should have
    // <OL>
    // <LI>TypeInfo table with TID, TableName and ClassName columns, 
    // which specifies the type mapping between Java classes and RDBMS table.
    // <LI>OidTable with a single row in the single OID column,
    // which is used for assigning OIDs to the inserted object.
    // <LI>OID and TID columns in each table, which store object and table identifiers
    // </OL>
    //
    public static String TYPE_TABLE_NAME        = "TypeInfo";
    public static String TYPE_ID_FIELD          = "TID";
    public static String TYPE_TABLE_NAME_FIELD  = "TableName";
    public static String TYPE_CLASS_NAME_FIELD  = "ClassName";

    public static String OID_TABLE_NAME         = "OidTable";
    public static String OID_TABLE_COLUMN_NAME  = "ID";

    public static String OBJECT_RID_NAME        = "RID";
    public static String OBJECT_TID_NAME        = "TID";
    public static String STRUCTURE_FIELDS_SEP   = "$";
    
    static class ColumnInfo { 
	static final int TYPE_BYTE         = 1;
	static final int TYPE_SHORT        = 2;
	static final int TYPE_CHAR         = 3;
	static final int TYPE_INT          = 4;
	static final int TYPE_LONG         = 5;
	static final int TYPE_BOOLEAN      = 6;
	static final int TYPE_FLOAT        = 7;
	static final int TYPE_DOUBLE       = 8;
	static final int TYPE_STRING       = 9;
	static final int TYPE_DATE         = 10;
	static final int TYPE_TIME         = 11;
	static final int TYPE_TIMESTAMP    = 12;
	static final int TYPE_CLOB         = 13;
	static final int TYPE_BLOB         = 14;
	static final int TYPE_ARRAY        = 15;
	static final int TYPE_READER       = 16;
	static final int TYPE_INPUT_STREAM = 17;
	static final int TYPE_STRUCTURE    = 18;
	static final int TYPE_REFERENCE    = 19;

	ColumnInfo nextComponent;
	ColumnInfo nextColumn;
	ColumnInfo components;

	int         type;
	String      name;
	Class       cls;
	Field       f;
	Constructor constructor;

	static ColumnInfo buildArrayColumnsList(Class type) throws Exception 
	{
	    ColumnInfo cinfo = new ColumnInfo();
	    cinfo.cls = type;
	    int javaType;
	    if (type == Byte.TYPE) { 
		javaType = TYPE_BYTE;
	    } else if (type == Short.TYPE) { 
		javaType = TYPE_SHORT;
	    } else if (type == Integer.TYPE) { 
		javaType = TYPE_INT;
	    } else if (type == Boolean.TYPE) { 
		javaType = TYPE_BOOLEAN;
	    } else if (type == Long.TYPE) { 
		javaType = TYPE_LONG;
	    } else if (type == Float.TYPE) { 
		javaType = TYPE_FLOAT;
	    } else if (type == Double.TYPE) { 
		javaType = TYPE_DOUBLE;
	    } else if (type == stringClass) { 
		javaType = TYPE_STRING;
	    } else if (persistentClass.isAssignableFrom(type)) { 
		javaType = TYPE_REFERENCE;
	    } else if (blobClass.isAssignableFrom(type)) { 
		javaType = TYPE_BLOB;
	    } else if (clobClass.isAssignableFrom(type)) { 
		javaType = TYPE_CLOB;
	    } else if (dateClass.isAssignableFrom(type)) { 
		javaType = TYPE_DATE;
	    } else if (timeClass.isAssignableFrom(type)) { 
		javaType = TYPE_TIME;
	    } else if (timestampClass.isAssignableFrom(type)) { 
		javaType = TYPE_TIMESTAMP;
	    } else if (inputStreamClass.isAssignableFrom(type)) { 
		javaType = TYPE_INPUT_STREAM;
	    } else if (readerClass.isAssignableFrom(type)) { 
		javaType = TYPE_READER;
	    } else if (type.isArray()) { 
		javaType = TYPE_ARRAY;
		cinfo.components = buildArrayColumnsList(type.getComponentType());
	    } else { 
		javaType = TYPE_STRUCTURE;
		buildColumnsList("", type, null, cinfo);
		cinfo.constructor = type.getDeclaredConstructor(new Class[0]); 
		if (setBypass != null) { 
		    setBypass.invoke(cinfo.constructor, bypassFlag);
		} 
	    }
	    cinfo.type = javaType;
	    return cinfo;
	}

	static ColumnInfo buildColumnsList(String prefix, Class cls, ColumnInfo columns,
					   ColumnInfo structure) throws Exception
	{ 
	    ColumnInfo components = null; 
	    do { 
		Field fields[] = cls.getDeclaredFields();
        Arrays.sort(fields, new Comparator<Field>() { public int compare(Field f1, Field f2) { return f1.getName().compareTo(f2.getName()); } });
		for (int i = fields.length; --i >= 0;) { 
		    if ((fields[i].getModifiers() 
			 & (Modifier.TRANSIENT|Modifier.STATIC)) == 0)
		    {
			ColumnInfo cinfo = new ColumnInfo();
			cinfo.f = fields[i];
			cinfo.nextComponent = components;
			components = cinfo;
			cinfo.name = prefix + fields[i].getName();
			Class type = cinfo.f.getType();
			cinfo.cls = type;
			int javaType;
			if (setBypass != null) { 
			    setBypass.invoke(fields[i], bypassFlag);
			} 
			if (type == Byte.TYPE) { 
			    javaType = TYPE_BYTE;
			} else if (type == Short.TYPE) { 
			    javaType = TYPE_SHORT;
			} else if (type == Integer.TYPE) { 
			    javaType = TYPE_INT;
			} else if (type == Boolean.TYPE) { 
			    javaType = TYPE_BOOLEAN;
			} else if (type == Long.TYPE) { 
			    javaType = TYPE_LONG;
			} else if (type == Float.TYPE) { 
			    javaType = TYPE_FLOAT;
			} else if (type == Double.TYPE) { 
			    javaType = TYPE_DOUBLE;
			} else if (type == stringClass) { 
			    javaType = TYPE_STRING;
			} else if (persistentClass.isAssignableFrom(type)) { 
			    javaType = TYPE_REFERENCE;
			} else if (blobClass.isAssignableFrom(type)) { 
			    javaType = TYPE_BLOB;
			} else if (clobClass.isAssignableFrom(type)) { 
			    javaType = TYPE_CLOB;
			} else if (dateClass.isAssignableFrom(type)) { 
			    javaType = TYPE_DATE;
			} else if (timeClass.isAssignableFrom(type)) { 
			    javaType = TYPE_TIME;
			} else if (timestampClass.isAssignableFrom(type)) { 
			    javaType = TYPE_TIMESTAMP;
			} else if (inputStreamClass.isAssignableFrom(type)) { 
			    javaType = TYPE_INPUT_STREAM;
			} else if (readerClass.isAssignableFrom(type)) { 
			    javaType = TYPE_READER;
			} else if (type.isArray()) { 
			    javaType = TYPE_ARRAY;
			    cinfo.components = buildArrayColumnsList(type.getComponentType());
			} else { 
			    javaType = TYPE_STRUCTURE;
			    columns = buildColumnsList(prefix + cinfo.name + STRUCTURE_FIELDS_SEP, 
						       type, columns, cinfo);
			    cinfo.constructor = type.getDeclaredConstructor(new Class[0]); 
			    if (setBypass != null) { 
				setBypass.invoke(cinfo.constructor, bypassFlag);
			    } 
			}
			cinfo.nextColumn = columns;
			columns = cinfo;
			cinfo.type = javaType;
		    }
		}
		cls = cls.getSuperclass();
	    } while (cls != null);

	    if (structure != null) { 
		structure.components = components;
	    }
	    return columns;
	}

	static void packStructure(ByteBuffer buf, ColumnInfo component, Object obj, 
				  JDBCObjectRelationalAdapter db)
	    throws Exception
	{
	    while (component != null) { 
		Field f = component.f;
		switch (component.type) { 
		  case TYPE_BYTE:
		    buf.putByte(f.getByte(obj));
		    break;
		  case TYPE_CHAR:
		    buf.putShort((short)f.getChar(obj));
		    break;
		  case TYPE_SHORT:
		    buf.putShort(f.getShort(obj));
		    break;
		  case TYPE_INT:
		    buf.putInt(f.getInt(obj));
		    break;
		  case TYPE_LONG:
		    buf.putLong(f.getLong(obj));
		    break;
		  case TYPE_BOOLEAN:
		    buf.putByte(f.getBoolean(obj) ? 1 : 0);
		    break;
		  case TYPE_FLOAT:
		    buf.putFloat(f.getFloat(obj));
		    break;
		  case TYPE_DOUBLE:
		    buf.putDouble(f.getDouble(obj));
		    break;
		  case TYPE_STRING:
		    buf.putString((String)f.get(obj));
		    break;
		  case TYPE_DATE:
		  case TYPE_TIME:
		  case TYPE_TIMESTAMP:
		  {
		      java.util.Date date = (java.util.Date)f.get(obj);
		      buf.putLong(date != null ? date.getTime() : 0);
		      break;
		  }
		  case TYPE_CLOB:
		  case TYPE_BLOB:
		  case TYPE_READER:
		  case TYPE_INPUT_STREAM:
		    // not supported
		    break;
		  case TYPE_ARRAY:
		    packArray(buf, component.components, f.get(obj), db);
		    break;
		  case TYPE_STRUCTURE:
		    packStructure(buf, component.components, f.get(obj), db);
		    break;
		  case TYPE_REFERENCE:
		    buf.putLong(db.store((Persistent)f.get(obj)));
		}		
		component = component.nextComponent;
	    } 
	}

	static void packArray(ByteBuffer buf, ColumnInfo element, Object array,
			      JDBCObjectRelationalAdapter db)
	    throws Exception
	{
	    byte[] dst;
	    if (array == null) { 
		buf.putInt(-1);
		return;
	    } 
	    int i, j, n = java.lang.reflect.Array.getLength(array);
	    buf.putInt(n);

	    switch (element.type) { 
	      case TYPE_BYTE:
		buf.putBytes((byte[])array);
		break;
	      case TYPE_CHAR:
		{
		    char src[] = (char[])array;
		    for (i = 0; i < n; i++) { 
			buf.putShort((short)src[i]);
		    }
		    break;
		}
	      case TYPE_SHORT:
		{
		    short src[] = (short[])array;
		    for (i = 0; i < n; i++) { 
			buf.putShort(src[i]);
		    }
		    break;
		}
	      case TYPE_INT:
		{
		    int src[] = (int[])array;
		    for (i = 0; i < n; i++) { 
			buf.putInt(src[i]);
		    }
		    break;
		}
	      case TYPE_LONG:
		{
		    long src[] = (long[])array;
		    for (i = 0; i < n; i++) { 
			buf.putLong(src[i]);
		    }
		    break;
		}
	      case TYPE_BOOLEAN:
		{
		    boolean src[] = (boolean[])array;
		    for (i = 0; i < n; i++) { 
			buf.putByte(src[i] ? 1 : 0);
		    }
		    break;
		}	      
	      case TYPE_FLOAT:
		{
		    float src[] = (float[])array;
		    for (i = 0; i < n; i++) { 
			buf.putFloat(src[i]);
		    }
		    break;
		}
	      case TYPE_DOUBLE:
		{
		    double src[] = (double[])array;
		    for (i = 0; i < n; i++) { 
			buf.putDouble(src[i]);
		    }
		    break;
		}
	      case TYPE_STRING:
	        {
		    String src[] = (String[])array;
		    for (i = 0; i < n; i++) { 
			buf.putString(src[i]);
		    }
		    return;
		}
	      case TYPE_DATE:
	      case TYPE_TIME:
	      case TYPE_TIMESTAMP:
	        {
		    java.util.Date src[] = (java.util.Date[])array;
		    for (i = 0; i < n; i++) { 
			buf.putLong(src[i].getTime());
		    }
		    break;
		}		      
	      case TYPE_CLOB:
	      case TYPE_BLOB:
	      case TYPE_READER:
	      case TYPE_INPUT_STREAM:
		// not supported
		break;
	      case TYPE_ARRAY:
		{
		    Object src[] = (Object[])array;
		    for (i = 0; i < n; i++) {
			packArray(buf, element.components, src[i], db);
		    }
		    break;
		}
	      case TYPE_STRUCTURE:
		{
		    Object src[] = (Object[])array;
		    for (i = 0; i < n; i++) {
			packStructure(buf, element.components, src[i], db);
		    }
		    break;
		}		
	      case TYPE_REFERENCE:
		{
		    Persistent src[] = (Persistent[])array;
		    for (i = 0; i < n; i++) { 
			buf.putLong(db.store(src[i]));
		    }
		}
	    }
	}

	static Object unpackStructure(ByteBuffer buf, ColumnInfo component, 
				      boolean loadClosure, JDBCObjectRelationalAdapter db)
	    throws Exception
	{
	    Object obj = component.constructor.newInstance(constructorParameters);

	    while (component != null) { 
		Field f = component.f;
		switch (component.type) { 
		  case TYPE_BYTE:
		    f.setByte(obj, buf.getByte());
		    break;
		  case TYPE_CHAR:
		    f.setChar(obj, (char)buf.getShort());
		    break;
		  case TYPE_SHORT:
		    f.setShort(obj, buf.getShort());
		    break;
		  case TYPE_INT:
		    f.setInt(obj, buf.getInt());
		    break;
		  case TYPE_LONG:
		    f.setLong(obj, buf.getLong());
		    break;
		  case TYPE_BOOLEAN:
		    f.setBoolean(obj, buf.getByte() != 0);
		    break;
		  case TYPE_FLOAT:
		    f.setFloat(obj, buf.getFloat());
		    break;
		  case TYPE_DOUBLE:
		    f.setDouble(obj, buf.getDouble());
		    break;
		  case TYPE_STRING:
		    f.set(obj, buf.getString());
		    break;
		  case TYPE_DATE:
		    f.set(obj, new java.sql.Date(buf.getLong()));
		    break;
		  case TYPE_TIME:
		    f.set(obj, new java.sql.Time(buf.getLong()));
		    break;
		  case TYPE_TIMESTAMP:
		    f.set(obj, new java.sql.Timestamp(buf.getLong()));
		    break;
		  case TYPE_CLOB:
		  case TYPE_BLOB:
		  case TYPE_READER:
		  case TYPE_INPUT_STREAM:
		    // not supported
		    break;
		  case TYPE_ARRAY:
		    f.set(obj, unpackArray(buf, component.components, loadClosure, db));
		    break;
		  case TYPE_STRUCTURE:
		    f.set(obj, unpackStructure(buf, component.components, loadClosure, db));
		    break;
		  case TYPE_REFERENCE:
		    f.set(obj, db.load(buf.getLong(), loadClosure));
		    break;
		}		
		component = component.nextComponent;
	    } 
	    return obj;
	}


	static Object unpackArray(ByteBuffer buf, ColumnInfo element,
				  boolean loadClosure, JDBCObjectRelationalAdapter db)
	    throws Exception
	{
	    int i;
	    int n = buf.getInt();
	    if (n < 0) { 
		return null;
	    }
	    Object array = java.lang.reflect.Array.newInstance(element.cls, n);

	    switch (element.type) { 
	      case TYPE_BYTE:
		buf.getBytes((byte[])array);
		break;
	      case TYPE_CHAR:
		{
		    char dst[] = (char[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = (char)buf.getShort();
		    }
		    break;
		}
	      case TYPE_SHORT:
		{
		    short dst[] = (short[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = buf.getShort();
		    }
		    break;
		}
	      case TYPE_INT:
		{
		    int dst[] = (int[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = buf.getInt();
		    }
		    break;
		}
	      case TYPE_LONG:
		{
		    long dst[] = (long[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = buf.getLong();
		    }
		    break;
		}
	      case TYPE_BOOLEAN:
		{
		    boolean dst[] = (boolean[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = buf.getByte() != 0;
		    }
		    break;
		}	      
	      case TYPE_FLOAT:
		{
		    float dst[] = (float[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = buf.getFloat();
		    }
		    break;
		}
	      case TYPE_DOUBLE:
		{
		    double dst[] = (double[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = buf.getDouble();
		    }
		    break;
		}
	      case TYPE_STRING:
	        {
		    String dst[] = (String[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = buf.getString();
		    }
		    break;
		}
	      case TYPE_DATE:
	        {
		    Object dst[] = (Object[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = new java.sql.Date(buf.getLong());
		    }
		    break;
		}		      
	      case TYPE_TIME:
	        {
		    Object dst[] = (Object[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = new java.sql.Time(buf.getLong());
		    }
		    break;
		}		      
	      case TYPE_TIMESTAMP:
	        {
		    Object dst[] = (Object[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = new java.sql.Timestamp(buf.getLong());
		    }
		    break;
		}		      
	      case TYPE_CLOB:
	      case TYPE_BLOB:
	      case TYPE_READER:
	      case TYPE_INPUT_STREAM:
		// not supported
		break;
	      case TYPE_ARRAY:
		{
		    Object dst[] = (Object[])array;
		    for (i = 0; i < n; i++) {
			dst[i] = unpackArray(buf, element.components, loadClosure, db);
		    }
		    break;
		}
	      case TYPE_STRUCTURE:
		{
		    Object dst[] = (Object[])array;
		    for (i = 0; i < n; i++) {
			dst[i] = unpackStructure(buf, element.components, loadClosure, db);
		    }
		    break;
		}		
	      case TYPE_REFERENCE:
		{
		    Persistent dst[] = (Persistent[])array;
		    for (i = 0; i < n; i++) { 
			dst[i] = db.load(buf.getLong(), loadClosure);
		    }
		}
	    }
	    return array;
	}


	static int specifyParameters(PreparedStatement stmt, ColumnInfo component, Object obj, 
				     int idx, JDBCObjectRelationalAdapter db)
	    throws Exception
	{
	    while (component != null) { 
		Field f = component.f;
		if (obj == null) { 
		    if (component.type == TYPE_STRUCTURE) { 
			idx = specifyParameters(stmt, component.components, null, idx, db) - 1;
		    }
		} else { 
		    switch (component.type) {
		      case TYPE_BYTE:
			stmt.setByte(idx, f.getByte(obj));
			break;
		      case TYPE_SHORT:
			stmt.setShort(idx, f.getShort(obj));
		    break;
		      case TYPE_CHAR:
			stmt.setInt(idx, f.getChar(obj));
			break;
		      case TYPE_INT:
			stmt.setInt(idx, f.getInt(obj));
			break;
		      case TYPE_BOOLEAN:
			stmt.setBoolean(idx, f.getBoolean(obj));
			break;
		      case TYPE_FLOAT:
			stmt.setFloat(idx, f.getFloat(obj));
			break;
		      case TYPE_DOUBLE:
			stmt.setDouble(idx, f.getDouble(obj));
			break;
		      case TYPE_STRING:
			stmt.setString(idx, (String)f.get(obj));
			break;
		      case TYPE_DATE:
			stmt.setDate(idx, (java.sql.Date)f.get(obj));
			break;
		      case TYPE_TIME:
			stmt.setTime(idx, (java.sql.Time)f.get(obj));
			break;
		      case TYPE_TIMESTAMP:
			stmt.setTimestamp(idx, (java.sql.Timestamp)f.get(obj));
			break;
		      case TYPE_CLOB:
			stmt.setClob(idx, (Clob)f.get(obj));
			break;
		      case TYPE_BLOB:
			stmt.setBlob(idx, (Blob)f.get(obj));
			break;		
		      case TYPE_ARRAY:
			if (component.components.type == TYPE_BYTE) { 
			    stmt.setBytes(idx, (byte[])f.get(obj));
			} else { 
			    ByteBuffer buf = new ByteBuffer();
			    packArray(buf, component.components, f.get(obj), db);
			    stmt.setBytes(idx, buf.getBytes());
			}
			break;
		      case TYPE_READER:
		      case TYPE_INPUT_STREAM:
			// not stored
			break;
		      case TYPE_STRUCTURE:
			idx = specifyParameters(stmt, component.components, f.get(obj), idx, db) - 1;
			break;
		      case TYPE_REFERENCE:
			stmt.setLong(idx, db.store((Persistent)f.get(obj)));
		    }
		}
		idx += 1;
		component = component.nextComponent;
	    } 
	    return idx;
	}
	
	static void load(Object obj, ColumnInfo component) throws Exception 
	{ 
	    Object o;
	    while (component != null) { 
		switch (component.type) { 
  		  case TYPE_ARRAY:
		    {
			ColumnInfo element = component.components;
			if (element.type == TYPE_REFERENCE) { 
			    Persistent arr[] = (Persistent[])component.f.get(obj);
			    if (arr != null) { 
				for (int i = 0; i < arr.length; i++) { 
				    Persistent.get(arr[i]);
				}
			    }
			} else if (element.type == TYPE_STRUCTURE) { 
			    Object arr[] = (Object[])component.f.get(obj);
			    if (arr != null) { 
				for (int i = 0; i < arr.length; i++) { 
				    if ((o = arr[i]) != null) { 
					load(o, element.components);
				    }
				}
			    }
			}
		    }
		    break;		    
		  case TYPE_STRUCTURE:
	            if ((o = component.f.get(obj)) != null) { 
			load(o, component.components);
		    }
		  case TYPE_REFERENCE:
		    Persistent.get((Persistent)component.f.get(obj));
		}
		component = component.nextComponent;
	    }
	}


	static void update(Object obj, ColumnInfo component, ResultSet cursor,
			   JDBCObjectRelationalAdapter db)
	    throws Exception
	{
	    while (component != null) { 
		Field f = component.f;
		String name = component.name;
		switch (component.type) {
		  case TYPE_BYTE:
		    cursor.updateByte(name, f.getByte(obj));
		    break;
		  case TYPE_SHORT:
		    cursor.updateShort(name, f.getShort(obj));
		    break;
		  case TYPE_CHAR:
		    cursor.updateInt(name, f.getChar(obj));
		    break;
		  case TYPE_INT:
		    cursor.updateInt(name, f.getInt(obj));
		    break;
		  case TYPE_BOOLEAN:
		    cursor.updateBoolean(name, f.getBoolean(obj));
		    break;
		  case TYPE_FLOAT:
		    cursor.updateFloat(name, f.getFloat(obj));
		    break;
		  case TYPE_DOUBLE:
		    cursor.updateDouble(name, f.getDouble(obj));
		    break;
		  case TYPE_STRING:
		    cursor.updateString(name, (String)f.get(obj));
		    break;
		  case TYPE_DATE:
		    cursor.updateDate(name, (java.sql.Date)f.get(obj));
		    break;
		  case TYPE_TIME:
		    cursor.updateTime(name, (java.sql.Time)f.get(obj));
		    break;
		  case TYPE_TIMESTAMP:
		    cursor.updateTimestamp(name, (java.sql.Timestamp)f.get(obj));
		    break;
		  case TYPE_ARRAY:
		    if (component.components.type == TYPE_BYTE) {
			cursor.updateBytes(name, (byte[])f.get(obj));
		    } else { 
			ByteBuffer buf = new ByteBuffer();
			packArray(buf, component.components, f.get(obj), db);
			cursor.updateBytes(name, buf.getBytes());
		    }
		    break;
		  case TYPE_CLOB:
		  case TYPE_BLOB:
		  case TYPE_READER:
		  case TYPE_INPUT_STREAM:
		    // not stored
		    break;
		  case TYPE_STRUCTURE:
		    update(f.get(obj), component.components, cursor, db);
		    break;
		  case TYPE_REFERENCE:
		    cursor.updateLong(name, db.store((Persistent)f.get(obj)));		    
		}
		component = component.nextComponent;
	    } 	    
	}


	static void fetch(Object obj, ColumnInfo component, ResultSet cursor,
			  boolean loadClosure, JDBCObjectRelationalAdapter db)
	    throws Exception
	{
	    while (component != null) { 
		Field f = component.f;
		String name = component.name;
		switch (component.type) {
		  case TYPE_BYTE:
		    f.setByte(obj, cursor.getByte(name));
		    break;
		  case TYPE_SHORT:
		    f.setShort(obj, cursor.getShort(name));
		    break;
		  case TYPE_CHAR:
		    f.setChar(obj, (char)cursor.getInt(name));
		    break;
		  case TYPE_INT:
		    f.setInt(obj, cursor.getInt(name));
		    break;
		  case TYPE_BOOLEAN:
		    f.setBoolean(obj, cursor.getBoolean(name));
		    break;
		  case TYPE_FLOAT:
		    f.setFloat(obj, cursor.getFloat(name));
		    break;
		  case TYPE_DOUBLE:
		    f.setDouble(obj, cursor.getDouble(name));
		    break;
		  case TYPE_STRING:
		    f.set(obj, cursor.getString(name));
		    break;
		  case TYPE_DATE:
		    f.set(obj, cursor.getDate(name));
		    break;
		  case TYPE_TIME:
		    f.set(obj, cursor.getTime(name));
		    break;
		  case TYPE_TIMESTAMP:
		    f.set(obj, cursor.getTimestamp(name));
		    break;
		  case TYPE_CLOB:
		    f.set(obj, cursor.getClob(name));
		    break;
		  case TYPE_BLOB:
		    f.set(obj, cursor.getBlob(name));
		    break;		
		  case TYPE_ARRAY:
		    if (component.components.type == TYPE_BYTE) { 
			f.set(obj, cursor.getBytes(name));
		    } else {
			f.set(obj, unpackArray(new ByteBuffer(cursor.getBytes(name)), 
					       component.components, loadClosure, db));
		    }
		    break;
		  case TYPE_READER:
		    f.set(obj, cursor.getCharacterStream(name));
		    break;
		  case TYPE_INPUT_STREAM:
		    f.set(obj, cursor.getBinaryStream(name));
		    break;
		  case TYPE_STRUCTURE:
		    fetch(component.constructor.newInstance(constructorParameters), 
			  component.components, cursor, loadClosure, db);
		    break;
		  case TYPE_REFERENCE:
		    f.set(obj, db.load(cursor.getLong(name), loadClosure));
		}
		component = component.nextComponent;
	    } 	    	    
	}
    }

    static class TypeInfo { 
	TypeInfo    next;
	String      tableName;
	Class       cls;	
	int         typeId;		
	ColumnInfo  columns;
	Constructor constructor;

	String            loadStmtSql;

	PreparedStatement updateStmt;
	PreparedStatement insertStmt;
	PreparedStatement deleteStmt;
	PreparedStatement getOidStmt;
	PreparedStatement incrementOidStmt;
    }
    

    /**
     * Constructor of the adapter class
     * @param con - opened JDBC connection
     */
    public JDBCObjectRelationalAdapter(Connection con) {
	this.con = con;
    }

    /**
     * Returns true if the database has already been initialized.
     * If database was not initilialized, the programmer has to create the
     * TypeInfo, OidTable and application tables, insert the information about class-table mapping
     * into the TypeInfo table and initialize the ID counter by inserting the value into OidTable.
     * Initialization can be done through the standard JDBC methods.
     */
    public boolean isInitialized() throws Exception
    {
	DatabaseMetaData meta = con.getMetaData();
	ResultSet cursor = meta.getTables(null, "", "TypeInfo", null);
	boolean initialized = cursor.next();
	cursor.close();
	return initialized;
    }
    
    /**
     * Opens the persistent object storage. This method should be called after checking
     * that storage is initialized. If it has not been initialized then it should be initialized.
     */
    public void open() throws Exception
    {
	TypeInfo tinfo;
	Statement stmt = con.createStatement();
	ResultSet cursor = stmt.executeQuery("select * from " + TYPE_TABLE_NAME);
	typeHash = new Hashtable();
	objectCache = new WeakHashTable(1013);
	int maxTid = 0;

	while (cursor.next()) { 
	    int i;
	    tinfo = new TypeInfo();
	    tinfo.tableName = cursor.getString(TYPE_TABLE_NAME_FIELD);
	    Class cls = Class.forName(cursor.getString(TYPE_CLASS_NAME_FIELD));	    
	    tinfo.cls = cls;
	    tinfo.typeId = cursor.getInt(TYPE_ID_FIELD);
	    if (tinfo.typeId > maxTid) { 
		maxTid = tinfo.typeId;
	    }
	    tinfo.next = types;
	    types = tinfo;
	    typeHash.put(cls, tinfo);
	    tinfo.constructor = cls.getDeclaredConstructor(new Class[0]);
	    if (setBypass != null) { 
		setBypass.invoke(tinfo.constructor, bypassFlag);
	    }
	    tinfo.columns = ColumnInfo.buildColumnsList("", cls, null, null);
	    
	    String loadStmtSql = "select";
	    String updateStmtSql = "update " + tinfo.tableName + " set";
	    String insertStmtSql = "insert into " + tinfo.tableName + "(";
	    String deleteStmtSql = "delete from " + tinfo.tableName + " where " + OBJECT_RID_NAME + "=?";
	    String values = "";
	    char sep = ' ';
	    int nColums = 0;

	    for (ColumnInfo col = tinfo.columns; col != null; col = col.nextColumn) { 
		loadStmtSql += sep;
		loadStmtSql += col.name;

		updateStmtSql += sep;
		updateStmtSql += col.name;
		updateStmtSql += "=?";

		insertStmtSql += sep;
		insertStmtSql += col.name;

		values += sep;
		values += "?";
		
		sep = ',';
	    }

	    loadStmtSql += " from " + tinfo.tableName + " where " + OBJECT_RID_NAME + "=?";
	    updateStmtSql += " where " + OBJECT_RID_NAME + "=?";
	    insertStmtSql += sep + OBJECT_RID_NAME + "," + OBJECT_TID_NAME 
		+ ") values (" + values + sep + "?,?)";

	    tinfo.loadStmtSql = loadStmtSql;

	    tinfo.updateStmt = con.prepareStatement(updateStmtSql);
	    tinfo.insertStmt = con.prepareStatement(insertStmtSql);
	    tinfo.deleteStmt = con.prepareStatement(deleteStmtSql);

	    tinfo.getOidStmt = con.prepareStatement("select " + OID_TABLE_COLUMN_NAME + " from " + OID_TABLE_NAME);
						      
	    tinfo.incrementOidStmt = con.prepareStatement("update " + OID_TABLE_NAME 
							  + " set " + OID_TABLE_COLUMN_NAME 
							  + "=" + OID_TABLE_COLUMN_NAME+"+1");
	}
	stmt.close();
	typeMap = new TypeInfo[maxTid+1];
	for (tinfo = types; tinfo != null; tinfo = tinfo.next) { 
	    typeMap[tinfo.typeId] = tinfo;
	}
    }

    /**
     * Fetch the current row of the cursor as a Java object
     */
    public Object fetchObject(ResultSet cursor) throws Exception { 
	int tid = cursor.getInt(OBJECT_TID_NAME);
	TypeInfo tinfo = typeMap[tid];	
	int rid = cursor.getInt(OBJECT_RID_NAME);
	synchronized(objectCache) { 
	    Object obj = objectCache.get(rid);
	    boolean loadClosure = false;
	    if (obj == null) { 
		obj = tinfo.constructor.newInstance(constructorParameters);
		if (obj instanceof Persistent) { 
		    Persistent po = (Persistent)obj;
		    po.oid = makeOID(tid, rid);
		    po.db = this;
		    po.state = Persistent.PO_LOADED;
		    po.type = tinfo;
		    loadClosure = po.loadClosure();
		}
		objectCache.put(rid, obj);		
	    } else if (!(obj instanceof Persistent)) { 
		return obj;
	    } else {
		Persistent po = (Persistent)obj;
		loadClosure = po.loadClosure();
		if (po.state != Persistent.PO_STUB) { 
		    return obj;
		} else { 
		    po.state = Persistent.PO_LOADED;
		}
	    }
	    ColumnInfo.fetch(obj, tinfo.columns, cursor, loadClosure, this);
	    return obj;
	}
    }

    /**
     * Fetch the current row of the cursor as a Java object
     */
    public void updateObject(Object obj, ResultSet cursor) throws Exception { 
	Class cls = obj.getClass();
	TypeInfo tinfo = (TypeInfo)typeHash.get(cls);
	if (tinfo == null) { 
	    throw new SQLException("No such table " + cls.getName() + " in the database");
	}
	ColumnInfo.update(obj, tinfo.columns, cursor, this);
	cursor.updateRow();
    }


    /**
     * Insert a new object into the database.
     * @return reference to the created object
     */
     public long insertObject(Object obj) throws Exception { 
	 Class cls = obj.getClass();
	 TypeInfo tinfo = (TypeInfo)typeHash.get(cls);
	 if (tinfo == null) { 
	     throw new SQLException("No such table " + cls.getName() + " in the database");
	 }
	 return insert(obj, tinfo);
    }
   

    /**
     * Load stub object or reload object from the database.
     */
    public void loadObject(Persistent obj) throws Exception { 
	TypeInfo tinfo = (TypeInfo)obj.type;
	PreparedStatement loadStmt = con.prepareStatement(tinfo.loadStmtSql);
	loadStmt.setInt(1, getRID(obj.oid));
	ResultSet cursor = loadStmt.executeQuery();
	if (cursor.next()) { 
	    ColumnInfo.fetch(obj, tinfo.columns, cursor, obj.loadClosure(), this);
	    loadStmt.close();
	} else { 
	    loadStmt.close();
	    throw new SQLException("Object with such record ID doesn't exists in table " + 
				   tinfo.tableName);
	}
    }

    /**
     * Loads the components of the object. This method should be called for the classes
     * which disable implicit load of object closure, that is the other objects referenced by the object being loaded. In this case referenced
     * components should be loaded explicitly by this method.
     */
    public void loadComponents(Persistent obj) throws Exception {  
	TypeInfo tinfo = (TypeInfo)obj.type;
	ColumnInfo.load(obj, tinfo.columns);
    }

    /** 
     * Clear all object fields (to make it possible for GC to collect 
     * unused persistent objects). If this method is not called and all persistent objects
     * in the database are referencing each other (for example linked in a L2-list),
     * then after some time all the data can be loaded into memory from the database.
     */
    public void clearComponents(Persistent obj) throws Exception { 
	TypeInfo tinfo = (TypeInfo)obj.type;
	for (ColumnInfo component = tinfo.columns; component != null; component = component.nextComponent)
	{
	    if (component.type >= ColumnInfo.TYPE_STRING) { // non-scalar type
		component.f.set(obj, null);
	    }
	}
    }


    /** 
     * Save object into the database. If the object refers to some some other persistent-capable objects
     * without assigned object identifiers, then these objects will also be recursively saved into
     * the database.
     */
    public void storeObject(Persistent obj) throws Exception { 
	if (obj.oid == 0) { 	
	    Class cls = obj.getClass();
	    TypeInfo tinfo = (TypeInfo)typeHash.get(cls);
	    if (tinfo == null) { 
		throw new SQLException("No such table " + cls.getName() + " in the database");
	    }
	    insert(obj, tinfo);
	} else { 
	    TypeInfo tinfo = (TypeInfo)obj.type;
	    int nParams = ColumnInfo.specifyParameters(tinfo.updateStmt, tinfo.columns, obj, 1, this);
	    tinfo.updateStmt.setInt(nParams, getRID(obj.oid));
	    tinfo.updateStmt.executeUpdate();
	} 
    }

    /**
     * Remove object from the database.
     */
    public void deleteObject(Persistent obj) throws SQLException { 
	TypeInfo tinfo = (TypeInfo)obj.type;
	int rid = getRID(obj.oid);
	objectCache.remove(rid);	
	tinfo.deleteStmt.setInt(1, rid);
	tinfo.deleteStmt.executeUpdate();
    }

    /**
     * Exchange references to the objects in the database. The result of execution A.becomeObject(B)
     * is that OIDs of A and B are exchanged, so all references to A in the database will
     * now refer to B, and vice versa. This operation doesn't affect references in main
     * memory. Object A should already have an assigned OID (being stored in the database).
     * If B is not yet stored in the database, it is explicitly assigned an OID by this method.
     * @param a object which OID will be exchanged with <code>b</code> object
     * @param b object which OID will be exchanged with <code>a</code> object
     */
    public void becomeObject(Persistent a, Persistent b) throws SQLException { 
	throw new ConsusNotImplemented("JDBCObjectRelationalAdapter.becomeObject");
    }

     /** 
     * Lock the database in exclusive mode. The lock is kept until the end of transaction.
     * This method should be called to prevent deadlocks caused by upgrading shared locks
     * to exclusive. For example, if you first issue a read-only statement and then,
     * within the same transaction issue an update statement, it can cause deadlock
     * if concurrent thread also tries to upgrade the shared lock to exclusive.
     */
    public void lock() { 
    } 
    

    public Object getObjectByOid(long oid) throws Exception {
	int rid = getRID(oid);
	Object obj = objectCache.get(rid);
	if (obj == null) { 
	    int tid = getTID(oid);
	    TypeInfo tinfo = typeMap[tid];
	    Statement stmt = con.createStatement();
	    ResultSet cursor = stmt.executeQuery("select * from " + tinfo.tableName 
						 + " where " + OBJECT_RID_NAME + "=" + rid);
	    if (cursor.next()) { 
		stmt.close();
		return fetchObject(cursor);
	    }
	    stmt.close();		
	} else if (obj instanceof Persistent) {
	    ((Persistent)obj).get();
	}
	return obj;
    }

    public void clearObjectCache() { 
	objectCache.clear();
    }

    public int getObjectCacheSize() { 
	return objectCache.size();
    }

    class JDBCTableIterator extends TableIterator 
    { 
	JDBCTableIterator(String tableName) throws SQLException { 
	    this.tableName = tableName;
	}

	public boolean next() throws SQLException { 
	    return getCursor().next();
	}

	public boolean previous() throws SQLException { 
	    return getCursor().previous();
	}
	    
	public void close()  throws SQLException { 
	    if (cursor != null) { 
		cursor.close();
		cursor = null;
	    }
	    if (stmt != null) { 
		stmt.close();
		stmt = null;
	    }
	    if (pstmt != null) { 
		pstmt.close();
		pstmt = null;
	    }
	}

	public boolean isBeforeFirst() throws SQLException { 
	    return getCursor().isBeforeFirst();
	}

	public boolean isAfterLast() throws SQLException { 
	    return getCursor().isAfterLast();
	}

	public boolean isFirst() throws SQLException { 
	    return getCursor().isFirst();
	}

	public boolean isLast() throws SQLException { 
	    return getCursor().isLast();
	}

	public void beforeFirst() throws SQLException { 
	    getCursor().beforeFirst();
	}

	public void afterLast() throws SQLException { 
	    getCursor().afterLast();
	}

	public boolean first() throws SQLException { 
	    return getCursor().first();
	}

	public boolean last() throws SQLException { 
	    return getCursor().last();
	}

	public int getRow() throws SQLException { 
	    return getRow();
	}


	public boolean absolute(int row) throws SQLException { 
	    return getCursor().absolute(row);
	}
    
	public boolean relative(int row) throws SQLException { 
	    return getCursor().relative(row);
	}

	public void deleteRow() throws SQLException { 
	    if (currObj instanceof Persistent) { 
		deleteObject(currObj);
	    } else { 
		cursor.deleteRow();
	    }
	}
	
	public void deleteAllSelectedRows() throws SQLException { 
	    stmt.executeUpdate("delete * from " + tableName);
	}

	public int getNumberOfRows() throws SQLException { 
	    if (stmt == null) { 
		stmt = con.createStatement();
	    }
	    ResultSet cursor = stmt.executeQuery("select count(*) from " + tableName);
	    int nRows = -1;
	    if (cursor.next()) { 
		nRows = cursor.getInt(1);
	    }
	    cursor.close();
	    return nRows;
	}
	    
	public Object selectByKey(String keyName, String keyValue) throws SQLException
	{
	    if (stmt == null) { 
		stmt = con.createStatement();
	    }
	    ResultSet cursor = stmt.executeQuery("select * from " + tableName
						+ " where " + keyName + "='"
						+ keyValue + "'");
	    Object obj = null;
	    try { 
		if (cursor.next()) { 
		    obj = fetchObject(cursor);
		}
	    } catch (Exception x) { 
		x.printStackTrace();
		throw new SQLException("Exception while fetching the object: " + x);
	    }
	    cursor.close();
	    return obj;
	}

	public Object selectByOid(long oid) throws SQLException
	{
	    if (pstmt == null) { 
		pstmt = con.prepareStatement("select * from " + tableName
					     + " where " + OBJECT_RID_NAME + "=?");
	    }
	    pstmt.setInt(1, getRID(oid));
	    ResultSet cursor = pstmt.executeQuery();
	    Object obj = null;
	    try { 
		if (cursor.next()) { 
		    obj = fetchObject(cursor);
		}
	    } catch (Exception x) { 
		x.printStackTrace();
		throw new SQLException("Exception while fetching the object: " + x);
	    }
	    cursor.close();
	    return obj;
	}

	public Object get() throws SQLException { 
	    try { 
		return fetchObject(cursor);
	    } catch (Exception x) { 
		x.printStackTrace();
		throw new SQLException("Exception while fetching the object: " + x);
	    }
	}

	ResultSet getCursor() throws SQLException { 
	    if (stmt == null) { 
		stmt = con.createStatement();
	    }
	    if (cursor == null) {		
		cursor = stmt.executeQuery("select * from " + tableName);
	    }
	    return cursor;
	}

	String            tableName;
	ResultSet         cursor;
	Statement         stmt;	
	PreparedStatement pstmt;
	Persistent        currObj;
    }

    /**
     * Get the iterator for the table. This is a more efficient way for iteration through the
     * the table then using "select * from" and ResultSet.
     */
    public TableIterator getTableIterator(String tableName) throws SQLException {
	return new JDBCTableIterator(tableName);
    }


    /**
     * Internal methods
     */

    static final int getRID(long oid) 
    { 
	return (int)(oid >>> 10);
    }
    
    static final int getTID(long oid) 
    { 
	return (int)oid & 0x3FF;
    }

    static final long makeOID(int tid, int rid) 
    { 
	return (((long)rid & 0xFFFFFFFF) << 10) | tid;
    }

    protected long insert(Object obj, TypeInfo tinfo) throws Exception 
    { 
	tinfo.incrementOidStmt.executeUpdate();
	ResultSet cursor = tinfo.getOidStmt.executeQuery();
	if (!cursor.next()) {
	    throw new SQLException("OidTable is empty");
	}
	int rid = cursor.getInt(1);
	cursor.close();
	long oid = makeOID(tinfo.typeId, rid);
	if (obj instanceof Persistent) { 
	    Persistent po = (Persistent)obj;
	    po.type = tinfo;
	    po.state = Persistent.PO_LOADED;
	    po.db = this;
	    po.oid = oid;
	}
	objectCache.put(rid, obj);
	int nParams = ColumnInfo.specifyParameters(tinfo.insertStmt, tinfo.columns, obj, 1, this);
	tinfo.insertStmt.setInt(nParams, rid);
	tinfo.insertStmt.setInt(nParams+1, tinfo.typeId);
	tinfo.insertStmt.executeUpdate();
	return oid;
    }

    protected Persistent load(long oid, boolean loadClosure) throws Exception 
    {
	if (oid == 0) { 
	    return null;
	}
	int rid = getRID(oid);
	synchronized(objectCache) { 
	    Persistent po = (Persistent)objectCache.get(rid);
	    if (po == null) { 
		int tid = getTID(oid);
		TypeInfo tinfo = typeMap[tid];
		po = (Persistent)tinfo.constructor.newInstance(constructorParameters);
		po.oid  = oid;
		po.type = tinfo;
		po.db   = this;
		objectCache.put(rid, po);
		if (loadClosure) { 
		    po.state = Persistent.PO_LOADED;
		    loadObject(po);
		} else { 
		    po.state = Persistent.PO_STUB;
		}
	    } else { 
		po.get();
	    }
	    return po;
	}
    }

    protected long store(Persistent po) throws Exception 
    { 
	if (po == null) { 
	    return 0;
	}
	if (po.oid == 0) { 
	    po.store(this);
	}
	return po.oid;
    } 


    /** 
     * Variables
     */
	
    static Class blobClass;
    static Class clobClass;
    static Class timeClass;
    static Class timestampClass;
    static Class dateClass;
    static Class stringClass;
    static Class persistentClass;
    static Class inputStreamClass;
    static Class readerClass;
    
    static Method         setBypass;
    static final Object[] bypassFlag = { new Boolean(true) };
    static final Object[] constructorParameters = {};

    static { 
        try { 
	    Class c = Class.forName("java.lang.reflect.AccessibleObject");
	    Class[] param = { Boolean.TYPE };
	    setBypass = c.getMethod("setAccessible", param);
        } catch(Exception ex) {}

	try { 
	    blobClass = Class.forName("java.sql.Blob");
	    clobClass = Class.forName("java.sql.Clob");
	    timeClass = Class.forName("java.sql.Time");
	    timestampClass = Class.forName("java.sql.Timestamp");
	    dateClass = Class.forName("java.sql.Date");
	    stringClass = Class.forName("java.lang.String");
	    persistentClass = Class.forName("org.garret.consus.Persistent");
	    inputStreamClass = Class.forName("java.io.InputStream");
	    readerClass = Class.forName("java.io.Reader");
	} catch(Exception ex) {}
    }

    Connection    con;
    Hashtable     typeHash;
    TypeInfo[]    typeMap;
    WeakHashTable objectCache;
    TypeInfo      types;

    static final Object[] constructorArgs = {};
}













