/* Consus - Pure Java Object Relational Database Management System
 * Interactive JSQL
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.consus;

import java.sql.*;
import java.io.*;

public class JSQL { 
    static BufferedReader in;

    static int lineNo;
    static boolean interactiveMode;

    static String input(String prompt) throws IOException {
	while (true) { 
	    if (interactiveMode) { 
		System.out.print(prompt);
	    }
	    String answer = in.readLine();
	    lineNo += 1;
	    if (answer == null) { 
		return null;
	    }
	    return answer;
	}
    }

    public static void main(String args[]) throws Exception {
	if (args.length < 1) { 
	    System.err.println("JSQL: interactive JSQL interpreter");
	    System.err.println("Usage: java JSQL (database-file-name | user:password@host:port) {sql-files...}");
	    return;
	} 
	String URL, driver;
	int p;
	java.util.Properties info = new java.util.Properties();
	URL = args[0];
	if ((p = URL.indexOf('@')) >= 0) { 
	    if (p > 0) { 
		int q = URL.lastIndexOf(':', p);
		if (q >= 0) { 
		    info.put("user", URL.substring(0, q));
		    info.put("password", URL.substring(q+1, p));
		} else { 
		    info.put("user", URL.substring(0, p));
		}
	    }
	    driver = System.getProperty("jsql.driver", "org.garret.consus.ClientDriver");
	    URL = "jdbc:" + System.getProperty("jsql.server", "consus-client") + ":" + URL.substring(p+1);	
	} else { 
	    driver = System.getProperty("jsql.driver", "org.garret.consus.ConsusDriver");	
	    URL = "jdbc:" + System.getProperty("jsql.server", "consus") + ":" + URL;	
	}
	Class.forName(driver);
	Connection connection = DriverManager.getConnection(URL, info);
	if (connection == null) { 
	    System.err.println("No connection with database " + args[0]);
	    return;
	}
	//((org.garret.internal.consus.ConsusConnection)connection).dbTraceEnabled = true;
	for (int i = 1; i < args.length; i++) { 
	    in = new BufferedReader(new FileReader(args[i])); 
	    if (!loop(connection)) { 
		connection.close();
		return;
	    }
	}
	interactiveMode = true;
	in = new BufferedReader(new InputStreamReader(System.in)); 
	loop(connection);
	connection.close();
    }
    
    final static int TAB = 8;

    static boolean loop(Connection connection) throws Exception { 
	Statement stmt = connection.createStatement();
	int i;
	lineNo = 0;
	while (true) { 
	    try { 
		if (interactiveMode) { 
		    lineNo = 0;
		}
		String sql = input("JSQL> ");
		if (sql == null) { 
		    return true;
		} 
		String cmd = sql.trim();
		if (cmd.length() == 0) { 
		    continue;
		}
		if (cmd.equalsIgnoreCase("exit")) { 
		    return false;
		}
		if (cmd.equalsIgnoreCase("describe")) { 
		    DatabaseMetaData dbMeta = connection.getMetaData();
		    ResultSet tableCursor = dbMeta.getTables(null, "%", "%", null);
		    String fill = "                   ";
		    int maxColumnNameLen = fill.length();
		    while (tableCursor.next()) { 
			String tableName = tableCursor.getString("TABLE_NAME");
			System.out.println("Table \"" + tableName + "\"");
			ResultSet columnCursor = dbMeta.getColumns(null, "%", tableName, "%");
			for (i = 1; columnCursor.next(); i++) { 
			    String columnName = columnCursor.getString("COLUMN_NAME");
			    if (columnCursor.getString("REMARKS").equals("<indexed>")) { 
				columnName += "(*)";
			    }
			    int columnNameLen = columnName.length();
			    if (columnNameLen >= maxColumnNameLen) { 
				columnNameLen = maxColumnNameLen-1;
			    }
			    System.out.println("  " + i + "\t" + columnName
					       + fill.substring(columnNameLen) 
					       + columnCursor.getString("TYPE_NAME"));
			}
			System.out.println("--------------------------------------");
			columnCursor.close();
		    }
		    tableCursor.close();
		    continue;
		}
		while (!sql.trim().endsWith(";")) { 
		    String more = input(">> ");
		    sql += "\n" + more;
		}
		sql = sql.substring(0, sql.indexOf(';'));
		if (stmt.execute(sql)) { 
		    ResultSet cursor = stmt.getResultSet();
		    ResultSetMetaData metaData = cursor.getMetaData();
		    int nColumns = metaData.getColumnCount();
		    //System.out.print("OID ");
		    String sep = "";
		    int columnWidth[] = new int[nColumns];
		    for (i = 1; i <= nColumns; i++) { 
			if (metaData.getColumnType(i) != Types.ARRAY) { 
			    String columnName = metaData.getColumnLabel(i);
			    if (columnName == null) { 
				columnName = "?";
			    }
			    columnWidth[i-1] = (columnName.length()+TAB) & ~(TAB-1);
			    System.out.print(sep + columnName);
			    sep = "\t";
			}
		    }		    
		    System.out.println("\n------------------------------------------------------------------------------");
		    int nRows = 0;
		    while (cursor.next()) { 
			//System.out.print(cursor.getRef(0));
			//System.out.print("\t");
			sep = "";
			for (i = 1; i <= nColumns; i++) { 
			    String val = ""; 
			    switch (metaData.getColumnType(i)) { 
			      case Types.BLOB:
				val = "<BLOB " + cursor.getBlob(i).length() + " bytes>";
				break;
			      case Types.CLOB:
				val = "<CLOB " + cursor.getClob(i).length() + " chars>";
				break;
			      case Types.ARRAY:
				continue;
			      default:
				try { 
				    val = cursor.getString(i);
				    if (val.length() == 0) { 
					val = "''";
				    }
				} catch (Throwable x) { 
				    val = "???";
				}
			    }
			    System.out.print(sep + val);
			    sep = "\t";
			    int len = (val.length()+TAB) & ~(TAB-1);
			    while (len < columnWidth[i-1]) { 
				len += TAB;
				sep += '\t';
			    }
			}		    
			nRows += 1;
			System.out.println("");
		    }
		    System.out.println("------------------------------------------------------------------------------");
		    System.out.println("Select " + nRows + " records");
		} else { 
		    if (interactiveMode) { 
			System.out.println("Update " + stmt.getUpdateCount() + " rows");
		    }
		}
	    } catch (Exception x) { 
		x.printStackTrace();
		System.err.println("EXCEPTION at line " + lineNo + ": " + x.getMessage());
	    }
	}
    }
}





