/* Consus - Pure Java Object Relational Database Management System
 * Subclass of POList that guarantees the ordering of the elements, based on
 * the compareTo method of the Comparable interface.
 * Created by Konstantin Knizhnik, Apr 2000.
 */

package org.garret.consus;

import java.util.*;

/** 
 * Objects inserted into this list should implement the Comparable interface and be derived from
 * the Persistent class. Alternatively you can specify the Comparator in the constructor.
 * Nulls cannot be placed in the sorted list.
 */
public class POSortedList extends POList implements Comparable
{
    /**
     * Constructs an empty list with the specified initial capacity.
     *
     * @param   initialCapacity   the initial capacity of the list.
     * @param   c the comparator to be associated with the list
     */
    public POSortedList(int initialCapacity, POComparator c) {
        super(initialCapacity);
	cmp = c;
    }

    /**
     * Constructs an empty list.
     */
    public POSortedList() {
    }

    /**
     * Constructs an empty list with the specified initial capacity.
     *
     * @param   initialCapacity   the initial capacity of the list.
     */
    public POSortedList(int initialCapacity) {
        super(initialCapacity);
    }

    /**
     * Constructs an empty list and associates a comparator with the list.
     */
    public POSortedList(POComparator c) {
	cmp = c;
    }

     /**
     * Constructs a list containing the elements of the specified
     * collection, in the order they are returned by the collections
     * iterator.  The <tt>POList</tt> instance has an initial capacity of
     * 110% of the size of the specified collection.
     */
    public POSortedList(Collection c) {
	super((c.size()*110)/100);	// Allow 10% room for growth

	deferredUpdate = 1;
	Iterator i = c.iterator();
	while (i.hasNext()) { 
	    add(i.next());
	}
	deferredUpdate = 0;
    }

    /**
     * Compare two ordered lists. Comparision is made by sequential comparison of elements
     * from each sorted list. 
     * If there is no index position at which they differ, then the shorter 
     * list lexicographically precedes the longer list. In this case,
     * <code>compareTo</code> returns the difference of the lengths of the 
     * lists -- that is, the value: 
     * <blockquote><pre>
     * this.length()-anotherList.length()
     * </pre></blockquote>
     *
     * @param   anotherList   the <code>POSortedList</code> to be compared.
     * @return  the value <code>0</code> if all elements in the first list are equal to
     *          the elements in the corresponding position in the second list and that the lists have the
     *          same size; a value less than <code>0</code> is returned if the first element is not equal to
     *          the corresponding element in the second list, is smaller than the element from the
     *          second list or there is not such index position in the second list; a value 
     *          greater than zero is returned otherwise.
     * @exception java.lang.NullPointerException if <code>anotherList</code> 
     *          is <code>null</code>.
     */
    public int compareTo(POSortedList anotherList) {
	int n = size < anotherList.size ? size : anotherList.size;
	if (cmp != null) { 
	    for (int i = 0; i < n; i++) { 
		int diff = cmp.compare(elementData[i].get(), anotherList.elementData[i].get());
		if (diff != 0) { 
		    return diff;
		}
	    }
	} else { 
	    for (int i = 0; i < n; i++) { 
		int diff = ((Comparable)elementData[i].get()).compareTo(anotherList.elementData[i].get());
		if (diff != 0) { 
		    return diff;
		}
	    }
	}
	return size - anotherList.size;
    }

    /** 
     * Compare two sorted lists
     */
    public int compareTo(Object o) {
	return compareTo((POSortedList)o);
    }

    /**
     * Find the position where the element should be inserted into the list
     * using the <tt>compare</tt> method of Comparator or the <tt>compareTo</tt> method of the elements.
     *
     * @param   elem   an object.
     * @return  position after the last smaller element
     * @see     Object#equals(Object)
     */
    public int findInsertPosition(Object elem) {
	int l = 0;
	int r = size;
	if (cmp == null) { 
	    Comparable c = (Comparable)elem; 
	    while (l < r) { 
		int i = (l + r) >>> 1;
		int diff = c.compareTo(elementData[i].get());
		if (diff > 0) { 
		    l = i+1;
		} else { 
		    r = i;
		}
	    }
	} else { 
	    while (l < r) { 
		int i = (l + r) >>> 1;
		int diff = cmp.compare(elem, elementData[i].get());
		if (diff > 0) { 
		    l = i+1;
		} else { 
		    r = i;
		}
	    }
	}
	return r;
    }
    /**
     * Searches for the first occurence of the given argument, testing 
     * for equality using the <tt>equals</tt> method. 
     *
     * @param   elem   an object.
     * @return  the index of the first occurrence of the argument in this
     *          list; returns <tt>-1</tt> if the object is not found.
     * @see     Object#equals(Object)
     */
    public int indexOf(Object elem) {
	int i = findInsertPosition(elem);
	return i < size 
	    && ((cmp != null && cmp.compare(elem, elementData[i].get()) == 0)
		|| (cmp == null && ((Comparable)elem).compareTo(elementData[i].get()) == 0))
	    ? i : -1;
    } 

    /**
     * Returns the index of the last occurrence of the specified object in
     * this list.
     *
     * @param   elem   the desired element.
     * @return  the index of the last occurrence of the specified object in
     *          this list; returns -1 if the object is not found.
     */
    public int lastIndexOf(Object elem) {
	int i, j = findInsertPosition(elem);
	if (j < size) { 
	    if (cmp == null) { 
		Comparable c = (Comparable)elem;
		for (i = j+1; i < size && c.compareTo(elementData[i].get()) == 0; i++);
	    } else { 
		for (i = j+1; i < size && cmp.compare(elem, elementData[i].get()) == 0; i++);
	    }
	    return i-1;
	}
	return -1;
    }

    /**
     * Appends the specified element to the end of this list.
     *
     * @param o element to be appended to this list.
     * @return <tt>true</tt> (as per the general contract of Collection.add).
     */
    public boolean add(Object o) {
	ensureCapacity0(size + 1);  // Increments modCount!!
	add(findInsertPosition(o), o);
	return true;
    }

    /**
     * Appends all of the elements in the specified Collection to the end of
     * this list, in the order that they are returned by the
     * specified Collections Iterator.  The behavior of this operation is
     * undefined if the specified Collection is modified while the operation
     * is in progress.  (This implies that the behavior of this call is
     * undefined if the specified Collection is this list, and this
     * list is nonempty.)
     *
     * @param c elements to be inserted into this list.
     * @throws    IndexOutOfBoundsException if index out of range <tt>(index
     *		  &lt; 0 || index &gt; size())</tt>.
     */
    public boolean addAll(Collection c) {
	modCount++;
	int numNew = c.size();
	ensureCapacity0(size + numNew);

	Iterator e = c.iterator();
	deferredUpdate += 1;
	try { 
	    for (int i=0; i<numNew; i++) { 
		add(e.next());
	    }
	} finally { 
	    deferredUpdate -= 1;
	}
	if (numNew != 0) { 
	    save();
	    return true;
	}
	return false;
    }

    protected POComparator cmp;
}
