/* Consus - Pure Java Object Relational Database Management System
 * Base class for all persistent classes
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.consus;

import java.sql.SQLException;

/**
 * Base class for all persistent-capable classes. You should derive your
 * class from the <code>Persistent</code> class to support transparent persistency.
 */
public class Persistent { 
    /** 
     * Specify whether the referenced objects should be loaded implicitly together
     * with the parent object or not. Default implementation of this method returns
     * <code>true</code>, so the complete tree of objects will be loaded into memory.
     * To prevent loading of all objects (and so exhausting the memory), redefine
     * this method and return <code>false</code>. In this case you will have
     * to load the components explicitly.
     * @return <code>true</code> to load all graph with referenced objects or 
     * <code>false</code> for loading stubs instead of referenced object.
     */
    public boolean loadClosure() { 
	return true;
    }

    /**
     * Check whether the object has not yet been stored in the database (assigned the persistent
     * object identifier). An object is concidered new starting from its creation and until 
     * stored in the database.
     * @return <code>true</code> if object has no assigned OID yet. 
     */ 
    public final boolean isNew() { 
	return state == PO_NEW;
    }

    /**
     * Check whether the object is a stub. A stub object has the same type as the normal
     * object, but has no data loaded.
     * @return <code>true</code> when object has not yet been loaded from the database,
     * and instead of it a stub was created.
     */     
    public final boolean isLoaded() { 
	return state == PO_LOADED || state == PO_NEW;
    }

    /**
     * Check if the object is marked as deleted.
     * @return <code>true</code> if remove method was invoked for the object.
     */     
    public final boolean isDeleted() { 
	return state == PO_DELETED;
    }

    /**
     * Load stub object from the database. If the object was already loaded, this method
     * does nothing.
     * @return pointer to <code>this</code>
     */
    public synchronized Persistent get() { 
	Assert.that(state != PO_DELETED);
	return state == PO_STUB ? load() : this;
    }

    
    /**
     * Load stub object from the database. If the object was already loaded, this method
     * does nothing. This method is static and checks its parameter for the <code>null</code> value.
     * @return pointer to <code>this</code>
     */
    public static Persistent get(Persistent obj) { 
	if (obj == null) { 
	    return null;
	}
	Assert.that(obj.state != PO_DELETED);
	return obj.state == PO_STUB ? obj.load() : obj;
    }

    /**
     * Load stub object or reload the object from the database.
     * @return pointer to <code>this</code>
     */
    public synchronized Persistent load() { 
	Assert.that(state != PO_DELETED);
	if (state == PO_NEW) { 
	    return this;
	}	
	try { 
	    state = PO_LOADED;
	    db.loadObject(this);
	    return this;
	} catch(Exception x) {
	    throw new POSError(x);
	}
    }

    /**
     * Load the components of the object. This method should be called for the classes
     * that disable implicit loading of object closure. In this case the referenced
     * components should be loaded explicitly by this method.
     */
    public synchronized void loadComponents() { 
	if (state == PO_NEW) { 
	    return;
	}
	Assert.that(state == PO_LOADED);
	try { 
	    db.loadComponents(this);
	} catch(Exception x) {
	    throw new POSError(x);
	}
    }

    /** 
     * Clear all object fields (to make it possible for GC to collect 
     * unused persistent objects). If this method is not called and all persistent objects
     * in the database are referencing each other (for example they are linked in a L2-list),
     * then after some time all the data can be loaded into memory from the database.
     */
    public synchronized void unget() { 
	try { 
	    db.clearComponents(this);
	} catch(Exception x) { 
	    throw new POSError(x);
	}
	state = PO_STUB;
    } 
    
    /** 
     * Lock database in exclusive mode. The lock is kept until the end of the transaction.
     * This method should be called to prevent deadlocks caused by upgradina a shared lock
     * to exclusive. For example, if you first issue read-only statement and then
     * within the same transaction, issue an update statement, it can cause a deadlock
     * if concurrent thread also tries to upgrade the shared lock to exclusive.
     */
    public void lock() { 
	try { 
	    db.lock();
	} catch(Exception x) { 
	    throw new POSError(x);
	}
    }

    /**
     * Store the object attached to the database to the storage. If object has not yet been
     * assigned to the database, this method does nothing.
     */
    public void store() { 
	if (db != null) { 
	    store(db);
	}
    }

    /**
     * Store the object in the database if the <code>deferredUpdate</code> flag is 0.
     * <code>deferredUpdate</code> flag indicates that the object will be saved by the
     * caller.
     * @return <code>true</code> if method stores the object, <code>false</code> otherwise
     */
    public synchronized boolean save() { 
	if (deferredUpdate == 0) { 
	    store();
	    return true;
	}
	return false;
    }

    /** 
     * Save the object in the database. If the object refers to some some other persistent-capable objects
     * without assigned object identifiers, then these objects will also be recursively saved in
     * the database.
     */
    public synchronized void store(PersistentObjectStorage db) { 
	Assert.that(db != null);
	Assert.that(state == PO_LOADED || state == PO_NEW);
	try { 
	    this.db = db;
	    db.storeObject(this);
	} catch(Exception x) {
	    x.printStackTrace();
	    throw new POSError(x);
	}
    }

    /**
     * Remove the object from the database.
     */
    public synchronized void remove() { 
	Assert.that(state != PO_DELETED);
	if (oid != 0) { 
	    try {
		db.deleteObject(this);
	    } catch(Exception x) { 
		throw new POSError(x);
	    }
	    oid = 0;
	}
	state = PO_DELETED;
    }

    /** 
     * Get the objects JDBC reference.
     * @return reference to the object as ConsusRef class.
     */
    public ConsusRef getRef() { 
	return new ConsusRef(db, oid); 
    }
    
    /**
     * Get the objects identifier.
     * @return object identifier
     */
    public long getOid() { 
	return oid; 
    }

    /**
     * Get the JDBC reference to the object.
     * @param obj object or <code>null</code>
     * @return reference to the object as ConsusRef class.
     */
    public static ConsusRef getRef(Persistent obj) { 
        return (obj == null) ? new ConsusRef(0) :  new ConsusRef(obj.db, obj.oid); 
    }
    
    /**
     * Get the objects identifier.
     * @param obj object or <code>null</code>
     * @return object identifier
     */   
    public static long getOid(Persistent obj) { 
	return obj == null ? 0 : obj.oid; 
    }

    /**
     * Assign a persistent object identifier to the object, if it hasn't been assigned yet.
     * @param db connection to the database in which object should be allocated
     */
    public synchronized long assignOID(PersistentObjectStorage db) { 
	if (oid == 0) { 
	    store(db);
	}
	return oid;
    }
	

    /**
     * Get the String representation of the persistent objects identifier.
     * @return class name '@' OID
     */
    public String toString() { 
	return getClass().getName() + "@" + Long.toHexString(oid); 
    }

    /**
     * Exchange references between objects in the database. The result of execution A.become(B)
     * is that thew OIDs of A and B are exchanged. All references to A in the database will
     * now refer to B, and vice versa. This operation doesn't affect references in main
     * memory. Object A should already have an assigned OID (being stored in the database)
     * If B is not yet stored in the database, it is explicitly assigned an OID by this method.
     * @param obj object which OID will be exchanged with <code>this</code> object
     */
    public synchronized void become(Persistent obj) { 
	long srcOID = oid;
	Assert.that(srcOID != 0);
	store();
	obj.store();
	try {
	    db.becomeObject(this, obj);
	} catch(Exception x) { 
	    throw new POSError(x);
	}
    }

    public final  void beginUpdate() { 
	deferredUpdate += 1;
    }
	
    public final void endUpdate() { 
	if (--deferredUpdate == 0) { 
	    store();
	}
    }
	
    public static final int PO_NEW     = 0;
    public static final int PO_STUB    = 1;
    public static final int PO_LOADED  = 2;
    public static final int PO_DELETED = 3;

    public transient long   oid;
    public transient int    state;
    public transient Object type;

    /**
     * Field used to group updates of persistent objects. In deferred update mode,
     * the method <code>Save</code> will not store objects immediately. The field
     * <code>deferredUpdate</code> is used as a counter
     * and, until it reaches zero, no stores will be performed. This is so that the most outer
     * method can save all changes, made by other invoked methods, using a single store
     * operation.
     */
    protected transient int                     deferredUpdate;
    public    transient PersistentObjectStorage db;
}
