/* Consus - Pure Java Object Relational Database Management System
 * Interface of persistent object storage. Collection of object oriented API methods.
 * Created by Konstantin Knizhnik, Apr 2001.
 */
package org.garret.consus;

import java.sql.*;

/**
 * The interface of persistent object storage. This interface declares methods
 * for the object oriented Consus API. This interface is implemented by ConsusConnection
 * and JDBCObjectRelationalAdapter. These methods can be used either with the
 * native Consus database engine or with any other DBMS with a JDBC interface
 * through the object-relational adapter.
 */
public interface PersistentObjectStorage { 
    /**
     * Returns true if the database has already been initialized.
     * If database was not initilialized, the programmer has to create
     * TypeInfo, OidTable and applications tables, insert information about class-table mapping
     * into TypeInfo table and initialize the ID counter by inserting a value in the OidTable.
     * Initialization can be done through standard JDBC methods.
     * @return <code>true</code> if database was already initialized, <code>false</code> otherwise
     */
    boolean isInitialized() throws Exception;

    /**
     * Opens the persistent object storage. This method should be called after checking
     * that the storage has been initialized, and if not, then it needs to initialize it.
     */
    void open() throws Exception;  

    /**
     * Fetch the current row of the cursor as a Java object.
     * @param cursor - result set with the current position corresponding to the fetched object
     * @return fetched object
     */
    Object fetchObject(ResultSet cursor) throws Exception;

    /**
     * Update the current row of the cursor with the new values passed in the Java object.
     * @param obj - object with the new values
     * @param cursor - result set with the current position pointing to the updated row
     */
    void updateObject(Object obj, ResultSet cursor) throws Exception; 


    /**
     * Load the stub object or reload the object from the database.
     * @param obj loaded persistent object
     */
     void loadObject(Persistent obj) throws Exception;

    /**
     * Load the components of the object. This method should be called for the classes
     * that disable the implicit loading of object closure. In this case the referenced
     * components should be loaded explicitly by this method.
     * @param obj persistent object which components are loaded
     */
     void loadComponents(Persistent obj) throws Exception; 


    /** 
     * Clear all object fields (to make it possible for GC to collect 
     * unused persistent object). If this method is not called and all pesistent object
     * in the database are referencing each other (for example linked in L2-list),
     * then after some time all data can be loaded to the memory from the database
     * @param obj persitent object which components are cleared
     */
     void clearComponents(Persistent obj) throws Exception; 


    /** 
     * Save object in the database. If object referes to some some other persistent-capable objcts
     * without assigned object identifiers, then these object will be also recursivley saved in
     * the database.
     * @param obj stored persitent object
     */
     void storeObject(Persistent obj) throws Exception;

    /**
     * Remove object from the database.
     * @param obj deleted persitent object
     */
     void deleteObject(Persistent obj) throws Exception;

    /**
     * Exchange references to the obejcts in the database. The result of execution A.become(B)
     * is that OIDs of A and B are exchanged, so that all references to the A in the database will
     * now refer to B, and visa versa. This operation doesn't affect references in the main 
     * memory. Object A should already have assigned OID (be stored in the database)
     * If B is not yet stored in the database, it is explicitly assigned OID by this method.
     * @param a object which OID will be exchanged with <code>b</code> object
     * @param b object which OID will be exchanged with <code>a</code> object
     */
     void becomeObject(Persistent a, Persistent b) throws Exception;

     /** 
     * Lock database in exclusive mode. The lock is kept until the end of transaction.
     * This method should be called to prevent deadlocks caused by upgradin shared lock 
     * to exclusive. For example, if you first issue read-only statement and then, 
     * within the same transaction, update statement, it can be a reason of deadlock
     * if concurrent thread also tries to upgrade shared lock to exclusive.
     */
     void lock() throws Exception;
    
    /**
     * Poll object cache size
     * @return number of object in object cache.
     */
    int getObjectCacheSize();

    /**
     * Throw all objects from the object cache. This method should be called
     * to prevent memory exhaustion caused by loading a lot of persistent interconnected 
     * objects (each of the objects is accessible from some set of root objects)
     * from the database. GC will not be able to deallocate such objects because they are
     * referencing each other. So finally all object from the database can be loaded to
     * the memory. Programmer should check number of loaded objects (by 
     * <code>getObjectCacheSize</code> or <code>getUsedMemorySize</code> method) and
     * if it axceeds some threshold, 
     * call <code>clearObjectCache</code> method to remove all objects from cache. 
     * Alternativly, progrtammer can call <code>Persistent.unget</code> method to replace
     * object with a stub, but this approach is less efficient because usually it throws 
     * away most recently used object. 
     */
    void clearObjectCache();

     
    /**
     * Get table iterator. More efficient way for iteration through the 
     * the table then using "select * from" and ResultSet. Provide sequential access and 
     * direct access by key.
     * @param tableName name of th table through which iteration will be performed
     * @return table iterator for specified table
     */
    TableIterator getTableIterator(String tableName) throws Exception; 

    /**
     * Get object by object identifier.
     * 
     * @param oid - identifier of object to be fetched
     * @return object with specified OID
     */
    Object getObjectByOid(long oid) throws Exception;
};





