/* Consus - Pure Java Object Relational Database Management System
 * Class for construction of parameterized queries
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.consus;

import java.sql.*;

/**
 * Class used for construction of parameterized queries.
 */
public class Query { 
    String     sql;
    Object[]   params;
    int        nParams;
    Connection conxn;
    boolean    firstConjunct;
    
    final void addParam(Object obj) { 
	if (nParams == params.length) { 
	    Object[] newParams = new Object[nParams*2];
	    System.arraycopy(params,0, newParams, 0, nParams);
	    nParams *= 2;
	    params = newParams;
	}
	params[nParams++] = obj;
    }

    /**
     * Reset the query.
     */
    public void reset() { 
	sql = "";
	firstConjunct = true;
	nParams = 0;
	if (params == null) { 
	    params = new Object[4];
	}
    }

    /**
     * Append String to the statement.
     * @param cond appended string 
     * @return <code>this</code>
     */
    public Query add(String cond) { 
	sql += cond;
	return this;
    }

    /**
     * Append a conjunct to the statement. If it is the first conjunct, it is prepended by
     * "<code> WHERE </code>", otherwise it is prepended by "<code> AND </code>".
     * @param cond appended conjunct
     * @return <code>this</code>
     */
    public Query and(String cond) { 
	if (firstConjunct) { 
	    sql += " where ";
	} else { 
	    sql += " and ";
	}
	sql += cond;
	firstConjunct = false;	
	return this;
    }
    
    /**
     * Bind the boolean parameter.
     * @param val parameter value
     * @return <code>this</code>
     */
    public Query param(boolean val) { 
	addParam(new Boolean(val));
	sql += " ? ";
	return this;
    } 

    /**
     * Bind the integer parameter.
     * @param val parameter value
     * @return <code>this</code>
     */
    public Query param(int val) { 
	addParam(new Integer(val));
	sql += " ? ";
	return this;
    } 

    /**
     * Bind the long parameter.
     * @param val parameter value
     * @return <code>this</code>
     */
    public Query param(long val) { 
	addParam(new Long(val));
	sql += " ? ";
	return this;
    } 
    /**
     * Bind the double parameter.
     * @param val parameter value
     * @return <code>this</code>
     */
    public Query param(double val) { 
	addParam(new Double(val));
	sql += " ? ";
	return this;
    } 

    /**
     * Bind the float parameter.
     * @param val parameter value
     * @return <code>this</code>
     */
    public Query param(float val) { 
	addParam(new Float(val));
	sql += " ? ";
	return this;
    } 
    /**
     * Bind any object parameter.
     * @param val parameter value
     * @return <code>this</code>
     */
    public Query param(Object val) { 
	addParam(val);
	sql += " ? ";
	return this;
    } 

    /**
     * Query constructor.
     * @param con database connection
     */
    public Query(Connection con) { 
	conxn = con;
	reset();
    }

    /**
     * Prepare the statement.
     * @return prepared statement
     */
    public ConsusPreparedStatement prepare() throws SQLException { 
	ConsusPreparedStatement pstmt = (ConsusPreparedStatement)conxn.prepareStatement(sql);
	for (int i = nParams; --i >= 0;) { 	    
	    pstmt.setObject(i+1, params[i]);
	}
	return pstmt;
    }

    /**  
     * Prepare and execute the query.
     * @return <code>ResultSet</code> with results of the query execution
     */
    public ConsusResultSet executeQuery() throws SQLException { 
	return (ConsusResultSet)prepare().executeQuery();
    }

    /**  
     * Prepare and execute the update statement.
     * @return number of updated records
     */
    public int executeUpdate() throws SQLException { 
	return prepare().executeUpdate();
    }
}
	




