/* Consus - Pure Java Object Relational Database Management System
 * Btree object container
 * Created by Konstantin Knizhnik, Jun 2000.
 */

package org.garret.consus;

import java.sql.SQLException;

public abstract class StringIndex extends Persistent implements Cloneable {
    /**
     * Remove all the objects from the index.
     */
    public abstract void clear();

    /**
     * Insert a new object in the index.
     * @param obj object to be placed in the index if unique constraint is not violated
     */
    public abstract boolean insertUnique(String key, Persistent obj);

    /**
     * Insert a new object in the index.
     * @param obj object to be placed in the index
     */
    public abstract void insert(String key, Persistent obj);
	
    /**
     * Remove an object from the index.
     * @param obj object to be removed from the index
     */
    public abstract boolean remove(String key, Persistent obj);

    /**
     * Select an object with a value of the key greater-or-equal than <code>firstKey</code>
     * and less-or-equal than <code>lastKey</code>.
     * @param firstKey  low bound of key rabge
     * @param lastKey   high bound of key rabge
     * @param forUpdate if objects are selected for update
     * @return ResultSet with selected objects
     */
    public abstract ConsusResultSet select(String firstKey, String lastKey, boolean forUpdate);

    /**
     * Select an object with a value of the key greater-or-equal than <code>firstKey</code>
     * and less-or-equal than <code>lastKey</code>.
     * @param firstKey  low bound of key rabge
     * @param lastKey   high bound of key rabge
     * @return ResultSet with selected objects
     */
    public abstract ConsusResultSet select(String firstKey, String lastKey);

    /**
     * Select for update an object with a value of the key greater-or-equal than <code>firstKey</code>
     * and less-or-equal than <code>lastKey</code>.
     * @param firstKey  low bound of key range
     * @param lastKey   high bound of key range
     * @return ResultSet with selected objects
     */
    public abstract ConsusResultSet selectForUpdate(String firstKey, String lastKey);

    /**
     * Select objects with the specified value of the key.
     * @param key searched value
     * @return ResultSet with selected objects
     */
    public abstract ConsusResultSet select(String key);

    /**
     * Select for update the objects with the specified value of the key.
     * @param key searched value
     * @return ResultSet with selected objects
     */
    public abstract ConsusResultSet selectForUpdate(String key);

    /**
     * Select the object with the specified value of the key.
     * @param key searched value
     * @return object with specified value of the key, <code>null</code> if there are no such object
     */
    public abstract Object selectUnique(String key);

    /**
     * Selects all indexed objects,
     * @param forUpdate if objects are selected for update
     * @return ResultSet with selected objects
     */
    public abstract ConsusResultSet select(boolean forUpdate);

    /**
     * Selects all indexed objects.
     * @return ResultSet with selected objects
     */
    public abstract ConsusResultSet select(); 
    
    /**
     * Selects all indexed objects for update.
     * @return ResultSet with selected objects
     */
    public ConsusResultSet selectForUpdate() {
	return select(true);
    }
    /**
     * Clone index.
     */
    public abstract Object clone() throws CloneNotSupportedException;
}








