/* Consus - Pure Java Object Relational Database Management System
 * Table iterator
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.consus;

import java.sql.SQLException;

/**
 * An Iterator through the table records (is equivalent to ResultSet for a single table select
 * without condition, but is more efficient).
 */
public abstract class TableIterator { 
    /**
     * Moves the cursor down one row from its current position.
     * A ResultSet cursor is initially positioned before the first row; the
     * first call to the next, makes the first row the current row; the
     * second call makes the second row the current row, and so on. 
     *
     * @return true if the new current row is valid; false if there
     * are no more rows 
     * @exception SQLException if a database access error occurs
     */
    public abstract boolean next() throws SQLException; 

    /**
     * <p>Moves the cursor to the previous row in the result set.  
     *
     * <p>Note: <code>previous()</code> is not the same as
     * <code>relative(-1)</code> because it
     * makes sense to call</code>previous()</code> when there is no current row.
     *
     * @return true if the cursor is on a valid row; false if it is off the result set
     * @exception SQLException if a database access error occurs or the
     * result set type is TYPE_FORWARD_ONLY
     */
    public abstract boolean previous() throws SQLException;

    /** 
     * Fetch the current table row as a Java object.
     */
    public abstract Object get() throws SQLException;

    /** 
     * Closes the iterator and releases all related resourses.
     */
    public abstract void close()  throws SQLException;
  
    /**
     * <p>Indicates whether the cursor is before the first row in the resultset.
     *
     * @return true if the cursor is before the first row, false otherwise. Returns
     * false when the resultset contains no rows.
     * @exception SQLException if a database access error occurs
     */
    public abstract boolean isBeforeFirst() throws SQLException;

    /**
     * <p>Indicates whether the cursor is after the last row in the resultset.
     *
     * @return true if the cursor is  after the last row, false otherwise.  Returns
     * false when the resultset contains no rows.
     * @exception SQLException if a database access error occurs
     */
    public abstract boolean isAfterLast() throws SQLException;
 
    /**
     * <p>Indicates whether the cursor is on the first row of the resultset.
     *
     * @return true if the cursor is on the first row, false otherwise.   
     * @exception SQLException if a database access error occurs
     */
    public abstract boolean isFirst() throws SQLException;
 
    /**
     * <p>Indicates whether the cursor is on the last row of the resultset.
     *
     * @return true if the cursor is on the last row, false otherwise. 
     * @exception SQLException if a database access error occurs
     */
    public abstract boolean isLast() throws SQLException;

    /**
     * <p>Moves the cursor to the front of the resultset, just before the
     * first row. Has no effect if the resultset contains no rows.
     *
     * @exception SQLException if a database access error occurs or the 
     * result set type is TYPE_FORWARD_ONLY
     */
    public abstract void beforeFirst() throws SQLException;

    /**
     * <p>Moves the cursor to the end of the resultset, just after the last
     * row.  Has no effect if the resultset contains no rows.
     *
     * @exception SQLException if a database access error occurs or the 
     * result set type is TYPE_FORWARD_ONLY 
     */
    public abstract void afterLast() throws SQLException;

    /**
     * <p>Moves the cursor to the first row in the resultset.
     *
     * @return true if the cursor is on a valid row; false if
     *         there are no rows in the result set
     * @exception SQLException if a database access error occurs or the 
     * result set type is TYPE_FORWARD_ONLY
     */
    public abstract boolean first() throws SQLException;

    /**
     * <p>Moves the cursor to the last row in the resultset.
     *
     * @return true if the cursor is on a valid row;
     * false if there are no rows in the resultset
     * @exception SQLException if a database access error occurs or the
     * result set type is TYPE_FORWARD_ONLY.
     */
    public abstract boolean last() throws SQLException;

    /**
     * <p>Retrieves the current row number.  The first row is number 1, the
     * second number 2, and so on.  
     *
     * @return the current row number; 0 if there is no current row
     * @exception SQLException if a database access error occurs
     */
    public abstract int getRow() throws SQLException;

    /**
     * <p>Moves the cursor to the given row number in the resultset.
     *
     * <p>If the row number is positive, the cursor moves to 
     * the given row number with respect to the
     * beginning of the resultset.  The first row is row 1, the second
     * is row 2, and so on. 
     *
     * <p>If the given row number is negative, the cursor moves to
     * an absolute row position with respect to
     * the end of the resultset.  For example, calling
     * <code>absolute(-1)</code> positions the 
     * cursor on the last row, <code>absolute(-2)</code> indicates the next-to-last
     * row, and so on.
     *
     * <p>An attempt to position the cursor beyond the first or last row in
     * the resultset leaves the cursor before or after the first or last
     * row, respectively.
     *
     * <p>Note: Calling <code>absolute(1)</code> is the same
     * as calling <code>first()</code>.
     * Calling <code>absolute(-1)</code> is the same as calling <code>last()</code>.
     *
     * @return true if the cursor is on the result set; false otherwise
     * @exception SQLException if a database access error occurs or 
     * row is 0, or result set type is TYPE_FORWARD_ONLY.
     */
    public abstract boolean absolute(int row) throws SQLException;

    /**
     * <p>Moves the cursor a relative number of rows, either positive or negative.
     * Attempting to move beyond the first or last row in the
     * result set positions the cursor before or after the
     * the first or last row. Calling <code>relative(0)</code> is valid, but does
     * not change the cursors position.
     *
     * <p>Note: Calling <code>relative(1)</code>
     * is different from calling <code>next()</code>
     * because it makes sense to call <code>next()</code> when there is no current row,
     * for example, when the cursor is positioned before the first row
     * or after the last row of the resultset.
     *
     * @return true if the cursor is on a row; false otherwise
     * @exception SQLException if a database access error occurs, there
     * is no current row, or the result set type is TYPE_FORWARD_ONLY
     */
    public abstract boolean relative(int rows) throws SQLException;

    /**
     * Deletes the current row from the resultset and the underlying
     * database.  Cannot be called when on the insert row.
     *
     * @exception SQLException if a database access error occurs or if
     * called when on the insert row.
     */
    public abstract void deleteRow() throws SQLException;

    /**
     * Deletes all records in the table.
     * 
     * @exception SQLException if a database access error occurs or if
     * called when on the insert row.
     */
    public abstract void deleteAllSelectedRows() throws SQLException;

    /**
     * Get the number of rows in the table.
     * @return number of rows in the table
     */
    public abstract int getNumberOfRows() throws SQLException;

    /**
     * Select an object from the table using a unique key.
     * @param keyName table column with a unique key
     * @param keyValue value of the key to look for
     * @return object with specified value of the key or <code>null</code> otherwise
     * @exception ConsusRuntimeException when there is no such column in the table or
     * it is not indexed; ConsusNotUniqueException when the key is not unique
     */
    public abstract Object selectByKey(String keyName, String keyValue) throws SQLException;

    /**
     * Select an object from the table using an object identifier.
     * @param oid object identifier
     * @return object with the specified identifier
     */
    public abstract Object selectByOid(long oid) throws SQLException;

    /**
     * Returns an array containing all of the selected elements.
     * The runtime type of the returned array is that of the
     * specified array. If the number of selected elements is not greater
     * then specified array length, then they are returned.
     * Otherwise, a new array is allocated with the runtime
     * type of the specified array and a length equal to the number of selected elements.<p>
     *
     * If the list fits into the specified array with room to spare (i.e. the
     * array has more elements than the selection), the element in the array
     * immediately following the end of the collection is set to
     * <tt>null</tt>.
     *
     * @return an array containing all of the selected records
     */
    public Object[] toArray(Object[] a) throws SQLException {
	int i, nSelected = getNumberOfRows();
        if (a.length < nSelected) { 
            a = (Object[])java.lang.reflect.Array.newInstance(
			     a.getClass().getComponentType(), nSelected);
	}
	for (i = 0; next(); i++) { 
	    a[i] = get();
	}	
        if (i < a.length) {
            a[i] = null;
	}
        return a;
    }


    /**
     * Returns an array of java.lang.Object objects, containing all the selected elements.
     *
     * @return an array containing all the selected records
     */
    public Object[] toArray() throws SQLException { 
	return toArray(new Object[0]);
    }
    
}








