/* Consus - Pure Java Object Relational Database Management System
 * Node for binary operations
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.SQLException;

class BinOpNode extends ExprNode { 
    ExprNode left;
    ExprNode right;

    String dump() { 
	return "(" + opMnem[tag] + " " + left.dump() + " " + right.dump() + ")";
    }
    boolean equalsNode(ExprNode node) { 
	return node.tag == tag && left.equalsNode(((BinOpNode)node).left) && right.equalsNode(((BinOpNode)node).right);
   }

    int dependsOn() { 
	return Math.max(left.dependsOn(), right.dependsOn());
    }

    long evaluateInt(ConsusStatement stmt) throws SQLException { 
	long lval = left.evaluateInt(stmt);
	long rval = right.evaluateInt(stmt);
	long res;
	switch (tag) { 
	  case opIntAdd:
	    return lval + rval;
	  case opIntSub:
	    return lval - rval;
	  case opIntMul:
	    return lval * rval;
	  case opIntDiv:
	    return lval / rval;
	  case opIntAnd:
	    return lval & rval;
	  case opIntOr:
	    return lval | rval;
	  case opIntPow:
	    res = 1;
	    if (rval < 0) {
		lval = 1/lval;
		rval = -rval;
	    }
	    while (rval != 0) {
		if ((rval & 1) != 0) { 
		    res *= lval;
		}
		lval *= lval;
		rval >>>= 1;
	    }
	    return res; 
	  default:
	    Assert.failed("Invalid tag" + tag);
	}
	return 0;
    }

    double evaluateReal(ConsusStatement stmt) throws SQLException { 
	double lval = left.evaluateReal(stmt);
	double rval = right.evaluateReal(stmt);
	switch (tag) { 
	  case opRealAdd:
	    return lval + rval;
	  case opRealSub:
	    return lval - rval;
	  case opRealMul:
	    return lval * rval;
	  case opRealDiv:
	    return lval / rval;
	  case opRealPow:
	    return Math.pow(lval, rval);
	  default:
	    Assert.failed("Invalid tag" + tag);
	}
	return 0.0;
    }

    String evaluateStr(ConsusStatement stmt) throws SQLException { 
	String lval = left.evaluateStr(stmt);
	String rval = right.evaluateStr(stmt);
	return lval + rval;
    }

    static boolean scanSet(ConsusStatement stmt, ExprNode subquery, Object val) throws SQLException { 
	byte[] saveRec = stmt.rec;
	Object[] set = subquery.evaluateSet(stmt);
	stmt.rec = saveRec;
	SelectNode select = (SelectNode)subquery;
	if ((select.order != null && select.order.columnNo == 0) 
	    || (select.resultColumns[0].field != null 
		&& select.resultColumns[0].field == select.indexField))
	{
	    int l = 0, r = set.length; 
	    if (select.order != null && !select.order.ascent) { 
		while (l < r) { 
		    int i = (l+r) >> 1;
		    int diff = ((Comparable)set[i]).compareTo(val);
		    if (diff <= 0) { 
			r = i;
		    } else { 
			l = i + 1;
		    }
		}
		stmt.rec = saveRec;
		return r < set.length && set[r].equals(val);
	    } else { 
		while (l < r) { 
		    int i = (l+r) >> 1;
		    int diff = ((Comparable)set[i]).compareTo(val);
		    if (diff < 0) { 
			l = i + 1;
		    } else { 
			r = i;
		    }
		}	
		stmt.rec = saveRec;
		return r < set.length && set[r].equals(val);
	    }		
	} else { 
	    for (int i = set.length; --i >= 0;) { 
		if (set[i].equals(val)) { 
                    stmt.rec = saveRec;
		    return true;
		}
	    }
	}
        stmt.rec = saveRec;
	return false;
    }

    boolean evaluateBool(ConsusStatement stmt) throws SQLException { 
	switch (tag) { 
	  case opBoolAnd:
	    return left.evaluateBool(stmt) && right.evaluateBool(stmt);
	  case opBoolOr:
	    return left.evaluateBool(stmt) || right.evaluateBool(stmt);

	  case opIntEq:
	    return left.evaluateInt(stmt) == right.evaluateInt(stmt);
	  case opIntNe:
	    return left.evaluateInt(stmt) != right.evaluateInt(stmt);
	  case opIntLt:
	    return left.evaluateInt(stmt) < right.evaluateInt(stmt);
	  case opIntLe:
	    return left.evaluateInt(stmt) <= right.evaluateInt(stmt);
	  case opIntGt:
	    return left.evaluateInt(stmt) > right.evaluateInt(stmt);
	  case opIntGe:
	    return left.evaluateInt(stmt) >= right.evaluateInt(stmt);

	  case opRealEq:
	    return left.evaluateReal(stmt) == right.evaluateReal(stmt);
	  case opRealNe:
	    return left.evaluateReal(stmt) != right.evaluateReal(stmt);
	  case opRealLt:
	    return left.evaluateReal(stmt) <  right.evaluateReal(stmt);
	  case opRealLe:
	    return left.evaluateReal(stmt) <= right.evaluateReal(stmt);
	  case opRealGt:
	    return left.evaluateReal(stmt) >  right.evaluateReal(stmt);
	  case opRealGe:
	    return left.evaluateReal(stmt) >= right.evaluateReal(stmt);

	  case opStrEq:
	    return left.evaluateStr(stmt).equals(right.evaluateStr(stmt));
	  case opStrNe:
	    return !left.evaluateStr(stmt).equals(right.evaluateStr(stmt));
	  case opStrLt:
	    return left.evaluateStr(stmt).compareTo(right.evaluateStr(stmt)) < 0;
	  case opStrLe:
	    return left.evaluateStr(stmt).compareTo(right.evaluateStr(stmt)) <= 0;
	  case opStrGt:
	    return left.evaluateStr(stmt).compareTo(right.evaluateStr(stmt)) > 0;
	  case opStrGe:
	    return left.evaluateStr(stmt).compareTo(right.evaluateStr(stmt)) >= 0;

	  case opBoolEq:
	    return left.evaluateBool(stmt) == right.evaluateBool(stmt);
	  case opBoolNe:
	    return left.evaluateBool(stmt) != right.evaluateBool(stmt);

	  case opOidEq:
	    return left.evaluateOid(stmt) == right.evaluateOid(stmt);
	  case opOidNe:
	    return left.evaluateOid(stmt) != right.evaluateOid(stmt);

	  case opScanSetBool:
	      return scanSet(stmt, right, new Boolean(left.evaluateBool(stmt)));
	  case opScanSetInt:
	      return scanSet(stmt, right, new Long(left.evaluateInt(stmt)));
	  case opScanSetReal:
	      return scanSet(stmt, right, new Double(left.evaluateReal(stmt)));
	  case opScanSetStr:
	      return scanSet(stmt, right, left.evaluateStr(stmt));
	  case opScanSetRef:
	      return scanSet(stmt, right, new ConsusRef(left.evaluateOid(stmt)));

	  case opScanArrayBool:
	  {
	    byte val = (byte)(left.evaluateBool(stmt) ? 1 : 0);
	    int offs = right.evaluateBase(stmt);  
	    int size = Varying.getSize(stmt.rec, offs);
	    offs = Varying.getOffs(stmt.rec, offs);
	    while (--size >= 0) {
		if (stmt.rec[offs++] == val) { 
		    return true;
		}
	    }
	    return false;
	  }
	  case opScanArrayInt1:
	  {
	    long val = left.evaluateInt(stmt);
	    int offs = right.evaluateBase(stmt);  
	    int size = Varying.getSize(stmt.rec, offs);
	    offs = Varying.getOffs(stmt.rec, offs);
	    while (--size >= 0) {
		if (stmt.rec[offs++] == val) { 
		    return true;
		}
	    }
	    return false;
	  }
	  case opScanArrayInt2:
	  {
	    long val = left.evaluateInt(stmt);
	    int offs = right.evaluateBase(stmt);  
	    int size = Varying.getSize(stmt.rec, offs);
	    offs = Varying.getOffs(stmt.rec, offs);
	    while (--size >= 0) {
		if (Bytes.unpack2(stmt.rec, offs) == val) { 
		    return true;
		}
		offs += 2;
	    }
	    return false;
	  }
	  case opScanArrayInt4:
	  {
	    long val = left.evaluateInt(stmt);
	    int offs = right.evaluateBase(stmt);  
	    int size = Varying.getSize(stmt.rec, offs);
	    offs = Varying.getOffs(stmt.rec, offs);
	    while (--size >= 0) {
		if (Bytes.unpack4(stmt.rec, offs) == val) { 
		    return true;
		}
		offs += 4;
	    }
	    return false;
	  }
	  case opScanArrayInt8:
	  {
	    long val = left.evaluateInt(stmt);
	    int offs = right.evaluateBase(stmt);  
	    int size = Varying.getSize(stmt.rec, offs);
	    offs = Varying.getOffs(stmt.rec, offs);
	    while (--size >= 0) {
		if (Bytes.unpack8(stmt.rec, offs) == val) { 
		    return true;
		}
		offs += 8;
	    }
	    return false;
	  }
	  case opScanArrayReal4:
	  {
	    double val = left.evaluateReal(stmt);
	    int offs = right.evaluateBase(stmt);  
	    int size = Varying.getSize(stmt.rec, offs);
	    offs = Varying.getOffs(stmt.rec, offs);
	    while (--size >= 0) {
		if (Float.intBitsToFloat(Bytes.unpack4(stmt.rec, offs)) == val) { 
		    return true;
		}
		offs += 4;
	    }
	    return false;
	  }
	  case opScanArrayReal8:
	  {
	    double val = left.evaluateReal(stmt);
	    int offs = right.evaluateBase(stmt);  
	    int size = Varying.getSize(stmt.rec, offs);
	    offs = Varying.getOffs(stmt.rec, offs);
	    while (--size >= 0) {
		if (Double.longBitsToDouble(Bytes.unpack8(stmt.rec, offs)) == val) { 
		    return true;
		}
		offs += 8;
	    }
	    return false;
	  }
	  case opScanArrayStr:
	  {
	    String val = left.evaluateStr(stmt);
	    int offs = right.evaluateBase(stmt);  
	    int size = Varying.getSize(stmt.rec, offs);
	    offs = Varying.getOffs(stmt.rec, offs);
	    int len = val.length();
	  scan:
	    while (--size >= 0) {
		int strLen = Varying.getSize(stmt.rec, offs);
		if (strLen == len) { 
		    int strOffs = Varying.getOffs(stmt.rec, offs);
		    for (int i = 0; i < len; i++, strOffs += 2) { 
			if (val.charAt(i) != (char)Bytes.unpack2(stmt.rec, strOffs)) {
			    offs += Varying.sizeof;
			    continue scan;
			}
		    }
		    return true;
		}
		offs += Varying.sizeof;
	    }
	    return false;
	  }
	  case opScanArrayRef:
	  {
	    int val = left.evaluateOid(stmt);
	    int offs = right.evaluateBase(stmt);  
	    int size = Varying.getSize(stmt.rec, offs);
	    offs = Varying.getOffs(stmt.rec, offs);
	    while (--size >= 0) {
		if (Bytes.unpack4(stmt.rec, offs) == val) { 
		    return true;
		}
		offs += 4;
	    }
	    return false;
	  }
	  case opInString:
	  {
	    String substr = left.evaluateStr(stmt);
	    String str = right.evaluateStr(stmt);
	    return str.indexOf(substr) >= 0;
	  }
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	return false;
    }

    BinOpNode(int type, int tag, ExprNode left, ExprNode right) {
	super(type, tag);
	this.left = left;
	this.right = right;
    }
}
