/* Consus - Pure Java Object Relational Database Management System
 * B-Tree implementation 
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;


class Btree { 
    static int getRoot(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof);
    }
    static int getHeight(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof+4);
    }
    static int getType(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof+8);
    }

    static void setRoot(byte[] rec, int root) { 
	Bytes.pack4(rec, Record.sizeof, root);
    }
    static void setHeight(byte[] rec, int height) { 
	Bytes.pack4(rec, Record.sizeof+4, height);
    }
    static void setType(byte[] rec, int type) { 
	Bytes.pack4(rec, Record.sizeof+8, type);
    }

    static final int sizeof = Record.sizeof+12;


    static final int op_done = 0;
    static final int op_overflow = 1;
    static final int op_underflow = 2;
    static final int op_not_found = 3;
    static final int op_duplicate = 4;

    static int allocate(ConsusConnection db, int type) throws SQLException
    {
	int oid = db.allocateId();
	long pos = db.allocate(sizeof, 0);
	db.setPos(oid, pos | ConsusConnection.dbModifiedFlag);
	byte[] tree = new byte[sizeof];
	setType(tree, type);
	Record.setSize(tree, 0, sizeof);
	db.pool.put(pos, tree);
	return oid;
    }

    static void find(ConsusConnection db, int treeId, SearchContext sc)
	throws SQLException
    {
	byte[] tree = db.get(treeId);
	int root = getRoot(tree);
	if (root != 0) {
	    BtreePage.find(db, root, sc, getType(tree), getHeight(tree));
	}
    }

    static boolean insert(ConsusConnection db, FieldDescriptor fd, int recordId)
	throws SQLException
    {
	byte[] tree = db.get(fd.index);
	SearchValue ins = new SearchValue();
	int type = getType(tree);
	byte[] rec = db.get(recordId);
	ins.extract(rec, fd.offs, type);
	ins.oid = recordId;
	return insert(db, fd.index, tree, ins, type, (fd.attr & FieldDescriptor.Unique) != 0);
    }


    static boolean insert(ConsusConnection db, int treeId, int recordId, String key, boolean unique)
	throws SQLException
    {
	byte[] tree = db.get(treeId);
	SearchValue ins = new SearchValue();
	ins.sval = key.toCharArray();
	ins.oid = recordId;
	return insert(db, treeId, tree, ins, FieldDescriptor.tpString, unique);
    }


    static boolean insert(ConsusConnection db, int treeId, byte[] tree, SearchValue ins, 
			  int type, boolean unique) throws SQLException
    {
	int root = getRoot(tree);

	if (type == FieldDescriptor.tpString) { 
	    if (ins.sval.length > BtreePage.maxStrKeySize) { 
		throw new ConsusRuntimeError("String key exceeds limit for maximal string key size");
	    }
	}

	if (root == 0) {
	    setRoot(tree, BtreePage.allocate(db, 0, type, ins));
	    setHeight(tree, 1);
	    db.update(treeId, tree);
	    return true;
	} else { 
	    int height = getHeight(tree);
	    int result = BtreePage.insert(db, root, type, ins, height, unique);
	    Assert.that(result != op_not_found);
	    if (result == op_overflow) { 
		setRoot(tree, BtreePage.allocate(db, root, type, ins));
		setHeight(tree, height + 1);
		db.update(treeId, tree);
	    }
	    return result != op_duplicate;
	}
    }
	    

    static boolean remove(ConsusConnection db, FieldDescriptor fd, int recordId)
	throws SQLException
    {
	byte[] tree = db.get(fd.index);
	byte[] rec = db.get(recordId);
	int type = getType(tree);
	SearchValue rem = new SearchValue();
	rem.extract(rec, fd.offs, type);
	rem.oid = recordId;
	return remove(db, fd.index, tree, rem, type);
    }


    static boolean remove(ConsusConnection db, int treeId, int recordId, String key)
	throws SQLException
    {
	byte[] tree = db.get(treeId);
	SearchValue rem = new SearchValue();
	rem.sval = key.toCharArray();
	rem.oid = recordId;
	return remove(db, treeId, tree, rem, FieldDescriptor.tpString);
    }



    static boolean remove(ConsusConnection db, int treeId, byte[] tree, SearchValue rem, int type)
	throws SQLException
    {
	int root = getRoot(tree);
	int height = getHeight(tree);
	int result = BtreePage.remove(db, root, type, rem, height);
	if (result == op_not_found) { 
	    return false;
	}
	if (result == op_underflow && height != 1) { 
	    Page pg = db.getPage(root);
	    if (BtreePage.getnItems(pg) == 0) { 			
		if (type == FieldDescriptor.tpString) { 
		    setRoot(tree, BtreePage.getKeyStrOid(pg, 0));
		} else { 
		    setRoot(tree, BtreePage.getRecord(pg, BtreePage.maxItems-1));
		}
		setHeight(tree, height - 1);
		db.update(treeId, tree);
		db.freePage(root);
	    }
	    db.pool.unfix(pg);
	} else if (result == op_overflow) { 
	    setRoot(tree, BtreePage.allocate(db, root, type, rem));
	    setHeight(tree, height + 1);
	    db.update(treeId, tree);
	}
	return true;
    }

    static void  purge(ConsusConnection db, int treeId) 
	throws SQLException
    { 
	byte[] tree = db.get(treeId);
	int root = getRoot(tree);
	if (root != 0) { 
	    BtreePage.purge(db, root, getType(tree), getHeight(tree));
	    setRoot(tree, 0);
	    setHeight(tree, 0);
	    db.update(treeId, tree);
	}
    }
    
    static void  drop(ConsusConnection db, int treeId) 
	throws SQLException
    { 
	purge(db, treeId);
	db.free(db.getPos(treeId) & ~ConsusConnection.dbFlagsMask, Btree.sizeof);
	db.freeId(treeId);
    }

    static int cloneTree(ConsusConnection db, int treeId) throws SQLException
    {
	int oid = db.allocateId();
	long pos = db.allocate(sizeof, 0);
	db.setPos(oid, pos | ConsusConnection.dbModifiedFlag);
	byte[] origTree = db.get(treeId);
	byte[] newTree = new byte[sizeof];
	int type = getType(origTree);
	int height = getHeight(origTree);
	setType(newTree, type);
	setHeight(newTree, height);
	int root = getRoot(origTree);
	if (root != 0) { 
	    setRoot(newTree, BtreePage.clonePage(db, root, type, height));
	}
	Record.setSize(newTree, 0, sizeof);
	db.pool.put(pos, newTree);
	return oid;
    }
    
    static void  traverseForward(ConsusConnection db, int treeId,
				 ConsusStatement stmt, ExprNode condition)
	throws SQLException
    {
	byte[] tree = db.get(treeId);
	int root = getRoot(tree);
	if (root != 0) { 
	    BtreePage.traverseForward(db, root, stmt, condition,
				      getType(tree), getHeight(tree));
	}
    }

    static void  traverseBackward(ConsusConnection db, int treeId,
				  ConsusStatement stmt, ExprNode condition)
	throws SQLException
    {
	byte[] tree = db.get(treeId);
	int root = getRoot(tree);
	if (root != 0) { 
	    BtreePage.traverseBackward(db, root, stmt, condition, 
				       getType(tree), getHeight(tree));
	}	
    }
}
