/* Consus - Pure Java Object Relational Database Management System
 * Packing/unpacking data types 
 * Created by Konstantin Knizhnik, Apr 2001.
 */
package org.garret.internal.consus;

/**
 * Communication buffer used by Java Gigabase CLI to communicate with server
 */
public class ByteBuffer {
    public final void extend(int len) {  
	if (pos + len > buf.length) { 
	    int newLen = pos + len > len*2 ? pos + len : len*2;
	    byte[] newBuf = new byte[newLen];
	    System.arraycopy(buf, 0, newBuf, 0, buf.length); 
	    buf = newBuf;
	}
    }

    public final void putByte(int val) {
	extend(1);
	buf[pos++] = (byte)val;
    }

    public final void putBytes(byte arr[]) { 
	int i = 0, j = pos;
	int len = arr.length;
	extend(len);
	byte[] dst = buf;
	while (i < len) { 
	    dst[j++] = arr[i++]; 
	}
	pos = j;
    }

    public final void putShort(int val) {
	extend(2);
	Bytes.pack2(buf, pos, (short)val);
	pos += 2;
    }

    public final void putInt(int val) {
	extend(4);
	Bytes.pack4(buf, pos, val);
	pos += 4;
    }

    public final void putLong(long val) {
	extend(8);
	Bytes.pack8(buf, pos, val);
	pos += 8;
    }

    public final void putDouble(double val) {
	extend(8);
	Bytes.pack8(buf, pos, Double.doubleToLongBits(val));
	pos += 8;
    }

    public final void putFloat(float val) {
	extend(4);
	Bytes.pack4(buf, pos, Float.floatToIntBits(val));
	pos += 4;
    }

    public final void putString(String str) { 
	if (str != null) { 
	    byte[] bytes = str.getBytes();
	    int len = bytes.length;
	    putInt(len);
	    int i = pos, j = 0;
	    extend(len);
	    byte[] dst = buf;
	    while (j < len) { 
		dst[i++] = bytes[j++];
	    }
	    pos = i;
	} else { 
	    putInt(-1);
	}
    }

    public final void putByteArray(byte[] arr) { 
	if (arr != null) { 	    
	    int len = arr.length;	
	    putInt(len);
	    int i = pos, j = 0;
	    extend(len);
	    byte[] dst = buf;
	    while (j < len) { 
		dst[i++] = arr[j++];
	    }
	    pos = i;
	} else { 
	    putInt(-1);
	}
    }

    public final byte getByte() { 
	return buf[pos++];
    }

    public final short getShort() { 
	short value = Bytes.unpack2(buf, pos);
	pos += 2;
	return value;
    }

    public final int getInt() { 
	int value = Bytes.unpack4(buf, pos);
	pos += 4;
	return value;
    }

    public final long getLong() { 
	long value = Bytes.unpack8(buf, pos);
	pos += 8;
	return value;
    }

    public final float getFloat() { 
	float value = Float.intBitsToFloat(Bytes.unpack4(buf, pos));
	pos += 4;
	return value;
    }

    public final double getDouble() { 
	double value = Double.longBitsToDouble(Bytes.unpack8(buf, pos));
	pos += 8;
	return value;
    }

    public final String getString() { 
	int len = getInt();
	if (len < 0) { 
	    return null;
	}
	String value = new String(buf, pos, len);
	pos += len;
	return value;
    }

    public final byte[] getByteArray() { 
	int i, j;
	int len = getInt();
	if (len < 0) { 
	    return null;
	}
	byte arr[] = new byte[len];
	byte[] src = buf;
	for (i = 0, j = pos; i < len; arr[i++] = src[j++]); 
	pos = j;
	return arr;
    }

    public final void getBytes(byte arr[]) { 
	int i, j;
	int len = arr.length;
	byte[] src = buf;
	for (i = 0, j = pos; i < len; arr[i++] = src[j++]); 
	pos = j;
    }


    public final byte[] getBytes() { 
	byte[] arr = new byte[pos];
	System.arraycopy(buf, 0, arr, 0, pos);
	return arr;
    }

    public final int size() { 
	return pos;
    }

    public ByteBuffer() { 
	buf = new byte[256];
	pos = 0;
    }

    public ByteBuffer(byte[] p) {
	buf = p;
	pos = 0;
    }

    protected byte buf[];
    protected int  pos;
}




