/* Consus - Pure Java Object Relational Database Management System
 * Connection protocol implementation (most of the database methods are implemeted here)
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;
import java.io.*;
import java.net.Socket;
import java.util.Hashtable;
import java.util.Properties;

/**
 * <P>A connection (session) with a specific
 * database. Within the context of a Connection, SQL statements are
 * executed and results are returned.
 *
 * <P>A Connection's database is able to provide information
 * describing its tables, its supported SQL grammar, its stored
 * procedures, the capabilities of this connection, and so on. This
 * information is obtained with the <code>getMetaData</code> method.
 *
 * <P><B>Note:</B> By default the Connection automatically commits
 * changes after executing each statement. If auto commit has been
 * disabled, an explicit commit must be done or database changes will
 * not be saved.
 *
 * @see Statement 
 * @see ResultSet
 * @see DatabaseMetaData
 */
public class ClientConnection extends ConsusConnection implements PersistentObjectStorage
{
    public static int lingerTime = 10; // linger parameter for the socket

    /**
     * Creates a <code>Statement</code> object for sending
     * SQL statements to the database.
     * SQL statements without parameters are normally
     * executed using Statement objects. If the same SQL statement 
     * is executed many times, it is more efficient to use a 
     * PreparedStatement
     *
     * JDBC 2.0
     *
     * Result sets created using the returned Statement will have
     * forward-only type, and read-only concurrency, by default.
     *
     * @return a new Statement object 
     * @exception SQLException if a database access error occurs
     */
     public Statement createStatement() throws SQLException 
     {
	 if (!opened) { 
	     throw new ConsusRuntimeError("Connection is closed");
	 }
         return new ClientStatement(this);
     }

    /**
     * JDBC 2.0
     *
     * Creates a <code>Statement</code> object that will generate
     * <code>ResultSet</code> objects with the given type and concurrency.
     * This method is the same as the <code>createStatement</code> method
     * above, but it allows the default result set
     * type and result set concurrency type to be overridden.
     *
     * @param resultSetType a result set type; see ResultSet.TYPE_XXX
     * @param resultSetConcurrency a concurrency type; see ResultSet.CONCUR_XXX
     * @return a new Statement object 
     * @exception SQLException if a database access error occurs
     */
    public Statement createStatement(int resultSetType, int resultSetConcurrency) 
      throws SQLException
    {
         return new ClientStatement(this, resultSetType, resultSetConcurrency);
    }


    /**
     * Creates a <code>PreparedStatement</code> object for sending
     * parameterized SQL statements to the database.
     * 
     * A SQL statement with or without IN parameters can be
     * pre-compiled and stored in a PreparedStatement object. This
     * object can then be used to efficiently execute this statement
     * multiple times.
     *
     * <P><B>Note:</B> This method is optimized for handling
     * parametric SQL statements that benefit from precompilation. If
     * the driver supports precompilation,
     * the method <code>prepareStatement</code> will send
     * the statement to the database for precompilation. Some drivers
     * may not support precompilation. In this case, the statement may
     * not be sent to the database until the <code>PreparedStatement</code> is
     * executed.  This has no direct effect on users; however, it does
     * affect which method throws certain SQLExceptions.
     *
     * JDBC 2.0
     *
     * Result sets created using the returned PreparedStatement will have
     * forward-only type and read-only concurrency, by default.
     *
     * @param sql a SQL statement that may contain one or more '?' IN
     * parameter placeholders
     * @return a new PreparedStatement object containing the
     * pre-compiled statement 
     * @exception SQLException if a database access error occurs
     */
     public PreparedStatement prepareStatement(String sql)
        throws SQLException
     {
	 if (!opened) { 
	     throw new ConsusRuntimeError("Connection is closed");
	 }
         return new ClientPreparedStatement(this, sql);
     }

    /**
     * JDBC 2.0
     *
     * Creates a <code>PreparedStatement</code> object that will generate
     * <code>ResultSet</code> objects with the given type and concurrency.
     * This method is the same as the <code>prepareStatement</code> method
     * above, but it allows the default result set
     * type and result set concurrency type to be overridden.
     *
     * @param resultSetType a result set type; see ResultSet.TYPE_XXX
     * @param resultSetConcurrency a concurrency type; see ResultSet.CONCUR_XXX
     * @return a new PreparedStatement object containing the
     * pre-compiled SQL statement 
     * @exception SQLException if a database access error occurs
     */
     public PreparedStatement prepareStatement(String sql, int resultSetType, 
					       int resultSetConcurrency)
       throws SQLException
     {
	 if (!opened) { 
	     throw new ConsusRuntimeError("Connection is closed");
	 }
	 return new ClientPreparedStatement(this, sql, resultSetType, resultSetConcurrency);
     }
     
     public void commitTransaction() throws SQLException
     {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	Message req = new Message();
 	req.tag = Message.REQ_COMMIT;
	req.opd1 = req.opd2 = null;
	send(req);
	receive();
    }
    
    public void prepareTransaction(int fmtid, byte[] gtrid, byte[] bqual) throws SQLException 
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	Message req = new Message();
 	req.tag = Message.REQ_PREPARE;
	req.opd1 = gtrid;
	req.opd2 = bqual;
	req.opd3 = fmtid;
	send(req);
	receive();	
    } 

    public ConsusTransId getTransId() throws SQLException
    { 
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	Message req = new Message();
 	req.tag = Message.REQ_RECOVER;
	send(req);
	Message reply = receive();
	Assert.that(reply.tag == Message.REP_RECOVERED);
	return new ConsusTransId(reply.opd3, (byte[])reply.opd1, (byte[])reply.opd2);
    }

    public void completeTransaction() throws SQLException 
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	Message req = new Message();
 	req.tag = Message.REQ_COMPLETE;
	req.opd1 = req.opd2 = null;
	send(req);
	receive();
    } 

    public DatabaseMetaData getMetaData() throws SQLException { 
	return new ClientDatabaseMetaData(this);
    } 

    /**
     * Drops all changes made since the previous
     * commit/rollback and releases any database locks currently held
     * by this Connection. This method should be used only when auto-
     * commit has been disabled.
     *
     * @exception SQLException if a database access error occurs
     * @see #setAutoCommit 
     */
    public void rollback() throws SQLException
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	rollbackTransaction();
    }

    public void rollbackTransaction() throws SQLException
    {
	Message req = new Message();
 	req.tag = Message.REQ_ROLLBACK;
	req.opd1 = req.opd2 = null;
	send(req);
	receive();
    }

    /**
     * Releases a Connection's database and JDBC resources
     * immediately instead of waiting for
     * them to be automatically released.
     *
     * <P><B>Note:</B> A Connection is automatically closed when it is
     * garbage collected. Certain fatal errors also result in a closed
     * Connection.
     *
     * @exception SQLException if a database access error occurs
     */
    public void close() throws SQLException
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	Message req = new Message();
 	req.tag = Message.REQ_CLOSE;
	req.opd1 = req.opd2 = null;
	send(req);
	receive();
	conHash.remove(in);
	try { 
	    in.close();
	    out.close();
	    socket.close();
	} catch (IOException x) {}
	opened = false;
	in = null;
	out = null;
	socket = null;
    }

    /** 
     * Commit nested subtransaction. This request release all locks set by the thread
     * and allows other thread to continue execution and see changes done by the thread.
     */
    public void commitSubtransaction() throws SQLException 
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	Message req = new Message();
 	req.tag = Message.REQ_SUBCOMMIT;
	req.opd1 = req.opd2 = null;
	send(req);
	receive();
    }


    public ClientConnection(String address, String URL, Properties info, Driver driver) 
	throws SQLException
    { 
	super(driver, URL, info);
	int pos = address.indexOf(':');
	if (pos < 0) { 
	    throw new ConsusRuntimeError("Invalid server address: " + address);
	}
	String host = address.substring(0, pos);
	int port;
	try { 
	    port = Integer.parseInt(address.substring(pos+1));
	} catch(NumberFormatException x) {
	    throw new ConsusRuntimeError("Invalid port  specification: " + address);
	}
	socket = null;
	if (host.equals("localhost")) { 
	    try { 
		socket = new LocalSocket(host, port);
	    } catch (Throwable x) {}
	} 
	if (socket == null) { 
	    try { 
		socket = new Socket(host, port);
	    } catch(IOException x) { 
		throw new ConsusRuntimeError("Failed to connect to " + address + " " + x.getMessage());
	    } 
	}
	try { 
	    socket.setTcpNoDelay(true);
	} catch (Exception x) {}
	try {
	    socket.setSoLinger(true, lingerTime);
	} catch (Exception x) {}

	try { 
	    out = new ObjectOutputStream(new BufferedOutputStream(socket.getOutputStream()));
	} catch(IOException x) { 
	    throw new ConsusRuntimeError("Failed to establish network connection: " + 
					 x.getMessage());
	}
	Message msg = new Message();
 	msg.tag = Message.REQ_LOGIN;
	msg.opd1 = info;
	send(msg);

	try { 
	    in = new ObjectInputStream(new BufferedInputStream(socket.getInputStream()));
	} catch(IOException x) { 
	    throw new ConsusRuntimeError("Failed to establish network connection: " + 
					 x.getMessage());
	}
	msg = receive();
	Assert.that(msg.tag == Message.REP_LOGIN);
	if (msg.opd1 != null && !((Boolean)msg.opd1).booleanValue()) { 
	    throw new ConsusLoginRefusedException(host, port, info);
	}
	conHash.put(in, this);
	opened = true;
    }

    void send(Message msg) throws SQLException { 
	try { 
	    out.writeObject(msg);
	    out.flush();
	    out.reset();
	} catch(IOException x) { 
	    throw new ConsusRuntimeError("Failed to send request " + x.getMessage());
	}
    }

    Message receive() throws SQLException { 
	Message reply;
	try { 
	    reply = (Message)in.readObject();
	} catch(Exception x) { 
	    throw new ConsusRuntimeError("Failed to reseive reply: " + x.getMessage());
	}
	if (reply.tag == Message.REP_ERROR) { 
	    if (reply.opd1 instanceof RuntimeException) {
		throw (RuntimeException)reply.opd1;
	    } else if (reply.opd1 instanceof SQLException) { 
		throw (SQLException)reply.opd1;
	    } else { 
		throw new ConsusRuntimeError("Exception " + reply.opd1.getClass() + " was raised at server");
	    }
	}
	return reply;
    }

    public CallableStatement prepareCall(String sql) throws SQLException 
    {
	throw new ConsusNotImplemented("Connection.prepareCall");
    } 

    public String nativeSQL(String sql) throws SQLException 
    {
        return sql;
    }

    static ClientConnection getConnectionByStream(java.io.ObjectInputStream stream)
    {
	return (ClientConnection)conHash.get(stream);
    }

    /*
     * PersistentObjectStorage methods
     */

    public Object fetchObject(ResultSet cursor) throws Exception 
    { 
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	return storage.fetchObject(cursor);
    }

    public void updateObject(Object obj, ResultSet cursor) throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	storage.updateObject(obj, cursor);
    }

    public void loadObject(Persistent obj) throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	storage.loadObject(obj);
    }

    public void loadComponents(Persistent obj) throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	storage.loadComponents(obj);
    }

    public void clearComponents(Persistent obj) throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	storage.clearComponents(obj);
    }

    public boolean isInitialized() throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	return storage.isInitialized();	
    }

    public void open() throws Exception 
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
        storage.open();	
    }


    public void storeObject(Persistent obj) throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	storage.storeObject(obj);
    }

    public void deleteObject(Persistent obj) throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	storage.deleteObject(obj);
    }

    public void becomeObject(Persistent a, Persistent b) throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	storage.becomeObject(a, b);
    }

    public void lock() throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	storage.lock();
    }

    public int getObjectCacheSize()
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	return storage.getObjectCacheSize();
    }

    public void clearObjectCache()
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	storage.clearObjectCache();
    }

    public org.garret.consus.TableIterator getTableIterator(String tableName) throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	return storage.getTableIterator(tableName);
    }

    public Object getObjectByOid(long oid) throws Exception
    {
	if (storage == null) { 
	    storage = new JDBCObjectRelationalAdapter(this);
	}
	return storage.getObjectByOid(oid);
    }

    /* 
     * Not implemented methods from ConsusConnection
     */

    public Ref insert(Object obj) throws SQLException { 
	throw new ConsusNotImplemented("ClientConnection.insert(Object)");
    }

    public void update(Ref ref, Object obj) throws SQLException { 
	throw new ConsusNotImplemented("ClientConnection.unpdate(Ref,Object)");
    }

    public Object get(Ref ref) throws SQLException { 
	throw new ConsusNotImplemented("ClientConnection.get(Ref)");
    }
	
    public void remove(Ref ref) throws SQLException { 
	throw new ConsusNotImplemented("ClientConnection.remove(ref)");
    }

    public void createObjectCache() {}

    public long getUsedSize() throws SQLException { 
	throw new ConsusNotImplemented("ClientConnection.getUsedSize()");
    }
    
    public void setGCRoots(String tableNames[]) throws SQLException { 
	throw new ConsusNotImplemented("ClientConnection.setGCRoots");
    }

    public void setGCThreshold(long threshold) throws SQLException { 
	throw new ConsusNotImplemented("ClientConnection.setGCThreshold");
    }

    public void doGarbageCollection() throws SQLException { 
	throw new ConsusNotImplemented("ClientConnection.doGarbageCollection");
    }

    public void checkOutTable(String tableName, Connection conxn) throws SQLException
    { 
	throw new ConsusNotImplemented("ClientConnection.checkOutTable");
    }

    public void checkInTable(String tableName, Connection conxn) throws SQLException
    {
	throw new ConsusNotImplemented("ClientConnection.checkInTable");
    }

    public org.garret.consus.StringIndex createStringIndex() throws SQLException
    {
	throw new ConsusNotImplemented("ClientConnection.createStringIndex");
    }

    /*
     * Instance variables
     */

    ObjectInputStream  in;
    ObjectOutputStream out;
    Socket             socket;
    
    PersistentObjectStorage storage;

    static Hashtable   conHash = new Hashtable();
}


