/* Consus - Pure Java Object Relational Database Management System
 * Implementation of DatabaseMetaData protocol for client connection. 
 * Created by Konstantin Knizhnik, Apr 2001.
 */
package org.garret.internal.consus;
import org.garret.consus.*;
import java.sql.*;

/**
 * Comprehensive information about the database as a whole.
 *
 * <P>Many of the methods here return lists of information in 
 * the form of <code>ResultSet</code> objects.
 * You can use the normal ResultSet methods such as getString and getInt 
 * to retrieve the data from these ResultSets.  If a given form of
 * metadata is not available, these methods should throw an SQLException.
 *
 * <P>Some of these methods take arguments that are String patterns.  These
 * arguments all have names such as fooPattern.  Within a pattern String, "%"
 * means match any substring of 0 or more characters, and "_" means match
 * any one character. Only metadata entries matching the search pattern 
 * are returned. If a search pattern argument is set to a null ref, 
 * that argument's criteria will be dropped from the search.
 * 
 * <P>An <code>SQLException</code> will be thrown if a driver does not support a meta
 * data method.  In the case of methods that return a ResultSet,
 * either a ResultSet (which may be empty) is returned or a
 * SQLException is thrown.
 */
public class ClientDatabaseMetaData extends ConsusDatabaseMetaData {
    /**
     * What's the url for this database?
     *
     * @return the url or null if it cannot be generated
     * @exception SQLException if a database access error occurs
     */
    public String getURL() throws SQLException { 
	return db.URL;
    }

    /**
     * What's the name of this JDBC driver?
     *
     * @return JDBC driver name
     * @exception SQLException if a database access error occurs
     */
    public String getDriverName() throws SQLException { 
	return db.driver.getClass().getName();
    }

    /**
     * What's this JDBC driver's major version number?
     *
     * @return JDBC driver major version
     */
    public int getDriverMajorVersion() { 
	return db.driver.getMajorVersion();
    }

    /**
     * What's this JDBC driver's minor version number?
     *
     * @return JDBC driver minor version number
     */
    public int getDriverMinorVersion() { 
	return db.driver.getMinorVersion();
    }

    /**
     * What's our user name as known to the database?
     *
     * @return our database user name
     * @exception SQLException if a database access error occurs
     */
    public String getUserName() throws SQLException { 
	return db.properties.getProperty("user", "public");
    }


    /**
     * Does the database store tables in a local file?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean usesLocalFiles() throws SQLException { 
	return false;
    }
    
    /**
     * Gets a description of tables available in a catalog.
     *
     * <P>Only table descriptions matching the catalog, schema, table
     * name and type criteria are returned.  They are ordered by
     * TABLE_TYPE, TABLE_SCHEM and TABLE_NAME.
     *
     * <P>Each table description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>TABLE_TYPE</B> String => table type.  Typical types are "TABLE",
     *			"VIEW",	"SYSTEM TABLE", "GLOBAL TEMPORARY", 
     *			"LOCAL TEMPORARY", "ALIAS", "SYNONYM".
     *	<LI><B>REMARKS</B> String => explanatory comment on the table
     *  </OL>
     *
     * <P><B>Note:</B> Some databases may not return information for
     * all tables.
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those
     * without a schema
     * @param tableNamePattern a table name pattern 
     * @param types a list of table types to include; null returns all types 
     * @return ResultSet - each row is a table description
     * @exception SQLException if a database access error occurs
     * @see #getSearchStringEscape 
     */
    public ResultSet getTables(String catalog, String schemaPattern,
			       String tableNamePattern, String types[]) throws SQLException
    {
	Message req = new Message();
	req.tag = Message.REQ_TABLES;
	req.opd1 = tableNamePattern;
	req.opd2 = types;
	db.send(req);
	Message reply = db.receive();
	ColumnNode columns[] = new ColumnNode[5];
	for (int i = 0; i < 5; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}
	columns[0].name = "TABLE_CAT";
	columns[1].name = "TABLE_SCHEM";
	columns[2].name = "TABLE_NAME";
	columns[3].name = "TABLE_TYPE";
	columns[4].name = "REMARKS";
	ConsusResultSet cursor = new ConsusResultSet(null);
	cursor.tuples = (Object[][])reply.opd1;
	cursor.tupleColumns = columns;
	cursor.nSelected = cursor.tuples.length;
	return cursor;
    }

    /**
     * Gets a description of table columns available in 
     * the specified catalog.
     *
     * <P>Only column descriptions matching the catalog, schema, table
     * and column name criteria are returned.  They are ordered by
     * TABLE_SCHEM, TABLE_NAME and ORDINAL_POSITION.
     *
     * <P>Each column description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>COLUMN_NAME</B> String => column name
     *	<LI><B>DATA_TYPE</B> short => SQL type from java.sql.Types
     *	<LI><B>TYPE_NAME</B> String => Data source dependent type name,
     *  for a UDT the type name is fully qualified
     *	<LI><B>COLUMN_SIZE</B> int => column size.  For char or date
     *	    types this is the maximum number of characters, for numeric or
     *	    decimal types this is precision.
     *	<LI><B>BUFFER_LENGTH</B> is not used.
     *	<LI><B>DECIMAL_DIGITS</B> int => the number of fractional digits
     *	<LI><B>NUM_PREC_RADIX</B> int => Radix (typically either 10 or 2)
     *	<LI><B>NULLABLE</B> int => is NULL allowed?
     *      <UL>
     *      <LI> columnNoNulls - might not allow NULL values
     *      <LI> columnNullable - definitely allows NULL values
     *      <LI> columnNullableUnknown - nullability unknown
     *      </UL>
     *	<LI><B>REMARKS</B> String => comment describing column (may be null)
     * 	<LI><B>COLUMN_DEF</B> String => default value (may be null)
     *	<LI><B>SQL_DATA_TYPE</B> int => unused
     *	<LI><B>SQL_DATETIME_SUB</B> int => unused
     *	<LI><B>CHAR_OCTET_LENGTH</B> int => for char types the 
     *       maximum number of bytes in the column
     *	<LI><B>ORDINAL_POSITION</B> int	=> index of column in table 
     *      (starting at 1)
     *	<LI><B>IS_NULLABLE</B> String => "NO" means column definitely 
     *      does not allow NULL values; "YES" means the column might 
     *      allow NULL values.  An empty string means nobody knows.
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those
     * without a schema
     * @param tableNamePattern a table name pattern 
     * @param columnNamePattern a column name pattern 
     * @return ResultSet - each row is a column description
     * @exception SQLException if a database access error occurs
     * @see #getSearchStringEscape 
     */
    public ResultSet getColumns(String catalog, String schemaPattern,
				String tableNamePattern, String columnNamePattern)
	throws SQLException
    {
	Message req = new Message();
	req.tag = Message.REQ_COLUMNS;
	req.opd1 = tableNamePattern;
	req.opd2 = columnNamePattern;
	db.send(req);
	Message reply = db.receive();
	ConsusResultSet cursor = new ConsusResultSet(null);
	cursor.tuples = (Object[][])reply.opd1;
	cursor.nSelected = cursor.tuples.length;
	ColumnNode columns[] = new ColumnNode[18];
	for (int i = 0; i < 18; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}
	columns[0].name = "TABLE_CAT";
	columns[1].name = "TABLE_SCHEM";
	columns[2].name = "TABLE_NAME";
	columns[3].name = "COLUMN_NAME";
	columns[4].name = "DATA_TYPE";
	columns[4].type = ExprNode.tpInt;
	columns[5].name = "TYPE_NAME";
	columns[6].name = "COLUMN_SIZE";
	columns[6].type = ExprNode.tpInt;
	columns[7].name = "BUFFER_LENGTH";
	columns[7].type = ExprNode.tpInt;
	columns[8].name = "DECIMAL_DIGITS";
	columns[8].type = ExprNode.tpInt;
	columns[9].name = "NUM_PREC_RADIX";
	columns[9].type = ExprNode.tpInt;
	columns[10].name = "NULLABLE";
	columns[10].type = ExprNode.tpInt;
	columns[11].name = "REMARKS";
	columns[12].name = "COLUMN_DEF";
	columns[13].name = "SQL_DATA_TYPE";
	columns[13].type = ExprNode.tpInt;
	columns[14].name = "SQL_DATETIME_SUB";
	columns[14].type = ExprNode.tpInt;
	columns[15].name = "CHAR_OCTET_LENGTH";
	columns[15].type = ExprNode.tpInt;
	columns[16].name = "ORDINAL_POSITION";
	columns[16].type = ExprNode.tpInt;
	columns[17].name = "IS_NULLABLE";
	cursor.tupleColumns = columns;
	return cursor;	
    }


    /**
     * Gets a description of a table's primary key columns.  They
     * are ordered by COLUMN_NAME.
     *
     * <P>Each primary key column description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>COLUMN_NAME</B> String => column name
     *	<LI><B>KEY_SEQ</B> short => sequence number within primary key
     *	<LI><B>PK_NAME</B> String => primary key name (may be null)
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those
     * without a schema
     * @param table a table name
     * @return ResultSet - each row is a primary key column description 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getPrimaryKeys(String catalog, String schema,
				    String tableName) throws SQLException
    {
	Message req = new Message();
	req.tag = Message.REQ_PRIMARY_KEYS;
	req.opd1 = tableName;
	req.opd2 = null;
	db.send(req);
	Message reply = db.receive();
	ConsusResultSet cursor = new ConsusResultSet(null);
	cursor.tuples = (Object[][])reply.opd1;
	cursor.nSelected = cursor.tuples.length;
	ColumnNode columns[] = new ColumnNode[6];
	for (int i = 0; i < 6; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}	
	columns[0].name = "TABLE_CAT";
	columns[1].name = "TABLE_SCHEM";
	columns[2].name = "TABLE_NAME";
	columns[3].name = "COLUMN_NAME";
	columns[4].name = "KEY_SEQ";
	columns[5].name = "PK_NAME";
	cursor.tupleColumns = columns;
	return cursor;
    }


    /**
     * Gets a description of the primary key columns that are
     * referenced by a table's foreign key columns (the primary keys
     * imported by a table).  They are ordered by PKTABLE_CAT,
     * PKTABLE_SCHEM, PKTABLE_NAME, and KEY_SEQ.
     *
     * <P>Each primary key column description has the following columns:
     *  <OL>
     *	<LI><B>PKTABLE_CAT</B> String => primary key table catalog 
     *      being imported (may be null)
     *	<LI><B>PKTABLE_SCHEM</B> String => primary key table schema
     *      being imported (may be null)
     *	<LI><B>PKTABLE_NAME</B> String => primary key table name
     *      being imported
     *	<LI><B>PKCOLUMN_NAME</B> String => primary key column name
     *      being imported
     *	<LI><B>FKTABLE_CAT</B> String => foreign key table catalog (may be null)
     *	<LI><B>FKTABLE_SCHEM</B> String => foreign key table schema (may be null)
     *	<LI><B>FKTABLE_NAME</B> String => foreign key table name
     *	<LI><B>FKCOLUMN_NAME</B> String => foreign key column name
     *	<LI><B>KEY_SEQ</B> short => sequence number within foreign key
     *	<LI><B>UPDATE_RULE</B> short => What happens to 
     *       foreign key when primary is updated:
     *      <UL>
     *      <LI> importedNoAction - do not allow update of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - change imported key to agree 
     *               with primary key update
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been updated
     *      <LI> importedKeySetDefault - change imported key to default values 
     *               if its primary key has been updated
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      </UL>
     *	<LI><B>DELETE_RULE</B> short => What happens to 
     *      the foreign key when primary is deleted.
     *      <UL>
     *      <LI> importedKeyNoAction - do not allow delete of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - delete rows that import a deleted key
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been deleted
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      <LI> importedKeySetDefault - change imported key to default if 
     *               its primary key has been deleted
     *      </UL>
     *	<LI><B>FK_NAME</B> String => foreign key name (may be null)
     *	<LI><B>PK_NAME</B> String => primary key name (may be null)
     *	<LI><B>DEFERRABILITY</B> short => can the evaluation of foreign key 
     *      constraints be deferred until commit
     *      <UL>
     *      <LI> importedKeyInitiallyDeferred - see SQL92 for definition
     *      <LI> importedKeyInitiallyImmediate - see SQL92 for definition 
     *      <LI> importedKeyNotDeferrable - see SQL92 for definition 
     *      </UL>
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those
     * without a schema
     * @param table a table name
     * @return ResultSet - each row is a primary key column description 
     * @exception SQLException if a database access error occurs
     * @see #getExportedKeys 
     */
    public ResultSet getImportedKeys(String catalog, String schema,
				     String tableName) throws SQLException
    {
	Message req = new Message();
	req.tag = Message.REQ_IMPORTED_KEYS;
	req.opd1 = tableName;
	req.opd2 = null;
	db.send(req);
	Message reply = db.receive();
	ConsusResultSet cursor = new ConsusResultSet(null);
	cursor.tuples = (Object[][])reply.opd1;
	cursor.nSelected = cursor.tuples.length;
	ColumnNode columns[] = new ColumnNode[14];
	for (int i = 0; i < 14; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}	
	columns[0].name = "PKTABLE_CAT";
	columns[1].name = "PKTABLE_SCHEM";
	columns[2].name = "PKTABLE_NAME";
	columns[3].name = "PKCOLUMN_NAME";
	columns[4].name = "FKTABLE_CAT";
	columns[5].name = "FKTABLE_SCHEM";
	columns[6].name = "FKTABLE_NAME";
	columns[7].name = "FKCOLUMN_NAME";
	columns[8].name = "KEY_SEQ";
	columns[9].name = "UPDATE_RULE";
	columns[10].name = "DELETE_RULE";
	columns[11].name = "FK_NAME";
	columns[12].name = "PK_NAME";
	columns[13].name = "DEFERRABILITY";
	cursor.tupleColumns = columns;
	return cursor;
    }


    /**
     * Gets a description of the foreign key columns that reference a
     * table's primary key columns (the foreign keys exported by a
     * table).  They are ordered by FKTABLE_CAT, FKTABLE_SCHEM,
     * FKTABLE_NAME, and KEY_SEQ.
     *
     * <P>Each foreign key column description has the following columns:
     *  <OL>
     *	<LI><B>PKTABLE_CAT</B> String => primary key table catalog (may be null)
     *	<LI><B>PKTABLE_SCHEM</B> String => primary key table schema (may be null)
     *	<LI><B>PKTABLE_NAME</B> String => primary key table name
     *	<LI><B>PKCOLUMN_NAME</B> String => primary key column name
     *	<LI><B>FKTABLE_CAT</B> String => foreign key table catalog (may be null)
     *      being exported (may be null)
     *	<LI><B>FKTABLE_SCHEM</B> String => foreign key table schema (may be null)
     *      being exported (may be null)
     *	<LI><B>FKTABLE_NAME</B> String => foreign key table name
     *      being exported
     *	<LI><B>FKCOLUMN_NAME</B> String => foreign key column name
     *      being exported
     *	<LI><B>KEY_SEQ</B> short => sequence number within foreign key
     *	<LI><B>UPDATE_RULE</B> short => What happens to 
     *       foreign key when primary is updated:
     *      <UL>
     *      <LI> importedNoAction - do not allow update of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - change imported key to agree 
     *               with primary key update
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been updated
     *      <LI> importedKeySetDefault - change imported key to default values 
     *               if its primary key has been updated
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      </UL>
     *	<LI><B>DELETE_RULE</B> short => What happens to 
     *      the foreign key when primary is deleted.
     *      <UL>
     *      <LI> importedKeyNoAction - do not allow delete of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - delete rows that import a deleted key
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been deleted
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      <LI> importedKeySetDefault - change imported key to default if 
     *               its primary key has been deleted
     *      </UL>
     *	<LI><B>FK_NAME</B> String => foreign key name (may be null)
     *	<LI><B>PK_NAME</B> String => primary key name (may be null)
     *	<LI><B>DEFERRABILITY</B> short => can the evaluation of foreign key 
     *      constraints be deferred until commit
     *      <UL>
     *      <LI> importedKeyInitiallyDeferred - see SQL92 for definition
     *      <LI> importedKeyInitiallyImmediate - see SQL92 for definition 
     *      <LI> importedKeyNotDeferrable - see SQL92 for definition 
     *      </UL>
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those
     * without a schema
     * @param table a table name
     * @return ResultSet - each row is a foreign key column description 
     * @exception SQLException if a database access error occurs
     * @see #getImportedKeys 
     */
    public ResultSet getExportedKeys(String catalog, String schema,
				     String tableName) throws SQLException    
    {
	Message req = new Message();
	req.tag = Message.REQ_EXPORTED_KEYS;
	req.opd1 = tableName;
	req.opd2 = null;
	db.send(req);
	Message reply = db.receive();
	ConsusResultSet cursor = new ConsusResultSet(null);
	cursor.tuples = (Object[][])reply.opd1;
	cursor.nSelected = cursor.tuples.length;
	ColumnNode columns[] = new ColumnNode[14];
	for (int i = 0; i < 14; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}	
	columns[0].name = "PKTABLE_CAT";
	columns[1].name = "PKTABLE_SCHEM";
	columns[2].name = "PKTABLE_NAME";
	columns[3].name = "PKCOLUMN_NAME";
	columns[4].name = "FKTABLE_CAT";
	columns[5].name = "FKTABLE_SCHEM";
	columns[6].name = "FKTABLE_NAME";
	columns[7].name = "FKCOLUMN_NAME";
	columns[8].name = "KEY_SEQ";
	columns[9].name = "UPDATE_RULE";
	columns[10].name = "DELETE_RULE";
	columns[11].name = "FK_NAME";
	columns[12].name = "PK_NAME";
	columns[13].name = "DEFERRABILITY";
	cursor.tupleColumns = columns;
	return cursor;
    }

    /**
     * Gets a description of the foreign key columns in the foreign key
     * table that reference the primary key columns of the primary key
     * table (describe how one table imports another's key.) This
     * should normally return a single foreign key/primary key pair
     * (most tables only import a foreign key from a table once.)  They
     * are ordered by FKTABLE_CAT, FKTABLE_SCHEM, FKTABLE_NAME, and
     * KEY_SEQ.
     *
     * <P>Each foreign key column description has the following columns:
     *  <OL>
     *	<LI><B>PKTABLE_CAT</B> String => primary key table catalog (may be null)
     *	<LI><B>PKTABLE_SCHEM</B> String => primary key table schema (may be null)
     *	<LI><B>PKTABLE_NAME</B> String => primary key table name
     *	<LI><B>PKCOLUMN_NAME</B> String => primary key column name
     *	<LI><B>FKTABLE_CAT</B> String => foreign key table catalog (may be null)
     *      being exported (may be null)
     *	<LI><B>FKTABLE_SCHEM</B> String => foreign key table schema (may be null)
     *      being exported (may be null)
     *	<LI><B>FKTABLE_NAME</B> String => foreign key table name
     *      being exported
     *	<LI><B>FKCOLUMN_NAME</B> String => foreign key column name
     *      being exported
     *	<LI><B>KEY_SEQ</B> short => sequence number within foreign key
     *	<LI><B>UPDATE_RULE</B> short => What happens to 
     *       foreign key when primary is updated:
     *      <UL>
     *      <LI> importedNoAction - do not allow update of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - change imported key to agree 
     *               with primary key update
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been updated
     *      <LI> importedKeySetDefault - change imported key to default values 
     *               if its primary key has been updated
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      </UL>
     *	<LI><B>DELETE_RULE</B> short => What happens to 
     *      the foreign key when primary is deleted.
     *      <UL>
     *      <LI> importedKeyNoAction - do not allow delete of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - delete rows that import a deleted key
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been deleted
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      <LI> importedKeySetDefault - change imported key to default if 
     *               its primary key has been deleted
     *      </UL>
     *	<LI><B>FK_NAME</B> String => foreign key name (may be null)
     *	<LI><B>PK_NAME</B> String => primary key name (may be null)
     *	<LI><B>DEFERRABILITY</B> short => can the evaluation of foreign key 
     *      constraints be deferred until commit
     *      <UL>
     *      <LI> importedKeyInitiallyDeferred - see SQL92 for definition
     *      <LI> importedKeyInitiallyImmediate - see SQL92 for definition 
     *      <LI> importedKeyNotDeferrable - see SQL92 for definition 
     *      </UL>
     *  </OL>
     *
     * @param primaryCatalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param primarySchema a schema name; "" retrieves those
     * without a schema
     * @param primaryTable the table name that exports the key
     * @param foreignCatalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param foreignSchema a schema name; "" retrieves those
     * without a schema
     * @param foreignTable the table name that imports the key
     * @return ResultSet - each row is a foreign key column description 
     * @exception SQLException if a database access error occurs
     * @see #getImportedKeys 
     */
    public ResultSet getCrossReference(String primaryCatalog, String primarySchema, String primaryTable,
				       String foreignCatalog, String foreignSchema, String foreignTable) 
	throws SQLException
    {
	Message req = new Message();
	req.tag = Message.REQ_CROSSREFS;
	req.opd1 = primaryTable;
	req.opd2 = foreignTable;
	db.send(req);
	Message reply = db.receive();
	ConsusResultSet cursor = new ConsusResultSet(null);
	cursor.tuples = (Object[][])reply.opd1;
	cursor.nSelected = cursor.tuples.length;
	ColumnNode columns[] = new ColumnNode[14];
	for (int i = 0; i < 14; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}	
	columns[0].name = "PKTABLE_CAT";
	columns[1].name = "PKTABLE_SCHEM";
	columns[2].name = "PKTABLE_NAME";
	columns[3].name = "PKCOLUMN_NAME";
	columns[4].name = "FKTABLE_CAT";
	columns[5].name = "FKTABLE_SCHEM";
	columns[6].name = "FKTABLE_NAME";
	columns[7].name = "FKCOLUMN_NAME";
	columns[8].name = "KEY_SEQ";
	columns[9].name = "UPDATE_RULE";
	columns[10].name = "DELETE_RULE";
	columns[11].name = "FK_NAME";
	columns[12].name = "PK_NAME";
	columns[13].name = "DEFERRABILITY";
	cursor.tupleColumns = columns;
	return cursor;
    }

    /**
     * Gets a description of a table's indices and statistics. They are
     * ordered by NON_UNIQUE, TYPE, INDEX_NAME, and ORDINAL_POSITION.
     *
     * <P>Each index column description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>NON_UNIQUE</B> boolean => Can index values be non-unique? 
     *      false when TYPE is tableIndexStatistic
     *	<LI><B>INDEX_QUALIFIER</B> String => index catalog (may be null); 
     *      null when TYPE is tableIndexStatistic
     *	<LI><B>INDEX_NAME</B> String => index name; null when TYPE is 
     *      tableIndexStatistic
     *	<LI><B>TYPE</B> short => index type:
     *      <UL>
     *      <LI> tableIndexStatistic - this identifies table statistics that are
     *           returned in conjuction with a table's index descriptions
     *      <LI> tableIndexClustered - this is a clustered index
     *      <LI> tableIndexHashed - this is a hashed index
     *      <LI> tableIndexOther - this is some other style of index
     *      </UL>
     *	<LI><B>ORDINAL_POSITION</B> short => column sequence number 
     *      within index; zero when TYPE is tableIndexStatistic
     *	<LI><B>COLUMN_NAME</B> String => column name; null when TYPE is 
     *      tableIndexStatistic
     *	<LI><B>ASC_OR_DESC</B> String => column sort sequence, "A" => ascending, 
     *      "D" => descending, may be null if sort sequence is not supported; 
     *      null when TYPE is tableIndexStatistic	
     *	<LI><B>CARDINALITY</B> int => When TYPE is tableIndexStatistic, then 
     *      this is the number of rows in the table; otherwise, it is the 
     *      number of unique values in the index.
     *	<LI><B>PAGES</B> int => When TYPE is  tableIndexStatisic then 
     *      this is the number of pages used for the table, otherwise it 
     *      is the number of pages used for the current index.
     *	<LI><B>FILTER_CONDITION</B> String => Filter condition, if any.  
     *      (may be null)
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those without a schema
     * @param table a table name  
     * @param unique when true, return only indices for unique values; 
     *     when false, return indices regardless of whether unique or not 
     * @param approximate when true, result is allowed to reflect approximate 
     *     or out of data values; when false, results are requested to be 
     *     accurate
     * @return ResultSet - each row is an index column description 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getIndexInfo(String catalog, String schema, String tableName,
				  boolean unique, boolean approximate)
	throws SQLException
    {
	Message req = new Message();
	req.tag = Message.REQ_INDICES;
	req.opd1 = tableName;
	req.opd2 = null;
	req.opd3 = unique ? 1 : 0;
	db.send(req);
	Message reply = db.receive();
	ConsusResultSet cursor = new ConsusResultSet(null);
	cursor.tuples = (Object[][])reply.opd1;
	cursor.nSelected = cursor.tuples.length;
	ColumnNode columns[] = new ColumnNode[13];
	for (int i = 0; i < 13; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}
	columns[0].name = "TABLE_CAT";
	columns[1].name = "TABLE_SCHEM";
	columns[2].name = "TABLE_NAME";
	columns[3].name = "NON_UNIQUE";
	columns[3].type = ExprNode.tpBool;
	columns[4].name = "INDEX_QUALIFIER";
	columns[5].name = "INDEX_NAME";
	columns[6].name = "TYPE";
	columns[6].type = ExprNode.tpInt;
	columns[7].name = "ORDINAL_POSITION";
	columns[7].type = ExprNode.tpInt;
	columns[8].name = "COLUMN_NAME";
	columns[9].name = "ASC_OR_DESC";
	columns[10].name = "CARDINALITY";
	columns[10].type = ExprNode.tpInt;
	columns[11].name = "PAGES";
	columns[11].type = ExprNode.tpInt;
	columns[12].name = "FILTER_CONDITION";
	cursor.tupleColumns = columns;
	return cursor;
    }


    /**
     * JDBC 2.0
	 * Retrieves the connection that produced this metadata object.
     *
     * @return the connection that produced this metadata object
     */
    public Connection getConnection() throws SQLException { 
	return db;
    }

    ClientDatabaseMetaData(ClientConnection conxn) { 
	super(conxn);
	db = conxn;
    }

    /** 
     * Instance variables 
     */
    ClientConnection db;
}





