/* Consus - Pure Java Object Relational Database Management System
 * PreparedStatement protocol implementation
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;
import java.io.*;
import java.math.BigDecimal;
import java.util.Calendar;

/**
 * An object that represents a precompiled SQL statement.
 * <P>A SQL statement is pre-compiled and stored in a
 * PreparedStatement object. This object can then be used to
 * efficiently execute this statement multiple times. 
 *
 * <P><B>Note:</B> The setXXX methods for setting IN parameter values
 * must specify types that are compatible with the defined SQL type of
 * the input parameter. For instance, if the IN parameter has SQL type
 * Integer, then the method <code>setInt</code> should be used.
 *
 * <p>If arbitrary parameter type conversions are required, the method
 * <code>setObject</code> should be used with a target SQL type.
 * <br>
 * Example of setting a parameter; <code>con</code> is an active connection  
 * <pre><code>
 *   PreparedStatement pstmt = con.prepareStatement("UPDATE EMPLOYEES
 *                                     SET SALARY = ? WHERE ID = ?");
 *   pstmt.setBigDecimal(1, 153833.00)
 *   pstmt.setInt(2, 110592)
 * </code></pre>
 *
 * @see Connection#prepareStatement
 * @see ResultSet 
 */
public class ClientPreparedStatement extends ConsusPreparedStatement
{
    Message          req = new Message();
    ClientConnection conxn;
    String           sql;
    Object[]         reqParams;

    ColumnNode[] getColumns(Object arg) { 
	ColumnInfo[] cinfo = (ColumnInfo[])arg;
	int n = cinfo.length;
	ColumnNode columns[] = new ColumnNode[n];
	while (-- n >= 0) { 
	    columns[n] = new ColumnNode(cinfo[n].type, cinfo[n].name, n+1);
	}
	return columns;
    }

    /**
     * Executes the SQL query in this <code>PreparedStatement</code> object
     * and returns the result set generated by the query.
     *
     * @return a ResultSet that contains the data produced by the
     * query; never null
     * @exception SQLException if a database access error occurs
     */
    public ResultSet executeQuery() throws SQLException
    {
	req.tag = Message.REQ_EXEC_QUERY_PARAM;
	req.opd1 = sql;
	req.opd2 = reqParams;
	req.opd3 = cursor.selectionLimit;
	conxn.send(req);
	Message reply = conxn.receive();
	cursor.tuples = (Object[][])reply.opd1;
	cursor.tupleColumns = getColumns(reply.opd2);
	cursor.nSelected = reply.opd3;
	oid = reply.opd4;
	cursor.currPos = -1;
	isQuery = true;
	return cursor;
    }


    /**
     * Executes the SQL INSERT, UPDATE or DELETE statement
     * in this <code>PreparedStatement</code> object.
     * In addition,
     * SQL statements that return nothing, such as SQL DDL statements,
     * can be executed.
     *
     * @return either the row count for INSERT, UPDATE or DELETE statements;
     * or 0 for SQL statements that return nothing
     * @exception SQLException if a database access error occurs
     */
    public int executeUpdate() throws SQLException { 
	req.tag = Message.REQ_EXEC_UPDATE_PARAM;
	req.opd1 = sql;
	req.opd2 = reqParams;
	conxn.send(req);
	Message reply = conxn.receive();
	isQuery = false;
	oid = reply.opd4;
	return updateCount = reply.opd3;
    }

    void setParameter(int index, Object obj) {
	if (reqParams == null || index > reqParams.length) { 
	    Object[] newParams = new Object[index];
	    if (reqParams != null) { 
		System.arraycopy(reqParams, 0, newParams, 0, reqParams.length);
	    }
	    reqParams = newParams;
	}
	reqParams[index-1] = obj;
    }
    
    /**
     * Sets the designated parameter to SQL NULL.
     *
     * <P><B>Note:</B> You must specify the parameter's SQL type.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param sqlType the SQL type code defined in java.sql.Types
     * @exception SQLException if a database access error occurs
     */
    public void setNull(int parameterIndex, int sqlType) throws SQLException { 	
	setParameter(parameterIndex, new ConsusRef(0));
    }

    /**
     * Sets the designated parameter to a Java boolean value.  The driver converts this
     * to an SQL BIT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setBoolean(int parameterIndex, boolean x) throws SQLException {
	setParameter(parameterIndex, new Boolean(x));
    }

    /**
     * Sets the designated parameter to a Java byte value.  The driver converts this
     * to an SQL TINYINT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setByte(int parameterIndex, byte x) throws SQLException {
	setParameter(parameterIndex, new Byte(x));
    }

    /**
     * Sets the designated parameter to a Java short value.  The driver converts this
     * to an SQL SMALLINT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setShort(int parameterIndex, short x) throws SQLException {
	setParameter(parameterIndex, new Short(x));
    }

    /**
     * Sets the designated parameter to a Java int value.  The driver converts this
     * to an SQL INTEGER value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setInt(int parameterIndex, int x) throws SQLException {
	setParameter(parameterIndex, new Integer(x));
    }

    /**
     * Sets the designated parameter to a Java long value.  The driver converts this
     * to an SQL BIGINT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setLong(int parameterIndex, long x) throws SQLException { 
	setParameter(parameterIndex, new Long(x));
    }


    /**
     * Sets the designated parameter to a Java float value.  The driver converts this
     * to an SQL FLOAT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setFloat(int parameterIndex, float x) throws SQLException {
	setParameter(parameterIndex, new Float(x));
    }


    /**
     * Sets the designated parameter to a Java double value.  The driver converts this
     * to an SQL DOUBLE value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setDouble(int parameterIndex, double x) throws SQLException {
	setParameter(parameterIndex, new Double(x));
    }

    /**
     * Sets the designated parameter to a java.lang.BigDecimal value.  
     * The driver converts this to an SQL NUMERIC value when
     * it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setBigDecimal(int parameterIndex, BigDecimal x) throws SQLException
    {
	setParameter(parameterIndex, x);
    }


    /**
     * Sets the designated parameter to a Java String value.  The driver converts this
     * to an SQL VARCHAR or LONGVARCHAR value (depending on the argument's
     * size relative to the driver's limits on VARCHARs) when it sends
     * it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setString(int parameterIndex, String x) throws SQLException { 
	setParameter(parameterIndex, x);
    }
	

    /**
     * Sets the designated parameter to a Java array of bytes.  The driver converts
     * this to an SQL VARBINARY or LONGVARBINARY (depending on the
     * argument's size relative to the driver's limits on VARBINARYs)
     * when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value 
     * @exception SQLException if a database access error occurs
     */
    public void setBytes(int parameterIndex, byte x[]) throws SQLException {
	setParameter(parameterIndex, x);
    }

    /**
     * Sets the designated parameter to a java.sql.Date value.  The driver converts this
     * to an SQL DATE value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setDate(int parameterIndex, java.sql.Date x)
        throws SQLException 
    {
	setParameter(parameterIndex, x);
    }

    /**
     * Sets the designated parameter to a java.sql.Time value.  The driver converts this
     * to an SQL TIME value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setTime(int parameterIndex, java.sql.Time x) 
        throws SQLException
    {
	setParameter(parameterIndex, x);
    }

    /**
     * Sets the designated parameter to a java.sql.Timestamp value.  The driver
     * converts this to an SQL TIMESTAMP value when it sends it to the
     * database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value 
     * @exception SQLException if a database access error occurs
     */
    public void setTimestamp(int parameterIndex, java.sql.Timestamp x)
        throws SQLException
    {
	setParameter(parameterIndex, x);
    }

    /**
     * Sets the designated parameter to the given input stream, which will have 
     * the specified number of bytes.
     * When a very large binary value is input to a LONGVARBINARY
     * parameter, it may be more practical to send it via a
     * java.io.InputStream. JDBC will read the data from the stream
     * as needed, until it reaches end-of-file.
     * 
     * <P><B>Note:</B> This stream object can either be a standard
     * Java stream object or your own subclass that implements the
     * standard interface.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the java input stream which contains the binary parameter value
     * @param length the number of bytes in the stream 
     * @exception SQLException if a database access error occurs
     */
    public void setBinaryStream(int parameterIndex, java.io.InputStream stream, 
				int length) throws SQLException
    {
	byte buf[] = new byte[length];
	int offs = 0;
	try { 
	    while (offs < length) {
		offs += stream.read(buf, offs, length - offs);
	    }
	} catch(IOException x) { 
	    throw new ConsusRuntimeError("Failed to read data from input stream: " 
					 + x.getMessage());
	}
	setParameter(parameterIndex, buf);
    }



    /**
     * Clears the current parameter values immediately.
     * <P>In general, parameter values remain in force for repeated use of a
     * Statement. Setting a parameter value automatically clears its
     * previous value.  However, in some cases it is useful to immediately
     * release the resources used by the current parameter values; this can
     * be done by calling clearParameters.
     *
     * @exception SQLException if a database access error occurs
     */
    public void clearParameters() throws SQLException {
	reqParams = null;
    }

    /**
     * <p>Sets the value of a parameter using an object; use the
     * java.lang equivalent objects for integral values.
     *
     * <p>The JDBC specification specifies a standard mapping from
     * Java Object types to SQL types.  The given argument java object
     * will be converted to the corresponding SQL type before being
     * sent to the database.
     *
     * <p>Note that this method may be used to pass datatabase-
     * specific abstract data types, by using a Driver-specific Java
     * type.
     *
     * If the object is of a class implementing SQLData,
     * the JDBC driver should call its method <code>writeSQL</code> to write it 
     * to the SQL data stream.
     * If, on the other hand, the object is of a class implementing
     * Ref, Blob, Clob, Struct, 
     * or Array, then the driver should pass it to the database as a value of the 
     * corresponding SQL type.
     *
     * This method throws an exception if there is an ambiguity, for example, if the
     * object is of a class implementing more than one of those interfaces.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the object containing the input parameter value 
     * @exception SQLException if a database access error occurs
     */
    public void setObject(int parameterIndex, Object x) throws SQLException
    {
	setParameter(parameterIndex, x);	
    }

    /**
     * Executes any kind of SQL statement.
     * Some prepared statements return multiple results; the execute
     * method handles these complex statements as well as the simpler
     * form of statements handled by executeQuery and executeUpdate.
     *
     * @exception SQLException if a database access error occurs
     * @see Statement#execute
     */
    public boolean execute() throws SQLException { 
	req.tag = Message.REQ_EXEC_PARAM;
	req.opd1 = sql;
	req.opd2 = reqParams;
	req.opd3 = cursor.selectionLimit;
	conxn.send(req);
	Message reply = conxn.receive();
	if (reply.tag == Message.REP_QUERY) { 
	    cursor.tuples = (Object[][])reply.opd1;
	    cursor.tupleColumns = getColumns(reply.opd2);
	    cursor.nSelected = reply.opd3;
	    oid = reply.opd4;
	    cursor.currPos = -1;
	    isQuery = true;
	    return true;
	} else { 
	    oid = reply.opd4;
	    updateCount = reply.opd3;
	    oid = reply.opd4;
	    isQuery = false;
	    return false;
	}
    }

    /**
     * JDBC 2.0
     *
     * Sets the designated parameter to the given <code>Reader</code>
     * object, which is the given number of characters long.
     * When a very large UNICODE value is input to a LONGVARCHAR
     * parameter, it may be more practical to send it via a
     * java.io.Reader. JDBC will read the data from the stream
     * as needed, until it reaches end-of-file.  The JDBC driver will
     * do any necessary conversion from UNICODE to the database char format.
     * 
     * <P><B>Note:</B> This stream object can either be a standard
     * Java stream object or your own subclass that implements the
     * standard interface.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the java reader which contains the UNICODE data
     * @param length the number of characters in the stream 
     * @exception SQLException if a database access error occurs
     */
    public void setCharacterStream(int parameterIndex,
				   java.io.Reader reader,
				   int length) throws SQLException
    {
	char[] buf = new char[length];
	int offs = 0;
	try { 
	    while (offs < length) {
		offs += reader.read(buf, offs, length - offs);
	    }
	} catch (java.io.IOException x) { 
	    throw new ConsusRuntimeError("Failed to read from stream: " + 
					 x.getMessage());
	}
	setParameter(parameterIndex, new String(buf));
    }

    /**
     * JDBC 2.0
     *
     * Sets a REF(&lt;structured-type&gt;) parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing data of an SQL REF Type
     * @exception SQLException if a database access error occurs
     */
    public void setRef (int i, Ref x) throws SQLException { 
	setParameter(i, x);
    }



    /**
     * JDBC 2.0
     *
     * Sets a BLOB parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing a BLOB
     * @exception SQLException if a database access error occurs
     */
    public void setBlob (int i, Blob x) throws SQLException
    {
	setParameter(i, x.getBytes(0, (int)x.length()));
    }

    /**
     * JDBC 2.0
     *
     * Sets a CLOB parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing a CLOB
     * @exception SQLException if a database access error occurs
     */
    public void setClob (int i, Clob x) throws SQLException
    {
	setParameter(i, x.getSubString(0, (int)x.length()));
    }


    /**
     * JDBC 2.0
     *
     * Sets the designated parameter to SQL NULL.  This version of setNull should
     * be used for user-named types and REF type parameters.  Examples
     * of user-named types include: STRUCT, DISTINCT, JAVA_OBJECT, and 
     * named array types.
     *
     * <P><B>Note:</B> To be portable, applications must give the
     * SQL type code and the fully-qualified SQL type name when specifying
     * a NULL user-defined or REF parameter.  In the case of a user-named type 
     * the name is the type name of the parameter itself.  For a REF 
     * parameter the name is the type name of the referenced type.  If 
     * a JDBC driver does not need the type code or type name information, 
     * it may ignore it.     
     *
     * Although it is intended for user-named and Ref parameters,
     * this method may be used to set a null parameter of any JDBC type.
     * If the parameter does not have a user-named or REF type, the given
     * typeName is ignored.
     *
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param sqlType a value from java.sql.Types
     * @param typeName the fully-qualified name of an SQL user-named type,
     *  ignored if the parameter is not a user-named type or REF 
     * @exception SQLException if a database access error occurs
     */
    public void setNull (int parameterIndex, int sqlType, String typeName) 
	throws SQLException
    {
	setParameter(parameterIndex, new ConsusRef(0));
    }

    ClientPreparedStatement(ClientConnection con, String sql) throws SQLException { 
	super(con);
	conxn = con;
	this.sql = sql;
    }

    ClientPreparedStatement(ClientConnection con, String sql, int resultSetType, int resultSetConcurrency) throws SQLException { 
	super(con, resultSetType, resultSetConcurrency);
	conxn = con;
	this.sql = sql;
    }
}
