/* Consus - Pure Java Object Relational Database Management System
 * Character stream returned by ConsusClob for remote client connection
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;
import java.io.*;
import java.sql.*;

/** 
 * Class used to transfer CLOB data from server to the client
 */
public class ClientReader extends Reader {
    ClientConnection conxn;
    long             pos;
    long             offs;
    long             size;
    long             markPos;

    ClientReader(ClientConnection conxn, long pos, long size) { 
	this.conxn = conxn;
	this.pos = pos;
	this.size = size;
	offs = 0;
	markPos = 0;
    }

    /**
     * Read a single character.  This method will block until a character is
     * available, an I/O error occurs, or the end of the stream is reached.
     *
     * <p> Subclasses that intend to support efficient single-character input
     * should override this method.
     *
     * @return     The character read, as an integer in the range 0 to 16383
     *             (<tt>0x00-0xffff</tt>), or -1 if the end of the stream has
     *             been reached
     *
     * @exception  IOException  If an I/O error occurs
     */
    public int read() throws IOException {
	char buf[] = new char[1];
	return read(buf, 0, 1) == 1 ? (buf[0] & 0xFF) : -1;
    }

    /**
     * Read characters into a portion of an array.  This method will block
     * until some input is available, an I/O error occurs, or the end of the
     * stream is reached.
     *
     * @param      cbuf  Destination buffer
     * @param      off   Offset at which to start storing characters
     * @param      len   Maximum number of characters to read
     *
     * @return     The number of characters read, or -1 if the end of the
     *             stream has been reached
     *
     * @exception  IOException  If an I/O error occurs
     */
    public int read(char cbuf[], int off, int len) throws IOException { 
	if (len == 0) { 
	    return 0;
	}
	if (offs >= size) { 
	    return -1;
	}
	Message msg = new Message();
	msg.tag = Message.REQ_CLOB_SEGMENT;
	msg.opd1 = new BlobInfo(pos, size, offs);
	msg.opd3 = len;
	try { 
	    conxn.send(msg);
	    msg = conxn.receive();
	} catch(SQLException x) { 
	    throw new IOException(x.getMessage());
	}
	if (msg.opd3 > 0) { 
	    offs += msg.opd3;
	    System.arraycopy(msg.opd1, 0, cbuf, off, msg.opd3);
	}
	return msg.opd3;
    }


    /**
     * Skip characters.  This method will block until some characters are
     * available, an I/O error occurs, or the end of the stream is reached.
     *
     * @param  n  The number of characters to skip
     *
     * @return    The number of characters actually skipped
     *
     * @exception  IOException  If an I/O error occurs
     */
    public long skip(long n) throws IOException {
	if (offs + n > size) { 
	    n = size - offs;
	    offs = size;
	} else if (offs + n < 0) { 
	    n = -offs;
	    offs = 0;
	} else { 
	    offs += n;
	}
	return n;
    }

    /**
     * Tell whether this stream is ready to be read.
     *
     * @return True if the next read() is guaranteed not to block for input,
     * false otherwise.  Note that returning false does not guarantee that the
     * next read will block.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public boolean ready() throws IOException {
	return offs < size;
    }

    /**
     * Tell whether this stream supports the mark() operation.
     */
    public boolean markSupported() {
	return true;
    }

    /**
     * Mark the present position in the stream.  Subsequent calls to reset()
     * will attempt to reposition the stream to this point.  Not all
     * character-input streams support the mark() operation.
     *
     * @param  readAheadLimit  Limit on the number of characters that may be
     *                         read while still preserving the mark.  After
     *                         reading this many characters, attempting to
     *                         reset the stream may fail.
     *
     * @exception  IOException  If the stream does not support mark(),
     *                          or if some other I/O error occurs
     */
    public void mark(int readAheadLimit) throws IOException {
	markPos = offs;
    }

    /**
     * Reset the stream.  If the stream has been marked, then attempt to
     * reposition it at the mark.  If the stream has not been marked, then
     * attempt to reset it in some way appropriate to the particular stream,
     * for example by repositioning it to its starting point.  Not all
     * character-input streams support the reset() operation, and some support
     * reset() without supporting mark().
     *
     * @exception  IOException  If the stream has not been marked,
     *                          or if the mark has been invalidated,
     *                          or if the stream does not support reset(),
     *                          or if some other I/O error occurs
     */
    public void reset() throws IOException {
	offs = markPos;
    }

    /** 
     * Close the stream and release all used resources.
     */
    public void close() {}
}






