/* Consus - Pure Java Object Relational Database Management System
 * Statement protocol implementation for remote connection
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;

/**
 * <P>The object used for executing a static SQL statement
 * and obtaining the results produced by it. 
 *
 * <P>Only one ResultSet per Statement can be open at any point in
 * time. Therefore, if the reading of one ResultSet is interleaved
 * with the reading of another, each must have been generated by
 * different Statements. All statement <code>execute</code> methods implicitly
 * close a statment's current ResultSet if an open one exists.
 *
 * @see Connection#createStatement
 * @see ResultSet 
 */
public class ClientStatement extends ConsusStatement {
    Message          req = new Message();
    ClientConnection conxn;
    /**
     * Executes a SQL statement that returns a single ResultSet.
     *
     * @param sql typically this is a static SQL SELECT statement
     * @return a ResultSet that contains the data produced by the
     * query; never null 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet executeQuery(String sql) throws SQLException { 
	req.tag = Message.REQ_EXEC_QUERY;
	req.opd1 = sql;
	req.opd2 = null;
	req.opd3 = cursor.selectionLimit;
	conxn.send(req);
	Message reply = conxn.receive();
	cursor.tuples = (Object[][])reply.opd1;
	cursor.tupleColumns = getColumns(reply.opd2);
	cursor.nSelected = reply.opd3;
	oid = reply.opd4;
	cursor.currPos = -1;
	isQuery = true;
	return cursor;
    }

    /**
     * Executes an SQL INSERT, UPDATE or DELETE statement. In addition,
     * SQL statements that return nothing, such as SQL DDL statements,
     * can be executed.
     *
     * @param sql a SQL INSERT, UPDATE or DELETE statement or a SQL
     * statement that returns nothing
     * @return either the row count for INSERT, UPDATE or DELETE or 0
     * for SQL statements that return nothing
     * @exception SQLException if a database access error occurs
     */
    public int executeUpdate(String sql) throws SQLException { 
	req.tag = Message.REQ_EXEC_UPDATE;
	req.opd1 = sql;
	req.opd2 = null;
	conxn.send(req);
	Message reply = conxn.receive();
	isQuery = false;
	oid = reply.opd4;
	return updateCount = reply.opd3;
    }
	

    /**
     * Releases this <code>Statement</code> object's database 
     * and JDBC resources immediately instead of waiting for
     * this to happen when it is automatically closed.
     * It is generally good practice to release resources as soon as
     * you are finished with them to avoid tying up database
     * resources.
     * <P><B>Note:</B> A Statement is automatically closed when it is
     * garbage collected. When a Statement is closed, its current
     * ResultSet, if one exists, is also closed.  
     *
     * @exception SQLException if a database access error occurs
     */
    public void close() throws SQLException { }

    //----------------------- Multiple Results --------------------------

    /**
     * Executes a SQL statement that may return multiple results.
     * Under some (uncommon) situations a single SQL statement may return
     * multiple result sets and/or update counts.  Normally you can ignore
     * this unless you are (1) executing a stored procedure that you know may
     * return multiple results or (2) you are dynamically executing an
     * unknown SQL string.  The  methods <code>execute</code>,
     * <code>getMoreResults</code>, <code>getResultSet</code>,
     * and <code>getUpdateCount</code> let you navigate through multiple results.
     *
     * The <code>execute</code> method executes a SQL statement and indicates the
     * form of the first result.  You can then use getResultSet or
     * getUpdateCount to retrieve the result, and getMoreResults to
     * move to any subsequent result(s).
     *
     * @param sql any SQL statement
     * @return true if the next result is a ResultSet; false if it is
     * an update count or there are no more results
     * @exception SQLException if a database access error occurs
     * @see #getResultSet
     * @see #getUpdateCount
     * @see #getMoreResults 
     */
    public boolean execute(String sql) throws SQLException 
    {
	req.tag = Message.REQ_EXEC;
	req.opd1 = sql;
	req.opd2 = null;
	req.opd3 = cursor.selectionLimit;
	conxn.send(req);
	Message reply = conxn.receive();
	if (reply.tag == Message.REP_QUERY) { 
	    cursor.tuples = (Object[][])reply.opd1;
	    cursor.tupleColumns = getColumns(reply.opd2);
	    cursor.nSelected = reply.opd3;
	    oid = reply.opd4;
	    cursor.currPos = -1;
	    isQuery = true;
	    return true;
	} else { 
	    updateCount = reply.opd3;
	    oid = reply.opd4;
	    isQuery = false;
	    return false;
	}
    }
    
    ColumnNode[] getColumns(Object arg) { 
	ColumnInfo[] cinfo = (ColumnInfo[])arg;
	int n = cinfo.length;
	ColumnNode columns[] = new ColumnNode[n];
	while (--n >= 0) { 
	    columns[n] = new ColumnNode(cinfo[n].type, cinfo[n].name, n+1);
	}
	return columns;
    }

    ClientStatement(ClientConnection con, int resultSetType, int resultSetConcurrency) { 
	super(con, resultSetType, resultSetConcurrency);
	conxn = con;
    }
    ClientStatement(ClientConnection con) { 
	super(con);
	conxn = con;
    }
} 



