/* Consus - Pure Java Object Relational Database Management System
 * Binary Large OBject  implemenation
 * Created by Konstantin Knizhnik, Mar 2000.
 */

package org.garret.internal.consus;
import org.garret.consus.*;
import java.sql.*;
import java.io.*;
 
/**
 * JDBC 2.0
 *
 * <p>
 * The representation (mapping) in
 * the Java<sup><font size=-2>TM</font></sup> programming
 * language of an SQL 
 * <code>BLOB</code>.  An SQL <code>BLOB</code> is a built-in type 
 * that stores a Binary Large Object as a column value in a row of 
 * a database table. The driver implements <code>Blob</code> using
 * an SQL <code>locator(BLOB)</code>, which means that a
 * <code>Blob</code> object contains a logical pointer to the
 * SQL <code>BLOB</code> data rather than the data itself.
 * A <code>Blob</code> object is valid for the duration of the
 * transaction in which is was created.
 * 
 * <P>Methods in the interfaces {@link ResultSet}, 
 * {@link CallableStatement}, and {@link PreparedStatement}, such as
 * <code>getBlob</code> and <code>setBlob</code> allow a programmer to 
 * access the SQL <code>BLOB</code>.
 * The <code>Blob</code> interface provides methods for getting the
 * length of an SQL <code>BLOB</code> (Binary Large Object) value,
 * for materializing a <code>BLOB</code> value on the client, and for
 * determining the position of a pattern of bytes within a 
 * <code>BLOB</code> value. 
 */
public class ConsusBlob implements Blob, java.io.Serializable {
    ConsusConnection db;
    long             size;
    long             pos;
    InputStream      stream;
    byte[]           arr;
    Page             pg;
    long             pagePos;
    ClientConnection conxn;

    private void writeObject(java.io.ObjectOutputStream stream)
	throws IOException
    {
	stream.writeLong(pos);
	stream.writeLong(size);	
    }

    private void readObject(java.io.ObjectInputStream stream)
	throws IOException, ClassNotFoundException
    {
        conxn = ClientConnection.getConnectionByStream(stream);
	pos = stream.readLong();
	size = stream.readLong();	
    }
   

    /** 
     * Internal constructor used by Consus
     */
    ConsusBlob(ConsusConnection db, long pos, long size) { 
	this.db   = db;
	this.size = size;
	this.pos  = pos;
    }
    
    /** 
     * Constructor used by application to replace BLOB field
     */ 
    public ConsusBlob(long length, InputStream in) { 
	size = length;
	stream = in;
    }

    /** 
     * Construct BLOB object from byte array
     */ 
    public ConsusBlob(byte[] b) { 
	arr = b;
	size = b.length;
    }

    /**
     * Returns the number of bytes in the <code>BLOB</code> value
     * designated by this <code>Blob</code> object.
     * @return length of the <code>BLOB</code> in bytes
     * @exception SQLException if there is an error accessing the
     * length of the <code>BLOB</code>
     */
    public long length() throws SQLException { 
	return size;
    }
     
    /**
     * Returns as an array of bytes part or all of the <code>BLOB</code>
     * value that this <code>Blob</code> object designates.  The byte
     * array contains up to <code>length</code> consecutive bytes
     * starting at position <code>pos</code>.
     * @param pos the ordinal position of the first byte in the 
     * <code>BLOB</code> value to be extracted; the first byte is at
     * position 1
     * @param length is the number of consecutive bytes to be copied
     * @return a byte array containing up to <code>length</code> 
     * consecutive bytes from the <code>BLOB</code> value designated
     * by this <code>Blob</code> object, starting with the
     * byte at position <code>pos</code>.
     * @exception SQLException if there is an error accessing the
     * <code>BLOB</code>
     */
    public byte[] getBytes(long offs, int length) throws SQLException { 
	if (conxn != null) { 
	    Message msg = new Message();
	    msg.tag = Message.REQ_BLOB_SEGMENT;
	    msg.opd1 = new BlobInfo(pos, size, offs);
	    msg.opd3 = length;
	    conxn.send(msg);
	    msg = conxn.receive();
	    return (byte[])msg.opd1;
	}
	offs -= 1;
	if (offs < 0 || offs > size) { 
	    return null;
	}
	if (offs + length > size) { 
	    length = (int)(size - offs);
	}
	byte b[] = new byte[length];
	if (length == 0) { 
	    return b;
	}
	if (db == null) { 
	    if (stream != null) { 
		try {
		    while (length > 0) { 
			int rc = stream.read(b, (int)offs, length);
			if (rc < 0) { 
			    throw new ConsusRuntimeError("Failed to read byte stream");
			}
			offs += rc;
			length -= rc;
		    }
		} catch(java.io.IOException x) { 
		    throw new ConsusRuntimeError("Failed to read byte stream");
		}
	    } else { 
		System.arraycopy(arr, (int)offs, b, 0, length);
	    }
	    return b;
	}
	int fetched = 0;
	while (true) { 
	    int pageOffs = (int)(pos + offs) & (Page.pageSize-1);
	    Page pg = db.pool.getPage(pos + offs - pageOffs);
	    if (length > Page.pageSize - pageOffs) { 
		System.arraycopy(pg.data, pageOffs, b, fetched, 
				 Page.pageSize - pageOffs);
		fetched += Page.pageSize - pageOffs;
		offs += Page.pageSize - pageOffs;
		length -= Page.pageSize - pageOffs;
		db.pool.unfix(pg);
	    } else { 
		System.arraycopy(pg.data, pageOffs, b, fetched, length);
		db.pool.unfix(pg);
		return b;
	    }
	}
    }

    /**
     * Retrieves the <code>BLOB</code> designated by this
     * <code>Blob</code> instance as a stream.
     * @return a stream containing the <code>BLOB</code> data
     * @exception SQLException if there is an error accessing the
     * <code>BLOB</code>
     */
    public java.io.InputStream getBinaryStream () throws SQLException { 
	if (stream != null) { 
	    return stream;
	} else if (arr != null) { 
	    return new java.io.ByteArrayInputStream(arr);
	} else if (conxn != null) {
	    return new ClientInputStream(conxn, pos, size);
	} else { 
	    return new ConsusInputStream(db, pos, size);
	}
    }

    /** 
     * Determines the byte position at which the specified byte 
     * <code>pattern</code> begins within the <code>BLOB</code>
     * value that this <code>Blob</code> object represents.  The
     * search for <code>pattern</code. begins at position
     * <code>start</code>.  
     * @param pattern the byte array for which to search
     * @param start the position at which to begin searching; the
     *        first position is 1
     * @return the position at which the pattern appears, else -1.
     * @exception SQLException if there is an error accessing the 
     * <code>BLOB</code>
     */
    final byte getByte(long offs) throws SQLException { 
	if (offs < pagePos || offs >= pagePos + Page.pageSize) { 
	    db.pool.unfix(pg);
	    pagePos = offs & ~(Page.pageSize-1);
	    pg = db.pool.getPage(pagePos);
	}
	return pg.data[(int)offs & (Page.pageSize-1)];
    }
    
    /** 
     * Determines the byte position at which the specified byte 
     * <code>pattern</code> begins within the <code>BLOB</code>
     * value that this <code>Blob</code> object represents.  The
     * search for <code>pattern</code> begins at position
     * <code>start</code>.  
     * @param pattern the byte array for which to search
     * @param start the position at which to begin searching; the
     *        first position is 1
     * @return the position at which the pattern appears, else -1
     * @exception SQLException if there is an error accessing the 
     * <code>BLOB</code>
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC 2.0 API</a>
     */
    public long position(byte pattern[], long start) throws SQLException { 
	if (conxn != null) { 
	    Message msg = new Message();
	    msg.tag = Message.REQ_BLOB_POSITION;
	    msg.opd1 = new BlobInfo(pos, size, start);
	    msg.opd2 = pattern;
	    conxn.send(msg);
	    msg = conxn.receive();
	    return ((Long)msg.opd1).longValue();
	}
	    
	Assert.that(db != null);
	int len = pattern.length - 1;
	if (len < 0) { 
	    return start;
	} 
	start -= 1;
	if (start < 0 || start + len > size) { 
	    return -1;
	}
	int n = (int)(size - start);
	int i, j, k;
	int shift[] = new int[256];
	long base = start + pos;
	pagePos = base & ~(Page.pageSize-1);
	pg = db.pool.getPage(pagePos);
	for (i = 0; i < 256; i++) { 
	    shift[i] = len+1;
	}
 	for (i = 0; i < len; i++) { 
	    shift[pattern[i] & 0xFF] = len-i;
	}
	for (i = len; i < n; i += shift[0xFF & getByte(base + i)]) { 
	    j = len;
	    k = i;

	    while (pattern[j] == getByte(base + k)) { 
		k -= 1;
		if (--j < 0) { 
		    db.pool.unfix(pg);
		    return start + k + 2; // first position is 1
		}
	    }
	}
	db.pool.unfix(pg);
	return -1;
    } 
    
    /** 
     * Determines the byte position in the <code>BLOB</code> value
     * designated by this <code>Blob</code> object at which 
     * <code>pattern</code> begins.  The search begins at position
     * <code>start</code>.
     * @param pattern the <code>Blob</code> object designating
     * the <code>BLOB</code> value for which to search
     * @param start the position in the <code>BLOB</code> value
     *        at which to begin searching; the first position is 1
     * @return the position at which the pattern begins, else -1
     * @exception SQLException if there is an error accessing the
     * <code>BLOB</code>
     */
    public long position(Blob pattern, long start) throws SQLException { 
	if (pattern.length() == 0) { 
	    return start;
	}
	if (start <= 0 || start - 1 + pattern.length() > size) { 
	    return -1;
	}
	return position(pattern.getBytes(1, (int)pattern.length()), start);
    } 

    public String toString() { 
	try { 
	    return size == 0 ? "" : new String(getBytes(1, (int)size));
	} catch(SQLException x) { 
	    return null;
	}
    }

    /**
     * Writes the given array of bytes to this <code>Blob</code> 
     * object, starting at position <code>pos</code>, and returns the number
     * of bytes written.
     *
     * @param pos the position in the <code>BLOB</code> object at which
     *        to start writing
     * @param bytes the array of bytes to be written to the <code>BLOB</code>
     *        value that this <code>Blob</code> object represents
     * @return the number of bytes written
     * @exception SQLException if there is an error accessing the
     *            <code>BLOB</code> value
     * @see #getBytes
     * @since 1.4
     */
    public int setBytes(long pos, byte[] bytes) throws SQLException
    {
	return setBytes(pos, bytes, 0, bytes.length);
    }

    /**
     * Writes all or part of the given <code>byte</code> array to the
     * <code>BLOB</code> value that this <code>Blob</code> object represents
     * and returns the number of bytes written.
     * Writing starts at position <code>pos</code> in the <code>BLOB</code>
     * value; <code>len</code> bytes from the given byte array are written.
     *
     * @param pos the position in the <code>BLOB</code> object at which
     *        to start writing
     * @param bytes the array of bytes to be written to this <code>BLOB</code>
     *        object
     * @param offset the offset into the array <code>bytes</code> at which
     *        to start reading the bytes to be set
     * @param len the number of bytes to be written to the <code>BLOB</code>
     *        value from the array of bytes <code>bytes</code>
     * @return the number of bytes written
     * @exception SQLException if there is an error accessing the
     *            <code>BLOB</code> value
     * @see #getBytes
     * @since 1.4
     */
    public int setBytes(long pos, byte[] bytes, int offset, int len) throws SQLException
    {
	try { 
	    OutputStream out = setBinaryStream(pos);
	    out.write(bytes, offset, len);
	    out.close();
	} catch (IOException x) { 
	    throw new ConsusRuntimeError("Failed to write byte stream: " + x);
	}
	return len;
    }

    /**
     * Retrieves a stream that can be used to write to the <code>BLOB</code> 
     * value that this <code>Blob</code> object represents.  The stream begins
     * at position <code>pos</code>.
     *
     * @param pos the position in the <code>BLOB</code> value at which
     *        to start writing
     * @return a <code>java.io.OutputStream</code> object to which data can 
     *         be written
     * @exception SQLException if there is an error accessing the
     *            <code>BLOB</code> value
     * @see #getBinaryStream
     * @since 1.4
     */
     public java.io.OutputStream setBinaryStream(long pos) throws SQLException
     {
	 throw new ConsusNotImplemented("ConsusBlob.setBinaryStream(long pos)");
     }

    /**
     * Truncates the <code>BLOB</code> value that this <code>Blob</code>
     * object represents to be <code>len</code> bytes in length.
     *
     * @param len the length, in bytes, to which the <code>BLOB</code> value
     *        that this <code>Blob</code> object represents should be truncated
     * @exception SQLException if there is an error accessing the
     *            <code>BLOB</code> value
     * @since 1.4
     */
    public void truncate(long len) throws SQLException
    {
	 throw new ConsusNotImplemented("ConsusBlob.truncate(long len)");
    }   

    public InputStream getBinaryStream(long pos, long length) throws SQLException
    {
	 throw new ConsusNotImplemented("ConsusBlob.getBinaryStream");
    }   

    public void free() throws SQLException {}
}









