/* Consus - Pure Java Object Relational Database Management System
 * Character Large OBject implemenation
 * Created by Konstantin Knizhnik, Mar 2000.
 */

package org.garret.internal.consus;
import org.garret.consus.*;
import java.sql.*;
import java.io.*;
 
 
/**
 * JDBC 2.0
 * The mapping in the Java<sup><font size=-2>TM</font></sup> programming language 
 * for the SQL <code>CLOB</code> type.
 * An SQL <code>CLOB</code> is a built-in type
 * that stores a Character Large Object as a column value in a row of
 * a database table.
 * The driver implements a <code>Clob</code> object using an SQL
 * <code>locator(CLOB)</code>, which means that a <code>Clob</code> object 
 * contains a logical pointer to the SQL <code>CLOB</code> data rather than
 * the data itself. A <code>Clob</code> object is valid for the duration
 * of the transaction in which it was created.
 * <P>The <code>Clob</code> interface provides methods for getting the
 * length of an SQL <code>CLOB</code> (Character Large Object) value,
 * for materializing a <code>CLOB</code> value on the client, and for
 * searching for a substring or <code>CLOB</code> object within a
 * <code>CLOB</code> value.
 * Methods in the interfaces {@link ResultSet},
 * {@link CallableStatement}, and {@link PreparedStatement}, such as
 * <code>getClob</code> and <code>setClob</code> allow a programmer to
 * access the SQL <code>CLOB</code>.
 */

public class ConsusClob implements Clob, java.io.Serializable {
    ConsusConnection db;
    long             size;
    long             pos;
    Reader           stream;
    String           str;
    Page             pg;
    long             pagePos;
    ClientConnection conxn;

    /**
     * Internal constructor used by Consus
     */
    ConsusClob(ConsusConnection db, long pos, long size) { 
	this.db   = db;
	this.size = size;
	this.pos = pos;
    }
    
    /** 
     * Constructor used by application to replace BLOB field
     */ 
    public ConsusClob(long length, Reader in) { 
	size = length;
	stream = in;
    }

    /**
     * Construct CLOB object from String
     */
    public ConsusClob(String s) { 
	size = s.length();
	str = s;
    }

    private void writeObject(java.io.ObjectOutputStream stream)
	throws IOException
    {
	stream.writeLong(pos);
	stream.writeLong(size);	
    }

    private void readObject(java.io.ObjectInputStream stream)
	throws IOException, ClassNotFoundException
    {
        conxn = ClientConnection.getConnectionByStream(stream);
	pos = stream.readLong();
	size = stream.readLong();	
    }
 
    /**
     * Returns the number of characters 
     * in the <code>CLOB</code> value
     * designated by this <code>Clob</code> object.
     * @return length of the <code>CLOB</code> in characters
     * @exception SQLException if there is an error accessing the
     * length of the <code>CLOB</code>
     */
    public long length() throws SQLException { 
	return size;
    }

    /**
     * Returns a copy of the specified substring 
     * in the <code>CLOB</code> value
     * designated by this <code>Clob</code> object.
     * The substring begins at position
     * <code>pos</code> and has up to <code>length</code> consecutive
     * characters.
     * @param pos the first character of the substring to be extracted.
     *            The first character is at position 1.
     * @param length the number of consecutive characters to be copied
     * @return a <code>String</code> that is the specified substring in
     *         the <code>CLOB</code> value designated by this <code>Clob</code> object
     * @exception SQLException if there is an error accessing the
     * <code>CLOB</code> 
     */
    public String getSubString(long offs, int length) throws SQLException { 
	if (conxn != null) { 
	    Message msg = new Message();
	    msg.tag = Message.REQ_CLOB_SEGMENT;
	    msg.opd1 = new BlobInfo(pos, size, offs);
	    msg.opd3 = length;
	    conxn.send(msg);
	    msg = conxn.receive();
	    return (String)msg.opd1;
	}
	offs -= 1;
	if (offs < 0 || offs > size) { 
	    return null;
	}
	if (offs + length > size) { 
	    length = (int)(size - offs);
	}
	if (length == 0) { 
	    return "";
	}
	char cbuf[] = new char[length];
	if (db == null) { 
	    if (stream != null) { 
		try { 
		    while (length > 0) { 
			int rc = stream.read(cbuf, (int)offs, length);
			if (rc < 0) { 
			    throw new ConsusRuntimeError("Failed to read byte stream");
			}
			offs += rc;
			length -= rc;
		    }
		} catch(java.io.IOException x) { 
		    throw new ConsusRuntimeError("Failed to read character stream"
						 + x.getMessage());
		}
		return new String(cbuf);
	    } else { 
		return str.substring((int)offs, (int)(offs + length));
	    }
	}
	int fetched = 0;
	while (true) { 
	    int pageOffs = (int)(pos + offs*2) & (Page.pageSize-1);
	    Page pg = db.pool.getPage(pos + offs*2 - pageOffs);
	    if ((length - fetched)*2 > Page.pageSize - pageOffs) { 
		offs -= fetched;
		while (pageOffs < Page.pageSize) { 
		    cbuf[fetched] = (char)Bytes.unpack2(pg.data, pageOffs); 
		    fetched += 1;
		    pageOffs += 2;
		}
		offs += fetched;
		db.pool.unfix(pg);
	    } else { 
		while (fetched < length) { 
		    cbuf[fetched] = (char)Bytes.unpack2(pg.data, pageOffs); 
		    fetched += 1;
		    pageOffs += 2;
		}		    
		db.pool.unfix(pg);
		return new String(cbuf);
	    }
	}
    }

    /**
     * Gets the <code>Clob</code> contents as a Unicode stream.
     * @return a Unicode stream containing the <code>CLOB</code> data
     * @exception SQLException if there is an error accessing the 
     * <code>CLOB</code>
     */
    public java.io.Reader getCharacterStream() throws SQLException { 
	if (stream != null) { 
	    return stream;
	} else if (str != null) { 
	    return new StringReader(str);
	} else if (conxn != null) {
	    return new ClientReader(conxn, pos, size);
	} else { 
	    return new ConsusReader(db, pos, size);
	}
    }

    /**
     * Gets the <code>CLOB</code> value designated by this <code>Clob</code>
     * object as a stream of Ascii bytes.
     * @return an ascii stream containing the <code>CLOB</code> data
     * @exception SQLException if there is an error accessing the 
     * <code>CLOB</code> value
     */
    public java.io.InputStream getAsciiStream() throws SQLException { 
	return new StringBufferInputStream(getSubString(1, (int)size));
    }
    
    /** 
     * Determines the character position at which the specified substring 
     * <code>searchstr</code> appears in the <code>CLOB</code>.  The search 
     * begins at position <code>start</code>.
     * @param searchstr the substring for which to search 
     * @param start the position at which to begin searching; the first position
     *              is 1
     * @return the position at which the substring appears, else -1; the first
     *         position is 1
     * @exception SQLException if there is an error accessing the       
     * <code>CLOB</code> value
     */
    final byte getByte(long offs) throws SQLException { 
	if (offs < pagePos || offs >= pagePos + Page.pageSize) { 
	    db.pool.unfix(pg);
	    pagePos = offs & ~(Page.pageSize-1);
	    pg = db.pool.getPage(pagePos);
	}
	return pg.data[(int)offs & (Page.pageSize-1)];
    }

    /** 
     * Determines the character position at which the specified substring 
     * <code>searchstr</code> appears in the SQL <code>CLOB</code> value
     * represented by this <code>Clob</code> object.  The search 
     * begins at position <code>start</code>.
     * @param searchstr the substring for which to search 
     * @param start the position at which to begin searching; the first position
     *              is 1
     * @return the position at which the substring appears, else -1; the first
     *         position is 1
     * @exception SQLException if there is an error accessing the       
     * <code>CLOB</code> value
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC 2.0 API</a>
     */
    public long position(String searchstr, long start) throws SQLException { 
	if (conxn != null) { 
	    Message msg = new Message();
	    msg.tag = Message.REQ_CLOB_POSITION;
	    msg.opd1 = new BlobInfo(pos, size, start);
	    msg.opd2 = searchstr;
	    conxn.send(msg);
	    msg = conxn.receive();
	    return ((Long)msg.opd1).longValue();
	}
	Assert.that(db != null);
	byte[] pattern = new byte[searchstr.length()*2];
	for (int i = searchstr.length(); --i >= 0;) { 
	    Bytes.pack2(pattern, i*2,  (short)searchstr.charAt(i));
	}
	int len = pattern.length - 1;
	if (len < 0) { 
	    return start;
	} 
	start = (start - 1)*2;
	if (start < 0 || start + len >= size*2) { 
	    return -1;
	}
	int n = (int)(size*2 - start);
	int i, j, k;
	int shift[] = new int[256];
	long base = start + pos;
	pagePos = base & ~(Page.pageSize-1);
	pg = db.pool.getPage(pagePos);
	for (i = 0; i < 256; i++) { 
	    shift[i] = len+1;
	}
 	for (i = 0; i < len; i++) { 
	    shift[pattern[i] & 0xFF] = len-i;
	}
	for (i = len; i < n; i += shift[0xFF & getByte(base + i)]) { 
	    j = len;
	    k = i;

	    while (pattern[j] == getByte(base + k)) { 
		k -= 1;
		if (--j < 0) { 
		    db.pool.unfix(pg);
		    return ((start + k + 1) >>> 1) + 1; // first position is 1
		}
	    }
	}
	db.pool.unfix(pg);
	return -1;
    }
    
    /** 
     * Determines the character position at which the specified  
     * <code>Clob</code> object <code>searchstr</code> appears in this 
     * <code>Clob</code> object.  The search begins at position 
     * <code>start</code>.
     * @param searchstr the <code>Clob</code> object for which to search
     * @param start the position at which to begin searching; the first
     *              position is 1
     * @return the position at which the <code>Clob</code> object appears, 
     * else -1; the first position is 1
     * @exception SQLException if there is an error accessing the 
     * <code>CLOB</code> value
     */
    public long position(Clob searchstr, long start) throws SQLException { 
	if (searchstr.length() == 0) { 
	    return start;
	}
	if (start <= 0 || start - 1 + searchstr.length() > size) { 
	    return -1;
	}
	return position(searchstr.getSubString(1, (int)size), start);	
    }

    /** 
     * Convert CLOB object to String 
     */
    public String toString() { 
	try { 
	    return getSubString(1, (int)size);
	} catch(SQLException x) { 
	    return null;
	}
    }

   /**
     * Writes the given Java <code>String</code> to the <code>CLOB</code>
     * value that this <code>Clob</code> object designates at the position 
     * <code>pos</code>.
     *
     * @param pos the position at which to start writing to the <code>CLOB</code>
     *         value that this <code>Clob</code> object represents
     * @param str the string to be written to the <code>CLOB</code>
     *        value that this <code>Clob</code> designates
     * @return the number of characters written
     * @exception SQLException if there is an error accessing the 
     *            <code>CLOB</code> value
     *
     * @since 1.4
     */
    public int setString(long pos, String str) throws SQLException
    {
	return setString(pos, str, 0, str.length());
    }

    /**
     * Writes <code>len</code> characters of <code>str</code>, starting 
     * at character <code>offset</code>, to the <code>CLOB</code> value
     * that this <code>Clob</code> designates.
     *
     * @param pos the position at which to start writing to this
     *        <code>CLOB</code> object
     * @param str the string to be written to the <code>CLOB</code> 
     *        value that this <code>Clob</code> object represents
     * @param offset the offset into <code>str</code> to start reading
     *        the characters to be written
     * @param len the number of characters to be written
     * @return the number of characters written
     * @exception SQLException if there is an error accessing the 
     *            <code>CLOB</code> value
     *
     * @since 1.4
     */
    public int setString(long pos, String str, int offset, int len) throws SQLException
    {
	try { 
	    java.io.Writer out = setCharacterStream(pos);
	    out.write(str, offset, len);
	    out.close();
	} catch (IOException x) { 
	    throw new ConsusRuntimeError("Failed to write character stream: " + x);
	}
	return len;
    }


    /**
     * Retrieves a stream to be used to write Ascii characters to the
     * <code>CLOB</code> value that this <code>Clob</code> object represents, 
     * starting at position <code>pos</code>.
     *
     * @param pos the position at which to start writing to this
     *        <code>CLOB</code> object
     * @return the stream to which ASCII encoded characters can be written
     * @exception SQLException if there is an error accessing the 
     *            <code>CLOB</code> value
     * @see #getAsciiStream
     *
     * @since 1.4
     */
    public java.io.OutputStream setAsciiStream(long pos) throws SQLException {
	throw new ConsusNotImplemented("ConsusClob.setAsciiStream(long pos)");
    }


    /**
     * Retrieves a stream to be used to write a stream of Unicode characters 
     * to the <code>CLOB</code> value that this <code>Clob</code> object
     * represents, at position <code>pos</code>.
     *
     * @param  pos the position at which to start writing to the
     *        <code>CLOB</code> value
     *
     * @return a stream to which Unicode encoded characters can be written
     * @exception SQLException if there is an error accessing the 
     *            <code>CLOB</code> value
     * @see #getCharacterStream
     *
     * @since 1.4
     */
    public java.io.Writer setCharacterStream(long pos) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusClob.setCharacterStream(long pos)");
    }

    /**
     * Truncates the <code>CLOB</code> value that this <code>Clob</code> 
     * designates to have a length of <code>len</code> 
     * characters.
     * @param len the length, in bytes, to which the <code>CLOB</code> value
     *        should be truncated
     * @exception SQLException if there is an error accessing the 
     *            <code>CLOB</code> value
     *
     * @since 1.4
     */
    public void truncate(long len) throws SQLException {
	throw new ConsusNotImplemented("ConsusClob.truncate(long len)");
    }

    public void free() throws SQLException {}

    public Reader getCharacterStream(long pos, long length) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusClob.getCharactterStream");
    }
        

}




