/* Consus - Pure Java Object Relational Database Management System
 * Connection protocol implementation (most of the database methods are implemeted here)
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;
import java.io.*;
import java.util.*;
import java.text.*;


/**
 * <P>A connection (session) with a specific
 * database. Within the context of a Connection, SQL statements are
 * executed and results are returned.
 *
 * <P>A Connection's database is able to provide information
 * describing its tables, its supported SQL grammar, its stored
 * procedures, the capabilities of this connection, and so on. This
 * information is obtained with the <code>getMetaData</code> method.
 *
 * <P><B>Note:</B> By default the Connection automatically commits
 * changes after executing each statement. If auto commit has been
 * disabled, an explicit commit must be done or database changes will
 * not be saved.
 *
 * @see Statement 
 * @see ResultSet
 * @see DatabaseMetaData
 */
public class ConsusConnection implements org.garret.consus.ConsusConnection, ConsusTransaction
{
    /**
     * Creates a <code>Statement</code> object for sending
     * SQL statements to the database.
     * SQL statements without parameters are normally
     * executed using Statement objects. If the same SQL statement 
     * is executed many times, it is more efficient to use a 
     * PreparedStatement
     *
     * JDBC 2.0
     *
     * Result sets created using the returned Statement will have
     * forward-only type, and read-only concurrency, by default.
     *
     * @return a new Statement object 
     * @exception SQLException if a database access error occurs
     */
     public Statement createStatement() throws SQLException 
     {
         return new ConsusStatement(this);
     }


    /**
     * Creates a <code>PreparedStatement</code> object for sending
     * parameterized SQL statements to the database.
     * 
     * A SQL statement with or without IN parameters can be
     * pre-compiled and stored in a PreparedStatement object. This
     * object can then be used to efficiently execute this statement
     * multiple times.
     *
     * <P><B>Note:</B> This method is optimized for handling
     * parametric SQL statements that benefit from precompilation. If
     * the driver supports precompilation,
     * the method <code>prepareStatement</code> will send
     * the statement to the database for precompilation. Some drivers
     * may not support precompilation. In this case, the statement may
     * not be sent to the database until the <code>PreparedStatement</code> is
     * executed.  This has no direct effect on users; however, it does
     * affect which method throws certain SQLExceptions.
     *
     * JDBC 2.0
     *
     * Result sets created using the returned PreparedStatement will have
     * forward-only type and read-only concurrency, by default.
     *
     * @param sql a SQL statement that may contain one or more '?' IN
     * parameter placeholders
     * @return a new PreparedStatement object containing the
     * pre-compiled statement 
     * @exception SQLException if a database access error occurs
     */
     public PreparedStatement prepareStatement(String sql)
        throws SQLException
     {
         return new ConsusPreparedStatement(this, sql);
     }

    /**
     * Creates a <code>CallableStatement</code> object for calling
     * database stored procedures.
     * The CallableStatement provides
     * methods for setting up its IN and OUT parameters, and
     * methods for executing the call to a stored procedure.
     *
     * <P><B>Note:</B> This method is optimized for handling stored
     * procedure call statements. Some drivers may send the call
     * statement to the database when the method <code>prepareCall</code>
     * is done; others
     * may wait until the CallableStatement is executed. This has no
     * direct effect on users; however, it does affect which method
     * throws certain SQLExceptions.
     *
     * JDBC 2.0
     *
     * Result sets created using the returned CallableStatement will have
     * forward-only type and read-only concurrency, by default.
     *
     * @param sql a SQL statement that may contain one or more '?'
     * parameter placeholders. Typically this  statement is a JDBC
     * function call escape string.
     * @return a new CallableStatement object containing the
     * pre-compiled SQL statement 
     * @exception SQLException if a database access error occurs
     */
     public CallableStatement prepareCall(String sql) throws SQLException 
     {
         throw new ConsusNotImplemented("Connection.prepareCall");
     } 
                        
    /**
     * Converts the given SQL statement into the system's native SQL grammar.
     * A driver may convert the JDBC sql grammar into its system's
     * native SQL grammar prior to sending it; this method returns the
     * native form of the statement that the driver would have sent.
     *
     * @param sql a SQL statement that may contain one or more '?'
     * parameter placeholders
     * @return the native form of this statement
     * @exception SQLException if a database access error occurs
     */
    public String nativeSQL(String sql) throws SQLException 
    {
        return sql;
    }

    /**
     * Sets this connection's auto-commit mode.
     * If a connection is in auto-commit mode, then all its SQL
     * statements will be executed and committed as individual
     * transactions.  Otherwise, its SQL statements are grouped into
     * transactions that are terminated by a call to either
     * the method <code>commit</code> or the method <code>rollback</code>.
     * By default, new connections are in auto-commit
     * mode.
     *
     * The commit occurs when the statement completes or the next
     * execute occurs, whichever comes first. In the case of
     * statements returning a ResultSet, the statement completes when
     * the last row of the ResultSet has been retrieved or the
     * ResultSet has been closed. In advanced cases, a single
     * statement may return multiple results as well as output
     * parameter values. In these cases the commit occurs when all results and
     * output parameter values have been retrieved.
     *
     * @param autoCommit true enables auto-commit; false disables
     * auto-commit.  
     * @exception SQLException if a database access error occurs
     */
    public void setAutoCommit(boolean autoCommit) throws SQLException { 
	this.autoCommit = autoCommit;
    }

    /**
     * Gets the current auto-commit state.
     *
     * @return the current state of auto-commit mode
     * @exception SQLException if a database access error occurs
     * @see #setAutoCommit 
     */
    public boolean getAutoCommit() throws SQLException { 
	return autoCommit;
    }

    /** 
     * Set autocommit mode only for the current thread. Transaction commit modes of other
     * threads are not changed. Transaction will be automatically committed after each
     * statement when either global auto-commit mode is set either current thread 
     * is in auto-commit state.
     */
    public synchronized void setThreadAutoCommit(boolean autoCommit) throws SQLException { 
	ThreadContext ctx = getContext();
	if (ctx.autoCommit != autoCommit) { 
	    checkThreadAutoCommitState += autoCommit ? +1 : -1;
	    ctx.autoCommit = autoCommit;
	}
    }

    /**
     * Gets the current auto-commit state.
     *
     * @return the current state of auto-commit mode
     * @exception SQLException if a database access error occurs
     * @see #setAutoCommit 
     */
    public boolean getThreadAutoCommit() throws SQLException { 
	return getContext().autoCommit;
    }

    void endStatement() throws SQLException { 
	if (autoCommit || (checkThreadAutoCommitState > 0 && getContext().autoCommit)) { 
	    if (isolationLevel == TRANSACTION_NONE) { 
		commitSubtransaction();
	    } else { 
		commitTransaction();	
	    }
	}
    }

    /**
     * Makes all changes made since the previous
     * commit/rollback permanent and releases any database locks
     * currently held by the Connection. This method should be
     * used only when auto-commit mode has been disabled.
     *
     * @exception SQLException if a database access error occurs
     * @see #setAutoCommit 
     */
    public void commit() throws SQLException
    {
	if (isolationLevel == TRANSACTION_NONE) {
	    commitSubtransaction();
	} else { 
	    commitTransaction();
	}
    }

    public void prepareTransaction(int fmtid, byte[] gtrid, byte[] bqual) throws SQLException 
    { 
	if (modified) { 
	    if (usedSize - usedBeforeLastGC > GCThreshold) { 
		doGarbageCollection();
	    }
	    for (TableDescriptor table = firstTable; table != null; table = table.next)
	    { 
		if (table.altered) { 
		    put(table.tableId, table.exportDescriptor());
		} else if (table.updated) { 
		    TableDescriptor.setNRows(table.desc, table.nRows);
		    TableDescriptor.setLastRow(table.desc, table.lastRow);
		    TableDescriptor.setFirstRow(table.desc, table.firstRow);
		    TableDescriptor.setTimestamp(table.desc, table.timestamp);
		    put(table.tableId, table.desc);
		}
	    }
	    // enable readers to proceed
	    synchronized(monitor) {
		ThreadContext ctx = getContext();
		Assert.that(monitor.nWriters == 1 && monitor.nReaders == 0);
		monitor.nWriters = 0;
		monitor.nReaders = 1;
		ctx.writeAccess = false;
		ctx.writeAccessByThisTransaction = false;
		ctx.readAccess = true;
		ctx.transactionPrepared = true;
		commitInProgress = true;
    		Assert.that(monitor.nWriters == 0 && !monitor.waitForUpgrade);
		if (monitor.nWaitReaders != 0) { 
		    monitor.notifyAll();
		}
	    } 
	    int curr = header.curr;
	    int i, j, n;
	    int[] map = dirtyPagesMap;
	    int oldIndexSize = header.root[curr].indexSize;
	    int newIndexSize = header.root[1-curr].indexSize;
	    int nPages = committedIndexSize >>> dbHandlesPerPageBits;
	    Page pg;
	    if (newIndexSize > oldIndexSize) { 
		long newIndex = allocate(newIndexSize*8, 0);
		header.root[1-curr].shadowIndex = newIndex;
		header.root[1-curr].shadowIndexSize = newIndexSize;
		cloneBitmap(header.root[curr].index, oldIndexSize*8);
		free(header.root[curr].index, oldIndexSize*8);
	    }
	    for (i = 0; i < nPages; i++) { 
		if ((map[i >> 5] & (1 << (i & 31))) != 0) { 
		    Page srcIndex = pool.getPage(header.root[1-curr].index+i*Page.pageSize);
		    Page dstIndex = pool.getPage(header.root[curr].index+i*Page.pageSize);
		    for (j = 0; j < Page.pageSize; j += 8) {
			long pos = Bytes.unpack8(dstIndex.data, j);
			if (Bytes.unpack8(srcIndex.data, j) != pos) { 
			    if ((pos & dbFreeHandleFlag) == 0) {
				if ((pos & dbPageObjectFlag) != 0) {  
				    free(pos & ~dbFlagsMask, Page.pageSize);
				} else { 
				    int offs = (int)pos & (Page.pageSize-1);
				    pg = pool.getPage(pos-offs);
				    free(pos, Record.getSize(pg.data, offs));
				    pool.unfix(pg);
				}
			    }
			}
		    }
		    pool.unfix(srcIndex);
		    pool.unfix(dstIndex);
		}
	    }
	    n = committedIndexSize & (dbHandlesPerPage-1);
	    if (n != 0 && (map[i >> 5] & (1 << (i & 31))) != 0) { 
		Page srcIndex = pool.getPage(header.root[1-curr].index+i*Page.pageSize);
		Page dstIndex = pool.getPage(header.root[curr].index+i*Page.pageSize);
		j = 0;
		do { 
		    long pos = Bytes.unpack8(dstIndex.data, j);
		    if (Bytes.unpack8(srcIndex.data, j) != pos) { 
			if ((pos & dbFreeHandleFlag) == 0) {
			    if ((pos & dbPageObjectFlag) != 0) { 
				free(pos & ~dbFlagsMask, Page.pageSize);
			    } else { 
				int offs = (int)pos & (Page.pageSize-1);
				pg = pool.getPage(pos - offs);
				free(pos, Record.getSize(pg.data, offs));
				pool.unfix(pg);
			    }
			}
		    }
		    j += 8;
		} while (--n != 0);

		pool.unfix(srcIndex);
		pool.unfix(dstIndex);
	    }
	    for (i = 0; i <= nPages; i++) { 
		if ((map[i >> 5] & (1 << (i & 31))) != 0) { 
		    pg = pool.putPage(header.root[1-curr].index+i*Page.pageSize);
		    for (j = 0; j < Page.pageSize; j += 8) {
			Bytes.pack8(pg.data, j, Bytes.unpack8(pg.data, j) & ~dbModifiedFlag);
		    }
		    pool.unfix(pg);
		}
	    }
	    if (currIndexSize > committedIndexSize) { 
		long page = (header.root[1-curr].index 
			   + committedIndexSize*8) & ~(Page.pageSize-1);
		long end = (header.root[1-curr].index + Page.pageSize - 1
			    + currIndexSize*8) & ~(Page.pageSize-1);
		while (page < end) { 
		    pg = pool.putPage(page);
		    for (j = 0; j < Page.pageSize; j += 8) {
			Bytes.pack8(pg.data, j, Bytes.unpack8(pg.data, j) & ~dbModifiedFlag);
		    }
		    pool.unfix(pg);
		    page += Page.pageSize;
		}
	    }
	    header.root[1-curr].usedSize = usedSize;
	    pg = pool.putPage(0);
	    header.pack(pg.data);
	    if (gtrid != null) { 
		pool.flush();
		pool.modify(pg);
		header.fmtid = fmtid;
		header.gtrid = gtrid;
		header.bqual = bqual;
		header.dirty = false;
		header.pack(pg.data);
	    }
	    pool.unfix(pg);
	    pool.flush();
	}
    }	    

    public void completeTransaction() throws SQLException 
    {
	completeTransaction(getContext());
    }

    void completeTransaction(ThreadContext ctx) throws SQLException 
    { 
	int i, j, n;

	if (ctx.transactionPrepared) {  
	    Assert.that(commitInProgress && !ctx.writeAccess && ctx.readAccess);
	    for (TableDescriptor table = firstTable; table != null; table = table.next)
	    { 
		table.updated = table.altered = table.created = false;
	    }
	    deletedTables = null;

	    int curr = header.curr;
	    Page pg = pool.putPage(0);
	    synchronized (backupEvent) { 
		while (backupInProgress) { 
		    try { 
			backupEvent.wait();
		    } catch(InterruptedException x) {}
		}
		pool.modify(pg);
		header.curr = curr ^= 1;
		header.dirty = true;
		header.gtrid = null;
		header.bqual = null;
		header.pack(pg.data);
	    }
	    pool.unfix(pg);
	    pool.flush();
	    
	    header.root[1-curr].size = header.root[curr].size;
	    header.root[1-curr].indexUsed = currIndexSize; 
	    header.root[1-curr].freeList  = header.root[curr].freeList; 
	    header.root[1-curr].bitmapEnd = header.root[curr].bitmapEnd; 
	    
	    int oldIndexSize = header.root[curr].indexSize;
	    int newIndexSize = header.root[1-curr].indexSize;
	    int[] map = dirtyPagesMap;

	    if (currIndexSize == 0 || newIndexSize != oldIndexSize) {
		if (currIndexSize == 0) { 
		    currIndexSize = header.root[1-curr].indexUsed;
		}
		header.root[1-curr].index = header.root[curr].shadowIndex;
		header.root[1-curr].indexSize = header.root[curr].shadowIndexSize;
		header.root[1-curr].shadowIndex = header.root[curr].index;
		header.root[1-curr].shadowIndexSize = header.root[curr].indexSize;
		pool.copy(header.root[1-curr].index, header.root[curr].index,
			  currIndexSize*8);
		i = (currIndexSize+dbHandlesPerPage*32-1) >>> (dbHandlesPerPageBits+5);
		while (--i >= 0) { 
		    map[i] = 0;
		}
	    } else { 
		int nPages = committedIndexSize >>> dbHandlesPerPageBits;
		for (i = 0; i < nPages; i++) { 
		    if ((map[i >> 5] & (1 << (i & 31))) != 0) { 
			map[i >> 5] -= (1 << (i & 31));
			pool.copy(header.root[1-curr].index + i*Page.pageSize,
				  header.root[curr].index + i*Page.pageSize,
				  Page.pageSize);
		    }
		}
		if (currIndexSize > i*dbHandlesPerPage &&
		    ((map[i >> 5] & (1 << (i & 31))) != 0
		     || currIndexSize != committedIndexSize))
		{
		    pool.copy(header.root[1-curr].index + i*Page.pageSize,
			      header.root[curr].index + i*Page.pageSize,
			      8*currIndexSize - i*Page.pageSize);
		    j = i>>>5;
		    n = (currIndexSize + dbHandlesPerPage*32 - 1) >>> (dbHandlesPerPageBits+5); 
		    while (j < n) { 
			map[j++] = 0;
		    }
		}
	    }
	    synchronized (this) { 
		modified = false;
		schemaAltered = false;
		currIndex = curr;
		concurrentTransId += 1;
		commitInProgress = false;	    
	    }
	}
	if (ctx.writeAccess || ctx.readAccess) { 
	    endTransaction(ctx);
	}
    }


    public void commitTransaction() throws SQLException
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	ThreadContext ctx = getContext();     
	boolean hasSomethingToCommit;
	synchronized (this) { 
	    hasSomethingToCommit = modified && !commitInProgress && (ctx.writeAccess || ctx.concurrentId == concurrentTransId);
	} 
	if (!hasSomethingToCommit) {
	    if (ctx.writeAccess || ctx.readAccess) { 
		endTransaction(ctx);
	    }
	}
	if (!ctx.writeAccess) {
	    beginTransaction(true);
	    if (ctx.concurrentId != concurrentTransId) { 
		endTransaction(ctx);
		return;
	    }
	}
	if (commitDelay != 0) {
	    if (ctx.commitDelayed) { 
		return;
	    }
	    synchronized (commitThreadSyncEvent) {
		if (!schemaAltered && monitor.nWaitWriters + monitor.nWaitReaders == 0) {  
		    synchronized (delayedCommitStartTimerEvent) { 
			delayedCommitContext = ctx;
			ctx.commitDelayed = true;
			delayedCommitStartTimerEvent.notify();
		    }
		    try { 
			commitThreadSyncEvent.wait();
		    } catch(InterruptedException x) {}
		    return;
		}
	    }
	} 
	prepareTransaction(0, null, null);
	completeTransaction(ctx);
    } 
    
    class DelayedCommitThread extends Thread { 
        public void run() { 
	    try { 
		synchronized(delayedCommitStartTimerEvent) { 
		    synchronized(commitThreadSyncEvent) { 
			commitThreadSyncEvent.notify();
		    }
		    while (true) { 
			delayedCommitStartTimerEvent.wait(); 
			if (delayedCommitContext == null) { 			
			    return;
			}
			synchronized(delayedCommitStopTimerEvent) { 
			    synchronized (commitThreadSyncEvent) {
				commitThreadSyncEvent.notify();
			    }
			    if (forceCommitCount == 0 && monitor.nWaitWriters + monitor.nWaitReaders == 0) {  
				//System.out.println("Delayed transaction: " + commitTimeout);
				commitTimerStarted = System.currentTimeMillis();
				delayedCommitStopTimerEvent.wait(commitTimeout);
			    }
			    // System.out.println("forceCommitCount = " +  forceCommitCount + " monitor.nWaitWriters + monitor.nWaitReaders= " + monitor.nWaitWriters + monitor.nWaitReaders);
			    ThreadContext ctx = delayedCommitContext;
			    if (ctx != null) { 
				//System.out.println("delayedCommitContext=" + delayedCommitContext);
				commitTimeout = commitDelay;
				delayedCommitContext = null;
                synchronized (ctxHash) { 
                    ctxHash.put(this, ctx);
                }
				prepareTransaction(0, null, null);
				completeTransaction(ctx);
				ctx.commitDelayed = false;
			    }
			}
		    }
		}
	    } catch (Exception x) { 
		System.err.println("Exception during delayed commit: ");
		x.printStackTrace(System.err);
	    }
	}
    }

    public ConsusTransId getTransId() throws SQLException
    { 
	return new ConsusTransId(header.fmtid, header.gtrid, header.bqual);
    }

    public void rollbackTransaction() throws SQLException
    {
	rollbackTransaction(getContext());
    }
    
    void rollbackTransaction(ThreadContext ctx) throws SQLException
    { 
	TableDescriptor table;
	int curr = header.curr;
	int[] map = dirtyPagesMap;
	if (committedIndexSize == 0 || header.root[1-curr].index != header.root[curr].shadowIndex) { 
	    committedIndexSize = header.root[curr].indexUsed;
	    pool.copy(header.root[curr].shadowIndex, header.root[curr].index, 8*committedIndexSize);
	} else { 
	    int nPages = (committedIndexSize + dbHandlesPerPage - 1) >>> dbHandlesPerPageBits;
	    for (int i = 0; i < nPages; i++) { 
		if ((map[i >> 5] & (1 << (i & 31))) != 0) { 
		    pool.copy(header.root[curr].shadowIndex + i*Page.pageSize,
			      header.root[curr].index + i*Page.pageSize,
			      Page.pageSize);
		}
	    }
	}
	for (int j = (currIndexSize+dbHandlesPerPage*32-1) >>> (dbHandlesPerPageBits+5);
	     --j >= 0;
	     map[j] = 0);
	header.root[1-curr].index = header.root[curr].shadowIndex;
	header.root[1-curr].indexSize = header.root[curr].shadowIndexSize;
	header.root[1-curr].indexUsed = committedIndexSize;
	header.root[1-curr].freeList  = header.root[curr].freeList; 
	header.root[1-curr].bitmapEnd = header.root[curr].bitmapEnd; 
	header.root[1-curr].size = header.root[curr].size;
	header.gtrid = null;
	header.bqual = null;
	header.dirty = true;
	modified = false;
	commitInProgress = false;
	usedSize = usedBeforeLastGC = header.root[curr].size;
	concurrentTransId += 1;

	currRBitmapPage = currPBitmapPage = dbBitmapId;
	currRBitmapOffs = currPBitmapOffs = 0;
	
	while (deletedTables != null) { 
	    table = deletedTables;
	    deletedTables = (TableDescriptor)deletedTables.next;
	    registerTable(table);
	}
	TableDescriptor next;
	for (table = firstTable; table != null; table = next) { 
	    next = table.next;
	    if (table.created) { 
		unlinkTable(table);
	    } else if (table.altered || table.updated) { 
		table.undoUpdate(new TableDescriptor(this, table.tableId));
	    }
	}
	restoreTablesConsistency();
	resetForeignKeyConstraints();
	endTransaction(ctx);
    }

    /**
     * Drops all changes made since the previous
     * commit/rollback and releases any database locks currently held
     * by this Connection. This method should be used only when auto-
     * commit has been disabled.
     *
     * @exception SQLException if a database access error occurs
     * @see #setAutoCommit 
     */    
    public void rollback() throws SQLException
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	ThreadContext ctx = getContext();
	if (modified && (ctx.writeAccessByThisTransaction || ctx.concurrentId == concurrentTransId))
	{ 
	    if (!ctx.writeAccess) {
		beginTransaction(true);
	    }
	    rollbackTransaction(ctx);
	} else { 
	    endTransaction(ctx);
	}
    }

    /**
     * Releases a Connection's database and JDBC resources
     * immediately instead of waiting for
     * them to be automatically released.
     *
     * <P><B>Note:</B> A Connection is automatically closed when it is
     * garbage collected. Certain fatal errors also result in a closed
     * Connection.
     *
     * @exception SQLException if a database access error occurs
     */
    public void close() throws SQLException
    {
	synchronized (backupCS) { 
	    if (backupThread != null) { 
		backupThread.fileName = null;
		backupCS.notify();
	    }
	}
	if (commitDelay != 0) { 
	    synchronized (delayedCommitStopTimerEvent) { 
		delayedCommitStopTimerEvent.notify();
	    }
	    synchronized (delayedCommitStartTimerEvent) { 
		delayedCommitContext = null;
		delayedCommitStartTimerEvent.notify();
	    }
	    commitDelay = 0;
	}

	if (modified) { 
	    beginTransaction(true);
	    commitTransaction();
	}
	opened = false;
	if (header.dirty) { 
	    Page pg = pool.putPage(0);
	    header.pack(pg.data);
	    pool.flush();
	    pool.modify(pg);
	    header.dirty = false;
	    header.pack(pg.data);
	    pool.unfix(pg);
	    pool.flush();
	}
	ctxHash.clear();
	pool.close(); 
    }

    /**
     * Tests to see if a Connection is closed.
     *
     * @return true if the connection is closed; false if it's still open
     * @exception SQLException if a database access error occurs
     */
    public boolean isClosed() throws SQLException { 
	return !opened;
    }

    //======================================================================
    // Advanced features:

    /**
     * Gets the metadata regarding this connection's database.
     * A Connection's database is able to provide information
     * describing its tables, its supported SQL grammar, its stored
     * procedures, the capabilities of this connection, and so on. This
     * information is made available through a DatabaseMetaData
     * object.
     *
     * @return a DatabaseMetaData object for this Connection 
     * @exception SQLException if a database access error occurs
     */
    public DatabaseMetaData getMetaData() throws SQLException { 
	return new ConsusDatabaseMetaData(this);
    } 


    /**
     * Puts this connection in read-only mode as a hint to enable 
     * database optimizations.
     *
     * <P><B>Note:</B> This method cannot be called while in the
     * middle of a transaction.
     *
     * @param readOnly true enables read-only mode; false disables
     * read-only mode.  
     * @exception SQLException if a database access error occurs
     */
    public void setReadOnly(boolean readOnly) throws SQLException {}

    /**
     * Tests to see if the connection is in read-only mode.
     *
     * @return true if connection is read-only and false otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean isReadOnly() throws SQLException { 
	return false;
    }

    /**
     * Sets a catalog name in order to select     
     * a subspace of this Connection's database in which to work.
     * If the driver does not support catalogs, it will
     * silently ignore this request.
     *
     * @exception SQLException if a database access error occurs
     */
    public void setCatalog(String catalog) throws SQLException {}

    /**
     * Returns the Connection's current catalog name.
     *
     * @return the current catalog name or null
     * @exception SQLException if a database access error occurs
     */
    public String getCatalog() throws SQLException
    {
        return null;
    }

    /**
     * Attempts to change the transaction
     * isolation level to the one given.
     * The constants defined in the interface <code>Connection</code>
     * are the possible transaction isolation levels.
     *
     * <P><B>Note:</B> This method cannot be called while
     * in the middle of a transaction.
     *
     * @param level one of the TRANSACTION_* isolation values with the
     * exception of TRANSACTION_NONE; some databases may not support
     * other values
     * @exception SQLException if a database access error occurs
     * @see DatabaseMetaData#supportsTransactionIsolationLevel 
     */
    public void setTransactionIsolation(int level) throws SQLException { 
	isolationLevel = level;
    }

    /**
     * Gets this Connection's current transaction isolation level.
     *
     * @return the current TRANSACTION_* mode value
     * @exception SQLException if a database access error occurs
     */
    public int getTransactionIsolation() throws SQLException 
    { 
        return isolationLevel;
    }

    /**
     * Returns the first warning reported by calls on this Connection.
     *
     * <P><B>Note:</B> Subsequent warnings will be chained to this
     * SQLWarning.
     *
     * @return the first SQLWarning or null 
     * @exception SQLException if a database access error occurs
     */
     public SQLWarning getWarnings() throws SQLException 
     { 
         return null;
     }

    /**
     * Clears all warnings reported for this <code>Connection</code> object.    
     * After a call to this method, the method <code>getWarnings</code>
     * returns null until a new warning is
     * reported for this Connection.  
     *
     * @exception SQLException if a database access error occurs
     */
     public void clearWarnings() throws SQLException {}


    //--------------------------JDBC 2.0-----------------------------

    /**
     * JDBC 2.0
     *
     * Creates a <code>Statement</code> object that will generate
     * <code>ResultSet</code> objects with the given type and concurrency.
     * This method is the same as the <code>createStatement</code> method
     * above, but it allows the default result set
     * type and result set concurrency type to be overridden.
     *
     * @param resultSetType a result set type; see ResultSet.TYPE_XXX
     * @param resultSetConcurrency a concurrency type; see ResultSet.CONCUR_XXX
     * @return a new Statement object 
     * @exception SQLException if a database access error occurs
     */
    public Statement createStatement(int resultSetType, int resultSetConcurrency) 
      throws SQLException
    {
         return new ConsusStatement(this, resultSetType, resultSetConcurrency);
    }

    /**
     * JDBC 2.0
     *
     * Creates a <code>PreparedStatement</code> object that will generate
     * <code>ResultSet</code> objects with the given type and concurrency.
     * This method is the same as the <code>prepareStatement</code> method
     * above, but it allows the default result set
     * type and result set concurrency type to be overridden.
     *
     * @param resultSetType a result set type; see ResultSet.TYPE_XXX
     * @param resultSetConcurrency a concurrency type; see ResultSet.CONCUR_XXX
     * @return a new PreparedStatement object containing the
     * pre-compiled SQL statement 
     * @exception SQLException if a database access error occurs
     */
     public PreparedStatement prepareStatement(String sql, int resultSetType, 
					       int resultSetConcurrency)
       throws SQLException
     {
          return new ConsusPreparedStatement(this, sql, resultSetType, resultSetConcurrency);
     }

    /**
     * JDBC 2.0
     *
     * Creates a <code>CallableStatement</code> object that will generate
     * <code>ResultSet</code> objects with the given type and concurrency.
     * This method is the same as the <code>prepareCall</code> method
     * above, but it allows the default result set
     * type and result set concurrency type to be overridden.
     *
     * @param resultSetType a result set type; see ResultSet.TYPE_XXX
     * @param resultSetConcurrency a concurrency type; see ResultSet.CONCUR_XXX
     * @return a new CallableStatement object containing the
     * pre-compiled SQL statement 
     * @exception SQLException if a database access error occurs
     */
    public CallableStatement prepareCall(String sql, int resultSetType, 
					 int resultSetConcurrency) throws SQLException
    {
        throw new ConsusNotImplemented("Connection.prepareCall");
    }

    /**
     * JDBC 2.0
     *
     * Gets the type map object associated with this connection.
     * Unless the application has added an entry to the type map,
     * the map returned will be empty.
     *
     * @return the <code>java.util.Map</code> object associated 
     *         with this <code>Connection</code> object
     */
    public java.util.Map getTypeMap() throws SQLException
    {
        throw new ConsusNotImplemented("Connection.getTypeMap");
    }

    /**
     * JDBC 2.0
     *
     * Installs the given type map as the type map for
     * this connection.  The type map will be used for the
     * custom mapping of SQL structured types and distinct types.
     *
     * @param the <code>java.util.Map</code> object to install
     *        as the replacement for this <code>Connection</code>
     *        object's default type map
     */
    public void setTypeMap(java.util.Map map) throws SQLException
    {
        throw new ConsusNotImplemented("Connection.setTypeMap");
    }

    /**
     * Changes the holdability of <code>ResultSet</code> objects
     * created using this <code>Connection</code> object to the given
     * holdability.
     *
     * @param holdability a <code>ResultSet</code> holdability constant; one of
     *        <code>ResultSet.HOLD_CURSORS_OVER_COMMIT</code> or
     *        <code>ResultSet.CLOSE_CURSORS_AT_COMMIT</code>
     * @throws SQLException if a database access occurs, the given parameter
     *         is not a <code>ResultSet</code> constant indicating holdability,
     *         or the given holdability is not supported
     * @see #getHoldability
     * @see ResultSet
     * @since 1.4
     */
    public void setHoldability(int holdability) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.setHoldability(int holdability)");
    }

    /**
     * Retrieves the current holdability of <code>ResultSet</code> objects
     * created using this <code>Connection</code> object.
     *
     * @return the holdability, one of
     *        <code>ResultSet.HOLD_CURSORS_OVER_COMMIT</code> or
     *        <code>ResultSet.CLOSE_CURSORS_AT_COMMIT</code>
     * @throws SQLException if a database access occurs
     * @see #setHoldability
     * @see ResultSet
     * @since 1.4
     */
    public int getHoldability() throws SQLException
    {
	return 2; // CLOSE_CURSORS_AT_COMMIT
    }


    /**
     * Creates an unnamed savepoint in the current transaction and 
     * returns the new <code>Savepoint</code> object that represents it.
     *
     * @return the new <code>Savepoint</code> object
     * @exception SQLException if a database access error occurs
     *            or this <code>Connection</code> object is currently in
     *            auto-commit mode
     * @see Savepoint
     * @since 1.4
     */
    public Savepoint setSavepoint() throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.setSavepoint()");
    }

    /**
     * Creates a savepoint with the given name in the current transaction
     * and returns the new <code>Savepoint</code> object that represents it.
     *
     * @param name a <code>String</code> containing the name of the savepoint
     * @return the new <code>Savepoint</code> object
     * @exception SQLException if a database access error occurs
     *            or this <code>Connection</code> object is currently in
     *            auto-commit mode
     * @see Savepoint
     * @since 1.4
     */
    public Savepoint setSavepoint(String name) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.setSavepoint(String name)");
    }

    /**
     * Undoes all changes made after the given <code>Savepoint</code> object
     * was set. 
     * <P>
     * This method should be used only when auto-commit has been disabled.
     *
     * @param savepoint the <code>Savepoint</code> object to roll back to
     * @exception SQLException if a database access error occurs,
     *            the <code>Savepoint</code> object is no longer valid,
     *            or this <code>Connection</code> object is currently in
     *            auto-commit mode
     * @see Savepoint
     * @see #rollback
     * @since 1.4
     */
    public void rollback(Savepoint savepoint) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.rollback(Savepoint savepoint)");
    }

    /**
     * Removes the given <code>Savepoint</code> object from the current 
     * transaction. Any reference to the savepoint after it have been removed 
     * will cause an <code>SQLException</code> to be thrown.
     *
     * @param savepoint the <code>Savepoint</code> object to be removed
     * @exception SQLException if a database access error occurs or
     *            the given <code>Savepoint</code> object is not a valid 
     *            savepoint in the current transaction
     * @since 1.4
     */
    public void releaseSavepoint(Savepoint savepoint) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.releaseSavepoint(Savepoint savepoint)");
    }
    
    /**
     * Creates a <code>Statement</code> object that will generate
     * <code>ResultSet</code> objects with the given type, concurrency,
     * and holdability.
     * This method is the same as the <code>createStatement</code> method
     * above, but it allows the default result set
     * type, concurrency, and holdability to be overridden.
     *
     * @param resultSetType one of the following <code>ResultSet</code> 
     *        constants:
     *         <code>ResultSet.TYPE_FORWARD_ONLY</code>, 
     *         <code>ResultSet.TYPE_SCROLL_INSENSITIVE</code>, or
     *         <code>ResultSet.TYPE_SCROLL_SENSITIVE</code>
     * @param resultSetConcurrency one of the following <code>ResultSet</code> 
     *        constants:
     *         <code>ResultSet.CONCUR_READ_ONLY</code> or
     *         <code>ResultSet.CONCUR_UPDATABLE</code>
     * @param resultSetHoldability one of the following <code>ResultSet</code> 
     *        constants:
     *         <code>ResultSet.HOLD_CURSORS_OVER_COMMIT</code> or
     *         <code>ResultSet.CLOSE_CURSORS_AT_COMMIT</code>
     * @return a new <code>Statement</code> object that will generate
     *         <code>ResultSet</code> objects with the given type,
     *         concurrency, and holdability
     * @exception SQLException if a database access error occurs
     *            or the given parameters are not <code>ResultSet</code> 
     *            constants indicating type, concurrency, and holdability
     * @see ResultSet
     * @since 1.4
     */
    public Statement createStatement(int resultSetType, int resultSetConcurrency, 
				     int resultSetHoldability) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.createStatement(int resultSetType, int resultSetConcurrency, int resultSetHoldability)");
    }

    /**
     * Creates a <code>PreparedStatement</code> object that will generate
     * <code>ResultSet</code> objects with the given type, concurrency,
     * and holdability.
     * <P>
     * This method is the same as the <code>prepareStatement</code> method
     * above, but it allows the default result set
     * type, concurrency, and holdability to be overridden.
     *
     * @param sql a <code>String</code> object that is the SQL statement to
     *            be sent to the database; may contain one or more ? IN
     *            parameters
     * @param resultSetType one of the following <code>ResultSet</code> 
     *        constants:
     *         <code>ResultSet.TYPE_FORWARD_ONLY</code>, 
     *         <code>ResultSet.TYPE_SCROLL_INSENSITIVE</code>, or
     *         <code>ResultSet.TYPE_SCROLL_SENSITIVE</code>
     * @param resultSetConcurrency one of the following <code>ResultSet</code> 
     *        constants:
     *         <code>ResultSet.CONCUR_READ_ONLY</code> or
     *         <code>ResultSet.CONCUR_UPDATABLE</code>
     * @param resultSetHoldability one of the following <code>ResultSet</code> 
     *        constants:
     *         <code>ResultSet.HOLD_CURSORS_OVER_COMMIT</code> or
     *         <code>ResultSet.CLOSE_CURSORS_AT_COMMIT</code>
     * @return a new <code>PreparedStatement</code> object, containing the
     *         pre-compiled SQL statement, that will generate
     *         <code>ResultSet</code> objects with the given type,
     *         concurrency, and holdability
     * @exception SQLException if a database access error occurs
     *            or the given parameters are not <code>ResultSet</code> 
     *            constants indicating type, concurrency, and holdability
     * @see ResultSet
     * @since 1.4
     */
    public PreparedStatement prepareStatement(String sql, int resultSetType, 
					      int resultSetConcurrency, int resultSetHoldability)
	throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.prepareStatement(String sql, int resultSetType, int resultSetConcurrency, int resultSetHoldability)");
    }

    /**
     * Creates a <code>CallableStatement</code> object that will generate
     * <code>ResultSet</code> objects with the given type and concurrency.
     * This method is the same as the <code>prepareCall</code> method
     * above, but it allows the default result set
     * type, result set concurrency type and holdability to be overridden.
     *
     * @param sql a <code>String</code> object that is the SQL statement to
     *            be sent to the database; may contain on or more ? parameters
     * @param resultSetType one of the following <code>ResultSet</code> 
     *        constants:
     *         <code>ResultSet.TYPE_FORWARD_ONLY</code>, 
     *         <code>ResultSet.TYPE_SCROLL_INSENSITIVE</code>, or
     *         <code>ResultSet.TYPE_SCROLL_SENSITIVE</code>
     * @param resultSetConcurrency one of the following <code>ResultSet</code> 
     *        constants:
     *         <code>ResultSet.CONCUR_READ_ONLY</code> or
     *         <code>ResultSet.CONCUR_UPDATABLE</code>
     * @param resultSetHoldability one of the following <code>ResultSet</code> 
     *        constants:
     *         <code>ResultSet.HOLD_CURSORS_OVER_COMMIT</code> or
     *         <code>ResultSet.CLOSE_CURSORS_AT_COMMIT</code>
     * @return a new <code>CallableStatement</code> object, containing the
     *         pre-compiled SQL statement, that will generate
     *         <code>ResultSet</code> objects with the given type,
     *         concurrency, and holdability
     * @exception SQLException if a database access error occurs
     *            or the given parameters are not <code>ResultSet</code> 
     *            constants indicating type, concurrency, and holdability
     * @see ResultSet
     * @since 1.4
     */
    public CallableStatement prepareCall(String sql, int resultSetType, 
					 int resultSetConcurrency, 
					 int resultSetHoldability) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.prepareCall(String sql, int resultSetType, int resultSetConcurrency, int resultSetHoldability)");
    }

    /**
     * Creates a default <code>PreparedStatement</code> object that has
     * the capability to retrieve auto-generated keys. The given constant
     * tells the driver whether it should make auto-generated keys
     * available for retrieval.  This parameter is ignored if the SQL 
     * statement is not an <code>INSERT</code> statement.
     * <P>
     * <B>Note:</B> This method is optimized for handling
     * parametric SQL statements that benefit from precompilation. If
     * the driver supports precompilation,
     * the method <code>prepareStatement</code> will send
     * the statement to the database for precompilation. Some drivers
     * may not support precompilation. In this case, the statement may
     * not be sent to the database until the <code>PreparedStatement</code> 
     * object is executed.  This has no direct effect on users; however, it does
     * affect which methods throw certain SQLExceptions.
     * <P>
     * Result sets created using the returned <code>PreparedStatement</code>
     * object will by default be type <code>TYPE_FORWARD_ONLY</code>
     * and have a concurrency level of <code>CONCUR_READ_ONLY</code>.
     *
     * @param sql an SQL statement that may contain one or more '?' IN
     *        parameter placeholders
     * @param autoGeneratedKeys a flag indicating whether auto-generated keys 
     *        should be returned; one of the following <code>Statement</code>
     *        constants:
     * @param autoGeneratedKeys a flag indicating that auto-generated keys should be returned, one of
     *        <code>Statement.RETURN_GENERATED_KEYS</code> or
     *	      <code>Statement.NO_GENERATED_KEYS</code>.  
     * @return a new <code>PreparedStatement</code> object, containing the
     *         pre-compiled SQL statement, that will have the capability of
     *         returning auto-generated keys
     * @exception SQLException if a database access error occurs
     *         or the given parameter is not a <code>Statement</code>
     *         constant indicating whether auto-generated keys should be
     *         returned
     * @since 1.4
     */
    public PreparedStatement prepareStatement(String sql, int autoGeneratedKeys)
	throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.prepareStatement(String sql, int autoGeneratedKeys)");
    }

    /**
     * Creates a default <code>PreparedStatement</code> object capable
     * of returning the auto-generated keys designated by the given array.
     * This array contains the indexes of the columns in the target
     * table that contain the auto-generated keys that should be made
     * available. This array is ignored if the SQL 
     * statement is not an <code>INSERT</code> statement.
     * <P>
     * An SQL statement with or without IN parameters can be
     * pre-compiled and stored in a <code>PreparedStatement</code> object. This
     * object can then be used to efficiently execute this statement
     * multiple times.
     * <P>
     * <B>Note:</B> This method is optimized for handling
     * parametric SQL statements that benefit from precompilation. If
     * the driver supports precompilation,
     * the method <code>prepareStatement</code> will send
     * the statement to the database for precompilation. Some drivers
     * may not support precompilation. In this case, the statement may
     * not be sent to the database until the <code>PreparedStatement</code> 
     * object is executed.  This has no direct effect on users; however, it does
     * affect which methods throw certain SQLExceptions.
     * <P>
     * Result sets created using the returned <code>PreparedStatement</code>
     * object will by default be type <code>TYPE_FORWARD_ONLY</code>
     * and have a concurrency level of <code>CONCUR_READ_ONLY</code>.
     *
     * @param sql an SQL statement that may contain one or more '?' IN
     *        parameter placeholders
     * @param columnIndexes an array of column indexes indicating the columns
     *        that should be returned from the inserted row or rows 
     * @return a new <code>PreparedStatement</code> object, containing the
     *         pre-compiled statement, that is capable of returning the
     *         auto-generated keys designated by the given array of column
     *         indexes
     * @exception SQLException if a database access error occurs
     *
     * @since 1.4
     */
    public PreparedStatement prepareStatement(String sql, int columnIndexes[])
	throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.prepareStatement(String sql, int columnIndexes[])");
    }

    /**
     * Creates a default <code>PreparedStatement</code> object capable
     * of returning the auto-generated keys designated by the given array.
     * This array contains the names of the columns in the target
     * table that contain the auto-generated keys that should be returned.
     * This array is ignored if the SQL 
     * statement is not an <code>INSERT</code> statement.
     * <P>
     * An SQL statement with or without IN parameters can be
     * pre-compiled and stored in a <code>PreparedStatement</code> object. This
     * object can then be used to efficiently execute this statement
     * multiple times.
     * <P>
     * <B>Note:</B> This method is optimized for handling
     * parametric SQL statements that benefit from precompilation. If
     * the driver supports precompilation,
     * the method <code>prepareStatement</code> will send
     * the statement to the database for precompilation. Some drivers
     * may not support precompilation. In this case, the statement may
     * not be sent to the database until the <code>PreparedStatement</code> 
     * object is executed.  This has no direct effect on users; however, it does
     * affect which methods throw certain SQLExceptions.
     * <P>
     * Result sets created using the returned <code>PreparedStatement</code>
     * object will by default be type <code>TYPE_FORWARD_ONLY</code>
     * and have a concurrency level of <code>CONCUR_READ_ONLY</code>.
     *
     * @param sql an SQL statement that may contain one or more '?' IN
     *        parameter placeholders
     * @param columnNames an array of column names indicating the columns
     *        that should be returned from the inserted row or rows 
     * @return a new <code>PreparedStatement</code> object, containing the
     *         pre-compiled statement, that is capable of returning the
     *         auto-generated keys designated by the given array of column
     *         names
     * @exception SQLException if a database access error occurs
     *
     * @since 1.4
     */
    public PreparedStatement prepareStatement(String sql, String columnNames[])
	throws SQLException
    {
	throw new ConsusNotImplemented("ConsusConnection.prepareStatement(String sql, String columnNames[])");
    }


    /**************************************************
     * C O N S U S  specific methods
     **************************************************/
      
    /** 
     * Insert specified object in the database. Table in which object will be inserted
     * is determined by the object class.
     * @param obj Java object to be inserted.
     * @return reference to the cerated object
     */
    public Ref insert(Object obj) throws SQLException { 
	if (obj instanceof Persistent) { 
	    Persistent po = (Persistent)obj;
	    po.store(this);
	    return po.getRef();
	}
	beginTransaction(true);
	TableDescriptor table = getTableByClass(obj.getClass());
	int size = table.calculateRecordSize(obj);
	byte[] rec = new byte[size];
	table.allocateBlobs(rec, null, obj);
	table.storeObject(rec, obj);
	int oid = allocateRow(table, rec);
	if (!table.insertInIndices(oid)) {
	    freeRow(table, oid);
	    throw new ConsusDuplicateException();
	} else { 
	    if (objectCache != null) { 
		objectCache.put(oid, obj);
	    }
	    return new ConsusRef(this, oid);
	}
    }

    /**
     * Update object in the table. 
     *
     * @param ref reference to the updated object
     * @param obj object with new values 
     */
    public void update(Ref ref, Object obj) throws SQLException { 
	beginTransaction(true);
	int oid = (int)((ConsusRef)ref).getOid();
	TableDescriptor table = getTableByClass(obj.getClass());
	int size = table.calculateRecordSize(obj);
	byte[] newRec = new byte[size];	
	byte[] oldRec = get(oid);
	if (Record.getType(oldRec, 0) != table.tableId) { 
	    throw new ConsusRuntimeError("Type of the object can not be changed");
	} 
	table.allocateBlobs(newRec, oldRec, obj);
	table.storeObject(newRec, obj);
	boolean updated[] = new boolean[table.nFields];
	table.markUpdatedFields(newRec, oldRec, updated);
	table.removeFromIndices(oid, updated);
	put(oid, newRec);
	if (!table.updateIndices(oid, updated)) { 
	    put(oid, oldRec);
	    table.updateIndices(oid, updated);
	    throw new ConsusDuplicateException();
	}
	if (objectCache != null) { 
	    objectCache.put(oid, obj);
	}
    }

    /**
     * Get object by OID
     *
     * @param ref reference to the object
     * @return reference to the fetched object
     */
    public Object get(Ref ref) throws SQLException { 
	Object obj;
	if (ref == null) { 
	    return null;
	}
	int oid = (int)((ConsusRef)ref).getOid();
	if (oid == 0) {
	    return null;
	}
	beginTransaction(false);
	if (objectCache != null) { 
	    synchronized (objectCache) { 
		obj = objectCache.get(oid);
		if (obj == null) { 
		    byte[] rec = get(oid);
		    TableDescriptor table = getTableByOid(Record.getType(rec, 0));
		    Assert.that(table != null);
		    obj = table.fetchObject(oid, rec);
		} else if (obj instanceof Persistent 
			   && ((Persistent)obj).state == Persistent.PO_STUB)
		{
		    ((Persistent)obj).load();
		}
	    }
	} else { 
	    byte[] rec = get(oid);
	    TableDescriptor table = getTableByOid(Record.getType(rec, 0));
	    Assert.that(table != null);
	    obj = table.fetchObject(oid, rec);
	}
	return obj;
    }

    /**
     * Remove object with specifed OID
     *
     * @param ref reference to the removed object
     */
    public void remove(Ref ref) throws SQLException { 
	if (ref == null) { 
	    return;
	}
	int oid = (int)((ConsusRef)ref).getOid();
	if (oid == 0) {
	    return;
	}
	beginTransaction(true);
	TableDescriptor table = getTableByOid(getType(oid));
	Assert.that(table != null);
	freeRow(table, oid);
    }

    /*
     * PersistentObjectStorage methods
     */
    public boolean isInitialized() throws Exception
    {
	return true;
    }
    
    public void open()  throws Exception
    {
    }

    public Object getObjectByOid(long oid) throws Exception {
	return get(new ConsusRef(this, oid));
    }

    public void loadObject(Persistent obj) throws Exception { 
	beginTransaction(false);
	byte[] rec = get((int)obj.oid);
	int tableId = Record.getType(rec, 0);
	TableDescriptor table = (TableDescriptor)obj.type;
	if (tableId != table.tableId) { 
	    obj.type = table = getTableByOid(tableId);
	    Assert.that(table != null);
	}
	table.fetchObject(obj, rec);
    }

    public Object fetchObject(ResultSet cursor) throws Exception 
    {
	return ((ConsusResultSet)cursor).getSelfObject();
    } 

    public void updateObject(Object obj, ResultSet cursor) throws Exception
    {	
	((ConsusResultSet)cursor).updateObject(0, obj);
	cursor.updateRow();
    }
    

    public void loadComponents(Persistent obj) throws Exception
    {
	TableDescriptor table = (TableDescriptor)obj.type;
	table.loadComponents(obj);
    }	

    public void clearComponents(Persistent obj) throws Exception
    {
	TableDescriptor table = (TableDescriptor)obj.type;
	for (int i = table.components.length; --i>= 0;) { 
	    FieldDescriptor fd = table.components[i];
	    if (fd.type > FieldDescriptor.tpReal8) { // non scalar
		fd.field.set(obj, null);
	    }
	}
    } 

    public void storeObject(Persistent obj) throws Exception
    {
	beginTransaction(true);
	TableDescriptor table = (TableDescriptor)obj.type;
	if (table == null) { 
	    obj.type = table = getTableByClass(obj.getClass());
	}
	int size = table.calculateRecordSize(obj);
	byte[] newRec = new byte[size];	
	int oid = (int)obj.oid;
	if (oid == 0) { 
	    table.allocateBlobs(newRec, null, obj);
	    oid = allocateId();
	    obj.oid = oid;
	    obj.state = Persistent.PO_LOADED;
	    objectCache.put(oid, obj);
	    table.storeObject(newRec, obj);
	    allocateRow(oid, table, newRec);
	    if (!table.insertInIndices(oid)) { 
		freeRow(table, oid);
		throw new ConsusDuplicateException();
	    }
	} else { 
	    byte[] oldRec = get(oid);
	    table.allocateBlobs(newRec, oldRec, obj);
	    table.storeObject(newRec, obj);
	    boolean updated[] = new boolean[table.nFields];
	    table.markUpdatedFields(newRec, oldRec, updated);
	    table.removeFromIndices(oid, updated);
	    put(oid, newRec);
	    if (!table.updateIndices(oid, updated)) {
		put(oid, oldRec);
		table.updateIndices(oid, updated);
		throw new ConsusDuplicateException();
	    }		    
	}
    }
    
    public void deleteObject(Persistent obj) throws Exception 
    {
	TableDescriptor table = (TableDescriptor)obj.type;
	beginTransaction(true);
	freeRow(table, (int)obj.oid);
    }   

    public void becomeObject(Persistent a, Persistent b) throws Exception 
    {
	beginTransaction(true);
	int srcOID = (int)a.oid;
	int dstOID = (int)b.assignOID(this);
	a.oid = dstOID;
	b.oid = srcOID;
	long srcPos = getPos(srcOID);
	long dstPos = getPos(dstOID);
	objectCache.put(srcOID, b);
	objectCache.put(dstOID, a);
	setPos(srcOID, (dstPos & ~ConsusConnection.dbFlagsMask) | (srcPos & ConsusConnection.dbFlagsMask));
	setPos(dstOID, (srcPos & ~ConsusConnection.dbFlagsMask) | (dstPos & ConsusConnection.dbFlagsMask));
    }


     /**
      * Create new String index in the database. 
      */
     public org.garret.consus.StringIndex createStringIndex() throws SQLException
     {
	 return new StringIndex(this);
     }

     /**
      *  Get table iterator
      */
     public org.garret.consus.TableIterator getTableIterator(String tableName)
	 throws Exception
     {
	 return new TableIterator(this, tableName);
     }

    /**
     * Get the name of the field containing referenence to the master object instance 
     * at the server (used for table replication)
     * if <code>null</code> thyen existed primary key will be used instead
     */
    public String getMasterFieldName()
    {
	return masterOID;
    }

    /**
     * Set name of the field containing refenence to the master object instance at the 
     * server (used for table replication)
     * @param <code>name</code> - name of the field, if <code>null</code>
     * then  some existed primary key will be used instead
     */
    public void setMasterFieldName(String name)
    {
	masterOID = name;
    }

    /**
     * Get the name of the timestamp field used for data replication
     * @return timestamp field name (recordTimestamp" by default)
     */
    public String getTimestampFieldName()
    {
	return recordTimestamp;
    }

    /**
     * Set name of the timestamp field used for data replication
     * @param <code>name</code> - name of the timestamp field
     */
    public void setTimestampFieldName(String name)
    {
	recordTimestamp = name;
    }

    /**
     * Name of the field containing refenence to the master object instance at the server 
     * (used for table replication)
     * Set it to null to use some existed primary key instead
     */
    protected String masterOID = "masterOID"; 
    
    /**
     * Name of the timestamp field used for data replication
     */
    protected String recordTimestamp = "recordTimestamp";

    /***********************************************************************
     * C O N S U S internal methods
     ***********************************************************************/

    /**
     * Size of page pool (number of pages in the pool).
     * Should be set before connection to the driver.
     */
    public static int dbPagePoolSize = 1024; // 4Mb page pool
    
    /**
     * Initialial database index size - increasing it reduce number of inde reallocation but increase
     * initial database size. Should be set before openning connection.
     */
    public static int        dbDefaultInitIndexSize = 1024;

    /**
     * Database extension quantum. Memory is allocate by scanning bitmap. If there is no
     * large enough hole, then database is extended by the value of dbDefaultExtensionQuantum 
     * This parameter should not be smaller than dbFirstUserId
     */
    public static long       dbDefaultExtensionQuantum = 4*1024*1024;

    /**
     * Initial capacity of object hash
     */
    static int dbObjectCacheInitSize = 1319;

    /** 
     * Enable database tracing 
     */
    public boolean    dbTraceEnabled = Boolean.getBoolean("consus.trace");

    static final int  dbDatabaseOffsetBits = 32;  // up to 4 gigabyte, 37 - up to 128 terabyte database

    static final int  dbAllocationQuantumBits = 6;
    static final int  dbAllocationQuantum = 1 << dbAllocationQuantumBits;
    static final int  dbBitmapSegmentBits = Page.pageBits + 3 + dbAllocationQuantumBits;
    static final int  dbBitmapSegmentSize = 1 << dbBitmapSegmentBits;
    static final int  dbBitmapPages = 1 << (dbDatabaseOffsetBits-dbBitmapSegmentBits);
    static final int  dbHandlesPerPageBits = Page.pageBits - 3;
    static final int  dbHandlesPerPage = 1 << dbHandlesPerPageBits;
    static final int  dbDirtyPageBitmapSize = 1 << (32-Page.pageBits-3);

    static final int  dbInvalidId   = 0;
    static final int  dbMetaTableId = 1; 
    static final int  dbBitmapId    = 2;
    static final int  dbFirstUserId = dbBitmapId + dbBitmapPages;
    
    static final int  dbPageObjectFlag = 1;
    static final int  dbModifiedFlag   = 2;
    static final int  dbFreeHandleFlag = 4;
    static final int  dbFlagsMask      = 7;
    static final int  dbFlagsBits      = 3;


    final long getPos(int oid) throws SQLException { 
	if (oid == 0 && oid >= currIndexSize) { 
	    throw new InvalidOIDError(oid);
	}
	Page pg = pool.getPage(header.root[1-currIndex].index 
			       + (oid >>> dbHandlesPerPageBits << Page.pageBits));
	long pos = Bytes.unpack8(pg.data, (oid & (dbHandlesPerPage-1)) << 3);
	pool.unfix(pg);
	return pos;
    }
    
    final void setPos(int oid, long pos) throws SQLException { 
	dirtyPagesMap[oid >>> (dbHandlesPerPageBits+5)] 
	    |= 1 << ((oid >>> dbHandlesPerPageBits) & 31);
	Page pg = pool.putPage(header.root[1-currIndex].index 
			       + (oid >>> dbHandlesPerPageBits << Page.pageBits));
	Bytes.pack8(pg.data, (oid & (dbHandlesPerPage-1)) << 3, pos);
	pool.unfix(pg);
    }

    final byte[] get(int oid) throws SQLException { 
	long pos = getPos(oid);
	if ((pos & (dbFreeHandleFlag|dbPageObjectFlag)) != 0) { 
	    throw new InvalidOIDError(oid, pos);
	}
	return pool.get(pos & ~dbFlagsMask);
    }
    
    final Page getPage(int oid) throws SQLException {  
	long pos = getPos(oid);
	if ((pos & (dbFreeHandleFlag|dbPageObjectFlag)) != dbPageObjectFlag) { 
	    throw new InvalidOIDError(oid, pos);
	}
	return pool.getPage(pos & ~dbFlagsMask);
    }

    final Page putPage(int oid) throws SQLException {  
	long pos = getPos(oid);
	if ((pos & (dbFreeHandleFlag|dbPageObjectFlag)) != dbPageObjectFlag) { 
	    throw new InvalidOIDError(oid, pos);
	}
	if ((pos & dbModifiedFlag) == 0) { 
	    dirtyPagesMap[oid >>> (dbHandlesPerPageBits+5)] 
		|= 1 << ((oid >>> dbHandlesPerPageBits) & 31);
	    allocate(Page.pageSize, oid);
	    cloneBitmap(pos & ~dbFlagsMask, Page.pageSize);
	    pos = getPos(oid);
	}
	modified = true;
	return pool.putPage(pos & ~dbFlagsMask);
    }


    final void update(int oid, byte[] rec) throws SQLException { 
	long pos = getPos(oid);
	if ((pos & dbModifiedFlag) == 0) { 
	    cloneBitmap(pos & ~dbFlagsMask, rec.length);
	    pos = allocate(rec.length, 0);
	    setPos(oid, pos | dbModifiedFlag);
	}	
	modified = true;
	pool.put(pos & ~dbFlagsMask, rec);
    }

    final void put(int oid, byte[] rec) throws SQLException { 
	long pos = getPos(oid);
	if ((pos & (dbFreeHandleFlag|dbPageObjectFlag)) != 0) { 
	    throw new InvalidOIDError(oid, pos);
	}
	int offs = (int)pos & (Page.pageSize-1);
	Page pg = pool.getPage(pos - offs);
	offs &= ~dbFlagsMask;
	int size = Record.getSize(pg.data, offs);
	int newSize = rec.length;
	modified = true;
	if ((pos & dbModifiedFlag) == 0) { 
	    cloneBitmap(pos & ~dbFlagsMask, size);
	    pos = allocate(newSize, 0);
	    setPos(oid, pos | dbModifiedFlag);
	} else { 
	    if (((newSize + dbAllocationQuantum - 1) & ~(dbAllocationQuantum-1))
		> ((size + dbAllocationQuantum - 1) & ~(dbAllocationQuantum-1)))
	    { 
		long newPos = allocate(newSize, 0);
		cloneBitmap(pos & ~dbFlagsMask, size);
		free(pos & ~dbFlagsMask, size);
		pos = newPos;
		setPos(oid, pos | dbModifiedFlag);
	    } else if (newSize < size) { 
		newSize = size;
	    }
	}
	Record.setSize(rec, 0, newSize);
	Record.setType(rec, 0, Record.getType(pg.data, offs));
	Record.setNext(rec, 0, Record.getNext(pg.data, offs));
	Record.setPrev(rec, 0, Record.getPrev(pg.data, offs));
	pool.unfix(pg);
	pool.put(pos & ~dbFlagsMask, rec);
    }
    

    int allocateRow(TableDescriptor table, byte[] rec) throws SQLException { 
	int oid = allocateId();
	allocateRow(oid, table, rec);
	return oid;
    }
			   

    void allocateRow(int oid, TableDescriptor table, byte[] rec) throws SQLException
    { 
	if (table.tableId == 0) { 
	    table.registerInDatabase();
	    resetForeignKeyConstraints();	    
	}
	long pos = allocate(rec.length, 0);
	setPos(oid, pos | dbModifiedFlag);
	Record.setSize(rec, 0, rec.length);
	Record.setType(rec, 0, table.tableId);
	while (true) { 
	    table.nRows += 1;
	    table.updated = true;
	    if (table.superClass == null) {
		break;
	    }
	    table = table.superClass;
	} 
	Record.setNext(rec, 0, 0);
	Record.setPrev(rec, 0, table.lastRow);
	if (table.lastRow != 0) { 
	    long lastPos = getPos(table.lastRow);
	    int offs = (int)lastPos & (Page.pageSize-1);
	    Assert.that((offs & (dbFreeHandleFlag|dbPageObjectFlag)) == 0);
	    Page pg = pool.putPage(lastPos - offs);
	    Record.setNext(pg.data, offs & ~dbFlagsMask, oid); 
	    pool.unfix(pg);
	} else { 
	    table.firstRow = oid;
	}
	table.lastRow = oid;
	pool.put(pos, rec);
	modified = true;
    }

    void freeRow(TableDescriptor table, int oid) throws SQLException { 
	long pos = getPos(oid);
	int offs = (int)pos & (Page.pageSize-1);
	Assert.that((offs & (dbFreeHandleFlag|dbPageObjectFlag)) == 0);
	Page pg = pool.getPage(pos - offs);
	offs &= ~dbFlagsMask;
	int prev = Record.getPrev(pg.data, offs); 
	int next = Record.getNext(pg.data, offs); 
	int size = Record.getSize(pg.data, offs);
	pool.unfix(pg);
	for (FieldDescriptor fd = table.fields; fd != null; fd = fd.next) { 
	    if (fd.index != 0) {
		Btree.remove(this, fd, oid);
	    }
	    if (fd.type == FieldDescriptor.tpBlob || fd.type == FieldDescriptor.tpClob)
	    { 
		byte[] rec = get(oid);
		long blobPos = Bytes.unpack8(rec, fd.offs);		
		if (blobPos != 0) { 
		    long blobSize = Bytes.unpack8(rec, fd.offs+8);

		    cloneBitmap(blobPos, blobSize);
		    free(blobPos, blobSize);
		}
	    }
	}
	while (true) { 
	    table.nRows -= 1;
	    table.updated = true;
	    if (table.superClass == null) {
		break;
	    }
	    table = table.superClass;
	} 
	if (prev != 0) { 
	    byte[] rec = get(prev);
	    Record.setNext(rec, 0, next);
	    update(prev, rec);
	} else { 
	    table.firstRow = next;
	}
	if (next != 0) { 
	    byte[] rec = get(next);
	    Record.setPrev(rec, 0, prev);
	    update(next, rec);
	} else { 
	    table.lastRow = prev;
	}
	if (objectCache != null) { 
	    objectCache.remove(oid);
	}
	for (FieldDescriptor fd = table.fields; fd != null; fd = fd.next) { 
	    if (fd.foreignKeys != null) { 
		cascadeDelete(fd.foreignKeys, fd.offs, oid);
	    }
	}
	cascadeDelete(table.foreignKeys, 0, oid);
	freeId(oid);
	if ((pos & dbModifiedFlag) != 0) { 
	    free(pos & ~dbFlagsMask, size);
	} else { 
	    cloneBitmap(pos, size);
	}
	modified = true;
    }

    final void cascadeDelete(FieldDescriptor fk, int offs, int oid) throws SQLException
    { 
	while (fk != null) { 
	    PreparedStatement pstmt = fk.cascadeDeleteStmt;
	    if (pstmt == null) { 
		pstmt = prepareStatement("delete from " + fk.table.name 
					 + " where " + fk.name + " = ?");
		fk.cascadeDeleteStmt = pstmt;
	    }
	    if (fk.fkKeyName.equals("oid")) { 
		pstmt.setRef(1, new ConsusRef(this, oid));
	    } else { 
		byte[] rec = get(oid);
		switch (fk.type) { 
		  case FieldDescriptor.tpInt1:
		    pstmt.setByte(1, rec[offs]);
		    break;
		  case FieldDescriptor.tpInt2:
		    pstmt.setShort(1, Bytes.unpack2(rec, offs));
		    break;
		  case FieldDescriptor.tpInt4:
		    pstmt.setInt(1, Bytes.unpack4(rec, offs));
		    break;
		  case FieldDescriptor.tpReal4:
		    pstmt.setFloat(1, Float.intBitsToFloat(Bytes.unpack4(rec, offs)));
		    break;
		  case FieldDescriptor.tpReal8:
		    pstmt.setDouble(1, Double.longBitsToDouble(Bytes.unpack8(rec, offs)));
		    break;
		  case FieldDescriptor.tpDate:
		  case FieldDescriptor.tpTime:
		  case FieldDescriptor.tpTimestamp:
		  case FieldDescriptor.tpInt8:
		    pstmt.setLong(1, Bytes.unpack8(rec, offs));
		    break;
		  case FieldDescriptor.tpReference:
		    pstmt.setRef(1, new ConsusRef(this, Bytes.unpack4(rec, offs)));
		    break;
		  case FieldDescriptor.tpString:
		    pstmt.setString(1, Varying.getString(rec, offs));
		    break;
		  default:
		    throw new ConsusRuntimeError("Invalid foreign key type " + fk.type);
		}
	    }
	    pstmt.executeUpdate();
	    fk = fk.nextForeignKey;
	}
    }	


    int allocatePage() throws SQLException { 
	int oid = allocateId();
	setPos(oid, allocate(Page.pageSize, 0) | dbPageObjectFlag | dbModifiedFlag);
	return oid;
    }

    void freePage(int oid) throws SQLException { 
	long pos = getPos(oid);
	if ((pos & dbModifiedFlag) != 0) { 
	    free(pos & ~dbFlagsMask, Page.pageSize);
	} else { 
	    cloneBitmap(pos & ~dbFlagsMask, Page.pageSize);
	} 
	freeId(oid);
    }

    final static byte firstHoleSize [] = {
	8,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	7,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0
    };
    final static byte lastHoleSize [] = {
        8,7,6,6,5,5,5,5,4,4,4,4,4,4,4,4,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
	2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    };
    final static byte maxHoleSize [] = {
	8,7,6,6,5,5,5,5,4,4,4,4,4,4,4,4,4,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
	5,4,3,3,2,2,2,2,3,2,2,2,2,2,2,2,4,3,2,2,2,2,2,2,3,2,2,2,2,2,2,2,
	6,5,4,4,3,3,3,3,3,2,2,2,2,2,2,2,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,1,
	5,4,3,3,2,2,2,2,3,2,1,1,2,1,1,1,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,1,
	7,6,5,5,4,4,4,4,3,3,3,3,3,3,3,3,4,3,2,2,2,2,2,2,3,2,2,2,2,2,2,2,
	5,4,3,3,2,2,2,2,3,2,1,1,2,1,1,1,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,1,
	6,5,4,4,3,3,3,3,3,2,2,2,2,2,2,2,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,1,
	5,4,3,3,2,2,2,2,3,2,1,1,2,1,1,1,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,0
    };
    final static byte maxHoleOffset [] = {
	0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,0,1,5,5,5,5,5,5,0,5,5,5,5,5,5,5,
        0,1,2,2,0,3,3,3,0,1,6,6,0,6,6,6,0,1,2,2,0,6,6,6,0,1,6,6,0,6,6,6,
        0,1,2,2,3,3,3,3,0,1,4,4,0,4,4,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,5,
	0,1,2,2,0,3,3,3,0,1,0,2,0,1,0,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,7,
	0,1,2,2,3,3,3,3,0,4,4,4,4,4,4,4,0,1,2,2,0,5,5,5,0,1,5,5,0,5,5,5,
	0,1,2,2,0,3,3,3,0,1,0,2,0,1,0,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,6,
	0,1,2,2,3,3,3,3,0,1,4,4,0,4,4,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,5,
	0,1,2,2,0,3,3,3,0,1,0,2,0,1,0,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,0
    };

    static final int pageBits = Page.pageSize*8;
    static final int inc = Page.pageSize/dbAllocationQuantum/8;

    static final void memset(Page pg, int offs, int pattern, int len) { 
	byte[] arr = pg.data;
	byte pat = (byte)pattern;
	while (--len >= 0) { 
	    arr[offs++] = pat;
	}
    }

    final void extend(long size)
    {
	if (size > header.root[1-currIndex].size) { 
	    header.root[1-currIndex].size = size;
	}
    }

    static class dbLocation { 
	long       pos;
	long       size;
	dbLocation next;
    };

    final boolean wasReserved(long pos, long size) 
    {
	for (dbLocation location = reservedChain; location != null; location = location.next) { 
	    if ((pos >= location.pos && pos - location.pos < location.size) 
		|| (pos <= location.pos && location.pos - pos < size)) 
	    {
		return true;
	    }
	}
	return false;
    }

    final void reserveLocation(long pos, long size)
    {
	dbLocation location = new dbLocation();
	location.pos = pos;
	location.size = size;
	location.next = reservedChain;
	reservedChain = location;
    }

    final void commitLocation()
    {
	reservedChain = reservedChain.next;
    }


    long allocate(long size, int oid) throws SQLException
    {
	size = (size + dbAllocationQuantum-1) & ~(dbAllocationQuantum-1);
	Assert.that(size != 0);
	int  objBitSize = (int)(size >> dbAllocationQuantumBits);
	long pos;    
	int  holeBitSize = 0;
	int  alignment = (int)size & (Page.pageSize-1);
	int  offs, firstPage, lastPage, i;
	int  holeBeforeFreePage  = 0;
	int  freeBitmapPage = 0;
	Page pg;

	lastPage = header.root[1-currIndex].bitmapEnd;
	usedSize += size;

	if (alignment == 0) { 
	    firstPage = currPBitmapPage;
	    offs = (currPBitmapOffs+inc-1) & ~(inc-1);
	} else { 
	    firstPage = currRBitmapPage;
	    offs = currRBitmapOffs;
	}
	
	while (true) { 
	    if (alignment == 0) { 
		// allocate page object 
		for (i = firstPage; i < lastPage; i++){
		    int spaceNeeded = objBitSize - holeBitSize < pageBits 
			? objBitSize - holeBitSize : pageBits;
		    if (bitmapPageAvailableSpace[i] <= spaceNeeded) {
			holeBitSize = 0;
			offs = 0;
			continue;
		    }
		    pg = getPage(i);
		    int startOffs = offs;	
		    while (offs < Page.pageSize) { 
			if (pg.data[offs++] != 0) { 
			    offs = (offs + inc - 1) & ~(inc-1);
			    holeBitSize = 0;
			} else if ((holeBitSize += 8) == objBitSize) { 
			    pos = (((long)(i-dbBitmapId)*Page.pageSize + offs)*8 
				   - holeBitSize) << dbAllocationQuantumBits;
			    if (wasReserved(pos, size)) { 
				offs += objBitSize >> 3;
				startOffs = offs = (offs + inc - 1) & ~(inc-1);
				holeBitSize = 0;
				continue;
			    }	
			    reserveLocation(pos, size);
			    currPBitmapPage = i;
			    currPBitmapOffs = offs;
			    extend(pos + size);
			    if (oid != 0) { 
				long prev = getPos(oid);
				int marker = (int)prev & dbFlagsMask;
				pool.copy(pos, prev - marker, size);
				setPos(oid, pos | marker | dbModifiedFlag);
			    }
			    pool.unfix(pg);
			    pg = putPage(i);
			    int holeBytes = holeBitSize >> 3;
			    if (holeBytes > offs) { 
				memset(pg, 0, 0xFF, offs);
				holeBytes -= offs;
				pool.unfix(pg);
				pg = putPage(--i);
				offs = Page.pageSize;
			    }
			    while (holeBytes > Page.pageSize) { 
				memset(pg, 0, 0xFF, Page.pageSize);
				holeBytes -= Page.pageSize;
				bitmapPageAvailableSpace[i] = 0;
				pool.unfix(pg);
				pg = putPage(--i);
			    }
			    memset(pg, offs-holeBytes, 0xFF, holeBytes);
			    commitLocation();
			    pool.unfix(pg);
			    return pos;
			}
		    }
		    if (startOffs == 0 && holeBitSize == 0
			&& spaceNeeded < bitmapPageAvailableSpace[i]) 
		    { 
			bitmapPageAvailableSpace[i] = spaceNeeded;
		    }
		    offs = 0;
		    pool.unfix(pg);
		}
	    } else { 
		for (i = firstPage; i < lastPage; i++){
		    int spaceNeeded = objBitSize - holeBitSize < pageBits 
			? objBitSize - holeBitSize : pageBits;
		    if (bitmapPageAvailableSpace[i] <= spaceNeeded) {
			holeBitSize = 0;
			offs = 0;
			continue;
		    }
		    pg = getPage(i);
		    int startOffs = offs;
		    while (offs < Page.pageSize) { 
			int mask = pg.data[offs] & 0xFF; 
			if (holeBitSize + firstHoleSize[mask] >= objBitSize) { 
			    pos = (((long)(i-dbBitmapId)*Page.pageSize + offs)*8 
				   - holeBitSize) << dbAllocationQuantumBits;
			    if (wasReserved(pos, size)) { 			    
				startOffs = offs += (objBitSize + 7) >> 3;
				holeBitSize = 0;
				continue;
			    }	
			    reserveLocation(pos, size);
			    currRBitmapPage = i;
			    currRBitmapOffs = offs;
			    extend(pos + size);
			    if (oid != 0) { 
				long prev = getPos(oid);
				int marker = (int)prev & dbFlagsMask;
				pool.copy(pos, prev - marker, size);
				setPos(oid, pos | marker | dbModifiedFlag);
			    }
			    pool.unfix(pg);
			    pg = putPage(i);
			    pg.data[offs] |= (byte)((1 << (objBitSize - holeBitSize)) - 1); 
			    if (holeBitSize != 0) { 
				if (holeBitSize > offs*8) { 
				    memset(pg, 0, 0xFF, offs);
				    holeBitSize -= offs*8;
				    pool.unfix(pg);
				    pg = putPage(--i);
				    offs = Page.pageSize;
				}
				while (holeBitSize > pageBits) { 
				    memset(pg, 0, 0xFF, Page.pageSize);
				    holeBitSize -= pageBits;
				    bitmapPageAvailableSpace[i] = 0;
				    pool.unfix(pg);
				    pg = putPage(--i);
				}
				while ((holeBitSize -= 8) > 0) { 
				    pg.data[--offs] = (byte)0xFF; 
				}
				pg.data[offs-1] |= (byte)~((1 << -holeBitSize) - 1);
			    }
			    pool.unfix(pg);
			    commitLocation();
			    return pos;
			} else if (maxHoleSize[mask] >= objBitSize) { 
			    int holeBitOffset = maxHoleOffset[mask];
			    pos = (((long)(i-dbBitmapId)*Page.pageSize + offs)*8 + 
				   holeBitOffset) << dbAllocationQuantumBits;
			    if (wasReserved(pos, size)) { 
				startOffs = offs += (objBitSize + 7) >> 3;
				holeBitSize = 0;
				continue;
			    }	
			    reserveLocation(pos, size);
			    currRBitmapPage = i;
			    currRBitmapOffs = offs;
			    extend(pos + size);
			    if (oid != 0) { 
				long prev = getPos(oid);
				int marker = (int)prev & dbFlagsMask;
				pool.copy(pos, prev - marker, size);
				setPos(oid, pos | marker | dbModifiedFlag);
			    }
			    pool.unfix(pg);
			    pg = putPage(i);
			    pg.data[offs] |= (byte)((1<<objBitSize) - 1) << holeBitOffset;
			    pool.unfix(pg);
			    commitLocation();
			    return pos;
			}
			offs += 1;
			if (lastHoleSize[mask] == 8) { 
			    holeBitSize += 8;
			} else { 
			    holeBitSize = lastHoleSize[mask];
			}
		    }
		    if (startOffs == 0 && holeBitSize == 0
			&& spaceNeeded < bitmapPageAvailableSpace[i]) 
		    {
			bitmapPageAvailableSpace[i] = spaceNeeded;
		    }
		    offs = 0;
		    pool.unfix(pg);
		}
	    }
	    if (firstPage == dbBitmapId) { 
		if (freeBitmapPage > i) { 
		    i = freeBitmapPage;
		    holeBitSize = holeBeforeFreePage;
		}
		if (i == dbBitmapId + dbBitmapPages) { 
		    throw new ConsusRuntimeError("Database virtual space exhausted");
		}
		long extension = (size > dbDefaultExtensionQuantum) 
		    ? size : dbDefaultExtensionQuantum;
		int morePages = (int) 
		    ((extension + Page.pageSize*(dbAllocationQuantum*8-1) - 1)
		     / (Page.pageSize*(dbAllocationQuantum*8-1)));
		
		if (i + morePages > dbBitmapId + dbBitmapPages) { 
		    morePages = (int)  
			((size + Page.pageSize*(dbAllocationQuantum*8-1) - 1)
			 / (Page.pageSize*(dbAllocationQuantum*8-1)));
		    if (i + morePages > dbBitmapId + dbBitmapPages) { 
			throw new ConsusRuntimeError("Database virtual space exhausted");
		    }
		}
		objBitSize -= holeBitSize;
		int skip = (objBitSize + Page.pageSize/dbAllocationQuantum - 1) 
		    & ~(Page.pageSize/dbAllocationQuantum - 1);
		pos = ((long)(i-dbBitmapId) 
		       << (Page.pageBits+dbAllocationQuantumBits+3)) 
		    + (skip << dbAllocationQuantumBits);
		extend(pos + morePages*Page.pageSize);
		int len = objBitSize >> 3;
		long adr = pos;
		while (len >= Page.pageSize) { 
		    pg = pool.putPage(adr);
		    memset(pg, 0, 0xFF, Page.pageSize);
		    pool.unfix(pg);
		    adr += Page.pageSize;
		    len -= Page.pageSize;
		}
		pg = pool.putPage(adr);
		memset(pg, 0, 0xFF, len);
		pg.data[len] = (byte)((1 << (objBitSize&7))-1);
		pool.unfix(pg);
		adr = pos + (skip>>3);
		len = morePages * (Page.pageSize/dbAllocationQuantum/8);
		while (true) { 
		    int off = (int)adr & (Page.pageSize-1);
		    pg = pool.putPage(adr - off);
		    if (Page.pageSize - off >= len) { 
			memset(pg, off, 0xFF, len);
			pool.unfix(pg);
			break;
		    } else { 
			memset(pg, off, 0xFF, Page.pageSize - off);
			pool.unfix(pg);
			adr += Page.pageSize - off;
			len -= Page.pageSize - off;
		    }
		}
		int j = i;
		while (--morePages >= 0) { 
		    setPos(j++, pos | dbPageObjectFlag | dbModifiedFlag);
		    pos += Page.pageSize;
		}
		freeBitmapPage = header.root[1-currIndex].bitmapEnd = j;
		j = i + objBitSize / pageBits; 
		if (alignment != 0) { 
		    currRBitmapPage = j;
		    currRBitmapOffs = 0;
		} else { 
		    currPBitmapPage = j;
		    currPBitmapOffs = 0;
		}
		while (j > i) { 
		    bitmapPageAvailableSpace[--j] = 0;
		}
		
		pos = ((long)(i-dbBitmapId)*Page.pageSize*8 - holeBitSize)
		    << dbAllocationQuantumBits;
		if (oid != 0) { 
		    long prev = getPos(oid);
		    int marker = (int)prev & dbFlagsMask;
		    pool.copy(pos, prev - marker, size);
		    setPos(oid, pos | marker | dbModifiedFlag);
		}
		
		if (holeBitSize != 0) { 
		    reserveLocation(pos, size);
		    while (holeBitSize > pageBits) { 
			holeBitSize -= pageBits;
			pg = putPage(--i);
			memset(pg, 0, 0xFF, Page.pageSize);
			bitmapPageAvailableSpace[i] = 0;
			pool.unfix(pg);
		    }
		    pg = putPage(--i);
		    offs = Page.pageSize;
		    while ((holeBitSize -= 8) > 0) { 
			pg.data[--offs] = (byte)0xFF; 
		    }
		    pg.data[offs-1] |= (byte)~((1 << -holeBitSize) - 1);
		    pool.unfix(pg);
		    commitLocation();
		}
		return pos;
	    } 
	    freeBitmapPage = i;
	    holeBeforeFreePage = holeBitSize;
	    holeBitSize = 0;
	    lastPage = firstPage + 1;
	    firstPage = dbBitmapId;
	    offs = 0;
	}
    } 



    void free(long pos, long size) throws SQLException
    {
	Assert.that(pos != 0 && (pos & (dbAllocationQuantum-1)) == 0);
	long quantNo = pos >>> dbAllocationQuantumBits;
	int  objBitSize = (int)((size+dbAllocationQuantum-1) >>> dbAllocationQuantumBits);
	int  pageId = dbBitmapId + (int)(quantNo >>> (Page.pageBits+3));
	int  offs = (int)(quantNo & (Page.pageSize*8-1)) >> 3;
	Page pg = putPage(pageId);
	int  bitOffs = (int)quantNo & 7;

	usedSize -= objBitSize << dbAllocationQuantumBits;

	if (((int)pos & (Page.pageSize-1)) == 0 && size >= Page.pageSize) { 
	    if (pageId == currPBitmapPage && offs < currPBitmapOffs) { 
		currPBitmapOffs = offs;
	    }
	} else { 
	    if (pageId == currRBitmapPage && offs < currRBitmapOffs) { 
		currRBitmapOffs = offs;
	    }
	}
	if (pageId == currRBitmapPage && offs < currRBitmapOffs) { 
	    currRBitmapOffs = offs;
	}
	bitmapPageAvailableSpace[pageId] = Integer.MAX_VALUE;
	
	if (objBitSize > 8 - bitOffs) { 
	    objBitSize -= 8 - bitOffs;
	    pg.data[offs++] &= (1 << bitOffs) - 1;
	    while (objBitSize + offs*8 > Page.pageSize*8) { 
		memset(pg, offs, 0, Page.pageSize - offs);
		pool.unfix(pg);
		pg = putPage(++pageId);
		bitmapPageAvailableSpace[pageId] = Integer.MAX_VALUE;
		objBitSize -= (Page.pageSize - offs)*8;
		offs = 0;
	    }
	    while ((objBitSize -= 8) > 0) { 
		pg.data[offs++] = (byte)0;
	    }
	    pg.data[offs] &= (byte)~((1 << (objBitSize + 8)) - 1);
	} else { 
	    pg.data[offs] &= (byte)~(((1 << objBitSize) - 1) << bitOffs); 
	}
	pool.unfix(pg);
    }

    void cloneBitmap(long pos, long size) throws SQLException
    {
	long quantNo = pos >>> dbAllocationQuantumBits;
	int  objBitSize = (int)((size+dbAllocationQuantum-1) >>> dbAllocationQuantumBits);
	int  pageId = (int)(dbBitmapId + (quantNo >>> (Page.pageBits + 3)));
	int  offs = (int)(quantNo & (Page.pageSize*8-1)) >> 3;
	int  bitOffs = (int)quantNo & 7;
	int  oid = pageId;
	pos = getPos(oid);
	if ((pos & dbModifiedFlag) == 0) { 
	    dirtyPagesMap[oid >>> (dbHandlesPerPageBits+5)] 
		|= 1 << ((oid >>> dbHandlesPerPageBits) & 31);
	    allocate(Page.pageSize, oid);
	    cloneBitmap(pos & ~dbFlagsMask, Page.pageSize);
	}
	
	if (objBitSize > 8 - bitOffs) { 
	    objBitSize -= 8 - bitOffs;
	    offs += 1;
	    while (objBitSize + offs*8 > Page.pageSize*8) { 
		oid = ++pageId;
		pos = getPos(oid);
		if ((pos & dbModifiedFlag) == 0) { 
		    dirtyPagesMap[oid >>> (dbHandlesPerPageBits+5)] 
			|= 1 << ((oid >>> dbHandlesPerPageBits) & 31);
		    allocate(Page.pageSize, oid);
		    cloneBitmap(pos & ~dbFlagsMask, Page.pageSize);
		}
		objBitSize -= (Page.pageSize - offs)*8;
		offs = 0;
	    }
	}
    }


    int allocateId() throws SQLException
    {
	int oid;
	int curr = 1-currIndex;
	if ((oid = header.root[curr].freeList) != 0) { 
	    header.root[curr].freeList = (int)(getPos(oid) >> dbFlagsBits);
	    dirtyPagesMap[oid >>> (dbHandlesPerPageBits+5)] 
		|= 1 << ((oid >>> dbHandlesPerPageBits) & 31);
	    return oid;
	}
	if (currIndexSize + 1 > header.root[curr].indexSize) {
	    int oldIndexSize = header.root[curr].indexSize;
	    int newIndexSize = oldIndexSize * 2;
	    while (newIndexSize < oldIndexSize + 1) { 
		newIndexSize = newIndexSize*2;
	    }
	    long newIndex = allocate(newIndexSize * 8, 0);
	    pool.copy(newIndex, header.root[curr].index, currIndexSize*8);
	    free(header.root[curr].index, oldIndexSize*8);
	    header.root[curr].index = newIndex;
	    header.root[curr].indexSize = newIndexSize;
	}
	oid = currIndexSize;
	header.root[curr].indexUsed = ++currIndexSize;
	modified = true;
	return oid;
    }
    
    void freeId(int oid) throws SQLException
    {
	setPos(oid, ((long)(header.root[1-currIndex].freeList) << dbFlagsBits)
	       | dbFreeHandleFlag);
	header.root[1-currIndex].freeList = oid;
	modified = true;
    }
    
    synchronized TableDescriptor getTableByName(String name) throws SQLException { 
	TableDescriptor table = (TableDescriptor)tableHashByName.get(name);
	if (table == null) { 
	    Class cls;
	    try { 
		cls = Class.forName(name);
	    } catch(Throwable x) {
		return null;
	    }
	    table = new TableDescriptor(this, cls);
	    table.addIndices(false);	
	    tableHashByName.put(name, table);
	}
	return table;
    }

    TableDescriptor getTableByClass(Class cls) throws SQLException { 
	return getTableByName(cls.getName());
    }

    TableDescriptor getTableByOid(int oid) { 
	int h = oid % dbTableHashByOidSize;
	TableDescriptor table;
	for (table = tableHashByOid[h]; table != null; table = table.collisionChain) { 
	    if (table.tableId == oid) { 
		return table;
	    }
	}
	return null;
    }

    void registerTable(TableDescriptor table) {
	int h = table.tableId % dbTableHashByOidSize;
	table.collisionChain = tableHashByOid[h];
	tableHashByOid[h] = table;
	tableHashByName.put(table.name, table);
	if (lastTable == null) { 
	    firstTable = lastTable = table;
	} else { 
	    lastTable = lastTable.next = table;
	}
	table.next = null;
	schemaAltered = true;
    }
    
    // This method updates records belongs to the altered table. 
    // It doesn't change records from the derived tables assuming that it will 
    // be then called for the decriptor of the derived table. It reinsert updated
    // fields in the indices.
    void reformatTable(TableDescriptor oldTable, TableDescriptor newTable) 
        throws SQLException
    { 
	int oid = oldTable.getRootTable().firstRow;
	System.out.println("Reformatting table " + oldTable.name);
	while (oid != 0) { 
	    byte[] oldRec = get(oid);
	    if (Record.getType(oldRec, 0) == oldTable.tableId) { 
		byte[] newRec = newTable.convertRecord(oldRec);
		put(oid, newRec);
		for (FieldDescriptor fd = newTable.fields; fd != null; fd = fd.next)
		{
		    if (fd.index != 0 && fd.type != fd.oldType) { 
			if (!Btree.insert(this, fd, oid)) { 
			    throw new ConsusDuplicateException();
			}
		    }
		}
	    }
	    oid = Record.getNext(oldRec, 0);
	}
	schemaAltered = true;
    }


    int deleteTable(TableDescriptor table) throws SQLException { 
	TableDescriptor rootTable = table.getRootTable();
	int oid = rootTable.firstRow;
	int nDeleted = 0;
	beginTransaction(true);	    
		
	if (rootTable == table && table.nSubclasses == 0 && table.numberOfBlobs == 0) {
	    while (oid != 0) { 
		long pos = getPos(oid);
		int offs = (int)pos & (Page.pageSize-1);
		Page pg = pool.getPage(pos - offs);
		int size = Record.getSize(pg.data, offs & ~dbFlagsMask); 
		pool.unfix(pg);
		
		if ((pos & dbModifiedFlag) != 0) {
		    free(pos & ~dbFlagsMask, size);
		} else { 
		    cloneBitmap(pos, size);
		}
		nDeleted += 1;
		freeId(oid);
		oid = Record.getNext(pg.data, offs & ~dbFlagsMask);
	    }
	    for (FieldDescriptor fd = table.fields; fd != null; fd = fd.next) { 
		if (fd.index != 0) { 
		    Btree.purge(this, fd.index);
		}
	    }
	    rootTable.firstRow = rootTable.lastRow = 0;
	} else { 
	    while (oid != 0) { 
		long pos = getPos(oid) & ~dbFlagsMask;
		int offs = (int)pos & (Page.pageSize-1);
		Page pg = pool.getPage(pos - offs);
		if (Record.getType(pg.data, offs) == table.tableId) { 
		    freeRow(table, oid);
		    nDeleted += 1;
		}
		oid = Record.getNext(pg.data, offs);
		pool.unfix(pg);
	    }
	}
	rootTable.updated = true;
	schemaAltered = true;
	return nDeleted;
    }

    void dropTable(TableDescriptor table) throws SQLException { 
	beginTransaction(true);
	if (table.nSubclasses != 0) { 
	    throw new ConsusRuntimeError("Table with existed derived subtables can not be dropped");
	}
    if (table.superClass != null) { 
        table.superClass.nSubclasses -= 1;
    }
	deleteTable(table);
	if (table.tableId != 0) { 
	    freeRow(metatable, table.tableId);
	}
	unlinkTable(table);
	table.next = deletedTables;
	deletedTables = table;
	resetForeignKeyConstraints();
	schemaAltered = true;
    }

    void unlinkTable(TableDescriptor table) {
	int h = table.tableId % dbTableHashByOidSize;
	TableDescriptor prev = null, curr = tableHashByOid[h];
	while (curr != table) { 
	    prev = curr;
	    curr = curr.collisionChain;
	}
	if (prev == null) { 
	    tableHashByOid[h] = table.collisionChain;
	} else { 
	    prev.collisionChain = table.collisionChain;
	}
	tableHashByName.remove(table.name);
        curr = firstTable;
	prev = null;
	while (curr != table) { 
	    prev = curr;
	    curr = curr.next;
	}
	if (prev == null) { 
	    firstTable = table.next;
	} else { 
	    prev.next = table.next;
	}
	if (table == lastTable) { 
	    lastTable = prev;
	}
    }

    void dropIndex(FieldDescriptor field) throws SQLException { 
	beginTransaction(true);
	if (field.index != 0) { 
	    modified = true;
	    Btree.drop(this, field.index);
	    field.index = 0;
	    field.table.altered = true;
	    schemaAltered = true;
	}
    }

    void addIndex(FieldDescriptor field) throws SQLException { 
	Assert.that((field.attr & FieldDescriptor.ComponentOfArray) == 0);
	beginTransaction(true);
	if (field.index != 0) { 
	    return;
	}
	modified = true;
	field.index = Btree.allocate(this, field.type);
	int oid = field.table.getRootTable().firstRow; 
	while (oid != 0) { 
	    byte[] rec = get(oid);
	    if (field.table.isSuperclassFor(getTableByOid(Record.getType(rec, 0)))) { 
		if (!Btree.insert(this, field, oid)) { 
		    Btree.drop(this, field.index);
		    field.index = 0;
		    throw new ConsusDuplicateException();
		}
	    }
	    oid = Record.getNext(rec, 0);
	}
	field.table.altered = true;
	schemaAltered = true;
    }

    public final void deleteContext() { 
        synchronized(ctxHash) { 
            ctxHash.remove(Thread.currentThread());
        }
    }

    final ThreadContext getContext() { 
        synchronized(ctxHash) { 
            Thread current = Thread.currentThread();
            ThreadContext ctx = (ThreadContext)ctxHash.get(current);
            if (ctx == null) { 
                ctx = new ThreadContext();
                //if (++nNewThreads > 100) { 
                nNewThreads = 0;
                java.util.Enumeration threads = ctxHash.keys();
                while (threads.hasMoreElements()) { 
                    Thread thr = (Thread)threads.nextElement();
                    if (!thr.isAlive()) { 
                        ctxHash.remove(thr);
                    }
                }
                //}
                ctxHash.put(current, ctx);
            }
            return ctx;
        }
    }
    
    final void openCursor(ConsusResultSet cursor) { 
	ThreadContext ctx = getContext();
	cursor.next = ctx.openCursors;
	ctx.openCursors = cursor;
	
    }
	
    final void closeCursor(ConsusResultSet cursor) { 
	ThreadContext ctx = getContext();
	ConsusResultSet c = ctx.openCursors, prev = null; 
	while (c != null) { 
	    if (cursor == c) { 
		if (prev != null) { 
		    prev.next = cursor.next;
		} else { 
		    ctx.openCursors = cursor.next;
		}
		return;
	    }
	    prev = c;
	    c = c.next;
	}	
    }

    void beginTransaction(boolean modify) throws SQLException
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	ThreadContext ctx = getContext();

	if (commitDelay != 0) { 
	    synchronized (delayedCommitStopTimerEvent) { 
		forceCommitCount += 1;
		if (delayedCommitContext == ctx) {
		    // skip delayed transaction because this thread is starting new transaction
		    delayedCommitContext = null;
		    ctx.commitDelayed = false;
		    ctx.writeAccessByThisTransaction = modify;
		    if (commitTimerStarted != 0) { 
			long elapsed = System.currentTimeMillis() - commitTimerStarted;
			commitTimeout -= elapsed;
			if (commitTimeout < 0) { 
			    commitTimeout = 0;
			}
		    }
		    delayedCommitStopTimerEvent.notify();
		} else { 
		    if (delayedCommitContext != null) { 
		    // force transaction commit because another thread is going to begin transaction
			delayedCommitStopTimerEvent.notify();
		    }
		}
	    }
	}
		
	if (modify) { 
	    if (!ctx.writeAccess) { 
		synchronized (monitor) { 
		    if (ctx.readAccess) { 
			Assert.that(monitor.nWriters == 0 && monitor.nReaders >= 1);
			if (dbTraceEnabled) { 
			    System.out.println("Upgrade of lock can be a reason of deadlock");
			}				
			if (monitor.nReaders != 1) {
			    if (monitor.waitForUpgrade) { 
				throw new ConsusRuntimeError("Deadlock detected");
			    }
			    monitor.waitForUpgrade = true;
			    monitor.nWaitWriters += 1;
			    do { 
				try { 
				    monitor.wait();
				} catch(InterruptedException x) {}
			    } while (monitor.nWriters != 0 || monitor.nReaders != 1);
			    monitor.waitForUpgrade = false;
			    monitor.nWaitWriters -= 1;
			} 
			monitor.nWriters = 1;		    
			monitor.nReaders = 0;
		    } else { 
			if (monitor.nWriters != 0 || monitor.nReaders != 0) { 
			    monitor.nWaitWriters += 1;
			    do { 
				try { 
				    monitor.wait();
				} catch(InterruptedException x) {}
			    } while (monitor.nWriters + monitor.nReaders != 0);
			    monitor.nWaitWriters -= 1;
			} 
			monitor.nWriters = 1;
		    }
		    if (!header.dirty) {
 			if (header.gtrid != null) { 
			    //
			    // There is prepared but not committed transaction.
			    // And looks like transaction manager is not going to recover it.
			    // Then just rollback this transaction.
			    //
			    rollbackTransaction();
			}
			header.dirty = true;
			Page pg = pool.putPage(0);
			header.pack(pg.data);
			pool.unfix(pg);
			pool.flush();
		    }
		    ctx.writeAccess = true;
		    ctx.writeAccessByThisTransaction = true;
		}
	    }
	} else { 
	    if (!ctx.readAccess && !ctx.writeAccess) { 
		synchronized(monitor) { 
		    if (monitor.nWriters + monitor.nWaitWriters != 0) {
			monitor.nWaitReaders += 1;
			do { 
			    try { 
				monitor.wait();
			    } catch(InterruptedException x) {}
			} while(monitor.nWriters != 0);
			monitor.nWaitReaders -= 1;
		    } 
		    monitor.nReaders += 1;
		}
		ctx.readAccess = true;
	    }
	}
	if (commitDelay != 0) { 
	    synchronized (delayedCommitStopTimerEvent) { 
		forceCommitCount -= 1;
	    }
	}
	int curr = currIndex;
	currIndexSize = header.root[1-curr].indexUsed;
	committedIndexSize = header.root[curr].indexUsed;
	Assert.that(currIndexSize >= committedIndexSize);
    }

    void endTransaction(ThreadContext ctx) throws SQLException
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	while (ctx.openCursors != null) { 
	    ctx.openCursors.reset(false);
	}	
	if (ctx.writeAccess) { 
	    synchronized(monitor) {
		Assert.that(monitor.nWriters == 1);
		ctx.transactionPrepared = false;	
		ctx.readAccess = false;
		ctx.writeAccess = false;
		ctx.writeAccessByThisTransaction = false;
		monitor.nWriters = 0;
		Assert.that(!monitor.waitForUpgrade);
		if (monitor.nWaitWriters + monitor.nWaitReaders != 0) { 
		    monitor.notifyAll();
		}
	    } 
	} else if (ctx.readAccess) { 
	    synchronized(monitor) { 
		Assert.that(monitor.nReaders > 0);
		ctx.readAccess = false;
		ctx.transactionPrepared = false;	
		monitor.nReaders -= 1;
		if ((monitor.nReaders == 1 && monitor.waitForUpgrade) ||
		    (monitor.nReaders == 0 && monitor.nWaitWriters != 0))
		{ 
		    monitor.notifyAll();
		} 
	    }
	}
    }

    /** 
     * Lock database in exclusive mode. Lock will be released by transaction commit/rollback
     */
    public void lock() throws Exception { 
	beginTransaction(true);
    }

    /**
     * Commit nested subtransaction. Exterior transaction will be committed when all nested
     * subtransaction were successfully committed.
     */ 
    public void commitSubtransaction() throws SQLException 
    {
	if (!opened) { 
	    throw new ConsusRuntimeError("Connection is closed");
	}
	ThreadContext ctx = getContext();
	if (ctx.writeAccess || ctx.readAccess) { 
	    if (ctx.writeAccess) { 
		if (usedSize - usedBeforeLastGC > GCThreshold) { 
		    doGarbageCollection();
		}
	    }
	    ctx.concurrentId = concurrentTransId; 
	    endTransaction(ctx);
	}
    }

    static final String DATE_FORMAT = "yyyy.MM.dd_HH.mm.ss";
    static final DateFormat formatter = new SimpleDateFormat(DATE_FORMAT, Locale.ENGLISH);

    class BackupThread extends Thread { 
	String  fileName;
	long    periodMsec;

	public void run() { 
	    setPriority(Thread.MIN_PRIORITY);

	    synchronized (backupCS) { 
		try { 		
		    while (true) { 
			long timeout = periodMsec;

			if (!fileName.endsWith("?")) {
			    File f = new File(fileName);
			    if (f.exists()) { 
				long howOld = System.currentTimeMillis() - f.lastModified();
				if ((timeout -= howOld) < 0) { 
				    timeout = 0;
				}
			    }
			}

			backupCS.wait(timeout);

			if (fileName != null) { 
			    if (fileName.endsWith("?")) { 
				java.util.Date currDate = new java.util.Date();
				File f = new File(fileName.substring(0, fileName.length()-1) 
						     + formatter.format(currDate));
				FileOutputStream out = new FileOutputStream(f);
				backup(out);
				out.close();
			    } else { 
				File fnew = new File(fileName + ".new");
				FileOutputStream out = new FileOutputStream(fnew);
				backup(out);
				out.close();
				File f = new File(fileName);
				f.delete();
				fnew.renameTo(f);
				timeout = periodMsec;
			    }
			} else { 
			    return;
			}
		    }
		} catch (InterruptedException x) { 
		    System.err.println("Backup thread was interrupted");
		} catch (SQLException x) { 
		    System.err.println("SQLException during backup: " + x);
		} catch (IOException x) { 
		    System.err.println("IOException during backup: " + x);
		}
		backupThread = null;
	    }
	}

	BackupThread(String name, long msec) {
	    fileName = name;
	    periodMsec = msec;
	}
    }

    public void scheduleBackup(String fileName, long periodMsec) { 
	synchronized (backupCS) { 
	    if (backupThread == null) { 
		backupThread = new BackupThread(fileName, periodMsec);
		backupThread.start();
	    }
	}
    }



    /**
     * Online backup of the database.
     * @param stream to which backup is performed
     */
    public void backup(java.io.OutputStream stream) 
      throws SQLException, java.io.IOException 
    {
	synchronized(backupEvent) { 
	    if (backupInProgress) { 
		throw new ConsusRuntimeError("Only one backup can be active at each moment of time");
	    }
	    backupInProgress = true;
	}
	long eof = header.root[currIndex].size;
	for (long pos = 0; pos < eof; pos += Page.pageSize) { 
	    Page pg = pool.getPage(pos);
	    stream.write(pg.data, 0, Page.pageSize);
	    pool.unfix(pg);
	}
	synchronized(backupEvent) { 
	     backupInProgress = false;
	     backupEvent.notifyAll();
	}
    }

    /**
     * Throw all objects from the object cache. This method should be called
     * to prevent memory exhaustion caused by loading a lot of persistent interconnected 
     * objects (each of the objects is accessible from some set of root objects)
     * from the database. GC will not be able to deallocate such objects because they are
     * referencing each other. So finally all object from the database can be loaded to
     * the memory. Programmer should check number of loaded objects (by 
     * <code>getObjectCacheSize</code> or <code>getUsedMemorySize</code> method) and
     * if it axceeds some threshold, 
     * call <code>clearObjectCache</code> method to remove all objects from cache. 
     * Alternativly, progrtammer can call <code>Persistent.unget</code> method to replace
     * object with a stub, but this approach is less efficient because usually it throws 
     * away most recently used object. 
     */
    public void clearObjectCache() { 
	if (objectCache != null) { 
	    objectCache.clear();
	}
    }

    /**
     * Poll object cache size
     * @return number of object in object cache.
     */
    public int getObjectCacheSize() { 
	return objectCache != null ? objectCache.size() : 0;
    }

    /**
     * Get size of memory usd by application. 
     * @return size of  memory in bytes used by application.
     */
    public static long getUsedMemorySize() { 
	if (runtime == null) { 
	    runtime = Runtime.getRuntime();
	}	    
	return runtime.totalMemory() - runtime.freeMemory();
    }
    
    /**
     * Allocate object cache (instance manager) if not allocated yet.
     */
    public void createObjectCache() { 
	if (objectCache == null) { 
	    try { 
		// Solution for MS jview with com.ms.vm.WeakReference class
		Class.forName("com.ms.vm.WeakReference");
		Object args[] = {new Integer(dbObjectCacheInitSize)};		    
		objectCache = (WeakHashTable)Class.forName("week.MsWeakHashTable")
		    .getConstructors()[0].newInstance(args);
	    } catch(Exception x1) { 
		try { 
		    Class.forName("sun.misc.Cache");
		    Object args[] = {new Integer(dbObjectCacheInitSize)};
		    objectCache = (WeakHashTable)Class.forName("week.sunWeakHashTable")
			.getConstructors()[0].newInstance(args);
		} catch (Exception x2) { 
		    try { 
			Class.forName("java.lang.ref.WeakReference");		
			objectCache = new WeakHashTable(dbObjectCacheInitSize);
		    } catch(Exception x3) { 
			System.out.println("Failed to create object cache");
		    }
		}
	    }
	}
    }

    /**
     * Get size used by objects in the database
     * @return total size of all objects in the database (bytes)
     */
    public long getUsedSize() throws SQLException { 
	return usedSize;
    }

    /**
     * Specify root tables for the garbage collection. All tables should already exist.
     * All records (objects) for these tables will be marked as accessible. The all
     * other objects in the database recursuvly referenced from these objects will be marked.
     * And at the final stage all unmarked objects will be removed.
     * @param tableNames array with table names
     */
    public void setGCRoots(String tableNames[]) throws SQLException
    {
	if (tableNames.length == 0) { 
	    throw new ConsusRuntimeError("At least one root table should be specified");
	}
	int n = tableNames.length;
	rootTables = new TableDescriptor[n];
	for (int i = 0; i < n; i++) { 
	    TableDescriptor table = getTableByName(tableNames[i]);
	    if (table == null) { 
		throw new ConsusRuntimeError("Table " + tableNames[i] + " not found");
	    }
	    rootTables[i] = table.getRootTable();
	}
    } 
    
    /** 
     * Specify size of allocated objects after which GC should be initiated. 
     * @param threshold difference between allocated and deallocated object since last GC
     * after which new GC is initiated. By default it is set to <code>Long.MAX_VALUE</code>
     * so GC is disable.
     */
    public void setGCThreshold(long threshold) throws SQLException
    { 
	GCThreshold = threshold;
    }

    /** 
     * Explicitly initiate new garbage collection (if not yet started)
     */
    public void doGarbageCollection() throws SQLException { 
	synchronized (this) { 
	    if (GCstarted) { 
		return;
	    }
	    GCstarted = true;
	}
	long start = 0;
	if (dbTraceEnabled) { 
	    start = System.currentTimeMillis();
	    System.out.println("Garbage collection started, used size: " + usedSize);
	}
	if (rootTables == null) { 
	    throw new ConsusRuntimeError("No root tables were specified");
	}
	int nObjects = currIndexSize;
	int bitmapSize = (currIndexSize + 31) >>> 5;
	int greyBitmap[] = new int[bitmapSize];
	int blackBitmap[] = new int[bitmapSize];
	int i;
	for (i = 0; i < rootTables.length; i++) { 
	    int oid = rootTables[i].firstRow;
	    while (oid != 0) { 
		greyBitmap[oid>>>5] |= 1 << (oid & 31);
		oid = getNext(oid);
	    }
	}
	if (objectCache != null) { 
	    objectCache.mark(greyBitmap);
	}
	/* GC mark */
	boolean toBeContinued;
	do { 
	    toBeContinued = false;
	    for (i = dbFirstUserId>>>5; i < bitmapSize; i++) { 
		int mask = greyBitmap[i];
		while (mask != 0) {
		    int oid = i << 5;
		    blackBitmap[i] |= mask;
		    greyBitmap[i] = 0;
		    do { 
			if ((mask & 1) != 0 && oid < nObjects) { 
			    long pos = getPos(oid);
			    if (pos != 0 
				&& (pos & (dbFreeHandleFlag|dbPageObjectFlag)) == 0)
			    {
				byte[] obj = get(oid);
				int tableId = Record.getType(obj, 0);
				if (tableId >= dbFirstUserId) { 
				    TableDescriptor table = getTableByOid(tableId);
				    Assert.that(table != null);
				    if (table.markObject(obj, greyBitmap, blackBitmap)) { 
					toBeContinued = true;
				    }
				}
			    }
			}
			oid += 1;
		    } while ((mask >>>= 1) != 0);
		    mask = greyBitmap[i];
		}
	    }
	} while (toBeContinued);

	TableDescriptor indexTable = 
	    (TableDescriptor)tableHashByName.get("com.togethesoft.internal.consus.StringIndex");
	/* GC sweep */
	for (i = dbFirstUserId>>>5; i < bitmapSize; i++) { 
	    int mask = ~blackBitmap[i];
	    if (mask != 0) { 
		int oid = i << 5;
		do { 
		    if ((mask & 1) != 0 && oid < nObjects) { 
			long pos = getPos(oid);
			if (pos != 0 
			    && (pos & (dbFreeHandleFlag|dbPageObjectFlag)) == 0)
			{
			    int tableId = getType(oid);
			    if (tableId >= dbFirstUserId) {
				TableDescriptor table = getTableByOid(tableId);
				Assert.that(table != null);
				if (table == indexTable) {
				    // deallocate btree pages used by StringIndex object
				    byte[] rec = get(oid);
				    int treeId = Bytes.unpack4(rec, Record.sizeof);
				    if (treeId != 0) { 
					Btree.drop(this, treeId);
				    }
				} 
				freeRow(table, oid);		
			    }
			}
		    }
		    oid += 1;
		} while ((mask >>>= 1) != 0);
	    }
	}     
	usedBeforeLastGC = usedSize;
	if (dbTraceEnabled) { 
	    System.out.println("Garbage collection finished in " + 
			       (System.currentTimeMillis() - start) + " milliseconds, " + 
			       "used size: " + usedSize);
	}
	GCstarted = false;
    }
    
    /** 
     * Update replica table by transfering all changes from the master table.
     * It is not necessary for master and replica table to be exactly the same. Replica table
     * can contain subset of master table columns. But they both should have 
     * <code>recordTimestamp</code> and optionally <code>masterOID</code> fields. 
     * The the application can close connection with the server and normally work with 
     * replica table autonomously.
     * 
     * @param tableName name of the table to check-out
     * @param conxn database connection
     */
    public void checkOutTable(String tableName, Connection conxn) 
	throws SQLException
    {
	int i, n;
	TableDescriptor table = getTableByName(tableName);
	if (table == null) { 
	    throw new ConsusRuntimeError("Table " + tableName + " not found");
	}
	table.setVersionAutoincrementMode(recordTimestamp);
	FieldDescriptor[] columns = table.columns;
	int nColumns = table.columns.length;
	FieldDescriptor primaryKey = null;
	String keyName = masterOID;
	long timestamp = table.timestamp;
	if (masterOID == null || table.getColumn(masterOID) == null) { 
	    primaryKey = table.getPrimaryKey();
	    if (primaryKey == null) { 
		throw new ConsusRuntimeError("Table " + tableName + " has no primary key");
	    }
	    keyName = primaryKey.longName;
	}
	Statement masterSelect = conxn.createStatement();
	ResultSet masterCursor = 
	    masterSelect.executeQuery("select * from " + tableName
				      + " where " + recordTimestamp + ">" + timestamp);
	
	PreparedStatement select = prepareStatement("select * from " + tableName + " where " + 
						    keyName + "=? for update");

	// switch off automatic incrementing of version number
	FieldDescriptor timestampField = table.timestampField;
	table.timestampField = null;
	
	while (masterCursor.next()) { 
	    select.setObject(1, masterCursor.getObject(keyName));
	    ResultSet cursor = select.executeQuery();
	    long t = masterCursor.getLong(recordTimestamp);
	    if (t > timestamp) { 
		timestamp = t;
	    }
	    if (cursor.next()) { 
		for (i = 0; i < nColumns; i++) { 
		    String name = columns[i].name;
		    if (!name.equals(keyName)) { 
			cursor.updateObject(i+1, masterCursor.getObject(name));
		    }
		}
		cursor.updateRow();
	    } else { 		    
		cursor.moveToInsertRow();
		for (i = 0; i < nColumns; i++) { 
		    cursor.updateObject(i+1, masterCursor.getObject(columns[i].name));
		}
		cursor.insertRow();
	    }
	}
	if (table.timestamp != timestamp) { 
	    table.timestamp = timestamp;
	    table.updated = true;
	}
	table.timestampField = timestampField;	
	masterSelect.close();
	select.close();
    }
	
    /** 
     * Synchronize replica and master table. 
     * This method will first transfer 
     * to the client all records, which were changed at the server from the moment of the 
     * check out. If some of these records was also changed by the client, then merge 
     * conflict is detected and <code>MergeConflictException</code> is thrown. 
     * Exception object is given value of the key of the record caused conflict. 
     * Then Consus will transfer all records which were changed by the client to the master 
     * table, assigning them the value of timestamp greater than of any other record in the 
     * master table. It is possible to perform check-ins several times synchronizing content 
     * of the master and replica tables.
     *
     * @param tableName name of the table to check-in
     * @param conxn database connection
     * @exception MergeConflictException
     */
    public void checkInTable(String tableName, Connection conxn) 
	throws SQLException
    {
	int i, j;
	FieldDescriptor fd;
	TableDescriptor table = getTableByName(tableName);
	if (table == null) { 
	    throw new ConsusRuntimeError("Table " + tableName + " not found");
	}
	FieldDescriptor[] columns = table.columns;
	int nColumns = table.columns.length;
	String keyName = masterOID;
	FieldDescriptor primaryKey = null;
	Statement stmt = createStatement();
	long timestamp = table.timestamp;
	ResultSet localChanges = stmt.executeQuery("select * from " + tableName + 
						   " where " + recordTimestamp + ">" + 
						   timestamp + " for update");
	if (masterOID == null || table.getColumn(masterOID) == null) { 
	    primaryKey = table.getPrimaryKey();
	    if (primaryKey == null) { 
		throw new ConsusRuntimeError("Table " + tableName + " has no primary key");
	    }
	    keyName = primaryKey.longName;
	}
	PreparedStatement select = prepareStatement("select * from " + tableName + " where " + 
						    keyName + "=? for update");

	PreparedStatement masterSelect = 
	    conxn.prepareStatement("select * from " + tableName + " where " 
				   + recordTimestamp + ">" + timestamp);
	ResultSet masterChanges = masterSelect.executeQuery();
	// switch off automatic incrementing of version number
	FieldDescriptor timestampField = table.timestampField;
	table.timestampField = null;

	while (masterChanges.next()) { 
	    long t = masterChanges.getLong(recordTimestamp);
	    if (t > timestamp) { 
		timestamp = t;
	    }
	    select.setObject(1, masterChanges.getObject(keyName));
	    ResultSet cursor = select.executeQuery();
	    if (cursor.next()) { 
		if (cursor.getLong(recordTimestamp) > table.timestamp) { 
		    throw new MergeConflictException(cursor.getObject(keyName));
		}
		for (i = 0; i < nColumns; i++) { 
		    String name = columns[i].name;
		    if (!name.equals(keyName)) { 
			cursor.updateObject(i+1, masterChanges.getObject(name));
		    }
		}
		cursor.updateRow();
	    } else { 		    
		cursor.moveToInsertRow();
		for (i = 0; i < nColumns; i++) { 
		    cursor.updateObject(i+1, masterChanges.getObject(columns[i].name));
		}
		cursor.insertRow();
	    }
	}
	select.close();
	masterSelect.close();

	char isep = ' ', usep = ' ';
	String insertSQL = "insert into " + tableName + "(";
	String updateSQL = "update " + tableName + " set";
	String params = "";
	
	if (((ConsusResultSet)localChanges).getSelectionSize() > 0) { 
	    timestamp += 1;
	}

	for (i = 0; i < nColumns; i++) { 
	    if (!columns[i].name.equals(masterOID)) { 
		params += isep;
		params += '?';
		insertSQL += isep;
		insertSQL += columns[i].name;
		isep = ',';							      
		if (columns[i] != primaryKey) { 
		    updateSQL += usep;
		    updateSQL += columns[i].name + "=?";
		    usep = ',';
		}
	    }
	}
	if (primaryKey == null) { 
	    insertSQL += isep;
	    insertSQL += masterOID;
	    params += isep;
	    params += "oid";
	    updateSQL += " where oid" + "=?";
	    masterSelect = conxn.prepareStatement("select * from " + tableName + 
						  " where oid=?");
	} else { 
	    updateSQL += " where " + keyName + "=?";
	    masterSelect = conxn.prepareStatement("select * from " + tableName + 
						  " where " + keyName + "=?");
	}
	insertSQL += ") values (" + params + ")";
	PreparedStatement insert = conxn.prepareStatement(insertSQL);
	PreparedStatement update = conxn.prepareStatement(updateSQL);

	while (localChanges.next()) { 
	    masterSelect.setObject(1, localChanges.getObject(keyName));
	    ResultSet masterCursor = masterSelect.executeQuery();
	    if (masterCursor.next()) { 
		for (i = 0, j = 1; i < nColumns; i++) {
		    if (columns[i] == timestampField) { 
			update.setLong(j++, timestamp);
		    } else if (!columns[i].name.equals(keyName)) { 
			update.setObject(j++, localChanges.getObject(i+1));
		    }
		}
		update.setObject(j, localChanges.getObject(keyName));
		update.executeUpdate();
	    } else { 
		// new record
		for (i = 0, j = 1; i < nColumns; i++) {
		    if (columns[i] == timestampField) { 
			insert.setLong(j++, timestamp);
		    } else if (!columns[i].name.equals(masterOID)) { 
			insert.setObject(j++, localChanges.getObject(i+1));
		    }
		}
		insert.executeUpdate();
		if (primaryKey == null) { 
		    localChanges.updateObject(masterOID, 
					      ((ConsusStatement)insert).getLastRef());
		}
	    }
	    localChanges.updateLong(recordTimestamp, timestamp);
	    localChanges.updateRow();
	}
	table.timestampField = timestampField;
	if (table.timestamp != timestamp) { 
	    table.timestamp = timestamp;
	    table.updated = true;
	}
	masterSelect.close();
	update.close();
	insert.close();
	stmt.close();
    }

    public ConsusConnection(Driver JDBCdriver, String url, Properties info) { 
	driver = JDBCdriver;
	URL = url;
	properties = info;
    }

    public ConsusConnection(String fileName, String url, Properties info, 
			    Driver driver) throws SQLException
    { 
	this(driver, url, info);
	Page pg;
	int i;
	int indexSize = dbDefaultInitIndexSize;
	if (indexSize < dbFirstUserId) { 
	    indexSize = dbFirstUserId;
	}
	this.fileName = fileName;
	indexSize = (indexSize + dbHandlesPerPage - 1) & ~(dbHandlesPerPage-1);

	dirtyPagesMap = new int[dbDirtyPageBitmapSize/4+1];
	bitmapPageAvailableSpace = new int[dbBitmapId + dbBitmapPages];
 	for (i = dbBitmapId + dbBitmapPages; --i >= 0;) { 
	    bitmapPageAvailableSpace[i] = Integer.MAX_VALUE;
	}
	
	currRBitmapPage = currPBitmapPage = dbBitmapId;
	currRBitmapOffs = currPBitmapOffs = 0;
	firstTable = lastTable = null;
	modified = false;
	GCThreshold = Long.MAX_VALUE;
	concurrentTransId = 1;
	
	monitor = new Monitor();
	monitor.nReaders = 0;
	monitor.nWriters = 0;
	monitor.nWaitReaders = 0;
	monitor.nWaitWriters = 0;
	monitor.waitForUpgrade = false;

	backupEvent = new Object();
	backupCS = new Object();
	
	delayedCommitStartTimerEvent = new Object();
	delayedCommitStopTimerEvent = new Object();
	commitThreadSyncEvent = new Object();

	ConsusFile file = new ConsusFile(fileName);
	String fileLock = info.getProperty("consus.file.lock");
        if (!"no".equalsIgnoreCase(fileLock) && !"false".equalsIgnoreCase(fileLock)) {             
            if (!file.lock()) { 
                throw new SQLException("File is used by some other application");
            }
	}
	header = new Header();
	byte[] buf = new byte[Header.sizeof];
	int rc = file.read(0, buf);
	if (rc > 0 && rc < Header.sizeof) { 
	    throw new ConsusIOError("Failed to read file header");
	}
	header.unpack(buf);
	if (header.curr < 0 || header.curr > 1) { 
	    throw new ConsusIOError("Database file corrupted");
	}
	if (!header.initialized) {	    
	    header.curr = currIndex = 0;
	    long used = Page.pageSize;
	    header.root[0].index = used;
	    header.root[0].indexSize = indexSize;
	    header.root[0].indexUsed = dbFirstUserId;
	    header.root[0].freeList = 0;
	    used += indexSize*8;
	    header.root[1].index = used;
	    header.root[1].indexSize = indexSize;
	    header.root[1].indexUsed = dbFirstUserId;
	    header.root[1].freeList = 0;
	    used += indexSize*8;
	
	    header.root[0].shadowIndex = header.root[1].index;
	    header.root[1].shadowIndex = header.root[0].index;
	    header.root[0].shadowIndexSize = indexSize;
	    header.root[1].shadowIndexSize = indexSize;
	    
	    int bitmapPages = 
		(int)((used + Page.pageSize*(dbAllocationQuantum*8-1) - 1)
		      / (Page.pageSize*(dbAllocationQuantum*8-1)));
	    int bitmapSize = bitmapPages*Page.pageSize;
	    int usedBitmapSize = (int)((used + bitmapSize) >>> (dbAllocationQuantumBits + 3));

	    byte[] bitmap = new byte[bitmapSize];
	    for (i = 0; i < usedBitmapSize; i++) { 
		bitmap[i] = (byte)0xFF;
	    }
	    file.write(used, bitmap);
	    int bitmapIndexSize = 
		((dbBitmapId + dbBitmapPages)*8 + Page.pageSize - 1)
		& ~(Page.pageSize - 1);
	    byte[] index = new byte[bitmapIndexSize];
	    Bytes.pack8(index, dbInvalidId*8, dbFreeHandleFlag);
	    for (i = 0; i < bitmapPages; i++) { 
		Bytes.pack8(index, (dbBitmapId+i)*8, used | dbPageObjectFlag | dbModifiedFlag);
		used += Page.pageSize;
	    }
	    header.root[0].bitmapEnd = dbBitmapId + i;
	    header.root[1].bitmapEnd = dbBitmapId + i;
	    while (i < dbBitmapPages) { 
		Bytes.pack8(index, (dbBitmapId+i)*8, dbFreeHandleFlag);
		i += 1;
	    }
	    file.write(header.root[1].index, index);
	    header.root[0].size = used;
	    header.root[1].size = used;
	    committedIndexSize = 0;
	    currIndexSize = dbFirstUserId;
	    pool.open(file, used);

	    initializeMetaTable();

	    long indexPage = header.root[1].index;
	    long lastIndexPage = indexPage + header.root[1].bitmapEnd*8;
	    while (indexPage < lastIndexPage) { 
		pg = pool.putPage(indexPage);
		for (i = 0; i < Page.pageSize; i += 8) { 
		    Bytes.pack8(pg.data, i, 
				Bytes.unpack8(pg.data, i) & ~dbModifiedFlag);
		}
		pool.unfix(pg);
		indexPage += Page.pageSize;
	    }
	    pool.copy(header.root[0].index, header.root[1].index, currIndexSize*8);
	    header.dirty = true;
	    header.root[0].size = header.root[1].size;
	    pg = pool.putPage(0);
	    header.pack(pg.data);
	    pool.flush();
	    pool.modify(pg);
	    header.initialized = true;
	    header.pack(pg.data);
	    pool.unfix(pg);
	    pool.flush();
	    opened = true;
	} else {
	    int curr = header.curr;
	    currIndex = curr;
	    if (header.root[curr].indexSize != header.root[curr].shadowIndexSize) {
		throw new ConsusIOError("Database header is corrupted");
	    }		
	    pool.open(file, header.root[curr].size);
	    if (header.dirty) { 
		System.err.println("Database was not normally closed: start recovery");
		header.root[1-curr].size = header.root[curr].size;
		header.root[1-curr].indexUsed = header.root[curr].indexUsed; 
		header.root[1-curr].freeList = header.root[curr].freeList; 
		header.root[1-curr].index = header.root[curr].shadowIndex;
		currIndexSize = header.root[1-curr].indexSize = 
		    header.root[curr].shadowIndexSize;
		header.root[1-curr].shadowIndex = header.root[curr].index;
		header.root[1-curr].shadowIndexSize = 
		    header.root[curr].indexSize;
		header.root[1-curr].bitmapEnd = header.root[curr].bitmapEnd;
		
		pg = pool.putPage(0);
		header.pack(pg.data);
		pool.unfix(pg);

		pool.copy(header.root[1-curr].index, header.root[curr].index, 
			  (header.root[curr].indexUsed*8 + Page.pageSize - 1) & ~(Page.pageSize-1));
		restoreTablesConsistency();
		System.err.println("Recovery completed");
	    } 
	    usedSize = usedBeforeLastGC = header.root[curr].size;
	    opened = true;
	    loadScheme();
	}

	String backupFileName = info.getProperty("consus.backup.name");
	if (backupFileName != null) { 
	    String backupPeriodStr = info.getProperty("consus.backup.period");
	    long backupPeriod;
	    try { 			
		if (backupPeriodStr == null) { 
		    backupPeriod = (long)24*60*60*1000; // one day
		} else { 
		    backupPeriod = Long.parseLong(backupPeriodStr);
		}
		scheduleBackup(backupFileName, backupPeriod);
	    } catch (NumberFormatException x) {
		System.err.println("Invalid backup period: " + backupPeriodStr);
	    }
	}

	String commitDelayStr = info.getProperty("consus.commit.delay", "0");
	try { 
	    commitDelay = Long.parseLong(commitDelayStr);
	    if (commitDelay != 0) { 
		commitTimeout = commitDelay;
		commitThread = new DelayedCommitThread();
		commitThread.start();
		synchronized(commitThreadSyncEvent) { 
		    try { 
			commitThreadSyncEvent.wait();
		    } catch(InterruptedException x) {}
		}
	    }
	} catch (NumberFormatException x) {
	    System.err.println("Invalid commit delay: " + commitDelayStr);
	}
    }

    void restoreTablesConsistency() throws SQLException
    {
	Page pg;
	//
	// Restore consistency of table rows l2-list 
	//
	byte[] tab = get(dbMetaTableId);
	int lastId = TableDescriptor.getLastRow(tab);
	int tableId = TableDescriptor.getFirstRow(tab);
	if (lastId != 0) { 
	    long pos = getPos(lastId);
	    int offs = (int)pos & (Page.pageSize-1);
	    pg = pool.getPage(pos - offs);
	    if (Record.getNext(pg.data, offs & ~dbFlagsMask) != 0)  { 
		pool.modify(pg);
		Record.setNext(pg.data, offs & ~dbFlagsMask, 0);
	    }
	    pool.unfix(pg);
	}
	while (tableId != 0) { 
	    tab = get(tableId);
	    lastId = TableDescriptor.getLastRow(tab);
	    tableId = Record.getNext(tab, 0);
	    if (lastId != 0) { 
		long pos = getPos(lastId);
		int offs = (int)pos & (Page.pageSize-1);
		pg = pool.getPage(pos - offs);
		if (Record.getNext(pg.data, offs & ~dbFlagsMask) != 0) { 
		    pool.modify(pg);
		    Record.setNext(pg.data, offs & ~dbFlagsMask, 0);
		}
		pool.unfix(pg);
	    }
	}
    }

    void initializeMetaTable() throws SQLException
    {
	try { 
	    metatable = new TableDescriptor(this, Class.forName("org.garret.consus.MetaTable"));
	    metatable.tableId = dbMetaTableId;
	} catch(ClassNotFoundException x) {
	    Assert.failed("No MetaTable class definition found");
	}
	metatable.name = "MetaTable";
	byte[] metadesc = metatable.exportDescriptor();
	long metaTableOffs = allocate(metadesc.length, 0);
	setPos(dbMetaTableId, metaTableOffs);
	pool.put(metaTableOffs, metadesc);
	registerTable(metatable);
    }
    
    void loadScheme() throws SQLException
    {
	TableDescriptor table;
	beginTransaction(true);
	try { 
	    metatable = new TableDescriptor(this, Class.forName("org.garret.consus.MetaTable"));
	} catch(ClassNotFoundException x) {
	    Assert.failed("No MetaTable class definition found");
	}
	metatable.name = "MetaTable";
	metatable.tableId = dbMetaTableId;
	table = new TableDescriptor(this, dbMetaTableId); 
	if (!metatable.equals(table)) { 
	    throw new ConsusRuntimeError("Incompatible format of MetaTable: " + 
					 "database was created by newer Consus version");
	}
	metatable.tableId = table.tableId;
	metatable.firstRow = table.firstRow;
	metatable.lastRow = table.lastRow;
	metatable.nRows = table.nRows;
	metatable.desc = table.desc;
	registerTable(metatable);
	int tableId = metatable.firstRow;
	while (tableId != 0) { 
	    table = new TableDescriptor(this, tableId);
	    if ((table.attr & TableDescriptor.TA_JAVA_CLASS) != 0) { 
		try { 
		    Class clazz = Class.forName(table.name);
		    TableDescriptor newTable = new TableDescriptor(this, clazz);
		    if (!newTable.equals(table)) { 
			System.out.println("Table descriptor " + table.name + " is changed");
			table.newTable = newTable;
			newTable.match(table);
			registerTable(table);
		    } else { 
			newTable.tableId = table.tableId;
			newTable.firstRow = table.firstRow;
			newTable.lastRow = table.lastRow;
			newTable.timestamp = table.timestamp;
			newTable.nRows = table.nRows;
			newTable.desc = table.desc;
			registerTable(newTable);
		    }
		} catch (ClassNotFoundException x) {
		    registerTable(table);
		}	    
	    } else { 
		registerTable(table);
	    }
	    tableId = Record.getNext(table.desc, 0);
	}
	TableDescriptor last = lastTable;
	int lastTableId = metatable.lastRow;
	int metatableLastRow = 0;
	metatable.lastRow = 0;
	for (table = firstTable.next; table != null; table = table.next) { 
	    if (table.newTable != null) { 
		if (table.formatChanged) { 
		    reformatTable(table, table.newTable);
		    for (FieldDescriptor fd = table.fields; fd != null; fd = fd.next) {
			if (fd.index != 0) { 
			    Btree.drop(this, fd.index);
			    fd.index = 0;
			}
		    }
		}
		table.update(table.newTable);
		//
		// exportDescriptor will register in the database all base classes
		// which were not registered before. These classes will be appended
		// after metatable.lastRow. We modify metatable.lastRow to point before
		// the current class to make sure that all base classes are linked in 
		// the list before derived classes. 
		//
		table.exportDescriptor();
		if (metatableLastRow != metatable.lastRow) { 
		    byte[] desc = get(metatable.lastRow);
		    Record.setNext(desc, 0, table.tableId);
		    update(metatable.lastRow, desc);
		    Record.setPrev(table.desc, 0, metatable.lastRow);
		    update(table.tableId, table.desc);
		}
	    }
	    metatable.lastRow = metatableLastRow = table.tableId;
	    if (table == last) { 
		// Ignore table descriptors added during execution of this loop
		break;
	    }
	}
	Assert.that(metatable.lastRow == lastTableId);
	for (table = firstTable; table != null; table = table.next) { 
	    table.addIndices(true);
	}
	setForeignKeyConstraints();
	commitSubtransaction();
	commitTransaction(); // commit schema changes
    }

    void resetForeignKeyConstraints() throws SQLException { 
	for (TableDescriptor table = firstTable; table != null; table = table.next) { 
	    table.foreignKeys = null;
	    for (FieldDescriptor fd = table.fields; fd != null; fd = fd.next) { 
		fd.foreignKeys = null;
		if (fd.cascadeDeleteStmt != null) { 
		    fd.cascadeDeleteStmt.close();
		    fd.cascadeDeleteStmt = null;
		}
	    }
	}
	setForeignKeyConstraints();
    }

    void setForeignKeyConstraints() throws SQLException { 
	for (TableDescriptor table = firstTable; table != null; table = table.next) { 
	    for (FieldDescriptor fk = table.fields; fk != null; fk = fk.next) { 
		if ((fk.attr & (FieldDescriptor.ForeignKey|FieldDescriptor.OnDeleteCascade))
		    == (FieldDescriptor.ForeignKey|FieldDescriptor.OnDeleteCascade))
		{
		    TableDescriptor refTable = 
			(TableDescriptor)tableHashByName.get(fk.fkTableName);
		    if (refTable == null) { 
			throw new ConsusRuntimeError("Unknown table " + fk.fkTableName
							 + " in foreign key specification");
		    }
		    if (fk.fkKeyName.equals("oid")) { 
			if (fk.type != FieldDescriptor.tpReference) { 
			    throw new ConsusRuntimeError("Foreign key should have reference type");
			}
			fk.nextForeignKey = refTable.foreignKeys;
			refTable.foreignKeys = fk;
		    } else { 
			FieldDescriptor pk = refTable.getColumn(fk.fkKeyName);
			if (pk == null) { 
			    throw new ConsusRuntimeError("No such field " + fk.fkKeyName
							     + " in table " + fk.fkTableName
							     + " in foreign key specification");
			}
			if (pk.type != fk.type) { 
			    throw new ConsusRuntimeError("Types of primary and foreign keys do not match");
			}
			fk.nextForeignKey = pk.foreignKeys;
			pk.foreignKeys = fk;
		    }
		}
	    }
	}
    }

    

    long storeBlob(Blob blob) throws SQLException { 
	long length = blob.length();
	if (length == 0) { 
	    return 0;
	}
	long pos = allocate(length, 0);
	long currPos = pos;
	InputStream in = blob.getBinaryStream();
	try { 
	    while (length > 0) { 
		int pageOffs = (int)currPos & (Page.pageSize-1);
		Page pg = pool.putPage(currPos - pageOffs);
		if (length > Page.pageSize - pageOffs) { 
		    int rc = in.read(pg.data, pageOffs, Page.pageSize - pageOffs);
		    if (rc < 0) { 
			in.close();
			throw new ConsusRuntimeError("Unexpected end of input stream");
		    }
		    currPos += rc;
		    length -= rc;
		    pool.unfix(pg);
		} else { 
		    int rc = in.read(pg.data, pageOffs, (int)length);
		    if (rc < 0) { 
			in.close();
			throw new ConsusRuntimeError("Unexpected end of input stream"); 
		    }
		    currPos += rc;
		    length -= rc;
		    pool.unfix(pg);
		}
	    }
	    in.close();
	    return pos;
	} catch (IOException x) { 
	    throw new ConsusRuntimeError("Exception while reading input stream: " 
					 + x.getMessage());
	}
    }
	
    long storeClob(Clob clob) throws SQLException { 
	long length = clob.length();
	if (length == 0) { 
	    return 0;
	}
	long pos = allocate(length*2, 0);
	long currPos = pos;
	if (clob instanceof ConsusClob) { 
	    ((ConsusClob)clob).pos = pos;
	}
	Reader in = clob.getCharacterStream();
	char[] buf = new char[Page.pageSize >>> 1];
	try { 
	    while (length > 0) { 
		int pageOffs = (int)currPos & (Page.pageSize-1);
		Page pg = pool.putPage(currPos - pageOffs);
		if (length*2 > Page.pageSize - pageOffs) { 
		    int nChars = (Page.pageSize - pageOffs) >>> 1;
		    int rc = in.read(buf, 0, nChars);
		    if (rc < 0) { 
			in.close();
			throw new ConsusRuntimeError("Unexpected end of input stream");
		    }
		    currPos += rc*2;
		    length -= rc;
		    for (int i = 0; i < rc; i++, pageOffs += 2) { 
			Bytes.pack2(pg.data, pageOffs, (short)buf[i]);
		    }
		    pool.unfix(pg);
		} else { 
		    int rc = in.read(buf, 0, (int)length);
		    if (rc < 0) { 
			in.close();
			throw new ConsusRuntimeError("Unexpected end of input stream");
		    }
		    currPos += rc*2;
		    length -= rc;
		    for (int i = 0; i < rc; i++, pageOffs += 2) { 
			Bytes.pack2(pg.data, pageOffs, (short)buf[i]);
		    }		    
		    pool.unfix(pg);
		}
	    }
	    return pos;
	} catch (IOException x) { 
	    throw new ConsusRuntimeError("Exception while reading input stream: " 
					 + x.getMessage());
	}
    }
	
    public Blob createBlob(long length, InputStream in) { 
	return new ConsusBlob(length, in);
    }

    public Clob createClob(long length, Reader in) { 
	return new ConsusClob(length, in);
    }

    public Clob createClob(String str) { 
	return new ConsusClob(str);
    } 
    
    public void enableTracing(boolean enabled) 
    {
	dbTraceEnabled = enabled;
    }

    public Struct createStruct(String typeName, Object[] attributes) throws SQLException
    {
        throw new ConsusNotImplemented("Connection.createStruct");
    }

    public Array createArrayOf(String typeName, Object[] elements) throws SQLException
    {
        throw new ConsusNotImplemented("Connection.createArrayOf");
    }

    public Properties getClientInfo() throws SQLException {
        return null;
    }

    public String getClientInfo(String name) throws SQLException {
        return null;
    }

    public void setClientInfo(Properties properties) {
    }
    
    public void setClientInfo(String name, String value) {
    }

    public boolean isValid(int timeout) throws SQLException {
        return true;
    }

    public SQLXML createSQLXML() throws SQLException
    {
        throw new ConsusNotImplemented("Connection.createSQLXML");
    }

    public NClob createNClob() throws SQLException
    {
        throw new ConsusNotImplemented("Connection.createNClob");
    }

    public Clob createClob() throws SQLException
    {
        throw new ConsusNotImplemented("Connection.createClob");
    }

    public Blob createBlob() throws SQLException
    {
        throw new ConsusNotImplemented("Connection.createBlob");
    }

    public boolean isWrapperFor(Class iface) throws java.sql.SQLException
    {
        return false;
    }
    
    public Object unwrap(java.lang.Class iface) throws java.sql.SQLException
    {
	return null;
    }

    final int getType(int oid) throws SQLException { 
	long pos = getPos(oid);
	int offs = (int)pos & (Page.pageSize-1);
	Page pg = pool.getPage(pos - offs);
	int tableId = Record.getType(pg.data, offs & ~dbFlagsMask);
	pool.unfix(pg);
	return tableId;
    }

    final int getNext(int oid) throws SQLException { 
	long pos = getPos(oid);
	int offs = (int)pos & (Page.pageSize-1);
	Page pg = pool.getPage(pos - offs);
	int next = Record.getNext(pg.data, offs & ~dbFlagsMask);
	pool.unfix(pg);
	return next;
    }

    final int getPrev(int oid) throws SQLException { 
	long pos = getPos(oid);
	int offs = (int)pos & (Page.pageSize-1);
	Page pg = pool.getPage(pos - offs);
	int prev = Record.getPrev(pg.data, offs & ~dbFlagsMask);
	pool.unfix(pg);
	return prev;
    }

    
    /*
     ***********************************************
     * Instance variables
     ***********************************************
     */

    Hashtable ctxHash = new Hashtable();
    int       nNewThreads = 0;

    PagePool pool = new PagePool(dbPagePoolSize);
    TableDescriptor metatable;
    TableDescriptor firstTable;
    TableDescriptor lastTable;
    TableDescriptor deletedTables;
    Hashtable tableHashByName = new Hashtable();

    static final int dbTableHashByOidSize = 1009;
    TableDescriptor tableHashByOid[] = new TableDescriptor[dbTableHashByOidSize];

    Header    header;           // base address of database file mapping
    int       dirtyPagesMap[];  // bitmap of changed pages in current index
    boolean   modified;
    String    fileName;

    int       concurrentTransId;

    int       currRBitmapPage;//current bitmap page for allocating records
    int       currRBitmapOffs;//offset in current bitmap page for allocating 
                              //unaligned records
    int       currPBitmapPage;//current bitmap page for allocating page objects
    int       currPBitmapOffs;//offset in current bitmap page for allocating 
                              //page objects

    int       forceCommitCount;
    long      commitDelay;     
    long      commitTimeout;
    long      commitTimerStarted;
    Object    commitThreadSyncEvent;   
    Object    delayedCommitStartTimerEvent; // object used to notify delayed commit thread to schdule delayed commit
    Object    delayedCommitStopTimerEvent;  // object used by delaued commit thread to wait for sepcified timeout
    ThreadContext delayedCommitContext;     // save context of delayed transaction

    dbLocation reservedChain;

    int       committedIndexSize;
    int       currIndexSize;

    int       currIndex;  // copy of header.root, used to allow read access to the database 
                          // during transaction commit
    boolean   commitInProgress; 

    boolean   schemaAltered;

    long      usedSize;   // total size of allocated objects since the beginning of the session
    long      usedBeforeLastGC; // total size of allocated objects before last GC
 
    long      GCThreshold; 
    boolean   GCstarted;
    TableDescriptor[] rootTables;

    int       checkThreadAutoCommitState; // counter of threads with enabled autocommit mode

    int[]     bitmapPageAvailableSpace;
    boolean   opened;
    boolean   autoCommit = false; // not compatible wiith JDBC specification
    int       isolationLevel = TRANSACTION_SERIALIZABLE;  

    WeakHashTable objectCache;

    Object              backupEvent;
    Object              backupCS;    
    BackupThread        backupThread;
    DelayedCommitThread commitThread;
    boolean             backupInProgress;

    Monitor          monitor;
    Driver           driver;
    String           URL;
    Properties       properties;

    static Runtime   runtime;
}

class Monitor { 
    int     nReaders;
    int     nWriters;
    int     nWaitReaders;
    int     nWaitWriters;
    boolean waitForUpgrade;
};

class RootPage { 
    long size;            // database file size
    long index;           // offset to object index
    long shadowIndex;     // offset to shadow index
    long usedSize;        // size used by objects
    int  indexSize;       // size of object index
    int  shadowIndexSize; // size of object index
    int  indexUsed;       // userd part of the index   
    int  freeList;        // L1 list of free descriptors
    int  bitmapEnd;       // index of last allocated bitmap page
    // 12 reserved bytes
} 

class Header { 
    int      curr;  // current root
    boolean  dirty; // database was not closed normally
    boolean  initialized; // database is initilaized

    int      fmtid;   // format of transaction indentifier
    byte[]   gtrid;   // global transaction identifier 
    byte[]   bqual;   // transaction branch identifier 
    RootPage root[];
    
    final static int MAXGTRIDSIZE = 64; 
    final static int MAXBQUALSIZE = 64; 
    final static int sizeof = 3 + 4 + 2 + MAXGTRIDSIZE + MAXBQUALSIZE + (4*8 + 5*4 + 12)*2;
    
    final void pack(byte[] rec) { 
	int offs = 0;
	rec[offs++] = (byte)curr;
	rec[offs++] = (byte)(dirty ? 1 : 0);
	rec[offs++] = (byte)(initialized ? 1 : 0);
	Bytes.pack4(rec, offs, fmtid);
	offs += 4;
	if (gtrid != null) { 
	    rec[offs++] = (byte)gtrid.length;
	    System.arraycopy(gtrid, 0, rec, offs, gtrid.length);
	} else { 
	    rec[offs++] = 0;
	}
	offs += MAXGTRIDSIZE;
	if (bqual != null) { 
	    rec[offs++] = (byte)bqual.length;
	    System.arraycopy(bqual, 0, rec, offs, bqual.length);
	}else { 
	    rec[offs++] = 0;
	}
	offs += MAXBQUALSIZE;
	for (int i = 0; i < 2; i++) { 
	    Bytes.pack8(rec, offs, root[i].size);
	    offs += 8;
	    Bytes.pack8(rec, offs, root[i].index);
	    offs += 8;
	    Bytes.pack8(rec, offs, root[i].shadowIndex);
	    offs += 8;
	    Bytes.pack8(rec, offs, root[i].usedSize);
	    offs += 8;
	    Bytes.pack4(rec, offs, root[i].indexSize);
	    offs += 4;
	    Bytes.pack4(rec, offs, root[i].shadowIndexSize);
	    offs += 4;
	    Bytes.pack4(rec, offs, root[i].indexUsed);
	    offs += 4;
	    Bytes.pack4(rec, offs, root[i].freeList);
	    offs += 4;
	    Bytes.pack4(rec, offs, root[i].bitmapEnd);
	    offs += 16;
	}
    }
    
    final void unpack(byte[] rec) { 
	int offs = 0;
	curr = rec[offs++];
	dirty = rec[offs++] != 0;
	initialized = rec[offs++] != 0;
	fmtid = Bytes.unpack4(rec, offs);
	offs += 4;
	int len = rec[offs++];
	if (len != 0) { 
	    gtrid = new byte[len];
	    System.arraycopy(rec, offs, gtrid, 0, len);
	}
	offs += MAXGTRIDSIZE;
	len = rec[offs++];
	if (len != 0) { 
	    bqual = new byte[len];
	    System.arraycopy(rec, offs, bqual, 0, len);
	}
	offs += MAXBQUALSIZE;
	    
	root = new RootPage[2];
	for (int i = 0; i < 2; i++) { 
	    root[i] = new RootPage();
	    root[i].size = Bytes.unpack8(rec, offs);
	    offs += 8;
	    root[i].index = Bytes.unpack8(rec, offs);
	    offs += 8;
	    root[i].shadowIndex = Bytes.unpack8(rec, offs);
	    offs += 8;
	    root[i].usedSize = Bytes.unpack8(rec, offs);
	    offs += 8;
	    root[i].indexSize = Bytes.unpack4(rec, offs);
	    offs += 4;
	    root[i].shadowIndexSize = Bytes.unpack4(rec, offs);
	    offs += 4;
	    root[i].indexUsed = Bytes.unpack4(rec, offs);
	    offs += 4;
	    root[i].freeList = Bytes.unpack4(rec, offs);
	    offs += 4;
	    root[i].bitmapEnd = Bytes.unpack4(rec, offs);
	    offs += 16;
	}
    }	
}





