/* Consus - Pure Java Object Relational Database Management System
 * Implementation of DatabaseMetaData protocol. Runtime information about Consus classes/tables
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;
import java.sql.*;

/**
 * Comprehensive information about the database as a whole.
 *
 * <P>Many of the methods here return lists of information in 
 * the form of <code>ResultSet</code> objects.
 * You can use the normal ResultSet methods such as getString and getInt 
 * to retrieve the data from these ResultSets.  If a given form of
 * metadata is not available, these methods should throw an SQLException.
 *
 * <P>Some of these methods take arguments that are String patterns.  These
 * arguments all have names such as fooPattern.  Within a pattern String, "%"
 * means match any substring of 0 or more characters, and "_" means match
 * any one character. Only metadata entries matching the search pattern 
 * are returned. If a search pattern argument is set to a null ref, 
 * that argument's criteria will be dropped from the search.
 * 
 * <P>An <code>SQLException</code> will be thrown if a driver does not support a meta
 * data method.  In the case of methods that return a ResultSet,
 * either a ResultSet (which may be empty) is returned or a
 * SQLException is thrown.
 */
public class ConsusDatabaseMetaData implements DatabaseMetaData {

    //----------------------------------------------------------------------
    // First, a variety of minor information about the target database.

    /**
     * Can all the procedures returned by getProcedures be called by the
     * current user?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean allProceduresAreCallable() throws SQLException { 
	return true;
    }

    /**
     * Can all the tables returned by getTable be SELECTed by the
     * current user?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean allTablesAreSelectable() throws SQLException { 
	return true;
    }

    /**
     * What's the url for this database?
     *
     * @return the url or null if it cannot be generated
     * @exception SQLException if a database access error occurs
     */
    public String getURL() throws SQLException { 
	return db.URL;
    }

    /**
     * What's our user name as known to the database?
     *
     * @return our database user name
     * @exception SQLException if a database access error occurs
     */
    public String getUserName() throws SQLException { 
	return db.properties.getProperty("user", "public");
    }

    /**
     * Is the database in read-only mode?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean isReadOnly() throws SQLException { 
	return false;
    }

    /**
     * Are NULL values sorted high?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean nullsAreSortedHigh() throws SQLException { 
	return true;
    }

    /**
     * Are NULL values sorted low?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean nullsAreSortedLow() throws SQLException { 
	return false;
    }

    /**
     * Are NULL values sorted at the start regardless of sort order?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean nullsAreSortedAtStart() throws SQLException { 
	return false;
    }

    /**
     * Are NULL values sorted at the end regardless of sort order?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean nullsAreSortedAtEnd() throws SQLException { 
	return false;
    }

    /**
     * What's the name of this database product?
     *
     * @return database product name
     * @exception SQLException if a database access error occurs
     */
    public String getDatabaseProductName() throws SQLException { 
	String name = db.properties.getProperty("jdbc.driver.productName");
	if (name != null) { 
	    return name;
	}
	name = System.getProperty("jdbc.driver.productName");
	if (name != null) { 
	    return name;
	}
	return "Consus";
    }

    /**
     * What's the version of this database product?
     *
     * @return database version
     * @exception SQLException if a database access error occurs
     */
    public String getDatabaseProductVersion() throws SQLException { 
	return "1.1";
    }

    /**
     * What's the name of this JDBC driver?
     *
     * @return JDBC driver name
     * @exception SQLException if a database access error occurs
     */
    public String getDriverName() throws SQLException { 
	return db.driver.getClass().getName();
    }

    /**
     * What's the version of this JDBC driver?
     *
     * @return JDBC driver version
     * @exception SQLException if a database access error occurs
     */
    public String getDriverVersion() throws SQLException { 
	return "" + getDriverMajorVersion() + "." + getDriverMinorVersion();
    }

    /**
     * What's this JDBC driver's major version number?
     *
     * @return JDBC driver major version
     */
    public int getDriverMajorVersion() { 
	return db.driver.getMajorVersion();
    }

    /**
     * What's this JDBC driver's minor version number?
     *
     * @return JDBC driver minor version number
     */
    public int getDriverMinorVersion() { 
	return db.driver.getMinorVersion();
    }

    /**
     * Does the database store tables in a local file?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean usesLocalFiles() throws SQLException { 
	return true;
    }

    /**
     * Does the database use a file for each table?
     *
     * @return true if the database uses a local file for each table
     * @exception SQLException if a database access error occurs
     */
    public boolean usesLocalFilePerTable() throws SQLException { 
	return false;
    }

    /**
     * Does the database treat mixed case unquoted SQL identifiers as
     * case sensitive and as a result store them in mixed case?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver will always return false.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsMixedCaseIdentifiers() throws SQLException { 
	return true;
    }

    /**
     * Does the database treat mixed case unquoted SQL identifiers as
     * case insensitive and store them in upper case?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean storesUpperCaseIdentifiers() throws SQLException { 
	return false;
    }

    /**
     * Does the database treat mixed case unquoted SQL identifiers as
     * case insensitive and store them in lower case?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean storesLowerCaseIdentifiers() throws SQLException { 
	return false;
    }

    /**
     * Does the database treat mixed case unquoted SQL identifiers as
     * case insensitive and store them in mixed case?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean storesMixedCaseIdentifiers() throws SQLException { 
	return false;
    }

    /**
     * Does the database treat mixed case quoted SQL identifiers as
     * case sensitive and as a result store them in mixed case?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver will always return true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsMixedCaseQuotedIdentifiers() throws SQLException { 
	return true;
    }

    /**
     * Does the database treat mixed case quoted SQL identifiers as
     * case insensitive and store them in upper case?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean storesUpperCaseQuotedIdentifiers() throws SQLException { 
	return false;
    }

    /**
     * Does the database treat mixed case quoted SQL identifiers as
     * case insensitive and store them in lower case?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean storesLowerCaseQuotedIdentifiers() throws SQLException { 
	return false;
    }

    /**
     * Does the database treat mixed case quoted SQL identifiers as
     * case insensitive and store them in mixed case?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean storesMixedCaseQuotedIdentifiers() throws SQLException { 
	return false;
    }
    
    /**
     * What's the string used to quote SQL identifiers?
     * This returns a space " " if identifier quoting isn't supported.
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> 
	 * driver always uses a double quote character.
     *
     * @return the quoting string
     * @exception SQLException if a database access error occurs
     */
    public String getIdentifierQuoteString() throws SQLException { 
	return " ";
    }

    /**
     * Gets a comma-separated list of all a database's SQL keywords
     * that are NOT also SQL92 keywords.
     *
     * @return the list 
     * @exception SQLException if a database access error occurs
     */
    public String getSQLKeywords() throws SQLException { 
	return "start,following,reference,to,string,blob,clob,boolean,first,last,sysdate";
    }

    /**
     * Gets a comma-separated list of math functions.  These are the 
     * X/Open CLI math function names used in the JDBC function escape 
     * clause.
     *
     * @return the list
     * @exception SQLException if a database access error occurs
     */
    public String getNumericFunctions() throws SQLException { 
	return "abs,integer,real,sin,cos,tan,asin,acos,atan,log,exp,ceil,floor";
    }

    /**
     * Gets a comma-separated list of string functions.  These are the 
     * X/Open CLI string function names used in the JDBC function escape 
     * clause.
     *
     * @return the list
     * @exception SQLException if a database access error occurs
     */
    public String getStringFunctions() throws SQLException { 
	return "string,length,lower,upper,substr";
    }


    /**
     * Gets a comma-separated list of system functions.  These are the 
     * X/Open CLI system function names used in the JDBC function escape 
     * clause.
     *
     * @return the list
     * @exception SQLException if a database access error occurs
     */
    public String getSystemFunctions() throws SQLException { 
	return "";
    }

    /**
     * Gets a comma-separated list of time and date functions.
     *
     * @return the list
     * @exception SQLException if a database access error occurs
     */
    public String getTimeDateFunctions() throws SQLException { 
	return "sysdate";
    }

    /**
     * Gets the string that can be used to escape wildcard characters.
     * This is the string that can be used to escape '_' or '%' in
     * the string pattern style catalog search parameters.
     *
     * <P>The '_' character represents any single character.
     * <P>The '%' character represents any sequence of zero or 
     * more characters.
     *
     * @return the string used to escape wildcard characters
     * @exception SQLException if a database access error occurs
     */
    public String getSearchStringEscape() throws SQLException { 
	return "\\";
    }

    /**
     * Gets all the "extra" characters that can be used in unquoted
     * identifier names (those beyond a-z, A-Z, 0-9 and _).
     *
     * @return the string containing the extra characters 
     * @exception SQLException if a database access error occurs
     */
    public String getExtraNameCharacters() throws SQLException { 
	return "$";
    }

    //--------------------------------------------------------------------
    // Functions describing which features are supported.

    /**
     * Is "ALTER TABLE" with add column supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsAlterTableWithAddColumn() throws SQLException { 
	return true;
    }
    
    /**
     * Is "ALTER TABLE" with drop column supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsAlterTableWithDropColumn() throws SQLException { 
	return true;
    }

    /**
     * Is column aliasing supported? 
     *
     * <P>If so, the SQL AS clause can be used to provide names for
     * computed columns or to provide alias names for columns as
     * required.
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsColumnAliasing() throws SQLException { 
	return true;
    }

    /**
     * Are concatenations between NULL and non-NULL values NULL?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean nullPlusNonNullIsNull() throws SQLException { 
	return true;
    }

    /**
     * Is the CONVERT function between SQL types supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsConvert() throws SQLException { 
	return false;
    }
    
    /**
     * Is CONVERT between the given SQL types supported?
     *
     * @param fromType the type to convert from
     * @param toType the type to convert to     
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsConvert(int fromType, int toType) throws SQLException { 
	return false;
    }

    /**
     * Are table correlation names supported?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsTableCorrelationNames() throws SQLException { 
	return true;
    }

    /**
     * If table correlation names are supported, are they restricted
     * to be different from the names of the tables?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsDifferentTableCorrelationNames() throws SQLException { 
	return false;
    }
    
    /**
     * Are expressions in "ORDER BY" lists supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsExpressionsInOrderBy() throws SQLException { 
	return false;
    }

    /**
     * Can an "ORDER BY" clause use columns not in the SELECT statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsOrderByUnrelated() throws SQLException { 
	return false;
    }

    /**
     * Is some form of "GROUP BY" clause supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsGroupBy() throws SQLException { 
	return true;
    }

    /**
     * Can a "GROUP BY" clause use columns not in the SELECT?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsGroupByUnrelated() throws SQLException { 
	return true;
    }
    
    /**
     * Can a "GROUP BY" clause add columns not in the SELECT
     * provided it specifies all the columns in the SELECT?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsGroupByBeyondSelect() throws SQLException { 
	return true;
    }

    /**
     * Is the escape character in "LIKE" clauses supported?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsLikeEscapeClause() throws SQLException { 
	return true;
    }

    /**
     * Are multiple ResultSets from a single execute supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsMultipleResultSets() throws SQLException { 
	return false;
    }

    /**
     * Can we have multiple transactions open at once (on different
     * connections)?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsMultipleTransactions() throws SQLException { 
	return true;
    }

    /**
     * Can columns be defined as non-nullable?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsNonNullableColumns() throws SQLException { 
	return true;
    }

    /**
     * Is the ODBC Minimum SQL grammar supported?
     *
     * All JDBC Compliant<sup><font size=-2>TM</font></sup> drivers must return true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsMinimumSQLGrammar() throws SQLException { 
	return true;
    }

    /**
     * Is the ODBC Core SQL grammar supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsCoreSQLGrammar() throws SQLException { 
	return true;
    }

    /**
     * Is the ODBC Extended SQL grammar supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsExtendedSQLGrammar() throws SQLException { 
	return false;
    }

    /**
     * Is the ANSI92 entry level SQL grammar supported?
     *
     * All JDBC Compliant<sup><font size=-2>TM</font></sup> drivers must return true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsANSI92EntryLevelSQL() throws SQLException { 
	return true;
    }

    /**
     * Is the ANSI92 intermediate SQL grammar supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsANSI92IntermediateSQL() throws SQLException { 
	return false;
    }
    
    /**
     * Is the ANSI92 full SQL grammar supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsANSI92FullSQL() throws SQLException { 
	return false;
    }

    /**
     * Is the SQL Integrity Enhancement Facility supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsIntegrityEnhancementFacility() throws SQLException { 
	return false;
    }

    /**
     * Is some form of outer join supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsOuterJoins() throws SQLException { 
	return false;
    }

    /**
     * Are full nested outer joins supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsFullOuterJoins() throws SQLException { 
	return false;
    }

    /**
     * Is there limited support for outer joins?  (This will be true
     * if supportFullOuterJoins is true.)
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsLimitedOuterJoins() throws SQLException { 
	return false;
    }

    /**
     * What's the database vendor's preferred term for "schema"?
     *
     * @return the vendor term
     * @exception SQLException if a database access error occurs
     */
    public String getSchemaTerm() throws SQLException { 
	return "";
    }

    /**
     * What's the database vendor's preferred term for "procedure"?
     *
     * @return the vendor term
     * @exception SQLException if a database access error occurs
     */
    public String getProcedureTerm() throws SQLException { 
	return "";
    }

    /**
     * What's the database vendor's preferred term for "catalog"?
     *
     * @return the vendor term
     * @exception SQLException if a database access error occurs
     */
    public String getCatalogTerm() throws SQLException { 
	return "";
    }

    /**
     * Does a catalog appear at the start of a qualified table name?
     * (Otherwise it appears at the end)
     *
     * @return true if it appears at the start 
     * @exception SQLException if a database access error occurs
     */
    public boolean isCatalogAtStart() throws SQLException { 
	return false;
    }

    /**
     * What's the separator between catalog and table name?
     *
     * @return the separator string
     * @exception SQLException if a database access error occurs
     */
    public String getCatalogSeparator() throws SQLException { 
	return "";
    }

    /**
     * Can a schema name be used in a data manipulation statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSchemasInDataManipulation() throws SQLException { 
	return false;
    }

    /**
     * Can a schema name be used in a procedure call statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSchemasInProcedureCalls() throws SQLException { 
	return false;
    }

    /**
     * Can a schema name be used in a table definition statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSchemasInTableDefinitions() throws SQLException { 
	return false;
    }

    /**
     * Can a schema name be used in an index definition statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSchemasInIndexDefinitions() throws SQLException { 
	return false;
    }

    /**
     * Can a schema name be used in a privilege definition statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSchemasInPrivilegeDefinitions() throws SQLException { 
	return false;
    }

    /**
     * Can a catalog name be used in a data manipulation statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsCatalogsInDataManipulation() throws SQLException { 
	return false;
    }

    /**
     * Can a catalog name be used in a procedure call statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsCatalogsInProcedureCalls() throws SQLException { 
	return false;
    }

    /**
     * Can a catalog name be used in a table definition statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsCatalogsInTableDefinitions() throws SQLException { 
	return false;
    }

    /**
     * Can a catalog name be used in an index definition statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsCatalogsInIndexDefinitions() throws SQLException { 
	return false;
    }

    /**
     * Can a catalog name be used in a privilege definition statement?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsCatalogsInPrivilegeDefinitions() throws SQLException { 
	return false;
    }


    /**
     * Is positioned DELETE supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsPositionedDelete() throws SQLException { 
	return true;
    }

    /**
     * Is positioned UPDATE supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsPositionedUpdate() throws SQLException { 
	return true;
    }

    /**
     * Is SELECT for UPDATE supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSelectForUpdate() throws SQLException {
	return true;
    }

    /**
     * Are stored procedure calls using the stored procedure escape
     * syntax supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsStoredProcedures() throws SQLException { 
	return false;
    }

    /**
     * Are subqueries in comparison expressions supported?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSubqueriesInComparisons() throws SQLException { 
	return true;
    }

    /**
     * Are subqueries in 'exists' expressions supported?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSubqueriesInExists() throws SQLException { 
	return true;
    }

    /**
     * Are subqueries in 'in' statements supported?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSubqueriesInIns() throws SQLException { 
	return true;
    }

    /**
     * Are subqueries in quantified expressions supported?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsSubqueriesInQuantifieds() throws SQLException { 
	return true;
    }

    /**
     * Are correlated subqueries supported?
     *
     * A JDBC Compliant<sup><font size=-2>TM</font></sup> driver always returns true.
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsCorrelatedSubqueries() throws SQLException { 
	return true;
    }

    /**
     * Is SQL UNION supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsUnion() throws SQLException { 
	return true;
    }

    /**
     * Is SQL UNION ALL supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsUnionAll() throws SQLException { 
	return true;
    }

    /**
     * Can cursors remain open across commits? 
     * 
     * @return <code>true</code> if cursors always remain open;
	 *       <code>false</code> if they might not remain open
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsOpenCursorsAcrossCommit() throws SQLException { 
	return false;
    }

    /**
     * Can cursors remain open across rollbacks?
     * 
     * @return <code>true</code> if cursors always remain open;
	 *       <code>false</code> if they might not remain open
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsOpenCursorsAcrossRollback() throws SQLException { 
	return false;
    }

    /**
     * Can statements remain open across commits?
     * 
     * @return <code>true</code> if statements always remain open;
	 *       <code>false</code> if they might not remain open
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsOpenStatementsAcrossCommit() throws SQLException { 
	return true;
    }

    /**
     * Can statements remain open across rollbacks?
     * 
     * @return <code>true</code> if statements always remain open;
     *       <code>false</code> if they might not remain open
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsOpenStatementsAcrossRollback() throws SQLException { 
	return true;
    }

	

    //----------------------------------------------------------------------
    // The following group of methods exposes various limitations 
    // based on the target database with the current driver.
    // Unless otherwise specified, a result of zero means there is no
    // limit, or the limit is not known.
	
    /**
     * How many hex characters can you have in an inline binary literal?
     *
     * @return max binary literal length in hex characters;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxBinaryLiteralLength() throws SQLException { 
	return 0;
    }

    /**
     * What's the max length for a character literal?
     *
     * @return max literal length;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxCharLiteralLength() throws SQLException { 
	return 0;
    }

    /**
     * What's the limit on column name length?
     *
     * @return max column name length;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxColumnNameLength() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum number of columns in a "GROUP BY" clause?
     *
     * @return max number of columns;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxColumnsInGroupBy() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum number of columns allowed in an index?
     *
     * @return max number of columns;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxColumnsInIndex() throws SQLException { 
	return 1;
    }

    /**
     * What's the maximum number of columns in an "ORDER BY" clause?
     *
     * @return max number of columns;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxColumnsInOrderBy() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum number of columns in a "SELECT" list?
     *
     * @return max number of columns;
	 *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxColumnsInSelect() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum number of columns in a table?
     *
     * @return max number of columns;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxColumnsInTable() throws SQLException { 
	return 0;
    }

    /**
     * How many active connections can we have at a time to this database?
     *
     * @return max number of active connections;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxConnections() throws SQLException { 
	return 1;
    }

    /**
     * What's the maximum cursor name length?
     *
     * @return max cursor name length in bytes;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxCursorNameLength() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum length of an index (in bytes)?	
     *
     * @return max index length in bytes;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxIndexLength() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum length allowed for a schema name?
     *
     * @return max name length in bytes;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxSchemaNameLength() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum length of a procedure name?
     *
     * @return max name length in bytes; 
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxProcedureNameLength() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum length of a catalog name?
     *
     * @return max name length in bytes;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxCatalogNameLength() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum length of a single row?
     *
     * @return max row size in bytes;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxRowSize() throws SQLException { 
	return 0;
    }

    /**
     * Did getMaxRowSize() include LONGVARCHAR and LONGVARBINARY
     * blobs?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean doesMaxRowSizeIncludeBlobs() throws SQLException { 
	return false;
    }

    /**
     * What's the maximum length of a SQL statement?
     *
     * @return max length in bytes;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxStatementLength() throws SQLException { 
	return 0;
    }
    
    /**
     * How many active statements can we have open at one time to this
     * database?
     *
     * @return the maximum number of statements that can be open at one time;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxStatements() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum length of a table name?
     *
     * @return max name length in bytes;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxTableNameLength() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum number of tables in a SELECT statement?
     *
     * @return the maximum number of tables allowed in a SELECT statement;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxTablesInSelect() throws SQLException { 
	return 0;
    }

    /**
     * What's the maximum length of a user name?
     *
     * @return max user name length  in bytes;
     *      a result of zero means that there is no limit or the limit is not known
     * @exception SQLException if a database access error occurs
     */
    public int getMaxUserNameLength() throws SQLException { 
	return 0;
    }

    //----------------------------------------------------------------------

    /**
     * What's the database's default transaction isolation level?  The
     * values are defined in <code>java.sql.Connection</code>.
     *
     * @return the default isolation level 
     * @exception SQLException if a database access error occurs
     * @see Connection
     */
    public int getDefaultTransactionIsolation() throws SQLException { 
	return Connection.TRANSACTION_SERIALIZABLE;
    }

    /**
     * Are transactions supported? If not, invoking the method
     * <code>commit</code> is a noop and the
     * isolation level is TRANSACTION_NONE.
     *
     * @return <code>true</code> if transactions are supported; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsTransactions() throws SQLException { 
	return true;
    }


    /**
     * Does this database support the given transaction isolation level?
     *
     * @param level the values are defined in <code>java.sql.Connection</code>
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     * @see Connection
     */
    public boolean supportsTransactionIsolationLevel(int level) throws SQLException { 
	return level == Connection.TRANSACTION_SERIALIZABLE || level == Connection.TRANSACTION_NONE;
    }

    /**
     * Are both data definition and data manipulation statements
     * within a transaction supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsDataDefinitionAndDataManipulationTransactions()
	throws SQLException
    {
	return true;
    }
    /**
     * Are only data manipulation statements within a transaction
     * supported?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean supportsDataManipulationTransactionsOnly()throws SQLException { 
	return false;
    }

    /**
     * Does a data definition statement within a transaction force the
     * transaction to commit?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean dataDefinitionCausesTransactionCommit() throws SQLException { 
	return false;
    }

    /**
     * Is a data definition statement within a transaction ignored?
     *
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     */
    public boolean dataDefinitionIgnoredInTransactions() throws SQLException { 
	return false;
    }


    /**
     * Gets a description of the stored procedures available in a
     * catalog.
     *
     * <P>Only procedure descriptions matching the schema and
     * procedure name criteria are returned.  They are ordered by
     * PROCEDURE_SCHEM, and PROCEDURE_NAME.
     *
     * <P>Each procedure description has the the following columns:
     *  <OL>
     *	<LI><B>PROCEDURE_CAT</B> String => procedure catalog (may be null)
     *	<LI><B>PROCEDURE_SCHEM</B> String => procedure schema (may be null)
     *	<LI><B>PROCEDURE_NAME</B> String => procedure name
     *  <LI> reserved for future use
     *  <LI> reserved for future use
     *  <LI> reserved for future use
     *	<LI><B>REMARKS</B> String => explanatory comment on the procedure
     *	<LI><B>PROCEDURE_TYPE</B> short => kind of procedure:
     *      <UL>
     *      <LI> procedureResultUnknown - May return a result
     *      <LI> procedureNoResult - Does not return a result
     *      <LI> procedureReturnsResult - Returns a result
     *      </UL>
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those
     * without a schema
     * @param procedureNamePattern a procedure name pattern 
     * @return ResultSet - each row is a procedure description 
     * @exception SQLException if a database access error occurs
     * @see #getSearchStringEscape 
     */
    public ResultSet getProcedures(String catalog, String schemaPattern,
				   String procedureNamePattern) throws SQLException
    {
	throw new ConsusNotImplemented("DatabaseMetaData.getProcedures");
    }


    /**
     * Gets a description of a catalog's stored procedure parameters
     * and result columns.
     *
     * <P>Only descriptions matching the schema, procedure and
     * parameter name criteria are returned.  They are ordered by
     * PROCEDURE_SCHEM and PROCEDURE_NAME. Within this, the return value,
     * if any, is first. Next are the parameter descriptions in call
     * order. The column descriptions follow in column number order.
     *
     * <P>Each row in the ResultSet is a parameter description or
     * column description with the following fields:
     *  <OL>
     *	<LI><B>PROCEDURE_CAT</B> String => procedure catalog (may be null)
     *	<LI><B>PROCEDURE_SCHEM</B> String => procedure schema (may be null)
     *	<LI><B>PROCEDURE_NAME</B> String => procedure name
     *	<LI><B>COLUMN_NAME</B> String => column/parameter name 
     *	<LI><B>COLUMN_TYPE</B> Short => kind of column/parameter:
     *      <UL>
     *      <LI> procedureColumnUnknown - nobody knows
     *      <LI> procedureColumnIn - IN parameter
     *      <LI> procedureColumnInOut - INOUT parameter
     *      <LI> procedureColumnOut - OUT parameter
     *      <LI> procedureColumnReturn - procedure return value
     *      <LI> procedureColumnResult - result column in ResultSet
     *      </UL>
     *  <LI><B>DATA_TYPE</B> short => SQL type from java.sql.Types
     *	<LI><B>TYPE_NAME</B> String => SQL type name, for a UDT type the
     *  type name is fully qualified
     *	<LI><B>PRECISION</B> int => precision
     *	<LI><B>LENGTH</B> int => length in bytes of data
     *	<LI><B>SCALE</B> short => scale
     *	<LI><B>RADIX</B> short => radix
     *	<LI><B>NULLABLE</B> short => can it contain NULL?
     *      <UL>
     *      <LI> procedureNoNulls - does not allow NULL values
     *      <LI> procedureNullable - allows NULL values
     *      <LI> procedureNullableUnknown - nullability unknown
     *      </UL>
     *	<LI><B>REMARKS</B> String => comment describing parameter/column
     *  </OL>
     *
     * <P><B>Note:</B> Some databases may not return the column
     * descriptions for a procedure. Additional columns beyond
     * REMARKS can be defined by the database.
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those
     * without a schema 
     * @param procedureNamePattern a procedure name pattern 
     * @param columnNamePattern a column name pattern 
     * @return ResultSet - each row describes a stored procedure parameter or 
     *      column
     * @exception SQLException if a database access error occurs
     * @see #getSearchStringEscape 
     */
    public ResultSet getProcedureColumns(String catalog,
					 String schemaPattern,
					 String procedureNamePattern, 
					 String columnNamePattern) throws SQLException
    {
	throw new ConsusNotImplemented("DatabaseMetaData.getProcedureColumns");
    }
    
    /**
     * Gets a description of tables available in a catalog.
     *
     * <P>Only table descriptions matching the catalog, schema, table
     * name and type criteria are returned.  They are ordered by
     * TABLE_TYPE, TABLE_SCHEM and TABLE_NAME.
     *
     * <P>Each table description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>TABLE_TYPE</B> String => table type.  Typical types are "TABLE",
     *			"VIEW",	"SYSTEM TABLE", "GLOBAL TEMPORARY", 
     *			"LOCAL TEMPORARY", "ALIAS", "SYNONYM".
     *	<LI><B>REMARKS</B> String => explanatory comment on the table
     *  </OL>
     *
     * <P><B>Note:</B> Some databases may not return information for
     * all tables.
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those
     * without a schema
     * @param tableNamePattern a table name pattern 
     * @param types a list of table types to include; null returns all types 
     * @return ResultSet - each row is a table description
     * @exception SQLException if a database access error occurs
     * @see #getSearchStringEscape 
     */
    public ResultSet getTables(String catalog, String schemaPattern,
			       String tableNamePattern, String types[]) throws SQLException
    {
	int i;
	ConsusResultSet cursor = new ConsusResultSet(null);
	TableDescriptor firstTable = db.firstTable;
	boolean skipSystemTables = true;

	if (types != null) { 
	    skipSystemTables = true;
	    for (i = types.length; --i >= 0;) { 
		if (types[i].equals("SYSTEM TABLE")) { 
		    skipSystemTables = false;
		    break;
		}
	    }
	}
	if (skipSystemTables) { 
	    firstTable = firstTable.next;
	}
	for (TableDescriptor table = firstTable; table != null; table = table.next) { 
	    if (Regexp.matchString(table.name, tableNamePattern, '\\')) { 
		Object[] tuple = new Object[5];
		tuple[0] = "";
		tuple[1] = "";
		tuple[2] = table.name;
		tuple[3] = table.tableId != ConsusConnection.dbMetaTableId 
		    ? "TABLE" : "SYSTEM TABLE";
		tuple[4] = "";
		cursor.addTuple(tuple);
	    }
	}
	ColumnNode columns[] = new ColumnNode[5];
	for (i = 0; i < 5; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}
	columns[0].name = "TABLE_CAT";
	columns[1].name = "TABLE_SCHEM";
	columns[2].name = "TABLE_NAME";
	columns[3].name = "TABLE_TYPE";
	columns[4].name = "REMARKS";
	cursor.tupleColumns = columns;
	return cursor;
    }

    /**
     * Gets the schema names available in this database.  The results
     * are ordered by schema name.
     *
     * <P>The schema column is:
     *  <OL>
     *	<LI><B>TABLE_SCHEM</B> String => schema name
     *  </OL>
     *
     * @return ResultSet - each row has a single String column that is a
     * schema name 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getSchemas() throws SQLException { 
	ConsusResultSet cursor = new ConsusResultSet(null);
        ColumnNode columns[] = new ColumnNode[1];
	columns[0] = new ColumnNode(1);
	columns[0].name = "TABLE_SCHEM";
 	cursor.nSelected = 0;
	cursor.tupleColumns = columns;
	return cursor;
    }

    /**
     * Gets the catalog names available in this database.  The results
     * are ordered by catalog name.
     *
     * <P>The catalog column is:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => catalog name
     *  </OL>
     *
     * @return ResultSet - each row has a single String column that is a
     * catalog name 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getCatalogs() throws SQLException { 
	ConsusResultSet cursor = new ConsusResultSet(null);
	ColumnNode columns[] = new ColumnNode[1];
	columns[0] = new ColumnNode(1);
	columns[0].name = "TABLE_CAT";
 	cursor.nSelected = 0;
	cursor.tupleColumns = columns;
	return cursor;
    }

    /**
     * Gets the table types available in this database.  The results
     * are ordered by table type.
     *
     * <P>The table type is:
     *  <OL>
     *	<LI><B>TABLE_TYPE</B> String => table type.  Typical types are "TABLE",
     *			"VIEW",	"SYSTEM TABLE", "GLOBAL TEMPORARY", 
     *			"LOCAL TEMPORARY", "ALIAS", "SYNONYM".
     *  </OL>
     *
     * @return ResultSet - each row has a single String column that is a
     * table type 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getTableTypes() throws SQLException { 
	ConsusResultSet cursor = new ConsusResultSet(null);
	Object[][] tuples = new Object[2][1];
	tuples[0][0] = "TABLE";
	tuples[1][0] = "SYSTEM TABLE";
	ColumnNode columns[] = new ColumnNode[1];
	columns[0] = new ColumnNode(1);
	columns[0].name = "TABLE_TYPE";
 	cursor.nSelected = 2;
	cursor.tuples = tuples;
	cursor.tupleColumns = columns;
	return cursor;
    }


    /**
     * Gets a description of table columns available in 
     * the specified catalog.
     *
     * <P>Only column descriptions matching the catalog, schema, table
     * and column name criteria are returned.  They are ordered by
     * TABLE_SCHEM, TABLE_NAME and ORDINAL_POSITION.
     *
     * <P>Each column description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>COLUMN_NAME</B> String => column name
     *	<LI><B>DATA_TYPE</B> short => SQL type from java.sql.Types
     *	<LI><B>TYPE_NAME</B> String => Data source dependent type name,
     *  for a UDT the type name is fully qualified
     *	<LI><B>COLUMN_SIZE</B> int => column size.  For char or date
     *	    types this is the maximum number of characters, for numeric or
     *	    decimal types this is precision.
     *	<LI><B>BUFFER_LENGTH</B> is not used.
     *	<LI><B>DECIMAL_DIGITS</B> int => the number of fractional digits
     *	<LI><B>NUM_PREC_RADIX</B> int => Radix (typically either 10 or 2)
     *	<LI><B>NULLABLE</B> int => is NULL allowed?
     *      <UL>
     *      <LI> columnNoNulls - might not allow NULL values
     *      <LI> columnNullable - definitely allows NULL values
     *      <LI> columnNullableUnknown - nullability unknown
     *      </UL>
     *	<LI><B>REMARKS</B> String => comment describing column (may be null)
     * 	<LI><B>COLUMN_DEF</B> String => default value (may be null)
     *	<LI><B>SQL_DATA_TYPE</B> int => unused
     *	<LI><B>SQL_DATETIME_SUB</B> int => unused
     *	<LI><B>CHAR_OCTET_LENGTH</B> int => for char types the 
     *       maximum number of bytes in the column
     *	<LI><B>ORDINAL_POSITION</B> int	=> index of column in table 
     *      (starting at 1)
     *	<LI><B>IS_NULLABLE</B> String => "NO" means column definitely 
     *      does not allow NULL values; "YES" means the column might 
     *      allow NULL values.  An empty string means nobody knows.
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those
     * without a schema
     * @param tableNamePattern a table name pattern 
     * @param columnNamePattern a column name pattern 
     * @return ResultSet - each row is a column description
     * @exception SQLException if a database access error occurs
     * @see #getSearchStringEscape 
     */
    public ResultSet getColumns(String catalog, String schemaPattern,
				String tableNamePattern, String columnNamePattern)
	throws SQLException
    {
	ConsusResultSet cursor = new ConsusResultSet(null);
	TableDescriptor table;
	int i;
	for (table = db.firstTable; table != null; table = table.next) { 
	    if (Regexp.matchString(table.name, tableNamePattern, '\\')) { 
		for (i = 0; i < table.columns.length; i++) {
		    Object[] tuple = new Object[18];
		    FieldDescriptor fd = table.columns[i];
		    if (!Regexp.matchString(fd.longName, columnNamePattern, '\\')) { 
			continue;
		    }
		    tuple[0] = "";
		    tuple[1] = "";
		    tuple[2] = table.name;
		    tuple[3] = fd.longName;
		    int type = Types.OTHER;
		    String typeName = "";
		    switch (fd.type) { 
		      case FieldDescriptor.tpBool:
			type = Types.BIT;
			typeName = "boolean";
			break;
		      case FieldDescriptor.tpInt1:
			type = Types.TINYINT;
			typeName = "tinyint";
			break;
		      case FieldDescriptor.tpInt2:
			type = Types.SMALLINT;
			typeName = "smallint";
			break;
		      case FieldDescriptor.tpInt4:
			type = Types.INTEGER;
			typeName = "integer";
			break;
		      case FieldDescriptor.tpInt8:
			type = Types.BIGINT;
			typeName = "bigint";
			break;
		      case FieldDescriptor.tpReal4:
			type = Types.FLOAT;
			typeName = "float";
			break;
		      case FieldDescriptor.tpReal8:
			type = Types.DOUBLE;
			typeName = "double";
			break;
		      case FieldDescriptor.tpString:
			type = Types.VARCHAR;
			typeName = "string";
			break;
		      case FieldDescriptor.tpReference:
			type = Types.REF;
			typeName = "reference";
			break;
		      case FieldDescriptor.tpDate:
			type = Types.DATE;
			typeName = "date";
			break;
		      case FieldDescriptor.tpTime:
			type = Types.TIME;
			typeName = "time";
			break;
		      case FieldDescriptor.tpTimestamp:
			type = Types.TIMESTAMP;
			typeName = "timestamp";
			break;
		      case FieldDescriptor.tpArray:
			if (fd.components[0].type == FieldDescriptor.tpInt1) { 
                            type = Types.VARBINARY;
			    typeName = "binary";
			} else { 
			    type = Types.ARRAY;
			    typeName = "";
			}
			break;
		      case FieldDescriptor.tpStructure:
			type = Types.STRUCT;
			typeName = "";
			break;
		      case FieldDescriptor.tpBlob:
			type = Types.BLOB;
			typeName = "blob";
			break;
		      case FieldDescriptor.tpClob:
			type = Types.CLOB;
			typeName = "clob";
		    }
		    tuple[4] = new Long(type);
		    tuple[5] = typeName;
		    tuple[6] = new Long(0); // column size
		    tuple[7] = new Long(0); // buffer length
		    tuple[8] = new Long(0); // decimal digits
		    tuple[9] = new Long(10); // radix
		    tuple[10] = new Long((fd.type == FieldDescriptor.tpReference)
					 ? columnNullable : columnNoNulls); // nullable
		    tuple[11] = (fd.attr & FieldDescriptor.Indexed) != 0 ? "<indexed>" : "";  // remarks
		    tuple[12] = "";  // column default value
		    tuple[13] = new Long(0);  // SQL_DATA_TYPE
		    tuple[14] = new Long(0);  // SQL_DATETIME_SUB
		    tuple[15] = new Long(2);  // CHAR_OCTET_LENGT
 		    tuple[16] = new Long(i+1);  // ORDINAL_POSITION
		    tuple[17] = (fd.type == FieldDescriptor.tpReference) ? "YES" : "NO";
		    cursor.addTuple(tuple);
		}
	    }
	}
	ColumnNode columns[] = new ColumnNode[18];
	for (i = 0; i < 18; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}
	columns[0].name = "TABLE_CAT";
	columns[1].name = "TABLE_SCHEM";
	columns[2].name = "TABLE_NAME";
	columns[3].name = "COLUMN_NAME";
	columns[4].name = "DATA_TYPE";
	columns[4].type = ExprNode.tpInt;
	columns[5].name = "TYPE_NAME";
	columns[6].name = "COLUMN_SIZE";
	columns[6].type = ExprNode.tpInt;
	columns[7].name = "BUFFER_LENGTH";
	columns[7].type = ExprNode.tpInt;
	columns[8].name = "DECIMAL_DIGITS";
	columns[8].type = ExprNode.tpInt;
	columns[9].name = "NUM_PREC_RADIX";
	columns[9].type = ExprNode.tpInt;
	columns[10].name = "NULLABLE";
	columns[10].type = ExprNode.tpInt;
	columns[11].name = "REMARKS";
	columns[12].name = "COLUMN_DEF";
	columns[13].name = "SQL_DATA_TYPE";
	columns[13].type = ExprNode.tpInt;
	columns[14].name = "SQL_DATETIME_SUB";
	columns[14].type = ExprNode.tpInt;
	columns[15].name = "CHAR_OCTET_LENGTH";
	columns[15].type = ExprNode.tpInt;
	columns[16].name = "ORDINAL_POSITION";
	columns[16].type = ExprNode.tpInt;
	columns[17].name = "IS_NULLABLE";
	cursor.tupleColumns = columns;
	return cursor;	
    }

    /**
     * Gets a description of the access rights for a table's columns.
     *
     * <P>Only privileges matching the column name criteria are
     * returned.  They are ordered by COLUMN_NAME and PRIVILEGE.
     *
     * <P>Each privilige description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>COLUMN_NAME</B> String => column name
     *	<LI><B>GRANTOR</B> => grantor of access (may be null)
     *	<LI><B>GRANTEE</B> String => grantee of access
     *	<LI><B>PRIVILEGE</B> String => name of access (SELECT, 
     *      INSERT, UPDATE, REFRENCES, ...)
     *	<LI><B>IS_GRANTABLE</B> String => "YES" if grantee is permitted 
     *      to grant to others; "NO" if not; null if unknown 
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those without a schema
     * @param table a table name
     * @param columnNamePattern a column name pattern 
     * @return ResultSet - each row is a column privilege description
     * @exception SQLException if a database access error occurs
     * @see #getSearchStringEscape 
     */
    public ResultSet getColumnPrivileges(String catalog, String schema,
					 String table, String columnNamePattern) throws SQLException
    {
	throw new ConsusNotImplemented("DatabaseResultSet.getColumnPrivileges");
    }


    /**
     * Gets a description of the access rights for each table available
     * in a catalog. Note that a table privilege applies to one or
     * more columns in the table. It would be wrong to assume that
     * this priviledge applies to all columns (this may be true for
     * some systems but is not true for all.)
     *
     * <P>Only privileges matching the schema and table name
     * criteria are returned.  They are ordered by TABLE_SCHEM,
     * TABLE_NAME, and PRIVILEGE.
     *
     * <P>Each privilige description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>GRANTOR</B> => grantor of access (may be null)
     *	<LI><B>GRANTEE</B> String => grantee of access
     *	<LI><B>PRIVILEGE</B> String => name of access (SELECT, 
     *      INSERT, UPDATE, REFRENCES, ...)
     *	<LI><B>IS_GRANTABLE</B> String => "YES" if grantee is permitted 
     *      to grant to others; "NO" if not; null if unknown 
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those
     * without a schema
     * @param tableNamePattern a table name pattern 
     * @return ResultSet - each row is a table privilege description
     * @exception SQLException if a database access error occurs
     * @see #getSearchStringEscape 
     */
    public ResultSet getTablePrivileges(String catalog, String schemaPattern,
					String tableNamePattern) throws SQLException
    {
	throw new ConsusNotImplemented("DatabaseResultSet.getTablePrivileges");
    }

    /**
     * Gets a description of a table's optimal set of columns that
     * uniquely identifies a row. They are ordered by SCOPE.
     *
     * <P>Each column description has the following columns:
     *  <OL>
     *	<LI><B>SCOPE</B> short => actual scope of result
     *      <UL>
     *      <LI> bestRowTemporary - very temporary, while using row
     *      <LI> bestRowTransaction - valid for remainder of current transaction
     *      <LI> bestRowSession - valid for remainder of current session
     *      </UL>
     *	<LI><B>COLUMN_NAME</B> String => column name
     *	<LI><B>DATA_TYPE</B> short => SQL data type from java.sql.Types
     *	<LI><B>TYPE_NAME</B> String => Data source dependent type name,
     *  for a UDT the type name is fully qualified
     *	<LI><B>COLUMN_SIZE</B> int => precision
     *	<LI><B>BUFFER_LENGTH</B> int => not used
     *	<LI><B>DECIMAL_DIGITS</B> short	 => scale
     *	<LI><B>PSEUDO_COLUMN</B> short => is this a pseudo column 
     *      like an Oracle ROWID
     *      <UL>
     *      <LI> bestRowUnknown - may or may not be pseudo column
     *      <LI> bestRowNotPseudo - is NOT a pseudo column
     *      <LI> bestRowPseudo - is a pseudo column
     *      </UL>
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those without a schema
     * @param table a table name
     * @param scope the scope of interest; use same values as SCOPE
     * @param nullable include columns that are nullable?
     * @return ResultSet - each row is a column description 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getBestRowIdentifier(String catalog, String schema,
					  String table, int scope, boolean nullable) throws SQLException
    {
	throw new ConsusNotImplemented("DatabaseResultSet.getBestRowIdentifier");
    }

    /**
     * Gets a description of a table's columns that are automatically
     * updated when any value in a row is updated.  They are
     * unordered.
     *
     * <P>Each column description has the following columns:
     *  <OL>
     *	<LI><B>SCOPE</B> short => is not used
     *	<LI><B>COLUMN_NAME</B> String => column name
     *	<LI><B>DATA_TYPE</B> short => SQL data type from java.sql.Types
     *	<LI><B>TYPE_NAME</B> String => Data source dependent type name
     *	<LI><B>COLUMN_SIZE</B> int => precision
     *	<LI><B>BUFFER_LENGTH</B> int => length of column value in bytes
     *	<LI><B>DECIMAL_DIGITS</B> short	 => scale
     *	<LI><B>PSEUDO_COLUMN</B> short => is this a pseudo column 
     *      like an Oracle ROWID
     *      <UL>
     *      <LI> versionColumnUnknown - may or may not be pseudo column
     *      <LI> versionColumnNotPseudo - is NOT a pseudo column
     *      <LI> versionColumnPseudo - is a pseudo column
     *      </UL>
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those without a schema
     * @param table a table name
     * @return ResultSet - each row is a column description 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getVersionColumns(String catalog, String schema,
				       String table) throws SQLException
    {
	throw new ConsusNotImplemented("ResultSet.getVersionColumns");
    }
	

    /**
     * Gets a description of a table's primary key columns.  They
     * are ordered by COLUMN_NAME.
     *
     * <P>Each primary key column description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>COLUMN_NAME</B> String => column name
     *	<LI><B>KEY_SEQ</B> short => sequence number within primary key
     *	<LI><B>PK_NAME</B> String => primary key name (may be null)
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those
     * without a schema
     * @param table a table name
     * @return ResultSet - each row is a primary key column description 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getPrimaryKeys(String catalog, String schema,
				    String tableName) throws SQLException
    {
	ConsusResultSet cursor = new ConsusResultSet(null);
	int i;
	for (TableDescriptor table = db.firstTable; table != null; table = table.next) { 
	    if (table.name.equals(tableName)) { 
		for (i = 0; i < table.columns.length; i++) {
		    FieldDescriptor fd = table.columns[i];
		    if ((fd.attr & FieldDescriptor.PrimaryKey) != 0) { 
			Object[] tuple = new Object[6];
			tuple[0] = "";
			tuple[1] = "";
			tuple[2] = table.name;
			tuple[3] = fd.name;
			tuple[4] = new Long(0);
			tuple[5] = fd.longName;
 			cursor.addTuple(tuple);
		    }
		}
		break;
	    }
	}
	ColumnNode columns[] = new ColumnNode[6];
	for (i = 0; i < 6; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}	
	columns[0].name = "TABLE_CAT";
	columns[1].name = "TABLE_SCHEM";
	columns[2].name = "TABLE_NAME";
	columns[3].name = "COLUMN_NAME";
	columns[4].name = "KEY_SEQ";
	columns[5].name = "PK_NAME";
	cursor.tupleColumns = columns;
	return cursor;
    }


    /**
     * Gets a description of the primary key columns that are
     * referenced by a table's foreign key columns (the primary keys
     * imported by a table).  They are ordered by PKTABLE_CAT,
     * PKTABLE_SCHEM, PKTABLE_NAME, and KEY_SEQ.
     *
     * <P>Each primary key column description has the following columns:
     *  <OL>
     *	<LI><B>PKTABLE_CAT</B> String => primary key table catalog 
     *      being imported (may be null)
     *	<LI><B>PKTABLE_SCHEM</B> String => primary key table schema
     *      being imported (may be null)
     *	<LI><B>PKTABLE_NAME</B> String => primary key table name
     *      being imported
     *	<LI><B>PKCOLUMN_NAME</B> String => primary key column name
     *      being imported
     *	<LI><B>FKTABLE_CAT</B> String => foreign key table catalog (may be null)
     *	<LI><B>FKTABLE_SCHEM</B> String => foreign key table schema (may be null)
     *	<LI><B>FKTABLE_NAME</B> String => foreign key table name
     *	<LI><B>FKCOLUMN_NAME</B> String => foreign key column name
     *	<LI><B>KEY_SEQ</B> short => sequence number within foreign key
     *	<LI><B>UPDATE_RULE</B> short => What happens to 
     *       foreign key when primary is updated:
     *      <UL>
     *      <LI> importedNoAction - do not allow update of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - change imported key to agree 
     *               with primary key update
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been updated
     *      <LI> importedKeySetDefault - change imported key to default values 
     *               if its primary key has been updated
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      </UL>
     *	<LI><B>DELETE_RULE</B> short => What happens to 
     *      the foreign key when primary is deleted.
     *      <UL>
     *      <LI> importedKeyNoAction - do not allow delete of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - delete rows that import a deleted key
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been deleted
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      <LI> importedKeySetDefault - change imported key to default if 
     *               its primary key has been deleted
     *      </UL>
     *	<LI><B>FK_NAME</B> String => foreign key name (may be null)
     *	<LI><B>PK_NAME</B> String => primary key name (may be null)
     *	<LI><B>DEFERRABILITY</B> short => can the evaluation of foreign key 
     *      constraints be deferred until commit
     *      <UL>
     *      <LI> importedKeyInitiallyDeferred - see SQL92 for definition
     *      <LI> importedKeyInitiallyImmediate - see SQL92 for definition 
     *      <LI> importedKeyNotDeferrable - see SQL92 for definition 
     *      </UL>
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those
     * without a schema
     * @param table a table name
     * @return ResultSet - each row is a primary key column description 
     * @exception SQLException if a database access error occurs
     * @see #getExportedKeys 
     */
    public ResultSet getImportedKeys(String catalog, String schema,
				     String tableName) throws SQLException
    {
	ConsusResultSet cursor = new ConsusResultSet(null);
	int i;
	for (TableDescriptor table = db.firstTable; table != null; table = table.next) { 
	    if (table.name.equals(tableName)) { 
		for (i = 0; i < table.columns.length; i++) {
		    FieldDescriptor fd = table.columns[i];
		    if ((fd.attr & FieldDescriptor.ForeignKey) != 0) { 
			Object[] tuple = new Object[14];
			tuple[0] = "";             // PKTABLE_CAT
			tuple[1] = "";             // PKTABLE_SCHEM
			tuple[2] = fd.fkTableName; // PKTABLE_NAME
			tuple[3] = fd.fkKeyName;   // PKCOLUMN_NAME
			tuple[4] = "";             // FKTABLE_CAT
			tuple[5] = "";             // FKTABLE_SCHEM
			tuple[6] = tableName;      // FKTABLE_NAME
			tuple[7] = fd.longName;    // FKCOLUMN_NAME
			tuple[8] = new Long(0);    // KEY_SEQ
			tuple[9] = new Long(DatabaseMetaData.importedKeyNoAction); // UPDATE_RULE
			tuple[10] = new Long(DatabaseMetaData.importedKeyNoAction); // DELETE_RULE
			tuple[11] = fd.longName;  // FK_NAME
			tuple[12] = fd.fkKeyName; // PK_NAME
			tuple[13] = new Long(DatabaseMetaData.importedKeyNotDeferrable); // DEFERRABILITY
 			cursor.addTuple(tuple);
		    }
		}
		break;
	    }
	}
	ColumnNode columns[] = new ColumnNode[14];
	for (i = 0; i < 14; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}	
	columns[0].name = "PKTABLE_CAT";
	columns[1].name = "PKTABLE_SCHEM";
	columns[2].name = "PKTABLE_NAME";
	columns[3].name = "PKCOLUMN_NAME";
	columns[4].name = "FKTABLE_CAT";
	columns[5].name = "FKTABLE_SCHEM";
	columns[6].name = "FKTABLE_NAME";
	columns[7].name = "FKCOLUMN_NAME";
	columns[8].name = "KEY_SEQ";
	columns[9].name = "UPDATE_RULE";
	columns[10].name = "DELETE_RULE";
	columns[11].name = "FK_NAME";
	columns[12].name = "PK_NAME";
	columns[13].name = "DEFERRABILITY";
	cursor.tupleColumns = columns;
	return cursor;
    }


    /**
     * Gets a description of the foreign key columns that reference a
     * table's primary key columns (the foreign keys exported by a
     * table).  They are ordered by FKTABLE_CAT, FKTABLE_SCHEM,
     * FKTABLE_NAME, and KEY_SEQ.
     *
     * <P>Each foreign key column description has the following columns:
     *  <OL>
     *	<LI><B>PKTABLE_CAT</B> String => primary key table catalog (may be null)
     *	<LI><B>PKTABLE_SCHEM</B> String => primary key table schema (may be null)
     *	<LI><B>PKTABLE_NAME</B> String => primary key table name
     *	<LI><B>PKCOLUMN_NAME</B> String => primary key column name
     *	<LI><B>FKTABLE_CAT</B> String => foreign key table catalog (may be null)
     *      being exported (may be null)
     *	<LI><B>FKTABLE_SCHEM</B> String => foreign key table schema (may be null)
     *      being exported (may be null)
     *	<LI><B>FKTABLE_NAME</B> String => foreign key table name
     *      being exported
     *	<LI><B>FKCOLUMN_NAME</B> String => foreign key column name
     *      being exported
     *	<LI><B>KEY_SEQ</B> short => sequence number within foreign key
     *	<LI><B>UPDATE_RULE</B> short => What happens to 
     *       foreign key when primary is updated:
     *      <UL>
     *      <LI> importedNoAction - do not allow update of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - change imported key to agree 
     *               with primary key update
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been updated
     *      <LI> importedKeySetDefault - change imported key to default values 
     *               if its primary key has been updated
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      </UL>
     *	<LI><B>DELETE_RULE</B> short => What happens to 
     *      the foreign key when primary is deleted.
     *      <UL>
     *      <LI> importedKeyNoAction - do not allow delete of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - delete rows that import a deleted key
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been deleted
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      <LI> importedKeySetDefault - change imported key to default if 
     *               its primary key has been deleted
     *      </UL>
     *	<LI><B>FK_NAME</B> String => foreign key name (may be null)
     *	<LI><B>PK_NAME</B> String => primary key name (may be null)
     *	<LI><B>DEFERRABILITY</B> short => can the evaluation of foreign key 
     *      constraints be deferred until commit
     *      <UL>
     *      <LI> importedKeyInitiallyDeferred - see SQL92 for definition
     *      <LI> importedKeyInitiallyImmediate - see SQL92 for definition 
     *      <LI> importedKeyNotDeferrable - see SQL92 for definition 
     *      </UL>
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those
     * without a schema
     * @param table a table name
     * @return ResultSet - each row is a foreign key column description 
     * @exception SQLException if a database access error occurs
     * @see #getImportedKeys 
     */
    public ResultSet getExportedKeys(String catalog, String schema,
				     String tableName) throws SQLException    
    {
	ConsusResultSet cursor = new ConsusResultSet(null);
	int i;
	for (TableDescriptor it = db.firstTable; it != null; it = it.next) { 
	    if (it.name.equals(tableName)) { 
		for (TableDescriptor et = db.firstTable; et != null; et = et.next) { 
		    if (it != et) { 
			for (i = 0; i < et.columns.length; i++) {
			    FieldDescriptor fd = et.columns[i];
			    if ((fd.attr & FieldDescriptor.ForeignKey) != 0 
				&& tableName.equals(fd.fkTableName)) 
			    { 
				Object[] tuple = new Object[14];
				tuple[0] = "";             // PKTABLE_CAT
				tuple[1] = "";             // PKTABLE_SCHEM
				tuple[2] = tableName;      // PKTABLE_NAME
				tuple[3] = fd.fkKeyName;   // PKCOLUMN_NAME
				tuple[4] = "";             // FKTABLE_CAT
				tuple[5] = "";             // FKTABLE_SCHEM
				tuple[6] = et.name;        // FKTABLE_NAME
				tuple[7] = fd.longName;    // FKCOLUMN_NAME
				tuple[8] = new Long(0);    // KEY_SEQ
				tuple[9] = new Long(DatabaseMetaData.importedKeyNoAction); // UPDATE_RULE
				tuple[10] = new Long(DatabaseMetaData.importedKeyNoAction); // DELETE_RULE
				tuple[11] = fd.longName;  // FK_NAME
				tuple[12] = fd.fkKeyName; // PK_NAME
				tuple[13] = new Long(DatabaseMetaData.importedKeyNotDeferrable); // DEFERRABILITY
				cursor.addTuple(tuple);
			    }
			}
		    }
		}
		break;
	    }
	}
	ColumnNode columns[] = new ColumnNode[14];
	for (i = 0; i < 14; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}	
	columns[0].name = "PKTABLE_CAT";
	columns[1].name = "PKTABLE_SCHEM";
	columns[2].name = "PKTABLE_NAME";
	columns[3].name = "PKCOLUMN_NAME";
	columns[4].name = "FKTABLE_CAT";
	columns[5].name = "FKTABLE_SCHEM";
	columns[6].name = "FKTABLE_NAME";
	columns[7].name = "FKCOLUMN_NAME";
	columns[8].name = "KEY_SEQ";
	columns[9].name = "UPDATE_RULE";
	columns[10].name = "DELETE_RULE";
	columns[11].name = "FK_NAME";
	columns[12].name = "PK_NAME";
	columns[13].name = "DEFERRABILITY";
	cursor.tupleColumns = columns;
	return cursor;
    }

    /**
     * Gets a description of the foreign key columns in the foreign key
     * table that reference the primary key columns of the primary key
     * table (describe how one table imports another's key.) This
     * should normally return a single foreign key/primary key pair
     * (most tables only import a foreign key from a table once.)  They
     * are ordered by FKTABLE_CAT, FKTABLE_SCHEM, FKTABLE_NAME, and
     * KEY_SEQ.
     *
     * <P>Each foreign key column description has the following columns:
     *  <OL>
     *	<LI><B>PKTABLE_CAT</B> String => primary key table catalog (may be null)
     *	<LI><B>PKTABLE_SCHEM</B> String => primary key table schema (may be null)
     *	<LI><B>PKTABLE_NAME</B> String => primary key table name
     *	<LI><B>PKCOLUMN_NAME</B> String => primary key column name
     *	<LI><B>FKTABLE_CAT</B> String => foreign key table catalog (may be null)
     *      being exported (may be null)
     *	<LI><B>FKTABLE_SCHEM</B> String => foreign key table schema (may be null)
     *      being exported (may be null)
     *	<LI><B>FKTABLE_NAME</B> String => foreign key table name
     *      being exported
     *	<LI><B>FKCOLUMN_NAME</B> String => foreign key column name
     *      being exported
     *	<LI><B>KEY_SEQ</B> short => sequence number within foreign key
     *	<LI><B>UPDATE_RULE</B> short => What happens to 
     *       foreign key when primary is updated:
     *      <UL>
     *      <LI> importedNoAction - do not allow update of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - change imported key to agree 
     *               with primary key update
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been updated
     *      <LI> importedKeySetDefault - change imported key to default values 
     *               if its primary key has been updated
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      </UL>
     *	<LI><B>DELETE_RULE</B> short => What happens to 
     *      the foreign key when primary is deleted.
     *      <UL>
     *      <LI> importedKeyNoAction - do not allow delete of primary 
     *               key if it has been imported
     *      <LI> importedKeyCascade - delete rows that import a deleted key
     *      <LI> importedKeySetNull - change imported key to NULL if 
     *               its primary key has been deleted
     *      <LI> importedKeyRestrict - same as importedKeyNoAction 
     *                                 (for ODBC 2.x compatibility)
     *      <LI> importedKeySetDefault - change imported key to default if 
     *               its primary key has been deleted
     *      </UL>
     *	<LI><B>FK_NAME</B> String => foreign key name (may be null)
     *	<LI><B>PK_NAME</B> String => primary key name (may be null)
     *	<LI><B>DEFERRABILITY</B> short => can the evaluation of foreign key 
     *      constraints be deferred until commit
     *      <UL>
     *      <LI> importedKeyInitiallyDeferred - see SQL92 for definition
     *      <LI> importedKeyInitiallyImmediate - see SQL92 for definition 
     *      <LI> importedKeyNotDeferrable - see SQL92 for definition 
     *      </UL>
     *  </OL>
     *
     * @param primaryCatalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param primarySchema a schema name; "" retrieves those
     * without a schema
     * @param primaryTable the table name that exports the key
     * @param foreignCatalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param foreignSchema a schema name; "" retrieves those
     * without a schema
     * @param foreignTable the table name that imports the key
     * @return ResultSet - each row is a foreign key column description 
     * @exception SQLException if a database access error occurs
     * @see #getImportedKeys 
     */
    public ResultSet getCrossReference(String primaryCatalog, String primarySchema, String primaryTable,
				       String foreignCatalog, String foreignSchema, String foreignTable) 
	throws SQLException
    {
	ConsusResultSet cursor = new ConsusResultSet(null);
	int i;
	for (TableDescriptor table = db.firstTable; table != null; table = table.next) { 
	    if (table.name.equals(foreignTable)) { 
		for (i = 0; i < table.columns.length; i++) {
		    FieldDescriptor fd = table.columns[i];
		    if ((fd.attr & FieldDescriptor.ForeignKey) != 0
			&& fd.fkTableName.equals(primaryTable))
		    { 
			Object[] tuple = new Object[14];
			tuple[0] = "";             // PKTABLE_CAT
			tuple[1] = "";             // PKTABLE_SCHEM
			tuple[2] = primaryTable;   // PKTABLE_NAME
			tuple[3] = fd.fkKeyName;   // PKCOLUMN_NAME
			tuple[4] = "";             // FKTABLE_CAT
			tuple[5] = "";             // FKTABLE_SCHEM
			tuple[6] = foreignTable;   // FKTABLE_NAME
			tuple[7] = fd.longName;    // FKCOLUMN_NAME
			tuple[8] = new Long(0);    // KEY_SEQ
			tuple[9] = new Long(DatabaseMetaData.importedKeyNoAction); // UPDATE_RULE
			tuple[10] = new Long(DatabaseMetaData.importedKeyNoAction); // DELETE_RULE
			tuple[11] = fd.longName;  // FK_NAME
			tuple[12] = fd.fkKeyName; // PK_NAME
			tuple[13] = new Long(DatabaseMetaData.importedKeyNotDeferrable); // DEFERRABILITY
 			cursor.addTuple(tuple);
		    }
		}
		break;
	    }
	}
	ColumnNode columns[] = new ColumnNode[14];
	for (i = 0; i < 14; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}	
	columns[0].name = "PKTABLE_CAT";
	columns[1].name = "PKTABLE_SCHEM";
	columns[2].name = "PKTABLE_NAME";
	columns[3].name = "PKCOLUMN_NAME";
	columns[4].name = "FKTABLE_CAT";
	columns[5].name = "FKTABLE_SCHEM";
	columns[6].name = "FKTABLE_NAME";
	columns[7].name = "FKCOLUMN_NAME";
	columns[8].name = "KEY_SEQ";
	columns[9].name = "UPDATE_RULE";
	columns[10].name = "DELETE_RULE";
	columns[11].name = "FK_NAME";
	columns[12].name = "PK_NAME";
	columns[13].name = "DEFERRABILITY";
	cursor.tupleColumns = columns;
	return cursor;
    }

    /**
     * Gets a description of all the standard SQL types supported by
     * this database. They are ordered by DATA_TYPE and then by how
     * closely the data type maps to the corresponding JDBC SQL type.
     *
     * <P>Each type description has the following columns:
     *  <OL>
     *	<LI><B>TYPE_NAME</B> String => Type name
     *	<LI><B>DATA_TYPE</B> short => SQL data type from java.sql.Types
     *	<LI><B>PRECISION</B> int => maximum precision
     *	<LI><B>LITERAL_PREFIX</B> String => prefix used to quote a literal 
     *      (may be null)
     *	<LI><B>LITERAL_SUFFIX</B> String => suffix used to quote a literal 
            (may be null)
     *	<LI><B>CREATE_PARAMS</B> String => parameters used in creating 
     *      the type (may be null)
     *	<LI><B>NULLABLE</B> short => can you use NULL for this type?
     *      <UL>
     *      <LI> typeNoNulls - does not allow NULL values
     *      <LI> typeNullable - allows NULL values
     *      <LI> typeNullableUnknown - nullability unknown
     *      </UL>
     *	<LI><B>CASE_SENSITIVE</B> boolean=> is it case sensitive?
     *	<LI><B>SEARCHABLE</B> short => can you use "WHERE" based on this type:
     *      <UL>
     *      <LI> typePredNone - No support
     *      <LI> typePredChar - Only supported with WHERE .. LIKE
     *      <LI> typePredBasic - Supported except for WHERE .. LIKE
     *      <LI> typeSearchable - Supported for all WHERE ..
     *      </UL>
     *	<LI><B>UNSIGNED_ATTRIBUTE</B> boolean => is it unsigned?
     *	<LI><B>FIXED_PREC_SCALE</B> boolean => can it be a money value?
     *	<LI><B>AUTO_INCREMENT</B> boolean => can it be used for an 
     *      auto-increment value?
     *	<LI><B>LOCAL_TYPE_NAME</B> String => localized version of type name 
     *      (may be null)
     *	<LI><B>MINIMUM_SCALE</B> short => minimum scale supported
     *	<LI><B>MAXIMUM_SCALE</B> short => maximum scale supported
     *	<LI><B>SQL_DATA_TYPE</B> int => unused
     *	<LI><B>SQL_DATETIME_SUB</B> int => unused
     *	<LI><B>NUM_PREC_RADIX</B> int => usually 2 or 10
     *  </OL>
     *
     * @return ResultSet - each row is a SQL type description 
     * @exception SQLException if a database access error occurs
     */
    static Object[] cloneTuple(Object[] tuple) { 
	Object[] newTuple = new Object[tuple.length];
	System.arraycopy(tuple, 0, newTuple, 0, tuple.length);
	return newTuple;
    }

    public ResultSet getTypeInfo() throws SQLException
    {
	ConsusResultSet cursor = new ConsusResultSet(null);
	Object[] tuple = new Object[18];
	tuple[0] = "BIT";
	tuple[1] = new Long(Types.BIT);
	tuple[2] = new Long(0);
	tuple[3] = "";
	tuple[4] = "";
	tuple[5] = "";
	tuple[6] = new Long(typeNoNulls);
	tuple[7] = new Boolean(true); // case sensetive
	tuple[8] = new Long(typeSearchable);
	tuple[9] = new Boolean(false); // unsigned 
	tuple[10] = new Boolean(false); // fixed precision
	tuple[11] = new Boolean(false); // auto increment
	tuple[12] = "";
	tuple[13] = new Long(0);
	tuple[14] = new Long(0);
	tuple[15] = new Long(0);
	tuple[16] = new Long(0);
	tuple[17] = new Long(10);  // radix
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "BOOLEAN";
	tuple[1] = new Long(Types.BIT);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "TINYINT";
	tuple[1] = new Long(Types.TINYINT);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "SMALLINT";
	tuple[1] = new Long(Types.SMALLINT);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "INTEGER";
	tuple[1] = new Long(Types.INTEGER);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "BIGINT";
	tuple[1] = new Long(Types.BIGINT);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "FLOAT";
	tuple[1] = new Long(Types.FLOAT);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "REAL";
	tuple[1] = new Long(Types.DOUBLE);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "DOUBLE";
	tuple[1] = new Long(Types.DOUBLE);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "DATE";
	tuple[1] = new Long(Types.DATE);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "TIME";
	tuple[1] = new Long(Types.TIME);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "TIMESTAMP";
	tuple[1] = new Long(Types.TIMESTAMP);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "STRING";
	tuple[1] = new Long(Types.VARCHAR);
	tuple[2] = new Long(Integer.MAX_VALUE);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "LONGVARCHAR";
	tuple[1] = new Long(Types.LONGVARCHAR);
	tuple[2] = new Long(Integer.MAX_VALUE);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "VARCHAR";
	tuple[1] = new Long(Types.VARCHAR);
	tuple[2] = new Long(Integer.MAX_VALUE);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "CHAR";
	tuple[1] = new Long(Types.VARCHAR);
	tuple[2] = new Long(Integer.MAX_VALUE);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "REFERENCE";
	tuple[1] = new Long(Types.REF);
	tuple[6] = new Long(typeNullable);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "BINARY";
	tuple[1] = new Long(Types.VARBINARY);
	tuple[2] = new Long(Integer.MAX_VALUE);
	tuple[6] = new Long(typeNoNulls);
	tuple[8] = new Long(typePredNone);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "VARBINARY";
	tuple[1] = new Long(Types.VARBINARY);
	tuple[2] = new Long(Integer.MAX_VALUE);
	tuple[6] = new Long(typeNoNulls);
	tuple[8] = new Long(typePredNone);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "LONGVARBINARY";
	tuple[1] = new Long(Types.LONGVARBINARY);
	tuple[2] = new Long(Integer.MAX_VALUE);
	tuple[6] = new Long(typeNoNulls);
	tuple[8] = new Long(typePredNone);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "BLOB";
	tuple[1] = new Long(Types.BLOB);
	cursor.addTuple(cloneTuple(tuple));

	tuple[0] = "CLOB";
	tuple[1] = new Long(Types.CLOB);
	cursor.addTuple(cloneTuple(tuple));

	ColumnNode columns[] = new ColumnNode[18];
	for (int i = 0; i < 18; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}	
	columns[0].name = "TYPE_NAME";
	columns[1].name = "DATA_TYPE";
	columns[1].type = ExprNode.tpInt;
	columns[2].name = "PRECISION";
	columns[2].type = ExprNode.tpInt;
	columns[3].name = "LITERAL_PREFIX";
	columns[4].name = "LITERAL_SUFFIX";
	columns[5].name = "CREATE_PARAMS";
	columns[6].name = "NULLABLE";
	columns[6].type = ExprNode.tpInt;
	columns[7].name = "CASE_SENSITIVE";
	columns[7].type = ExprNode.tpBool;
	columns[8].name = "SEARCHABLE";
	columns[8].type = ExprNode.tpInt;
	columns[9].name = "UNSIGNED_ATTRIBUTE";
	columns[9].type = ExprNode.tpBool;
	columns[10].name = "FIXED_PREC_SCALE";
	columns[10].type = ExprNode.tpBool;
	columns[11].name = "AUTO_INCREMENT";
	columns[11].type = ExprNode.tpBool;
	columns[12].name = "LOCAL_TYPE_NAME";
	columns[13].name = "MINIMUM_SCALE";
	columns[13].type = ExprNode.tpInt;
	columns[14].name = "MAXIMUM_SCALE";
	columns[14].type = ExprNode.tpInt;
	columns[15].name = "SQL_DATA_TYPE";
	columns[15].type = ExprNode.tpInt;
	columns[16].name = "SQL_DATETIME_SUB";
	columns[16].type = ExprNode.tpInt;
	columns[17].name = "NUM_PREC_RADIX";
	columns[17].type = ExprNode.tpInt;
	cursor.tupleColumns = columns;
	return cursor;
    }

    /**
     * Gets a description of a table's indices and statistics. They are
     * ordered by NON_UNIQUE, TYPE, INDEX_NAME, and ORDINAL_POSITION.
     *
     * <P>Each index column description has the following columns:
     *  <OL>
     *	<LI><B>TABLE_CAT</B> String => table catalog (may be null)
     *	<LI><B>TABLE_SCHEM</B> String => table schema (may be null)
     *	<LI><B>TABLE_NAME</B> String => table name
     *	<LI><B>NON_UNIQUE</B> boolean => Can index values be non-unique? 
     *      false when TYPE is tableIndexStatistic
     *	<LI><B>INDEX_QUALIFIER</B> String => index catalog (may be null); 
     *      null when TYPE is tableIndexStatistic
     *	<LI><B>INDEX_NAME</B> String => index name; null when TYPE is 
     *      tableIndexStatistic
     *	<LI><B>TYPE</B> short => index type:
     *      <UL>
     *      <LI> tableIndexStatistic - this identifies table statistics that are
     *           returned in conjuction with a table's index descriptions
     *      <LI> tableIndexClustered - this is a clustered index
     *      <LI> tableIndexHashed - this is a hashed index
     *      <LI> tableIndexOther - this is some other style of index
     *      </UL>
     *	<LI><B>ORDINAL_POSITION</B> short => column sequence number 
     *      within index; zero when TYPE is tableIndexStatistic
     *	<LI><B>COLUMN_NAME</B> String => column name; null when TYPE is 
     *      tableIndexStatistic
     *	<LI><B>ASC_OR_DESC</B> String => column sort sequence, "A" => ascending, 
     *      "D" => descending, may be null if sort sequence is not supported; 
     *      null when TYPE is tableIndexStatistic	
     *	<LI><B>CARDINALITY</B> int => When TYPE is tableIndexStatistic, then 
     *      this is the number of rows in the table; otherwise, it is the 
     *      number of unique values in the index.
     *	<LI><B>PAGES</B> int => When TYPE is  tableIndexStatisic then 
     *      this is the number of pages used for the table, otherwise it 
     *      is the number of pages used for the current index.
     *	<LI><B>FILTER_CONDITION</B> String => Filter condition, if any.  
     *      (may be null)
     *  </OL>
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schema a schema name; "" retrieves those without a schema
     * @param table a table name  
     * @param unique when true, return only indices for unique values; 
     *     when false, return indices regardless of whether unique or not 
     * @param approximate when true, result is allowed to reflect approximate 
     *     or out of data values; when false, results are requested to be 
     *     accurate
     * @return ResultSet - each row is an index column description 
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getIndexInfo(String catalog, String schema, String tableName,
				  boolean unique, boolean approximate)
	throws SQLException
    {
	ConsusResultSet cursor = new ConsusResultSet(null);
	for (TableDescriptor table = db.firstTable; table != null; table = table.next) { 
	    if (table.name.equals(tableName)) { 
		for (FieldDescriptor fd = table.fields; fd != null; fd = fd.next) { 
		    if (fd.index != 0 
			&& (!unique || (fd.attr & (FieldDescriptor.PrimaryKey|FieldDescriptor.Unique)) != 0))
		    { 
			Object[] tuple = new Object[13];
			tuple[0] = ""; // TABLE_CAT
			tuple[1] = ""; // TABLE_SCHEM
			tuple[2] = table.name; // TABLE_NAME
			tuple[3] = new Boolean(true); // NON_UNIQUE
			tuple[4] = ""; // INDEX_QUALIFIER
			tuple[5] = fd.name; // INDEX_NAME
			tuple[6] = new Long(tableIndexClustered); // TYPE
			tuple[7] = new Long(1); // ORDINAL_POSITION
			tuple[8] = fd.name; // COLUMN_NAME
			tuple[9] = "A"; // ASC_OR_DESC
			tuple[10] = new Long(0); // CARDINALITY
			tuple[11] = new Long(0); // PAGES
			tuple[12] = ""; // FILTER_CONDITION
			cursor.addTuple(tuple);
		    }
		}
		break;
	    }
	}
	ColumnNode columns[] = new ColumnNode[13];
	for (int i = 0; i < 13; i++) { 
	    columns[i] = new ColumnNode(i+1);
	}
	columns[0].name = "TABLE_CAT";
	columns[1].name = "TABLE_SCHEM";
	columns[2].name = "TABLE_NAME";
	columns[3].name = "NON_UNIQUE";
	columns[3].type = ExprNode.tpBool;
	columns[4].name = "INDEX_QUALIFIER";
	columns[5].name = "INDEX_NAME";
	columns[6].name = "TYPE";
	columns[6].type = ExprNode.tpInt;
	columns[7].name = "ORDINAL_POSITION";
	columns[7].type = ExprNode.tpInt;
	columns[8].name = "COLUMN_NAME";
	columns[9].name = "ASC_OR_DESC";
	columns[10].name = "CARDINALITY";
	columns[10].type = ExprNode.tpInt;
	columns[11].name = "PAGES";
	columns[11].type = ExprNode.tpInt;
	columns[12].name = "FILTER_CONDITION";
	cursor.tupleColumns = columns;
	return cursor;
    }

    //--------------------------JDBC 2.0-----------------------------

    /**
     * JDBC 2.0
     *
     * Does the database support the given result set type?
     *
     * @param type defined in <code>java.sql.ResultSet</code>
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     * @see Connection
     */
    public boolean supportsResultSetType(int type) throws SQLException { 
	return type != ResultSet.TYPE_SCROLL_SENSITIVE;
    }

    /**
     * JDBC 2.0
     *
     * Does the database support the concurrency type in combination
     * with the given result set type?
     *
     * @param type defined in <code>java.sql.ResultSet</code>
     * @param concurrency type defined in <code>java.sql.ResultSet</code>
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     * @see Connection
     */
    public boolean supportsResultSetConcurrency(int type, int concurrency) throws SQLException
    {
	return type != ResultSet.TYPE_SCROLL_SENSITIVE;
    }

    /**
     * JDBC 2.0
     *
     * Indicates whether a result set's own updates are visible.
     *
     * @param result set type, i.e. ResultSet.TYPE_XXX
     * @return <code>true</code> if updates are visible for the result set type;
	 *        <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean ownUpdatesAreVisible(int type) throws SQLException { 
	return true;
    }

    /**
     * JDBC 2.0
     *
     * Indicates whether a result set's own deletes are visible.
     *
     * @param result set type, i.e. ResultSet.TYPE_XXX
     * @return <code>true</code> if deletes are visible for the result set type;
	 *        <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean ownDeletesAreVisible(int type) throws SQLException { 
	return true;
    }
    /**
     * JDBC 2.0
     *
     * Indicates whether a result set's own inserts are visible.
     *
     * @param result set type, i.e. ResultSet.TYPE_XXX
     * @return <code>true</code> if inserts are visible for the result set type;
	 *        <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean ownInsertsAreVisible(int type) throws SQLException { 
	return false;
    }

    /**
     * JDBC 2.0
     *
     * Indicates whether updates made by others are visible.
     *
     * @param result set type, i.e. ResultSet.TYPE_XXX
     * @return <code>true</code> if updates made by others
	 * are visible for the result set type;
	 *        <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean othersUpdatesAreVisible(int type) throws SQLException { 
	return false;
    }

    /**
     * JDBC 2.0
     *
     * Indicates whether deletes made by others are visible.
     *
     * @param result set type, i.e. ResultSet.TYPE_XXX
     * @return <code>true</code> if deletes made by others
	 * are visible for the result set type;
	 *        <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean othersDeletesAreVisible(int type) throws SQLException { 
	return false;
    }
    /**
     * JDBC 2.0
     *
     * Indicates whether inserts made by others are visible.
     *
     * @param result set type, i.e. ResultSet.TYPE_XXX
     * @return true if updates are visible for the result set type
     * @return <code>true</code> if inserts made by others
     * are visible for the result set type;
     *        <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean othersInsertsAreVisible(int type) throws SQLException { 
	return false;
    }

    /**
     * JDBC 2.0
     *
     * Indicates whether or not a visible row update can be detected by 
     * calling the method <code>ResultSet.rowUpdated</code>.
     *
     * @param result set type, i.e. ResultSet.TYPE_XXX
     * @return <code>true</code> if changes are detected by the result set type;
     *         <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean updatesAreDetected(int type) throws SQLException { 
	return false;
    }

    /**
     * JDBC 2.0
     *
     * Indicates whether or not a visible row delete can be detected by 
     * calling ResultSet.rowDeleted().  If deletesAreDetected()
     * returns false, then deleted rows are removed from the result set.
     *
     * @param result set type, i.e. ResultSet.TYPE_XXX
     * @return true if changes are detected by the resultset type
     * @exception SQLException if a database access error occurs
     */
    public boolean deletesAreDetected(int type) throws SQLException { 
	return false;
    }

    /**
     * JDBC 2.0
     *
     * Indicates whether or not a visible row insert can be detected
     * by calling ResultSet.rowInserted().
     *
     * @param result set type, i.e. ResultSet.TYPE_XXX
     * @return true if changes are detected by the resultset type
     * @exception SQLException if a database access error occurs
     */
    public boolean insertsAreDetected(int type) throws SQLException { 
	return false;
    }

    /**
     * JDBC 2.0
     *
     * Indicates whether the driver supports batch updates.
     * @return true if the driver supports batch updates; false otherwise
     */
    public boolean supportsBatchUpdates() throws SQLException { 
	return false;
    }

    /**
     * JDBC 2.0
     *
     * Gets a description of the user-defined types defined in a particular
     * schema.  Schema-specific UDTs may have type JAVA_OBJECT, STRUCT, 
     * or DISTINCT.
     *
     * <P>Only types matching the catalog, schema, type name and type  
     * criteria are returned.  They are ordered by DATA_TYPE, TYPE_SCHEM 
     * and TYPE_NAME.  The type name parameter may be a fully-qualified 
     * name.  In this case, the catalog and schemaPattern parameters are
     * ignored.
     *
     * <P>Each type description has the following columns:
     *  <OL>
     *	<LI><B>TYPE_CAT</B> String => the type's catalog (may be null)
     *	<LI><B>TYPE_SCHEM</B> String => type's schema (may be null)
     *	<LI><B>TYPE_NAME</B> String => type name
     *  <LI><B>CLASS_NAME</B> String => Java class name
     *	<LI><B>DATA_TYPE</B> String => type value defined in java.sql.Types.  
     *  One of JAVA_OBJECT, STRUCT, or DISTINCT
     *	<LI><B>REMARKS</B> String => explanatory comment on the type
     *  </OL>
     *
     * <P><B>Note:</B> If the driver does not support UDTs, an empty
     * result set is returned.
     *
     * @param catalog a catalog name; "" retrieves those without a
     * catalog; null means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those
     * without a schema
     * @param typeNamePattern a type name pattern; may be a fully-qualified
     * name
     * @param types a list of user-named types to include (JAVA_OBJECT, 
     * STRUCT, or DISTINCT); null returns all types 
     * @return ResultSet - each row is a type description
     * @exception SQLException if a database access error occurs
     */
    public ResultSet getUDTs(String catalog, String schemaPattern, 
			     String typeNamePattern, int[] types) 
      throws SQLException
    {
	throw new ConsusNotImplemented("DatabaseMetaData.getUDTs");
    }

    /**
     * JDBC 2.0
	 * Retrieves the connection that produced this metadata object.
     *
     * @return the connection that produced this metadata object
     */
    public Connection getConnection() throws SQLException { 
	return db;
    }


    // ------------------- JDBC 3.0 -------------------------

    /**
     * Retrieves whether this database supports savepoints.
     *
     * @return <code>true</code> if savepoints are supported; 
     *         <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public boolean supportsSavepoints() throws SQLException
    {
	return false;
    }

    /**
     * Retrieves whether this database supports named parameters to callable 
     * statements.
     *
     * @return <code>true</code> if named parameters are supported; 
     *         <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public boolean supportsNamedParameters() throws SQLException
    {
	return true;
    }

    /**
     * Retrieves whether it is possible to have multiple <code>ResultSet</code> objects
     * returned from a <code>CallableStatement</code> object
     * simultaneously.
     *
     * @return <code>true</code> if a <code>CallableStatement</code> object
     *         can return multiple <code>ResultSet</code> objects
     *         simultaneously; <code>false</code> otherwise
     * @exception SQLException if a datanase access error occurs
     * @since 1.4
     */
    public boolean supportsMultipleOpenResults() throws SQLException
    {
	return false;
    }

    /**
     * Retrieves whether auto-generated keys can be retrieved after 
     * a statement has been executed.
     *
     * @return <code>true</code> if auto-generated keys can be retrieved
     *         after a statement has executed; <code>false</code> otherwise
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public boolean supportsGetGeneratedKeys() throws SQLException
    {
	return false;
    }

    /**
     * Retrieves a description of the user-defined type (UDT) hierarchies defined in a 
     * particular schema in this database. Only the immediate super type/ 
     * sub type relationship is modeled.
     * <P>
     * Only supertype information for UDTs matching the catalog, 
     * schema, and type name is returned. The type name parameter
     * may be a fully-qualified name. When the UDT name supplied is a 
     * fully-qualified name, the catalog and schemaPattern parameters are 
     * ignored. 
     * <P>
     * If a UDT does not have a direct super type, it is not listed here.
     * A row of the <code>ResultSet</code> object returned by this method
     * describes the designated UDT and a direct supertype. A row has the following 
     * columns:
     *  <OL>
     *  <LI><B>TYPE_CAT</B> String => the UDT's catalog (may be <code>null</code>)
     *  <LI><B>TYPE_SCHEM</B> String => UDT's schema (may be <code>null</code>)
     *  <LI><B>TYPE_NAME</B> String => type name of the UDT
     *  <LI><B>SUPERTYPE_CAT</B> String => the direct super type's catalog 
     *                           (may be <code>null</code>)
     *  <LI><B>SUPERTYPE_SCHEM</B> String => the direct super type's schema 
     *                             (may be <code>null</code>)
     *  <LI><B>SUPERTYPE_NAME</B> String => the direct super type's name
     *  </OL>
     *
     * <P><B>Note:</B> If the driver does not support type hierarchies, an 
     * empty result set is returned.
     *
     * @param catalog a catalog name; "" retrieves those without a catalog;
     *        <code>null</code> means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those 
     *        without a schema
     * @param typeNamePattern a UDT name pattern; may be a fully-qualified
     *        name
     * @return a <code>ResultSet</code> object in which a row gives information
     *         about the designated UDT
     * @throws SQLException if a database access error occurs
     * @since 1.4
     */
    public ResultSet getSuperTypes(String catalog, String schemaPattern, 
				   String typeNamePattern) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusDatabaseMetaData.getSuperTypes");
    }
    
    /**
     * Retrieves a description of the table hierarchies defined in a particular 
     * schema in this database.
     *
     * <P>Only supertable information for tables matching the catalog, schema
     * and table name are returned. The table name parameter may be a fully-
     * qualified name, in which case, the catalog and schemaPattern parameters
     * are ignored. If a table does not have a super table, it is not listed here.
     * Supertables have to be defined in the same catalog and schema as the 
     * sub tables. Therefore, the type description does not need to include
     * this information for the supertable.
     *
     * <P>Each type description has the following columns:
     *  <OL>
     *  <LI><B>TABLE_CAT</B> String => the type's catalog (may be <code>null</code>)
     *  <LI><B>TABLE_SCHEM</B> String => type's schema (may be <code>null</code>)
     *  <LI><B>TABLE_NAME</B> String => type name
     *  <LI><B>SUPERTABLE_NAME</B> String => the direct super type's name
     *  </OL>
     *
     * <P><B>Note:</B> If the driver does not support type hierarchies, an 
     * empty result set is returned.
     *
     * @param catalog a catalog name; "" retrieves those without a catalog;
     *        <code>null</code> means drop catalog name from the selection criteria
     * @param schemaPattern a schema name pattern; "" retrieves those 
     *        without a schema
     * @param tableNamePattern a table name pattern; may be a fully-qualified
     *        name
     * @return a <code>ResultSet</code> object in which each row is a type description
     * @throws SQLException if a database access error occurs
     * @since 1.4
     */
    public ResultSet getSuperTables(String catalog, String schemaPattern,
				    String tableNamePattern) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusDatabaseMetaData.getSuperTables");
    }


    /**
     * Retrieves a description of the given attribute of the given type 
     * for a user-defined type (UDT) that is available in the given schema 
     * and catalog.
     * <P>
     * Descriptions are returned only for attributes of UDTs matching the 
     * catalog, schema, type, and attribute name criteria. They are ordered by
     * TYPE_SCHEM, TYPE_NAME and ORDINAL_POSITION. This description
     * does not contain inherited attributes.
     * <P>
     * The <code>ResultSet</code> object that is returned has the following 
     * columns:
     * <OL>
     *  <LI><B>TYPE_CAT</B> String => type catalog (may be <code>null</code>)
     *	<LI><B>TYPE_SCHEM</B> String => type schema (may be <code>null</code>)
     *	<LI><B>TYPE_NAME</B> String => type name
     *	<LI><B>ATTR_NAME</B> String => attribute name
     *	<LI><B>DATA_TYPE</B> short => attribute type SQL type from java.sql.Types
     *	<LI><B>ATTR_TYPE_NAME</B> String => Data source dependent type name.
     *  For a UDT, the type name is fully qualified. For a REF, the type name is 
     *  fully qualified and represents the target type of the reference type.
     *	<LI><B>ATTR_SIZE</B> int => column size.  For char or date
     *	    types this is the maximum number of characters; for numeric or
     *	    decimal types this is precision.
     *	<LI><B>DECIMAL_DIGITS</B> int => the number of fractional digits
     *	<LI><B>NUM_PREC_RADIX</B> int => Radix (typically either 10 or 2)
     *	<LI><B>NULLABLE</B> int => whether NULL is allowed
     *      <UL>
     *      <LI> attributeNoNulls - might not allow NULL values
     *      <LI> attributeNullable - definitely allows NULL values
     *      <LI> attributeNullableUnknown - nullability unknown
     *      </UL>
     *	<LI><B>REMARKS</B> String => comment describing column (may be <code>null</code>)
     * 	<LI><B>ATTR_DEF</B> String => default value (may be <code>null</code>)
     *	<LI><B>SQL_DATA_TYPE</B> int => unused
     *	<LI><B>SQL_DATETIME_SUB</B> int => unused
     *	<LI><B>CHAR_OCTET_LENGTH</B> int => for char types the 
     *       maximum number of bytes in the column
     *	<LI><B>ORDINAL_POSITION</B> int	=> index of column in table 
     *      (starting at 1)
     *	<LI><B>IS_NULLABLE</B> String => "NO" means column definitely 
     *      does not allow NULL values; "YES" means the column might 
     *      allow NULL values.  An empty string means unknown.
     *  <LI><B>SCOPE_CATALOG</B> String => catalog of table that is the
     *      scope of a reference attribute (<code>null</code> if DATA_TYPE isn't REF)
     *  <LI><B>SCOPE_SCHEMA</B> String => schema of table that is the 
     *      scope of a reference attribute (<code>null</code> if DATA_TYPE isn't REF)
     *  <LI><B>SCOPE_TABLE</B> String => table name that is the scope of a 
     *      reference attribute (<code>null</code> if the DATA_TYPE isn't REF)
     * <LI><B>SOURCE_DATA_TYPE</B> short => source type of a distinct type or user-generated
     *      Ref type,SQL type from java.sql.Types (<code>null</code> if DATA_TYPE 
     *      isn't DISTINCT or user-generated REF)
     *  </OL>
     * @param catalog a catalog name; must match the catalog name as it
     *        is stored in the database; "" retrieves those without a catalog;
     *        <code>null</code> means that the catalog name should not be used to narrow
     *        the search
     * @param schemaPattern a schema name pattern; must match the schema name
     *        as it is stored in the database; "" retrieves those without a schema;
     *        <code>null</code> means that the schema name should not be used to narrow
     *        the search
     * @param typeNamePattern a type name pattern; must match the
     *        type name as it is stored in the database 
     * @param attributeNamePattern an attribute name pattern; must match the attribute
     *        name as it is declared in the database
     * @return a <code>ResultSet</code> object in which each row is an 
     *         attribute description
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public ResultSet getAttributes(String catalog, String schemaPattern,
				   String typeNamePattern, String attributeNamePattern) 
	throws SQLException
    {
	throw new ConsusNotImplemented("ConsusDatabaseMetaData.getAttributes");
    }

    /**
     * Retrieves whether this database supports the given result set holdability.
     *
     * @param holdability one of the following constants:
     *          <code>ResultSet.HOLD_CURSORS_OVER_COMMIT</code> or
     *          <code>ResultSet.CLOSE_CURSORS_AT_COMMIT<code>
     * @return <code>true</code> if so; <code>false</code> otherwise 
     * @exception SQLException if a database access error occurs
     * @see Connection
     * @since 1.4
     */
    public boolean supportsResultSetHoldability(int holdability) throws SQLException
    {
	return holdability == 2; // ResultSet.CLOSE_CURSORS_AT_COMMIT
    }

    /**
     * Retrieves the default holdability of this <code>ResultSet</code>
     * object.
     *
     * @return the default holdability; either 
     *         <code>ResultSet.HOLD_CURSORS_OVER_COMMIT</code> or
     *         <code>ResultSet.CLOSE_CURSORS_AT_COMMIT</code>
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public int getResultSetHoldability() throws SQLException {
	return 2; // ResultSet.CLOSE_CURSORS_AT_COMMIT
    }

    /**
     * Retrieves the major version number of the underlying database.
     *
     * @return the underlying database's major version
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public int getDatabaseMajorVersion() throws SQLException
    {
	return 1;
    }

    /**
     * Retrieves the minor version number of the underlying database.
     *
     * @return underlying database's minor version
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public int getDatabaseMinorVersion() throws SQLException
    {
	return 2;
    }

    /**
     * Retrieves the major JDBC version number for this
     * driver.
     * 
     * @return JDBC version major number
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public int getJDBCMajorVersion() throws SQLException
    {
	return 2;
    }

    /**
     * Retrieves the minor JDBC version number for this
     * driver.
     * 
     * @return JDBC version minor number
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public int getJDBCMinorVersion() throws SQLException
    {
	return 0;
    }


    /**
     * Indicates whether the SQLSTATEs returned by <code>SQLException.getSQLState</code>
     * is X/Open (now known as Open Group) SQL CLI or SQL99.
     * @return the type of SQLSTATEs, one of:
     *        sqlStateXOpen or
     *        sqlStateSQL99
     * @throws SQLException if a database access error occurs
     * @since 1.4
     */
    public int getSQLStateType() throws SQLException
    {
	return 1; // sqlStateXOpen
    }


    /**
     * Indicates whether updates made to a LOB are made on a copy or directly 
     * to the LOB.
     * @return <code>true</code> if updates are made to a copy of the LOB;
     *         <code>false</code> if updates are made directly to the LOB
     * @throws SQLException if a database access error occurs
     * @since 1.4
     */
    public boolean locatorsUpdateCopy()
    { 
	return false;
    }

    /**
     * Retrieves weather this database supports statement pooling.
     *
     * @return <code>true</code> is so;
	       <code>false</code> otherwise
     * @throws SQLExcpetion if a database access error occurs
     * @since 1.4
     */
    public boolean supportsStatementPooling() 
    { 
	return true;
    }

    public ResultSet getFunctionColumns(String catalog,
                                        String schemaPattern,
                                        String functionNamePattern, 
                                        String columnNamePattern) throws SQLException
    {
        throw new ConsusNotImplemented("DatabaseMetaData.getFunctionColumns");
    }


    public ResultSet getFunctions(String catalog, String schemaPattern,
			    String functionNamePattern) throws SQLException
    {
        throw new ConsusNotImplemented("DatabaseMetaData.getFunctions");
    }

    public ResultSet getClientInfoProperties() throws SQLException
    {
        throw new ConsusNotImplemented("DatabaseMetaData.getClientInfoProperties");
    }

    public boolean autoCommitFailureClosesAllResultSets() throws SQLException
    {
        return false;
    }
	
    public boolean supportsStoredFunctionsUsingCallSyntax() throws SQLException
    {
        return false;
    }
     
    public ResultSet getSchemas(String catalog, String schemaPattern) throws SQLException
    {
        throw new ConsusNotImplemented("DatabaseMetaData.getSchemas");
    }

    public RowIdLifetime getRowIdLifetime() throws SQLException
    {
        throw new ConsusNotImplemented("DatabaseMetaData.getRowIdLifetime");
    }

    public boolean isWrapperFor(Class iface) throws java.sql.SQLException
    {
        return false;
    }
    
    public Object unwrap(java.lang.Class iface) throws java.sql.SQLException
    {
	return null;
    }

    ConsusDatabaseMetaData(ConsusConnection conxn) { 
	db = conxn;
    }

    /** 
     * Instance variables 
     */
    ConsusConnection db;
}





