/* Consus - Pure Java Object Relational Database Management System
 * PreparedStatement protocol implementation
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;
import java.io.*;
import java.math.BigDecimal;
import java.util.Calendar;

/**
 * An object that represegnts a precompiled SQL statement.
 * <P>A SQL statement is pre-compiled and stored in a
 * PreparedStatement object. This object can then be used to
 * efficiently execute this statement multiple times. 
 *
 * <P><B>Note:</B> The setXXX methods for setting IN parameter values
 * must specify types that are compatible with the defined SQL type of
 * the input parameter. For instance, if the IN parameter has SQL type
 * Integer, then the method <code>setInt</code> should be used.
 *
 * <p>If arbitrary parameter type conversions are required, the method
 * <code>setObject</code> should be used with a target SQL type.
 * <br>
 * Example of setting a parameter; <code>con</code> is an active connection  
 * <pre><code>
 *   PreparedStatement pstmt = con.prepareStatement("UPDATE EMPLOYEES
 *                                     SET SALARY = ? WHERE ID = ?");
 *   pstmt.setBigDecimal(1, 153833.00)
 *   pstmt.setInt(2, 110592)
 * </code></pre>
 *
 * @see Connection#prepareStatement
 * @see ResultSet 
 */

public class ConsusPreparedStatement extends ConsusStatement
    implements org.garret.consus.ConsusPreparedStatement 
{
    /**
     * Executes the SQL query in this <code>PreparedStatement</code> object
     * and returns the result set generated by the query.
     *
     * @return a ResultSet that contains the data produced by the
     * query; never null
     * @exception SQLException if a database access error occurs
     */
    public ResultSet executeQuery() throws SQLException
    {
	if (stmt.tag != ExprNode.opSelect) { 
	    throw new ConsusCompileError("Select statement expected", 0);
	}
	checkParameters();
	execCount += 1;
	stmt.executeStmt(this);
	isQuery = true;
	oid = 0;
	rec = null;
	select = null;
	return cursor;
    }

    /**
     * Executes the SQL INSERT, UPDATE or DELETE statement
     * in this <code>PreparedStatement</code> object.
     * In addition,
     * SQL statements that return nothing, such as SQL DDL statements,
     * can be executed.
     *
     * @return either the row count for INSERT, UPDATE or DELETE statements;
     * or 0 for SQL statements that return nothing
     * @exception SQLException if a database access error occurs
     */
    public int executeUpdate() throws SQLException { 
	if (stmt.tag == ExprNode.opSelect) { 
	    throw new ConsusCompileError("Use executeQuery for select statement", 0);
	}
	checkParameters();
	isQuery = false;
	oid = 0;
	rec = null;
	select = null;
	execCount += 1;
	return updateCount = stmt.executeStmt(this);
    }

    /**
     * Sets the designated parameter to SQL NULL.
     *
     * <P><B>Note:</B> You must specify the parameter's SQL type.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param sqlType the SQL type code defined in java.sql.Types
     * @exception SQLException if a database access error occurs
     */
    public void setNull(int parameterIndex, int sqlType) throws SQLException { 
	setParameter(--parameterIndex);
	if (params[parameterIndex].type != ExprNode.tpRef) { 
	    throw new ConsusIncompatibleTypesError();
	}
	params[parameterIndex].oid = 0;
    }

    /**
     * Sets the designated parameter to a Java boolean value.  The driver converts this
     * to an SQL BIT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setBoolean(int parameterIndex, boolean x) throws SQLException {
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpBool:
	    params[parameterIndex].bval = x;
	    break;
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = x ? 1 : 0;
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = x ? 1 : 0;
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = String.valueOf(x);
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Sets the designated parameter to a Java byte value.  The driver converts this
     * to an SQL TINYINT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setByte(int parameterIndex, byte x) throws SQLException {
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpBool:
	    params[parameterIndex].bval = x != 0;
	    break;
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = x;
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = (double)x;
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = String.valueOf(x);
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Sets the designated parameter to a Java short value.  The driver converts this
     * to an SQL SMALLINT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setShort(int parameterIndex, short x) throws SQLException {
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpBool:
	    params[parameterIndex].bval = x != 0;
	    break;
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = x;
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = (double)x;
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = String.valueOf(x);
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Sets the designated parameter to a Java int value.  The driver converts this
     * to an SQL INTEGER value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setInt(int parameterIndex, int x) throws SQLException {
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpBool:
	    params[parameterIndex].bval = x != 0;
	    break;
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = x;
	    break;
	  case ExprNode.tpRef:
	    params[parameterIndex].oid = x;
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = (double)x;
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = String.valueOf(x);
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Sets the designated parameter to a Java long value.  The driver converts this
     * to an SQL BIGINT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setLong(int parameterIndex, long x) throws SQLException { 
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpBool:
	    params[parameterIndex].bval = x != 0;
	    break;
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = x;
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = (double)x;
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = String.valueOf(x);
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }


    /**
     * Sets the designated parameter to a Java float value.  The driver converts this
     * to an SQL FLOAT value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setFloat(int parameterIndex, float x) throws SQLException {
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpBool:
	    params[parameterIndex].bval = x != 0;
	    break;
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = (long)x;
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = x;
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = String.valueOf(x);
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }


    /**
     * Sets the designated parameter to a Java double value.  The driver converts this
     * to an SQL DOUBLE value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setDouble(int parameterIndex, double x) throws SQLException {
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpBool:
	    params[parameterIndex].bval = x != 0;
	    break;
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = (long)x;
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = x;
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = String.valueOf(x);
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Sets the designated parameter to a java.lang.BigDecimal value.  
     * The driver converts this to an SQL NUMERIC value when
     * it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setBigDecimal(int parameterIndex, BigDecimal x) throws SQLException
    {
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = x.longValue();
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = x.doubleValue();
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = x.toString();
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }


    /**
     * Sets the designated parameter to a Java String value.  The driver converts this
     * to an SQL VARCHAR or LONGVARCHAR value (depending on the argument's
     * size relative to the driver's limits on VARCHARs) when it sends
     * it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setString(int parameterIndex, String x) throws SQLException { 
	setParameter(--parameterIndex);
	if (x == null) { 
	    x = "";
	}
	try { 
	    switch (params[parameterIndex].type) { 
	      case ExprNode.tpStr:
		params[parameterIndex].sval = x;
		break;
	      case ExprNode.tpInt:
		params[parameterIndex].ival = Long.parseLong(x, 10);
		break;
	      case ExprNode.tpReal:
		params[parameterIndex].dval = Double.parseDouble(x);
		break;
	      case ExprNode.tpBool:
		params[parameterIndex].bval = x.equalsIgnoreCase("true");
		break;
	      case ExprNode.tpClob:
		params[parameterIndex].obj = new ConsusClob(x);
		break;
	      case ExprNode.tpBlob:
		params[parameterIndex].obj = new ConsusBlob(x.getBytes());
		break;
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	} catch(NumberFormatException ex) {
	    throw new ConsusIncompatibleTypesError();
	}
    }
	

    /**
     * Sets the designated parameter to a Java array of bytes.  The driver converts
     * this to an SQL VARBINARY or LONGVARBINARY (depending on the
     * argument's size relative to the driver's limits on VARBINARYs)
     * when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value 
     * @exception SQLException if a database access error occurs
     */
    public void setBytes(int parameterIndex, byte x[]) throws SQLException {
	setParameter(--parameterIndex);
	if (x == null) { 
	    x = new byte[0];
	}
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpArrayInt1:
	    params[parameterIndex].aval = x;
	    break; 
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = new String(x);
	    break;
	  case ExprNode.tpBlob:
	    params[parameterIndex].obj = new ConsusBlob(x);
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Sets the designated parameter to a java.sql.Date value.  The driver converts this
     * to an SQL DATE value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setDate(int parameterIndex, java.sql.Date x)
        throws SQLException 
    {
	setParameter(--parameterIndex);	
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = x.getTime();
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = (double)x.getTime();
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = x.toString();
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Sets the designated parameter to a java.sql.Time value.  The driver converts this
     * to an SQL TIME value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database access error occurs
     */
    public void setTime(int parameterIndex, java.sql.Time x) 
        throws SQLException
    {
	setParameter(--parameterIndex);	
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = x.getTime();
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = (double)x.getTime();
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = x.toString();
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Sets the designated parameter to a java.sql.Timestamp value.  The driver
     * converts this to an SQL TIMESTAMP value when it sends it to the
     * database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value 
     * @exception SQLException if a database access error occurs
     */
    public void setTimestamp(int parameterIndex, java.sql.Timestamp x)
        throws SQLException
    {
	setParameter(--parameterIndex);	
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpInt:
	    params[parameterIndex].ival = x.getTime();
	    break;
	  case ExprNode.tpReal:
	    params[parameterIndex].dval = (double)x.getTime();
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = x.toString();
	    break;	    
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }


    /**
     * Sets the designated parameter to the given input stream, which will have 
     * the specified number of bytes.
     * When a very large ASCII value is input to a LONGVARCHAR
     * parameter, it may be more practical to send it via a
     * java.io.InputStream. JDBC will read the data from the stream
     * as needed, until it reaches end-of-file.  The JDBC driver will
     * do any necessary conversion from ASCII to the database char format.
     * 
     * <P><B>Note:</B> This stream object can either be a standard
     * Java stream object or your own subclass that implements the
     * standard interface.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the Java input stream that contains the ASCII parameter value
     * @param length the number of bytes in the stream 
     * @exception SQLException if a database access error occurs
     */
    public void setAsciiStream(int parameterIndex, java.io.InputStream x, int length)
        throws SQLException
    {
	setBinaryStream(parameterIndex, x, length);
    }

    /**
     * Sets the designated parameter to the given input stream, which will have 
     * the specified number of bytes.
     * When a very large UNICODE value is input to a LONGVARCHAR
     * parameter, it may be more practical to send it via a
     * java.io.InputStream. JDBC will read the data from the stream
     * as needed, until it reaches end-of-file.  The JDBC driver will
     * do any necessary conversion from UNICODE to the database char format.
     * The byte format of the Unicode stream must be Java UTF-8, as
     * defined in the Java Virtual Machine Specification.
     * 
     * <P><B>Note:</B> This stream object can either be a standard
     * Java stream object or your own subclass that implements the
     * standard interface.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...  
     * @param x the java input stream which contains the
     * UNICODE parameter value 
     * @param length the number of bytes in the stream 
     * @exception SQLException if a database access error occurs
     * @deprecated
     */
    public void setUnicodeStream(int parameterIndex, java.io.InputStream x, 
				 int length) throws SQLException
    {
	setBinaryStream(parameterIndex, x, length);
    }


    /**
     * Sets the designated parameter to the given input stream, which will have 
     * the specified number of bytes.
     * When a very large binary value is input to a LONGVARBINARY
     * parameter, it may be more practical to send it via a
     * java.io.InputStream. JDBC will read the data from the stream
     * as needed, until it reaches end-of-file.
     * 
     * <P><B>Note:</B> This stream object can either be a standard
     * Java stream object or your own subclass that implements the
     * standard interface.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the java input stream which contains the binary parameter value
     * @param length the number of bytes in the stream 
     * @exception SQLException if a database access error occurs
     */
    public void setBinaryStream(int parameterIndex, java.io.InputStream stream, 
				int length) throws SQLException
    {
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpBlob: 
	    params[parameterIndex].obj = new ConsusBlob(length, stream);
	    break;
	  case ExprNode.tpStr:
	  case ExprNode.tpArrayInt1:
	    byte[] buf = new byte[length];
	    try { 
		int offs = 0;
		while (offs < length) {
		    int rc = stream.read(buf, offs, length - offs);
		    if (rc <= 0) { 
			throw new ConsusRuntimeError("Failed to read all bytes from stream");
		    }
		    offs += rc;
		}
	    } catch (java.io.IOException x) { 
		throw new ConsusRuntimeError("Failed to read from stream: " + 
					     x.getMessage());
	    }
	    if (params[parameterIndex].type == ExprNode.tpStr) { 
		params[parameterIndex].sval = new String(buf);
	    } else { 
		params[parameterIndex].aval = buf;
	    }
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }



    /**
     * Clears the current parameter values immediately.
     * <P>In general, parameter values remain in force for repeated use of a
     * Statement. Setting a parameter value automatically clears its
     * previous value.  However, in some cases it is useful to immediately
     * release the resources used by the current parameter values; this can
     * be done by calling clearParameters.
     *
     * @exception SQLException if a database access error occurs
     */
    public void clearParameters() throws SQLException {
	nSetParams = 0;
	params = null;
    }

    //----------------------------------------------------------------------
    // Advanced features:

    /**
     * <p>Sets the value of a parameter using an object. The second
     * argument must be an object type; for integral values, the
     * java.lang equivalent objects should be used.
     *
     * <p>The given Java object will be converted to the targetSqlType
     * before being sent to the database.
     *
     * If the object has a custom mapping (is of a class implementing SQLData),
     * the JDBC driver should call its method <code>writeSQL</code> to write it 
     * to the SQL data stream.
     * If, on the other hand, the object is of a class implementing
     * Ref, Blob, Clob, Struct, 
     * or Array, the driver should pass it to the database as a value of the 
     * corresponding SQL type.
     *
     * <p>Note that this method may be used to pass datatabase-
     * specific abstract data types. 
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the object containing the input parameter value
     * @param targetSqlType the SQL type (as defined in java.sql.Types) to be 
     * sent to the database. The scale argument may further qualify this type.
     * @param scale for java.sql.Types.DECIMAL or java.sql.Types.NUMERIC types,
     *          this is the number of digits after the decimal point.  For all other
     *          types, this value will be ignored.
     * @exception SQLException if a database access error occurs
     */
    public void setObject(int parameterIndex, Object x, int targetSqlType, int scale)
            throws SQLException
    {
	setObject(parameterIndex, x, targetSqlType);
    }

   /**
     * Sets the value of the designated parameter with the given object.
     * This method is like setObject above, except that it assumes a scale of zero.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the object containing the input parameter value
     * @param targetSqlType the SQL type (as defined in java.sql.Types) to be 
     *                      sent to the database
     * @exception SQLException if a database access error occurs
     */
    public void setObject(int parameterIndex, Object x, int targetSqlType) 
      throws SQLException
    {
	try {
	    switch (targetSqlType) { 
	      case Types.BIT:
		if (!(x instanceof Boolean)) {
		    if (x instanceof Number) { 
			x = new Boolean(((Number)x).longValue() != 0);
		    } else if (x instanceof String) { 
			x = new Boolean((String)x);
		    } else { 
			throw new ConsusIncompatibleTypesError();
		    }
		}
		break;
	      case Types.TINYINT:
	      case Types.SMALLINT:
	      case Types.INTEGER:
	      case Types.BIGINT:
		if (!(x instanceof Long)) { 
		    if (x instanceof Number) { 
			x = new Long(((Number)x).longValue());
		    } else if (x instanceof String) { 
			x = new Long((String)x);
		    } else { 
			throw new ConsusIncompatibleTypesError();
		    }
		}
		break;
	      case Types.FLOAT:
	      case Types.REAL:
	      case Types.DOUBLE:
	      case Types.NUMERIC:
	      case Types.DECIMAL:
		if (!(x instanceof Double)) { 
		    if (x instanceof Number) { 
			x = new Double(((Number)x).doubleValue());
		    } else if (x instanceof String) { 
			x = new Double((String)x);
		    } else { 
			throw new ConsusIncompatibleTypesError();
		    }
		}
		break;
	      case Types.CHAR:
	      case Types.VARCHAR:
	      case Types.LONGVARCHAR:
		if (!(x instanceof String)) { 
		    x = x.toString();
		}
		break;
	      case Types.DATE:
		if (!(x instanceof java.util.Date)) { 
		    if (x instanceof Number) { 
			x = new java.util.Date(((Number)x).longValue());
		    } else if (x instanceof String) { 
			x = java.sql.Date.valueOf((String)x);
		    } else { 
			throw new ConsusIncompatibleTypesError();
		    }
		}
		break;
	      case Types.TIME:
		if (!(x instanceof java.util.Date)) { 
		    if (x instanceof Number) { 
			x = new java.util.Date(((Number)x).longValue());
		    } else if (x instanceof String) { 
			x = java.sql.Time.valueOf((String)x);
		    } else { 
			throw new ConsusIncompatibleTypesError();
		    }
		}
		break;
	      case Types.TIMESTAMP:
		if (!(x instanceof java.util.Date)) { 
		    if (x instanceof Number) { 
			x = new java.util.Date(((Number)x).longValue());
		    } else if (x instanceof String) { 
			x = java.sql.Timestamp.valueOf((String)x);
		    } else { 
			throw new ConsusIncompatibleTypesError();
		    }
		}
		break;
	      case Types.BINARY:
	      case Types.VARBINARY:
	      case Types.LONGVARBINARY:
		if (!(x instanceof byte[])) {
		    if (x instanceof Serializable) { 
			try { 
			    ByteArrayOutputStream bs = new ByteArrayOutputStream();
			    ObjectOutputStream os = new ObjectOutputStream(bs);
			    os.writeObject(x);
			    x = bs.toByteArray();
			    os.close();
			} catch (IOException ex) {
			    throw new ConsusRuntimeError("Serialization error");
			}
		    } else { 
			throw new ConsusIncompatibleTypesError();
		    }
		}
		break;
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	    setObject(parameterIndex, x);
	} catch(NumberFormatException ex) {
	    throw new ConsusIncompatibleTypesError();
	}
    }


    /**
     * <p>Sets the value of a parameter using an object; use the
     * java.lang equivalent objects for integral values.
     *
     * <p>The JDBC specification specifies a standard mapping from
     * Java Object types to SQL types.  The given argument java object
     * will be converted to the corresponding SQL type before being
     * sent to the database.
     *
     * <p>Note that this method may be used to pass datatabase-
     * specific abstract data types, by using a Driver-specific Java
     * type.
     *
     * If the object is of a class implementing SQLData,
     * the JDBC driver should call its method <code>writeSQL</code> to write it 
     * to the SQL data stream.
     * If, on the other hand, the object is of a class implementing
     * Ref, Blob, Clob, Struct, 
     * or Array, then the driver should pass it to the database as a value of the 
     * corresponding SQL type.
     *
     * This method throws an exception if there is an ambiguity, for example, if the
     * object is of a class implementing more than one of those interfaces.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the object containing the input parameter value 
     * @exception SQLException if a database access error occurs
     */
    public void setObject(int parameterIndex, Object x) throws SQLException
    {
	setParameter(--parameterIndex);	
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpBool:
	    if (x instanceof Boolean) { 
		params[parameterIndex].bval = ((Boolean)x).booleanValue();
	    } else if (x instanceof Number) {
		params[parameterIndex].bval = ((Number)x).longValue() != 0;
	    } else { 		
		throw new ConsusIncompatibleTypesError(x);
	    }
	    break;
	    
	  case ExprNode.tpInt:
	    if (x instanceof Number) { 
		params[parameterIndex].ival = ((Number)x).longValue();
	    } else if (x instanceof java.util.Date) {
		params[parameterIndex].ival = ((java.util.Date)x).getTime();
	    } else { 		
		throw new ConsusIncompatibleTypesError(x);
	    }
	    break;
	  case ExprNode.tpReal:
	    if (x instanceof Number) { 
		params[parameterIndex].dval = ((Number)x).doubleValue();
	    } else if (x instanceof java.util.Date) {
		params[parameterIndex].dval = (double)((java.util.Date)x).getTime();
	    } else { 		
		throw new ConsusIncompatibleTypesError(x);
	    }		
	    break;
	  case ExprNode.tpStr:
	    params[parameterIndex].sval = x == null ? "" : x.toString();
	    break;
	  case ExprNode.tpArrayInt1:
	    if (x instanceof byte[]) { 
		params[parameterIndex].aval = (byte[])x;
	    } else { 		
		throw new ConsusIncompatibleTypesError(x);
	    }		
	    break;	    
	  case ExprNode.tpBlob:
	    if (x instanceof Blob) { 
		params[parameterIndex].obj = x;
	    } else if (x instanceof byte[]) { 
		params[parameterIndex].obj = new ConsusBlob((byte[])x);
	    } else { 		
		throw new ConsusIncompatibleTypesError(x);
	    }		
	    break;
	  case ExprNode.tpClob:
	    if (x instanceof Clob) { 
		params[parameterIndex].obj = x;
	    } else if (x instanceof String) { 
		params[parameterIndex].obj = new ConsusClob((String)x);
	    } else { 		
		throw new ConsusIncompatibleTypesError(x);
	    }		
	    break;
	  case ExprNode.tpRef:	  
	    if (x instanceof Ref) { 
		params[parameterIndex].oid = (int)((ConsusRef)x).getOid();
	    } else if (x == null) { 
		params[parameterIndex].oid = 0;
	    } else { 		
		throw new ConsusIncompatibleTypesError(x);
	    }		
	    break;
	  case ExprNode.tpCursor:
	    if (x instanceof ConsusRef) { 
		params[parameterIndex].oid = (int)((ConsusRef)x).getOid();
		params[parameterIndex].itype = ExprNode.tpRef;
	    } else if (x instanceof ResultSet) { 
		params[parameterIndex].obj = x;		
		params[parameterIndex].itype = ExprNode.tpCursor;
	    } else if (x instanceof Ref[]) {
		params[parameterIndex].obj = x;		
		params[parameterIndex].itype = ExprNode.tpArrayRef;
	    } else if (x == null) { 
		params[parameterIndex].oid = 0;
		params[parameterIndex].itype = ExprNode.tpRef;
	    } else {
		throw new ConsusIncompatibleTypesError(x);
	    }		
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError(x);
	}
    }

    /**
     * Executes any kind of SQL statement.
     * Some prepared statements return multiple results; the execute
     * method handles these complex statements as well as the simpler
     * form of statements handled by executeQuery and executeUpdate.
     *
     * @exception SQLException if a database access error occurs
     * @see Statement#execute
     */
    public boolean execute() throws SQLException { 
	checkParameters();	
	execCount += 1;
	updateCount = stmt.executeStmt(this);
	isQuery = stmt.tag == ExprNode.opSelect; 
	oid = 0;
	rec = null;
	select = null;
	return isQuery;
    }


    //--------------------------JDBC 2.0-----------------------------

    /**
     * JDBC 2.0
     *
     * Adds a set of parameters to the batch.
     * 
     * @exception SQLException if a database access error occurs
     * @see Statement#addBatch
     */
    public void addBatch() throws SQLException
    {
	if (batchResults == null) { 
	    batchResults = new int[1];
	} else { 
	    int[] newResults = new int[batchResults.length+1];
	    System.arraycopy(batchResults, 0, newResults, 0, batchResults.length);
	    batchResults = newResults;
	}
	batchResults[batchResults.length-1] = execute() ? getUpdateCount() : -1;
    }


    /**
     * JDBC 2.0
     *
     * Sets the designated parameter to the given <code>Reader</code>
     * object, which is the given number of characters long.
     * When a very large UNICODE value is input to a LONGVARCHAR
     * parameter, it may be more practical to send it via a
     * java.io.Reader. JDBC will read the data from the stream
     * as needed, until it reaches end-of-file.  The JDBC driver will
     * do any necessary conversion from UNICODE to the database char format.
     * 
     * <P><B>Note:</B> This stream object can either be a standard
     * Java stream object or your own subclass that implements the
     * standard interface.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the java reader which contains the UNICODE data
     * @param length the number of characters in the stream 
     * @exception SQLException if a database access error occurs
     */
    public void setCharacterStream(int parameterIndex,
				   java.io.Reader reader,
				   int length) throws SQLException
    {
	setParameter(--parameterIndex);
	switch (params[parameterIndex].type) { 
	  case ExprNode.tpClob:
 	    params[parameterIndex].obj = new ConsusClob(length, reader);
	    break;
	  case ExprNode.tpStr:
	    try { 
		char[] buf = new char[length];
		int offs = 0;
		while (offs < length) {
		    int rc = reader.read(buf, offs, length - offs);
		    if (rc <= 0) { 
			throw new ConsusRuntimeError("Failed to read all characters from stream");
		    }
		    offs += rc;
		}
		params[parameterIndex].sval = new String(buf);
	    } catch (java.io.IOException x) { 
		throw new ConsusRuntimeError("Failed to read from stream: " + 
					     x.getMessage());
	    }
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }



    /**
     * JDBC 2.0
     *
     * Sets a REF(&lt;structured-type&gt;) parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing data of an SQL REF Type
     * @exception SQLException if a database access error occurs
     */
    public void setRef (int i, Ref x) throws SQLException { 
	setParameter(--i);
	if (params[i].type == ExprNode.tpCursor) { 
	    params[i].itype = ExprNode.tpRef;
	} else if (params[i].type != ExprNode.tpRef) { 
	    throw new ConsusIncompatibleTypesError();
	}
	params[i].oid = (x == null) ? 0 : (int)((ConsusRef)x).getOid();
    }



    /**
     * JDBC 2.0
     *
     * Sets a BLOB parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing a BLOB
     * @exception SQLException if a database access error occurs
     */
    public void setBlob (int i, Blob x) throws SQLException
    {
	setParameter(--i);
	if (params[i].type != ExprNode.tpBlob) { 
	    throw new ConsusIncompatibleTypesError();
	}
	if (x == null) { 
	    x = new ConsusBlob(db, 0, 0);
	}
	params[i].obj = x;
    }



    /**
     * JDBC 2.0
     *
     * Sets a CLOB parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing a CLOB
     * @exception SQLException if a database access error occurs
     */
    public void setClob (int i, Clob x) throws SQLException
    {
	setParameter(--i);
	if (params[i].type != ExprNode.tpClob) { 
	    throw new ConsusIncompatibleTypesError();
	}
	if (x == null) { 
	    x = new ConsusClob(db, 0, 0);
	}
	params[i].obj = x;
    }



    /**
     * JDBC 2.0
     *
     * Sets an Array parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing an SQL array
     * @exception SQLException if a database access error occurs
     */
    public void setArray (int i, Array x) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setArray");
    }



    /**
     * JDBC 2.0
     *
     * Gets the number, types and properties of a ResultSet's columns.
     *
     * @return the description of a ResultSet's columns
     * @exception SQLException if a database access error occurs
     */
    public ResultSetMetaData getMetaData() throws SQLException
    {
	return cursor.getMetaData();
    }

    /**
     * JDBC 2.0
     *
     * Sets the designated parameter to a java.sql.Date value,
     * using the given <code>Calendar</code> object.  The driver uses
     * the <code>Calendar</code> object to construct an SQL DATE,
     * which the driver then sends to the database.  With a
     * a <code>Calendar</code> object, the driver can calculate the date
     * taking into account a custom timezone and locale.  If no
     * <code>Calendar</code> object is specified, the driver uses the default
     * timezone and locale.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @param cal the <code>Calendar</code> object the driver will use
     *            to construct the date
     * @exception SQLException if a database access error occurs
     */
    public void setDate(int parameterIndex, java.sql.Date x, Calendar cal)
        throws SQLException
    {
	throw new ConsusNotImplemented("PreparedStatement.setDate(calendar)");
    }

    /**
     * JDBC 2.0
     *
     * Sets the designated parameter to a java.sql.Time value,
     * using the given <code>Calendar</code> object.  The driver uses
     * the <code>Calendar</code> object to construct an SQL TIME,
     * which the driver then sends to the database.  With a
     * a <code>Calendar</code> object, the driver can calculate the time
     * taking into account a custom timezone and locale.  If no
     * <code>Calendar</code> object is specified, the driver uses the default
     * timezone and locale.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @param cal the <code>Calendar</code> object the driver will use
     *            to construct the time
     * @exception SQLException if a database access error occurs
     */
    public void setTime(int parameterIndex, java.sql.Time x, Calendar cal) 
        throws SQLException
    {
	throw new ConsusNotImplemented("PreparedStatement.setTime(calendar)");
    }

    /**
     * JDBC 2.0
     *
     * Sets the designated parameter to a java.sql.Timestamp value,
     * using the given <code>Calendar</code> object.  The driver uses
     * the <code>Calendar</code> object to construct an SQL TIMESTAMP,
     * which the driver then sends to the database.  With a
     * a <code>Calendar</code> object, the driver can calculate the timestamp
     * taking into account a custom timezone and locale.  If no
     * <code>Calendar</code> object is specified, the driver uses the default
     * timezone and locale.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value 
     * @param cal the <code>Calendar</code> object the driver will use
     *            to construct the timestamp
     * @exception SQLException if a database access error occurs
     */
    public void setTimestamp(int parameterIndex, java.sql.Timestamp x, Calendar cal)
        throws SQLException
    {
	throw new ConsusNotImplemented("PreparedStatement.setTimestamp(calendar)");
    }


    /**
     * JDBC 2.0
     *
     * Sets the designated parameter to SQL NULL.  This version of setNull should
     * be used for user-named types and REF type parameters.  Examples
     * of user-named types include: STRUCT, DISTINCT, JAVA_OBJECT, and 
     * named array types.
     *
     * <P><B>Note:</B> To be portable, applications must give the
     * SQL type code and the fully-qualified SQL type name when specifying
     * a NULL user-defined or REF parameter.  In the case of a user-named type 
     * the name is the type name of the parameter itself.  For a REF 
     * parameter the name is the type name of the referenced type.  If 
     * a JDBC driver does not need the type code or type name information, 
     * it may ignore it.     
     *
     * Although it is intended for user-named and Ref parameters,
     * this method may be used to set a null parameter of any JDBC type.
     * If the parameter does not have a user-named or REF type, the given
     * typeName is ignored.
     *
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param sqlType a value from java.sql.Types
     * @param typeName the fully-qualified name of an SQL user-named type,
     *  ignored if the parameter is not a user-named type or REF 
     * @exception SQLException if a database access error occurs
     */
    public void setNull (int parameterIndex, int sqlType, String typeName) 
	throws SQLException
    {
	setParameter(--parameterIndex);
	if (params[parameterIndex].type != ExprNode.tpRef) { 
	    throw new ConsusIncompatibleTypesError();
	}
	params[parameterIndex].oid = 0;
    }


    /**
     * Sets the designated parameter to the given <code>java.net.URL</code> value. 
     * The driver converts this to an SQL <code>DATALINK</code> value
     * when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the <code>java.net.URL</code> object to be set
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */ 
    public void setURL(int parameterIndex, java.net.URL x) throws SQLException
    {
	setString(parameterIndex, x.toString());
    }

    /**
     * Retrieves the number, types and properties of this 
     * <code>PreparedStatement</code> object's parameters.
     *
     * @return a <code>ParameterMetaData</code> object that contains information
     *         about the number, types and properties of this 
     *         <code>PreparedStatement</code> object's parameters
     * @exception SQLException if a database access error occurs
     * @see ParameterMetaData
     * @since 1.4
     */
    public ParameterMetaData getParameterMetaData() throws SQLException
    {
	throw new ConsusNotImplemented("ConsusPreparedStatement.getParameterMetaData()");
    }
    
    public void setNClob(int parameterIndex, NClob value) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setNClob");
    }

    public void setNClob(int parameterIndex, Reader reader, long length) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setNClob");
    }

    public void setNClob(int parameterIndex, Reader reader) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setNClob");
    }

    public void setBlob(int parameterIndex, InputStream inputStream, long length) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setBlob");
    }

    public void setSQLXML(int parameterIndex, SQLXML xmlObject) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setSQLXML");
    }

    public void setAsciiStream(int parameterIndex, java.io.InputStream x, long length) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setAsciiStream");
    }
    
    public void setBinaryStream(int parameterIndex, java.io.InputStream x, long length) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setBinaryStream");
    }    

    public void setCharacterStream(int parameterIndex,
                                   java.io.Reader reader,
                                   long length) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setCharacterStream");
    }    

    public void setAsciiStream(int parameterIndex, java.io.InputStream x) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setAsciiStream");
    }

    public void setBinaryStream(int parameterIndex, java.io.InputStream x) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setBinaryStream");
    }    

    public void setCharacterStream(int parameterIndex, java.io.Reader reader) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setCharacterStream");
    }    

    public void setNCharacterStream(int parameterIndex, Reader value, long length) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setNCharacterStream");
    }    

    public void setNCharacterStream(int parameterIndex, Reader value) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setNCharacterStream");
    }    

    public void setClob(int parameterIndex, Reader reader) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setClob");
    }    

    public void setClob(int parameterIndex, Reader reader, long length) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setClob");
    }    

    public void setBlob(int parameterIndex, InputStream inputStream) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setBlob");
    }    

    public void setNString(int parameterIndex, String value) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setNString");
    }    

    public void setRowId(int parameterIndex, RowId x) throws SQLException
    {
        throw new ConsusNotImplemented("PreparedStatement.setRawId");
    }    


    final void setParameter(int i) throws SQLException 
    { 
	if (i < 0 || i >= params.length) { 
	    throw new ConsusRuntimeError("Parameter index " + i + " is out of range");
	}
	if (!params[i].isSet) { 
	    nSetParams += 1;
	    params[i].isSet = true;
	}
    }
    
    final void checkParameters() throws SQLException { 
	if (params != null && nSetParams < params.length) { 
	    throw new ConsusRuntimeError("Not all query parameters were set");
	}
	if (db.dbTraceEnabled) { 
	    StringBuffer buf = new StringBuffer();
	    int p = 0;
	    for (int i = 0; i < params.length; i++) { 
		int q = sql.indexOf('?', p);
		System.out.print(sql.substring(p, q));
		p = q + 1;
		System.out.print(params[i].toString());
	    }
	    System.out.println();
	}
    }

    ConsusPreparedStatement(ConsusConnection conxn, String sql, int resultSetType, int resultSetConcurrency) throws SQLException 
    { 
	super(conxn, resultSetType, resultSetConcurrency);
	Compiler compiler = new Compiler(conxn, sql);
	stmt = compiler.statement();
	params = compiler.params;
	nSetParams = 0;
	this.sql = sql;
    }

    ConsusPreparedStatement(ClientConnection conxn) { 
	super(conxn);
    }

     ConsusPreparedStatement(ClientConnection conxn, int resultSetType, int resultSetConcurrency) { 
	super(conxn, resultSetType, resultSetConcurrency);
    }

   ConsusPreparedStatement(ConsusConnection conxn, String sql) throws SQLException 
    {
	super(conxn);
	Compiler compiler = new Compiler(conxn, sql);
	stmt = compiler.statement();
	params = compiler.params;
	nSetParams = 0;
	this.sql = sql;
    }

   /*
    ******************************************************
    * Instance variables 
    ******************************************************
    */
    ParameterNode[]  params;
    int              nSetParams;    
    StmtNode         stmt;
    String           sql;
}
