/* Consus - Pure Java Object Relational Database Management System
 * Character stream returned by ConsusClob
 * Created by Konstantin Knizhnik, Mar 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;
import java.io.*;
import java.sql.*;


/**
 * Reader for CLOB object
 */
public class ConsusReader extends Reader {
    ConsusConnection db;
    long             pos;
    long             offs;
    long             size;
    long             markPos;

    ConsusReader(ConsusConnection db, long pos, long size) { 
	this.db = db;
	this.pos = pos;
	this.size = size*2;
	offs = 0;
	markPos = 0;
    }

    /**
     * Read a single character.  This method will block until a character is
     * available, an I/O error occurs, or the end of the stream is reached.
     *
     * <p> Subclasses that intend to support efficient single-character input
     * should override this method.
     *
     * @return     The character read, as an integer in the range 0 to 16383
     *             (<tt>0x00-0xffff</tt>), or -1 if the end of the stream has
     *             been reached
     *
     * @exception  IOException  If an I/O error occurs
     */
    public int read() throws IOException {
	if (offs + 1 >= size) { 
	    return -1;
	}
	try { 
	    int pageOffs = (int)(pos + offs) & (Page.pageSize-1);
	    Page pg = db.pool.getPage(pos + offs - pageOffs);
	    int val = (char)Bytes.unpack2(pg.data, pageOffs);
	    offs += 2;
	    db.pool.unfix(pg);
	    return val;
	} catch(SQLException x) { 
	    throw new IOException(x.getMessage());
	}
    }

    /**
     * Read characters into a portion of an array.  This method will block
     * until some input is available, an I/O error occurs, or the end of the
     * stream is reached.
     *
     * @param      cbuf  Destination buffer
     * @param      off   Offset at which to start storing characters
     * @param      len   Maximum number of characters to read
     *
     * @return     The number of characters read, or -1 if the end of the
     *             stream has been reached
     *
     * @exception  IOException  If an I/O error occurs
     */
    public int read(char cbuf[], int off, int len) throws IOException { 
	int i;
	if (len == 0) { 
	    return 0;
	}
	if (offs + 1 >= size) { 
	    return -1;
	}
	int fetched = 0;
	if (offs + len*2 > size) { 
	    len = (int)(size - offs) >>> 1;
	}
	try { 
	    while (true) { 
		int pageOffs = (int)(pos + offs) & (Page.pageSize-1);
		Page pg = db.pool.getPage(pos + offs - pageOffs);
		if ((len - fetched)*2 > Page.pageSize - pageOffs) { 
		    offs += Page.pageSize - pageOffs;
		    while (pageOffs < Page.pageSize) { 
			cbuf[off + fetched] = (char)Bytes.unpack2(pg.data, pageOffs); 
			fetched += 1;
			pageOffs += 2;
		    }
		    db.pool.unfix(pg);
		} else { 
		    while (fetched < len) { 
			cbuf[off + fetched] = (char)Bytes.unpack2(pg.data, pageOffs); 
			fetched += 1;
			pageOffs += 2;
		    }		    
		    offs += len*2;
		    db.pool.unfix(pg);
		    return fetched;
		}
	    }
	} catch(SQLException x) { 
	    throw new IOException(x.getMessage());
	}
    }


    /**
     * Skip characters.  This method will block until some characters are
     * available, an I/O error occurs, or the end of the stream is reached.
     *
     * @param  n  The number of characters to skip
     *
     * @return    The number of characters actually skipped
     *
     * @exception  IOException  If an I/O error occurs
     */
    public long skip(long n) throws IOException {
	n <<= 1;
	if (offs + n > size) { 
	    n = size - offs;
	    offs = size;
	} else if (offs + n < 0) { 
	    n = -offs;
	    offs = 0;
	} else { 
	    offs += n;
	}
	return n >>> 1;
    }

    /**
     * Tell whether this stream is ready to be read.
     *
     * @return True if the next read() is guaranteed not to block for input,
     * false otherwise.  Note that returning false does not guarantee that the
     * next read will block.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public boolean ready() throws IOException {
	return offs < size;
    }

    /**
     * Tell whether this stream supports the mark() operation.
     */
    public boolean markSupported() {
	return true;
    }

    /**
     * Mark the present position in the stream.  Subsequent calls to reset()
     * will attempt to reposition the stream to this point.  Not all
     * character-input streams support the mark() operation.
     *
     * @param  readAheadLimit  Limit on the number of characters that may be
     *                         read while still preserving the mark.  After
     *                         reading this many characters, attempting to
     *                         reset the stream may fail.
     *
     * @exception  IOException  If the stream does not support mark(),
     *                          or if some other I/O error occurs
     */
    public void mark(int readAheadLimit) throws IOException {
	markPos = offs;
    }

    /**
     * Reset the stream.  If the stream has been marked, then attempt to
     * reposition it at the mark.  If the stream has not been marked, then
     * attempt to reset it in some way appropriate to the particular stream,
     * for example by repositioning it to its starting point.  Not all
     * character-input streams support the reset() operation, and some support
     * reset() without supporting mark().
     *
     * @exception  IOException  If the stream has not been marked,
     *                          or if the mark has been invalidated,
     *                          or if the stream does not support reset(),
     *                          or if some other I/O error occurs
     */
    public void reset() throws IOException {
	offs = markPos;
    }

    /**
     * Close reader and release all resources
     */
    public void close() {}
}






