/* Consus - Pure Java Object Relational Database Management System
 * Implementation of ResultSet protocol. Bidirectional, random accessible database cursor.
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;

import org.garret.consus.*;

import java.sql.*;
import java.math.BigDecimal;
import java.util.Calendar;
import java.util.Arrays;
import java.io.Reader;
import java.io.InputStream;

/**
 * <P>A ResultSet provides access to a table of data.  A ResultSet
 * object is usually generated by executing a Statement. 
 * 
 * <P>A ResultSet maintains a cursor pointing to its current row of
 * data.  Initially the cursor is positioned before the first row.
 * The 'next' method moves the cursor to the next row.
 *
 * <P>The getXXX methods retrieve column values for the current
 * row.  You can retrieve values using either the index number of the
 * column or the name of the column.  In general, using the 
 * column index will be more efficient.  Columns are numbered from 1.
 *
 * <P>For maximum portability, ResultSet columns within each row should be
 * read in left-to-right order and each column should be read only once.
 *
 * <P>For the getXXX methods, the JDBC driver attempts to convert the
 * underlying data to the specified Java type and returns a suitable
 * Java value.  See the JDBC specification for allowable mappings
 * from SQL types to Java types with the ResultSet.getXXX methods.
 *
 * <P>Column names used as input to getXXX methods are case
 * insensitive.  When performing a getXXX using a column name, if
 * several columns have the same name, then the value of the first
 * matching column will be returned. The column name option is
 * designed to be used when column names are used in the SQL
 * query. For columns that are NOT explicitly named in the query, it
 * is best to use column numbers. If column names are used, there is
 * no way for the programmer to guarantee that they actually refer to
 * the intended columns.
 *
 * <P>A ResultSet is automatically closed by the Statement that
 * generated it when that Statement is closed, re-executed, or used
 * to retrieve the next result from a sequence of multiple results.
 * 
 * <P>The number, types and properties of a ResultSet's columns are
 * provided by the ResulSetMetaData object returned by the getMetaData
 * method.
 *
 * @see Statement#executeQuery 
 * @see Statement#getResultSet 
 * @see ResultSetMetaData 
 */

public class ConsusResultSet implements org.garret.consus.ConsusResultSet {
    final void fetch() throws SQLException { 
	if (tuples == null) { 
	    oid = selection[currPos];
	    currRec = db.get(oid);
	    obj = null;
	    if (updateMode) { 
		Arrays.fill(stmt.updated, false);
		updateColumns = false;
	    }
	    cursorState = CS_CURRENT_ROW;
	}
    }	

    public int getHoldability() throws SQLException
    {
        return 2; // CLOSE_CURSORS_AT_COMMIT
    }

    public boolean isClosed() { 
        return !opened;
    }

    /**
     * Moves the cursor down one row from its current position.
     * A ResultSet cursor is initially positioned before the first row; the
     * first call to next makes the first row the current row; the
     * second call makes the second row the current row, and so on. 
     *
     * @return true if the new current row is valid; false if there
     * are no more rows 
     * @exception SQLException if a database access error occurs
     */
    public boolean next() throws SQLException { 
	int newCurrPos = (cursorState == CS_DELETED_ROW) ? currPos : currPos + 1;
	if (newCurrPos >= nSelected) { 
	    if (cursorState == CS_CURRENT_ROW) { 
		cursorState = CS_AFTER_LAST;
	    }
	    return false;
	}
	currPos = newCurrPos;
	fetch();
	return true;
    }

    /**
     * Releases this <code>ResultSet</code> object's database and
     * JDBC resources immediately instead of waiting for
     * this to happen when it is automatically closed.
     *
     * <P><B>Note:</B> A ResultSet is automatically closed by the
     * Statement that generated it when that Statement is closed,
     * re-executed, or is used to retrieve the next result from a
     * sequence of multiple results. A ResultSet is also automatically
     * closed when it is garbage collected.  
     *
     * @exception SQLException if a database access error occurs
     */
    public void close() throws SQLException { 
	reset(true);
    }

    final void reset(boolean endStatement) throws SQLException { 
	if (db != null && opened) { 	 
	    if (endStatement) { 
		db.endStatement();
	    }
	    db.closeCursor(this);
	}
	table = null;
	nSelected = 0;
	currPos = -1;
	cursorState = CS_BEFORE_FIRST;
	obj = null;
	oid = 0;
	currRec = null;
	selectMap = null;
	tupleColumns = null;
	tuples = null;
	opened = false;
    }

    /**
     * Reports whether
     * the last column read had a value of SQL NULL.
     * Note that you must first call getXXX on a column to try to read
     * its value and then call wasNull() to see if the value read was
     * SQL NULL.
     *
     * @return true if last column read was SQL NULL and false otherwise
     * @exception SQLException if a database access error occurs
     */
    public boolean wasNull() throws SQLException { 
	return false;
    }
    
    //======================================================================
    // Methods for accessing results by column index
    //======================================================================

    /**
     * Gets the value of a column in the current row as a Java String.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public String getString(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    return tuples[currPos][columnIndex-1].toString();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	try { 
	    Object value = fd.fetch(currRec, 0, true);
	    return value != null ? String.valueOf(value) : null;
	} catch(Exception x) { 
	    throw new ConsusRuntimeError(x.getMessage());
	}	 
    }

    /**
     * Gets the value of a column in the current row as a Java boolean.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is false
     * @exception SQLException if a database access error occurs
     */
    public boolean getBoolean(int columnIndex) throws SQLException {
	checkColumnIndex(columnIndex);
	if (tuples != null)  { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpBool:
		return ((Boolean)tuples[currPos][columnIndex-1]).booleanValue();
	      case ExprNode.tpInt:
		return ((Long)tuples[currPos][columnIndex-1]).longValue() != 0;
	      case ExprNode.tpReal:
		return ((Double)tuples[currPos][columnIndex-1]).doubleValue() != 0;
	      case ExprNode.tpStr:
		return Boolean.getBoolean((String)tuples[currPos][columnIndex-1]);
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpBool:
	    return currRec[fd.offs] != 0;
	  case FieldDescriptor.tpString:
	    return Boolean.getBoolean(Varying.getString(currRec, fd.offs));
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Gets the value of a column in the current row as a Java byte.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public byte getByte(int columnIndex) throws SQLException {
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return ((Number)tuples[currPos][columnIndex-1]).byteValue();
	      case ExprNode.tpStr:
		try { 
		    return Byte.parseByte((String)tuples[currPos][columnIndex-1]);
		} catch(NumberFormatException x) {
		    throw new ConsusRuntimeError("Failed to convert from string: " 
						 + x.getMessage());
		}
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    return currRec[fd.offs];
	  case FieldDescriptor.tpInt2:
	    return (byte)Bytes.unpack2(currRec, fd.offs);
	  case FieldDescriptor.tpInt4:
	    return (byte)Bytes.unpack4(currRec, fd.offs);
	  case FieldDescriptor.tpInt8:
	    return (byte)Bytes.unpack8(currRec, fd.offs);
	  case FieldDescriptor.tpReal4:
	    return (byte)Float.intBitsToFloat(Bytes.unpack4(currRec, fd.offs));
	  case FieldDescriptor.tpReal8:
	    return (byte)Double.longBitsToDouble(Bytes.unpack8(currRec, fd.offs));
	  case FieldDescriptor.tpString:
	    try { 
		return Byte.parseByte(Varying.getString(currRec, fd.offs));
	    } catch(NumberFormatException x) {
		throw new ConsusRuntimeError("Failed to convert from string: " 
					     + x.getMessage());
	    }
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }


    /**
     * Gets the value of a column in the current row as a Java short.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public short getShort(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return ((Number)tuples[currPos][columnIndex-1]).shortValue();
	      case ExprNode.tpStr:
		try { 
		    return Short.parseShort((String)tuples[currPos][columnIndex-1]);
		} catch(NumberFormatException x) {
		    throw new ConsusRuntimeError("Failed to convert from string: " 
						 + x.getMessage());
		}
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    return currRec[fd.offs];
	  case FieldDescriptor.tpInt2:
	    return Bytes.unpack2(currRec, fd.offs);
	  case FieldDescriptor.tpInt4:
	    return (short)Bytes.unpack4(currRec, fd.offs);
	  case FieldDescriptor.tpInt8:
	    return (short)Bytes.unpack8(currRec, fd.offs);
	  case FieldDescriptor.tpReal4:
	    return (short)Float.intBitsToFloat(Bytes.unpack4(currRec, fd.offs));
	  case FieldDescriptor.tpReal8:
	    return (short)Double.longBitsToDouble(Bytes.unpack8(currRec, fd.offs));
	  case FieldDescriptor.tpString:
	    try { 
		return Short.parseShort(Varying.getString(currRec, fd.offs));
	    } catch(NumberFormatException x) {
		throw new ConsusRuntimeError("Failed to convert from string: " 
					     + x.getMessage());
	    }
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Gets the value of a column in the current row as a Java int.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public int getInt(int columnIndex) throws SQLException {
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return ((Number)tuples[currPos][columnIndex-1]).intValue();
	      case ExprNode.tpStr:
		try { 
		    return Integer.parseInt((String)tuples[currPos][columnIndex-1]);
		} catch(NumberFormatException x) {
		    throw new ConsusRuntimeError("Failed to convert from string: " 
						 + x.getMessage());
		}
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    return currRec[fd.offs];
	  case FieldDescriptor.tpInt2:
	    return Bytes.unpack2(currRec, fd.offs);
	  case FieldDescriptor.tpInt4:
	    return Bytes.unpack4(currRec, fd.offs);
	  case FieldDescriptor.tpInt8:
	    return (int)Bytes.unpack8(currRec, fd.offs);
	  case FieldDescriptor.tpReal4:
	    return (int)Float.intBitsToFloat(Bytes.unpack4(currRec, fd.offs));
	  case FieldDescriptor.tpReal8:
	    return (int)Double.longBitsToDouble(Bytes.unpack8(currRec, fd.offs));
	  case FieldDescriptor.tpString:
	    try { 
		return Integer.parseInt(Varying.getString(currRec, fd.offs));
	    } catch(NumberFormatException x) {
		throw new ConsusRuntimeError("Failed to convert from string: " 
					     + x.getMessage());
	    }
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Gets the value of a column in the current row as a Java long.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public long getLong(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return ((Number)tuples[currPos][columnIndex-1]).longValue();
	      case ExprNode.tpStr:
		try { 
		    return Long.parseLong((String)tuples[currPos][columnIndex-1]);
		} catch(NumberFormatException x) {
		    throw new ConsusRuntimeError("Failed to convert from string: " 
						 + x.getMessage());
		}
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    return currRec[fd.offs];
	  case FieldDescriptor.tpInt2:
	    return Bytes.unpack2(currRec, fd.offs);
	  case FieldDescriptor.tpInt4:
	    return Bytes.unpack4(currRec, fd.offs);
	  case FieldDescriptor.tpInt8:
	    return Bytes.unpack8(currRec, fd.offs);
	  case FieldDescriptor.tpReal4:
	    return (long)Float.intBitsToFloat(Bytes.unpack4(currRec, fd.offs));
	  case FieldDescriptor.tpReal8:
	    return (long)Double.longBitsToDouble(Bytes.unpack8(currRec, fd.offs));
	  case FieldDescriptor.tpString:
	    try { 
		return Long.parseLong(Varying.getString(currRec, fd.offs));
	    } catch(NumberFormatException x) {
		throw new ConsusRuntimeError("Failed to convert from string: " 
					     + x.getMessage());
	    }
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Gets the value of a column in the current row as a Java float.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public float getFloat(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return ((Number)tuples[currPos][columnIndex-1]).floatValue();
	      case ExprNode.tpStr:
		try { 
		    return Float.parseFloat((String)tuples[currPos][columnIndex-1]);
		} catch(NumberFormatException x) {
		    throw new ConsusRuntimeError("Failed to convert from string: " 
						 + x.getMessage());
		}
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    return (float)currRec[fd.offs];
	  case FieldDescriptor.tpInt2:
	    return (float)Bytes.unpack2(currRec, fd.offs);
	  case FieldDescriptor.tpInt4:
	    return (float)Bytes.unpack4(currRec, fd.offs);
	  case FieldDescriptor.tpInt8:
	    return (float)Bytes.unpack8(currRec, fd.offs);
	  case FieldDescriptor.tpReal4:
	    return Float.intBitsToFloat(Bytes.unpack4(currRec, fd.offs));
	  case FieldDescriptor.tpReal8:
	    return (float)Double.longBitsToDouble(Bytes.unpack8(currRec, fd.offs));
	  case FieldDescriptor.tpString:
	    try { 
		return Float.parseFloat(Varying.getString(currRec, fd.offs));
	    } catch(NumberFormatException x) {
		throw new ConsusRuntimeError("Failed to convert from string: " 
					     + x.getMessage());
	    }
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }


    /**
     * Gets the value of a column in the current row as a Java double.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public double getDouble(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return ((Number)tuples[currPos][columnIndex-1]).doubleValue();
	      case ExprNode.tpStr:
		try { 
		    return Double.parseDouble((String)tuples[currPos][columnIndex-1]);
		} catch(NumberFormatException x) {
		    throw new ConsusRuntimeError("Failed to convert from string: " 
						 + x.getMessage());
		}
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return ((Number)tuples[currPos][columnIndex-1]).doubleValue();
	      case ExprNode.tpStr:
		try { 
		    return Double.parseDouble((String)tuples[currPos][columnIndex-1]);
		} catch(NumberFormatException x) {
		    throw new ConsusRuntimeError("Failed to convert from string: " 
						 + x.getMessage());
		}
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    return (double)currRec[fd.offs];
	  case FieldDescriptor.tpInt2:
	    return (double)Bytes.unpack2(currRec, fd.offs);
	  case FieldDescriptor.tpInt4:
	    return (double)Bytes.unpack4(currRec, fd.offs);
	  case FieldDescriptor.tpInt8:
	    return (double)Bytes.unpack8(currRec, fd.offs);
	  case FieldDescriptor.tpReal4:
	    return Float.intBitsToFloat(Bytes.unpack4(currRec, fd.offs));
	  case FieldDescriptor.tpReal8:
	    return Double.longBitsToDouble(Bytes.unpack8(currRec, fd.offs));
	  case FieldDescriptor.tpString:
	    try { 
		return Double.parseDouble(Varying.getString(currRec, fd.offs));
	    } catch(NumberFormatException x) {
		throw new ConsusRuntimeError("Failed to convert from string: " 
					     + x.getMessage());
	    }
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * Gets the value of a column in the current row as a java.math.BigDecimal object.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param scale the number of digits to the right of the decimal
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     * @deprecated
     */
    public BigDecimal getBigDecimal(int columnIndex, int scale) throws SQLException
    {
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
		return BigDecimal.valueOf(((Long)tuples[currPos][columnIndex-1]).longValue(), 
					  scale);
	      case ExprNode.tpReal:
		return new BigDecimal(((Double)tuples[currPos][columnIndex-1]).doubleValue());
	      case ExprNode.tpStr:
		return new BigDecimal((String)tuples[currPos][columnIndex-1]);
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    return BigDecimal.valueOf(currRec[fd.offs], scale);
	  case FieldDescriptor.tpInt2:
	    return BigDecimal.valueOf(Bytes.unpack2(currRec, fd.offs), scale);
	  case FieldDescriptor.tpInt4:
	    return BigDecimal.valueOf(Bytes.unpack4(currRec, fd.offs), scale);
	  case FieldDescriptor.tpInt8:
	    return BigDecimal.valueOf(Bytes.unpack8(currRec, fd.offs), scale);
	  case FieldDescriptor.tpReal4:
	    return new BigDecimal(Float.intBitsToFloat(Bytes.unpack4(currRec, fd.offs)));
	  case FieldDescriptor.tpReal8:
	    return new BigDecimal(Double.longBitsToDouble(Bytes.unpack8(currRec, fd.offs)));
	  case FieldDescriptor.tpString:
	    return new BigDecimal(Varying.getString(currRec, fd.offs));
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }


    /**
     * Gets the value of a column in the current row as a Java byte array.
     * The bytes represent the raw values returned by the driver.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public byte[] getBytes(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpArrayInt1:
		return (byte[])tuples[currPos][columnIndex-1]; 
	      case ExprNode.tpStr:
		return ((String)tuples[currPos][columnIndex-1]).getBytes();
	      case ExprNode.tpBlob:
		{
		    Blob blob = (Blob)tuples[currPos][columnIndex-1];
		    return blob.getBytes(1, (int)blob.length());
		}
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	byte[] result;
	switch (fd.type) { 
	  case FieldDescriptor.tpString:
	    result = new byte[2*Varying.getSize(currRec, fd.offs)];
	    System.arraycopy(currRec, Varying.getOffs(currRec, fd.offs), 
			     result, 0, result.length);
	    break;
	  case FieldDescriptor.tpArray:
	    result = new byte[fd.components[0].size*Varying.getSize(currRec, fd.offs)];
	    System.arraycopy(currRec, Varying.getOffs(currRec, fd.offs), 
			     result, 0, result.length);
	    break;
	  case FieldDescriptor.tpBlob:
	    {
		ConsusBlob blob = new ConsusBlob(db, Bytes.unpack8(currRec, fd.offs), 
						 Bytes.unpack8(currRec, fd.offs + 8));
		return blob.getBytes(1, (int)blob.length());
	    }
	  default:
	    result = new byte[fd.size];
	    System.arraycopy(currRec, fd.offs, result, 0, result.length);
	}    
	return result;
    }
	

    /**
     * Gets the value of a column in the current row as a java.sql.Date object.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Date getDate(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return new java.sql.Date(((Number)tuples[currPos][columnIndex-1]).longValue()); 
	      case ExprNode.tpStr:
		return java.sql.Date.valueOf((String)tuples[currPos][columnIndex-1]);
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	long timestamp;
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    timestamp = currRec[fd.offs];
	    break;
	  case FieldDescriptor.tpInt2:
	    timestamp = Bytes.unpack2(currRec, fd.offs);
	    break;
	  case FieldDescriptor.tpInt4:
	    timestamp = Bytes.unpack4(currRec, fd.offs);
	    break;
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    timestamp = Bytes.unpack8(currRec, fd.offs);
	    break;
	  case FieldDescriptor.tpString:
	    return java.sql.Date.valueOf(Varying.getString(currRec, fd.offs));
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
	return new java.sql.Date(timestamp);
    }

    /**
     * Gets the value of a column in the current row as a java.sql.Time object.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Time getTime(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return new java.sql.Time(((Number)tuples[currPos][columnIndex-1]).longValue()); 
	      case ExprNode.tpStr:
		return java.sql.Time.valueOf((String)tuples[currPos][columnIndex-1]);
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	long timestamp;
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    timestamp = currRec[fd.offs];
	    break;
	  case FieldDescriptor.tpInt2:
	    timestamp = Bytes.unpack2(currRec, fd.offs);
	    break;
	  case FieldDescriptor.tpInt4:
	    timestamp = Bytes.unpack4(currRec, fd.offs);
	    break;
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    timestamp = Bytes.unpack8(currRec, fd.offs);
	    break;
	  case FieldDescriptor.tpString:
	    return java.sql.Time.valueOf(Varying.getString(currRec, fd.offs));
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
	return new java.sql.Time(timestamp);
    }

    /**
     * Gets the value of a column in the current row as a java.sql.Timestamp object.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Timestamp getTimestamp(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (tuples != null) { 
	    switch (tupleColumns[columnIndex-1].type) { 
	      case ExprNode.tpInt:
	      case ExprNode.tpReal:
		return new java.sql.Timestamp(((Number)tuples[currPos][columnIndex-1]).longValue()); 
	      case ExprNode.tpStr:
		return java.sql.Timestamp.valueOf((String)tuples[currPos][columnIndex-1]);
	      default:
		throw new ConsusIncompatibleTypesError();
	    }
	}
	FieldDescriptor fd = columns[columnIndex-1];
	long timestamp;
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	    timestamp = currRec[fd.offs];
	    break;
	  case FieldDescriptor.tpInt2:
	    timestamp = Bytes.unpack2(currRec, fd.offs);
	    break;
	  case FieldDescriptor.tpInt4:
	    timestamp = Bytes.unpack4(currRec, fd.offs);
	    break;
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    timestamp = Bytes.unpack8(currRec, fd.offs);
	    break;
	  case FieldDescriptor.tpString:
	    return java.sql.Timestamp.valueOf(Varying.getString(currRec, fd.offs));
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
	return new java.sql.Timestamp(timestamp);
    }

    /**
     * Gets the value of a column in the current row as a stream of
     * ASCII characters. The value can then be read in chunks from the
     * stream. This method is particularly
     * suitable for retrieving large LONGVARCHAR values.  The JDBC driver will
     * do any necessary conversion from the database format into ASCII.
     *
     * <P><B>Note:</B> All the data in the returned stream must be
     * read prior to getting the value of any other column. The next
     * call to a get method implicitly closes the stream.  Also, a
     * stream may return 0 when the method <code>available</code>
     * is called whether there is data
     * available or not.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return a Java input stream that delivers the database column value
     * as a stream of one byte ASCII characters.  If the value is SQL NULL
     * then the result is null.  
     * @exception SQLException if a database access error occurs
     */
    public java.io.InputStream getAsciiStream(int columnIndex) throws SQLException
    {
	return getBinaryStream(columnIndex);
    }


    /**
     * Gets the value of a column in the current row as a stream of
     * Unicode characters. The value can then be read in chunks from the
     * stream. This method is particularly
     * suitable for retrieving large LONGVARCHAR values.  The JDBC driver will
     * do any necessary conversion from the database format into Unicode.
     * The byte format of the Unicode stream must Java UTF-8,
     * as specified in the Java Virtual Machine Specification.
     *
     * <P><B>Note:</B> All the data in the returned stream must be
     * read prior to getting the value of any other column. The next
     * call to a get method implicitly closes the stream. Also, a
     * stream may return 0 when the method <code>available</code>
     * is called whether there is data
     * available or not.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return a Java input stream that delivers the database column value
     * as a stream of two-byte Unicode characters.  If the value is SQL NULL
     * then the result is null.  
     * @exception SQLException if a database access error occurs
     * @deprecated
     */
    public java.io.InputStream getUnicodeStream(int columnIndex) throws SQLException
    {
	return getBinaryStream(columnIndex);
    }

    /**
     * Gets the value of a column in the current row as a stream of
     * uninterpreted bytes. The value can then be read in chunks from the
     * stream. This method is particularly
     * suitable for retrieving large LONGVARBINARY values.
     *
     * <P><B>Note:</B> All the data in the returned stream must be
     * read prior to getting the value of any other column. The next
     * call to a get method implicitly closes the stream. Also, a
     * stream may return 0 when the method <code>available</code>
     * is called whether there is data
     * available or not.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return a Java input stream that delivers the database column value
     * as a stream of uninterpreted bytes.  If the value is SQL NULL
     * then the result is null.  
     * @exception SQLException if a database access error occurs
     */
    public java.io.InputStream getBinaryStream(int columnIndex)
        throws SQLException
    {
	checkColumnIndex(columnIndex);
	switch (tuples != null ? tupleColumns[columnIndex-1].type : columns[columnIndex-1].type) { 
	  case FieldDescriptor.tpBlob: 
	    return getBlob(columnIndex).getBinaryStream();
	  case FieldDescriptor.tpClob: 
	    return getClob(columnIndex).getAsciiStream();
	  default:
	    byte[] bytes = getBytes(columnIndex);
	    if (bytes == null) { 
		bytes = new byte[0];
	    }
	    return new java.io.ByteArrayInputStream(bytes);
	}
    }


    //======================================================================
    // Methods for accessing results by column name
    //======================================================================

    /**
     * Gets the value of a column in the current row as a Java String.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public String getString(String columnName) throws SQLException { 
	return getString(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java boolean.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is false
     * @exception SQLException if a database access error occurs
     */
    public boolean getBoolean(String columnName) throws SQLException { 
	return getBoolean(findColumn(columnName));
    }


    /**
     * Gets the value of a column in the current row as a Java byte.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public byte getByte(String columnName) throws SQLException { 
	return getByte(findColumn(columnName));
    }


    /**
     * Gets the value of a column in the current row as a Java short.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public short getShort(String columnName) throws SQLException { 
	return getShort(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java int.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public int getInt(String columnName) throws SQLException { 
	return getInt(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java long.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public long getLong(String columnName) throws SQLException { 
	return getLong(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java float.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public float getFloat(String columnName) throws SQLException { 
	return getFloat(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a Java double.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is 0
     * @exception SQLException if a database access error occurs
     */
    public double getDouble(String columnName) throws SQLException { 
	return getDouble(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a java.math.BigDecimal 
     * object.
     *
     * @param columnName the SQL name of the column
     * @param scale the number of digits to the right of the decimal
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     * @deprecated
     */
    public BigDecimal getBigDecimal(String columnName, int scale) throws SQLException { 
	return getBigDecimal(findColumn(columnName), scale);
    }

    /**
     * Gets the value of a column in the current row as a Java byte array.
     * The bytes represent the raw values returned by the driver.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public byte[] getBytes(String columnName) throws SQLException { 
	return getBytes(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a java.sql.Date object.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Date getDate(String columnName) throws SQLException { 
	return getDate(findColumn(columnName));
    }
	

    /**
     * Gets the value of a column in the current row as a java.sql.Time object.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Time getTime(String columnName) throws SQLException { 
	return getTime(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a java.sql.Timestamp object.
     *
     * @param columnName the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Timestamp getTimestamp(String columnName) throws SQLException { 
	return getTimestamp(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a stream of
     * ASCII characters. The value can then be read in chunks from the
     * stream. This method is particularly
     * suitable for retrieving large LONGVARCHAR values.  The JDBC driver will
     * do any necessary conversion from the database format into ASCII.
     *
     * <P><B>Note:</B> All the data in the returned stream must be
     * read prior to getting the value of any other column. The next
     * call to a get method implicitly closes the stream. Also, a
     * stream may return 0 when the method <code>available</code>
     * is called whether there is data
     * available or not.
     *
     * @param columnName the SQL name of the column
     * @return a Java input stream that delivers the database column value
     * as a stream of one byte ASCII characters.  If the value is SQL NULL
     * then the result is null.
     * @exception SQLException if a database access error occurs
     */
    public java.io.InputStream getAsciiStream(String columnName) throws SQLException { 
	return getAsciiStream(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a stream of
     * Unicode characters. The value can then be read in chunks from the
     * stream. This method is particularly
     * suitable for retrieving large LONGVARCHAR values.  The JDBC driver will
     * do any necessary conversion from the database format into Unicode.
     * The byte format of the Unicode stream must be Java UTF-8,
     * as defined in the Java Virtual Machine Specification.
     *
     * <P><B>Note:</B> All the data in the returned stream must be
     * read prior to getting the value of any other column. The next
     * call to a get method implicitly closes the stream. Also, a
     * stream may return 0 when the method <code>available</code>
     * is called whether there is data
     * available or not.
     *
     * @param columnName the SQL name of the column
     * @return a Java input stream that delivers the database column value
     * as a stream of two-byte Unicode characters.  If the value is SQL NULL
     * then the result is null.
     * @exception SQLException if a database access error occurs
     * @deprecated
     */
    public java.io.InputStream getUnicodeStream(String columnName) throws SQLException { 
	return getUnicodeStream(findColumn(columnName));
    }

    /**
     * Gets the value of a column in the current row as a stream of
     * uninterpreted bytes. The value can then be read in chunks from the
     * stream. This method is particularly
     * suitable for retrieving large LONGVARBINARY values.  The JDBC driver will
     * do any necessary conversion from the database format into uninterpreted
     * bytes.
     *
     * <P><B>Note:</B> All the data in the returned stream must be
     * read prior to getting the value of any other column. The next
     * call to a get method implicitly closes the stream. Also, a
     * stream may return 0 when the method <code>available</code>
     * is called whether there is data
     * available or not.
     *
     * @param columnName the SQL name of the column
     * @return a Java input stream that delivers the database column value
     * as a stream of uninterpreted bytes.  If the value is SQL NULL
     * then the result is null.
     * @exception SQLException if a database access error occurs
     */
    public java.io.InputStream getBinaryStream(String columnName) throws SQLException { 
	return getUnicodeStream(findColumn(columnName));
    }


    //=====================================================================
    // Advanced features:
    //=====================================================================

    /**
     * <p>The first warning reported by calls on this ResultSet is
     * returned. Subsequent ResultSet warnings will be chained to this
     * SQLWarning.
     *
     * <P>The warning chain is automatically cleared each time a new
     * row is read.
     *
     * <P><B>Note:</B> This warning chain only covers warnings caused
     * by ResultSet methods.  Any warning caused by statement methods
     * (such as reading OUT parameters) will be chained on the
     * Statement object. 
     *
     * @return the first SQLWarning or null 
     * @exception SQLException if a database access error occurs
     */
    public SQLWarning getWarnings() throws SQLException
    {
        return null;
    }

    /**
     * After this call getWarnings returns null until a new warning is
     * reported for this ResultSet.  
     *
     * @exception SQLException if a database access error occurs
     */
    public void clearWarnings() throws SQLException {}

    /**
     * Gets the name of the SQL cursor used by this ResultSet.
     *
     * <P>In SQL, a result table is retrieved through a cursor that is
     * named. The current row of a result can be updated or deleted
     * using a positioned update/delete statement that references the
     * cursor name. To insure that the cursor has the proper isolation
     * level to support update, the cursor's select statement should be 
     * of the form 'select for update'. If the 'for update' clause is 
     * omitted the positioned updates may fail.
     * 
     * <P>JDBC supports this SQL feature by providing the name of the
     * SQL cursor used by a ResultSet. The current row of a ResultSet
     * is also the current row of this SQL cursor.
     *
     * <P><B>Note:</B> If positioned update is not supported a
     * SQLException is thrown
     *
     * @return the ResultSet's SQL cursor name
     * @exception SQLException if a database access error occurs
     */
    public String getCursorName() throws SQLException { 
	return null;
    }

    /**
     * Retrieves the  number, types and properties of a ResultSet's columns.
     *
     * @return the description of a ResultSet's columns
     * @exception SQLException if a database access error occurs
     */
    public ResultSetMetaData getMetaData() throws SQLException
    {
	return tupleColumns != null 
	    ? new ConsusResultSetMetaData(tupleColumns)
	    : new ConsusResultSetMetaData(table);
    }

    /**
     * <p>Gets the value of a column in the current row as a Java object.
     *
     * <p>This method will return the value of the given column as a
     * Java object.  The type of the Java object will be the default
     * Java object type corresponding to the column's SQL type,
     * following the mapping for built-in types specified in the JDBC 
     * spec.
     *
     * <p>This method may also be used to read datatabase-specific
     * abstract data types.
     *
     * JDBC 2.0
     *
     * 
     * In the JDBC 2.0 API, the behavior of method
     * <code>getObject</code> is extended to materialize  
     * data of SQL user-defined types.  When the a column contains
     * a structured or distinct value, the behavior of this method is as 
     * if it were a call to: getObject(columnIndex, 
     * this.getStatement().getConnection().getTypeMap()).
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * <B>CONSUS specific:</B> zero value of the parameter refers 
     *                         to the current selected object itself
     * @return a java.lang.Object holding the column value  
     * @exception SQLException if a database access error occurs
     */
    public Object getObject(int columnIndex) throws SQLException { 
	checkColumnIndex0(columnIndex);
	if (tuples != null) { 
	    return tuples[currPos][columnIndex-1];
	}
	if (columnIndex == 0) {
	    return getSelfObject();
	} else { 
	    return columns[columnIndex-1].fetch(currRec, 0, true);
	}
    }

    /**
     * <p>Gets the value of a column in the current row as a Java object.
     *
     * <p>This method will return the value of the given column as a
     * Java object.  The type of the Java object will be the default
     * Java object type corresponding to the column's SQL type,
     * following the mapping for built-in types specified in the JDBC 
     * spec.
     *
     * <p>This method may also be used to read datatabase-specific
     * abstract data types.
     *
     * JDBC 2.0
     *
     * In the JDBC 2.0 API, the behavior of method
     * <code>getObject</code> is extended to materialize  
     * data of SQL user-defined types.  When the a column contains
     * a structured or distinct value, the behavior of this method is as 
     * if it were a call to: getObject(columnIndex, 
     * this.getStatement().getConnection().getTypeMap()).
     *
     * @param columnName the SQL name of the column
     * <B>CONSUS specific:</B> "this" value of the parameter refers to 
     *                         to the current selected object itself
     * @return a java.lang.Object holding the column value.  
     * @exception SQLException if a database access error occurs
     */
    public Object getObject(String columnName) throws SQLException { 
	return getObject(findColumn(columnName));
    }

    //----------------------------------------------------------------

    /**
     * Maps the given Resultset column name to its ResultSet column index.
     *
     * @param columnName the name of the column
     * @return the column index
     * @exception SQLException if a database access error occurs
     */
    public int findColumn(String columnName) throws SQLException { 
	if (tuples != null) { 
	    if (columnName.indexOf('.') != -1) { 
		for (int i = tupleColumns.length; --i >= 0;) { 
		    ColumnNode col = tupleColumns[i];
		    if (columnName.equals(col.name)
			|| (col.field != null 
			    && columnName.equals(col.field.table.name + '.' + tupleColumns[i].name)))
		    {
			return i + 1;
		    }
		}
	    } else { 
		for (int i = tupleColumns.length; --i >= 0;) { 
		    ColumnNode col = tupleColumns[i];
		    if (columnName.equals(col.name)) {
			return i + 1;
		    }
		}
	    }
	    throw new ConsusColumnNotFoundError(columnName);
	}
	FieldDescriptor fd = table.getColumn(columnName);
	if (fd == null) {
	    if (columnName.equals("this")) { 
		return 0;
	    }
	    throw new ConsusColumnNotFoundError(columnName);
	}
	return fd.columnIndex + 1;
    }

    //--------------------------JDBC 2.0-----------------------------------

    //---------------------------------------------------------------------
    // Getter's and Setter's
    //---------------------------------------------------------------------

    /**
     * JDBC 2.0
     *
     * <p>Gets the value of a column in the current row as a java.io.Reader.
     * @param columnIndex the first column is 1, the second is 2, ...
     */
    public java.io.Reader getCharacterStream(int columnIndex) throws SQLException
    {
	checkColumnIndex(columnIndex);
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpClob: 
	    return getClob(columnIndex).getCharacterStream();
	  default: 
	    return new java.io.StringReader(getString(columnIndex));
	}
    }


    /**
     * JDBC 2.0
     *
     * <p>Gets the value of a column in the current row as a java.io.Reader.
     * @param columnName the name of the column
     * @return the value in the specified column as a <code>java.io.Reader</code>
     */
    public java.io.Reader getCharacterStream(String columnName) throws SQLException
    {
	return getCharacterStream(findColumn(columnName));
    }


    /**
     * JDBC 2.0
     *
     * Gets the value of a column in the current row as a java.math.BigDecimal 
     * object with full precision.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value (full precision); if the value is SQL NULL, 
     * the result is null
     * @exception SQLException if a database access error occurs
     */
    public BigDecimal getBigDecimal(int columnIndex) throws SQLException
    {
	return getBigDecimal(columnIndex, 0);
    }


    /**
     * JDBC 2.0
     *
     * Gets the value of a column in the current row as a java.math.BigDecimal 
     * object with full precision.
     * @param columnName the column name
     * @return the column value (full precision); if the value is SQL NULL, 
     * the result is null
     * @exception SQLException if a database access error occurs
     *
     */
    public BigDecimal getBigDecimal(String columnName) throws SQLException {
	return getBigDecimal(findColumn(columnName));
    }


    //---------------------------------------------------------------------
    // Traversal/Positioning
    //---------------------------------------------------------------------

    /**
     * JDBC 2.0
     *
     * <p>Indicates whether the cursor is before the first row in the result 
     * set.   
     *
     * @return true if the cursor is before the first row, false otherwise. Returns
     * false when the result set contains no rows.
     * @exception SQLException if a database access error occurs
     */
    public boolean isBeforeFirst() throws SQLException { 
	return currPos < 0;
    }
      
    /**
     * JDBC 2.0
     *
     * <p>Indicates whether the cursor is after the last row in the result 
     * set.   
     *
     * @return true if the cursor is  after the last row, false otherwise.  Returns
     * false when the result set contains no rows.
     * @exception SQLException if a database access error occurs
     */
    public boolean isAfterLast() throws SQLException {
	return currPos >= nSelected;
    }
 
    /**
     * JDBC 2.0
     *
     * <p>Indicates whether the cursor is on the first row of the result set.   
     *
     * @return true if the cursor is on the first row, false otherwise.   
     * @exception SQLException if a database access error occurs
     */
    public boolean isFirst() throws SQLException { 
	return nSelected != 0 && currPos == 0;
    }
 
    /**
     * JDBC 2.0
     *
     * <p>Indicates whether the cursor is on the last row of the result set.   
     * @return true if the cursor is on the last row, false otherwise. 
     * @exception SQLException if a database access error occurs
     */
    public boolean isLast() throws SQLException { 
	return nSelected != 0 && currPos == nSelected-1;
    }

    /**
     * JDBC 2.0
     *
     * <p>Moves the cursor to the front of the result set, just before the
     * first row. Has no effect if the result set contains no rows.
     *
     * @exception SQLException if a database access error occurs or the 
     * result set type is TYPE_FORWARD_ONLY
     */
    public void beforeFirst() throws SQLException { 
        currPos = -1;
    }
    

    /**
     * JDBC 2.0
     *
     * <p>Moves the cursor to the end of the result set, just after the last
     * row.  Has no effect if the result set contains no rows.
     *
     * @exception SQLException if a database access error occurs or the 
     * result set type is TYPE_FORWARD_ONLY 
     */
    public void afterLast() throws SQLException { 
	currPos = nSelected;
    }
	
    /**
     * JDBC 2.0
     *
     * <p>Moves the cursor to the first row in the result set.  
     *
     * @return true if the cursor is on a valid row; false if
     *         there are no rows in the result set
     * @exception SQLException if a database access error occurs or the 
     * result set type is TYPE_FORWARD_ONLY
     */
    public boolean first() throws SQLException { 
	if (nSelected > 0) { 
	    currPos = 0;
	    fetch();
	    return true;
	}
	return false;
    }

    /**
     * JDBC 2.0
     *
     * <p>Moves the cursor to the last row in the result set.  
     *
     * @return true if the cursor is on a valid row;
     * false if there are no rows in the result set
     * @exception SQLException if a database access error occurs or the
     * result set type is TYPE_FORWARD_ONLY.
     */
    public boolean last() throws SQLException { 
	if (nSelected > 0) { 
	    currPos = nSelected - 1;
	    fetch();
	    return true;
	}
	return false;
    }
    /**
     * JDBC 2.0
     *
     * <p>Retrieves the current row number.  The first row is number 1, the
     * second number 2, and so on.  
     *
     * @return the current row number; 0 if there is no current row
     * @exception SQLException if a database access error occurs
     */
    public int getRow() throws SQLException { 
	return currPos + 1;
    }

    /**
     * JDBC 2.0
     *
     * <p>Moves the cursor to the given row number in the result set.
     *
     * <p>If the row number is positive, the cursor moves to 
     * the given row number with respect to the
     * beginning of the result set.  The first row is row 1, the second
     * is row 2, and so on. 
     *
     * <p>If the given row number is negative, the cursor moves to
     * an absolute row position with respect to
     * the end of the result set.  For example, calling
     * <code>absolute(-1)</code> positions the 
     * cursor on the last row, <code>absolute(-2)</code> indicates the next-to-last
     * row, and so on.
     *
     * <p>An attempt to position the cursor beyond the first/last row in
     * the result set leaves the cursor before/after the first/last
     * row, respectively.
     *
     * <p>Note: Calling <code>absolute(1)</code> is the same
     * as calling <code>first()</code>.
     * Calling <code>absolute(-1)</code> is the same as calling <code>last()</code>.
     *
     * @return true if the cursor is on the result set; false otherwise
     * @exception SQLException if a database access error occurs or 
     * row is 0, or result set type is TYPE_FORWARD_ONLY.
     */
    public boolean absolute( int row ) throws SQLException { 
	if (row == -1) { 
	    return last();
	}	    
	if (row <= 0) { 
	    throw new ConsusRuntimeError("Row index should be positive number");
	}
	row -= 1;
	if (row >= nSelected) { 
	    return false;
	}
	currPos = row;
	fetch();
	return true;
    }

    /**
     * JDBC 2.0
     *
     * <p>Moves the cursor a relative number of rows, either positive or negative.
     * Attempting to move beyond the first/last row in the
     * result set positions the cursor before/after the
     * the first/last row. Calling <code>relative(0)</code> is valid, but does
     * not change the cursor position.
     *
     * <p>Note: Calling <code>relative(1)</code>
     * is different from calling <code>next()</code>
     * because is makes sense to call <code>next()</code> when there is no current row,
     * for example, when the cursor is positioned before the first row
     * or after the last row of the result set.
     *
     * @return true if the cursor is on a row; false otherwise
     * @exception SQLException if a database access error occurs, there
     * is no current row, or the result set type is TYPE_FORWARD_ONLY
     */
    public boolean relative( int rows ) throws SQLException {
	return absolute(currPos + rows);
    }

    /**
     * JDBC 2.0
     *
     * <p>Moves the cursor to the previous row in the result set.  
     *
     * <p>Note: <code>previous()</code> is not the same as
     * <code>relative(-1)</code> because it
     * makes sense to call</code>previous()</code> when there is no current row.
     *
     * @return true if the cursor is on a valid row; false if it is off the result set
     * @exception SQLException if a database access error occurs or the
     * result set type is TYPE_FORWARD_ONLY
     */
    public boolean previous() throws SQLException { 
	 if (currPos <= 0) { 
	    return false;
	 }
	 currPos -= 1;
	 fetch();
	 return true;
    }


    //---------------------------------------------------------------------
    // Properties
    //---------------------------------------------------------------------

    /**
     * Gives a hint as to the direction in which the rows in this
	 * <code>ResultSet</code> object will be processed. 
	 * The initial value is determined by the 
	 * <code>Statement</code> object
     * that produced this <code>ResultSet</code> object.
	 * The fetch direction may be changed at any time.
     *
     * @exception SQLException if a database access error occurs or
     * the result set type is <code>TYPE_FORWARD_ONLY</code> and the fetch
	 * direction is not <code>FETCH_FORWARD</code>
     * @since 1.2
     * @see <a href="package-summary.html#2.0 API">What Is in the JDBC
     *      2.0 API</a>
	 * @see Statement#setFetchDirection
     */
    public void setFetchDirection(int direction) throws SQLException {}

    /**
     * JDBC 2.0
     *
     * Returns the fetch direction for this result set.
     *
     * @return the current fetch direction for this result set
     * @exception SQLException if a database access error occurs
     */
    public int getFetchDirection() throws SQLException { 
	return FETCH_UNKNOWN;
    }

    /**
     * JDBC 2.0
     *
     * @param rows the number of rows to fetch
     * @exception SQLException if a database access error occurs or the
     * condition 0 <= rows <= this.getMaxRows() is not satisfied.
     */
    public void setFetchSize(int rows) throws SQLException {}

    /**
     * JDBC 2.0
     *
     * Returns the fetch size for this result set.
     *
     * @return the current fetch size for this result set
     * @exception SQLException if a database access error occurs
     */
    public int getFetchSize() throws SQLException { 
	return 1;
    }

    /**
     * JDBC 2.0
     *
     * Returns the type of this result set.  The type is determined by
     * the statement that created the result set.
     *
     * @return TYPE_FORWARD_ONLY, TYPE_SCROLL_INSENSITIVE, or
     * TYPE_SCROLL_SENSITIVE
     * @exception SQLException if a database access error occurs
     */
    public int getType() throws SQLException
    {
        return resultSetType;
    }

    /**
     * JDBC 2.0
     *
     * Returns the concurrency mode of this result set.  The concurrency
     * used is determined by the statement that created the result set.
     *
     * @return the concurrency type, CONCUR_READ_ONLY or CONCUR_UPDATABLE
     * @exception SQLException if a database access error occurs
     */
    public int getConcurrency() throws SQLException
    {
        return resultSetConcurrency;
    }

    //---------------------------------------------------------------------
    // Updates
    //---------------------------------------------------------------------

    /**
     * JDBC 2.0
     *
     * Indicates whether the current row has been updated.  The value returned 
     * depends on whether or not the result set can detect updates.
     *
     * @return true if the row has been visibly updated by the owner or
     * another, and updates are detected
     * @exception SQLException if a database access error occurs
     * 
     * @see DatabaseMetaData#updatesAreDetected
     */
    public boolean rowUpdated() throws SQLException { 
	if (oid != 0) { 
	    return (db.getPos(oid) & ConsusConnection.dbModifiedFlag) != 0;
	}
	return false;
    }

    /**
     * JDBC 2.0
     *
     * Indicates whether the current row has had an insertion.  The value returned 
     * depends on whether or not the result set can detect visible inserts.
     *
     * @return true if a row has had an insertion and insertions are detected
     * @exception SQLException if a database access error occurs
     * 
     * @see DatabaseMetaData#insertsAreDetected
     */
    public boolean rowInserted() throws SQLException { 
	return cursorState == CS_INSERTED_ROW;
    }
   
    /**
     * JDBC 2.0
     *
     * Indicates whether a row has been deleted.  A deleted row may leave
     * a visible "hole" in a result set.  This method can be used to
     * detect holes in a result set.  The value returned depends on whether 
     * or not the result set can detect deletions.
     *
     * @return true if a row was deleted and deletions are detected
     * @exception SQLException if a database access error occurs
     * 
     * @see DatabaseMetaData#deletesAreDetected
     */
    public boolean rowDeleted() throws SQLException { 
	return cursorState == CS_DELETED_ROW;
    }

    final void updateColumn(FieldDescriptor fd, Object val) { 
	stmt.updated[fd.id] = true;
	stmt.columnValue[fd.id] = val;
	updateColumns = true;
    }

    /**
     * JDBC 2.0
     * 
     * Give a nullable column a null value.
     * 
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @exception SQLException if a database access error occurs
     */
    public void updateNull(int columnIndex) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	if (fd.type != FieldDescriptor.tpReference) { 
	    throw new ConsusIncompatibleTypesError();
	}
	obj = null; // prevent usage of object instance for update
	updateColumn(fd, null);
    }

    /**
     * JDBC 2.0
     * 
     * Updates a column with a boolean value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateBoolean(int columnIndex, boolean x) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	if (fd.type != FieldDescriptor.tpBool) { 
	    throw new ConsusIncompatibleTypesError();
	}
	obj = null; // prevent usage of object instance for update
	updateColumn(fd, new Boolean(x));
    }

    /**
     * JDBC 2.0
     *   
     * Updates a column with a byte value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateByte(int columnIndex, byte x) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long(x));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double((double)x));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * JDBC 2.0
     *   
     * Updates a column with a short value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateShort(int columnIndex, short x) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long(x));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double((double)x));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }
    /**
     * JDBC 2.0
     *   
     * Updates a column with an integer value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateInt(int columnIndex, int x) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long(x));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double((double)x));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * JDBC 2.0
     *   
     * Updates a column with a long value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateLong(int columnIndex, long x) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long(x));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double((double)x));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a float value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateFloat(int columnIndex, float x) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long((long)x));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double(x));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a Double value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
     public void updateDouble(int columnIndex, double x) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long((long)x));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double(x));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a BigDecimal value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateBigDecimal(int columnIndex, BigDecimal x) throws SQLException
    {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long(x.longValue()));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double(x.doubleValue()));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a String value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateString(int columnIndex, String x) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	if (fd.type != FieldDescriptor.tpString) { 
	    throw new ConsusIncompatibleTypesError();
	}
	updateColumn(fd, x);
    }


    /**
     * JDBC 2.0
     *  
     * Updates a column with a byte array value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateBytes(int columnIndex, byte x[]) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	if (fd.type != FieldDescriptor.tpArray || fd.components[0].type != FieldDescriptor.tpInt1) { 
	    throw new ConsusIncompatibleTypesError();
	}
	updateColumn(fd, x);
    }


    /**
     * JDBC 2.0
     *  
     * Updates a column with a Date value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateDate(int columnIndex, java.sql.Date x) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long(x.getTime()));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double((double)x.getTime()));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }
	
    /**
     * JDBC 2.0
     *  
     * Updates a column with a Time value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateTime(int columnIndex, java.sql.Time x) throws SQLException {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}	
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long(x.getTime()));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double((double)x.getTime()));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a Timestamp value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateTimestamp(int columnIndex, java.sql.Timestamp x) throws SQLException { 
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpInt1:
	  case FieldDescriptor.tpInt2:
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	    updateColumn(fd, new Long(x.getTime()));
	    break;
	  case FieldDescriptor.tpReal4:
	  case FieldDescriptor.tpReal8:
	    updateColumn(fd, new Double((double)x.getTime()));
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /** 
     * JDBC 2.0
     *  
     * Updates a column with an ascii stream value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @param length the length of the stream
     * @exception SQLException if a database access error occurs
     */
    public void updateAsciiStream(int columnIndex, 
				  java.io.InputStream x, 
				  int length) throws SQLException
    {
	updateBinaryStream(columnIndex, x, length);
    }

 
    /** 
     * JDBC 2.0
     *  
     * Updates a column with a binary stream value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value     
     * @param length the length of the stream
     * @exception SQLException if a database access error occurs
     */
    public void updateBinaryStream(int columnIndex, 
				   java.io.InputStream stream,
				   int length) throws SQLException
    {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpBlob:
	    updateColumn(fd, new ConsusBlob(length, stream));
	    break;
	  case FieldDescriptor.tpString:
	    try { 
		byte[] buf = new byte[length];
		if (stream.read(buf, 0, length) != length) { 
		    throw new ConsusRuntimeError("Failed to read all bytes from stream");
		}
		updateColumn(fd, new String(buf));
	    } catch (java.io.IOException x) { 
		throw new ConsusRuntimeError("Failed to read from stream: " + 
					     x.getMessage());
	    }
	    break;
	  case FieldDescriptor.tpArray:
	    if (fd.components[0].type == FieldDescriptor.tpInt1) { 
		try { 
		    byte[] buf = new byte[length];
		    if (stream.read(buf, 0, length) != length) { 
			throw new ConsusRuntimeError("Failed to read all bytes from stream");
		    }
		    updateColumn(fd, buf);
		} catch (java.io.IOException x) { 
		    throw new ConsusRuntimeError("Failed to read from stream: " + 
					     x.getMessage());
		}
		break;
	    } 
	    // no break
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    public void updateBinaryStream(int columnIndex, 
				   java.io.InputStream stream,
				   long length) throws SQLException
    {
        updateBinaryStream(columnIndex, stream, (int)length);
    }

    public void updateBinaryStream(String columnLabel, 
				   java.io.InputStream stream,
				   long length) throws SQLException
    {
        updateBinaryStream(columnLabel, stream, (int)length);
    }

    public void updateAsciiStream(int columnIndex, 
				   java.io.InputStream stream,
				   long length) throws SQLException
    {
        updateBinaryStream(columnIndex, stream, (int)length);
    }

    public void updateAsciiStream(String columnLabel, 
				   java.io.InputStream stream,
				   long length) throws SQLException
    {
         updateBinaryStream(columnLabel, stream, (int)length);
    }

     /**
     * JDBC 2.0
     *  
     * Updates a column with a character stream value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @param length the length of the stream
     * @exception SQLException if a database access error occurs
     */
    public void updateCharacterStream(int columnIndex,
				      java.io.Reader stream,
				      int length) throws SQLException
    {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpClob:
	    updateColumn(fd, new ConsusClob(length, stream));
	    break;
	  case FieldDescriptor.tpString:
	    try { 
		char[] buf = new char[length];
		if (stream.read(buf, 0, length) != length) { 
		    throw new ConsusRuntimeError("Failed to read all characters from stream");
		}
		updateColumn(fd, new String(buf));
	    } catch (java.io.IOException x) { 
		throw new ConsusRuntimeError("Failed to read from stream: " + 
					     x.getMessage());
	    }
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    public void updateCharacterStream(int columnIndex,
				      java.io.Reader stream,
				      long length) throws SQLException
    {
        updateCharacterStream(columnIndex, stream, (int)length);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with an Object value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @param scale For java.sql.Types.DECIMAL or java.sql.Types.NUMERIC types
     *  this is the number of digits after the decimal.  For all other
     *  types this value will be ignored.
     * @exception SQLException if a database access error occurs
     */
    public void updateObject(int columnIndex, Object x, int scale)
      throws SQLException
    {
        throw new ConsusNotImplemented("ResultSet.updateObject(scale)");
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with an Object value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * <B>CONSUS specific:</B> zero value of the parameter refers 
     * to the current selected object itself. In case of the updated record, 
     * values of the updated record fields are replaced with the values of 
     * specified object fields. In case of the inserted record, specified 
     * object is inserted into the table.
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateObject(int columnIndex, Object x) throws SQLException {
	checkColumnIndex0(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	if (columnIndex == 0) { 	    
	    if (updateColumns) { 
		throw new ConsusRuntimeError("Mixed row-level and object updates are not possible");
	    }
	    obj = x;
	} else { 
	    FieldDescriptor fd = columns[columnIndex-1];
	    switch (fd.type) { 
	      case FieldDescriptor.tpBool:
		if (x instanceof Number) { 
		    updateColumn(fd, new Boolean(((Number)x).longValue() != 0));
		} else if (x instanceof Boolean) { 
		    updateColumn(fd, x);
		} else if (x instanceof String) { 
		    updateColumn(fd, new Boolean(((String)x).equalsIgnoreCase("true")));
		} else { 
		    throw new ConsusIncompatibleTypesError(x);
		}
		break;
	      case FieldDescriptor.tpInt1:
	      case FieldDescriptor.tpInt2:
	      case FieldDescriptor.tpInt4:
	      case FieldDescriptor.tpInt8:
	      case FieldDescriptor.tpDate:
	      case FieldDescriptor.tpTime:
	      case FieldDescriptor.tpTimestamp:
		if (x instanceof Long) { 
		    updateColumn(fd, x);
		} else if (x instanceof Boolean) { 
		    updateColumn(fd, new Long(((Boolean)x).booleanValue() ? 1 : 0));
		} else if (x instanceof String) { 
		    try { 
			String str = (String)x;
			switch (fd.type) { 
			  case FieldDescriptor.tpDate:
			    updateColumn(fd, new Long(Date.valueOf(str).getTime()));
			  case FieldDescriptor.tpTime:
			    updateColumn(fd, new Long(Time.valueOf(str).getTime()));
			  case FieldDescriptor.tpTimestamp:
			    updateColumn(fd, new Long(Timestamp.valueOf(str).getTime()));
			  default:
			    updateColumn(fd, new Long(str));
			}
		    } catch(Exception ex) { 
			throw new ConsusIncompatibleTypesError(x);
		    }
		} else if (x instanceof Number) { 
		    updateColumn(fd, new Long(((Number)x).longValue()));
		} else if (x instanceof java.util.Date) { 
		    updateColumn(fd, new Long(((java.util.Date)x).getTime()));
		} else { 
		    throw new ConsusIncompatibleTypesError(x);
		}
		break;
	      case FieldDescriptor.tpReal4:
	      case FieldDescriptor.tpReal8:
		if (x instanceof Double) { 
		    updateColumn(fd, x);
		} else if (x instanceof Boolean) { 
		    updateColumn(fd, new Double(((Boolean)x).booleanValue() ? 1 : 0));
		} else if (x instanceof String) { 
		    try { 
			updateColumn(fd, new Double((String)x));
		    } catch(Exception ex) { 
			throw new ConsusIncompatibleTypesError(x);
		    }
		} else if (x instanceof Number) { 
		    updateColumn(fd, new Double(((Number)x).doubleValue()));
		} else if (x instanceof java.util.Date) { 
		    updateColumn(fd, new Double((double)((java.util.Date)x).getTime()));
		} else { 
		    throw new ConsusIncompatibleTypesError(x);
		}
		break;
	      case FieldDescriptor.tpReference:
		if (x instanceof ConsusRef) { 
		    updateColumn(fd, x);
		} else if (x instanceof Persistent) { 
		    updateColumn(fd, x);
		} else if (x == null) { 
		    updateColumn(fd, null);
		} else { 
		    throw new ConsusIncompatibleTypesError(x);
		}
		break;
	      case FieldDescriptor.tpString:
		updateColumn(fd, x != null ? String.valueOf(x) : null);
		break;
	      case FieldDescriptor.tpArray:
		if (fd.components[0].type == FieldDescriptor.tpInt1
		    && (x instanceof byte[]))
		{
		    updateColumn(fd, x);
		} else { 
		    throw new ConsusIncompatibleTypesError(x);
		}
		break;
	      case FieldDescriptor.tpBlob:
		if (x instanceof Blob) { 
		    updateColumn(fd, x);
		} else { 
		    throw new ConsusIncompatibleTypesError(x);
		}
		break;		    
	      case FieldDescriptor.tpClob:
		if (x instanceof Clob) { 
		    updateColumn(fd, x);
		} else { 
		    throw new ConsusIncompatibleTypesError(x);
		}
		break;		    
	      default:
		throw new ConsusIncompatibleTypesError(x);
	    }
	}
    }

	
    /**
     * JDBC 2.0
     *  
     * Updates a column with a null value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @exception SQLException if a database access error occurs
     */
    public void updateNull(String columnName) throws SQLException { 
	updateNull(findColumn(columnName));
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a boolean value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateBoolean(String columnName, boolean x) throws SQLException { 
	updateBoolean(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a byte value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateByte(String columnName, byte x) throws SQLException { 
	updateByte(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a short value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateShort(String columnName, short x) throws SQLException { 
	updateShort(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with an integer value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateInt(String columnName, int x) throws SQLException { 
	updateInt(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a long value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateLong(String columnName, long x) throws SQLException { 
	updateLong(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a float value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateFloat(String columnName, float x) throws SQLException { 
	updateFloat(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a double value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateDouble(String columnName, double x) throws SQLException {
	updateDouble(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a BigDecimal value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateBigDecimal(String columnName, BigDecimal x) throws SQLException {
	updateBigDecimal(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a String value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateString(String columnName, String x) throws SQLException {
	updateString(findColumn(columnName), x);
    }
	
    /**
     * JDBC 2.0
     *  
     * Updates a column with a byte array value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateBytes(String columnName, byte x[]) throws SQLException {
	updateBytes(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a Date value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateDate(String columnName, java.sql.Date x) throws SQLException {
	updateDate(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a Time value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateTime(String columnName, java.sql.Time x) throws SQLException { 
	updateTime(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a Timestamp value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateTimestamp(String columnName, java.sql.Timestamp x) throws SQLException {
	updateTimestamp(findColumn(columnName), x);
    }

    /** 
     * JDBC 2.0
     *  
     * Updates a column with an ascii stream value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @param length of the stream
     * @exception SQLException if a database access error occurs
     */
    public void updateAsciiStream(String columnName, 
				  java.io.InputStream x, 
				  int length) throws SQLException
    {
	updateAsciiStream(findColumn(columnName), x, length);
    }


    /** 
     * JDBC 2.0
     *  
     * Updates a column with a binary stream value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @param length of the stream
     * @exception SQLException if a database access error occurs
     */
    public void updateBinaryStream(String columnName, 
				   java.io.InputStream x,
				   int length) throws SQLException
    {
	updateBinaryStream(findColumn(columnName), x, length);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with a character stream value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @param length of the stream
     * @exception SQLException if a database access error occurs
     */
    public void updateCharacterStream(String columnName,
				      java.io.Reader reader,
				      int length) throws SQLException
    {
	updateCharacterStream(findColumn(columnName), reader, length);
    }

    public void updateCharacterStream(String columnName,
				      java.io.Reader reader,
				      long length) throws SQLException
    {
	updateCharacterStream(findColumn(columnName), reader, (int)length);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with an Object value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @param scale For java.sql.Types.DECIMAL or java.sql.Types.NUMERIC types
     *  this is the number of digits after the decimal.  For all other
     *  types this value will be ignored.
     * @exception SQLException if a database access error occurs
     */
    public void updateObject(String columnName, Object x, int scale)
      throws SQLException
    {
	updateObject(findColumn(columnName), x, scale);
    }

    /**
     * JDBC 2.0
     *  
     * Updates a column with an Object value.
     *
     * The <code>updateXXX</code> methods are used to update column values in the
     * current row, or the insert row.  The <code>updateXXX</code> methods do not 
     * update the underlying database; instead the <code>updateRow</code> or <code>insertRow</code>
     * methods are called to update the database.
     *
     * @param columnName the name of the column
     * <B>CONSUS specific:</B> "this" value of the parameter refers 
     * to the current selected object itself. In case of the updated record, 
     * values of the updated record fields are replaced with the values of 
     * specified object fields. In case of the inserted record, specified 
     * object is inserted into the table.
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     */
    public void updateObject(String columnName, Object x) throws SQLException {
	updateObject(findColumn(columnName), x);
    }

    /**
     * JDBC 2.0
     *
     * Inserts the contents of the insert row into the result set and
     * the database.  Must be on the insert row when this method is called.
     *
     * @exception SQLException if a database access error occurs,
     * if called when not on the insert row, or if not all of non-nullable columns in
     * the insert row have been given a value
     */
    public void insertRow() throws SQLException {
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	if (updateColumns) {
	    byte[] rec = table.updateRecord(stmt, null);
	    int oid = db.allocateRow(table, rec);
	    if (!table.insertInIndices(oid)) { 
		db.freeRow(table, oid);
		throw new ConsusDuplicateException();
	    }
	} else if (obj != null) { 
	    insertObject(obj);
	    obj = null;
	} else { 
	    throw new ConsusRuntimeError("No object was specified for insertRow");
	}
    }


    /**
     * JDBC 2.0
     *
     * Updates the underlying database with the new contents of the
     * current row.  Cannot be called when on the insert row.
     *
     * @exception SQLException if a database access error occurs or
     * if called when on the insert row
     */
    public void updateRow() throws SQLException {
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	if (oid == 0) { 
	    throw new ConsusRuntimeError("Cursor doesn't point to any record");
	}
	if (updateColumns) { 
	    int tableId = Record.getType(currRec, 0);
	    TableDescriptor selfTable = 
		(tableId == table.tableId) ? table : db.getTableByOid(tableId);
	    Assert.that(selfTable != null);
	    table.removeFromIndices(oid, stmt.updated);
	    byte[] newRec = selfTable.updateRecord(stmt, currRec);
	    db.put(oid, newRec);
	    if (!table.updateIndices(oid, stmt.updated)) { 
		db.put(oid, currRec);
		table.updateIndices(oid, stmt.updated);
		throw new ConsusDuplicateException();
	    }
	    currRec = newRec;
	} else if (obj != null) { 
	    updateSelfObject(obj);
	    obj = null;
	}
     }

    /**
     * JDBC 2.0
     *
     * Deletes the current row from the result set and the underlying
     * database. Cannot be called when on the insert row.
     *
     * @exception SQLException if a database access error occurs or if
     * called when on the insert row.
     */
    public void deleteRow() throws SQLException{
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	if (cursorState == CS_CURRENT_ROW && oid == selection[currPos]) { 
	    cursorState = CS_DELETED_ROW;
	    int tableId = Record.getType(currRec, 0);
	    TableDescriptor selfTable = 
		(tableId == table.tableId) ? table : db.getTableByOid(tableId);
	    db.freeRow(selfTable, oid);
	    nSelected -= 1;
	    currRec = null;
	    obj = null;
	    oid = 0;
	    System.arraycopy(selection, currPos+1, selection, currPos, 
			     nSelected - currPos);
	} else { 
	    throw new ConsusRuntimeError("Cursor doesn't point to any record");
	}
    }

    /**
     * Deletes all selected records<BR>
     * <B>This method is Consus specific and is not present in JDBC.
     * 
     * @exception SQLException if a database access error occurs or if
     * called when on the insert row.
     */
    public void deleteAllSelectedRows() throws SQLException
    {
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	if (tuples != null) { 
	    throw new ConsusRuntimeError("Positioned delete is not possible for tuples");
	}
	for (int i = 0; i < nSelected; i++) {
	    int oid = selection[i];
	    int tableId = db.getType(oid);
	    TableDescriptor selfTable = 
		(tableId == table.tableId) ? table : db.getTableByOid(tableId);
	    db.freeRow(selfTable, oid);
	} 
	nSelected = 0;
	currRec = null;
	obj = null;
	oid = 0;
	cursorState = CS_DELETED_ROW;
    }

    /**
     * JDBC 2.0
     *
     * Refreshes the current row with its most recent value in 
     * the database.  Cannot be called when on the insert row.
     *
     * The <code>refreshRow</code> method provides a way for an application to 
     * explicitly tell the JDBC driver to refetch a row(s) from the
     * database.  An application may want to call <code>refreshRow</code> when 
     * caching or prefetching is being done by the JDBC driver to
     * fetch the latest value of a row from the database.  The JDBC driver 
     * may actually refresh multiple rows at once if the fetch size is 
     * greater than one.
     * 
     * All values are refetched subject to the transaction isolation 
     * level and cursor sensitivity.  If <code>refreshRow</code> is called after
     * calling <code>updateXXX</code>, but before calling <code>updateRow</code>, then the
     * updates made to the row are lost.  Calling the method <code>refreshRow</code> frequently
     * will likely slow performance.
     *
     * @exception SQLException if a database access error occurs or if
     * called when on the insert row
     */
    public void refreshRow() throws SQLException {}

    /**
     * JDBC 2.0
     *
     * Cancels the updates made to a row.
     * This method may be called after calling an
     * <code>updateXXX</code> method(s) and before calling <code>updateRow</code> to rollback 
     * the updates made to a row.  If no updates have been made or 
     * <code>updateRow</code> has already been called, then this method has no 
     * effect.
     *
     * @exception SQLException if a database access error occurs or if
     * called when on the insert row
     *
     */
     public void cancelRowUpdates() throws SQLException {
	 Arrays.fill(stmt.updated, false);
     } 

    /**
     * JDBC 2.0
     *
     * Moves the cursor to the insert row.  The current cursor position is 
     * remembered while the cursor is positioned on the insert row.
     *
     * The insert row is a special row associated with an updatable
     * result set.  It is essentially a buffer where a new row may
     * be constructed by calling the <code>updateXXX</code> methods prior to 
     * inserting the row into the result set.  
     *
     * Only the <code>updateXXX</code>, <code>getXXX</code>,
     * and <code>insertRow</code> methods may be 
     * called when the cursor is on the insert row.  All of the columns in 
     * a result set must be given a value each time this method is
     * called before calling <code>insertRow</code>.  
     * The method <code>updateXXX</code> must be called before a
     * <code>getXXX</code> method can be called on a column value.
     *
     * @exception SQLException if a database access error occurs
     * or the result set is not updatable
     */
    public void moveToInsertRow() throws SQLException {
	saveObj = obj;
	saveRec = currRec;
	saveOid = oid;
	saveState = cursorState;
	currRec = null;
	obj = null;
	oid = 0;
	cursorState = CS_INSERTED_ROW;
    }

    /**
     * JDBC 2.0
     *
     * Moves the cursor to the remembered cursor position, usually the
     * current row.  This method has no effect if the cursor is not on the insert 
     * row. 
     *
     * @exception SQLException if a database access error occurs
     * or the result set is not updatable
     */
    public void moveToCurrentRow() throws SQLException {
	obj = saveObj;
	currRec = saveRec;
	saveRec = null;
	oid = saveOid;
	cursorState = saveState;
    }

    /**
     * JDBC 2.0
     *
     * Returns the Statement that produced this <code>ResultSet</code> object.
     * If the result set was generated some other way, such as by a
     * <code>DatabaseMetaData</code> method, this method returns <code>null</code>.
     *
     * @return the Statment that produced the result set or
     * null if the result set was produced some other way
     * @exception SQLException if a database access error occurs
     */
    public Statement getStatement() throws SQLException {
	return stmt;
    }

    /**
     * JDBC 2.0
     *
     * Returns the value of a column in the current row as a Java object.  
     * This method uses the given <code>Map</code> object
     * for the custom mapping of the
     * SQL structured or distinct type that is being retrieved.
     *
     * @param i the first column is 1, the second is 2, ...
     * @param map the mapping from SQL type names to Java classes
     * @return an object representing the SQL value
     */
    public Object getObject(int i, java.util.Map map) throws SQLException
    {
        throw new ConsusNotImplemented("ResultSet.getObject(int,java,utilMap)");
    }


    /**
     * JDBC 2.0
     *
     * Gets a REF(&lt;structured-type&gt;) column value from the current row.
     *
     * @param i the first column is 1, the second is 2, ...
     * <B>CONSUS specific:</B> zero value of the parameter refers 
     *                         to the current selected object itself
     * @return a <code>Ref</code> object representing an SQL REF value
     */
    public Ref getRef(int i) throws SQLException { 
	checkColumnIndex0(i);
	if (tuples != null) { 
	    if (tupleColumns[i-1].type != ExprNode.tpRef) { 
		throw new ConsusIncompatibleTypesError();
	    }
	    return (Ref)tuples[currPos][i-1];
	}
	if (i == 0) { 
	    return new ConsusRef(db, oid);
	} 
	FieldDescriptor fd = columns[i-1];
	if (fd.type != FieldDescriptor.tpReference) { 
	    throw new ConsusIncompatibleTypesError();
	}
	return new ConsusRef(db, Bytes.unpack4(currRec, fd.offs));
    }

    /**
     * JDBC 2.0
     *
     * Gets a BLOB value in the current row of this <code>ResultSet</code> object.
     *
     * @param i the first column is 1, the second is 2, ...
     * @return a <code>Blob</code> object representing the SQL BLOB value in
     *         the specified column
     */
    public Blob getBlob(int i) throws SQLException
    {
	checkColumnIndex(i);
	if (tuples != null) { 
	    if (tupleColumns[i-1].type != ExprNode.tpBlob) { 
		throw new ConsusIncompatibleTypesError();
	    }
	    return (Blob)tuples[currPos][i-1];
	}
	FieldDescriptor fd = columns[i-1];
	if (fd.type != FieldDescriptor.tpBlob) { 
	    throw new ConsusIncompatibleTypesError();
	}
	return new ConsusBlob(db, Bytes.unpack8(currRec, fd.offs), 
			      Bytes.unpack8(currRec, fd.offs + 8));
    }


    /**
     * JDBC 2.0
     *
     * Gets a CLOB value in the current row of this <code>ResultSet</code> object.
     *
     * @param i the first column is 1, the second is 2, ...
     * @return a <code>Clob</code> object representing the SQL CLOB value in
     *         the specified column
     */
    public Clob getClob(int i) throws SQLException
    {
	checkColumnIndex(i);
	if (tuples != null) { 
	    if (tupleColumns[i-1].type != ExprNode.tpClob) { 
		throw new ConsusIncompatibleTypesError();
	    }
	    return (Clob)tuples[currPos][i-1];
	}
	FieldDescriptor fd = columns[i-1];
	if (fd.type != FieldDescriptor.tpClob) { 
	    throw new ConsusIncompatibleTypesError();
	}
	return new ConsusClob(db, Bytes.unpack8(currRec, fd.offs), 
			      Bytes.unpack8(currRec, fd.offs + 8) >>> 1);
    }


    /**
     * JDBC 2.0
     *
     * Gets an SQL ARRAY value from the current row of this <code>ResultSet</code> object.
     *
     * @param i the first column is 1, the second is 2, ...
     * @return an <code>Array</code> object representing the SQL ARRAY value in
     *         the specified column
     */
    public Array getArray(int i) throws SQLException
    {
        throw new ConsusNotImplemented("ResultSet.getArray");
    }


    /**
     * JDBC 2.0
     *
     * Returns the value in the specified column as a Java object.  
     * This method uses the specified <code>Map</code> object for
     * custom mapping if appropriate.
     *
     * @param colName the name of the column from which to retrieve the value
     * @param map the mapping from SQL type names to Java classes
     * @return an object representing the SQL value in the specified column
     */
    public Object getObject(String colName, java.util.Map map) throws SQLException
    {
        throw new ConsusNotImplemented("ResultSet.getObject(String,java.util.Map)");
    }


    /**
     * JDBC 2.0
     *
     * Gets a REF(&lt;structured-type&gt;) column value from the current row.
     *
     * @param colName the column name
     * <B>CONSUS specific:</B> "this" value of the parameter refers 
     *                         to the current selected object itself
     * @return a <code>Ref</code> object representing the SQL REF value in
     *         the specified column
     */
    public Ref getRef(String colName) throws SQLException { 
	return getRef(findColumn(colName));
    }

    /**
     * JDBC 2.0
     *
     * Gets a BLOB value in the current row of this <code>ResultSet</code> object.
     *
     * @param colName the name of the column from which to retrieve the value
     * @return a <code>Blob</code> object representing the SQL BLOB value in
     *         the specified column
     */
    public Blob getBlob(String colName) throws SQLException { 
	return getBlob(findColumn(colName));
    }


    /**
     * JDBC 2.0
     *
     * Gets a CLOB value in the current row of this <code>ResultSet</code> object.
     *
     * @param colName the name of the column from which to retrieve the value
     * @return a <code>Clob</code> object representing the SQL CLOB value in
     *         the specified column
     */
    public Clob getClob(String colName) throws SQLException {
	return getClob(findColumn(colName));
    }

    /**
     * JDBC 2.0
     *
     * Gets an SQL ARRAY value in the current row of this <code>ResultSet</code> object.
     *
     * @param colName the name of the column from which to retrieve the value
     * @return an <code>Array</code> object representing the SQL ARRAY value in
     *         the specified column
     */
    public Array getArray(String colName) throws SQLException {
	return getArray(findColumn(colName));
    }

    /**
     * JDBC 2.0
     *
     * Gets the value of a column in the current row as a java.sql.Date 
     * object. This method uses the given calendar to construct an appropriate millisecond
     * value for the Date if the underlying database does not store
     * timezone information.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param cal the calendar to use in constructing the date
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Date getDate(int columnIndex, Calendar cal) throws SQLException {
	throw new ConsusNotImplemented("ResultSet.getDate(calendar)");
    }

    /**
     * Gets the value of a column in the current row as a java.sql.Date 
     * object. This method uses the given calendar to construct an appropriate millisecond
     * value for the Date, if the underlying database does not store
     * timezone information.
     *
     * @param columnName the SQL name of the column from which to retrieve the value
     * @param cal the calendar to use in constructing the date
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Date getDate(String columnName, Calendar cal) throws SQLException {
        return getDate(findColumn(columnName), cal);
    }

    /**
     * Gets the value of a column in the current row as a java.sql.Time 
     * object. This method uses the given calendar to construct an appropriate millisecond
     * value for the Time if the underlying database does not store
     * timezone information.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param cal the calendar to use in constructing the time
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Time getTime(int columnIndex, Calendar cal) throws SQLException { 
	throw new ConsusNotImplemented("ResultSet.getDate(calendar)");
    }


    /**
     * Gets the value of a column in the current row as a java.sql.Time 
     * object. This method uses the given calendar to construct an appropriate millisecond
     * value for the Time if the underlying database does not store
     * timezone information.
     *
     * @param columnName the SQL name of the column
     * @param cal the calendar to use in constructing the time
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Time getTime(String columnName, Calendar cal) throws SQLException { 
        return getTime(findColumn(columnName), cal);
    }


    /**
     * Gets the value of a column in the current row as a java.sql.Timestamp 
     * object. This method uses the given calendar to construct an appropriate millisecond
     * value for the Timestamp if the underlying database does not store
     * timezone information.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param cal the calendar to use in constructing the timestamp
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Timestamp getTimestamp(int columnIndex, Calendar cal) 
	throws SQLException 
    {
	throw new ConsusNotImplemented("ResultSet.getTimestamp(calendar)");
    }

    /**
     * Gets the value of a column in the current row as a java.sql.Timestamp 
     * object. This method uses the given calendar to construct an appropriate millisecond
     * value for the Timestamp if the underlying database does not store
     * timezone information.
     *
     * @param columnName the SQL name of the column
     * @param cal the calendar to use in constructing the timestamp
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database access error occurs
     */
    public java.sql.Timestamp getTimestamp(String columnName, Calendar cal)
      throws SQLException
    {
	return getTimestamp(findColumn(columnName), cal);
    }

   /****************************************************
    * CONSUS specific methods 
    ****************************************************/

    /**
     * Extract current table row as Java object
     * @return current object
     */
    public Object getSelfObject() throws SQLException { 
	if (obj == null) { 
	    if (oid == 0) { 
		return null;
	    }
	    int tableId = Record.getType(currRec, 0);
	    TableDescriptor selfTable = (table != null && tableId == table.tableId) 
		? table : db.getTableByOid(tableId);
	    Assert.that(selfTable != null);
	    if (db.objectCache != null) { 
		synchronized(db.objectCache) {
		    obj = db.objectCache.get(oid);
		    if (obj == null) { 
			obj = selfTable.fetchObject(oid, currRec);
		    } else if (obj instanceof Persistent 
			       && ((Persistent)obj).state == Persistent.PO_STUB)
		    {
			selfTable.fetchObject(obj, currRec);
		    }
		}
	    } else { 
		obj = selfTable.fetchObject(oid, currRec);
	    }
	}
	return obj;
    }

    /**
     * Insert new object in the table
     * @param <code>obj</code> inserted object
     * @return OID of the object in the database
     */
    public Ref insertObject(Object obj) throws SQLException { 
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	if (obj instanceof Persistent) { 
	    Persistent po = (Persistent)obj;
	    po.store(db);
	    return po.getRef();
	}
	TableDescriptor selfTable = db.getTableByClass(obj.getClass());
	if (table != null && !table.isSuperclassFor(selfTable)) { 
	    throw new ConsusRuntimeError("Class " + selfTable.name + " is not superclass for " 
					 + table.name);
	}
	int size = selfTable.calculateRecordSize(obj);
	byte[] rec = new byte[size];
	selfTable.allocateBlobs(rec, null, obj);
	selfTable.storeObject(rec, obj);
	int oid = db.allocateRow(selfTable, rec);
	if (!selfTable.insertInIndices(oid)) {
	    db.freeRow(selfTable, oid);
	    throw new ConsusDuplicateException();
	}
	if (db.objectCache != null) { 
	    db.objectCache.put(oid, obj);
	}
	return new ConsusRef(db, oid);
    }

    /**
     * Update current row in the result set
     * @param <code>obj</code> new version of the object. Should be of the same type as
     * original object
     */
    public void updateSelfObject(Object obj) throws SQLException { 
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	TableDescriptor selfTable = db.getTableByClass(obj.getClass());
	if (Record.getType(currRec, 0) != selfTable.tableId) { 
	    throw new ConsusRuntimeError("Type of the object can not be changed to " + obj.getClass());
	} 
	if (table != null && !table.isSuperclassFor(selfTable)) { 
	    throw new ConsusRuntimeError("Class " + selfTable.name + " is not superclass for " 
					 + table.name);
	}
	int size = selfTable.calculateRecordSize(obj);
	byte[] newRec = new byte[size];	
	selfTable.allocateBlobs(newRec, currRec, obj);
	selfTable.storeObject(newRec, obj);
	boolean updated[] = new boolean[selfTable.nFields];
	selfTable.markUpdatedFields(newRec, currRec, updated);
	selfTable.removeFromIndices(oid, updated);
	db.put(oid, newRec);
	if (!selfTable.updateIndices(oid, updated)) { 
	    db.put(oid, currRec);
	    table.updateIndices(oid, updated);
	    throw new ConsusDuplicateException();
	}
	currRec = newRec;
	if (db.objectCache != null) { 
	    db.objectCache.put(oid, obj);
	}
    }

    /**
     * Get OID of the current record (object) in the result set
     * @return <code>ConsusRef</code> with the OID of the current object
     */
    public Ref getSelfRef() { 
	return new ConsusRef(db, oid);
    }

    /**
     * Get number of records (object) in the result set
     * @return selection size
     */
    public int getSelectionSize() { 
	return nSelected;
    }

    /**
     * Retrieves the value of the designated column in the current row
     * of this <code>ResultSet</code> object as a <code>java.net.URL</code>
     * object in the Java programming language.
     * 
     * @param columnIndex the index of the column 1 is the first, 2 is the second,...
     * @return the column value as a <code>java.net.URL</code> object;
     * if the value is SQL <code>NULL</code>,
     * the value returned is <code>null</code> in the Java programming language
     * @exception SQLException if a database access error occurs or if a URL is malformed
     * @since 1.4
     */
    public java.net.URL getURL(int columnIndex) throws SQLException
    {
	try { 
	    return new java.net.URL(getString(columnIndex));
	} catch(java.net.MalformedURLException x) { 
	    throw new SQLException("Bad URL: " + x);
	}
    }

    /**
     * Retrieves the value of the designated column in the current row
     * of this <code>ResultSet</code> object as a <code>java.net.URL</code>
     * object in the Java programming language.
     * 
     * @param columnName the SQL name of the column
     * @return the column value as a <code>java.net.URL</code> object;
     * if the value is SQL <code>NULL</code>,
     * the value returned is <code>null</code> in the Java programming language
     * @exception SQLException if a database access error occurs or if a URL is malformed
     * @since 1.4
     */
    public java.net.URL getURL(String columnName) throws SQLException
    {
	return getURL(findColumn(columnName));
    }


    /**
     * Updates the designated column with a <code>java.sql.Ref</code> value.
     * The updater methods are used to update column values in the
     * current row or the insert row.  The updater methods do not 
     * update the underlying database; instead the <code>updateRow</code> or
     * <code>insertRow</code> methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public void updateRef(int columnIndex, java.sql.Ref x) throws SQLException
    {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	obj = null; // prevent usage of object instance for update
	switch (fd.type) { 
	  case FieldDescriptor.tpReference:	      
	    updateColumn(fd, x);
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }
    
    /** 
     * Updates the designated column with a <code>java.sql.Ref</code> value.
     * The updater methods are used to update column values in the
     * current row or the insert row.  The updater methods do not 
     * update the underlying database; instead the <code>updateRow</code> or
     * <code>insertRow</code> methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public void updateRef(String columnName, java.sql.Ref x) throws SQLException {
	updateRef(findColumn(columnName), x);
    }

    /**
     * Updates the designated column with a <code>java.sql.Blob</code> value.
     * The updater methods are used to update column values in the
     * current row or the insert row.  The updater methods do not 
     * update the underlying database; instead the <code>updateRow</code> or
     * <code>insertRow</code> methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public void updateBlob(int columnIndex, java.sql.Blob x) throws SQLException
    {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpBlob:
	    updateColumn(fd, (ConsusBlob)x);
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /** 
     * Updates the designated column with a <code>java.sql.Blob</code> value.
     * The updater methods are used to update column values in the
     * current row or the insert row.  The updater methods do not 
     * update the underlying database; instead the <code>updateRow</code> or
     * <code>insertRow</code> methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public void updateBlob(String columnName, java.sql.Blob x) throws SQLException
    {
	updateBlob(findColumn(columnName), x);
    }

    /**
     * Updates the designated column with a <code>java.sql.Clob</code> value.
     * The updater methods are used to update column values in the
     * current row or the insert row.  The updater methods do not 
     * update the underlying database; instead the <code>updateRow</code> or
     * <code>insertRow</code> methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public void updateClob(int columnIndex, java.sql.Clob x) throws SQLException
    {
	checkColumnIndex(columnIndex);
	if (!updateMode) { 
	    throw new ConsusNotInUpdateModeError();
	}
	FieldDescriptor fd = columns[columnIndex-1];
	switch (fd.type) { 
	  case FieldDescriptor.tpClob:
	    updateColumn(fd, (ConsusClob)x);
	    break;
	  default:
	    throw new ConsusIncompatibleTypesError();
	}
    }

    /** 
     * Updates the designated column with a <code>java.sql.Clob</code> value.
     * The updater methods are used to update column values in the
     * current row or the insert row.  The updater methods do not 
     * update the underlying database; instead the <code>updateRow</code> or
     * <code>insertRow</code> methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public void updateClob(String columnName, java.sql.Clob x) throws SQLException
    {
	updateClob(findColumn(columnName), x);
    }

    public void updateClob(int columnIndex,  Reader reader, long length) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateClob");
    }

    public void updateClob(String columnLabel,  Reader reader, long length) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateClob");
    }

    public void updateNClob(int columnIndex,  Reader reader, long length) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateNClob");
    }

    public void updateNClob(String columnLabel,  Reader reader, long length) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateNClob");
    }

    public void updateClob(int columnIndex,  Reader reader) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateClob");
    }

    public void updateClob(String columnLabel,  Reader reader) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateClob");
    }

    public void updateNClob(int columnIndex,  Reader reader) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateNClob");
    }

    public void updateNClob(String columnLabel,  Reader reader) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateNClob");
    }

    public void updateBlob(int columnIndex, InputStream inputStream, long length) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateBlob");
    }

    public void updateBlob(String columnLabel, InputStream inputStream, long length) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateBlob");
    }

    public void updateBlob(int columnIndex, InputStream inputStream) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateBlob");
    }

    public void updateBlob(String columnLabel, InputStream inputStream) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateBlob");
    }




    /**
     * Updates the designated column with a <code>java.sql.Array</code> value.
     * The updater methods are used to update column values in the
     * current row or the insert row.  The updater methods do not 
     * update the underlying database; instead the <code>updateRow</code> or
     * <code>insertRow</code> methods are called to update the database.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public void updateArray(int columnIndex, java.sql.Array x) throws SQLException
    {
	throw new ConsusNotImplemented("ConsusResultSet.updateArray(int columnIndex, java.sql.Array x)");
    }

    /** 
     * Updates the designated column with a <code>java.sql.Array</code> value.
     * The updater methods are used to update column values in the
     * current row or the insert row.  The updater methods do not 
     * update the underlying database; instead the <code>updateRow</code> or
     * <code>insertRow</code> methods are called to update the database.
     *
     * @param columnName the name of the column
     * @param x the new column value
     * @exception SQLException if a database access error occurs
     * @since 1.4
     */
    public void updateArray(String columnName, java.sql.Array x) throws SQLException
    {
	updateArray(findColumn(columnName), x);
    }

    public void updateNClob(int columnIndex, NClob nClob) throws SQLException 
    {
         throw new ConsusNotImplemented("ResultSet.updateNClob");
    }

    public void updateNClob(String columnLabel, NClob nClob) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateNClob");
    }

    public NClob getNClob(int columnIndex) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.getNClob");
    }

    public NClob getNClob(String columnLabel) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.getNClob");
    }

    public java.io.Reader getNCharacterStream(int columnIndex) throws SQLException
    {
        return getCharacterStream(columnIndex);
    }

    public java.io.Reader getNCharacterStream(String columnLabel) throws SQLException
    {
        return getCharacterStream(columnLabel);
    }

    public String getNString(int columnIndex) throws SQLException
    {
        return getString(columnIndex);
    }

    public String getNString(String columnLabel) throws SQLException
    {
        return getString(columnLabel);
    }

    public void updateCharacterStream(int columnIndex,
                                      java.io.Reader x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateCharacterStream");
    }

    public void updateCharacterStream(String columnLabel,
                                      java.io.Reader x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateCharacterStream");
    }

    public void updateNCharacterStream(int columnIndex,
                                      java.io.Reader x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateNCharacterStream");
    }

    public void updateNCharacterStream(String columnLabel,
                                      java.io.Reader x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateNCharacterStream");
    }


    public void updateNCharacterStream(int columnIndex,
                                       java.io.Reader x, long length) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateNCharacterStream");
    }

    public void updateNCharacterStream(String columnLabel,
                                       java.io.Reader x, long length) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateNCharacterStream");
    }

    public void updateBinaryStream(int columnIndex, 
                                   java.io.InputStream x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateBinaryStream");
    }

    public void updateBinaryStream(String columnLabel, 
                                   java.io.InputStream x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateBinaryStream");
    }


    public void updateAsciiStream(int columnIndex, 
                                   java.io.InputStream x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateAsciiStream");
    }

    public void updateAsciiStream(String columnLabel, 
                                   java.io.InputStream x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateAsciiStream");
    }

    public void updateSQLXML(int columnIndex, SQLXML xmlObject) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateSQLXML");
    }

    public void updateSQLXML(String columnLabel, SQLXML xmlObject) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateSQLXML");
    }

    public void updateNString(int columnIndex, String nString) throws SQLException
    {
        updateString(columnIndex, nString);
    }
            
    public void updateNString(String columnLabel, String nString) throws SQLException
    {
        updateString(columnLabel, nString);
    }
            
    public SQLXML getSQLXML(String columnLabel) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.getSQLXML");
    }

    public SQLXML getSQLXML(int columnIndex) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.getSQLXML");
    }

    public void updateRowId(int columnIndex, RowId x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateRowId");
    }

    public void updateRowId(String columnLabel, RowId x) throws SQLException
    {
         throw new ConsusNotImplemented("ResultSet.updateRowId");
    }

    public RowId getRowId(int columnIndex) throws SQLException
    {
        throw new ConsusNotImplemented("ResultSet.getRowId");
    }

    public RowId getRowId(String columnLabel) throws SQLException
    {
        throw new ConsusNotImplemented("ResultSet.getRowId");
    }

    public boolean isWrapperFor(Class iface) throws java.sql.SQLException
    {
        return table.cls != null && iface.isAssignableFrom(table.cls);
    }
    
    public Object unwrap(java.lang.Class iface) throws java.sql.SQLException
    {
	return obj;
    }


   /*
    ***************************************************
    * I M P L E M E N T A T I O N 
    ***************************************************
    */

    ConsusResultSet  next;

    int              selection[];
    int              nSelected;
    int              selectionLimit;
    int              currPos;
    byte[]           currRec;    
    Object           obj;
    int              oid;
    Object           saveObj;
    byte[]           saveRec;
    int              saveOid;
    int              saveState;
    boolean          opened;

    static final int CS_BEFORE_FIRST    = 0;
    static final int CS_AFTER_LAST      = 1;
    static final int CS_INSERTED_ROW = 2;
    static final int CS_DELETED_ROW  = 3;
    static final int CS_CURRENT_ROW  = 4;
    int              cursorState;
        
    Object[]         tuples[];
    ColumnNode       tupleColumns[];

    FieldDescriptor  columns[];

    boolean          updateMode;
    boolean          updateColumns;
     
    int              selectMap[];

    int              staticTableId;

    int              resultSetType;
    int              resultSetConcurrency; 
    
    TableDescriptor  table;
    ConsusConnection db;
    ConsusStatement  stmt;

    ConsusResultSet(ConsusStatement stmt) { 
	this(stmt, TYPE_SCROLL_SENSITIVE, CONCUR_READ_ONLY);
    }

    ConsusResultSet(ConsusStatement stmt, int resultSetType, int resultSetConcurrency) { 
	this.stmt = stmt;
	this.resultSetType = resultSetType;
	this.resultSetConcurrency = resultSetConcurrency;
	selectionLimit = Integer.MAX_VALUE;
	currPos = -1;
	cursorState = CS_BEFORE_FIRST;
	if (stmt != null) { 
	    db = stmt.db;
	}
    }

    final void open(TableDescriptor desc, boolean forUpdate) { 
	db.openCursor(this);
	table = desc;
	updateMode = forUpdate || (resultSetConcurrency == CONCUR_UPDATABLE);
	if (desc != null) { 
	    columns = desc.columns;
	    if (updateMode) { 
		stmt.updated = new boolean[table.nFields];
		stmt.columnValue = new Object[table.nFields];
	    }
	} else { 
	    columns = null; 
	    stmt.updated = null;
	    stmt.columnValue = null;
	}
	if (selection == null) { 
	    selection = new int[256];
	}
	opened = true;
    }

    final void open(ColumnNode[] columns) { 
	db.openCursor(this);
	tupleColumns = columns;
	updateMode = false;
	opened = true;
    }

    final void enableCheckForDuplicates() {
	if (selectMap == null) { 
	    selectMap = new int[(db.currIndexSize - db.dbFirstUserId + 31) >> 5];
	}
    }

    final boolean isDuplicate(int oid) { 
	if (selectMap != null) { 
	    oid -= db.dbFirstUserId;
	    if ((selectMap[oid >>> 5] & (1 << (oid & 31))) != 0) { 
		return true;
	    }
	    selectMap[oid >>> 5] |= 1 << (oid & 31);	    
	}
	return false;
    }

    final boolean checkType(int oid) throws SQLException { 
	if (isDuplicate(oid)) { 
	    return false;
	}
	if (table != null && (table.superClass != null || table.nSubclasses != 0)) { 
	    if (!table.isSuperclassFor(db.getTableByOid(db.getType(oid)))) { 
		return false;
	    }
	}
	return true;
    }

    final void checkColumnIndex(int columnIndex) throws SQLException
    {	
	int nColumns = (tupleColumns != null) ? tupleColumns.length : columns.length;
	if (columnIndex <= 0 || columnIndex > nColumns) { 
	    throw new ConsusRuntimeError("Column index " + columnIndex + " is out of range");
	}
	if (currPos < 0 && cursorState != CS_INSERTED_ROW) { 
	    throw new ConsusRuntimeError("No record was located");
	}
    }
	    
    final void checkColumnIndex0(int columnIndex) throws SQLException
    {	
	int nColumns = (tupleColumns != null) ? tupleColumns.length : columns.length;
	if (columnIndex < 0 || columnIndex > nColumns || (tupleColumns != null && columnIndex == 0)) 
	{ 
	    throw new ConsusRuntimeError("Column index " + columnIndex + " is out of range");
	}
	if (currPos < 0 && cursorState != CS_INSERTED_ROW) { 
	    throw new ConsusRuntimeError("No record was located");
	}
    }
	    
    final boolean add(int oid) { 
	if (nSelected == selection.length) { 
	    int[] newSelection = new int[nSelected*2];
	    System.arraycopy(selection, 0, newSelection, 0, nSelected);
	    selection = newSelection;
	}
	selection[nSelected++] = oid;
	return nSelected < selectionLimit;
    }

    final boolean addTuple(Object[] tuple) { 
	if (tuples == null) { 
	    tuples = new Object[256][];
	}
	if (nSelected == tuples.length) { 
	    Object[] newTuples[] = new Object[nSelected*2][];
	    System.arraycopy(tuples, 0, newTuples, 0, nSelected);
	    tuples = newTuples;
	}
	tuples[nSelected++] = tuple;
	return nSelected < selectionLimit;
    }
}
