/* Consus - Pure Java Object Relational Database Management System
 * Implementation of ResultSetMetaData protocol. Runtime information about Consus classes/tables
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus; 
import org.garret.consus.*;

import java.sql.*;

/**
 * An object that can be used to find out about the types 
 * and properties of the columns in a ResultSet.
 */

public class ConsusResultSetMetaData  implements ResultSetMetaData {

    /**
     * Returns the number of columns in this ResultSet.
     *
     * @return the number of columns
     * @exception SQLException if a database access error occurs
     */
    public int getColumnCount() throws SQLException { 
	return columns != null ? columns.length : table.columns.length;
    }

    /**
     * Indicates whether the column is automatically numbered, thus read-only.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return true if so
     * @exception SQLException if a database access error occurs
     */
     public boolean isAutoIncrement(int column) throws SQLException { 
	 return false;
     }

    /**
     * Indicates whether a column's case matters.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return true if so
     * @exception SQLException if a database access error occurs
     */
    public boolean isCaseSensitive(int column) throws SQLException { 
	return true;
    }

    /**
     * Indicates whether the column can be used in a where clause.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return true if so
     * @exception SQLException if a database access error occurs
     */
    public boolean isSearchable(int column) throws SQLException { 
	if (columns != null) { 
	    return columns[column-1].type == ExprNode.tpInt 
		|| columns[column-1].type == ExprNode.tpReal
		|| columns[column-1].type == ExprNode.tpRef
		|| columns[column-1].type == ExprNode.tpStr;
	}
	return table.columns[column-1].type <= FieldDescriptor.tpReference;
    }

    /**
     * Indicates whether the column is a cash value.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return true if so
     * @exception SQLException if a database access error occurs
     */
    public boolean isCurrency(int column) throws SQLException { 
	return false;
    }

    /**
     * Indicates the nullability of values in the designated column.		
     *
     * @param column the first column is 1, the second is 2, ...
     * @return the nullability status of the given column; one of columnNoNulls,
     *          columnNullable or columnNullableUnknown
     * @exception SQLException if a database access error occurs
     */
    public int isNullable(int column) throws SQLException { 
	if (columns != null) { 
	    return columns[column-1].type == ExprNode.tpRef 
		? DatabaseMetaData.columnNullable : DatabaseMetaData.columnNoNulls;
	}
	return table.columns[column-1].type == FieldDescriptor.tpReference 
	    ? DatabaseMetaData.columnNullable : DatabaseMetaData.columnNoNulls;
    }

    /**
     * Indicates whether values in the column are signed numbers.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return true if so
     * @exception SQLException if a database access error occurs
     */
    public boolean isSigned(int column) throws SQLException { 
	if (columns != null) { 
	    return columns[column-1].type == ExprNode.tpInt 
		|| columns[column-1].type == ExprNode.tpReal;
	}
	int type = table.columns[column-1].type;
	return type >= FieldDescriptor.tpInt1 && type <= FieldDescriptor.tpReal8;
    }

    /**
     * Indicates the column's normal max width in chars.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return the normal maximum number of characters allowed as the width
     *          of the designated column
     * @exception SQLException if a database access error occurs
     */
    public int getColumnDisplaySize(int column) throws SQLException { 
	throw new ConsusNotImplemented("ResultSetMetaData.getColumnDisplaySize");
    }

    /**
     * Gets the suggested column title for use in printouts and
     * displays.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return the suggested column title
     * @exception SQLException if a database access error occurs
     */
    public String getColumnLabel(int column) throws SQLException { 
	return columns != null ? columns[column-1].name : table.columns[column-1].name;
    }

    /**
     * Gets a column's name.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return column name
     * @exception SQLException if a database access error occurs
     */
    public String getColumnName(int column) throws SQLException { 
	return columns != null ? columns[column-1].name : table.columns[column-1].longName;
    }
	

    /**
     * Gets a column's table's schema.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return schema name or "" if not applicable
     * @exception SQLException if a database access error occurs
     */
    public String getSchemaName(int column) throws SQLException { 
	return "";
    }

    /**
     * Gets a column's number of decimal digits.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return precision
     * @exception SQLException if a database access error occurs
     */
    public int getPrecision(int column) throws SQLException { 
	throw new ConsusNotImplemented("ResultSetMetaData.getPresision");
    }

    /**
     * Gets a column's number of digits to right of the decimal point.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return scale
     * @exception SQLException if a database access error occurs
     */
    public int getScale(int column) throws SQLException {
	throw new ConsusNotImplemented("ResultSetMetaData.getPresision");
    }

    /**
     * Gets a column's table name. 
     *
     * @param column the first column is 1, the second is 2, ...
     * @return table name or "" if not applicable
     * @exception SQLException if a database access error occurs
     */
    public String getTableName(int column) throws SQLException { 
	if (columns != null) { 
	    if (columns[column-1].field != null) { 
		return columns[column-1].field.table.name;
	    } 
	    return null;
	}
	return table.name;
    }

    /**
     * Gets a column's table's catalog name.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return column name or "" if not applicable.
     * @exception SQLException if a database access error occurs
     */
    public String getCatalogName(int column) throws SQLException { 
	return "";
    }

    /**
     * Retrieves a column's SQL type.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return SQL type from java.sql.Types
     * @exception SQLException if a database access error occurs
     */
    public int getColumnType(int column) throws SQLException { 
	FieldDescriptor fd;
	if (columns != null) { 
	    fd = columns[column-1].field;
	    if (fd == null) { 
		switch (columns[column-1].type) { 
		  case ExprNode.tpInt:
		    return Types.BIGINT;
		  case ExprNode.tpReal:
		    return Types.DOUBLE;
		  case ExprNode.tpStr:
		    return Types.VARCHAR;
		  case ExprNode.tpRef:
		    return Types.REF;
		  case ExprNode.tpBlob:
		    return Types.BLOB;
		  case ExprNode.tpClob:
		    return Types.CLOB;
		  case ExprNode.tpArrayInt1: 
		    return Types.VARBINARY; 
		  case ExprNode.tpBool:
		    return Types.BIT;
		  default:
		    return Types.OTHER;
		}
	    }
	} else { 
	    fd = table.columns[column-1];
	}
	switch (fd.type) { 
	  case FieldDescriptor.tpBool:
	    return Types.BIT;
	  case FieldDescriptor.tpInt1:
	    return Types.TINYINT;
	  case FieldDescriptor.tpInt2:
	    return Types.SMALLINT;
	  case FieldDescriptor.tpInt4:
	    return Types.INTEGER;
	  case FieldDescriptor.tpInt8:
	    return Types.BIGINT;
	  case FieldDescriptor.tpReal4:
	    return Types.FLOAT;
	  case FieldDescriptor.tpReal8:
	    return Types.DOUBLE;
	  case FieldDescriptor.tpString:
	    return Types.VARCHAR;
	  case FieldDescriptor.tpReference:
	    return Types.REF;
	  case FieldDescriptor.tpDate:
	    return Types.DATE;
	  case FieldDescriptor.tpTime:
	    return Types.TIME;
	  case FieldDescriptor.tpTimestamp:
	    return Types.TIMESTAMP;
	  case FieldDescriptor.tpArray:
	    return fd.components[0].type == FieldDescriptor.tpInt1 
		? Types.VARBINARY : Types.ARRAY;
	  case FieldDescriptor.tpStructure:
	    return Types.STRUCT;
	  case FieldDescriptor.tpBlob:
	    return Types.BLOB;
	  case FieldDescriptor.tpClob:
	    return Types.CLOB;
	  default:
	    return Types.OTHER;
	}
    }

    /**
     * Retrieves a column's database-specific type name.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return type name used by the database. If the column type is
     * a user-defined type, then a fully-qualified type name is returned.
     * @exception SQLException if a database access error occurs
     */
    public String getColumnTypeName(int column) throws SQLException {
	FieldDescriptor fd;
	if (columns != null) { 
	    fd = columns[column-1].field;
	    if (fd == null) { 
		switch (columns[column-1].type) { 
		  case ExprNode.tpInt:
		    return "long";
		  case ExprNode.tpReal:
		    return "double";
		  case ExprNode.tpStr:
		    return "String";
		  case ExprNode.tpRef:
		    return "Ref";
		  case ExprNode.tpBlob:
		    return "Blob";
		  case ExprNode.tpClob:
		    return "Clob";
		  case ExprNode.tpArrayInt1: 
		    return "Binary"; 
		  case ExprNode.tpBool:
		    return "boolean";
		  default:
		    return null;
		}
	    }
	} else { 
	    fd = table.columns[column-1];
	}
	switch (fd.type) { 
	  case FieldDescriptor.tpBool:
	    return "boolean";
	  case FieldDescriptor.tpInt1:
	    return "byte";
	  case FieldDescriptor.tpInt2:
	    return "short";
	  case FieldDescriptor.tpInt4:
	    return "int";
	  case FieldDescriptor.tpInt8:
	    return "long";
	  case FieldDescriptor.tpReal4:
	    return "float";
	  case FieldDescriptor.tpReal8:
	    return "double";
	  case FieldDescriptor.tpString:
	    return "String";
	  case FieldDescriptor.tpReference:
	    return "Ref";
	  case FieldDescriptor.tpDate:
	    return "Date";
	  case FieldDescriptor.tpTime:
	    return "Time";
	  case FieldDescriptor.tpTimestamp:
	    return "Timestamp";
	  case FieldDescriptor.tpArray:
	    return (fd.components[0].type == FieldDescriptor.tpInt1) ? "Binary" : "Array";
	  case FieldDescriptor.tpStructure:
	    return table.columns[column-1].className;
	  default:
	    return "";
	}
    }

    /**
     * Indicates whether a column is definitely not writable.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return true if so
     * @exception SQLException if a database access error occurs
     */
    public boolean isReadOnly(int column) throws SQLException { 
	return false;
    }

    /**
     * Indicates whether it is possible for a write on the column to succeed.
     *
     * @param column the first column is 1, the second is 2, ...
     * @return true if so
     * @exception SQLException if a database access error occurs
     */
    public boolean isWritable(int column) throws SQLException { 
	return true;
    }

    /**
     * Indicates whether a write on the column will definitely succeed.	
     *
     * @param column the first column is 1, the second is 2, ...
     * @return true if so
     * @exception SQLException if a database access error occurs
     */
    public boolean isDefinitelyWritable(int column) throws SQLException { 
	return true;
    }

    //--------------------------JDBC 2.0-----------------------------------

    /**
     * JDBC 2.0
     *
     * <p>Returns the fully-qualified name of the Java class whose instances 
     * are manufactured if the method <code>ResultSet.getObject</code>
     * is called to retrieve a value 
     * from the column.  <code>ResultSet.getObject</code> may return a subclass of the
     * class returned by this method.
     *
     * @return the fully-qualified name of the class in the Java programming
     *         language that would be used by the method 
     * <code>ResultSet.getObject</code> to retrieve the value in the specified
     * column. This is the class name used for custom mapping.
     * @exception SQLException if a database access error occurs
     */
    public String getColumnClassName(int column) throws SQLException { 
	FieldDescriptor fd;
	if (columns != null) { 
	    fd = columns[column-1].field;
	    if (fd == null) { 
		switch (columns[column-1].type) { 
		  case ExprNode.tpInt:
		    return "long";
		  case ExprNode.tpReal:
		    return "double";
		  case ExprNode.tpStr:
		    return "java.lang.String";
		  case ExprNode.tpRef:
		    return "java.sql.Ref";
		  case ExprNode.tpBlob:
		    return "java.sql.Blob";
		  case ExprNode.tpClob:
		    return "java.sql.Clob";
		  case ExprNode.tpArrayInt1: 
		    return "byte[]";
		  case ExprNode.tpBool:
		    return "boolean";
		  default:
		    return "";
		}
	    }
	} else { 
	    fd = table.columns[column-1];
	}
	switch (fd.type) { 
	  case FieldDescriptor.tpBool:
	    return "boolean";
	  case FieldDescriptor.tpInt1:
	    return "byte";
	  case FieldDescriptor.tpInt2:
	    return "short";
	  case FieldDescriptor.tpInt4:
	    return "int";
	  case FieldDescriptor.tpInt8:
	    return "long";
	  case FieldDescriptor.tpReal4:
	    return "float";
	  case FieldDescriptor.tpReal8:
	    return "double";
	  case FieldDescriptor.tpString:
	    return "java.lang.String";
	  case FieldDescriptor.tpReference:
	    return "java.sql.Ref";
	  case FieldDescriptor.tpDate:
	    return "java.sql.Date";
	  case FieldDescriptor.tpTime:
	    return "java.sql.Time";
	  case FieldDescriptor.tpTimestamp:
	    return "java.sql.Timestamp";
	  case FieldDescriptor.tpArray:
	    return "java.sql.Array";
	  case FieldDescriptor.tpStructure:
	    return table.columns[column-1].className;
	  default:
	    return "";
	}
    }	

    public boolean isWrapperFor(Class iface) throws java.sql.SQLException
    {
        return table.cls != null && iface.isAssignableFrom(table.cls);
    }
    
    public Object unwrap(java.lang.Class iface) throws java.sql.SQLException
    {
	return null;
    }

    ConsusResultSetMetaData(TableDescriptor table) { 
	this.table = table;
    }
    
    ConsusResultSetMetaData(ColumnNode columns[]) { 
	this.columns = columns;
    }
    
    TableDescriptor table;
    ColumnNode[]    columns;
}
