/* Consus - Pure Java Object Relational Database Management System
 * Field descriptor used to store information about Consus table/object fields.
 * This class is used for fetching/storing/updating/comparing records, calculating record size. 
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;
import java.lang.reflect.*;
import java.util.Comparator;
import java.util.Arrays;


class FieldDescriptor { 
    int    id;          // field ID (starting from 0)
    int    columnIndex; // index of the column (starting from 0)

    String name;
    String longName;  // full path to the fields including structures and arrays names
    String className; // only for structures and references
    Class  clazz;     // class for structure component
    Field  field;     // associated java.lang.reflect.Field (if any)
    int    offs;      // offset withing packed image of the field
    int    size;      // packed size of the fields
    int    index;     // OID of Btree intex (0 if none)

    FieldDescriptor components[]; // for structures or arrays
    FieldDescriptor next;         // list of all components
    FieldDescriptor nextIndexed;  // list of fields to be indexed

    TableDescriptor table;
    TableDescriptor refTable;     // referenced table for reference fields or foreign keys

    String          fkTableName;  // name of referenced table for foreign key
    String          fkKeyName;    // name of primary key for correpondent foreign key
    
    FieldDescriptor nextForeignKey; // list of foreing key associated with primary key
    PreparedStatement cascadeDeleteStmt;

    FieldDescriptor foreignKeys; // list of foreign keys for this primary key

    static final int ComponentOfArray   = 0x001;
    static final int HasArrayComponents = 0x002;
    static final int IsPersistentObject = 0x004;
    static final int PrimaryKey         = 0x010;
    static final int Unique             = 0x020;
    static final int Indexed            = 0x040;
    static final int ForeignKey         = 0x080;
    static final int OnDeleteCascade    = 0x100;
    static final int ExportMask         = 0x1F0;

    int attr;
 
    static final int tpUnknown = 0;
    static final int tpBool = 1;
    static final int tpInt1 = 2;
    static final int tpInt2 = 3;
    static final int tpInt4 = 4;
    static final int tpInt8 = 5;
    static final int tpReal4 = 6; 
    static final int tpReal8 = 7; 
    static final int tpDate =  8;
    static final int tpTime =  9;
    static final int tpTimestamp = 10;
    static final int tpString = 11;
    static final int tpReference = 12;
    static final int tpArray = 13;
    static final int tpStructure = 14;
    static final int tpBlob = 15;
    static final int tpClob = 16;

    static final int sizeof[] = { 
	0, // tpUnknown
	1, // tpBool
	1, // tpInt1
	2, // tpInt2
	4, // tpInt4
	8, // tpInt8
	4, // tpReal4
	8, // tpReal8
	8, // tpDate
	8, // tpTime
	8, // tpTimestamp
	8, // tpString
	4, // tpReference
	8, // tpArray
	0, // tpStructure
	16,// tpBlob
	16 // tpClob
    };

    int    type;

    // scheme evaluation
    int    oldType;
    int    oldOffs;
    int    oldSize;

    //
    // Calculate size of the packed record for Java object
    //
    static int calculateRecordSize(FieldDescriptor[] fields, Object obj, int objOffs) 
	throws SQLException
    {
	for (int i = 0; i < fields.length; i++) { 
	    FieldDescriptor fd = fields[i];	    
	    if (fd.type == tpArray || fd.type == tpString || fd.type == tpStructure) { 
		Object o = null;
		try { 
		    o = fd.field.get(obj);
		} catch(Exception x) { 
		    throw new ConsusRuntimeError("Failed to get object component: "
						 + x.getMessage());
		}
		objOffs = fd.calculateRecordSize(o, objOffs);
	    }
	}
	return objOffs;
    }

    int calculateRecordSize(Object obj, int objOffs) throws SQLException
    {
	if (obj != null) { 
	    if (type == tpArray) { 
		int nElems = java.lang.reflect.Array.getLength(obj);
		objOffs += components[0].size*nElems;
		if ((components[0].attr & HasArrayComponents) != 0) { 
		    Object[] arr = (Object[])obj;
		    for (int j = 0; j < nElems; j++) { 
			objOffs = components[0].calculateRecordSize(arr[j], objOffs);
		    }
		}
	    } else if (type == tpString) { 
		objOffs += ((String)obj).length()*2;
	    } else if ((attr & HasArrayComponents) != 0) { 
		objOffs = calculateRecordSize(components, obj, objOffs);
	    }
	}
	return objOffs;
    }

    //
    // Calculate size of packed record for the new object format (used during scheme evaluation)
    //
    static int calculateNewRecordSize(FieldDescriptor fields[], byte obj[], 
				      int oldOffs, int newOffs)  
    {
	for (int i = 0; i < fields.length; i++) { 
	    FieldDescriptor fd = fields[i];
	    if (fd.type == tpArray) { 
		if (fd.oldType == tpUnknown) { 
		    continue;
		}
		int nElems = Varying.getSize(obj, fd.oldOffs + oldOffs);
		newOffs += nElems*fd.components[0].size;
		if ((fd.components[0].attr & HasArrayComponents) != 0) {
		    int elems = Varying.getOffs(obj, fd.oldOffs + oldOffs);
		    while (--nElems >= 0) { 
			newOffs = calculateNewRecordSize(fd.components, obj, elems, newOffs);
			elems += fd.components[0].size;
		    }
		}
	    } else if (fd.type == tpString) { 
		if (fd.oldType != tpUnknown) { 
		    newOffs += Varying.getSize(obj, fd.oldOffs + oldOffs)*2;
		}
	    } else if ((fd.attr & HasArrayComponents) != 0) {
		newOffs = calculateNewRecordSize(fd.components, obj, oldOffs, newOffs);
	    }
	}
	return newOffs;
    }

    //
    // Calculate packed size of the updated record. New values of the fields are stored in 
    // "value" component of the field descriptor. Only scalar or string components can be changed in 
    // such way. This method is also called for inserted record, in this case obj == null
    //
    static int calculateUpdatedRecordSize(ConsusStatement stmt, 
					  FieldDescriptor fields[], byte obj[], 
					  int oldOffs, int newOffs)  
    {
	for (int i = 0; i < fields.length; i++) { 
	    FieldDescriptor fd = fields[i];
	    if (fd.type == tpArray) { 
		if (fd.id < stmt.updated.length && stmt.updated[fd.id]) {
		    Assert.that(fd.components[0].type == tpInt1);
		    if (stmt.columnValue[fd.id] != null) { 
			newOffs += ((byte[])stmt.columnValue[fd.id]).length;
		    }
		} else if (obj != null) { 
		    int nElems = Varying.getSize(obj, fd.offs + oldOffs);
		    newOffs += nElems*fd.components[0].size;
		    if ((fd.components[0].attr & HasArrayComponents) != 0) {
			int elems = Varying.getOffs(obj, fd.offs + oldOffs);
			while (--nElems >= 0) { 
			    newOffs = calculateUpdatedRecordSize(stmt, fd.components, obj, elems, newOffs);
			    elems += fd.components[0].size;
			}
		    } 
		}
	    } else if (fd.type == tpString) { 
		if (fd.id < stmt.updated.length && stmt.updated[fd.id])  { 
		    if (stmt.columnValue[fd.id] != null) {
			newOffs += ((String)stmt.columnValue[fd.id]).length()*2;
		    }
		} else if (obj != null) { 
		    newOffs += Varying.getSize(obj, fd.offs + oldOffs)*2;
		}
	    } else if ((fd.attr & HasArrayComponents) != 0) {
		newOffs = calculateUpdatedRecordSize(stmt, fd.components, obj, oldOffs, newOffs);
	    }
	}
	return newOffs;
    }

    final Object construct() { 
	try { 
	    return constructor.newInstance(constructorArgs);
	} 
	catch(IllegalAccessException ex) { throw new IllegalAccessError(); }
	catch(InstantiationException ex) { throw new InstantiationError(); }
	catch(Exception ex) { 
	    throw new InstantiationError("Exception was thrown by constructor of " + className);
	}
    }


    //
    // Extract packed recor field to Java object
    //
    Object fetch(byte[] src, int base, boolean closure) throws SQLException { 
	int p = base + offs;
	switch (type) { 
	  case tpBool:
	    return new Boolean(src[p] != 0);
	  case tpInt1:
	    return new Byte(src[p]);
	  case tpInt2:
	    return new Short(Bytes.unpack2(src, p));
	  case tpInt4:
	    return new Integer(Bytes.unpack4(src, p));
	  case tpInt8:
	    return new Long(Bytes.unpack8(src, p));
	  case tpReal4:
	    return new Float(Float.intBitsToFloat(Bytes.unpack4(src, p)));
	  case tpReal8:
	    return new Double(Double.longBitsToDouble(Bytes.unpack8(src, p)));
	  case tpString:
	    return Varying.getStringOrNull(src, p);
	  case tpArray:
	    {
	      int bytes = Varying.getOffs(src, p);
	      if (bytes == 0) { 
		  return null;
	      }
	      int len = Varying.getSize(src, p);
	      switch (components[0].type) { 
		case tpBool:
		  {
		      boolean arr[] = new boolean[len];
		      for (int i = 0; i < len; i++) { 
			  arr[i] = src[bytes++] != 0;
		      }
		      return arr;
		  }
		case tpInt1:
		  {
		      byte arr[] = new byte[len];
		      for (int i = 0; i < len; i++) { 
			  arr[i] = src[bytes++];
		      }
		      return arr;
		  }
		case tpInt2:
  		  {
		      short arr[] = new short[len];
		      for (int i = 0; i < len; i++) { 
			  arr[i] = Bytes.unpack2(src, bytes);
			  bytes += 2;
		      }
		      return arr;
		  }
		case tpInt4:
		  {
		      int arr[] = new int[len];
		      for (int i = 0; i < len; i++) { 
			  arr[i] = Bytes.unpack4(src, bytes);
			  bytes += 4;
		      }
		      return arr;
		  }
		case tpInt8:
		  {
		      long arr[] = new long[len];
		      for (int i = 0; i < len; i++) { 
			  arr[i] = Bytes.unpack8(src, bytes);
			  bytes += 8;
		      }
		      return arr;
		  }
		case tpReal4:
		  {
		      float arr[] = new float[len];
		      for (int i = 0; i < len; i++) { 
			  arr[i] = Float.intBitsToFloat(Bytes.unpack4(src, bytes));
			  bytes += 4;
		      }
		      return arr;
		  }
		case tpReal8:
		  {
		      double arr[] = new double[len];
		      for (int i = 0; i < len; i++) { 
			  arr[i] = Double.longBitsToDouble(Bytes.unpack8(src, bytes));
			  bytes += 8;
		      }
		      return arr;
		  }
		default:
		  {
		      Object[] arr = (Object[])java.lang.reflect.Array.newInstance
			  (components[0].clazz, len);
		      for (int i = 0; i < len; i++) { 
			  arr[i] = components[0].fetch(src, bytes, closure);
			  bytes += components[0].size;
		      }
		      return arr;
		  }	    
	      }
	    }
	  case tpStructure:
	    {
		Object obj = construct();
		fetchRecordFields(components, obj, src, base, closure);
		return obj;
	    }
	  case tpDate:
	    return new java.sql.Date(Bytes.unpack8(src, p));
	  case tpTime:
	    return new java.sql.Time(Bytes.unpack8(src, p));
	  case tpTimestamp:
	    return new java.sql.Timestamp(Bytes.unpack8(src, p));
	  case tpReference:
	    if ((attr & IsPersistentObject) != 0) { 
		int oid = Bytes.unpack4(src, p);
		ConsusConnection db = table.db;
		Persistent po = null;
		if (oid != 0) { 
		    if (refTable == null) { 
			refTable = db.getTableByName(className);
		    }
		    Assert.that(db.objectCache != null); 
		    synchronized(db.objectCache) { 
			po = (Persistent)db.objectCache.get(oid);
			if (po != null && (po.state != Persistent.PO_STUB || !closure)) { 
			    return po;
			}
			if (closure) { 
			    byte[] rec = table.db.get(oid);
			    int tableId = Record.getType(rec, 0);
			    TableDescriptor desc = refTable;
			    if (tableId != desc.tableId) { 
				desc = db.getTableByOid(tableId);
			    }
			    if (po == null) { 
				po = (Persistent)desc.fetchObject(oid, rec);
			    } else {
				desc.fetchObject(po, rec);
			    }
			} else { 
			    int tableId = db.getType(oid);
			    TableDescriptor desc = refTable;
			    if (tableId != desc.tableId) { 
				desc = db.getTableByOid(tableId);
			    }
			    po = (Persistent)desc.construct();
			    po.type = desc;
			    po.state = Persistent.PO_STUB;
			    db.objectCache.put(oid, po);
			}
			po.oid = oid;
			po.db = db;
		    } 
		}
		return po;
	    } else { 
		return new ConsusRef(table.db, Bytes.unpack4(src, p)); 
	    }
	  case tpBlob:
	    return new ConsusBlob(table.db, Bytes.unpack8(src, p), 
				  Bytes.unpack8(src, p+8));
	  case tpClob:
	    return new ConsusClob(table.db, Bytes.unpack8(src, p), 
				  Bytes.unpack8(src, p+8) >>> 1);
	  default:
	    Assert.failed("Unknown type " + type);
	}
	return null;
    }

    void loadArrayComponents(Object elem) throws SQLException { 
	switch (type) { 
	  case tpReference:
	    if ((attr & IsPersistentObject) != 0) { 
		Persistent.get((Persistent)elem);
	    }
	    break;
	  case tpArray:
	    Object[] arr = (Object[])elem;
	    for (int i = arr.length; --i >= 0;) { 
		components[0].loadArrayComponents(arr[i]);
	    }
	    break;
	  case tpStructure:
	    loadComponents(components, elem);
	}
    }	    
	    
    //
    // Load stub components of the persistent object
    //
    static void loadComponents(FieldDescriptor fields[], Object obj)
	throws SQLException
    { 
	try { 
	    for (int i = fields.length; --i >= 0;) { 
		FieldDescriptor fd = fields[i];
		Field f = fd.field;
		switch (fd.type) { 
		  case tpReference:
		    if ((fd.attr & IsPersistentObject) != 0) { 
		       Persistent.get((Persistent)f.get(obj));
		    }
		    continue;
		  case tpArray:
		    switch (fd.components[0].type) { 
		      case tpReference:
		      case tpArray:
		      case tpStructure:
		        Object[] arr = (Object[])f.get(obj);
			if (arr != null) { 
			    for (int j = arr.length; --j >= 0;) { 
				fd.components[0].loadArrayComponents(arr[j]);
			    }
			}
		    }
		    continue;
		  case tpStructure:
		    loadComponents(fd.components, f.get(obj));
		}
	    }
	} catch(Exception x) { 
	    x.printStackTrace();
	    throw new ConsusRuntimeError("Failed to get Java object components");
	}
    }

    //
    // Mark referenced objects
    //
    static boolean markObjects(FieldDescriptor fields[], byte[] obj, int offs,
			       int greyBitmap[], int blackBitmap[])
    {
	boolean newGreyRefs = false;

	for (int i = fields.length; --i >= 0;) { 
	    FieldDescriptor fd = fields[i];
	    switch (fd.type) { 
	      case tpStructure:
		if (markObjects(fd.components, obj, offs, greyBitmap, blackBitmap)) { 
		    newGreyRefs = true;
		}
		break;
	      case tpArray:
		int len = Varying.getSize(obj, offs + fd.offs);
		int bytes = Varying.getOffs(obj, offs + fd.offs);
		while (--len >= 0) { 
		    if (markObjects(fd.components, obj, bytes, greyBitmap, blackBitmap)) { 
			newGreyRefs = true;
		    }
		    bytes += fd.components[0].size;
		}
		break;
	      case tpReference:
		int oid = Bytes.unpack4(obj, offs + fd.offs);
		if (oid != 0 && (blackBitmap[oid >>> 5] & (1 << (oid & 31))) == 0) { 
		    newGreyRefs = true;
		    greyBitmap[oid >>> 5] |= (1 << (oid & 31));
		}
	    }
	}
	return newGreyRefs;
    }

    //
    // Unpack record fields to the correpondetn Java object componetns
    //
    static void fetchRecordFields(FieldDescriptor fields[], 
				  Object obj, byte[] src, int offs, boolean closure)
	throws SQLException
    { 
	try { 
	    for (int i = 0; i < fields.length; i++) { 
		FieldDescriptor fd = fields[i];
		Field f = fd.field;
		switch (fd.type) { 
		  case tpBool:
		    f.setBoolean(obj, src[offs + fd.offs] != 0);
		    break;
		  case tpInt1:
		    f.setByte(obj, src[offs + fd.offs]);
		    break;
		  case tpInt2:
		    f.setShort(obj, Bytes.unpack2(src, offs + fd.offs));
		    break;
		  case tpInt4:
		    f.setInt(obj, Bytes.unpack4(src, offs + fd.offs));
		    break;
		  case tpInt8:
		    f.setLong(obj, Bytes.unpack8(src, offs + fd.offs));
		    break;
		  case tpReal4:
		    f.setFloat(obj, Float.intBitsToFloat
			       (Bytes.unpack4(src, offs + fd.offs)));
		    break;
		  case tpReal8:
		    f.setDouble(obj, Double.longBitsToDouble
				(Bytes.unpack8(src, offs + fd.offs)));
		    break;
		  default:
		    Object fo = fd.fetch(src, offs, closure);
    		    f.set(obj, fo);
		    break;
		}
	    }
	} catch(Exception x) { 
	    x.printStackTrace();
	    throw new ConsusRuntimeError("Failed to set Java object components");
	}
    }
    
    //
    // Pack Java object to the database record. This method is used to store non-scalar
    // componets of the Java object representing database record.
    //
    int store(byte[] dst, int dstOffs, int dstSize, Object obj) throws SQLException { 
	int i;
	switch (type) { 
	  case tpString:
	    {
	       String str = (String)obj;
	       int len = str.length();
	       Varying.setSize(dst, dstOffs + offs, len);
	       Varying.setOffs(dst, dstOffs + offs, dstSize);
	       for (i = 0; i < len; i++) { 
		   Bytes.pack2(dst, dstSize, (short)str.charAt(i));
		   dstSize += 2;
	       }
	    }
	    break;
	  case tpArray:
	    {
		int len = java.lang.reflect.Array.getLength(obj);
		Varying.setSize(dst, dstOffs + offs, len);
		Varying.setOffs(dst, dstOffs + offs, dstSize);
		switch (components[0].type) {
		  case tpBool:
		    {
			boolean arr[] = (boolean[])obj;
			for (i = 0; i < len; i++) { 
			    dst[dstSize++] = (byte)(arr[i] ? 1 : 0);
			}
		    }
		    break;
		  case tpInt1:
		    {
			byte arr[] = (byte[])obj;
			for (i = 0; i < len; i++) { 
			    dst[dstSize++] = arr[i];
			}
		    }
		    break;
		  case tpInt2:
		    {
			short arr[] = (short[])obj;
			for (i = 0; i < len; i++) { 
			    Bytes.pack2(dst, dstSize, arr[i]);
			    dstSize += 2;
			}
		    }
		    break;
		  case tpInt4:
		    {
			int arr[] = (int[])obj;
			for (i = 0; i < len; i++) { 
			    Bytes.pack4(dst, dstSize, arr[i]);
			    dstSize += 4;
			}
		    }
		    break;
		  case tpInt8:
		    {
			long arr[] = (long[])obj;
			for (i = 0; i < len; i++) { 
			    Bytes.pack8(dst, dstSize, arr[i]);
			    dstSize += 8;
			}
		    }
		    break;
		  case tpReal4:
		    {
			float arr[] = (float[])obj;
			for (i = 0; i < len; i++) { 
			    Bytes.pack4(dst, dstSize, Float.floatToIntBits(arr[i]));
			    dstSize += 4;
			}
		    }
		    break;
		  case tpReal8:
		    {
			double arr[] = (double[])obj;
			for (i = 0; i < len; i++) { 
			    Bytes.pack8(dst, dstSize, Double.doubleToLongBits(arr[i]));
			    dstSize += 8;
			}
		    }
		    break;
		  default:
		    {
			Object arr[] = (Object[])obj;
			int elemSize = components[0].size;
			int elemOffs = dstSize;
			dstSize += elemSize*len;
			for (i = 0; i < len; i++) { 
			    if (arr[i] != null) { 
				dstSize = components[0].store(dst, elemOffs, dstSize, arr[i]);
			    }
			    elemOffs += elemSize;
			}		
		    }
		}
		break;
	    }
	  case tpStructure:
	    dstSize = storeRecordFields(components, dst, dstOffs, dstSize, obj);
	    break;
	  case tpReference:
	    if (obj instanceof Persistent) { 
		if (((Persistent)obj).oid == 0) { 
		    ((Persistent)obj).store(table.db);
		}
		Bytes.pack4(dst, dstOffs + offs, (int)((Persistent)obj).oid);
	    } else { 
		Bytes.pack4(dst, dstOffs + offs, (int)((ConsusRef)obj).getOid());
	    }
	    break;
	  case tpTime:
	    Bytes.pack8(dst, dstOffs + offs, ((Time)obj).getTime());
	    break;
	  case tpDate:
	    Bytes.pack8(dst, dstOffs + offs, ((Date)obj).getTime());
	    break;
	  case tpTimestamp:
	    Bytes.pack8(dst, dstOffs + offs, ((Timestamp)obj).getTime());
	    break;
	  case tpBlob:
	  case tpClob:
	    // These fields are set by TableDescriptor.allocateBlobs() 
	    break;
	  default:
	    Assert.failed("Invalid type " + type);
	}
	return dstSize;
    }

    //
    // Store compoents of the Java object to the database record
    //
    static int storeRecordFields(FieldDescriptor fields[], byte[] dst, int offs, 
				 int size, Object obj) throws SQLException
    { 
	try { 
	    for (int i = 0; i < fields.length; i++) { 
		FieldDescriptor fd = fields[i];
		Field f = fd.field;
		Object o;
		switch (fd.type) { 
		  case tpBool:
		    dst[offs + fd.offs] = (byte)(f.getBoolean(obj) ? 1 : 0); 
		    break;
		  case tpInt1:
		    dst[offs + fd.offs] = f.getByte(obj);
		    break;
		  case tpInt2:
		    Bytes.pack2(dst, offs + fd.offs, f.getShort(obj));
		    break;
		  case tpInt4:
		    Bytes.pack4(dst, offs + fd.offs, f.getInt(obj));
		    break;
		  case tpInt8:
		    Bytes.pack8(dst, offs + fd.offs, f.getLong(obj));
		    break;
		  case tpReal4:
		    Bytes.pack4(dst, offs + fd.offs, 
				Float.floatToIntBits(f.getFloat(obj)));
		    break;
		  case tpReal8:
		    Bytes.pack8(dst, offs + fd.offs, 
				Double.doubleToLongBits(f.getDouble(obj)));
		    break;
		  default:
		    if ((o = f.get(obj)) != null) { 
			size = fd.store(dst, offs, size, o);
		    }
		}
	    } 
	} catch(Exception x) { 
	    x.printStackTrace();	    
	    throw new ConsusRuntimeError("Failed to get object component: " + x.getMessage());
	}
	return size;
    }

    // 
    // Pack updated record component. This method takes new values of the updated fields from the 
    // 'value' component of field descriptors.
    //
    int updateRecord(ConsusStatement stmt, byte[] dst, int dstOffs, byte[] src, int srcOffs, int objSize) 
	throws SQLException
    {
	long pos;
	if (id < stmt.updated.length && stmt.updated[id]) { 
	    Object value = stmt.columnValue[id];
	    if (value instanceof String) { 
		String s = (String)value;
		try { 
		    switch (type) { 
		      case tpBool:
			value = new Boolean(s);
			break;
		      case tpInt1:
		      case tpInt2:
		      case tpInt4:
			value = new Integer(Integer.parseInt(s, 10));
			break;
		      case tpInt8:
			value = new Long(Long.parseLong(s, 10));
			break;
		      case tpReal8:
		      case tpReal4:
			value = new Double(Double.parseDouble(s));
			break;
		      case FieldDescriptor.tpDate:
			value = Date.valueOf(s);
			break;
		      case FieldDescriptor.tpTime:
			value = Time.valueOf(s);
			break;
		      case FieldDescriptor.tpTimestamp:
		        value = Timestamp.valueOf(s);
			break;
		    }
		} catch(NumberFormatException x) { 
		    throw new ConsusRuntimeError("Failed to convert to scalar value string '" + s + "'");
		} catch(IllegalArgumentException x) { 
		    throw new ConsusRuntimeError("Failed to convert to date/time value string '" + s + "'");
		}
	    }
	    switch (type) { 
	      case tpBool:
		dst[dstOffs + offs] = (byte)(((Boolean)value).booleanValue() ? 1 : 0);
		break;
	      case tpInt1:
		dst[dstOffs + offs] = ((Number)value).byteValue();
		break;
	      case tpInt2:
		Bytes.pack2(dst, dstOffs + offs, ((Number)value).shortValue());
		break;
	      case tpInt4:
		Bytes.pack4(dst, dstOffs + offs, ((Number)value).intValue());
		break;
	      case tpDate:
	      case tpTime:
	      case tpTimestamp:
		if (value instanceof java.util.Date) { 
		    Bytes.pack8(dst, dstOffs + offs, ((java.util.Date)value).getTime());
		    break;
		}
		// no break
	      case tpInt8:
	        Bytes.pack8(dst, dstOffs + offs, ((Number)value).longValue());
		break;
	      case tpReal4:
		Bytes.pack4(dst, dstOffs + offs, 
			    Float.floatToIntBits(((Number)value).floatValue()));
		break;
	      case tpReal8:
		Bytes.pack8(dst, dstOffs + offs, 
			    Double.doubleToLongBits(((Number)value).doubleValue()));
		break;
	      case tpString:
	        if (value == null) {
		    Varying.setSize(dst, dstOffs + offs, 0);
		    Varying.setOffs(dst, dstOffs + offs, 0);
		} else { 
		    String str = (String)value;		   
		    int len = str.length();
		    Varying.setSize(dst, dstOffs + offs, len);
		    Varying.setOffs(dst, dstOffs + offs, objSize);
		    for (int i = 0; i < len; i++) { 
			Bytes.pack2(dst, objSize, (short)str.charAt(i));
			objSize += 2;
		    }
		}
	        break;
	      case tpArray:
		if (value != null) {
		    byte[] arr = (byte[])value;
		    int len = arr.length;
		    Varying.setSize(dst, dstOffs + offs, len);
		    Varying.setOffs(dst, dstOffs + offs, objSize);
		    System.arraycopy(arr, 0, dst, objSize, len);
		    objSize += len;
		} else { 
		    Varying.setSize(dst, dstOffs + offs, 0);
		    Varying.setOffs(dst, dstOffs + offs, 0);
		}
		break;
	      case tpReference:
		if (value == null) { 
		    Bytes.pack4(dst, dstOffs + offs, 0);
		} else if (value instanceof Persistent) { 
		    if (((Persistent)value).oid == 0) { 
			((Persistent)value).store(table.db);
		    }
		    Bytes.pack4(dst, dstOffs + offs, (int)((Persistent)value).oid);
		} else { 
		    Bytes.pack4(dst, dstOffs + offs, (int)((ConsusRef)value).getOid());
		}
		break;
	      case tpBlob:
		pos = Bytes.unpack8(src, srcOffs + offs);
		if ((value instanceof ConsusBlob) 
		    && ((ConsusBlob)value).pos == pos
		    && ((ConsusBlob)value).db == table.db) 
		{   // BLOB was not changed
		    System.arraycopy(src, srcOffs + offs, dst, dstOffs + offs, 16);
		} else {
		    if (pos != 0) { 
			long size = Bytes.unpack8(src, srcOffs + offs + 8);
			table.db.cloneBitmap(pos, size);
			table.db.free(pos, size);
		    }
		    Bytes.pack8(dst, dstOffs + offs, table.db.storeBlob((Blob)value));
		    Bytes.pack8(dst, dstOffs + offs + 8, ((Blob)value).length());
		}
	        break;
	      case tpClob:
		pos = Bytes.unpack8(src, srcOffs + offs);
		if ((value instanceof ConsusClob) 
		    && ((ConsusClob)value).pos == pos
		    && ((ConsusBlob)value).db == table.db)
		{   // CLOB was not changed
		    System.arraycopy(src, srcOffs + offs, dst, dstOffs + offs, 16);
		} else { 
		    if (pos != 0) { 
			long size = Bytes.unpack8(src, srcOffs + offs + 8);
			table.db.cloneBitmap(pos, size);
			table.db.free(pos, size);
		    }
		    Bytes.pack8(dst, dstOffs + offs, table.db.storeClob((Clob)value));
		    Bytes.pack8(dst, dstOffs + offs + 8, ((Clob)value).length()*2);
		}
		break;
	      default:
		Assert.failed("Invalid type " + type);
	    }
	} else { 
	    switch (type) { 
	      case tpArray:
		if (src != null) {
		    int len = Varying.getSize(src, srcOffs + offs);
		    int elemSize = components[0].size;
		    int srcElemOffs = Varying.getOffs(src, srcOffs + offs);
		    int dstElemOffs = objSize;
		    Varying.setSize(dst, dstOffs + offs, len);
		    Varying.setOffs(dst, dstOffs + offs, objSize);
		    objSize += len*elemSize;
		    if ((components[0].attr & HasArrayComponents) != 0) { 
			while (--len >= 0) { 
			    objSize = components[0].updateRecord(stmt, dst, dstElemOffs, src, srcElemOffs, objSize);
			    dstElemOffs += elemSize;
			    srcElemOffs += elemSize;
			}
		    } else { 
			System.arraycopy(src, srcElemOffs, dst, dstElemOffs, len*elemSize);
		    }
		} else { 
		    Varying.setSize(dst, dstOffs + offs, 0);
		    Varying.setOffs(dst, dstOffs + offs, 0);
		}
	        break;	  
	      case tpStructure:
		objSize = updateRecord(stmt, components, dst, dstOffs, src, srcOffs, objSize);
		break;
	      case tpString:
		if (src != null) {
		    int len = Varying.getSize(src, srcOffs + offs);
		    int strOffs = Varying.getOffs(src, srcOffs + offs);
		    Varying.setSize(dst, dstOffs + offs, len);
		    Varying.setOffs(dst, dstOffs + offs, objSize);
		    while (--len >= 0) { 
			dst[objSize++] = src[strOffs++];
			dst[objSize++] = src[strOffs++];
		    }
		} else { 
		    Varying.setSize(dst, dstOffs + offs, 0);
		    Varying.setOffs(dst, dstOffs + offs, 0);
		}
	        break;
	      default:
		if (src != null) {
		    for (int n = sizeof[type]; --n >= 0;) { 
			dst[dstOffs + offs + n] = src[srcOffs + offs + n];
		    }
		}
	    }
	}
	return objSize;
    }

    // 
    // Pack compoentns of the updated record. 
    // This method takes new values of the updated fields from the 'value' component of 
    // field descriptors.
    //
    static int updateRecord(ConsusStatement stmt, 
			    FieldDescriptor fields[], byte[] dst, int dstOffs, 
			    byte[] src, int srcOffs, int objSize) 
	throws SQLException
    {
	for (int i = 0; i < fields.length; i++) { 
	    objSize = fields[i].updateRecord(stmt, dst, dstOffs, src, srcOffs, objSize);
	}
	return objSize;
    }

    // 
    // Convert record fields to the new format (scheme evaluation)
    //
    int convertRecord(byte[] dst, int dstOffs, byte[] src, int srcOffs, int objSize) 
    {
	switch (type) { 
	  case tpBool:
	    switch (oldType) { 
	      case tpBool:
		dst[dstOffs + offs] = src[srcOffs + oldOffs];
		break;
	      case tpInt1:
		dst[dstOffs + offs] = (byte)(src[srcOffs + oldOffs] != 0 ? 1 : 0);
		break;
	      case tpInt2:
		dst[dstOffs + offs] = (byte)(Bytes.unpack2(src, srcOffs + oldOffs) != 0 ? 1 : 0);
		break;
	      case tpInt4:
		dst[dstOffs + offs] = (byte)(Bytes.unpack4(src, srcOffs + oldOffs) != 0 ? 1 : 0);
		break;
	      case tpDate:
	      case tpTime:
	      case tpTimestamp:
	      case tpInt8:
		dst[dstOffs + offs] = (byte)(Bytes.unpack8(src, srcOffs + oldOffs) != 0 ? 1 : 0);
		break;
	      case tpReal4:
		dst[dstOffs + offs] = (byte)(Float.intBitsToFloat(Bytes.unpack4(src, srcOffs + oldOffs)) != 0 ? 1 : 0);
		break;
	      case tpReal8:
		dst[dstOffs + offs] = (byte)(Double.longBitsToDouble(Bytes.unpack8(src, srcOffs + oldOffs)) != 0 ? 1 : 0);
		break;
	      default:
		dst[dstOffs + offs] = 0;
	    }
	    break;

	  case tpInt1:
	    switch (oldType) { 
	      case tpBool:	
	      case tpInt1:
		dst[dstOffs + offs] = src[srcOffs + oldOffs];
		break;
	      case tpInt2:
		dst[dstOffs + offs] = (byte)Bytes.unpack2(src, srcOffs + oldOffs);
		break;
	      case tpInt4:
		dst[dstOffs + offs] = (byte)Bytes.unpack4(src, srcOffs + oldOffs);
		break;
	      case tpInt8:
	      case tpDate:
	      case tpTime:
	      case tpTimestamp:
		dst[dstOffs + offs] = (byte)Bytes.unpack8(src, srcOffs + oldOffs);
		break;
	      case tpReal4:
		dst[dstOffs + offs] = 
		    (byte)Float.intBitsToFloat(Bytes.unpack4(src, srcOffs + oldOffs));
		break;
	      case tpReal8:
		dst[dstOffs + offs] = (byte)Double.longBitsToDouble
		    (Bytes.unpack8(src, srcOffs + oldOffs));
		break;
	      default:
		dst[dstOffs + offs] = 0;
	    }
	    break;

	  case tpInt2:
	    switch (oldType) { 
	      case tpBool:	
	      case tpInt1:
		Bytes.pack2(dst, dstOffs + offs, src[srcOffs + oldOffs]);
		break;
	      case tpInt2:
		Bytes.pack2(dst, dstOffs + offs, 
			    Bytes.unpack2(src, srcOffs + oldOffs));
		break;
	      case tpInt4:
		Bytes.pack2(dst, dstOffs + offs, 
			    (short)Bytes.unpack4(src, srcOffs + oldOffs));
		break;
	      case tpInt8:
	      case tpDate:
	      case tpTime:
	      case tpTimestamp:
		Bytes.pack2(dst, dstOffs + offs, 
			    (short)Bytes.unpack8(src, srcOffs + oldOffs));
		break;
	      case tpReal4:
		Bytes.pack2(dst, dstOffs + offs, (short)Float.intBitsToFloat
			    (Bytes.unpack4(src, srcOffs + oldOffs)));
		break;
	      case tpReal8:
		Bytes.pack2(dst, dstOffs + offs, (short)Double.longBitsToDouble
			    (Bytes.unpack8(src, srcOffs + oldOffs)));
		break;
	      default:
		dst[dstOffs + offs] = 0;
		dst[dstOffs + offs + 1] = 0;
	    }
	    break;

	  case tpInt4:
	    switch (oldType) { 
	      case tpBool:	
	      case tpInt1:
		Bytes.pack4(dst, dstOffs + offs, src[srcOffs + oldOffs]);
		break;
	      case tpInt2:
		Bytes.pack4(dst, dstOffs + offs, 
			    Bytes.unpack2(src, srcOffs + oldOffs));
		break;
	      case tpInt4:
		Bytes.pack4(dst, dstOffs + offs, 
			    Bytes.unpack4(src, srcOffs + oldOffs));
		break;
	      case tpInt8:
	      case tpDate:
	      case tpTime:
	      case tpTimestamp:
		Bytes.pack4(dst, dstOffs + offs, 
			    (int)Bytes.unpack8(src, srcOffs + oldOffs));
		break;
	      case tpReal4:
		Bytes.pack4(dst, dstOffs + offs, (int)Float.intBitsToFloat
			    (Bytes.unpack4(src, srcOffs + oldOffs)));
		break;
	      case tpReal8:
		Bytes.pack4(dst, dstOffs + offs, (int)Double.longBitsToDouble
			    (Bytes.unpack8(src, srcOffs + oldOffs)));
		break;
	      default:
		dst[dstOffs + offs] = 0;
		dst[dstOffs + offs + 1] = 0;
		dst[dstOffs + offs + 2] = 0;
		dst[dstOffs + offs + 3] = 0;
	    }
	    break;

	  case tpInt8:
	  case tpDate:
	  case tpTime:
	  case tpTimestamp:
	    switch (oldType) { 
	      case tpBool:	
	      case tpInt1:
		Bytes.pack8(dst, dstOffs + offs, src[srcOffs + oldOffs]);
		break;
	      case tpInt2:
		Bytes.pack8(dst, dstOffs + offs, 
			    Bytes.unpack2(src, srcOffs + oldOffs));
		break;
	      case tpInt4:
		Bytes.pack8(dst, dstOffs + offs, 
			    Bytes.unpack4(src, srcOffs + oldOffs));
		break;
	      case tpInt8:
	      case tpDate:
	      case tpTime:
	      case tpTimestamp:
		Bytes.pack8(dst, dstOffs + offs, 
			    Bytes.unpack8(src, srcOffs + oldOffs));
		break;
	      case tpReal4:
		Bytes.pack8(dst, dstOffs + offs, (long)Float.intBitsToFloat
			    (Bytes.unpack4(src, srcOffs + oldOffs)));
		break;
	      case tpReal8:
		Bytes.pack8(dst, dstOffs + offs, (long)Double.longBitsToDouble
			    (Bytes.unpack8(src, srcOffs + oldOffs)));
		break;
 	      default:
		Bytes.pack8(dst, dstOffs + offs, 0);
	    }
	    break;

	  case tpReal4:
	    switch (oldType) { 
	      case tpBool:	
	      case tpInt1:
		Bytes.pack4(dst, dstOffs + offs, 
			    Float.floatToIntBits((float)src[srcOffs + oldOffs]));
		break;
	      case tpInt2:
		Bytes.pack4(dst, dstOffs + offs, Float.floatToIntBits
			    ((float)Bytes.unpack2(src, srcOffs + oldOffs)));
		break;
	      case tpInt4:
		Bytes.pack4(dst, dstOffs + offs, Float.floatToIntBits
			    ((float)Bytes.unpack4(src, srcOffs + oldOffs)));
		break;
	      case tpInt8:
		Bytes.pack4(dst, dstOffs + offs, Float.floatToIntBits
			    ((float)Bytes.unpack8(src, srcOffs + oldOffs)));
		break;
	      case tpReal4:
		dst[dstOffs + offs] = src[srcOffs + offs];
		dst[dstOffs + offs + 1] = src[srcOffs + offs + 1];
		dst[dstOffs + offs + 2] = src[srcOffs + offs + 2];
		dst[dstOffs + offs + 3] = src[srcOffs + offs + 3];
		break;
	      case tpReal8:
		Bytes.pack4(dst, dstOffs + offs,  Float.floatToIntBits
			    ((float)Double.longBitsToDouble
			     (Bytes.unpack8(src, srcOffs + oldOffs))));
		break;
 	      default:
		Bytes.pack4(dst, dstOffs + offs, 0);
	    }
	    break;

	  case tpReal8:
	    switch (oldType) { 
	      case tpBool:	
	      case tpInt1:
		Bytes.pack8(dst, dstOffs + offs, 
			    Double.doubleToLongBits((double)src[srcOffs + oldOffs]));
		break;
	      case tpInt2:
		Bytes.pack8(dst, dstOffs + offs, Double.doubleToLongBits
			    ((double)Bytes.unpack2(src, srcOffs + oldOffs)));
		break;
	      case tpInt4:
		Bytes.pack8(dst, dstOffs + offs, Double.doubleToLongBits
			    ((double)Bytes.unpack4(src, srcOffs + oldOffs)));
		break;
	      case tpInt8:
		Bytes.pack8(dst, dstOffs + offs, Double.doubleToLongBits
			    ((double)Bytes.unpack8(src, srcOffs + oldOffs)));
		break;
	      case tpReal4:
		Bytes.pack8(dst, dstOffs + offs,  Double.doubleToLongBits
			    (Float.intBitsToFloat
			     (Bytes.unpack4(src, srcOffs + oldOffs))));
		break;
	      case tpReal8:
		Bytes.pack8(dst, dstOffs + offs,
			    Bytes.unpack8(src, srcOffs + oldOffs));
		break;
 	      default:
		Bytes.pack8(dst, dstOffs + offs, 0);
	    }
	    break;

	  case tpString:
	    if (oldType == tpUnknown) { 
		Varying.setSize(dst, dstOffs + offs, 0);
		Varying.setOffs(dst, dstOffs + offs, 0);
	    } else { 
		int len = Varying.getSize(src, srcOffs + oldOffs);
		int ofs = Varying.getOffs(src, srcOffs + oldOffs);
		Varying.setSize(dst, dstOffs + offs, len);
		Varying.setOffs(dst, dstOffs + offs, objSize);
		while (--len >= 0) { 
		    dst[objSize++] = src[ofs++];
		    dst[objSize++] = src[ofs++];
		}
	    }
	    break;

	  case tpArray:
	    if (oldType == tpUnknown) { 
		Varying.setSize(dst, dstOffs + offs, 0);
		Varying.setOffs(dst, dstOffs + offs, 0);
	    } else { 
		int len = Varying.getSize(src, srcOffs + oldOffs);
		int elemSize = components[0].size;
		int dstElemOffs = objSize;
		int srcElemOffs = Varying.getOffs(src, srcOffs + oldOffs);
		Varying.setSize(dst, dstOffs + offs, len);
		Varying.setOffs(dst, dstOffs + offs, objSize);
		objSize += len*elemSize;
		while (--len >= 0) { 
		    objSize = components[0].convertRecord(dst, dstElemOffs, src, srcElemOffs, objSize);
		    dstElemOffs += elemSize;
		    srcElemOffs += components[0].oldSize;
		}
	    }
	    break;

	  case tpStructure:
	    objSize = convertRecord(components, dst, dstOffs, src, srcOffs, objSize);
	    break;

	  case tpReference:
	    if (oldType == tpUnknown) { 
		Bytes.pack4(dst, dstOffs + offs, 0);
	    } else { 
		Bytes.pack4(dst, dstOffs + offs, 
			    Bytes.unpack4(src, srcOffs + oldOffs));
	    }
	    break;

	  case tpBlob:
	  case tpClob:
	    if (oldType == tpUnknown) { 
		Bytes.pack8(dst, dstOffs + offs, 0);
		Bytes.pack8(dst, dstOffs + offs + 8, 0);
	    } else { 
		System.arraycopy(src, srcOffs + oldOffs, dst, dstOffs + offs, 16);
	    }
	    break;

	  default:
	    Assert.failed("Invalid type " + type);
	}
	return objSize;
    }
		
 
    // 
    // Convert fields of the record to the new format (scheme evaluation)
    //   
    static int convertRecord(FieldDescriptor fields[], byte[] dst, int dstOffs, 
			     byte[] src, int srcOffs, int objSize) 
    {
	for (int i = 0; i < fields.length; i++) { 
	    objSize = fields[i].convertRecord(dst, dstOffs, src, srcOffs, objSize);
	}
	return objSize;
    }

    static Class blobClass, clobClass, refClass, persistentClass, timeClass, timestampClass, dateClass;

    static { 
	try { 
	    blobClass = Class.forName("java.sql.Blob");
	    clobClass = Class.forName("java.sql.Clob");
	    refClass = Class.forName("java.sql.Ref");
	    timeClass = Class.forName("java.sql.Time");
	    timestampClass = Class.forName("java.sql.Timestamp");
	    dateClass = Class.forName("java.sql.Date");
	    persistentClass = Class.forName("org.garret.consus.Persistent");
	} catch(ClassNotFoundException x) { 
	    x.printStackTrace();
	}
    }
	    
    //
    // Constructor used while creating table descriptor using Java class
    //
    FieldDescriptor(TableDescriptor desc, String prefix, String nm, Class cls, int inheritedAttr) 
    { 
	table = desc;
	name = nm;
	longName = prefix + nm;
	attr = inheritedAttr;
	desc.nSelfFields += 1;
	next = desc.fields;
	desc.fields = this;
	offs = desc.fixedSize;
	clazz = cls;

	String c = cls.getName();
	if (c.equals("byte")) { 
	    type = tpInt1;
	} else if (c.equals("short")) { 
	    type = tpInt2;
	} else if (c.equals("int")) { 
	    type = tpInt4;
	} else if (c.equals("long")) { 
	    type = tpInt8;
	} else if (c.equals("float")) { 
	    type = tpReal4;
	} else if (c.equals("double")) { 
	    type = tpReal8;
	} else if (c.equals("boolean")) { 
	    type = tpBool;
	} else if (persistentClass.isAssignableFrom(cls)) { 
	    type = tpReference;
	    attr |= IsPersistentObject;
	    className = c;
	} else if (refClass.isAssignableFrom(cls)) { 
	    type = tpReference;
	    int dol = nm.indexOf('$', 0);
	    if (dol >= 0) { 
		className = nm.substring(dol+1);
	    }
	} else if (blobClass.isAssignableFrom(cls)) { 
	    Assert.that("BLOB is not array component", 
			(inheritedAttr & ComponentOfArray) == 0);
	    type = tpBlob;
	    table.numberOfBlobs += 1;
	} else if (clobClass.isAssignableFrom(cls)) { 
	    Assert.that("CLOB is not array component", 
			(inheritedAttr & ComponentOfArray) == 0);
	    type = tpClob;
	    table.numberOfBlobs += 1;
	} else if (dateClass.isAssignableFrom(cls)) { 
	    type = tpDate;
	} else if (timeClass.isAssignableFrom(cls)) { 
	    type = tpTime;
	} else if (timestampClass.isAssignableFrom(cls)) { 
	    type = tpTimestamp;
	} else if (c.equals("java.lang.String")) { 
	    type = tpString;
	    attr |= HasArrayComponents;
	} else if (cls.isArray()) { 
	    type = tpArray;
	    attr |= HasArrayComponents;
	    components = new FieldDescriptor[1];
	    int fixedSize = desc.fixedSize;
	    desc.fixedSize = 0;
	    FieldDescriptor comp = new FieldDescriptor(desc, longName, "[]", 
						       cls.getComponentType(), 
						       ComponentOfArray);
	    if (comp.type == tpReference) { 
		int dol = nm.indexOf('$', 0);
		if (dol >= 0) { 
		    comp.className = nm.substring(dol+1);
		}
	    }
	    components[0] = comp;
	    desc.fixedSize = fixedSize;
	} else {
	    type = tpStructure;
	    try { 
		constructor = cls.getDeclaredConstructor(new Class[0]);
                if (setBypass != null) { 
                    setBypass.invoke(constructor, bypassFlag);
                }
	    } catch(Exception ex) {}
	    className = c;
	    buildStructureComponentsArray(desc, longName + ".", cls, 0);
	}
	size += sizeof[type]; // correctly calculate size in case of structures
	desc.fixedSize += size;

	if (type != tpStructure && (inheritedAttr & ComponentOfArray) == 0) { 
	    desc.nColumns += 1;
	}
    }

    FieldDescriptor(TableDescriptor desc, String prefix, Field f, int inheritedAttr) 
    { 
	this(desc, prefix, f.getName(), f.getType(), inheritedAttr);
	field = f;
	try { 
            if (setBypass != null) { 
                setBypass.invoke(f, bypassFlag);
            }
	} catch(Exception ex) {}
    }

    int buildStructureComponentsArray(TableDescriptor desc, String prefix, Class clazz, int n) 
    { 
	int i, j;
	Field[] allFields = clazz.getDeclaredFields();
    Arrays.sort(allFields, new Comparator<Field>() { public int compare(Field f1, Field f2) { return f1.getName().compareTo(f2.getName()); } });
	for (i = 0; i < allFields.length; i++) { 
	    if ((allFields[i].getModifiers() 
		 & (Modifier.TRANSIENT|Modifier.STATIC)) == 0)
	    {
		n += 1;
	    }
	}
	Class superclass = clazz.getSuperclass();
	if (superclass != null) { 
	    n = buildStructureComponentsArray(desc, prefix, superclass, n);
	} else { 
	    components = new FieldDescriptor[n];
	    n = 0;
	}
	for (i = 0; i < allFields.length; i++) { 
	    if ((allFields[i].getModifiers() 
		 & (Modifier.TRANSIENT|Modifier.STATIC)) == 0)
	    {
		components[n] = new FieldDescriptor(desc, prefix, allFields[i], attr & ComponentOfArray);
		size += components[n].size;
		attr |= components[n].attr & HasArrayComponents;
		n += 1;
	    }
	}
	return n;
    }
	
    //
    // Unpack components of the record field 
    //
    int unpackFields(TableDescriptor table, byte[] desc, int offs, int inheritedAttr)  
    {
	int componentsSize = 0;
	for (int i = 0; i < components.length; i++) { 
	    FieldDescriptor fd = new FieldDescriptor(table, desc, offs, inheritedAttr);
	    components[i] = fd;
	    offs += descSize;
	    switch (fd.type) { 
	      case tpArray:
		attr |= HasArrayComponents;
		offs = fd.unpackFields(table, desc, offs, ComponentOfArray);
		break;
	      case tpStructure:
		offs = fd.unpackFields(table, desc, offs, inheritedAttr);
		attr |= fd.attr & HasArrayComponents;
		break;
	      case tpString:
		attr |= HasArrayComponents;
	    }
	    componentsSize += fd.size;
	}
	if (type == tpStructure) { 
	    size = componentsSize;
	}
	return offs;
    }
    
    //
    // Find structure component by name
    //
    final FieldDescriptor getComponent(String name) { 
	for (int i = components.length; --i >= 0;) { 
	    if (components[i].name.equals(name)) { 
		return components[i];
	    }
	}
	return null;
    }

    //
    // Unpack field descriptor from the database packed format
    //   
    FieldDescriptor(TableDescriptor tab, byte[] desc, int descOffs, int inheritedAttr)
    {
	table = tab;
	longName = getName(desc, descOffs);
	name = longName.substring(longName.lastIndexOf('.')+1);
	attr = inheritedAttr;
	next = tab.fields;
	tab.fields = this;
	type = getType(desc, descOffs);
	attr |= getAttr(desc, descOffs);
	offs = getOffset(desc, descOffs);
	index = getIndex(desc, descOffs);
	size = sizeof[type];
	if (type == tpStructure) { 
	    className = getClassName(desc, descOffs);
	    int nComponents = getNcomponents(desc, descOffs);
	    components = new FieldDescriptor[nComponents];    
	} else if (type == tpArray) { 
	    components = new FieldDescriptor[1];		
	} else if (type == tpBlob || type == tpClob) { 
	    tab.numberOfBlobs += 1;
	} else if (type == tpReference) { 
	    className = getClassName(desc, descOffs);
	    if (className.length() == 0) { 
		className = null;
	    }
	}
	fkTableName = getFkTableName(desc, descOffs);
	if (fkTableName.length() != 0) { 
	    fkKeyName = getFkKeyName(desc, descOffs);
	} else { 
	    fkTableName = null;
	}
    } 

    //
    // Constructor for field descriptor for "create table"
    //
    FieldDescriptor(TableDescriptor table, String name) { 
	this.table = table;
	this.name = this.longName = name;
    }

    void calculateAttributes() { 
	offs = table.fixedSize;
	size = sizeof[type];
	table.fixedSize += size;
	table.nSelfFields += 1;
	table.nColumns += 1;
	next = table.fields;
	table.fields = this;
	if (type == tpArray) { 
	    FieldDescriptor fd = new FieldDescriptor(table, name + "[]");
	    fd.type = tpInt1;
	    fd.attr = ComponentOfArray;
	    fd.name = "[]";
	    fd.offs = 0;
	    fd.size = 1;
	    table.nSelfFields += 1;
	    fd.next = table.fields;
	    table.fields = fd;
	    components = new FieldDescriptor[1];
	    components[0] = fd;
	} else { 
	    table.columnMapping.put(longName, this);
	}
    } 

    static Method         setBypass;
    static final Object[] bypassFlag = { new Boolean(true) };
    static final Object[] constructorArgs = {};
    Constructor           constructor;

    static { 
        try { 
	    Class c = Class.forName("java.lang.reflect.AccessibleObject");
	    Class[] param = { Boolean.TYPE };
	    setBypass = c.getMethod("setAccessible", param);
        } catch(Exception ex) {}
    }      

    static int setName(byte[] rec, int pos, String str, int varyingOffs) { 
	return Varying.setString(rec, pos, str, varyingOffs);
    }
    static int setClassName(byte[] rec, int pos, String str, int varyingOffs) {
	return Varying.setString(rec, pos+8, str, varyingOffs);
    }
    static void setAttr(byte[] rec, int pos, int attr) { 
	Bytes.pack2(rec, pos+16, (short)attr);
    }
    static void setType(byte[] rec, int pos, int type) { 
	Bytes.pack2(rec, pos+18, (short)type);
    }
    static void setOffset(byte[] rec, int pos, int offset) { 
	Bytes.pack4(rec, pos+20, offset);
    }
    static void setNcomponents(byte[] rec, int pos, int nComponents) { 
	Bytes.pack4(rec, pos+24, nComponents);
    }
    static void setIndex(byte[] rec, int pos, int index) { 
	Bytes.pack4(rec, pos+28, index);
    }
    static int setFkTableName(byte[] rec, int pos, String str, int varyingOffs) {
	return Varying.setString(rec, pos+32, str, varyingOffs);
    }
    static int setFkKeyName(byte[] rec, int pos, String str, int varyingOffs) {
	return Varying.setString(rec, pos+40, str, varyingOffs);
    }
    
    static String getName(byte[] rec, int pos) { 
	return Varying.getString(rec, pos);
    }
    static String getClassName(byte[] rec, int pos) { 
	return Varying.getString(rec, pos+8);
    }
    static int getAttr(byte[] rec, int pos) { 
	return Bytes.unpack2(rec, pos+16);
    }
    static int getType(byte[] rec, int pos) { 
	return Bytes.unpack2(rec, pos+18);
    }
    static int getOffset(byte[] rec, int pos) { 
	return Bytes.unpack4(rec, pos+20);
    }
    static int getNcomponents(byte[] rec, int pos) { 
	return Bytes.unpack4(rec, pos+24);
    }
    static int getIndex(byte[] rec, int pos) { 
	return Bytes.unpack4(rec, pos+28);
    }
    static String getFkTableName(byte[] rec, int pos) {
	return Varying.getString(rec, pos+32);
    }
    static String getFkKeyName(byte[] rec, int pos) {
	return Varying.getString(rec, pos+40);
    }

    static final int descSize = 48;

    public boolean equals(Object obj) { 
	if (!(obj instanceof FieldDescriptor)) { 
	    return false;
	}
	FieldDescriptor fd = (FieldDescriptor)obj; // descriptor loaded from database
	if (longName.equals(fd.longName) && offs == fd.offs && type == fd.type 
	    && (className == null || className.equals(fd.className))
	    && (fkKeyName == fd.fkKeyName || (fkKeyName != null && fkKeyName.equals(fd.fkKeyName)))
	    && (fkTableName == fd.fkTableName || (fkTableName != null && fkTableName.equals(fd.fkTableName))))
	{
	    if (className == null) { 
		className = fd.className;
	    }
	    if (index == 0) { 
		index = fd.index;
	    }
	    return true;
	}
	return false;
    }
}
