/* Consus - Pure Java Object Relational Database Management System
 * Node for group functions
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;
import  java.sql.*;

class GroupNode extends ExprNode { 
    ExprNode   expr;
    int        columnNo;

    GroupNode(int type, int tag, ExprNode e) { 
	super(type, tag);
	expr = e;
	if (e instanceof ColumnNode) { 
	    columnNo = ((ColumnNode)e).index-1;
	} else { 
	    columnNo = -1;
	}
    }


    String dump() { 
	return "(" + opMnem[tag] + (expr != null ? (expr.dump() + ":" + columnNo) : "") + ")";
    }
    int dependsOn() { return expr.dependsOn(); }

    boolean equalsNode(ExprNode node) { 
	return node.tag == tag && expr.equalsNode(((GroupNode)node).expr);
    }

    /**
     * Swaps x[a] with x[b].
     */
    final static void swap(Object[][] x, int column, int a, int b) {
	Object t = x[a][column];
	x[a][column] = x[b][column];
	x[b][column] = t;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(Object[][] x, int column, int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++)
	    swap(x, column, a, b);
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final static int med3(Object[][] x, int column, int a, int b, int c) throws SQLException {
	int diff_ab = ((Comparable)x[a][column]).compareTo(x[b][column]);
	int diff_bc = ((Comparable)x[b][column]).compareTo(x[c][column]);
	int diff_ac = ((Comparable)x[a][column]).compareTo(x[c][column]);
	return (diff_ab < 0 ?
		(diff_bc < 0 ? b : diff_ac < 0 ? c : a) :
		(diff_bc > 0 ? b : diff_ac > 0 ? c : a));
    }

    final static void qSort(Object[][] x, int column, int off, int len) throws SQLException
    {
	// Insertion sort on smallest arrays
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && ((Comparable)x[j-1][column]).compareTo(x[j][column]) > 0; j--)
		    swap(x, column, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, column, l,     l+s, l+2*s);
		m = med3(x, column, m-s,   m,   m+s);
		n = med3(x, column, n-2*s, n-s, n);
	    }
	    m = med3(x, column, l, m, n); // Mid-size, med of 3
	}
	Comparable v = (Comparable)x[m][column];

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	int diff;
	while(true) {
	    while (b <= c && (diff = v.compareTo(x[b][column])) >= 0) {
		if (diff == 0)
		    swap(x, column, a++, b);
		b++;
	    }
	    while (c >= b && (diff = v.compareTo(x[c][column])) <= 0) {
		if (diff == 0)
		    swap(x, column, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(x, column, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(x, column, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(x, column, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    qSort(x, column, off, s);
	if ((s = d-c) > 1)
	    qSort(x, column, n-s, s);
    }
	

    long evaluateInt(ConsusStatement stmt) throws SQLException { 
	if (columnNo < 0) { 
	    throw new ConsusRuntimeError("Invalid use of GROUP function");
	}
	int begin = stmt.groupBegin;
	int end = stmt.groupEnd;
	if (end == 0) { 
	    end = stmt.cursor.nSelected;
	}
	Object[][] tuples = stmt.cursor.tuples;
	long result = 0;
	long prev;
	int i; 
	switch (tag) { 
	  case opCountAll:
	    return end - begin;
	  case opCountDis:
	    if (begin != end) { 
		qSort(tuples, columnNo, begin, end - begin);
		result = 1;
		for (i = begin+1; i < end; i++) { 
		    if (!tuples[i][columnNo].equals(tuples[i-1][columnNo])) { 
			result += 1;
		    }
		}
	    }
	    break;
	  case opIntMinAll:
	  case opIntMinDis:
	    result = Long.MAX_VALUE;
	    for (i = begin; i < end; i++) { 
		long val = ((Long)tuples[i][columnNo]).longValue();
		if (val < result) { 
		    result = val;
		}
	    }
	    break;
	  case opIntMaxAll:
	  case opIntMaxDis:
	    result = Long.MIN_VALUE;
	    for (i = begin; i < end; i++) { 
		long val = ((Long)tuples[i][columnNo]).longValue();
		if (val > result) { 
		    result = val;
		}
	    }
	    break;
	  case opIntSumDis:
	    qSort(tuples, columnNo, begin, end - begin);
	    prev = 0;
	    for (i = begin; i < end; i++) { 
		long val = ((Long)tuples[i][columnNo]).longValue();
		if (val != prev) { 
		    prev = val;
		    result += prev;
		}
	    }
	    break;
	  case opIntSumAll:
	    for (i = begin; i < end; i++) { 
		result += ((Long)tuples[i][columnNo]).longValue();
	    }
	    break;	    
	  case opIntAvgDis:
	    if (begin != end) { 
		qSort(tuples, columnNo, begin, end - begin);
		int count = 1;
		result = prev = ((Long)tuples[begin][columnNo]).longValue();
		for (i = begin+1; i < end; i++) { 
		    long val = ((Long)tuples[i][columnNo]).longValue();
		    if (val != prev) { 
			prev = val;
			result += prev;
			count += 1;
		    }
		}
		result /= count;
	    }
	    break;
	  case opIntAvgAll:
	    if (begin != end) { 
		for (i = begin; i < end; i++) { 
		    result += ((Long)tuples[i][columnNo]).longValue();
		}
		result /= (end - begin);
	    }
	    break;
	  default:
	    Assert.failed("Unknown group function code");
	}
	return result;
    }

    double evaluateReal(ConsusStatement stmt) throws SQLException { 
	if (columnNo < 0) { 
	    throw new ConsusRuntimeError("Invalid use of GROUP function");
	}
	int begin = stmt.groupBegin;
	int end = stmt.groupEnd;
	if (end == 0) { 
	    end = stmt.cursor.nSelected;
	}
	Object[][] tuples = stmt.cursor.tuples;
	double prev;
	double result = 0;
	int i; 
	switch (tag) { 
	  case opRealMinAll:
	  case opRealMinDis:
	    result = Double.MAX_VALUE;
	    for (i = begin; i < end; i++) { 
		double val = ((Double)tuples[i][columnNo]).doubleValue();
		if (val < result) { 
		    result = val;
		}
	    }
	    break;
	  case opRealMaxAll:
	  case opRealMaxDis:
	    result = Double.MIN_VALUE;
	    for (i = begin; i < end; i++) { 
		double val = ((Double)tuples[i][columnNo]).doubleValue();
		if (val > result) { 
		    result = val;
		}
	    }
	    break;
	  case opRealSumDis:
	    qSort(tuples, columnNo, begin, end - begin);
	    prev = 0;
	    for (i = begin; i < end; i++) { 
		double val = ((Double)tuples[i][columnNo]).doubleValue();
		if (val != prev) { 
		    prev = val;
		    result += prev;
		}
	    }
	    break;
	  case opRealSumAll:
	    for (i = begin; i < end; i++) { 
		result += ((Double)tuples[i][columnNo]).doubleValue();
	    }
	    break;	    
	  case opRealAvgDis:
	    if (begin != end) { 
		qSort(tuples, columnNo, begin, end - begin);
		int count = 1;
		result = prev = ((Double)tuples[begin][columnNo]).doubleValue();
		for (i = begin+1; i < end; i++) { 
		    double val = ((Double)tuples[i][columnNo]).doubleValue();
		    if (val != prev) { 
			prev = val;
			result += prev;
		    }
		}
		result /= count;
	    }
	    break;
	  case opRealAvgAll:
	    if (begin != end) { 
		for (i = begin; i < end; i++) { 
		    result += ((Double)tuples[i][columnNo]).doubleValue();
		}
		result /= (end - begin);
	    }
	    break;
	  default:
	    Assert.failed("Unknown group function code");
	}
	return result;
    }
}


