/* Consus - Pure Java Object Relational Database Management System
 * Node for INSERT statement
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.*;
    
class InsertNode extends StmtNode { 
    ColumnValue values;
    TableDescriptor table;
    
    int executeStmt(ConsusStatement stmt) throws SQLException { 
	stmt.db.beginTransaction(true);
	ColumnValue cv;
	int size = table.fixedSize;
	int oid = stmt.db.allocateId();
	stmt.oid = oid;
	stmt.updated = new boolean[table.nFields];
	stmt.columnValue = new Object[table.nFields];
	for (cv = values; cv != null; cv = cv.next) { 
	    cv.evaluate(stmt);
	    Object value = stmt.columnValue[cv.field.id];
	    switch (cv.field.type) { 
	      case FieldDescriptor.tpArray:
		Assert.that(cv.field.components[0].type == FieldDescriptor.tpInt1);
		if (value != null) {
		    size += ((byte[])value).length;
		}
		break;
	      case FieldDescriptor.tpString:
		if (value != null) {
		    size += ((String)value).length()*2;
		}
	    }
	}
	byte[] rec = new byte[size];
	size = table.fixedSize;
	for (cv = values; cv != null; cv = cv.next) { 
	    int offs = cv.field.offs;
	    Object value = stmt.columnValue[cv.field.id];
	    if (value instanceof String) { 
		String s = (String)value;
		try { 
		    switch (cv.field.type) { 
		      case FieldDescriptor.tpBool:
			value = new Boolean(s);
			break;
		      case FieldDescriptor.tpInt1:
		      case FieldDescriptor.tpInt2:
		      case FieldDescriptor.tpInt4:
			value = new Integer(Integer.parseInt(s, 10));
			break;
		      case FieldDescriptor.tpInt8:
			value = new Long(Long.parseLong(s, 10));
			break;
		      case FieldDescriptor.tpReal8:
		      case FieldDescriptor.tpReal4:
			value = new Double(Double.parseDouble(s));
			break;
		      case FieldDescriptor.tpDate:
			value = Date.valueOf(s);
			break;
		      case FieldDescriptor.tpTime:
			value = Time.valueOf(s);
			break;
		      case FieldDescriptor.tpTimestamp:
		        value = Timestamp.valueOf(s);
			break;
		    }
		} catch(NumberFormatException x) { 
		    throw new ConsusRuntimeError("Failed to convert to scalar value string '" + s + "'");
		} catch(IllegalArgumentException x) { 
		    throw new ConsusRuntimeError("Failed to convert to date/time value string '" + s + "'");
		}
	    }
	    switch (cv.field.type) { 
	      case FieldDescriptor.tpBool:
		rec[offs] = (byte)(((Boolean)value).booleanValue() ? 1 : 0);
		break;
	      case FieldDescriptor.tpInt1:
		rec[offs] = ((Number)value).byteValue();
		break;
	      case FieldDescriptor.tpInt2:
		Bytes.pack2(rec, offs, ((Number)value).shortValue());
		break;
	      case FieldDescriptor.tpInt4:
		Bytes.pack4(rec, offs, ((Number)value).intValue());
		break;
	      case FieldDescriptor.tpDate:
	      case FieldDescriptor.tpTime:
	      case FieldDescriptor.tpTimestamp:
		if (value instanceof java.util.Date) { 
		    Bytes.pack8(rec, offs, ((java.util.Date)value).getTime());
		    break;
		}
		// no break
	      case FieldDescriptor.tpInt8:
		Bytes.pack8(rec, offs, ((Number)value).longValue());
		break;
	      case FieldDescriptor.tpReal4:
		Bytes.pack4(rec, offs, 
			    Float.floatToIntBits(((Number)value).floatValue()));
		break;
	      case FieldDescriptor.tpReal8:
		Bytes.pack8(rec, offs, 
			    Double.doubleToLongBits(((Number)value).doubleValue()));
		break;
	      case FieldDescriptor.tpString:
	        if (value != null) {
		    String str = (String)value;
		    int len = str.length();
		    Varying.setSize(rec, offs, len);
		    Varying.setOffs(rec, offs, size);
		    for (int i = 0; i < len; i++) { 
		        Bytes.pack2(rec, size, (short)str.charAt(i));
		        size += 2;
		    }
	        } else { 
		    Varying.setSize(rec, offs, 0);
		}
		break;
	      case FieldDescriptor.tpArray:
	        if (value != null) {
		    byte[] arr = (byte[])value;
		    Varying.setSize(rec, offs, arr.length);
		    Varying.setOffs(rec, offs, size);
		    System.arraycopy(arr, 0, rec, size, arr.length);
		    size += arr.length;
	        } else { 
		    Varying.setSize(rec, offs, 0);
	        }
		break;
	      case FieldDescriptor.tpReference:
	        if (value != null) {
		    if (value instanceof Persistent) { 
			Persistent po = (Persistent)value;
			if (po.oid == 0) { 
			    po.store(stmt.db);			    
			}
			Bytes.pack4(rec, offs, (int)po.oid);
		    } else { 
			Bytes.pack4(rec, offs, (int)((ConsusRef)value).getOid());
		    }
		} else { 
		    Bytes.pack4(rec, offs, 0);
		}
		break;
	      case FieldDescriptor.tpBlob:
		Bytes.pack8(rec, offs, stmt.db.storeBlob((Blob)value));
		Bytes.pack8(rec, offs+8, ((Blob)value).length());
		break;
	      case FieldDescriptor.tpClob:
		Bytes.pack8(rec, offs, stmt.db.storeClob((Clob)value));
		Bytes.pack8(rec, offs+8, ((Clob)value).length()*2);
		break;
	    }
	}
	table.incrementVersionNumber(rec);
	stmt.db.allocateRow(oid, table, rec);
	int count = 1;
	if (!table.insertInIndices(oid)) {
	    stmt.db.freeRow(table, oid);
	    count = 0;
	}
	stmt.db.endStatement();
	return count;
    }

    InsertNode(TableDescriptor desc, ColumnValue values) { 
	super(tpVoid, opInsert);
	table = desc;
	this.values = values;
    }
}





