/* Consus - Pure Java Object Relational Database Management System
 * Access to the record field
 * Created by Konstantin Knizhnik, Feb 2000.
 */

package org.garret.internal.consus;
import org.garret.consus.*;
import java.sql.SQLException;

class LoadNode extends ExprNode {
    FieldDescriptor field;
    ExprNode        base;

    FieldDescriptor getField() { 
	return field;
    }

    public String dump() { 
	return "(" + opMnem[tag] + " " + (base != null ? (base.dump() + " ") : "") + field.longName + ")";
    }

    boolean equalsNode(ExprNode node) { 
	return node.tag == tag && base == ((LoadNode)node).base 
	    && (base == null || ((LoadNode)node).base.equalsNode(base));
    }

    int dependsOn() { return base.dependsOn(); }

    int evaluateBase(ConsusStatement stmt) throws SQLException { 
	int offs = field.offs;
	if (base != null) { 
	    offs += base.evaluateBase(stmt);
	}
	return offs;
    }

    long evaluateInt(ConsusStatement stmt) throws SQLException { 
	byte[] saveRec = stmt.rec;
	int offs = evaluateBase(stmt);
	long val = 0;
	switch (tag) { 
	  case opLoadInt1:
	    val = stmt.rec[offs];
	    break;
	  case opLoadInt2:
	    val = Bytes.unpack2(stmt.rec, offs);
	    break;
	  case opLoadInt4:
	    val = Bytes.unpack4(stmt.rec, offs);
	    break;
	  case opLoadInt8:
	    val = Bytes.unpack8(stmt.rec, offs);
	    break;
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	stmt.rec = saveRec; 
	return val;
    }

    double evaluateReal(ConsusStatement stmt) throws SQLException { 
	byte[] saveRec = stmt.rec;
	int offs = evaluateBase(stmt);
	double val = 0;
	switch (tag) { 
	  case opLoadReal4:
	    val = Float.intBitsToFloat(Bytes.unpack4(stmt.rec, offs));
	    break;
	  case opLoadReal8:
	    val = Double.longBitsToDouble(Bytes.unpack8(stmt.rec, offs));
	    break;
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	stmt.rec = saveRec; 	
	return val;
    }
    
    boolean evaluateBool(ConsusStatement stmt) throws SQLException { 
	Assert.that(tag == opLoadBool);
	byte[] saveRec = stmt.rec;
	int offs = evaluateBase(stmt);
	boolean val = stmt.rec[offs] != 0;
	stmt.rec = saveRec; 	
	return val;
    }

    int evaluateOid(ConsusStatement stmt) throws SQLException { 
	Assert.that(tag == opLoadInt4);
	byte[] saveRec = stmt.rec;
	int offs = evaluateBase(stmt);
	int oid = Bytes.unpack4(stmt.rec, offs);
	stmt.rec = saveRec; 	
	return oid;
    }

    String evaluateStr(ConsusStatement stmt) throws SQLException { 
	Assert.that(tag == opLoadStr);
	byte[] saveRec = stmt.rec;
	int offs = evaluateBase(stmt);
	String val = Varying.getString(stmt.rec, offs);
	stmt.rec = saveRec; 	
	return val;
    }
    
    byte[] evaluateBinary(ConsusStatement stmt) throws SQLException { 
	Assert.that(tag == opLoadBin);
	byte[] saveRec = stmt.rec;
	int offs = evaluateBase(stmt);
	int size = Varying.getSize(stmt.rec, offs);
	int arrOffs = Varying.getOffs(stmt.rec, offs);
	byte[] val = new byte[size];
	System.arraycopy(stmt.rec, arrOffs, val, 0, size);
	stmt.rec = saveRec; 	
	return val;
    }
    
    Object evaluateObj(ConsusStatement stmt) throws SQLException { 
	byte[] saveRec = stmt.rec;
	int offs = evaluateBase(stmt);	
	Object val = null;
	switch (tag) { 
	  case opLoadBlob:
	    val = new ConsusBlob(stmt.db, Bytes.unpack8(stmt.rec, offs), 
				 Bytes.unpack8(stmt.rec, offs+8)); 
	    break;
	  case opLoadClob:
	    val = new ConsusClob(stmt.db, Bytes.unpack8(stmt.rec, offs), 
				 Bytes.unpack8(stmt.rec, offs+8) >>> 1); 
	    break;
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	stmt.rec = saveRec; 	
	return val;
    }

    static LoadNode create(FieldDescriptor fd, ExprNode base) 
    {
	int type = tpVoid;
	int tag = opNop;
	switch (fd.type) { 
	  case FieldDescriptor.tpBool:
	    type = tpBool;
	    tag = opLoadBool;
	    break;
	  case FieldDescriptor.tpInt1:
	    type = tpInt;
	    tag = opLoadInt1;
	    break;
	  case FieldDescriptor.tpInt2:
	    type = tpInt;
	    tag = opLoadInt2;
	    break;
	  case FieldDescriptor.tpInt4:
	    type = tpInt;
	    tag = opLoadInt4;
	    break;
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	  case FieldDescriptor.tpDate:
	    type = tpInt;
	    tag = opLoadInt8;
	    break;
	  case FieldDescriptor.tpReal4:
	    type = tpReal;
	    tag = opLoadReal4;
	    break;
	  case FieldDescriptor.tpReal8:
	    type = tpReal;
	    tag = opLoadReal8;
	    break;
	  case FieldDescriptor.tpString:
	    type = tpStr;
	    tag = opLoadStr;
	    break;
	  case FieldDescriptor.tpReference:
	    type = tpRef;
	    tag = opLoadInt4;
	    break;
	  case FieldDescriptor.tpBlob:
	    type = tpBlob;
	    tag = opLoadBlob;
	    break;
	  case FieldDescriptor.tpClob:
	    type = tpClob;
	    tag = opLoadClob;
	    break;
	  case FieldDescriptor.tpArray:
	    tag = opLoadBase;
	    switch (fd.components[0].type) { 
	      case FieldDescriptor.tpBool:
		type = tpArrayBool;
		break;
	      case FieldDescriptor.tpInt1:
		tag = opLoadBin;
		type = tpArrayInt1;
		break;
	      case FieldDescriptor.tpInt2:
		type = tpArrayInt2;
		break;
	      case FieldDescriptor.tpInt4:
		type = tpArrayInt4;
		break;
	      case FieldDescriptor.tpInt8:
	      case FieldDescriptor.tpTime:
	      case FieldDescriptor.tpTimestamp:
	      case FieldDescriptor.tpDate:
		type = tpArrayInt8;
		break;
	      case FieldDescriptor.tpReal4:
		type = tpArrayReal4;
		break;
	      case FieldDescriptor.tpReal8:
		type = tpArrayReal8;
		break;
	      case FieldDescriptor.tpString:
		type = tpArrayStr;
		break;
	      case FieldDescriptor.tpReference:
		type = tpArrayRef;
		break;
	      case FieldDescriptor.tpArray:		
	      case FieldDescriptor.tpStructure:
		type = tpArray;
		break;
	      default:
		Assert.failed("Invalid component type " + fd.components[0].type);
	    }
	    break;
	  default:
	    Assert.failed("Invalid type " + fd.type);
	}
	return new LoadNode(type, tag, fd, base);
    }


    LoadNode(int type, int tag, FieldDescriptor f, ExprNode base) { 
	super(type, tag);
	field = f;
	this.base = base;
    }
}

