/* Consus - Pure Java Object Relational Database Management System
 * Page pool for caching most recently used pages
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;
import java.io.*;

class PagePool { 
    LRU  lru;
    Page freePages;
    Page hashTable[];
    int  poolSize;
    long fileSize;
    String     fileName;
    ConsusFile files;

    int  nDirtyPages;
    Page dirtyPages[];
    
    boolean flushing;

    PagePool(int poolSize) { 
	this.poolSize = poolSize;
    }

    final Page find(long addr, int state) throws SQLException { 	
	Assert.that((addr & (Page.pageSize-1)) == 0);
	Page pg;
	int hashCode = (int)(addr >>> Page.pageBits) % poolSize;
	ConsusFile file = null;

	synchronized (this) { 		
	    for (pg = hashTable[hashCode]; pg != null; pg = pg.collisionChain) 
	    { 
		if (pg.offs == addr) {
		    if (pg.accessCount++ == 0) { 
			pg.unlink();
		    }
		    break;
		}
	    }
	    if (pg == null) { 
		file = getFile();		
		pg = freePages;
		if (pg != null) { 
		    freePages = (Page)pg.next;
		} else { 
		    Assert.that("unfixed page available", lru.prev != lru);
		    pg = (Page)lru.prev;
		    pg.unlink();
		    synchronized (pg) { 
			if ((pg.state & Page.psDirty) != 0) { 
			    pg.state = 0;
			    file.write(pg.offs, pg.data);
			    if (!flushing) { 
				dirtyPages[pg.writeQueueIndex] = dirtyPages[--nDirtyPages];
				dirtyPages[pg.writeQueueIndex].writeQueueIndex = pg.writeQueueIndex;
			    }
			    if (pg.offs >= fileSize) { 
				fileSize = pg.offs + Page.pageSize;
			    }
			}
		    }
		    int h = (int)(pg.offs >> Page.pageBits) % poolSize;
		    Page curr = hashTable[h], prev = null;
		    while (curr != pg) { 
			prev = curr;
			curr = curr.collisionChain;
		    }
		    if (prev == null) { 
			hashTable[h] = pg.collisionChain;
		    } else { 
			prev.collisionChain = pg.collisionChain;
		    }
		}
		pg.accessCount = 1;
		pg.offs = addr;
		pg.state = Page.psRaw;
		pg.collisionChain = hashTable[hashCode];
		hashTable[hashCode] = pg;
	    }
	    if ((pg.state & Page.psDirty) == 0 && (state & Page.psDirty) != 0)
	    {
		Assert.that(!flushing);
		dirtyPages[nDirtyPages] = pg;
		pg.writeQueueIndex = nDirtyPages++;
		pg.state |= Page.psDirty;
	    }
	    if (file == null && (pg.state & Page.psRaw) != 0) { 
		file = getFile();
	    }
	}		    
	synchronized (pg) { 
	    if ((pg.state & Page.psRaw) != 0) { 
		if (file.read(pg.offs, pg.data) < Page.pageSize) {
		    for (int i = 0; i < Page.pageSize; i++) { 
			pg.data[i] = 0;
		    }
		}
		pg.state &= ~Page.psRaw;
	    }		
	}
	if (file != null) { 
	    synchronized (this) { 
		releaseFile(file);
	    }
	}
	return pg;
    }


    final void copy(long dst, long src, long size) throws SQLException
    {
	int dstOffs = (int)dst & (Page.pageSize-1);
	int srcOffs = (int)src & (Page.pageSize-1);
	dst -= dstOffs;
	src -= srcOffs;
	Page dstPage = find(dst, Page.psDirty);
	Page srcPage = find(src, 0);
	do { 
	    if (dstOffs == Page.pageSize) { 
		unfix(dstPage);
		dst += Page.pageSize;
		dstPage = find(dst, Page.psDirty);
		dstOffs = 0;
	    }
	    if (srcOffs == Page.pageSize) { 
		unfix(srcPage);
		src += Page.pageSize;
		srcPage = find(src, 0);
		srcOffs = 0;
	    }
	    long len = size;
	    if (len > Page.pageSize - srcOffs) { 
		len = Page.pageSize - srcOffs; 
	    }
	    if (len > Page.pageSize - dstOffs) { 
		len = Page.pageSize - dstOffs; 
	    }
	    System.arraycopy(srcPage.data, srcOffs, dstPage.data, dstOffs, (int)len);
	    srcOffs += len;
	    dstOffs += len;
	    size -= len;
	} while (size != 0);
	unfix(dstPage);
	unfix(srcPage);
    }

    final void open(ConsusFile f, long size) throws SQLException 
    {
	files = f;
	fileName = f.name;
	fileSize = size;
	hashTable = new Page[poolSize];
	dirtyPages = new Page[poolSize];
	nDirtyPages = 0;
	lru = new LRU();
	freePages = null;
	for (int i = poolSize; --i >= 0; ) { 
	    Page pg = new Page();
	    pg.next = freePages;
	    freePages = pg;
	}
    }

    final synchronized void close() throws SQLException { 
	for (ConsusFile file = files; file != null; file = file.next) { 
	    file.close();
	}
	hashTable = null;
	dirtyPages = null;
	lru = null;
	freePages = null;
    }

    final synchronized void unfix(Page pg) { 
	Assert.that(pg.accessCount > 0);
	if (--pg.accessCount == 0) { 
	    lru.link(pg);
	}
    }

    final synchronized void modify(Page pg) { 
	Assert.that(pg.accessCount > 0);
	if ((pg.state & Page.psDirty) == 0) { 
	    Assert.that(!flushing);
	    pg.state |= Page.psDirty;
	    dirtyPages[nDirtyPages] = pg;
	    pg.writeQueueIndex = nDirtyPages++;
	}
    }
    
    final Page getPage(long addr) throws SQLException { 
	return find(addr, 0);
    }
    
    final Page putPage(long addr) throws SQLException { 
	return find(addr, Page.psDirty);
    }
    
    final byte[] get(long pos) throws SQLException { 
	Assert.that(pos != 0);
	int offs = (int)pos & (Page.pageSize-1);
	Page pg = find(pos - offs, 0);
	int size = Record.getSize(pg.data, offs);
	Assert.that(size >= Record.sizeof);
	byte[] obj = new byte[size];
	int dst = 0;
	while (size > Page.pageSize - offs) { 
	    System.arraycopy(pg.data, offs, obj, dst, Page.pageSize - offs);
	    unfix(pg);
	    size -= Page.pageSize - offs;
	    pos += Page.pageSize - offs;
	    dst += Page.pageSize - offs;
	    pg = find(pos, 0);
	    offs = 0;
	}
	System.arraycopy(pg.data, offs, obj, dst, size);
	unfix(pg);
	return obj;
    }

    final void put(long pos, byte[] obj) throws SQLException { 
	int offs = (int)pos & (Page.pageSize-1);
	Page pg = find(pos - offs, Page.psDirty);
	int size = obj.length;
	int src = 0;
	while (size > Page.pageSize - offs) { 
	    System.arraycopy(obj, src, pg.data, offs, Page.pageSize - offs);
	    unfix(pg);
	    size -= Page.pageSize - offs;
	    pos += Page.pageSize - offs;
	    src += Page.pageSize - offs;
	    pg = find(pos, Page.psDirty);
	    offs = 0;
	}
	System.arraycopy(obj, src, pg.data, offs, size);
	unfix(pg);
    }

    void flush() throws SQLException { 
	long maxOffs;
	ConsusFile file;
	synchronized (this) { 
	    flushing = true;
	    java.util.Arrays.sort(dirtyPages, 0, nDirtyPages); 
	    maxOffs = fileSize;
	    file = getFile();
	}
	for (int i = 0; i < nDirtyPages; i++) { 
	    Page pg = dirtyPages[i];
	    synchronized (pg) { 
		if ((pg.state & Page.psDirty) != 0) { 
		    file.write(pg.offs, pg.data);
		    pg.state &= ~Page.psDirty;
		    if (pg.offs >= maxOffs) { 
			maxOffs = pg.offs + Page.pageSize;
		    }
		}
	    }
	}	    
	file.sync();
	synchronized (this) { 
	    releaseFile(file);
	    nDirtyPages = 0;
	    flushing = false;
	    if (maxOffs > fileSize) { 
		fileSize = maxOffs;
	    }
	}
    }

    ConsusFile getFile() throws SQLException { 
	ConsusFile file = files;
	if (file == null) { 
	    return new ConsusFile(fileName);
	}
	files = file.next;
	return file;
    }

    void releaseFile(ConsusFile file) { 
	file.next = files;
	files = file;
    }    
}







