/* Consus - Pure Java Object Relational Database Management System
 * Implementation of RAID 0 file
 * Created by Konstantin Knizhnik, May 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.io.*;

class Raid0File extends RandomAccessMultiFile {
    int  raidBlockSize;   // RAID 0 block size

    /**
     * Creates a random access file stream to read from, and optionally 
     * to write to, a file with the specified name. A new 
     * {@link FileDescriptor} object is created to represent the 
     * connection to the file.
     * <p>
     * The mode argument must either be equal to <code>"r"</code> or 
     * <code>"rw"</code>, indicating that the file is to be opened for 
     * input only or for both input and output, respectively. The 
     * write methods on this object will always throw an 
     * <code>IOException</code> if the file is opened with a mode of 
     * <code>"r"</code>. If the mode is <code>"rw"</code> and the 
     * file does not exist, then an attempt is made to create it.
     * An <code>IOException</code> is thrown if the name argument
     * refers to a directory.
     * <p>
     * If there is a security manager, its <code>checkRead</code> method
     * is called with the <code>name</code> argument
     * as its argument to see if read access to the file is allowed.
     * If the mode is "rw", the security manager's
     * <code>checkWrite</code> method
     * is also called with the <code>name</code> argument
     * as its argument to see if write access to the file is allowed.
     *
     * @param      name   the system-dependent filename.
     * @param      mode   the access mode.
     * @exception  IllegalArgumentException  if the mode argument is not equal
     *               to <code>"r"</code> or to <code>"rw"</code>.
     * @exception  FileNotFoundException  if the file exists but is a directory
     *                   rather than a regular file, or cannot be opened or
     *                   created for any other reason
     * @exception  SecurityException         if a security manager exists and its
     *               <code>checkRead</code> method denies read access to the file
     *               or the mode is "rw" and the security manager's
     *               <code>checkWrite</code> method denies write access to the file.
     */
    public Raid0File(String name, String mode)
	throws FileNotFoundException, java.io.IOException
    {
	super(name);
	StreamTokenizer in = 
	    new StreamTokenizer(new BufferedReader(new FileReader(getFD())));
	segment = new MultiFileSegment[0];
	int tkn = in.nextToken();
	if (tkn != StreamTokenizer.TT_NUMBER && tkn != '"') { 
	    throw new IOException("RAID 0 segment size expected");
	}
	raidBlockSize = (int)in.nval*1024; // kilobytes
	tkn = in.nextToken();
	do {
	    MultiFileSegment seg = new MultiFileSegment();
	    if (tkn != StreamTokenizer.TT_WORD) { 
		throw new IOException("RAID segment name expected");
	    }
	    seg.name = in.sval;
	    seg.f = new RandomAccessFile(seg.name, mode);
	    segment = (MultiFileSegment[])AnyArray.append(segment, seg);
	    tkn = in.nextToken();
	    if (tkn == StreamTokenizer.TT_NUMBER) { 
		seg.offs = (long)in.nval*1024; // kilobytes
		tkn = in.nextToken();
	    }
	} while (tkn != StreamTokenizer.TT_EOF);
    }
   
    /**
     * Creates a random access file stream to read from, and optionally 
     * to write to, the file specified by the <code>File</code> argument. 
     * A new {@link FileDescriptor} object is created to represent 
     * this file connection. 
     * <p>
     * The mode argument must either be equal to <code>"r"</code> or 
     * <code>"rw"</code>, indicating that the file is to be opened for 
     * input only or for both input and output, respectively. The 
     * write methods on this object will always throw an 
     * <code>IOException</code> if the file is opened with a mode of 
     * <code>"r"</code>. If the mode is <code>"rw"</code> and the 
     * file does not exist, then an attempt is made to create it.
     * An <code>IOException</code> is thrown if the file argument
     * refers to a directory.
     * <p>
     * If there is a security manager, its <code>checkRead</code> method
     * is called with the pathname of the <code>file</code>
     * argument as its argument to see if read access to the file is allowed.
     * If the mode is "rw", the security manager's
     * <code>checkWrite</code> method
     * is also called with the path argument
     * to see if write access to the file is allowed.
     *
     * @param      file   the file object.
     * @param      mode   the access mode.
     * @exception  IllegalArgumentException  if the mode argument is not equal
     *               to <code>"r"</code> or to <code>"rw"</code>.
     * @exception  FileNotFoundException  if the file exists but is a directory
     *                   rather than a regular file, or cannot be opened or
     *                   created for any other reason
     * @exception  SecurityException         if a security manager exists and its
     *               <code>checkRead</code> method denies read access to the file
     *               or the mode is "rw" and the security manager's
     *               <code>checkWrite</code> method denies write access to the file.
     * @see        java.io.File#getPath()
     */
    public Raid0File(File file, String mode)
	throws FileNotFoundException, java.io.IOException
    {
	this(file.getPath(), mode);
    }


    public int read(byte b[], int off, int size) throws IOException {
	int startOff = off;
	while (true) { 
	    int i = (int)(currPos / raidBlockSize) % segment.length;
	    int offs = (int)currPos % raidBlockSize;
	    int available = raidBlockSize - offs;
	    if (available >= size) {
		segment[i].f.seek(segment[i].offs + currPos / (raidBlockSize*segment.length) * raidBlockSize + offs);
		currPos += size;
		return segment[i].f.read(b, off, size);	    
	    }
	    segment[i].f.seek(segment[i].offs + currPos / (raidBlockSize*segment.length) * raidBlockSize + offs);
	    int rc = segment[i].f.read(b, off, available);
	    if (rc < 0) {
		return rc;
	    } else { 
		currPos += rc;
		off += rc;
		size -= rc;
		if (rc < available) { 
		    return off - startOff;
		}
	    }
	}
    } 
    
    /**
     * Attempts to skip over <code>n</code> bytes of input discarding the 
     * skipped bytes. 
     * <p>
     * 
     * This method may skip over some smaller number of bytes, possibly zero. 
     * This may result from any of a number of conditions; reaching end of 
     * file before <code>n</code> bytes have been skipped is only one 
     * possibility. This method never throws an <code>EOFException</code>. 
     * The actual number of bytes skipped is returned.  If <code>n</code> 
     * is negative, no bytes are skipped.
     *
     * @param      n   the number of bytes to be skipped.
     * @return     the actual number of bytes skipped.
     * @exception  IOException  if an I/O error occurs.
     */
    public int skipBytes(int n) throws IOException {
	if (n <= 0) {
	    return 0;
	}
	long currLength = length();
	if (currPos + n > currLength) { 
	    n = (int)(currLength - currPos);
	} 
	currPos += n;
	return n;
    } 

    /**
     * Writes <code>len</code> bytes from the specified byte array 
     * starting at offset <code>off</code> to this file. 
     *
     * @param      b     the data.
     * @param      off   the start offset in the data.
     * @param      len   the number of bytes to write.
     * @exception  IOException  if an I/O error occurs.
     */
    public void write(byte b[], int off, int size) throws IOException 
    {
	while (true) { 
	    int i = (int)(currPos / raidBlockSize) % segment.length;
	    int offs = (int)currPos % raidBlockSize;
	    int available = raidBlockSize - offs;
	    if (available >= size) {
		segment[i].f.seek(segment[i].offs + currPos / (raidBlockSize*segment.length) * raidBlockSize + offs);
		currPos += size;
		segment[i].f.write(b, off, size);	    
		return;
	    }
	    segment[i].f.seek(segment[i].offs + currPos / (raidBlockSize*segment.length) * raidBlockSize + offs);
	    segment[i].f.write(b, off, available);
	    currPos += available;
	    off += available;
	    size -= available;
	}
    }

    /**
     * Returns the current offset in this file. 
     *
     * @return     the offset from the beginning of the file, in bytes,
     *             at which the next read or write occurs.
     * @exception  IOException  if an I/O error occurs.
     */
    public long getFilePointer() throws IOException { 
	return currPos;
    }

    /**
     * Sets the file-pointer offset, measured from the beginning of this 
     * file, at which the next read or write occurs.  The offset may be 
     * set beyond the end of the file. Setting the offset beyond the end 
     * of the file does not change the file length.  The file length will 
     * change only by writing after the offset has been set beyond the end 
     * of the file. 
     *
     * @param      pos   the offset position, measured in bytes from the 
     *                   beginning of the file, at which to set the file 
     *                   pointer.
     * @exception  IOException  if <code>pos</code> is less than 
     *                          <code>0</code> or if an I/O error occurs.
     */
    public void seek(long pos) throws IOException {
	currPos = pos;
    }


    /**
     * Returns the length of this file.
     *
     * @return     the length of this file, measured in bytes.
     * @exception  IOException  if an I/O error occurs.
     */
    public long length() throws IOException { 
	long totalLength = 0;
	for (int i = 0; i < segment.length; i++) { 
	    totalLength += segment[i].f.length();
	}
	return totalLength;
    }
    
    /**
     * Sets the length of this file.
     *
     * <p> If the present length of the file as returned by the
     * <code>length</code> method is greater than the <code>newLength</code>
     * argument then the file will be truncated.  In this case, if the file
     * offset as returned by the <code>getFilePointer</code> method is greater
     * then <code>newLength</code> then after this method returns the offset
     * will be equal to <code>newLength</code>.
     *
     * <p> If the present length of the file as returned by the
     * <code>length</code> method is smaller than the <code>newLength</code>
     * argument then the file will be extended.  In this case, the contents of
     * the extended portion of the file are not defined.
     *
     * @param      newLength    The desired length of the file
     * @exception  IOException  If an I/O error occurs
     * @since      1.2
     */
    public void setLength(long newLength) throws IOException { 
	for (int i = 0; i < segment.length; i++) { 
	    int offs = 0;
	    if (i == (int)(newLength / raidBlockSize) % segment.length) { 
		offs = (int)newLength % raidBlockSize;		
	    }
	    segment[i].f.setLength(segment[i].offs + newLength / (raidBlockSize*segment.length) * raidBlockSize + offs);
	}
    }    
}







