/* Consus - Pure Java Object Relational Database Management System
 * Implementation of multisegment file
 * Created by Konstantin Knizhnik, May 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.io.*;

/**
 * Instances of this class support both reading and writing to a 
 * random access file. A random access file behaves like a large 
 * array of bytes stored in the file system. There is a kind of cursor, 
 * or index into the implied array, called the <em>file pointer</em>; 
 * input operations read bytes starting at the file pointer and advance 
 * the file pointer past the bytes read. If the random access file is 
 * created in read/write mode, then output operations are also available; 
 * output operations write bytes starting at the file pointer and advance 
 * the file pointer past the bytes written. Output operations that write 
 * past the current end of the implied array cause the array to be 
 * extended. The file pointer can be read by the 
 * <code>getFilePointer</code> method and set by the <code>seek</code> 
 * method. 
 * <p>
 * It is generally true of all the reading routines in this class that 
 * if end-of-file is reached before the desired number of bytes has been 
 * read, an <code>EOFException</code> (which is a kind of 
 * <code>IOException</code>) is thrown. If any byte cannot be read for 
 * any reason other than end-of-file, an <code>IOException</code> other 
 * than <code>EOFException</code> is thrown. In particular, an 
 * <code>IOException</code> may be thrown if the stream has been closed.
 * <P>
 * This class allows to overcome operating system limitation for maximal file 
 * size and to split the database between several partitiones.  
 */

public class RandomAccessMultiFile extends RandomAccessFile {
    static class MultiFileSegment { 
	RandomAccessFile f;
	String           name;
	long             size;
	long             offs;
    }

    MultiFileSegment segment[];
    long             currPos;
    long             currOffs;
    long             fixedSize;
    int              currSeg;

    /**
     * Creates a random access file stream to read from, and optionally 
     * to write to, a file with the specified name. A new 
     * {@link FileDescriptor} object is created to represent the 
     * connection to the file.
     * <p>
     * The mode argument must either be equal to <code>"r"</code> or 
     * <code>"rw"</code>, indicating that the file is to be opened for 
     * input only or for both input and output, respectively. The 
     * write methods on this object will always throw an 
     * <code>IOException</code> if the file is opened with a mode of 
     * <code>"r"</code>. If the mode is <code>"rw"</code> and the 
     * file does not exist, then an attempt is made to create it.
     * An <code>IOException</code> is thrown if the name argument
     * refers to a directory.
     * <p>
     * If there is a security manager, its <code>checkRead</code> method
     * is called with the <code>name</code> argument
     * as its argument to see if read access to the file is allowed.
     * If the mode is "rw", the security manager's
     * <code>checkWrite</code> method
     * is also called with the <code>name</code> argument
     * as its argument to see if write access to the file is allowed.
     *
     * @param      name   the system-dependent filename.
     * @param      mode   the access mode.
     * @exception  IllegalArgumentException  if the mode argument is not equal
     *               to <code>"r"</code> or to <code>"rw"</code>.
     * @exception  FileNotFoundException  if the file exists but is a directory
     *                   rather than a regular file, or cannot be opened or
     *                   created for any other reason
     * @exception  SecurityException         if a security manager exists and its
     *               <code>checkRead</code> method denies read access to the file
     *               or the mode is "rw" and the security manager's
     *               <code>checkWrite</code> method denies write access to the file.
     */
    public RandomAccessMultiFile(String name, String mode)
	throws FileNotFoundException, java.io.IOException
    {
	super(name, "r");
	StreamTokenizer in = 
	    new StreamTokenizer(new BufferedReader(new FileReader(getFD())));
	segment = new MultiFileSegment[0];
	int tkn = in.nextToken();
	do {
	    MultiFileSegment seg = new MultiFileSegment();
	    if (tkn != StreamTokenizer.TT_WORD && tkn != '"') { 
		throw new IOException("Multifile segment name expected");
	    }
	    seg.name = in.sval;
	    tkn = in.nextToken();
	    if (tkn != StreamTokenizer.TT_EOF) { 
		if (tkn != StreamTokenizer.TT_NUMBER) { 
		    throw new IOException("Multifile segment size expected");
		}
		seg.size = (long)in.nval*1024; // kilobytes
		tkn = in.nextToken();
	    }
	    fixedSize += seg.size;
	    seg.f = new RandomAccessFile(seg.name, mode);
	    segment = (MultiFileSegment[])AnyArray.append(segment, seg);
	} while (tkn != StreamTokenizer.TT_EOF);
	fixedSize -= segment[segment.length-1].size;
	segment[segment.length-1].size = Long.MAX_VALUE;
    }

    protected RandomAccessMultiFile(String name) 
	throws FileNotFoundException, java.io.IOException
    {
	super(name, "r");
    }

    /**
     * Creates a random access file stream to read from, and optionally 
     * to write to, the file specified by the <code>File</code> argument. 
     * A new {@link FileDescriptor} object is created to represent 
     * this file connection. 
     * <p>
     * The mode argument must either be equal to <code>"r"</code> or 
     * <code>"rw"</code>, indicating that the file is to be opened for 
     * input only or for both input and output, respectively. The 
     * write methods on this object will always throw an 
     * <code>IOException</code> if the file is opened with a mode of 
     * <code>"r"</code>. If the mode is <code>"rw"</code> and the 
     * file does not exist, then an attempt is made to create it.
     * An <code>IOException</code> is thrown if the file argument
     * refers to a directory.
     * <p>
     * If there is a security manager, its <code>checkRead</code> method
     * is called with the pathname of the <code>file</code>
     * argument as its argument to see if read access to the file is allowed.
     * If the mode is "rw", the security manager's
     * <code>checkWrite</code> method
     * is also called with the path argument
     * to see if write access to the file is allowed.
     *
     * @param      file   the file object.
     * @param      mode   the access mode.
     * @exception  IllegalArgumentException  if the mode argument is not equal
     *               to <code>"r"</code> or to <code>"rw"</code>.
     * @exception  FileNotFoundException  if the file exists but is a directory
     *                   rather than a regular file, or cannot be opened or
     *                   created for any other reason
     * @exception  SecurityException         if a security manager exists and its
     *               <code>checkRead</code> method denies read access to the file
     *               or the mode is "rw" and the security manager's
     *               <code>checkWrite</code> method denies write access to the file.
     * @see        java.io.File#getPath()
     */
    public RandomAccessMultiFile(File file, String mode)
	throws FileNotFoundException, java.io.IOException
    {
	this(file.getPath(), mode);
    }

    // 'Read' primitives
    /**
     * Reads a byte of data from this file. The byte is returned as an 
     * integer in the range 0 to 255 (<code>0x00-0x0ff</code>). This 
     * method blocks if no input is yet available. 
     * <p>
     * Although <code>RandomAccessFile</code> is not a subclass of 
     * <code>InputStream</code>, this method behaves in exactly the same 
     * way as the {@link InputStream#read()} method of 
     * <code>InputStream</code>.
     *
     * @return     the next byte of data, or <code>-1</code> if the end of the
     *             file has been reached.
     * @exception  IOException  if an I/O error occurs. Not thrown if  
     *                          end-of-file has been reached.
     */
    public int read() throws IOException { 
	byte[] b = new byte[1];
	return read(b, 0, 1) != 1 ? -1 : (b[0] & 0xFF);
    }

    /**
     * Reads up to <code>len</code> bytes of data from this file into an 
     * array of bytes. This method blocks until at least one byte of input 
     * is available. 
     * <p>
     * Although <code>RandomAccessFile</code> is not a subclass of 
     * <code>InputStream</code>, this method behaves in the exactly the 
     * same way as the {@link InputStream#read(byte[], int, int)} method of 
     * <code>InputStream</code>.
     *
     * @param      b     the buffer into which the data is read.
     * @param      off   the start offset of the data.
     * @param      len   the maximum number of bytes read.
     * @return     the total number of bytes read into the buffer, or
     *             <code>-1</code> if there is no more data because the end of
     *             the file has been reached.
     * @exception  IOException  if an I/O error occurs.
     */
    public int read(byte b[], int off, int len) throws IOException {
	int totalRead = 0;
	while (len > 0) { 
	    int toRead = len;
	    if (len + currOffs > segment[currSeg].size) {
		toRead = (int)(segment[currSeg].size - currOffs);
	    }
	    int rc = segment[currSeg].f.read(b, off, toRead);
	    if (rc >= 0) { 
		currPos += rc;
		currOffs += rc;
		totalRead += rc;
		if (currOffs == segment[currSeg].size) {
		    segment[++currSeg].f.seek(0);
		    currOffs = 0;
		}
	    } else { 
		return (totalRead == 0) ? rc : totalRead;
	    }
	    if (rc != toRead) { 
		return totalRead;
	    }
	    off += rc;
	    len -= rc;
	}
	return totalRead;
    }

    /**
     * Reads up to <code>b.length</code> bytes of data from this file 
     * into an array of bytes. This method blocks until at least one byte 
     * of input is available. 
     * <p>
     * Although <code>RandomAccessFile</code> is not a subclass of 
     * <code>InputStream</code>, this method behaves in the exactly the 
     * same way as the {@link InputStream#read(byte[])} method of 
     * <code>InputStream</code>.
     *
     * @param      b   the buffer into which the data is read.
     * @return     the total number of bytes read into the buffer, or
     *             <code>-1</code> if there is no more data because the end of
     *             this file has been reached.
     * @exception  IOException  if an I/O error occurs.
     */
    public int read(byte b[]) throws IOException {
	return read(b, 0, b.length);
    }

    /**
     * Attempts to skip over <code>n</code> bytes of input discarding the 
     * skipped bytes. 
     * <p>
     * 
     * This method may skip over some smaller number of bytes, possibly zero. 
     * This may result from any of a number of conditions; reaching end of 
     * file before <code>n</code> bytes have been skipped is only one 
     * possibility. This method never throws an <code>EOFException</code>. 
     * The actual number of bytes skipped is returned.  If <code>n</code> 
     * is negative, no bytes are skipped.
     *
     * @param      n   the number of bytes to be skipped.
     * @return     the actual number of bytes skipped.
     * @exception  IOException  if an I/O error occurs.
     */
    public int skipBytes(int n) throws IOException {
	if (n <= 0) {
	    return 0;
	}
	int skipped = n;
	while (n + currOffs > segment[currSeg].size) { 
	    currPos += segment[currSeg].size - currOffs;
	    n -= segment[currSeg].size - currOffs;
	    currSeg += 1;
	    currOffs = 0;
	}
	segment[currSeg].f.seek(n);
	long pos = segment[currSeg].f.getFilePointer();
	currOffs += pos;
	currPos += pos;
	return skipped - n + (int)pos;
    }

    // 'Write' primitives
    /**
     * Writes the specified byte to this file. The write starts at 
     * the current file pointer.
     *
     * @param      b   the <code>byte</code> to be written.
     * @exception  IOException  if an I/O error occurs.
     */
    public void write(int b) throws IOException {
	byte arr[] = new byte[1];
	arr[0] = (byte)b;
	write(arr, 0, 1);
    }

    /**
     * Force all system buffers to synchronize with the underlying
     * device.  This method returns after all modified data and
     * attributes of this FileDescriptor have been written to the
     * relevant device(s).  In particular, if this FileDescriptor
     * refers to a physical storage medium, such as a file in a file
     * system, sync will not return until all in-memory modified copies
     * of buffers associated with this FileDesecriptor have been
     * written to the physical medium.
     *
     * sync is meant to be used by code that requires physical
     * storage (such as a file) to be in a known state  For
     * example, a class that provided a simple transaction facility
     * might use sync to ensure that all changes to a file caused
     * by a given transaction were recorded on a storage medium.
     *
     * sync only affects buffers downstream of this FileDescriptor.  If
     * any in-memory buffering is being done by the application (for
     * example, by a BufferedOutputStream object), those buffers must
     * be flushed into the FileDescriptor (for example, by invoking
     * OutputStream.flush) before that data will be affected by sync.
     *
     * @exception SyncFailedException
     *	      Thrown when the buffers cannot be flushed,
     *	      or because the system cannot guarantee that all the
     *	      buffers have been synchronized with physical media.
     * @since     JDK1.1
     */
    public void sync() throws IOException {
	for (int i = segment.length; --i >= 0;) { 
	    segment[i].f.getFD().sync();
	}
    }
	
    /**
     * Writes <code>b.length</code> bytes from the specified byte array 
     * to this file, starting at the current file pointer. 
     *
     * @param      b   the data.
     * @exception  IOException  if an I/O error occurs.
     */
    public void write(byte b[]) throws IOException {
	write(b, 0, b.length); 
    }

    /**
     * Writes <code>len</code> bytes from the specified byte array 
     * starting at offset <code>off</code> to this file. 
     *
     * @param      b     the data.
     * @param      off   the start offset in the data.
     * @param      len   the number of bytes to write.
     * @exception  IOException  if an I/O error occurs.
     */
    public void write(byte b[], int off, int len) throws IOException {
	while (len > 0) { 
	    int toWrite = len;
	    if (len + currOffs > segment[currSeg].size) {
		toWrite = (int)(segment[currSeg].size - currOffs);
	    }
	    segment[currSeg].f.write(b, off, toWrite);
	    currPos += toWrite;
	    currOffs += toWrite;
	    off += toWrite;
	    len -= toWrite;
	    if (currOffs == segment[currSeg].size) {
		segment[++currSeg].f.seek(0);
		currOffs = 0;
	    }
	}
    }

    // 'Random access' stuff
    /**
     * Returns the current offset in this file. 
     *
     * @return     the offset from the beginning of the file, in bytes,
     *             at which the next read or write occurs.
     * @exception  IOException  if an I/O error occurs.
     */
    public long getFilePointer() throws IOException { 
	return currPos;
    }

    /**
     * Sets the file-pointer offset, measured from the beginning of this 
     * file, at which the next read or write occurs.  The offset may be 
     * set beyond the end of the file. Setting the offset beyond the end 
     * of the file does not change the file length.  The file length will 
     * change only by writing after the offset has been set beyond the end 
     * of the file. 
     *
     * @param      pos   the offset position, measured in bytes from the 
     *                   beginning of the file, at which to set the file 
     *                   pointer.
     * @exception  IOException  if <code>pos</code> is less than 
     *                          <code>0</code> or if an I/O error occurs.
     */
    public void seek(long pos) throws IOException {
	currSeg = 0;
	currOffs = 0;
	currPos = 0;
	while (pos > segment[currSeg].size) { 
	    currPos += segment[currSeg].size;
	    pos -= segment[currSeg].size;
	    currSeg += 1;
	}
	segment[currSeg].f.seek(pos);
	pos = segment[currSeg].f.getFilePointer();
	currOffs += pos;
	currPos += pos;
    }

    /**
     * Returns the length of this file.
     *
     * @return     the length of this file, measured in bytes.
     * @exception  IOException  if an I/O error occurs.
     */
    public long length() throws IOException { 
	return fixedSize + segment[segment.length-1].f.length();
    }

    /**
     * Sets the length of this file.
     *
     * <p> If the present length of the file as returned by the
     * <code>length</code> method is greater than the <code>newLength</code>
     * argument then the file will be truncated.  In this case, if the file
     * offset as returned by the <code>getFilePointer</code> method is greater
     * then <code>newLength</code> then after this method returns the offset
     * will be equal to <code>newLength</code>.
     *
     * <p> If the present length of the file as returned by the
     * <code>length</code> method is smaller than the <code>newLength</code>
     * argument then the file will be extended.  In this case, the contents of
     * the extended portion of the file are not defined.
     *
     * @param      newLength    The desired length of the file
     * @exception  IOException  If an I/O error occurs
     * @since      1.2
     */
    public void setLength(long newLength) throws IOException { 
	if (newLength < fixedSize) { 
	    throw new IOException("Segment of the multifile can not be removed online");
	}
	segment[segment.length-1].f.setLength(newLength - fixedSize);
    }

    /**
     * Closes this random access file stream and releases any system 
     * resources associated with the stream. A closed random access 
     * file cannot perform input or output operations and cannot be 
     * reopened.
     *
     * @exception  IOException  if an I/O error occurs.
     */
    public void close() throws IOException { 
	super.close();
	for (int i = segment.length; --i >= 0;) { 
	    segment[i].f.close();
	}
    }
}
