/* Consus - Pure Java Object Relational Database Management System
 * Boundary value for search operation in B-Tree, 
 *   also used for data propagation during B-Tree reconstruction
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

class SearchValue { 
    boolean bval; 
    long    ival;
    double  dval;
    char[]  sval;
    int     oval;

    int     oid;
    int     inclusion;
    boolean subqueryResult;

    final int compare(Page pg, int i, int type) { 
	long i8;
	int  i4;
	double d;
	switch (type) {
	  case FieldDescriptor.tpBool:
	    return (bval ? 1 : 0) - pg.data[BtreePage.firstKeyOffs + i];
	  case FieldDescriptor.tpInt1:
	    return (byte)ival - pg.data[BtreePage.firstKeyOffs + i];
	  case FieldDescriptor.tpInt2:
	    return (short)ival - Bytes.unpack2(pg.data, BtreePage.firstKeyOffs + i*2);
	  case FieldDescriptor.tpReference:
	    return oval - Bytes.unpack4(pg.data, BtreePage.firstKeyOffs + i*4);
	  case FieldDescriptor.tpInt4:
	    i4 = Bytes.unpack4(pg.data, BtreePage.firstKeyOffs + i*4);
	    return ival < i4 ? -1 : ival == i4 ? 0 : 1;
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	  case FieldDescriptor.tpInt8:
	    i8 = Bytes.unpack8(pg.data, BtreePage.firstKeyOffs + i*8);
	    return ival < i8 ? -1 : ival == i8 ? 0 : 1;
	  case FieldDescriptor.tpReal4:
	    d =  Float.intBitsToFloat(Bytes.unpack4(pg.data, 
						       BtreePage.firstKeyOffs + i*4));
	    return dval < d ? -1 : dval == d ? 0 : 1;
	  case FieldDescriptor.tpReal8:
	    d = Double.longBitsToDouble(Bytes.unpack8(pg.data, 
							  BtreePage.firstKeyOffs + i*8));
	    return dval < d ? -1 : dval == d ? 0 : 1;
	}
	Assert.failed("Invalid type");
	return 0;
    }


    final int compareStr(Page pg, int i) { 
	int alen = sval.length;
	int blen = BtreePage.getKeyStrSize(pg, i);
	int minlen = alen < blen ? alen : blen;
	int offs = BtreePage.getKeyStrOffs(pg, i) + BtreePage.firstKeyOffs;
	byte[] b = pg.data;
	for (int j = 0; j < minlen; j++) { 
	    int diff = sval[j] - (char)Bytes.unpack2(b, offs);
	    if (diff != 0) { 
		return diff;
	    }
	    offs += 2;
	}
	return alen - blen;
    }

    final void getStr(Page pg, int i) { 
	int len = BtreePage.getKeyStrSize(pg, i);
	int offs = BtreePage.firstKeyOffs + BtreePage.getKeyStrOffs(pg, i);
	sval = new char[len];
	for (int j = 0; j < len; j++) { 
	    sval[j] = (char)Bytes.unpack2(pg.data, offs);
	    offs += 2;
	}
    }


    final void extract(byte[] rec, int offs, int type) { 
	int    i, len;
	char[] chars;

	switch (type) {
	  case FieldDescriptor.tpBool:
	    bval = rec[offs] != 0;
	    break;
	  case FieldDescriptor.tpInt1:
	    ival = rec[offs];
	    break;
	  case FieldDescriptor.tpInt2:
	    ival = Bytes.unpack2(rec, offs);
	    break;
	  case FieldDescriptor.tpInt4:
	    ival = Bytes.unpack4(rec, offs);
	    break;
	  case FieldDescriptor.tpReference:
	    oval = Bytes.unpack4(rec, offs);
	    break;
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	  case FieldDescriptor.tpInt8:
	    ival = Bytes.unpack8(rec, offs);
	    break;
	  case FieldDescriptor.tpReal4:
	    dval = Float.intBitsToFloat(Bytes.unpack4(rec, offs));
	    break;
	  case FieldDescriptor.tpReal8:
	    dval = Double.longBitsToDouble(Bytes.unpack8(rec, offs));
	    break;
	  case FieldDescriptor.tpString:
	    len = Varying.getSize(rec, offs);
	    offs = Varying.getOffs(rec, offs);
	    chars = new char[len];
	    for (i = 0; i < len; i++) { 
		chars[i] = (char)Bytes.unpack2(rec, offs);
		offs += 2;
	    }
	    sval = chars;
	    break;
	  default:
	    Assert.failed("Invalid type");
	}
    }

    final void pack(Page pg, int i, int type) { 
	byte[] dst = pg.data;
	switch (type) { 
	  case FieldDescriptor.tpBool:
	    dst[BtreePage.firstKeyOffs + i] = (byte)(bval ? 1 : 0);
	    break;
	  case FieldDescriptor.tpInt1:
	    dst[BtreePage.firstKeyOffs + i] = (byte)ival;
	    break;
	  case FieldDescriptor.tpInt2:
	    Bytes.pack2(dst, BtreePage.firstKeyOffs + i*2, (short)ival);
	    break;
	  case FieldDescriptor.tpInt4:
	    Bytes.pack4(dst, BtreePage.firstKeyOffs + i*4, (int)ival);
	    break;
	  case FieldDescriptor.tpReference:
	    Bytes.pack4(dst, BtreePage.firstKeyOffs + i*4, oval);
	    break;
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	  case FieldDescriptor.tpInt8:
	    Bytes.pack8(dst, BtreePage.firstKeyOffs + i*8, ival);
	    break;
	  case FieldDescriptor.tpReal4:
	    Bytes.pack4(dst, BtreePage.firstKeyOffs + i*4, Float.floatToIntBits((float)dval));
	    break;
	  case FieldDescriptor.tpReal8:
	    Bytes.pack8(dst, BtreePage.firstKeyOffs + i*8, Double.doubleToLongBits(dval));
	    break;
	  default:
	    Assert.failed("Invalid type");
	}
	Bytes.pack4(dst, BtreePage.firstKeyOffs + (BtreePage.maxItems - i - 1)*4, oid);
    }
}

	    

