/* Consus - Pure Java Object Relational Database Management System
 * Select implementation
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;
import java.util.Arrays;

class SelectNode extends StmtNode { 
    TableDescriptor  table;
    ExprNode         condition;
    OrderNode        order;
    IteratorNode     iterator;
    ConsusConnection db;
    boolean          forUpdate;
    int              selectType;

    SelectNode       unionWith;
    boolean          isUnion; 
    boolean          isSetUnion; // union with eliminating of duplicates
    boolean          existsOptimization;

    TableNode        tables[];
    ColumnNode       columns[];
    ColumnNode       resultColumns[];
    ColumnNode       groupColumns[];
    
    int              outDep;
    SelectNode       outer;
    
    OrderNode        groupBy;
    ExprNode         having;

    ColumnNode       groups;
    int              traceTableNo;
    int              searchTableNo;

    FieldDescriptor  indexField; // index of the field used in query which also specifies order of the columns

    int              subqueryExecCount;
    Object[]         subqueryResult;

    static final int SELECT_DEFAULT  = 0;
    static final int SELECT_DISTINCT = 1;
    static final int SELECT_ALL      = 2;
    static final int SELECT_COUNT    = 3;

    static class DistinctOrderNode extends OrderNode {
        int nColumns;
        DistinctOrderNode(int nColumns) { 
            this.nColumns = nColumns;
        }
        
        int compareTuples(Object[] t1, Object[] t2) { 
            for (int i = 0, n = nColumns; i < n; i++) { 
                int diff = ((Comparable)t1[i]).compareTo(t2[i]);
                if (diff != 0) { 
                    return diff;
                }
            }
            return 0;
        }
    }
       

    void setColumns(ConsusStatement stmt, byte[] rec) throws SQLException { 
	tables[searchTableNo-1].rec = rec;
	for (int j = 0; j < columns.length; j++) {
	    if (columns[j].tableDep == searchTableNo) { 
		columns[j].evaluate(stmt);
	    }
	}
    }


    int[] applyIndex(ConsusStatement stmt, ExprNode expr, int tableNo, boolean toAll) 
	throws SQLException 
    { 
	SearchValue s;
	ExprNode cond = expr;
	ExprNode and = null;
	ExprNode refine = null;
	Object[] set = null;
	boolean searchConstant = false;
	int[] mergeWith = null;
	if (expr.tag == opBoolAnd) {
	    refine = and = ((BinOpNode)expr).right;
	    mergeWith = applyIndex(stmt, and, tableNo, true);
	    if (mergeWith != null) { 
		and = null;
	    } else if (toAll) { 
		return null;
	    }
	    expr = ((BinOpNode)expr).left;
	}
	SearchContext ctx;
	FieldDescriptor fd = null;
	if (expr instanceof BinOpNode) { 
	    ExprNode left = ((BinOpNode)expr).left;
	    ExprNode right = ((BinOpNode)expr).right;
	    int leftDep = left.dependsOn();
	    int rightDep = right.dependsOn();
	    boolean swap = false;
	    if (leftDep < rightDep) { 
		swap = true;
		ExprNode tmp = left;
		left = right;
		right = tmp;
		int dep = leftDep;
		leftDep = rightDep;
		rightDep = dep;
	    }
	    Assert.that(leftDep == tableNo && rightDep <= tableNo);
	    if (rightDep == tableNo) { 
		return null;
	    } else if (rightDep == 0) { 
		searchConstant = true;
	    }
	    if (left instanceof ColumnNode) { 
		fd = ((ColumnNode)left).field;
	    } else if (left instanceof LoadNode && ((LoadNode)left).base instanceof TableNode) { 
		fd = ((LoadNode)left).field;
	    }
	    if (fd == null || fd.index == 0) { 
		return null;
	    }
	    s = new SearchValue();
	    ctx = new SearchContext();
	    ctx.condition = and;
	    switch (expr.tag) { 
	      case opBoolEq:
		s.bval = right.evaluateBool(stmt);
		ctx.firstKey = ctx.lastKey = s;
		s.inclusion = 1;
		break;
	      case opIntEq:
		s.ival = right.evaluateInt(stmt);
		ctx.firstKey = ctx.lastKey = s;
		s.inclusion = 1;
		break;
	      case opRealEq:
		s.dval = right.evaluateReal(stmt);
		ctx.firstKey = ctx.lastKey = s;
		s.inclusion = 1;
		break;
	      case opStrEq:
		s.sval = right.evaluateStr(stmt).toCharArray();
		ctx.firstKey = ctx.lastKey = s;
		s.inclusion = 1;
		break;
	      case opOidEq:
		s.oval = right.evaluateOid(stmt);
		ctx.firstKey = ctx.lastKey = s;
		s.inclusion = 1;
		break;
	      case opIntGt:
		s.ival = right.evaluateInt(stmt);
		ctx.firstKey = s;
		ctx.lastKey = null;
		s.inclusion = 0;		    
		break;
	      case opRealGt:
		s.dval = right.evaluateReal(stmt);
		ctx.firstKey = s;
		ctx.lastKey = null;
		s.inclusion = 0;
		break;
	      case opStrGt:
		s.sval = right.evaluateStr(stmt).toCharArray();
		ctx.firstKey = s;
		ctx.lastKey = null;
		s.inclusion = 0;
		break;
	      case opIntGe:
		s.ival = right.evaluateInt(stmt);
		ctx.firstKey = s;
		ctx.lastKey = null;
		s.inclusion = 1;
		break;
	      case opRealGe:
		s.dval = right.evaluateReal(stmt);
		ctx.firstKey = s;
		ctx.lastKey = null;
		s.inclusion = 1;
		break;
	      case opStrGe:
		s.sval = right.evaluateStr(stmt).toCharArray();
		ctx.firstKey = s;
		ctx.lastKey = null;
		s.inclusion = 1;
		break;
	      case opIntLt:
		s.ival = right.evaluateInt(stmt);
		ctx.firstKey = null;
		ctx.lastKey = s;
		s.inclusion = 0;
		break;
	      case opRealLt:
		s.dval = right.evaluateReal(stmt);
		ctx.firstKey = null;
		ctx.lastKey = s;
		s.inclusion = 0;
		break;
	      case opStrLt:
		s.sval = right.evaluateStr(stmt).toCharArray();
		ctx.firstKey = null;
		ctx.lastKey = s;
		s.inclusion = 0;
		break;
	      case opIntLe:
		s.ival = right.evaluateInt(stmt);
		ctx.firstKey = null;
		ctx.lastKey = s;
		s.inclusion = 1;
		break;
	      case opRealLe:
		s.dval = right.evaluateReal(stmt);
		ctx.firstKey = null;
		ctx.lastKey = s;
		s.inclusion = 1;
		break;
	      case opStrLe:
		s.sval = right.evaluateStr(stmt).toCharArray();
		ctx.firstKey = null;
		ctx.lastKey = s;
		s.inclusion = 1;
		break;
	      case opScanSetInt:
	      case opScanSetReal:
	      case opScanSetStr:
	      case opScanSetRef:
		set = right.evaluateSet(stmt);
		break;
	      default:
		return null;
	    }
	    if (swap) { 
		s = ctx.firstKey;
		ctx.firstKey = ctx.lastKey;
		ctx.lastKey = s;
	    }
	} else if (expr instanceof TripleOpNode) { 
	    TripleOpNode cmp = (TripleOpNode)expr;
	    int dep2, dep3 = 0;
	    if (cmp.o1 instanceof ColumnNode) { 
		fd = ((ColumnNode)cmp.o1).field;
	    } else if (cmp.o1 instanceof LoadNode && ((LoadNode)cmp.o1).base instanceof TableNode) { 
		fd = ((LoadNode)cmp.o1).field;
	    }
	    if (fd == null || fd.index == 0) { 
		return null;
	    }
	    if (cmp.o1.dependsOn() != tableNo 
		|| (dep2 = cmp.o2.dependsOn()) >= tableNo
		|| (cmp.o3 != null && (dep3 = cmp.o2.dependsOn()) >= tableNo))
	    { 
		return null;
	    }
	    if (dep2 + dep3 == 0) { 
		searchConstant = true;
	    }
	    s = new SearchValue();
	    ctx = new SearchContext();
	    char esc = '\\';
	    char[] pattern;
	    int i;
	    ctx.condition = and;
	    switch (cmp.tag) { 
	      case opIntBetween:
		ctx.firstKey = s;
		s.ival = cmp.o2.evaluateInt(stmt);
		s.inclusion = 1;
		s = new SearchValue();
		ctx.lastKey = s;
		s.ival = cmp.o3.evaluateInt(stmt);
		s.inclusion = 1;
		break;
	      case opRealBetween:
		ctx.firstKey = s;
		s.dval = cmp.o2.evaluateReal(stmt);
		s.inclusion = 1;
		s = new SearchValue();
		ctx.lastKey = s;
		s.dval = cmp.o3.evaluateReal(stmt);
		s.inclusion = 1;
		break;
	      case opStrBetween:	    
		ctx.firstKey = s;
		s.sval = cmp.o2.evaluateStr(stmt).toCharArray();
		s.inclusion = 1;
		s = new SearchValue();
		ctx.lastKey = s;
		s.sval = cmp.o3.evaluateStr(stmt).toCharArray();
		s.inclusion = 1;
		break;
	      case opStrLikeEsc:
		esc = cmp.o3.evaluateStr(stmt).charAt(0);
		// no break
	      case opStrLike:
		ctx.firstKey = s;
		s.inclusion = 1;
		pattern = cmp.o2.evaluateStr(stmt).toCharArray();
		i = findWildcard(pattern, esc);
		s.sval = pattern;
		if (i < 0) { 
		    ctx.lastKey = s;
		} else if (i > 0 && pattern[i-1] != (char)0xffff) {
		    int j;
		    char[] prefix = new char[i];
		    for (j = 0; j < i; j++) { 
			prefix[j] = pattern[j];
		    }
		    s.sval = prefix;
		    ctx.firstKey = s;
		    s = new SearchValue();
		    prefix = new char[i];
		    for (j = 0; j < i; j++) { 
			prefix[j] = pattern[j];
		    }
		    prefix[i-1] += 1;
		    s.sval = prefix;
		    s.inclusion = 0;
		    ctx.lastKey = s;
		    ctx.condition = cond; 
		} else { 
		    return null;
		}
		break;
	      default:
		return null;
	    }
	} else { 
	    return null;
	}
	ctx.stmt = stmt;	
	int saveSelectCount = stmt.cursor.nSelected;
	int saveSelectionLimit = stmt.cursor.selectionLimit;
	if (mergeWith != null) { 
	    stmt.cursor.selectionLimit = Integer.MAX_VALUE;
	}
	stmt.cursor.nSelected = 0;
	if (stmt.cursor.selection == null) { 
	    stmt.cursor.selection = new int[256];
	}
	stmt.select = this;
	searchTableNo = tableNo;
	stmt.cursor.table = tables[tableNo-1].table;
	if (db.dbTraceEnabled && traceTableNo < tableNo) { 
	    System.out.println("Index search through table '" + tables[tableNo-1].name + "'"
			       + "\n\texpression: " + expr.dump() 
			       + (and != null ? ("\n\tfilter: " + and.dump()) 
				  : refine != null ? ("\n\tindexed merge: " + refine.dump()) : ""));
	    if (!toAll) { 
		traceTableNo = tableNo;
	    }
	} 
	if (set != null) { 
	    int i;
	    s = new SearchValue();
	    ctx.firstKey = ctx.lastKey = s;
	    s.inclusion = 1;
	    for (i = 0; i < set.length; i++) { 
		Object setElem = set[i];
		switch (expr.tag) { 
		  case opScanSetInt:
		    s.ival = ((Long)setElem).longValue();
		    break;
		  case opScanSetReal:
		    s.dval = ((Double)setElem).doubleValue();
		    break;
		  case opScanSetStr:
		    s.sval = ((String)setElem).toCharArray();
		    break;
		  case opScanSetRef:
		    s.oval = (int)((ConsusRef)setElem).getOid();
		    break;
		}
		Btree.find(stmt.db, fd.index, ctx);
	    }
	} else { 
	    Btree.find(stmt.db, fd.index, ctx);
	}
	stmt.select = null;
	int selection[] = stmt.cursor.selection;
	int[] projection;
	if (mergeWith != null) { 
	    Arrays.sort(mergeWith);
	    Arrays.sort(selection, 0, stmt.cursor.nSelected);
	    int n = 0;
	    for (int i1 = 0, i2 = 0, n1 = mergeWith.length, n2 = stmt.cursor.nSelected;
		 i1 < n1 && i2 < n2;)
	    {
		if (mergeWith[i1] < selection[i2]) { 
		    i1 += 1;
		} else if (mergeWith[i1] > selection[i2]) { 
		    i2 += 1;
		} else { 
		    selection[n++] = mergeWith[i1];
		    if (n >= saveSelectionLimit) { 
			break;
		    }
		    i1 += 1;
		    i2 += 1;
		}
	    }
	    projection = new int[n];
	    System.arraycopy(selection, 0, projection, 0, n);
	    stmt.cursor.selectionLimit = saveSelectionLimit;
	} else { 
	    if (tableNo == 1) { 
		indexField = fd;
	    }
	    projection = new int[stmt.cursor.nSelected];
	    System.arraycopy(selection, 0, projection, 0, projection.length);
	    if (searchConstant && and == null && !toAll) { 
		tables[tableNo-1].projection = projection;
	    }
	}
	stmt.cursor.nSelected = saveSelectCount;
	return projection;
    }


    void joinTables(int tableNo, ExprNode condition, ConsusStatement stmt) 
	throws SQLException 
    { 
	int i, j;
	if (tableNo == tables.length) {
	    if (condition == null || condition.evaluateBool(stmt)) { 
		if (selectType == SELECT_COUNT) { 
		    stmt.cursor.nSelected += 1;
		} else { 
		    Object[] tuple = new Object[columns.length];
		    for (i = columns.length; --i >= 0;) {  
			tuple[i] = columns[i].getObject();
		    }
		    stmt.cursor.addTuple(tuple);
		}
	    }
	    return;
	} 
	
	TableNode tnode = tables[tableNo];
	TableDescriptor table = tnode.table;
	tableNo += 1;
	ExprNode filter = null;
	BinOpNode and = null, prev = null;
	try { 
	    if (condition != null) { 
		if (condition.tag == opBoolAnd) {
		    and = (BinOpNode)condition;
		    BinOpNode pprev = null; // previous previuos
		    while (and.left.dependsOn() <= tableNo) { 
			if (and.right.tag == opBoolAnd) { 
			    pprev = prev;
			    prev = and;
			    and = (BinOpNode)and.right;
			} else { 
			    if (and.right.dependsOn() <= tableNo) { 
				filter = condition;
				condition = null;
			    } else { 
				if (prev != null) { 
				    prev.right = and.left;
				    filter = condition;
				} else { 
				    filter = and.left;
				}
				condition = and.right; 
			    }
			    break;
			}
		    }	
		    if (filter == null && prev != null) { 
			if (pprev != null) { 
			    pprev.right = prev.left;
			    and = prev;
			    prev = pprev;
			    filter = condition;
			    condition = and.right;
			} else { 
			    filter = prev.left;
			    condition = and;
			    prev = null;
			}
		    }
		} else if (condition.dependsOn() <= tableNo) { 
		    filter = condition;
		    condition = null;
		}
		if (filter != null && !tnode.sequentialSearch) { 
		    int[] selection = tnode.projection != null 
			? tnode.projection : applyIndex(stmt, filter, tableNo, false);
		    if (selection != null) { 
			for (i = 0; i < selection.length; i++) { 
			    tnode.rec = stmt.db.get(selection[i]);
			    for (j = 0; j < columns.length; j++) {
				if (columns[j].tableDep == tableNo) { 
				    columns[j].evaluate(stmt);
				}
			    }
			    joinTables(tableNo, condition, stmt);
			}
			return;
		    } 
		    tnode.sequentialSearch = true;
		}
	    }
	    if (db.dbTraceEnabled && traceTableNo < tableNo) { 
		System.out.println("Sequential search through table '" + tnode.name + "'"
				   + (filter != null ? ("\n\tfilter: " + filter.dump()) : "")
				   + (condition != null ? ("\n\tcondition: " + condition.dump()) : ""));
		traceTableNo = tableNo;
	    } 
	    if (table.superClass != null || table.nSubclasses != 0) { 
		int oid = table.getRootTable().firstRow;
		while (oid != 0) { 
		    byte[] rec = stmt.db.get(oid);
		    int tableId = Record.getType(rec, 0);
		    if (table.isSuperclassFor(db.getTableByOid(tableId))) { 
			tnode.rec = rec;
			for (j = 0; j < columns.length; j++) {
			    if (columns[j].tableDep == tableNo) { 
				columns[j].evaluate(stmt);
			    }
			}
			if (filter == null || filter.evaluateBool(stmt)) { 
			    joinTables(tableNo, condition, stmt);
			}
		    }
		    oid = Record.getNext(rec, 0);
		}
	    } else { 
		int oid = table.firstRow;
		while (oid != 0) { 
		    stmt.rec = tnode.rec = stmt.db.get(oid);
		    for (j = 0; j < columns.length; j++) {
			if (columns[j].tableDep == tableNo) { 
			    columns[j].evaluate(stmt);
			}
		    }
		    if (filter == null || filter.evaluateBool(stmt)) { 
			joinTables(tableNo, condition, stmt);
		    }
		    oid = Record.getNext(tnode.rec, 0);
		}
	    }	 
	} finally {    
	    if (prev != null) { 
		prev.right = and;
	    }
	}
    }
    
    int dependsOn() { return outDep; }

    boolean evaluateBool(ConsusStatement stmt) throws SQLException { 
	Object[] result = executeSubquery(stmt);
	if (result.length != 1) { 
	    throw new ConsusNotSingleException("Subquery doesn't return the single value");
	}
	return ((Boolean)result[0]).booleanValue();
    }

    long evaluateInt(ConsusStatement stmt) throws SQLException { 
	Object[] result = executeSubquery(stmt);
	if (result.length != 1) { 
	    throw new ConsusNotSingleException("Subquery doesn't return the single value");
	}
	return ((Long)result[0]).longValue();
    }

    double evaluateReal(ConsusStatement stmt) throws SQLException { 
	Object[] result = executeSubquery(stmt);
	if (result.length != 1) { 
	    throw new ConsusNotSingleException("Subquery doesn't return the single value");
	}
	return ((Double)result[0]).doubleValue();
    }

    String evaluateStr(ConsusStatement stmt) throws SQLException { 
	Object[] result = executeSubquery(stmt);
	if (result.length != 1) { 
	    throw new ConsusNotSingleException("Subquery doesn't return the single value");
	}
	return (String)result[0];
    }
    
    int evaluateOid(ConsusStatement stmt) throws SQLException { 
	Object[] result = executeSubquery(stmt);
	if (result.length != 1) { 
	    throw new ConsusNotSingleException("Subquery doesn't return the single value");
	}
	return (int)((ConsusRef)result[0]).getOid();
    }

    Object[] evaluateSet(ConsusStatement stmt) throws SQLException { 
	return executeSubquery(stmt);
    }

    Object[] executeSubquery(ConsusStatement stmt) throws SQLException { 
	if (stmt.execCount != subqueryExecCount || outDep != 0) { 
	    ConsusResultSet saveCursor = stmt.cursor;	
	    try { 
		stmt.cursor = new ConsusResultSet(stmt);
		stmt.cursor.tupleColumns = resultColumns;
		executeStmt(stmt);
		subqueryExecCount = stmt.execCount;
		int n = stmt.cursor.nSelected;
		subqueryResult = new Object[n];
		for (int i = 0; i < n; i++) { 
		    subqueryResult[i] = stmt.cursor.tuples[i][0];
		}
	    } finally { 
		stmt.cursor = saveCursor;
	    }
	}
	return subqueryResult;
    }

    int executeStmt(ConsusStatement stmt) throws SQLException { 
	int i, j, n;
	db = stmt.db;
	if (!isUnion) { 
	    stmt.cursor.reset(true);
	    db.beginTransaction(forUpdate); 
	}
	int nSelectedBefore = stmt.cursor.nSelected;
	indexField = null;
	ColumnNode group;

	if (columns != null) { 
	    SelectNode save = stmt.select;
	    stmt.select = null;
	    if (!isUnion) { 
		stmt.cursor.open(resultColumns);
	    }
	    for (i = tables.length; --i >= 0;) { 
		tables[i].projection = null;
		tables[i].sequentialSearch = false;
	    }
	    for (i = columns.length; --i >= 0;) {  
		if (columns[i].tableDep == 0) { 
		    columns[i].evaluate(stmt);
		}
	    }
            int saveGroupBeg = stmt.groupBegin;
            int saveGroupEnd = stmt.groupEnd;
	    stmt.groupBegin = stmt.groupEnd = 0;
	    if (groupBy != null) { 
		OrderNode gp, saveOrder = order;
		try { 
		    order = groupBy;
		    joinTables(0, condition, stmt);
		    int nSelected = stmt.cursor.nSelected;
		    if (order.next != null || indexField == null ||
			columns[order.columnNo].field != indexField)
		    {
			qSort(stmt.cursor.tuples, nSelectedBefore, nSelected - nSelectedBefore);
		    }
		    Object[][] tuples = stmt.cursor.tuples;
		    for (n = nSelectedBefore, i = n; i < nSelected;) {  
			for (j = columns.length; --j >= nSelectedBefore;) { 
			    columns[j].assign(tuples[i][j]);
			}
			stmt.groupBegin = i;
		      locate_group:
			while (true) { 
			    if (++i == nSelected) { 
				break locate_group;
			    }
			    for (gp = groupBy; gp != null; gp = gp.next) { 
				if (!tuples[i-1][gp.columnNo].equals(tuples[i][gp.columnNo])) { 
				    break locate_group;
				} 
			    }
			}
			stmt.groupEnd = i;
			for (group = groups; group != null; group = group.next) { 
			    group.evaluate(stmt);
			}
			if (having == null || having.evaluateBool(stmt)) { 
			    for (j = 0; j < groupColumns.length; j++) { 
				tuples[n][j] = groupColumns[j].getObject();
			    }
			    n += 1;
			}
		    }
		    stmt.cursor.nSelected = n;
		} finally { 
		    order = saveOrder;
		}
		if (order != null) { 
		    if (db.dbTraceEnabled) { 
			System.out.println("Sorting result");
		    } 
		    qSort(stmt.cursor.tuples, nSelectedBefore, 
			  stmt.cursor.nSelected - nSelectedBefore);
		}
	    } else { 
		if (selectType == SELECT_COUNT && condition == null && tables.length == 1) { 
		    int nRows = tables[0].table.nRows;
		    if (isUnion && existsOptimization) { 
			stmt.cursor.tuples[0][0] = 
			    new Long(stmt.cursor.nSelected + nRows);
		    } else { 
			Object[] tuple = new Object[1];
			tuple[0] = new Long(nRows);
			stmt.cursor.addTuple(tuple);
		    }
		} else { 		    
		    joinTables(0, condition, stmt);
		    if (selectType == SELECT_COUNT) { 
			if (isUnion && existsOptimization) { 
			    stmt.cursor.tuples[0][0] = 
				new Long(stmt.cursor.nSelected 
					 + ((Long)stmt.cursor.tuples[0][0]).longValue());
			} else { 
			    Object[] tuple = new Object[1];
			    tuple[0] = new Long(stmt.cursor.nSelected);
			    stmt.cursor.nSelected = nSelectedBefore;
			    stmt.cursor.addTuple(tuple);
			}
		    } else if (groups != null) { 
			stmt.groupEnd = stmt.cursor.nSelected;
			for (group = groups; group != null; group = group.next) { 
			    group.evaluate(stmt);
			}
			Object[] tuple = new Object[groupColumns.length];
			for (i = groupColumns.length; --i >= 0;) { 
			    tuple[i] = groupColumns[i].getObject();
			}
			stmt.cursor.nSelected = nSelectedBefore;
			stmt.cursor.addTuple(tuple);		
		    } else {
			if (order != null) {	
			    if (order.next != null || indexField == null ||
				columns[order.columnNo].field != indexField ||
				(order.columnNo < resultColumns.length 
				 && resultColumns[order.columnNo].field != indexField))
			    {
				if (db.dbTraceEnabled) { 
				    System.out.println("Sorting result");
				} 
				qSort(stmt.cursor.tuples, nSelectedBefore, 
				      stmt.cursor.nSelected - nSelectedBefore);
			    }
			}		
		    }
		}
	    }
            stmt.groupBegin = saveGroupBeg;
            stmt.groupEnd = saveGroupEnd;
	    if (selectType == SELECT_DISTINCT || isSetUnion) { 
                if (db.dbTraceEnabled) { 
                    System.out.println("Sorting union tables");
                } 
                OrderNode saveOrder = order;
                order = new DistinctOrderNode(resultColumns.length);
                qSort(stmt.cursor.tuples, 0, stmt.cursor.nSelected);
                order = saveOrder;
                nSelectedBefore = 0;

		if ((n = stmt.cursor.nSelected) > nSelectedBefore + 1) { 
		    if (db.dbTraceEnabled) { 
			System.out.println("Eliminating duplicates");
		    } 
		    
		    int l = resultColumns.length;
		    Object[][] tuples = stmt.cursor.tuples;
		    for (i = nSelectedBefore, j = i + 1; j < n; j++) { 
			for (int k = 0; k < l; k++) { 
			    if (!tuples[i][k].equals(tuples[j][k])) { 
				tuples[++i] = tuples[j];
				break;
			    } 
			}
		    }
		    stmt.cursor.nSelected = i+1;
		}
	    } 
	    stmt.select = save;
	} else { 
	    TableDescriptor saveTable = stmt.cursor.table;
	    if (!isUnion) { 
		stmt.cursor.open(table, forUpdate);
	    } else { 
		stmt.cursor.table = table;
	    }
	    if (iterator != null) { 
		if (selectType == SELECT_DISTINCT) { 
		    stmt.cursor.enableCheckForDuplicates();
		}
		iterator.iterate(stmt, condition);
	    } else { 
		if (condition == null) { 
		    if (order != null && order.next == null && order.field.index != 0) { 
			if (order.ascent) { 
			    Btree.traverseForward(db, order.field.index, stmt, null);
			} else { 
			    Btree.traverseBackward(db, order.field.index, stmt, null);
			}
			if (!isUnion || unionWith != null) { 
			    return stmt.cursor.nSelected;
			}
		    }
		    int oid = table.getRootTable().firstRow;
		    while (oid != 0) {
			if (stmt.cursor.checkType(oid)) { 
			    if (!stmt.cursor.add(oid)) { 
				break;
			    }
			}
			long pos = db.getPos(oid);
			int offs = (int)pos & (Page.pageSize-1);
			Page pg = db.pool.getPage(pos - offs);
			oid = Record.getNext(pg.data, offs & ~ConsusConnection.dbFlagsMask);
			db.pool.unfix(pg);
		    }			
		} else {
		    if (!isIndexApplicable(stmt, condition, false)) { 	
			if (db.dbTraceEnabled) { 
			    System.out.println("Perform sequential search");
			} 
			if (order != null && order.next == null && order.field.index != 0) { 
			    if (order.ascent) { 
				Btree.traverseForward(db, order.field.index, stmt, condition);
			    } else { 
				Btree.traverseBackward(db, order.field.index, stmt, condition);
			    }
			    if (!isUnion && unionWith == null) { 
				return stmt.cursor.nSelected;
			    }
			} else { 
			    int oid = table.getRootTable().firstRow;
			    while (oid != 0) {
				stmt.rec = null;
				if (stmt.cursor.checkType(oid) && stmt.evaluate(condition, oid)) { 
				    if (!stmt.cursor.add(oid)) { 
					break;
				    }
				}
				if (stmt.rec == null) { 
				    long pos = db.getPos(oid);
				    int offs = (int)pos & (Page.pageSize-1);
				    Page pg = db.pool.getPage(pos - offs);
				    oid = Record.getNext(pg.data, offs & ~ConsusConnection.dbFlagsMask);
				    db.pool.unfix(pg);
				} else { 
				    oid = Record.getNext(stmt.rec, 0);
				}
			    }		
			    stmt.rec = null;
			}
		    }
		}
	    }
	    if (order != null) { 
		if (isUnion || order.next != null || order.field != indexField) { 
		    quickSort(stmt.cursor.selection, stmt.cursor.nSelected);
		}
	    }
	    if (isUnion) { 
		stmt.cursor.table = saveTable;
	    }
	}
	if (unionWith != null) { 
	    unionWith.executeStmt(stmt);
	}
	return stmt.cursor.nSelected;
    }
    

    final static int findWildcard(char[] arr, char esc) { 
	for (int i = 0; i < arr.length; i++) { 
	    if (arr[i] == '%' || arr[i] == '_' || arr[i] == esc) { 
		return i;
	    }
	}
	return -1;
    }

    final SearchValue getSearchParameter(ConsusStatement stmt, ExprNode expr) 
       throws SQLException 
    { 
	SearchValue s = new SearchValue();
	switch (expr.tag) { 
	  case opFalse:
	    s.bval = false;
	    break;
	  case opTrue:
	    s.bval = true;
	    break;
	  case opParameter:
	  {
	      ParameterNode param = (ParameterNode)expr;
	      switch (param.type) { 
		case tpBool:
		  s.bval = param.bval;
		  break;
		case tpInt:
		  s.ival = param.ival;
		  break;
		case tpRef:
		  s.oval = param.oid;
		  break;
		case tpReal:
		  s.dval = param.dval;
		  break;
		case tpStr:
		  s.sval = param.sval.toCharArray();
		  break;
		default:
		  return null;
	      }
	      break;
	  }
	  case opStrConst:
	    s.sval = expr.evaluateStr(stmt).toCharArray();
	    break;
	  case opIntConst:
	    s.ival = expr.evaluateInt(stmt);
	    s.oval = (int)s.ival;
	    break;
	  case opRealConst:
	    s.dval = expr.evaluateReal(stmt);
	    break;
	  case opSelect:
	    if (((SelectNode)expr).outDep != 0) { 
		return null;
	    }
	    s.subqueryResult = true;
	    break;
	  default:
	    return null;
	}
	return s;
    }


    final SearchContext createSearchContext(ConsusStatement stmt, ExprNode expr, SearchValue s)
	throws SQLException
    { 
	int i;
	char esc = '\\';
	char[] pattern;
	SearchContext ctx = new SearchContext();
	ctx.stmt = stmt;
	switch (expr.tag) { 
	  case opBoolEq:
	  case opIntEq:
	  case opRealEq:
	  case opStrEq:
	  case opOidEq:
	    ctx.firstKey = ctx.lastKey = s;
	    s.inclusion = 1;
	    break;
	  case opIntGt:
	  case opRealGt:
	  case opStrGt:
	    ctx.firstKey = s;
	    ctx.lastKey = null;
	    s.inclusion = 0;
	    break;
	  case opIntGe:
	  case opRealGe:
	  case opStrGe:
	    ctx.firstKey = s;
	    ctx.lastKey = null;
	    s.inclusion = 1;
	    break;
	  case opIntLt:
	  case opRealLt:
	  case opStrLt:
	    ctx.firstKey = null;
	    ctx.lastKey = s;
	    s.inclusion = 0;
	    break;
	  case opIntLe:
	  case opRealLe:
	  case opStrLe:
	    ctx.firstKey = null;
	    ctx.lastKey = s;
	    s.inclusion = 1;
	    break;
	  case opIntBetween:
	  case opRealBetween:
	  case opStrBetween:	    
	    ctx.firstKey = s;
	    s.inclusion = 1;
	    s = getSearchParameter(stmt, ((TripleOpNode)expr).o3);
	    if (s == null) { 
		return null;
	    }
	    s.inclusion = 1;
	    ctx.lastKey = s;
	    break;
	  case opScanSetInt:
	  case opScanSetReal:
	  case opScanSetStr:
	  case opScanSetRef:
	    s.inclusion = 1;
	    ctx.firstKey = ctx.lastKey = s;
	    break;
	  case opStrLikeEsc:
	    if (((TripleOpNode)expr).o3.tag != opStrConst) { 
		return null;
	    }
	    esc = ((StrLiteralNode)((TripleOpNode)expr).o3).value.charAt(0);
	    // no break
	  case opStrLike:
	    ctx.firstKey = s;
	    s.inclusion = 1;
	    pattern = s.sval;
	    i = findWildcard(pattern, esc);
	    if (i < 0) { 
		ctx.lastKey = s;
	    } else if (i > 0 && pattern[i-1] != (char)0xffff) {
		int j;
		char[] prefix = new char[i];
		for (j = 0; j < i; j++) { 
		    prefix[j] = pattern[j];
		}
		s.sval = prefix;
		ctx.firstKey = s;
		s = new SearchValue();
		prefix = new char[i];
		for (j = 0; j < i; j++) { 
		    prefix[j] = pattern[j];
		}
		prefix[i-1] += 1;
		s.sval = prefix;
		s.inclusion = 0;
		ctx.lastKey = s;
		ctx.condition = expr; 
	    } else { 
		return null;
	    }
	    break;
	  default:
	    return null;
	}
	return ctx;
    }

    final boolean isIndexApplicable(ConsusStatement stmt, ExprNode condition, boolean toAll)
	throws SQLException 
    { 
	ExprNode expr = condition;
	ExprNode and = null;
	ExprNode or = null;
	ConsusResultSet cursor = stmt.cursor;
	int mergeStart = 0, mergeEnd = -1, selectionLimit = 0;

	if (expr instanceof BinOpNode) { 
	    BinOpNode op = (BinOpNode)expr;	    
	    if (expr.tag == opBoolOr) { 
		or = op.right;
		expr = op.left;
	    } else if (expr.tag == opBoolAnd) { 
		and = op.right;
		expr = op.left;
	    } else if (expr.tag == opOidEq && op.left.tag == opGetOid) {  
                if (op.right.tag == opFirst) { 
		    int oid = table.getRootTable().firstRow;
		    if (oid != 0) { 
			if (cursor.checkType(oid)) { 
			    cursor.add(oid);
			}
		    }
		    return true;
		} else if (op.right.tag == opLast) { 
		    int oid = table.getRootTable().lastRow;
		    if (oid != 0) { 
			if (cursor.checkType(oid)) { 
			    cursor.add(oid);
			}
		    }
		    return true;
		} else if (op.right.tag == opParameter || op.right.tag == opIntConst) { 
		    int oid = op.right.evaluateOid(stmt);
		    if (oid != 0) { 
			if (cursor.checkType(oid)) { 
			    cursor.add(oid);
			}
		    }
		    return true;
		}
	    }
	}
	LoadNode load;
	ExprNode rightOpd;
	if (expr instanceof BinOpNode) { 
	    BinOpNode cmp = (BinOpNode)expr;
	    int loadCop = cmp.left.tag;
	    if (loadCop < opLoadBool || loadCop > opLoadStr) { 
		return false;
	    }
	    load = (LoadNode)cmp.left;
	    rightOpd = cmp.right;
	} else if (expr instanceof TripleOpNode) { 
	    TripleOpNode cmp = (TripleOpNode)expr;
	    int loadCop = cmp.o1.tag;
	    if (loadCop < opLoadBool || loadCop > opLoadStr) { 
		return false;
	    }
	    load = (LoadNode)cmp.o1;
	    rightOpd = cmp.o2; 
	} else { 
	    return false;
	}
	if (load.field.index == 0) { 
	    return false;
	}
	if (load.base != null) { 
	    LoadNode node = load;
	    do { 
		if (node.base.tag != opDeref) { 
		    return false;
		}
		node = (LoadNode)((DerefNode)load.base).base;
		if (node.field.index == 0) { 
		    return false;
		}
	    } while (node.base != null);

	    LoadNode ref = (LoadNode)((DerefNode)load.base).base;
	    SearchValue s = getSearchParameter(stmt, rightOpd);
	    if (s == null || s.subqueryResult) { 
		return false;
	    }
	    SearchContext ctx = createSearchContext(stmt, expr, s);
	    if (ctx == null) { 
		return false;
	    }
	    if (or != null) { 
		mergeStart = cursor.nSelected;
		selectionLimit = cursor.selectionLimit;
		if (!isIndexApplicable(stmt, or, true)) { 
                    return false;
		} else { 
                    mergeEnd = cursor.nSelected;
                }
	    } else if (and != null) { 
		mergeStart = cursor.nSelected;
		selectionLimit = cursor.selectionLimit;
		cursor.selectionLimit = Integer.MAX_VALUE;
		if (!isIndexApplicable(stmt, and, true)) { 
		    if (toAll) { 
			return false;
		    }
		} else { 
		    mergeEnd = cursor.nSelected;
		    and = null;
		}
	    }	    
	    indexField = null;
	    TableDescriptor refTable = ref.field.refTable;
	    Assert.that(refTable != null);
	    ConsusResultSet tmpCursor = new ConsusResultSet(stmt);
	    tmpCursor.open(refTable, false);
	    ExprNode saveBase = load.base;
	    load.base = null;
	    try { 
		stmt.cursor = tmpCursor;
		Btree.find(db, load.field.index, ctx);
		ctx.firstKey = ctx.lastKey = s;
		ctx.condition = null;
		s.inclusion = 1;
		while (true) { 
		    ConsusResultSet tmpCursor2 = null;
		    LoadNode ref2 = null;
		    if (ref.base == null) { 
			ctx.condition = and; 
			stmt.cursor = cursor;
		    } else { 
			ref2 = (LoadNode)((DerefNode)ref.base).base;
			tmpCursor2 = new ConsusResultSet(stmt);
			tmpCursor2.open(ref2.field.refTable, false);
			stmt.cursor = tmpCursor2;
		    }			    
		    int n = tmpCursor.nSelected;
		    for (int i = 0; i < n; i++) { 
			s.oval = tmpCursor.selection[i];
			Btree.find(db, ref.field.index, ctx);
		    }
		    tmpCursor.reset(false);
		    if (ref2 != null) { 
			ref = ref2;
			tmpCursor = tmpCursor2;
		    } else { 
			break;
		    }
		}
	    } finally { 
		load.base = saveBase;
		stmt.cursor = cursor;
	    }
	} else { 		   
	    SearchValue s = getSearchParameter(stmt, rightOpd);
	    if (s == null) { 
		return false;
	    }
	    SearchContext ctx = createSearchContext(stmt, expr, s);
	    if (ctx == null || (toAll && ctx.condition != null)) { 
		return false;
	    }
	    if (or != null) { 
		mergeStart = cursor.nSelected;
		selectionLimit = cursor.selectionLimit;
		if (!isIndexApplicable(stmt, or, true)) { 
		    return false;
		} else { 
		    mergeEnd = cursor.nSelected;
		}
	    } else if (and != null) { 
		mergeStart = cursor.nSelected;
		selectionLimit = cursor.selectionLimit;
		cursor.selectionLimit = Integer.MAX_VALUE;
		if (!isIndexApplicable(stmt, and, true)) { 		
		    if (toAll) { 
			return false;
		    }
		} else { 
		    mergeEnd = cursor.nSelected;
		    and = null;
		}
	    } else { 
		indexField = load.field;
	    }
	    if (ctx.condition == null) { 
		ctx.condition = and;
	    } else { 
		ctx.condition = condition;
	    }
	    if (s.subqueryResult) { 
		Object[] set = rightOpd.evaluateSet(stmt);
		for (int i = 0; i < set.length; i++) { 
		    Object setElem = set[i];
		    switch (expr.tag) { 
		      case opScanSetInt:
			s.ival = ((Long)setElem).longValue();
			break;
		      case opScanSetReal:
			s.dval = ((Double)setElem).doubleValue();
			break;
		      case opScanSetStr:
			s.sval = ((String)setElem).toCharArray();
			break;
		      case opScanSetRef:
			s.oval = (int)((ConsusRef)setElem).getOid();
			break;
		    }
		    Btree.find(db, load.field.index, ctx);
		}
	    } else { 
		Btree.find(db, load.field.index, ctx);
	    }
	} 
	if (mergeEnd >= 0) { 
	    int n = cursor.nSelected;
	    int[] selection = cursor.selection;
	    Arrays.sort(selection, mergeStart, mergeEnd);
	    Arrays.sort(selection, mergeEnd, n);
            if (or != null) { 
                for (int i = mergeStart, j = mergeEnd; i < mergeEnd; i++) { 
                    if (selection[i] == selection[j]) { 
                        n -= 1;
                        System.arraycopy(selection, j+1, selection, j, n-j);
                    } else { 
                        j += 1;
                    }
                }
                cursor.nSelected = n;
            } else { 
                for (int i = mergeStart, j = mergeEnd; i < mergeEnd && j < n;) { 
                    if (selection[i] > selection[j]) { 
                        j += 1;
                    } else if (selection[i] < selection[j]) { 
                        i += 1;
                    } else { 
                        selection[mergeStart++] = selection[i];
                        if (mergeStart >= selectionLimit) { 
                            break;
                        }
                        i += 1;
                        j += 1;
                    }
                }
                cursor.nSelected = mergeStart;
            }
	    cursor.selectionLimit = selectionLimit;
	}
	return true;
    }

    final void revert(int[] ref, int n) { 
	for (int i = 0, j = n-1; i < j; i++, j--) { 
	    int t = ref[i];
	    ref[i] = ref[j];
	    ref[j] = t;
	}
    }

    final static int strSortPrefix = 8;

    final void quickSort(int[] ref, int n) throws SQLException
    {
	int i, j;
	byte[] obj;
	if (n <= 1) { 
	    return;
	}
	int offs = order.field.offs;
	int type = order.field.type;
	switch (type) {
	  case FieldDescriptor.tpBool:
	  case FieldDescriptor.tpInt1:
	  {
	      byte arr[] = new byte[n];
	      for (i = 0; i < n; i++) { 
		  obj = db.get(ref[i]);
		  arr[i] = obj[offs];
	      }
	      QSort.sort(ref, arr, 0, n);
	      if (!order.ascent) { 
		  revert(ref, n);
	      }
	      if (order.next != null) { 
		  OrderNode save = order;
		  order = order.next;
		  for (i = 0; i < n; i = j) { 
		      for (j = i+1; j < n && arr[i] == arr[j]; j++);
		      qSort(ref, i, j-i);
		  }
		  order = save;
	      }
	      break;
	  }
	  case FieldDescriptor.tpInt2:
	  {
	      short arr[] = new short[n];
	      for (i = 0; i < n; i++) { 
		  obj = db.get(ref[i]);
		  arr[i] = Bytes.unpack2(obj, offs);
	      }
	      QSort.sort(ref, arr, 0, n);
	      if (!order.ascent) { 
		  revert(ref, n);
	      }
	      if (order.next != null) { 
		  OrderNode save = order;
		  order = order.next;
		  for (i = 0; i < n; i = j) { 
		      for (j = i+1; j < n && arr[i] == arr[j]; j++);
		      qSort(ref, i, j-i);
		  }
		  order = save;
	      }
	      break;
	  }
	  case FieldDescriptor.tpInt4:
	  case FieldDescriptor.tpReference:
	  {
	      int arr[] = new int[n];
	      for (i = 0; i < n; i++) { 
		  obj = db.get(ref[i]);
		  arr[i] = Bytes.unpack4(obj, offs);
	      }
	      QSort.sort(ref, arr, 0, n);
	      if (!order.ascent) { 
		  revert(ref, n);
	      }
	      if (order.next != null) { 
		  OrderNode save = order;
		  order = order.next;
		  for (i = 0; i < n; i = j) { 
		      for (j = i+1; j < n && arr[i] == arr[j]; j++);
		      qSort(ref, i, j-i);
		  }
		  order = save;
	      }
	      break;
	  }	    
	  case FieldDescriptor.tpInt8:
	  case FieldDescriptor.tpDate:
	  case FieldDescriptor.tpTime:
	  case FieldDescriptor.tpTimestamp:
	  {
	      long arr[] = new long[n];
	      for (i = 0; i < n; i++) { 
		  obj = db.get(ref[i]);
		  arr[i] = Bytes.unpack8(obj, offs);
	      }
	      QSort.sort(ref, arr, 0, n);
	      if (!order.ascent) { 
		  revert(ref, n);
	      }
	      if (order.next != null) { 
		  OrderNode save = order;
		  order = order.next;
		  for (i = 0; i < n; i = j) { 
		      for (j = i+1; j < n && arr[i] == arr[j]; j++);
		      qSort(ref, i, j-i);
		  }
		  order = save;
	      }
	      break;
	  }
	  case FieldDescriptor.tpReal4:
	  {
	      float arr[] = new float[n];
	      for (i = 0; i < n; i++) { 
		  obj = db.get(ref[i]);
		  arr[i] = Float.intBitsToFloat(Bytes.unpack4(obj, offs));
	      }
	      QSort.sort(ref, arr, 0, n);
	      if (!order.ascent) { 
		  revert(ref, n);
	      }
	      if (order.next != null) { 
		  OrderNode save = order;
		  order = order.next;
		  for (i = 0; i < n; i = j) { 
		      for (j = i+1; j < n && arr[i] == arr[j]; j++);
		      qSort(ref, i, j-i);
		  }
		  order = save;
	      }
	      break;
	  }
	  case FieldDescriptor.tpReal8:
	  {
	      double arr[] = new double[n];
	      for (i = 0; i < n; i++) { 
		  obj = db.get(ref[i]);
		  arr[i] = Double.longBitsToDouble(Bytes.unpack8(obj, offs));
	      }
	      QSort.sort(ref, arr, 0, n);
	      if (!order.ascent) { 
		  revert(ref, n);
	      }
	      if (order.next != null) { 
		  OrderNode save = order;
		  order = order.next;
		  for (i = 0; i < n; i = j) { 
		      for (j = i+1; j < n && arr[i] == arr[j]; j++);
		      qSort(ref, i, j-i);
		  }
		  order = save;
	      }
	      break;
	  }
	  case FieldDescriptor.tpString:
	  {
	      char arr[] = new char[n*strSortPrefix];
	      for (i = 0; i < n; i++) { 
		  obj = db.get(ref[i]);
		  int len = Varying.getSize(obj, offs);
		  int ofs = Varying.getOffs(obj, offs);
		  if (len > strSortPrefix) { 
		      len = strSortPrefix;
		  }
		  long val = 0;
		  for (j = 0; j < len; j++) { 
		      arr[i*strSortPrefix+j] = (char)Bytes.unpack2(obj, ofs);
		      ofs += 2;
		  }
	      }
	      QSort.sort(ref, arr, strSortPrefix, 0, n);	      	
	      if (!order.ascent) { 
		  revert(ref, n);
	      }
	      for (i = 0; i < n; i = j) { 
	        same_prefix: 
		  for (j = i+1; j < n; j++) { 
		      int len = strSortPrefix;
		      int k = j*8; 
		      while (--len >= 0) { 
			  k -= 1;
			  if (arr[k] != arr[k+strSortPrefix]) { 
			      break same_prefix;
			  }
		      }
		  }
		  qSort(ref, i, j-i);
	      }
	  }
	}
    }


    /**
     * Swaps x[a] with x[b].
     */
    final static void swap(int x[], int a, int b) {
	int t = x[a];
	x[a] = x[b];
	x[b] = t;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(int x[], int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++)
	    swap(x, a, b);
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final int med3(int x[], int a, int b, int c) throws SQLException {
	byte[] ra = db.get(x[a]);
	byte[] rb = db.get(x[b]);
	byte[] rc = db.get(x[c]);
	int diff_ab = order.compare(ra, rb);
	int diff_bc = order.compare(rb, rc);
	int diff_ac = order.compare(ra, rc);
	return (diff_ab < 0 ?
		(diff_bc < 0 ? b : diff_ac < 0 ? c : a) :
		(diff_bc > 0 ? b : diff_ac > 0 ? c : a));
    }

    final void qSort(int[] x, int off, int len) throws SQLException
    {
	// Insertion sort on smallest arrays
	if (len < 7) {
	    if (len > 1) { 
		byte[] rec2 = db.get(x[off]);
		for (int i=off+1; i<len+off; i++) { 
		    byte[] rec1 = rec2;
		    rec2 = db.get(x[i]);
		    int j = i;
		    if (order.compare(rec1, rec2) > 0) { 
			byte[] tmp = rec2;
			rec2 = rec1;
			do { 
			    swap(x, j, j-1);
			    if (--j == off) { 
				break;
			    }
			    rec1 = db.get(x[j-1]);
			} while (order.compare(rec1, tmp) > 0);
		    }
		}
	    }
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, l,     l+s, l+2*s);
		m = med3(x, m-s,   m,   m+s);
		n = med3(x, n-2*s, n-s, n);
	    }
	    m = med3(x, l, m, n); // Mid-size, med of 3
	}
	byte[] v = db.get(x[m]);

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	int diff;
	while(true) {
	    while (b <= c && (diff = order.compare(db.get(x[b]), v)) <= 0) {
		if (diff == 0)
		    swap(x, a++, b);
		b++;
	    }
	    while (c >= b && (diff = order.compare(db.get(x[c]), v)) >= 0) {
		if (diff == 0)
		    swap(x, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(x, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(x, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(x, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    qSort(x, off, s);
	if ((s = d-c) > 1)
	    qSort(x, n-s, s);
    }

	
   /**
     * Swaps x[a] with x[b].
     */
    final static void swap(Object[][] x, int a, int b) {
	Object[] t = x[a];
	x[a] = x[b];
	x[b] = t;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(Object[][] x, int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++)
	    swap(x, a, b);
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final int med3(Object[][] x, int a, int b, int c) throws SQLException {
	int diff_ab = order.compareTuples(x[a], x[b]);
	int diff_bc = order.compareTuples(x[b], x[c]);
	int diff_ac = order.compareTuples(x[a], x[c]);
	return (diff_ab < 0 ?
		(diff_bc < 0 ? b : diff_ac < 0 ? c : a) :
		(diff_bc > 0 ? b : diff_ac > 0 ? c : a));
    }

    final void qSort(Object[][] x, int off, int len) throws SQLException
    {
	// Insertion sort on smallest arrays
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && order.compareTuples(x[j-1], x[j]) > 0; j--)
		    swap(x, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, l,     l+s, l+2*s);
		m = med3(x, m-s,   m,   m+s);
		n = med3(x, n-2*s, n-s, n);
	    }
	    m = med3(x, l, m, n); // Mid-size, med of 3
	}
	Object[] v = x[m];

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	int diff;
	while(true) {
	    while (b <= c && (diff = order.compareTuples(x[b], v)) <= 0) {
		if (diff == 0)
		    swap(x, a++, b);
		b++;
	    }
	    while (c >= b && (diff = order.compareTuples(x[c], v)) >= 0) {
		if (diff == 0)
		    swap(x, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(x, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(x, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(x, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    qSort(x, off, s);
	if ((s = d-c) > 1)
	    qSort(x, n-s, s);
    }
	
    SelectNode(SelectNode scope) { 
	super(tpVoid, opSelect);
	outer = scope;
    }

    SelectNode(int tag, TableDescriptor desc, ExprNode criteria) { 
	super(tpVoid, tag);
	table = desc;
	condition = criteria;
	forUpdate = true;
    }   
}
