/* Consus - Pure Java Object Relational Database Management System
 * Implementation of Consus server for remote connections
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.io.*;
import java.net.*;

/** 
 * Server of remote client's connections
 */
public class ServerManager extends Thread { 
    ServerSocket        globalSocket;
    LocalSocket         localSocket;
    ConsusConnection    db;
    ConsusAuthenticator authenticator;
    boolean             running;
    int                 port;
    
    /** 
     * Linger parameter for the socket
     */
    public static int lingerTime = 10; 

    /**
     * Construct server manager
     * @param con database connection
     * @param port to accept connections
     * @param local <code>true</code> if manager should accept only local connections
     * @param auth authenticator object (<code>null</code> disables authentication)
     */
    public ServerManager(org.garret.consus.ConsusConnection con, int port, boolean local, 
			 ConsusAuthenticator auth) 
	throws IOException 
    { 
	db  = (ConsusConnection)con;
	this.port = port;
	if (local) { 
	    localSocket = LocalSocket.create(port);
	} else { 
	    globalSocket = new ServerSocket(port);
	}
	authenticator = auth; 
    }

    /**
     * Construct server manager
     * @param con database connection
     * @param port to accept connections
     * @param local <code>true</code> if manager should accept only local connections
     */
    public ServerManager(ConsusConnection con, int port, boolean local) throws IOException { 
	this(con, port, local, null);
    }

    /**
     * Construct server manager
     * @param con database connection
     * @param port to accept connections
     */
    public ServerManager(ConsusConnection con, int port) throws IOException { 
	this(con, port, false, null);
    }

    /**
     * Main server thread method. It accepts client connections to the port and create
     * new thread for each client.
     */
    public void run() { 
	running = true;
	try { 
	    while (running) { 
		Socket s = globalSocket != null ? globalSocket.accept() : localSocket.accept();	
		if (!running) { 
		    break;
		}
		if (db.dbTraceEnabled) { 
		    System.out.println("Connect client " + s.getInetAddress());
		}
		try { 
		    s.setTcpNoDelay(true);
		} catch (NoSuchMethodError er) {}
		try {
		    s.setSoLinger(true, lingerTime);
		} catch (NoSuchMethodError er) {}
		new ServerThread(db, s, authenticator).start();
	    }
	    if (globalSocket != null) { 
		globalSocket.close();
	    } else { 
		localSocket.close();
	    }
	} catch(IOException x) { 
	    x.printStackTrace();	    
	}
    }

    /**
     * Terminate the server
     */
    public void terminate() { 
	running = false;
	Socket s = null;
	try { 
	    // wakeup accept
	    if (localSocket != null) { 
		localSocket.cancelAccept();
	    } else { 
		s = new Socket("localhost", port);
	    }
	} catch(IOException x) {}

	try { 
	    join();
	} catch(InterruptedException x) {}

	try { 
	    if (s != null) { 
		s.close();
	    }
	} catch(IOException x) {}
    }
}






