/* Consus - Pure Java Object Relational Database Management System
 * Implementation of client agent at the server site
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.io.*;
import java.net.*;
import java.sql.*;
import java.util.Properties;
import java.sql.SQLException;

class ServerThread extends Thread { 
    ConsusConnection    db;
    Socket              socket;
    ObjectInputStream   in;
    ObjectOutputStream  out;
    Message             reply;
    ConsusAuthenticator authenticator;

    ServerThread(ConsusConnection conxn, Socket socket, ConsusAuthenticator auth) 
	throws IOException 
    { 
	db = conxn;
	this.socket = socket;
	authenticator = auth;
	in = new ObjectInputStream(new BufferedInputStream(socket.getInputStream()));
	out = new ObjectOutputStream(new BufferedOutputStream(socket.getOutputStream()));
	reply = new Message();
	setDaemon(true);
    }

    void setParameters(ConsusPreparedStatement pstmt, Object arg) throws SQLException { 
	if (arg != null) { 
	    Object[] params = (Object[])arg;
	    for (int i = 0; i < params.length; i++) { 
		pstmt.setObject(i+1, params[i]);
	    }
	}
    }

    ColumnInfo[] getColumns(ConsusStatement stmt) { 
	ColumnNode columns[] = stmt.cursor.tupleColumns;
	int n = columns.length;
	ColumnInfo cinfo[] = new ColumnInfo[n];
	while (--n >= 0) { 
	    cinfo[n] = new ColumnInfo(columns[n].name, columns[n].type);
	}
	return cinfo;
    }

    public void run() { 
	db.getContext().selectTuplesMode = true;
	DatabaseMetaData meta = null;
	ConsusResultSet cursor;
	try { 
	    boolean online = true;
	    ConsusPreparedStatement pstmt;
	    ConsusStatement stmt = null;
	    ConsusBlob blob;
	    ConsusClob clob;	    
	    BlobInfo binfo;

	    while (online) { 
		Message req = (Message)in.readObject();
		try { 
		    if (req.opd1 instanceof String && db.dbTraceEnabled) { 
			System.out.println("Execute statement: " + req.opd1);
		    }
		    switch (req.tag) { 
		      case Message.REQ_LOGIN:
			reply.tag = Message.REP_LOGIN;
			if (authenticator != null) { 
			    Properties userInfo = (Properties)req.opd1;
			    userInfo.setProperty("host", socket.getInetAddress().toString());
			    reply.opd1 = new Boolean(authenticator.authenticate(db, userInfo));
			    db.commit();
			}
			break;
		      case Message.REQ_CLOSE:
			db.commitTransaction();
			online = false;
			reply.tag = Message.REP_CLOSE;
			break;
		      case Message.REQ_COMMIT:
			db.commitTransaction();
			reply.tag = Message.REP_OK;
			break;
		      case Message.REQ_SUBCOMMIT:
			db.commitSubtransaction();
			reply.tag = Message.REP_OK;
			break;
		      case Message.REQ_ROLLBACK:
			db.rollback();
			reply.tag = Message.REP_OK;
			break;
		      case Message.REQ_PREPARE:
			db.prepareTransaction(req.opd3, (byte[])req.opd1, (byte[])req.opd2);
			reply.tag = Message.REP_OK;
			break;
		      case Message.REQ_COMPLETE:
			db.completeTransaction();
			reply.tag = Message.REP_OK;
			break;
		      case Message.REQ_RECOVER:
			ConsusTransId tid = db.getTransId();
			reply.tag = Message.REP_RECOVERED;
			reply.opd1 = tid.getGlobalTransactionId();
			reply.opd2 = tid.getBranchQualifier();
			reply.opd3 = tid.getFormatId();
			break;			
		      case Message.REQ_EXEC:
			if (stmt == null) { 
			    stmt = (ConsusStatement)db.createStatement();
			}
			stmt.setMaxRows(req.opd3);
			if (stmt.execute((String)req.opd1)) { 
			    reply.tag = Message.REP_QUERY;
			    reply.opd1 = stmt.cursor.tuples;
			    reply.opd2 = getColumns(stmt);
			    reply.opd3 = stmt.cursor.nSelected;
			    reply.opd4 = stmt.oid;
			    stmt.cursor.reset(true);
			} else { 
			    reply.tag = Message.REP_UPDATE;
			    reply.opd3 = stmt.getUpdateCount();
			    reply.opd4 = stmt.oid;
			}
			break;
		      case Message.REQ_EXEC_QUERY:
			if (stmt == null) { 
			    stmt = (ConsusStatement)db.createStatement();
			}
			stmt.setMaxRows(req.opd3);
			stmt.executeQuery((String)req.opd1); 
			reply.tag = Message.REP_QUERY;
			reply.opd1 = stmt.cursor.tuples;
			reply.opd2 = getColumns(stmt);
			reply.opd3 = stmt.cursor.nSelected;
			reply.opd4 = stmt.oid;
			stmt.cursor.reset(true);
			break;
		      case Message.REQ_EXEC_UPDATE:
			if (stmt == null) { 
			    stmt = (ConsusStatement)db.createStatement();
			}
			reply.tag = Message.REP_UPDATE;
			reply.opd3 = stmt.executeUpdate((String)req.opd1);
			reply.opd4 = stmt.oid;
			break;
		      case Message.REQ_EXEC_PARAM:
			pstmt = (ConsusPreparedStatement)db.prepareStatement((String)req.opd1);
			setParameters(pstmt, req.opd2);
			pstmt.setMaxRows(req.opd3);
			if (pstmt.execute()) { 
			    reply.tag = Message.REP_QUERY;
			    reply.opd1 = pstmt.cursor.tuples;
			    reply.opd2 = getColumns(pstmt);
			    reply.opd3 = pstmt.cursor.nSelected;
			    reply.opd4 = pstmt.oid;
			    pstmt.cursor.reset(true);
			} else { 
			    reply.tag = Message.REP_UPDATE;
			    reply.opd3 = pstmt.getUpdateCount();
			    reply.opd4 = pstmt.oid;
			}
			break;
		      case Message.REQ_EXEC_QUERY_PARAM:
			pstmt = (ConsusPreparedStatement)db.prepareStatement((String)req.opd1);
			setParameters(pstmt, req.opd2);
			pstmt.setMaxRows(req.opd3);
			pstmt.executeQuery(); 
			reply.tag = Message.REP_QUERY;
			reply.opd1 = pstmt.cursor.tuples;
			reply.opd2 = getColumns(pstmt);
			reply.opd3 = pstmt.cursor.nSelected;
			reply.opd4 = pstmt.oid;
			pstmt.cursor.reset(true);
			break;
		      case Message.REQ_EXEC_UPDATE_PARAM:
			pstmt = (ConsusPreparedStatement)db.prepareStatement((String)req.opd1);
			setParameters(pstmt, req.opd2);
			reply.tag = Message.REP_UPDATE;
			reply.opd3 = pstmt.executeUpdate();
			reply.opd4 = pstmt.oid;
			break;
		      case Message.REQ_BLOB_POSITION:
			binfo = (BlobInfo)req.opd1;
			blob = new ConsusBlob(db, binfo.pos, binfo.size);
			reply.tag = Message.REP_BLOB_POSITION;
			reply.opd1 = new Long(blob.position((byte[])req.opd2, binfo.offs));
			break;
		      case Message.REQ_BLOB_SEGMENT:
			binfo = (BlobInfo)req.opd1;
			blob = new ConsusBlob(db, binfo.pos, binfo.size);
			reply.tag = Message.REP_BLOB_SEGMENT;
			reply.opd1 = blob.getBytes(binfo.offs+1, req.opd3);
			reply.opd3 = ((byte[])reply.opd1).length;
			break;
		      case Message.REQ_CLOB_POSITION:
			binfo = (BlobInfo)req.opd1;
			clob = new ConsusClob(db, binfo.pos, binfo.size);
			reply.tag = Message.REP_CLOB_POSITION;
			reply.opd1 = new Long(clob.position((String)req.opd2, binfo.offs));
			break;
		      case Message.REQ_CLOB_SEGMENT:
			binfo = (BlobInfo)req.opd1;
			clob = new ConsusClob(db, binfo.pos, binfo.size);
			reply.tag = Message.REP_CLOB_SEGMENT;
			reply.opd1 = clob.getSubString(binfo.offs+1, req.opd3);
			reply.opd3 = ((String)reply.opd1).length();
			break;
		      case Message.REQ_TABLES:
			if (meta == null) {
			    meta = db.getMetaData();
			}
			cursor = (ConsusResultSet)meta.getTables(null, "", (String)req.opd1,
								 (String[])req.opd2);
			reply.tag = Message.REP_TABLES;
			reply.opd1 = new Object[cursor.nSelected][];
			if (cursor.tuples != null) { 
			    System.arraycopy(cursor.tuples, 0, reply.opd1, 0, cursor.nSelected);
			}
			cursor.close();
			break;
		      case Message.REQ_COLUMNS:
			if (meta == null) {
			    meta = db.getMetaData();
			}
			cursor = (ConsusResultSet)meta.getColumns(null, "", (String)req.opd1,
								 (String)req.opd2);
			reply.tag = Message.REP_COLUMNS;
			reply.opd1 = new Object[cursor.nSelected][];
			System.arraycopy(cursor.tuples, 0, reply.opd1, 0, cursor.nSelected);
			cursor.close();
			break;
		      case Message.REQ_PRIMARY_KEYS:
			if (meta == null) {
			    meta = db.getMetaData();
			}
			cursor = (ConsusResultSet)meta.getPrimaryKeys(null, "", (String)req.opd1);
			reply.tag = Message.REP_PRIMARY_KEYS;
			reply.opd1 = new Object[cursor.nSelected][];
			System.arraycopy(cursor.tuples, 0, reply.opd1, 0, cursor.nSelected);
			cursor.close();
			break;
		      case Message.REQ_IMPORTED_KEYS:
			if (meta == null) {
			    meta = db.getMetaData();
			}
			cursor = (ConsusResultSet)meta.getImportedKeys(null, "", (String)req.opd1);
			reply.tag = Message.REP_IMPORTED_KEYS;
			reply.opd1 = new Object[cursor.nSelected][];
			System.arraycopy(cursor.tuples, 0, reply.opd1, 0, cursor.nSelected);
			cursor.close();
			break;
		      case Message.REQ_EXPORTED_KEYS:
			if (meta == null) {
			    meta = db.getMetaData();
			}
			cursor = (ConsusResultSet)meta.getExportedKeys(null, "", (String)req.opd1);
			reply.tag = Message.REP_EXPORTED_KEYS;
			reply.opd1 = new Object[cursor.nSelected][];
			System.arraycopy(cursor.tuples, 0, reply.opd1, 0, cursor.nSelected);
			cursor.close();
			break;
		      case Message.REQ_CROSSREFS:
			if (meta == null) {
			    meta = db.getMetaData();
			}
			cursor = (ConsusResultSet)meta.getCrossReference(null, "", (String)req.opd1,
									 null, "", (String)req.opd2);
			reply.tag = Message.REP_CROSSREFS;
			reply.opd1 = new Object[cursor.nSelected][];
			System.arraycopy(cursor.tuples, 0, reply.opd1, 0, cursor.nSelected);
			cursor.close();
			break;
		      case Message.REQ_INDICES:
			if (meta == null) {
			    meta = db.getMetaData();
			}
			cursor = (ConsusResultSet)meta.getIndexInfo(null, "", (String)req.opd1,
								    req.opd3 != 0, false);
			reply.tag = Message.REP_INDICES;
			reply.opd1 = new Object[cursor.nSelected][];
			System.arraycopy(cursor.tuples, 0, reply.opd1, 0, cursor.nSelected);
			cursor.close();
			break;
		      default:
			reply.tag = Message.REP_ERROR;
			reply.opd1 = new ConsusRuntimeError("Invalid message tag: " + req.tag);
		    }
		} catch(Exception x) { 
		    x.printStackTrace();
		    reply.tag = Message.REP_ERROR;
		    reply.opd1 = x;
		    reply.opd2 = null;
		}
		out.writeObject(reply);
		out.flush();
		out.reset();
		reply.opd1 = reply.opd2 = null;
	    }
	} catch(Exception x) { 
	    x.printStackTrace();
	    try { 
		db.rollback();
	    } catch(SQLException ex) {}
	}
	if (db.dbTraceEnabled) { 
	    System.out.println("Disconnect client " + socket.getInetAddress());
	}
	try { 
	    in.close();
	    out.close();
	    socket.close();
	} catch(IOException x) {}
    }
}




