/* Consus - Pure Java Object Relational Database Management System
 * Btree object container
 * Created by Konstantin Knizhnik, Jun 2000.
 */

package org.garret.internal.consus;
import org.garret.consus.*;
import java.sql.SQLException;

public class StringIndex extends org.garret.consus.StringIndex {
    int    treeOID;
    
    StringIndex() {} // Constructor for Consus
    ConsusConnection db;

    StringIndex(StringIndex a) {
	try { 
	    treeOID = Btree.cloneTree(a.db, a.treeOID);
	} catch(SQLException x) { 
	    throw new POSError(x);
	}
    }

   /**
     * Create object index with string key.
     * @param conxn database connection
     */
    public StringIndex(ConsusConnection conxn) { 
	try { 
	    db = conxn;
	    conxn.beginTransaction(true);
	    treeOID = Btree.allocate(conxn, FieldDescriptor.tpString);
	    store(conxn);
	} catch(SQLException x) { 
	    throw new POSError(x);
	}
    }

    /**
     * Remove all object from the index
     */
    public synchronized void clear() {
	try { 
	    db.beginTransaction(true);
	    Btree.purge(db, treeOID);
	} catch(SQLException x) { 
	    throw new POSError(x);
	}	    
    }

    /**
     * Deallocate index
     */
    public void remove() { 
	try { 
	    db.beginTransaction(true);
	    Btree.drop(db, treeOID);
	    treeOID = 0;
	    super.remove();
	} catch(SQLException x) { 
	    throw new POSError(x);
	}
    }

    /**
     * Insert new object in the index
     * @param obj object to be placed in the index if unique constraint is not violated
     */
    public boolean insertUnique(String key, Persistent obj) { 
	try { 
	    db.beginTransaction(true);
	    return Btree.insert(db, treeOID, (int)obj.assignOID(db), key, true);
	} catch(SQLException x) { 
	    throw new POSError(x);
	}	
    }

    /**
     * Insert new object in the index
     * @param obj object to be placed in the index
     */
    public void insert(String key, Persistent obj) { 
	try { 
	    db.beginTransaction(true);
	    Btree.insert(db, treeOID, (int)obj.assignOID(db), key, false);
	} catch(SQLException x) { 
	    throw new POSError(x);
	}
    }
	
    /**
     * Remove object from the index
     * @param obj object to be removed from the index
     */
    public boolean remove(String key, Persistent obj) { 
	try { 
	    db.beginTransaction(true);
	    return Btree.remove(db, treeOID, (int)obj.oid, key);
	} catch(SQLException x) { 
	    throw new POSError(x);
	}
    }

    /**
     * Select object with value of the key greater-or-equal than <code>firstKey</code>
     * and less-or-equal than <code>lastKey</code>
     * @param firstKey  low bound of key rabge
     * @param lastKey   high bound of key rabge
     * @param forUpdate if objects are sel;ected for update
     * @return ResultSet with selected objects
     */
    public org.garret.consus.ConsusResultSet select(String firstKey, String lastKey, boolean forUpdate) 
    {
	try { 
	    db.beginTransaction(forUpdate); 
	    ConsusStatement stmt = new ConsusStatement(db);
	    stmt.cursor.reset(false);
	    stmt.cursor.open(null, forUpdate);
	    SearchContext sc = new SearchContext();
	    if (firstKey != null) {
		sc.firstKey = new SearchValue();
		sc.firstKey.sval = firstKey.toCharArray();
		sc.firstKey.inclusion = 1;
	    } 
	    if (lastKey != null) {
		sc.lastKey = new SearchValue();
		sc.lastKey.sval = lastKey.toCharArray();
		sc.lastKey.inclusion = 1;
	    } 
	    sc.stmt = stmt;
	    sc.condition = null;
	    Btree.find(db, treeOID, sc);
	    return stmt.cursor;
	} catch (SQLException x) { 
	    throw new POSError(x);
	} 
    }

    /**
     * Select object with value of the key greater-or-equal than <code>firstKey</code>
     * and less-or-equal than <code>lastKey</code>
     * @param firstKey  low bound of key rabge
     * @param lastKey   high bound of key rabge
     * @return ResultSet with selected objects
     */
    public org.garret.consus.ConsusResultSet select(String firstKey, String lastKey) {
	return select(firstKey, lastKey, false);
    }

    /**
     * Select for update object with value of the key greater-or-equal than <code>firstKey</code>
     * and less-or-equal than <code>lastKey</code>
     * @param firstKey  low bound of key range
     * @param lastKey   high bound of key range
     * @return ResultSet with selected objects
     */
    public org.garret.consus.ConsusResultSet selectForUpdate(String firstKey, String lastKey) {
	return select(firstKey, lastKey, true);
    }

    /**
     * Select objects with specified value of the key 
     * @param key searched value
     * @return ResultSet with selected objects
     */
    public org.garret.consus.ConsusResultSet select(String key) { 
	return select(key, key, false);
    }

    /**
     * Select for update objects with specified value of the key 
     * @param key searched value
     * @return ResultSet with selected objects
     */
    public org.garret.consus.ConsusResultSet selectForUpdate(String key) { 
	return select(key, key, true);
    }

    /**
     * Select object with specified value of the key 
     * @param key searched value
     * @return object with specified value of the key, <code>null</code> if there are no such object
     */
    public Object selectUnique(String key) { 
	try { 
	    ConsusResultSet cursor = (ConsusResultSet)select(key, key, false);
	    Object obj = null;
	    if (cursor.nSelected > 1) { 
		cursor.close();
		throw new ConsusNotSingleException("Key is not unique");
	    } else if (cursor.next()) { 
		obj = cursor.getSelfObject();
	    }
	    cursor.close();
	    return obj;
	} catch (SQLException x) { 
	    throw new POSError(x);
	} 
    }

    /**
     * Selects all indexed objects
     * @param forUpdate if objects are selected for update
     * @return ResultSet with selected objects
     */
    public org.garret.consus.ConsusResultSet select(boolean forUpdate) { 
	try { 
	    db.beginTransaction(forUpdate); 
	    ConsusStatement stmt = new ConsusStatement(db);
	    stmt.cursor.reset(false);
	    stmt.cursor.open(null, forUpdate);
	    Btree.traverseForward(db, treeOID, stmt, null);
	    return stmt.cursor;
	} catch (SQLException x) { 
	    throw new POSError(x);
	} 
    }

    /**
     * Selects all indexed objects
     * @return ResultSet with selected objects
     */
    public org.garret.consus.ConsusResultSet select() { 
	return select(false);
    }
    
    /**
     * Clone index
     */
    public Object clone() throws CloneNotSupportedException { 
	return new StringIndex(this);
    }
}








