/* Consus - Pure Java Object Relational Database Management System
 * Consus Table(class) descriptor. Runtime information about Consus tables,
 *   for storing/fetching/updating records, manipulations with indices
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;
import java.lang.reflect.*;
import java.util.Hashtable;
import java.util.Comparator;
import java.util.Arrays;

class TableDescriptor { 
    String           name;
    // Associated Java class (null if none)
    Class            cls;
    // List of atomic fields including components of structures and base classes
    FieldDescriptor  columns[]; 
    // List of all components of the class and base classes (structure and array fields are not included)
    FieldDescriptor  components[];
    // List of all fields descriptors (including base classes, structure and array fields)
    // This list is linked in reverse order
    FieldDescriptor  fields;
    
    
    int              fixedSize;      // Size of all fixed size compoennts of the class,
                                     // including sizeof(Record)
    TableDescriptor  superClass;     // Base class
    TableDescriptor  collisionChain; // collision chain for hashtable by table OID
    TableDescriptor  next;           // list of registered tables

    TableDescriptor  newTable;       // new version of table descriptor
    boolean          formatChanged;  // whether table records should be reformated 

    int              attr;    
    static final int TA_JAVA_CLASS = 1;    // table was created from Java class

    Hashtable        columnMapping = new Hashtable(); // hahstable for location column by name

    int              tableId;        // table OID
    int              nSubclasses;    // number of derived tables
    int              firstRow;       // first row OID
    int              lastRow;        // last row OID (records are inserted adter the last record)

    int              nRows;          // number of rows in the table
    int              nFields;        // number of all fields (length of fields list)
    int              nSelfFields;    // number of self fields (excluding superclasses)
    int              nColumns;       // length of columns array

    Constructor      constructor;    // constructor for creating object instances
    static final Object[] constructorArgs = {};
    
    byte[]           desc;           // byte array with table desriptor packed image

    boolean          updated;        // changes of table descriptor fields are not immediatly 
                                     // saved in the file, but instead of this 'updated' flag is set.
                                     // While transaction commit, updated table descriptors will be 
                                     // stored to the file

    boolean          created;        // table that was created within transaction which is not yet committed
    boolean          altered;        // table that was altered within transaction which is not yet committed

    FieldDescriptor  timestampField; // Field used to support replication, 
                                     // it is automatically updated by Consus
    long             timestamp;

    int              numberOfBlobs;  // numebr of BLOB or CLOB fields in the record

    FieldDescriptor  foreignKeys;    // list of foreign keys for this table which 
                                     // references OID of the record

    ConsusConnection db;             // associated database connection 


    void update(TableDescriptor tbl) { 
	cls = tbl.cls;
	columns = tbl.columns;
	components = tbl.components;
	fields = tbl.fields;
	fixedSize = tbl.fixedSize;
	superClass = tbl.superClass;
	nSubclasses = tbl.nSubclasses;
	columnMapping = tbl.columnMapping;
	nFields = tbl.nFields;
	nSelfFields = tbl.nSelfFields;
	nColumns = tbl.nColumns;
	constructor = tbl.constructor;
	numberOfBlobs = tbl.numberOfBlobs;
	altered = true;
    }

    void undoUpdate(TableDescriptor tbl) { 
	columns = tbl.columns;
	components = tbl.components;
	fields = tbl.fields;
	fixedSize = tbl.fixedSize;
	superClass = tbl.superClass;
	nSubclasses = tbl.nSubclasses;
	columnMapping = tbl.columnMapping;
	nFields = tbl.nFields;
	nSelfFields = tbl.nSelfFields;
	nColumns = tbl.nColumns;
	numberOfBlobs = tbl.numberOfBlobs;
	firstRow = tbl.firstRow;
	lastRow = tbl.lastRow;  
	nRows = tbl.nRows;
	timestamp = tbl.timestamp;
	altered = updated = false;
	desc = tbl.desc;
	for (FieldDescriptor fd = fields; fd != null; fd = fd.next) { 
	    fd.table = this;
	}
   }

    //
    // Build columns mapping
    //
    void mapColumns() { 
	columns = new FieldDescriptor[nColumns];
	int i = nFields, n = nColumns;
	for (FieldDescriptor fd = fields; fd != null; fd = fd.next) { 
	    fd.id = --i;
	    if (fd.type != FieldDescriptor.tpStructure 
		&& (fd.attr & FieldDescriptor.ComponentOfArray) == 0) 
	    {
		columnMapping.put(fd.longName, fd);	
		fd.columnIndex = --n;
		columns[n] = fd;
	    }
	}
	Assert.that(n == 0 && i == 0);
    }

    //
    // Constructor for table descriptor from Java class
    //
    TableDescriptor(ConsusConnection conxn, Class clazz) throws SQLException { 
	int i, j, n = 0;
	cls = clazz;
	name = clazz.getName();
	db = conxn;
	created = true;
	attr = TA_JAVA_CLASS;
	Class base = clazz.getSuperclass();
	if (FieldDescriptor.persistentClass.isAssignableFrom(clazz)) { 
	    conxn.createObjectCache();
	}
	TableDescriptor superClass = null;
	if (base != null && !base.getName().equals("java.lang.Object")
	    && !base.getName().equals("org.garret.consus.Persistent"))
	{ 
	    this.superClass = superClass = (TableDescriptor)conxn.getTableByClass(base);
	    if (superClass.newTable != null) { 
		superClass = superClass.newTable;
	    }
	    superClass.nSubclasses += 1; 
	}
	Field[] allFields = clazz.getDeclaredFields();
    Arrays.sort(allFields, new Comparator<Field>() { public int compare(Field f1, Field f2) { return f1.getName().compareTo(f2.getName()); } });
	int nComponents = 0;
	for (i = 0; i < allFields.length; i++) { 
	    if ((allFields[i].getModifiers() 
		 & (Modifier.TRANSIENT|Modifier.STATIC)) == 0)
	    {
		nComponents += 1;
	    }
	}
	
	if (superClass != null) { 
	    nComponents += n = superClass.components.length;
	    nColumns = superClass.nColumns;
	}
	components = new FieldDescriptor[nComponents];
	if (superClass != null) { 
	    for (i = 0; i < n; i++) { 
		components[i] = superClass.components[i];
	    }
	    fields = superClass.fields;
	    nFields = superClass.nFields;
	    fixedSize = superClass.fixedSize;
	} else { 
	    fixedSize = Record.sizeof;
	}
	for (i = n, j = 0; j < allFields.length; j++) { 
	    if ((allFields[j].getModifiers() 
		 & (Modifier.TRANSIENT|Modifier.STATIC)) == 0)
	    {
		components[i++] = new FieldDescriptor(this, "", allFields[j], 0);
	    }
	}
	nFields += nSelfFields;
	mapColumns();
	
	try { 
	    Field f = clazz.getDeclaredField("indices");
            if (FieldDescriptor.setBypass != null) { 
                FieldDescriptor.setBypass.invoke(f, FieldDescriptor.bypassFlag);	    
            }
	    String[] indices = (String[])f.get(null);
	    for (i = indices.length; --i >= 0; ) { 
		FieldDescriptor fd = getColumn(indices[i]);
		if (fd != null) { 
		    if (fd.type >= FieldDescriptor.tpArray) { 
			throw new ConsusRuntimeError("Binary fields can not be indexed");
		    }
		    fd.attr |= FieldDescriptor.Indexed;
		} else { 
		    System.err.println("Indices specification is not correct for " + indices[i]);
		}
	    }
	} catch(Exception ex) {}

	try { 
	    Field f = clazz.getDeclaredField("primaryKeys");
            if (FieldDescriptor.setBypass != null) { 
                FieldDescriptor.setBypass.invoke(f, FieldDescriptor.bypassFlag);	    
            }
	    String[] primaryKeys = (String[])f.get(null);
	    for (i = primaryKeys.length; --i >= 0; ) { 
		FieldDescriptor fd = getColumn(primaryKeys[i]);
		if (fd != null) { 
		    fd.attr |= FieldDescriptor.PrimaryKey;
		} else { 
		    System.err.println("Indices specification is not correct for " 
				       + primaryKeys[i]);
		}
	    }
	} catch(Exception ex) {}

	String constraints = null;
	try { 
	    Field f = clazz.getDeclaredField("constraints");
            if (FieldDescriptor.setBypass != null) { 
                FieldDescriptor.setBypass.invoke(f, FieldDescriptor.bypassFlag);	    
            }
	    constraints = (String)f.get(null);
	} catch(Exception ex) {}

	if (constraints != null) { 
	    Compiler compiler = new Compiler(conxn, constraints);
	    compiler.table = this;
	    compiler.parseConstraints();
	}

	try { 
	    constructor = cls.getDeclaredConstructor(new Class[0]);
            if (FieldDescriptor.setBypass != null) { 
                FieldDescriptor.setBypass.invoke(constructor, FieldDescriptor.bypassFlag);
            }
	} catch(Exception ex) {}
    }

    //
    // Get table column by name
    //
    final FieldDescriptor getColumn(String name) { 
	return (FieldDescriptor)columnMapping.get(name);
    }

    //
    // Rename table column
    //
    final void renameColumn(FieldDescriptor fd, String newName) { 
	columnMapping.remove(fd.longName);
	columnMapping.put(newName, fd);	
    }

    //
    // Check if 'table' parameter is derived from or equal with this table descriptor
    //
    final boolean isSuperclassFor(TableDescriptor table) { 
	if (table == this) { 
	    return true;
	}
	while ((table = table.superClass) != null) { 
	    if (table == this) { 
		return true;
	    }
	}
	return false;
    }

    //
    // Get the root of thr table(class) hierarchy for this table desciptor
    //
    final TableDescriptor getRootTable() { 
	TableDescriptor table = this;
	while (table.superClass != null) { 
	    table = table.superClass;
	}
	return table;
    }

    //
    // Create declared but not existed indices
    //
    final void addIndices(boolean updateTable) throws SQLException { 
	FieldDescriptor indexedFields = null;
	FieldDescriptor fd = fields; 
	for (int n = nSelfFields; --n >= 0; fd = fd.next) { 
	    if ((fd.attr & FieldDescriptor.Indexed) != 0 && fd.index == 0) { 
		fd.index = Btree.allocate(db, fd.type);
		fd.nextIndexed = indexedFields;
		indexedFields = fd;
	    }
	}
	if (indexedFields != null) { 
	    if (updateTable) { 
		int oid = getRootTable().firstRow; 
		while (oid != 0) { 
		    byte[] rec = db.get(oid);
		    if (this.isSuperclassFor(db.getTableByOid(Record.getType(rec, 0))))
		    { 
			for (fd = indexedFields; fd != null; fd = fd.nextIndexed) {   
			    if (!Btree.insert(db, fd, oid)) { 
				throw new ConsusDuplicateException();
			    }
			}
		    }
		    oid = Record.getNext(rec, 0);
		}
	    }
	    altered = true;
	}
    }

    static int setName(byte[] rec, String str, int varyingOffs) { 
	return Varying.setString(rec, Record.sizeof, str, varyingOffs);
    }
    static void setFieldsSize(byte[] rec, int size) { 
	Varying.setSize(rec, Record.sizeof + 8, size);
    }
    static void setFieldsOffs(byte[] rec, int offs) { 
	Varying.setOffs(rec, Record.sizeof + 8, offs);
    } 
    static void setNColumns(byte[] rec, int nColumns) { 
	Bytes.pack4(rec, Record.sizeof + 16, nColumns);
    }
    static void setNComponents(byte[] rec, int nComponents) { 
	Bytes.pack4(rec, Record.sizeof + 20, nComponents);
    }
    static void setFirstRow(byte[] rec, int firstRow) { 
	Bytes.pack4(rec, Record.sizeof + 24, firstRow);
    }
    static void setLastRow(byte[] rec, int lastRow) { 
	Bytes.pack4(rec, Record.sizeof + 28, lastRow);
    }
    static void setSuperclass(byte[] rec, int superclass) { 
	Bytes.pack4(rec, Record.sizeof + 32, superclass);
    }
    static void setFixedSize(byte[] rec, int fixedSize) { 
	Bytes.pack4(rec, Record.sizeof + 36, fixedSize);
    }
    static void setAttr(byte[] rec, int attr) { 
	Bytes.pack4(rec, Record.sizeof + 40, attr);
    }
    static void setNRows(byte[] rec, int nRows) { 
	Bytes.pack4(rec, Record.sizeof + 44, nRows);
    }
    static void setTimestamp(byte[] rec, long timestamp) { 
	Bytes.pack8(rec, Record.sizeof + 48, timestamp);
    }

    static String getName(byte[] rec) { 
	return Varying.getString(rec, Record.sizeof);
    }
    static int getFieldsSize(byte[] rec) { 
	return Varying.getSize(rec, Record.sizeof + 8);
    }
    static int getFieldsOffs(byte[] rec) { 
	return Varying.getOffs(rec, Record.sizeof + 8);
    } 
    static int getNColumns(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof + 16);
    }
    static int getNComponents(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof + 20);
    }
    static int getFirstRow(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof + 24);
    }
    static int getLastRow(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof + 28);
    } 
    static int getSuperclass(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof + 32);
    }
    static int getFixedSize(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof + 36);
    }
    static int getAttr(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof + 40);
    }
    static int getNRows(byte[] rec) { 
	return Bytes.unpack4(rec, Record.sizeof + 44);
    }
    static long getTimestamp(byte[] rec) { 
	return Bytes.unpack8(rec, Record.sizeof + 48);
    }
    static final int descSize = Record.sizeof + 64; 
    

    void registerInDatabase() throws SQLException { 
	addIndices(false);
	byte[] desc = exportDescriptor();
	tableId = db.allocateRow(db.metatable, desc); 
	db.registerTable(this);
    }

    //
    // Produce packed image of table descriptor to be stored in the database file
    //
    byte[] exportDescriptor() throws SQLException 
    {
	if (superClass != null && superClass.tableId == 0) { 
	    superClass.registerInDatabase();
	}	
	int n;
	FieldDescriptor fd = fields;
	int totalLength = name.length();

	for (n = nSelfFields; --n >= 0; fd = fd.next) {
	    totalLength += fd.longName.length();
	    if (fd.className != null)   totalLength += fd.className.length();
	    if (fd.fkTableName != null) totalLength += fd.fkTableName.length();
	    if (fd.fkKeyName != null)   totalLength += fd.fkKeyName.length();
	}
	int size = TableDescriptor.descSize + totalLength*2
            	 + FieldDescriptor.descSize*nSelfFields;
	desc = new byte[size];
	Record.setSize(desc, 0, size);
	Record.setType(desc, 0, ConsusConnection.dbMetaTableId);
	int offs = TableDescriptor.descSize;
	offs = setName(desc, name, offs);
	setFieldsSize(desc, nSelfFields);
	setFieldsOffs(desc, offs);
	int fieldsOffs = offs += nSelfFields*FieldDescriptor.descSize;
	fd = fields;
	for (n = nSelfFields; --n >= 0; fd = fd.next) {
	    fieldsOffs -= FieldDescriptor.descSize;
	    offs = FieldDescriptor.setName(desc, fieldsOffs, fd.longName, offs);
	    offs = FieldDescriptor.setClassName(desc, fieldsOffs, 
						fd.className != null ? fd.className : "", offs);
	    FieldDescriptor.setType(desc, fieldsOffs, fd.type);
	    FieldDescriptor.setAttr(desc, fieldsOffs, fd.attr & FieldDescriptor.ExportMask);
	    FieldDescriptor.setOffset(desc, fieldsOffs, fd.offs);
	    FieldDescriptor.setNcomponents(desc, fieldsOffs, 
					   fd.components != null ? fd.components.length : 0);
	    FieldDescriptor.setIndex(desc, fieldsOffs, fd.index);	    
	    offs = FieldDescriptor.setFkTableName(desc, fieldsOffs, 
						  fd.fkTableName != null ? fd.fkTableName : "", offs);
	    offs = FieldDescriptor.setFkKeyName(desc, fieldsOffs, 
						fd.fkKeyName != null ? fd.fkKeyName : "", offs);
	}
	setNColumns(desc, nColumns);
	setNComponents(desc, components.length);
	setFirstRow(desc, firstRow);
	setLastRow(desc, lastRow);
	setNRows(desc, nRows);
	setTimestamp(desc, timestamp);
	setAttr(desc, attr);
	setSuperclass(desc, superClass != null ? superClass.tableId : 0);
	setFixedSize(desc, fixedSize);
	Assert.that(size == offs);
	return desc;
    }
    
    //
    // Unpack table deriptor from the database file
    //
    TableDescriptor(ConsusConnection conxn, int tableId) throws SQLException { 
	this.tableId = tableId;
	desc = conxn.get(tableId);
	db = conxn;
	name = getName(desc);
	firstRow = getFirstRow(desc);
	lastRow = getLastRow(desc);
	nRows = getNRows(desc);
	timestamp = getTimestamp(desc);
	nColumns = getNColumns(desc);
	components = new FieldDescriptor[getNComponents(desc)];
	nSelfFields = nFields = getFieldsSize(desc);
	fixedSize = getFixedSize(desc);
	attr = getAttr(desc);
	created = true;

	int superClassId = getSuperclass(desc);
	int n, i = 0;
	if (superClassId != 0) { 
	    superClass = conxn.getTableByOid(superClassId);
	    Assert.that(superClass != null);
	    superClass.nSubclasses += 1; 
	    nFields += superClass.nFields;
	    fields = superClass.fields;
	    for (n = superClass.components.length; --n >= 0; i++) {
		components[i] = superClass.components[i];
	    }
	}
	int fieldsOffs = getFieldsOffs(desc);
	for (n = components.length; i < n; i++) { 
	    FieldDescriptor fd = new FieldDescriptor(this, desc, fieldsOffs, 0);
	    components[i] = fd;
	    fieldsOffs += FieldDescriptor.descSize;
	    if (fd.type == FieldDescriptor.tpStructure) { 
		fieldsOffs = fd.unpackFields(this, desc, fieldsOffs, 0);
	    } else if (fd.type == FieldDescriptor.tpArray) {
		fieldsOffs = fd.unpackFields(this, desc, fieldsOffs, 
					     FieldDescriptor.ComponentOfArray);
	    }	    
	}
	mapColumns();
    }

    //
    // Constructor used for "create table". Information about table fields will be specified later.
    // After it completeDeclaration method will be called to complete initialization of table descriptor
    //
    TableDescriptor(ConsusConnection conxn, String name) { 
	this.name = name;
	db = conxn;
	fixedSize = Record.sizeof; 
    }

    //
    // Complete initialization of the table descriptor while processing "create table" statement
    //
    void completeDeclaration() { 
	int n = nColumns;
	// As far as not array or structure fields are allowed in "create table", columns == components 
	nFields = nSelfFields;
	int i = nFields;
	columns = components = new FieldDescriptor[n]; 
	FieldDescriptor fd = fields;
	while (--n >= 0) { 
	    fd.id = --i;
	    if ((fd.attr & FieldDescriptor.ComponentOfArray) != 0) { 
		fd = fd.next; 
		fd.id = --i;
	    }
	    columns[n] = fd;
	    fd.columnIndex = n;
	    columnMapping.put(fd.longName, fd);
	    fd = fd.next;
	}
	Assert.that(i == 0);
    }
    
    //
    // Try to match two table descriptors. Matching is based on the names of the fields.
    // If two field descriptors are the same in two tables, then index for this fields can be reused.
    // This method is invoked only if equals method returns false and used to prepare table descriptor
    // to perform convertion to the new format.
    //
    void match(TableDescriptor old) throws SQLException { 
	FieldDescriptor fd, old_fd;
	int nMatches = 0;
	boolean formatChanged = (nFields != old.nFields);

	if (numberOfBlobs != old.numberOfBlobs) { 
	    // BLOB fields can not be removed
	    throw new ConsusIncompatibleTypesError();
	}

	for (fd = fields; fd != null; fd = fd.next) { 
	    fd.oldType = FieldDescriptor.tpUnknown;	
	    for (old_fd = old.fields; old_fd != null; old_fd = old_fd.next) { 
		if (fd.longName.equals(old_fd.longName)) { 
		    if ((fd.type < FieldDescriptor.tpString 
			 && old_fd.type < FieldDescriptor.tpString)
			|| (fd.type == old_fd.type 
			    && fd.type >= FieldDescriptor.tpString))
		    {
			fd.oldType = old_fd.type;
			fd.oldOffs = old_fd.offs;
			fd.oldSize = old_fd.size;
			if (fd.className == null) {
			    fd.className = old_fd.className;
			}
			if (fd.type != fd.oldType || fd.offs != fd.oldOffs) { 
			    formatChanged = true;
			}
			nMatches += 1;
			if (fd.type == fd.oldType) { 
			    if (old_fd.index != 0) { 
				// reuse index
				fd.index = old_fd.index; 
				old_fd.index = 0; 
			    }
			} else if (old_fd.index != 0 && fd.index == 0) { // create new index
			    fd.index = Btree.allocate(db, fd.type);
			}
			break;
		    } else { 
			throw new ConsusIncompatibleTypesError();
		    }
		}
	    }
	}
	if (nMatches != nFields) { 
	    formatChanged = true;
	}
	old.formatChanged = formatChanged;
    }
	
    // 
    // Test if descriptor of two tables are the same
    //
    boolean equals(TableDescriptor old) { 
	FieldDescriptor fd = fields, old_fd = old.fields;
	if (nFields != old.nFields || fixedSize != old.fixedSize) {
	    return false;
	}
	while (fd != null) { 
	    if (old_fd == null || !fd.equals(old_fd)) { 
		return false;
	    }
	    if (old_fd.index != 0) { 
		fd.index = old_fd.index;
	    }
	    fd = fd.next;
	    old_fd = old_fd.next;
	}
	return old_fd == null;
    }	
    
    //
    // Calculate size of the table record for the specified Java object
    //
    int calculateRecordSize(Object obj) throws SQLException
    {
	return FieldDescriptor.calculateRecordSize(components, obj, fixedSize);
    } 

    //
    // Create byte array with packed database image of the object
    //
    final void storeObject(byte[] rec, Object obj) throws SQLException
    { 
	FieldDescriptor.storeRecordFields(components, rec, 0, fixedSize, obj);
	incrementVersionNumber(rec);
    }
    
    final Object construct() throws SQLException { 
	if (constructor == null) { 
	    throw new ConsusRuntimeError("No Java class is mapped to the table " + name);
	}
	try { 
	    return constructor.newInstance(constructorArgs);
	}  
	catch(IllegalAccessException ex) { throw new IllegalAccessError(); }
	catch(InstantiationException ex) { throw new InstantiationError(); }
	catch(Exception ex) { 
	    throw new InstantiationError("Exception was thrown by constructor of " + name);
	}
    }


    //
    // Unpack object from the packed image
    //
    final Object fetchObject(int oid, byte[] src) throws SQLException
    { 
	Object obj = construct();
	if (obj instanceof Persistent) { 
	    Persistent po = (Persistent)obj;
	    po.oid  = oid;
	    po.type = this;
	    po.db   = db;
	    po.state = Persistent.PO_LOADED;
	}
	if (db.objectCache != null) { 
	    db.objectCache.put(oid, obj);
	}
	fetchObject(obj, src);
	return obj; 
    }

    final void fetchObject(Object obj, byte[] src) throws SQLException
    { 
	if (obj instanceof Persistent) { 
	    Persistent po = (Persistent)obj;
	    po.state = Persistent.PO_LOADED;
	    FieldDescriptor.fetchRecordFields(components, po, src, 0, po.loadClosure());
	} else { 
	    FieldDescriptor.fetchRecordFields(components, obj, src, 0, false);
	}
    }

    final void loadComponents(Object obj) throws SQLException 
    {
	FieldDescriptor.loadComponents(components, obj);
    }

    final boolean markObject(byte[] rec, int[] greyBitmap, int[] blackBitmap)
    {
	return FieldDescriptor.markObjects(components, rec, 0, greyBitmap, blackBitmap);
    }

    //
    // Convert record to the new format (scheme evaluation). Conversion is nased on the information 
    // prepared by match() method.
    //
    byte[] convertRecord(byte[] oldRec) throws SQLException { 
	byte[] newRec = 
	    new byte[FieldDescriptor.calculateNewRecordSize(components, oldRec, 0, fixedSize)];
	FieldDescriptor.convertRecord(components, newRec, 0, oldRec, 0, fixedSize);
	return newRec;
    }

    void allocateBlobs(byte[] newRec, byte[] oldRec, Object obj) throws SQLException { 
	if (numberOfBlobs != 0) { 
	    try { 
		for (FieldDescriptor fd = fields; fd != null; fd = fd.next) { 
		    if (fd.type == FieldDescriptor.tpBlob) { 
			Blob blob = (Blob)fd.field.get(obj); 
			if (oldRec != null) { 
			    long pos = Bytes.unpack8(oldRec, fd.offs);
			    if (pos != 0) { 
				if (!(blob instanceof ConsusBlob) || ((ConsusBlob)blob).pos != pos) 
				{
				    long size = Bytes.unpack8(oldRec, fd.offs+8);
				    db.cloneBitmap(pos, size);
				    db.free(pos, size);
				} else { 
				    System.arraycopy(oldRec, fd.offs, newRec, fd.offs, 16);
				    continue;
				}
			    }
			}
			if (blob == null || blob.length() == 0) { 
			    Bytes.pack8(newRec, fd.offs, 0);
			    Bytes.pack8(newRec, fd.offs+8, 0);
			} else { 
			    long pos = db.storeBlob(blob);
			    long size = blob.length();
			    Bytes.pack8(newRec, fd.offs, pos);
			    Bytes.pack8(newRec, fd.offs+8, size);
			    fd.field.set(obj, new ConsusBlob(db, pos, size));
			}
		    } else if (fd.type == FieldDescriptor.tpClob) { 
			Clob clob = (Clob)fd.field.get(obj); 
			long pos = Bytes.unpack8(oldRec, fd.offs);
			if (oldRec != null) { 
			    if (pos != 0) { 
				if (!(clob instanceof ConsusClob)|| ((ConsusClob)clob).pos != pos) 
				{
				    long size = Bytes.unpack8(oldRec, fd.offs+8);
				    db.cloneBitmap(pos, size);
				    db.free(pos, size);
				} else { 
				    System.arraycopy(oldRec, fd.offs, newRec, fd.offs, 16);
				    continue;
				}
			    }
			}
			if (clob == null || clob.length() == 0) { 
			    Bytes.pack8(newRec, fd.offs, 0);
			    Bytes.pack8(newRec, fd.offs+8, 0);
			} else { 
			    Bytes.pack8(newRec, fd.offs, db.storeClob(clob));
			    Bytes.pack8(newRec, fd.offs+8, clob.length()*2);
			}
		    }
		}
	    } catch(Exception x) { 
		throw new ConsusRuntimeError("Failed to get object component: " + x.getMessage());
	    }
	}
    }
			    
	

    //
    // Compare new and old record data to mark updated fields in table descriptor,
    // This method is used to update indices.
    //
    void markUpdatedFields(byte[] newRec, byte[] oldRec, boolean updated[]) 
    {
	for (FieldDescriptor fd = fields; fd != null; fd = fd.next) { 
	    if (fd.index != 0) { 
		int offs = fd.offs;
		if (fd.type == FieldDescriptor.tpString) { 
		    int len = Varying.getSize(oldRec, offs);
		    if (len == Varying.getSize(newRec, offs)) { 
			len *= 2;
			int newOffs = Varying.getSize(newRec, offs);
			int oldOffs = Varying.getSize(oldRec, offs);
			while (--len >= 0) { 
			    if (newRec[newOffs++] != oldRec[oldOffs++]) { 
				updated[fd.id] = true;
				break;
			    }
			} 
		    } else { 
			updated[fd.id] = true;
		    }
		} else { 
		    int len = FieldDescriptor.sizeof[fd.type];
		    while (--len >= 0) { 
			if (newRec[offs] != oldRec[offs]) { 
			    updated[fd.id] = true;
			    break;
			}
                        offs += 1;
		    }
		}
	    }
	}
    }

    //
    // Remove record from the indices for update fields
    //
    void removeFromIndices(int oid, boolean updated[]) throws SQLException { 
	if (timestampField != null && !updated[timestampField.id]
	    && timestampField.index != 0) 
	{ 
	    boolean success = Btree.remove(db, timestampField, oid);
	    Assert.that(success);
	}
	for (FieldDescriptor fd = fields; fd != null; fd = fd.next) { 
	    if (fd.index != 0 && updated[fd.id]) { 
		boolean success = Btree.remove(db, fd, oid);
		Assert.that(success);
	    }
	}
    }

    //
    // Reinsert updated object in indices according to the new values of updated columns
    //
    boolean updateIndices(int oid, boolean updated[]) throws SQLException { 
	if (timestampField != null && !updated[timestampField.id]
	    && timestampField.index != 0) 
	{ 
	    Btree.insert(db, timestampField, oid);
	}
	for (FieldDescriptor fd = fields; fd != null; fd = fd.next) { 
	    if (updated[fd.id] && fd.index != 0) { 
		if (!Btree.insert(db, fd, oid)) { 
		    for (FieldDescriptor fd2 = fields; fd2 != fd; fd2 = fd2.next) { 
			if (updated[fd2.id] && fd2.index != 0) { 
			    boolean success = Btree.remove(db, fd2, oid);
			    Assert.that(success);
			}
		    }
		    if (timestampField != null) { 
			updated[timestampField.id] = false;
		    }
		    return false;			
		}
	    }
	}
	return true;
    }

    //
    // Insert new record in indices.
    //
    boolean insertInIndices(int oid) throws SQLException { 
	for (FieldDescriptor fd = fields; fd != null; fd = fd.next) { 
	    if (fd.index != 0) { 
		if (!Btree.insert(db, fd, oid)) {
		    if (db.dbTraceEnabled) { 
			System.out.println("UNIQUE constrian violated for field " + fd.name);
		    }
		    return false;
		}
	    }
	}
	return true;
    }



    //
    // Create updated packed image of the database record. 
    // New value of the updated field of the record 
    // is stored in "value" component of FieldDescriptor.
    //
    byte[] updateRecord(ConsusStatement stmt, byte[] oldRec) throws SQLException { 
	byte[] newRec = 
	    new byte[FieldDescriptor.calculateUpdatedRecordSize(stmt, components, oldRec, 0, fixedSize)];
	FieldDescriptor.updateRecord(stmt, components, newRec, 0, oldRec, 0, fixedSize);
	incrementVersionNumber(newRec);
	return newRec;
    }

    //
    // Let Consus automatically inclrement objectVersionNumber fields at each 
    // insert or update operation.
    //
    void setVersionAutoincrementMode(String timestampFieldName) { 
	if (timestampField == null) { 
	    timestampField = getColumn(timestampFieldName);
	}
    }

    //
    // This method is invoked if version autoincrement mode is switched on
    // for the table and record is updated or inserted
    //
    void incrementVersionNumber(byte[] rec) { 
	if (timestampField != null) { 
	    switch (timestampField.type) { 
	      case FieldDescriptor.tpInt8:
		Bytes.pack8(rec, timestampField.offs, timestamp+1);
		break;
	      case FieldDescriptor.tpInt4:
		Bytes.pack4(rec, timestampField.offs, (int)timestamp+1);
		break;
	      default:
		Assert.failed("Timestamp should be of integer or long type");
	    }
	}
    }

    //
    // Get first suitable primary key for the table
    //
    FieldDescriptor getPrimaryKey() { 
	for (int i = 0, n = columns.length; i < n; i++) { 
	    if ((columns[i].attr & FieldDescriptor.PrimaryKey) != 0) { 
		return columns[i];
	    }
	}
	return null;
    }
}




