/* Consus - Pure Java Object Relational Database Management System
 * Table iterator
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.SQLException;

/**
 * Iterator through table records (is equivalent to RsultSet for single table select 
 * without condition but is more efficient)
 */
public class TableIterator extends org.garret.consus.TableIterator { 
    ConsusConnection db;
    TableDescriptor  table;
    TableDescriptor  baseTable;
    int              currPos;
    int              currOid;

    /**
     * Construct table iterator
     * @param <code>conxn</code> database connection
     * @param <code>tableName</code> name of the iterated table
     */
    public TableIterator(ConsusConnection conxn, String tableName) throws SQLException { 
	db = conxn;
	table = conxn.getTableByName(tableName);
	baseTable = table.getRootTable();
    }
       
    /**
     * Moves the cursor down one row from its current position.
     * A ResultSet cursor is initially positioned before the first row; the
     * first call to next makes the first row the current row; the
     * second call makes the second row the current row, and so on. 
     *
     * @return true if the new current row is valid; false if there
     * are no more rows 
     * @exception SQLException if a database access error occurs
     */
    public boolean next() throws SQLException { 
	if (currPos > table.nRows) { 
	    return false;
	}
	if (currPos == 0) { 
	    currOid = baseTable.firstRow;
	} else { 
	    Assert.that(currOid != 0);
	    currOid = db.getNext(currOid);
	}
	if (table != baseTable) { 
	    while (currOid != 0) { 
		int tableId = db.getType(currOid);
		if (table.isSuperclassFor(db.getTableByOid(tableId))) { 
		    break;
		}
		currOid = db.getNext(currOid);
	    }
	}
	if (currPos != 0 || currOid != 0) { 
	    currPos += 1;
	}
	return currOid != 0;
    }

    /**
     * <p>Moves the cursor to the previous row in the result set.  
     *
     * <p>Note: <code>previous()</code> is not the same as
     * <code>relative(-1)</code> because it
     * makes sense to call</code>previous()</code> when there is no current row.
     *
     * @return true if the cursor is on a valid row; false if it is off the result set
     * @exception SQLException if a database access error occurs or the
     * result set type is TYPE_FORWARD_ONLY
     */
    public boolean previous() throws SQLException { 
	if (currPos == 0) { 
	    return false;
	}
	if (currPos == table.nRows+1) { 
	    currOid = baseTable.lastRow;
	} else { 
	    Assert.that(currOid != 0);
	    currOid = db.getPrev(currOid);
	}
	if (table != baseTable) { 
	    while (currOid != 0) { 
		int tableId = db.getType(currOid);
		if (table.isSuperclassFor(db.getTableByOid(tableId))) { 
		    break;
		}
		currOid = db.getPrev(currOid);
	    }
	}
	currPos -= 1;
	return currOid != 0;
    }

    /** 
     * Fetch current table row as Java object
     */
    public Object get() throws SQLException { 
	if (currOid == 0) { 
	    return null;
	}
	return db.get(new ConsusRef(db, currOid));
    }

    /** 
     * Closes iterator and release all related resourses. 
     */
    public void close()  throws SQLException { 
    }  
  
    /**
     * <p>Indicates whether the cursor is before the first row in the result 
     * set.   
     *
     * @return true if the cursor is before the first row, false otherwise. Returns
     * false when the result set contains no rows.
     * @exception SQLException if a database access error occurs
     */
    public boolean isBeforeFirst() throws SQLException { 
	return currPos == 0;
    }
      
    /**
     * <p>Indicates whether the cursor is after the last row in the result 
     * set.   
     *
     * @return true if the cursor is  after the last row, false otherwise.  Returns
     * false when the result set contains no rows.
     * @exception SQLException if a database access error occurs
     */
    public boolean isAfterLast() throws SQLException {
	return currPos > table.nRows;
    }
 
    /**
     * <p>Indicates whether the cursor is on the first row of the result set.   
     *
     * @return true if the cursor is on the first row, false otherwise.   
     * @exception SQLException if a database access error occurs
     */
    public boolean isFirst() throws SQLException { 
	return currPos == 1;
    }
 
    /**
     * <p>Indicates whether the cursor is on the last row of the result set.   
     *
     * @return true if the cursor is on the last row, false otherwise. 
     * @exception SQLException if a database access error occurs
     */
    public boolean isLast() throws SQLException { 
	return currPos == table.nRows;
    }

    /**
     * <p>Moves the cursor to the front of the result set, just before the
     * first row. Has no effect if the result set contains no rows.
     *
     * @exception SQLException if a database access error occurs or the 
     * result set type is TYPE_FORWARD_ONLY
     */
    public void beforeFirst() throws SQLException { 
        currPos = 0;
	currOid = 0;
    }
    

    /**
     * <p>Moves the cursor to the end of the result set, just after the last
     * row.  Has no effect if the result set contains no rows.
     *
     * @exception SQLException if a database access error occurs or the 
     * result set type is TYPE_FORWARD_ONLY 
     */
    public void afterLast() throws SQLException { 
	currPos = table.nRows+1;
	currOid = 0;
    }
	
    /**
     * <p>Moves the cursor to the first row in the result set.  
     *
     * @return true if the cursor is on a valid row; false if
     *         there are no rows in the result set
     * @exception SQLException if a database access error occurs or the 
     * result set type is TYPE_FORWARD_ONLY
     */
    public boolean first() throws SQLException { 
	beforeFirst();
	return next();
    }

    /**
     * <p>Moves the cursor to the last row in the result set.  
     *
     * @return true if the cursor is on a valid row;
     * false if there are no rows in the result set
     * @exception SQLException if a database access error occurs or the
     * result set type is TYPE_FORWARD_ONLY.
     */
    public boolean last() throws SQLException { 
	afterLast();
	return previous();
    }

    /**
     * <p>Retrieves the current row number.  The first row is number 1, the
     * second number 2, and so on.  
     *
     * @return the current row number; 0 if there is no current row
     * @exception SQLException if a database access error occurs
     */
    public int getRow() throws SQLException { 
	return currPos;
    }

    /**
     * <p>Moves the cursor to the given row number in the result set.
     *
     * <p>If the row number is positive, the cursor moves to 
     * the given row number with respect to the
     * beginning of the result set.  The first row is row 1, the second
     * is row 2, and so on. 
     *
     * <p>If the given row number is negative, the cursor moves to
     * an absolute row position with respect to
     * the end of the result set.  For example, calling
     * <code>absolute(-1)</code> positions the 
     * cursor on the last row, <code>absolute(-2)</code> indicates the next-to-last
     * row, and so on.
     *
     * <p>An attempt to position the cursor beyond the first/last row in
     * the result set leaves the cursor before/after the first/last
     * row, respectively.
     *
     * <p>Note: Calling <code>absolute(1)</code> is the same
     * as calling <code>first()</code>.
     * Calling <code>absolute(-1)</code> is the same as calling <code>last()</code>.
     *
     * @return true if the cursor is on the result set; false otherwise
     * @exception SQLException if a database access error occurs or 
     * row is 0, or result set type is TYPE_FORWARD_ONLY.
     */
    public boolean absolute(int row) throws SQLException { 
	while (row > currPos) { 
	    if (!next()) { 
		return false;
	    }
	}
	while (row < currPos) { 
	    if (!previous()) { 
		return false;
	    }
	}
	return true;
    }

    /**
     * <p>Moves the cursor a relative number of rows, either positive or negative.
     * Attempting to move beyond the first/last row in the
     * result set positions the cursor before/after the
     * the first/last row. Calling <code>relative(0)</code> is valid, but does
     * not change the cursor position.
     *
     * <p>Note: Calling <code>relative(1)</code>
     * is different from calling <code>next()</code>
     * because is makes sense to call <code>next()</code> when there is no current row,
     * for example, when the cursor is positioned before the first row
     * or after the last row of the result set.
     *
     * @return true if the cursor is on a row; false otherwise
     * @exception SQLException if a database access error occurs, there
     * is no current row, or the result set type is TYPE_FORWARD_ONLY
     */
    public boolean relative(int rows) throws SQLException {
	while (rows > 0) { 
	    rows -= 1;
	    if (!next()) { 
		return false;
	    }
	}
	while (rows < 0) { 
	    rows += 1;
	    if (!previous()) { 
		return false;
	    }
	}
	return true;
    }

    /**
     * Deletes the current row from the result set and the underlying
     * database.  Cannot be called when on the insert row.
     *
     * @exception SQLException if a database access error occurs or if
     * called when on the insert row.
     */
    public void deleteRow() throws SQLException {
	Assert.that(currOid != 0);
	int oid = currOid;
	next();
	db.remove(new ConsusRef(oid));
	currPos -= 1;
    }

    /**
     * Deletes all records in the table
     * 
     * @exception SQLException if a database access error occurs or if
     * called when on the insert row.
     */
    public void deleteAllSelectedRows() throws SQLException
    {
	db.deleteTable(table);
    }

    /**
     * Get number of rows in the table
     * @return number of rows in the table
     */
    public int getNumberOfRows() throws SQLException { 
	return table.nRows;
    }

    /**
     * Select object from the table using unique key 
     * @param keyName table column with unique key
     * @param keyValue value of the key to look for
     * @return object with specified value of the key or <code>null</code> otherwise
     * @exception ConsusRuntimeException when there is no such column in the table or
     * it is not indexed; ConsusNotUniqueException when key is not unique
     */
    public Object selectByKey(String keyName, String keyValue) throws SQLException { 
	db.beginTransaction(false);
	ConsusStatement stmt = new ConsusStatement(db);
	stmt.cursor.open(table, false);
	FieldDescriptor fd = table.getColumn(keyName);
	if (fd == null) { 
	    throw new ConsusRuntimeError("No such field " + keyName + " in table " + table.name);	    
	}
	if (fd.index == 0) { 
	    throw new ConsusRuntimeError("No index for field " + keyName + " in table " + table.name);	    
	}
	SearchContext sc = new SearchContext();
	SearchValue sv = new SearchValue();
	sv.sval = keyValue.toCharArray();
	sv.inclusion = 1;	
	sc.firstKey = sc.lastKey = sv;
	sc.stmt = stmt;
	sc.condition = null;       	
	Btree.find(db, fd.index, sc);
	Object obj = null;
	if (stmt.cursor.nSelected > 1) { 
	    stmt.close();
	    throw new ConsusNotSingleException("Key " + keyName + " is not unique");
	} else if (stmt.cursor.next()) { 
	    obj = stmt.cursor.getSelfObject();
	}
	stmt.close();
	return obj;
    }

    public Object selectByOid(long oid) throws SQLException
    {
	return db.get(new ConsusRef(oid));
    }
}








