/* Consus - Pure Java Object Relational Database Management System
 * Node for unary operations
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.SQLException;


class UnaryOpNode extends ExprNode { 
    ExprNode opd;

    String dump() { 
	return "(" + opMnem[tag] + " " + opd.dump() + ")";
    }
    boolean equalsNode(ExprNode node) { 
	return node.tag == tag && opd.equalsNode(((UnaryOpNode)node).opd);
    }

    int dependsOn() { return opd.dependsOn(); }

    long evaluateInt(ConsusStatement stmt) throws SQLException { 
	long val;
	switch (tag) { 
	  case opIntNot:
	    return ~opd.evaluateInt(stmt);
	  case opIntNeg:
	    return -opd.evaluateInt(stmt);
	  case opIntAbs:
	    val = opd.evaluateInt(stmt);
	    return val < 0 ? -val : val;
	  case opRealToInt:
	    return (long)opd.evaluateReal(stmt);
	  case opLength:
	    return Varying.getSize(stmt.rec, opd.evaluateBase(stmt));
	  case opStrLength:
	    return opd.evaluateStr(stmt).length();
	  case opStrToInt:
	    try { 
		return Long.parseLong(opd.evaluateStr(stmt), 10);
	    } catch (NumberFormatException x) { 
		throw new ConsusRuntimeError("Failed to convert string to integer");
	    }
	  case opBoolToInt:
	    return opd.evaluateBool(stmt) ? 1 : 0;

	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	return 0;
    }
    
    double evaluateReal(ConsusStatement stmt) throws SQLException { 
	double val;
	switch (tag) { 
	  case opRealNeg:
	    return -opd.evaluateReal(stmt);
	  case opRealAbs:
	    val = opd.evaluateReal(stmt);
	    return val < 0 ? -val : val;
	  case opRealSin:
	    return Math.sin(opd.evaluateReal(stmt));
	  case opRealCos:
	    return Math.cos(opd.evaluateReal(stmt));
	  case opRealTan:
	    return Math.tan(opd.evaluateReal(stmt));
	  case opRealAsin:
	    return Math.asin(opd.evaluateReal(stmt));
	  case opRealAcos:
	    return Math.acos(opd.evaluateReal(stmt));
	  case opRealAtan:
	    return Math.atan(opd.evaluateReal(stmt));
	  case opRealExp:
	    return Math.exp(opd.evaluateReal(stmt));
	  case opRealLog:
	    return Math.log(opd.evaluateReal(stmt));
	  case opRealSqrt:
	    return Math.sqrt(opd.evaluateReal(stmt));
	  case opRealCeil:
	    return Math.ceil(opd.evaluateReal(stmt));
	  case opRealFloor:
	    return Math.floor(opd.evaluateReal(stmt));
	  case opIntToReal:
	    return (double)opd.evaluateInt(stmt);
	  case opStrToReal:
	    try { 
		return Double.parseDouble(opd.evaluateStr(stmt));
	    } catch (NumberFormatException x) { 
		throw new ConsusRuntimeError("Failed to convert string to real");
	    }
	  case opBoolToReal:
	    return opd.evaluateBool(stmt) ? 1.0 : 0.0;
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	return 0;
    }
    
    String evaluateStr(ConsusStatement stmt) throws SQLException { 
	switch (tag) { 
	  case opStrUpper:
	    return opd.evaluateStr(stmt).toUpperCase();
	  case opStrLower:
	    return opd.evaluateStr(stmt).toLowerCase();
	  case opIntToStr:
	    return Long.toString(opd.evaluateInt(stmt), 10);
	  case opRealToStr:
	    return Double.toString(opd.evaluateReal(stmt));
	  case opBoolToStr:
	    return opd.evaluateBool(stmt) ? "true" : "false";
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	return null;
    }
    
    boolean evaluateBool(ConsusStatement stmt) throws SQLException { 
	switch (tag) { 
	  case opBoolNot:
	    return !opd.evaluateBool(stmt);
	  case opIsNull:
	    return opd.evaluateOid(stmt) == 0;
	  case opSetNotEmpty:
	    return opd.evaluateSet(stmt).length != 0;
	  case opIntToBool:
	    return opd.evaluateInt(stmt) != 0;
	  case opStrToBool:
	    return opd.evaluateStr(stmt).equalsIgnoreCase("true");
	  case opRealToBool:
	    return opd.evaluateReal(stmt) != 0;
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	return false;
    }

    UnaryOpNode(int type, int tag, ExprNode node) { 
	super(type, tag);
	opd = node;
    }
}
