/* Consus - Pure Java Object Relational Database Management System
 * Test of JSQL iterators, instance manager, lock manager and memory allocation.
 * Created by Konstantin Knizhnik, Mar 2000.
 */

import java.sql.*;
import org.garret.consus.*;

class Queue { 
    Ref head;
    int nElems;
    transient Consumer consumer;
    transient Producer producer;

    void waitProducer(Consumer proc) throws Exception { 
	synchronized (proc) { 
	    consumer = proc;
	    proc.db.commitSubtransaction();
	    proc.wait();
	}
    }
    void waitConsumer(Producer proc) throws Exception { 
	synchronized (proc) { 
	    producer = proc;
	    proc.db.commitSubtransaction();
	    proc.wait();
	}
    }
    void notifyConsumer() { 
	if (consumer != null) {
	    synchronized(consumer) { 
		consumer.notify();
		consumer = null;
	    }
	}
    }
    void notifyProducer() { 
	if (producer != null) {
	    synchronized(producer) { 
		producer.notify();
		producer = null;
	    }
	}
    }
}

class QueueNode { 
    Ref  next;
    int  prty;
    byte data[];
};

class Consumer implements Runnable {
    ConsusConnection db;
    int checksum;
    int n;
    
    Consumer(ConsusConnection conxn, int nItemsToProceed) { 
	db = conxn;
	n = nItemsToProceed;
    }

    public void run() {
	try { 
	    consume();
	} catch(Exception x) { 
	    x.printStackTrace();
	}
    }

    void consume() throws Exception { 
	ConsusPreparedStatement pstmt = (ConsusPreparedStatement)
	    db.prepareStatement("select * from Queue where OID=FIRST for update");
	ConsusPreparedStatement pstmt2 = (ConsusPreparedStatement)
	    db.prepareStatement("select * from QueueNode start from ? following by next for update");
	Queue queue;
	while (true) { 
	    ConsusResultSet cursor = (ConsusResultSet)pstmt.executeQuery();
	    if (!cursor.next()) { 
	        queue = new Queue();
		queue.consumer = this;
		pstmt.insert(queue);
	    } else { 
		queue = (Queue)cursor.getSelfObject();
	    }
	    if (queue.nElems == 0) { 
		queue.waitProducer(this);
	    } else { 
		pstmt2.setRef(1, queue.head);
		ConsusResultSet cursor2 = (ConsusResultSet)pstmt2.executeQuery();
		int maxPos = 0;
		int maxPrty = -1;
		QueueNode maxNode = null;
		Assert.that(cursor2.getSelectionSize() == queue.nElems);
		for (int pos = 1; cursor2.next(); pos++) { 
		    QueueNode node = (QueueNode)cursor2.getSelfObject();
		    if (node.prty > maxPrty) { 
			maxPrty = node.prty;
			maxNode = node;
			maxPos = pos;
		    }
		}
		cursor2.absolute(maxPos);
		cursor2.deleteRow();
		if (maxPos == 1) { 
		    queue.head = maxNode.next;
		} else { 
		    cursor2.absolute(maxPos-1);
		    QueueNode node = (QueueNode)cursor2.getSelfObject();
		    node.next = maxNode.next;
		    cursor2.updateRow();
		}
		queue.nElems -= 1;
		checksum ^= maxNode.prty;
		cursor.updateRow();
		queue.notifyProducer();
		db.commitSubtransaction();
		if (--n <= 0) {
		    return;
		}
	    }
	}
	
    }
}

class Producer implements Runnable { 
    ConsusConnection db;
    int checksum;
    int n;
    static final int maxQueueSize = 50;
    
    Producer(ConsusConnection conxn, int nItemsToProceed) { 
	db = conxn;
	n = nItemsToProceed;
    }

    public void run() {
	try { 
	    produce();
	} catch(Exception x) { 
	    x.printStackTrace();
	}
    }

    void produce() throws Exception { 
	ConsusPreparedStatement pstmt = (ConsusPreparedStatement)
	    db.prepareStatement("select * from Queue where OID=FIRST for update");
	Queue queue;
	java.util.Random rand = new java.util.Random(2000);
	int i = 0;
	while (true) { 
	    ConsusResultSet cursor = (ConsusResultSet)pstmt.executeQuery();
	    if (!cursor.next()) { 
		queue = new Queue();
		pstmt.insert(queue);
		continue;
	    } else { 
		queue = (Queue)cursor.getSelfObject();
	    }
	    if (queue.nElems == maxQueueSize) { 
		queue.waitConsumer(this);
	    } else { 
		QueueNode node = new QueueNode();
		int r = rand.nextInt(10000);
		Assert.that(r < 10000);
		node.next = queue.head;
		node.prty = r;
		node.data = new byte[r];
		queue.head = pstmt.insert(node);
		queue.nElems += 1;
		Assert.that(queue.nElems <= maxQueueSize);
		cursor.updateRow();
		checksum ^= r;	
		queue.notifyConsumer();
		db.commitSubtransaction();
		if ((++i % 100) == 0) { 
		    System.out.print("Proceed " + i + " items...\r");
		    System.out.flush();
		}
		if (i == n) {
		    return;
		}
	    }
	}
    }
}

public class TestAlloc { 
    static int nIterations = 100000;
    
    static public void main(String[] args) throws Exception {
	Class.forName("org.garret.consus.ConsusDriver");
	ConsusConnection connection = 
	    (ConsusConnection)DriverManager.getConnection("jdbc:consus:TestAlloc.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database TestAlloc");
	    return;
	}
	connection.createObjectCache();
	Consumer consumer = new Consumer(connection, nIterations);
	Producer producer = new Producer(connection, nIterations);
	Thread t1 = new Thread(consumer);
	t1.start();
	Thread t2 = new Thread(producer);
	t2.start();
	t1.join();
	t2.join();
	Assert.that(consumer.checksum == producer.checksum);
	Statement stmt = connection.createStatement();
	ConsusResultSet cursor = 
	    (ConsusResultSet)stmt.executeQuery("select * from MetaTable where name='Queue'");
	Assert.that(cursor.next());
	MetaTable meta = (MetaTable)cursor.getSelfObject(); 
	Assert.that(meta.fields.length == 2);
	Assert.that(meta.fields[0].name.equals("head"));
	Assert.that(meta.fields[1].name.equals("nElems"));
	System.out.println("\nTest passed");
	connection.close();
    }
}
	
   
