/* Consus - Pure Java Object Relational Database Management System
 * Test of Consus Persistent Object Storage interface
 * Created by Konstantin Knizhnik, Apr 2000.
 */

import org.garret.consus.*;
import java.io.*;
import java.sql.*;
import java.util.Iterator;

class Location { 
    String country;
    String city;
    String address;

    public String toString() {
	return "country=" + country + " city=" + city + " address=" + address;
    }
}

class Company extends Persistent { 
    String       name;
    Location     location;
    POSortedList production;

    static final String[] indices={"name"};

    public String toString() {
	return "name=" + name + " " + location + " " + production;
    }
}

class Model extends Persistent { 
    String   modelId;
    Date     launchDate;
    
    public String toString() {
	return "model=" + modelId + " launched=" + launchDate;
    }    
}

class Production extends Persistent implements Comparable { 
    String     name;
    Model      model;
    Company    madeBy;
    POList     consistsOf;
    Production partOf;

    static final String[] indices={"name"};

    public String toString() {
	return getClass().getName() + " name=" + name + " " + model + " " 
	    + "made by=" + madeBy.name + " " + consistsOf;
    }
    
    public int compareTo(Object obj) { 
	return name.compareTo(((Production)obj).name);
    }
}

class Computer extends Production { 
    String category; // home-pc, office, server, ...
    public String toString() {
	return super.toString() + " category=" + category;
    }
}

class Printer extends Production { 
    String type; // matric, incjet, ...

    public String toString() {
	return super.toString() + " type=" + type;
    }
}

class Modem extends Production { 
    int speed;

    public String toString() {
	return super.toString() + " speed=" + speed;
    }
}

class Disk extends Production { 
    int capacity;
    int rotateSpeed;

    public String toString() {
	return super.toString() + " capacity=" + capacity 
	    + " rotate speed=" + rotateSpeed;
    }
}

class Resolution { 
    int width;
    int height;
    int freq;

    public String toString() {
	return "" + width + "x" + height + " at " + freq;
    }
}

class Monitor extends Production { 
    Resolution mode[];
    int        size;

    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append(super.toString());
	buf.append(" viewable size=");
	buf.append(size);
	buf.append(" resolutions=");
	String sep = "[";
	for (int i = 0; i < mode.length; i++) { 
	    buf.append(sep);
	    buf.append(mode[i].toString());
	    sep = ",";
	}
	buf.append("]");
	return buf.toString();
    }
}

class CPU extends Production { 
    int        clock;
    int        L1cache;

    public String toString() {
	return super.toString() + " clock=" + clock + " L1-cache=" + L1cache;
    }
}

class Memory extends Production { 
    int        size;
    int        speed; 

    public String toString() {
	return super.toString() + " size=" + size + " speed=" + speed;
    }
}

class Mainboard extends Production { 
    String     socketType;

    public String toString() {
	return super.toString() + " socket=" + socketType;
    }
}

public class TestPOS { 
    static BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 
    static PreparedStatement pstmt1;
    static PreparedStatement pstmt2;

    static String input(String prompt) {
	while (true) { 
	    try { 
		System.out.print(prompt);
		String answer = in.readLine().trim();
		if (answer.length() != 0) { 
		    return answer;
		}
	    } catch (IOException x) {}
	}
    }

    static int inputInt(String prompt) { 
	while (true) { 
	    try { 
		return Integer.parseInt(input(prompt));
	    } catch(NumberFormatException x) { 
		System.out.println("Bad number format");
	    }
	}
    }

    static Date inputDate(String prompt) { 
	while (true) { 
	    try { 
		return Date.valueOf(input(prompt));	    
	    } catch(Exception x) {
		System.out.println("Invalid date (YYYY-MM-DD expected)");
	    }
	}
    }	

    static boolean inputProduction(Production prod) throws Exception { 
	ResultSet cursor1 = null, cursor2 = null;
	prod.consistsOf = new POList();
	prod.name = input("Name: ");
	prod.model = new Model();
	prod.model.modelId = input("ModelId: ");
	prod.model.launchDate = inputDate("Lauch date: ");
	if (!(prod instanceof Computer)) { 
	    pstmt1.setString(1, input("Part of: "));
	    cursor1 = pstmt1.executeQuery();
	    if (!cursor1.next()) { 
		System.out.println("No such product");
		return false;
	    }
	}
	pstmt2.setString(1, input("Made by: "));
	cursor2 = pstmt2.executeQuery();
	if (!cursor2.next()) { 
	    System.out.println("No such company");
	    return false;
	}
	if (!(prod instanceof Computer)) { 	
	    prod.partOf = (Production)cursor1.getObject(0);
	    prod.partOf.consistsOf.add(prod);
	}
	prod.madeBy = (Company)cursor2.getObject(0);
	prod.madeBy.production.add(prod);
	return true;
    }

    static public void main(String args[]) throws Exception { 
	int i;
	Class.forName("org.garret.consus.ConsusDriver");
	ConsusConnection connection = 
	    (ConsusConnection)DriverManager.getConnection("jdbc:consus:TestPOS.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database TestPOS");
	    return;
	}
	Statement stmt = connection.createStatement();
	ResultSet cursor;

	cursor = stmt.executeQuery("select count(*) from Company");
	cursor.next();
	System.out.println("Totally " + cursor.getInt(1) + " companies");
	    
	cursor = stmt.executeQuery("select count(*) from Production");
	cursor.next();
	System.out.println("Totally " + cursor.getInt(1) + " products");
	stmt.close();

	pstmt1 = connection.prepareStatement("select * from Production where name like ?");
	pstmt2 = connection.prepareStatement("select * from Company where name like ?");

	Company   company;
	Computer  computer;
	CPU       cpu;
	Memory    memory;
	Mainboard mainboard;
	Disk      disk;
	Monitor   monitor;
	Printer   printer;
	Modem     modem;

	while (true) { 
	    System.out.println("Menu:\n" + 
			       "1.  Add company\n" + 
			       "2.  Add computer\n" + 
			       "3.  Add CPU\n" + 
			       "4.  Add memory\n" + 
			       "5.  Add mainboard\n" + 
			       "6.  Add disk\n" + 
			       "7.  Add monitor\n" + 
			       "8.  Add printer\n" + 
			       "9.  Add modem\n" + 
			       "10. Search for production\n" + 
			       "11. Search for company\n" + 
			       "12. Buy a company\n" + 
			       "13. Exit");
	    switch (inputInt(">")) { 
	      case 1:
		company = new Company();
		company.name = input("Company: ");
		company.location = new Location();
		company.location.country = input("Country: ");
		company.location.city    = input("City: ");
		company.location.address = input("Address: ");
		company.production = new POSortedList();
		company.store(connection);
		break;
	      case 2:
		computer = new Computer();
		if (inputProduction(computer)) { 
		    computer.category = input("Category: "); 
		    computer.store(connection);
		}
		break;
	      case 3:
		cpu = new CPU();
		if (inputProduction(cpu)) { 
		    cpu.clock = inputInt("Clock frequency (Mhz): ");
		    cpu.L1cache = inputInt("L1 cache size (Kb): ");
		    cpu.store(connection);
		}
		break;		
	      case 4:
		memory = new Memory();
		if (inputProduction(memory)) { 
		    memory.size = inputInt("Memory size (Mb): ");
		    memory.speed = inputInt("Memory frequency (MHz): ");
		    memory.store(connection);
		}
		break;
	      case 5:
		mainboard = new Mainboard();
		if (inputProduction(mainboard)) { 
		    mainboard.socketType = input("Socket type: ");
		    mainboard.store(connection);
		}
		break;
	      case 6:
		disk = new Disk();
		if (inputProduction(disk)) { 
		    disk.capacity = inputInt("Capacity (Mb): ");
		    disk.rotateSpeed = inputInt("Rotate speed: ");
		    disk.store(connection);
		}
		break;
	      case 7:
		monitor = new Monitor();
		if (inputProduction(monitor)) { 
		    int nModes = inputInt("Number of supported modes: ");
		    monitor.mode = new Resolution[nModes];
		    for (i = 0; i < nModes; i++) { 
			System.out.println("Mode " + (i+1));
			monitor.mode[i] = new Resolution();
			monitor.mode[i].width = inputInt("Width: "); 
			monitor.mode[i].height = inputInt("Height: "); 
			monitor.mode[i].freq = inputInt("Frequency: "); 
		    }
		    monitor.size = inputInt("Viewable size: ");
		    monitor.store(connection);
		}
		break;
	      case 8:
		printer = new Printer();
		if (inputProduction(printer)) { 
		    printer.type = input("Printer type: ");
		    printer.store(connection);
		}
		break;
	      case 9:
		modem = new Modem();
		if (inputProduction(modem)) { 
		    modem.speed = inputInt("Speed: ");
		    modem.store(connection);
		}
		break;
	      case 10:
		pstmt1.setString(1, input("Production name: "));
		cursor = pstmt1.executeQuery();
		while (cursor.next()) { 
		    System.out.println(cursor.getObject(0));
		}
		System.out.println("--------------------------------------");
		break;
	      case 11:
		pstmt2.setString(1, input("Company name: "));
		cursor = pstmt2.executeQuery();		
		while (cursor.next()) { 
		    System.out.println(cursor.getObject(0));
		}
		System.out.println("--------------------------------------");
		break;
	      case 12:
		pstmt2.setString(1, input("Company name: "));
		cursor = pstmt2.executeQuery();		
		if (!cursor.next()) { 
		    System.out.println("No such company!");
		} else { 
		    company = (Company)cursor.getObject(0);
		    pstmt2.setString(1, input("Which company to bue: "));
		    cursor = pstmt2.executeQuery();		
		    while (cursor.next()) { 
			Company swallowedCompany = (Company)cursor.getObject(0);
			System.out.println("Swallowed company " +  swallowedCompany.name);
			company.production.addAll(swallowedCompany.production);
			Iterator itr = swallowedCompany.production.iterator();
			while (itr.hasNext()) { 
			    Production product = (Production)itr.next();
			    product.madeBy = company;
			    product.store();
			}
			swallowedCompany.production.remove();
			swallowedCompany.remove();
		    }		    
		}
		break;
	      case 13:
		connection.close();
		return;
	    }
	    connection.commit();
	}
    }
}
