/* Consus - Pure Java Object Relational Database Management System
 * Hash table with weak reference toi the object. Weak reference is replaces with null
 *    by GC when there are no hard references to the object.
 * Created by Konstantin Knizhnik, May 2000.
 */
package consus;

import com.ms.vm.WeakReference;

class MsWeakHashTable extends WeakHashTable { 
    MsEntry table[];

    MsWeakHashTable(int initialCapacity) { 
	super(0);
	threshold = (int)(initialCapacity * loadFactor);
	table = new MsEntry[initialCapacity];
    }
    
    synchronized boolean remove(int oid) {
	MsEntry tab[] = table;
	int index = (oid & 0x7FFFFFFF) % tab.length;
	for (MsEntry e = tab[index], prev = null; e != null; prev = e, e = e.next) {
	    if (e.oid == oid) {
		count -= 1;
		if (prev != null) {
		    prev.next = e.next;
		} else {
		    tab[index] = e.next;
		}
	        return true;
	    }
	}
	return false;
    }

    synchronized void put(int oid, Object obj) { 
	WeakReference ref = new WeakReference(obj);
	MsEntry tab[] = table;
	int index = (oid & 0x7FFFFFFF) % tab.length;
	for (MsEntry e = tab[index]; e != null; e = e.next) {
	    if (e.oid == oid) {
		e.ref = ref;
		return;
	    }
	}
	if (count >= threshold) {
	    // Rehash the table if the threshold is exceeded
	    rehash();
            tab = table;
            index = (oid & 0x7FFFFFFF) % tab.length;
	} 

	// Creates the new MsEntry.
	tab[index] = new MsEntry(oid, ref, tab[index]);
	count++;
    }

    synchronized Object get(int oid) {
	MsEntry tab[] = table;
	int index = (oid & 0x7FFFFFFF) % tab.length;
	for (MsEntry e = tab[index] ; e != null ; e = e.next) {
	    if (e.oid == oid) {
		return e.ref.getReferent();
	    }
	}
	return null;
    }

    void rehash() {
	int oldCapacity = table.length;
	MsEntry oldMap[] = table;
	int i;
	for (i = oldCapacity; --i >= 0;) {
	    for (MsEntry prev = null, e = oldMap[i]; e != null; e = e.next) { 
		if (e.ref.getReferent() == null) { 
		    count -= 1;
		    if (prev == null) { 
			oldMap[i] = e.next;
		    } else { 
			prev.next = e.next;
		    }
		} else { 
		    prev = e;
		}
	    }
	}
	
	if (count <= (threshold >>> 1)) {
	    return;
	}
	int newCapacity = oldCapacity * 2 + 1;
	MsEntry newMap[] = new MsEntry[newCapacity];

	threshold = (int)(newCapacity * loadFactor);
	table = newMap;

	for (i = oldCapacity; --i >= 0 ;) {
	    for (MsEntry old = oldMap[i]; old != null; ) {
		MsEntry e = old;
		old = old.next;

		int index = (e.oid & 0x7FFFFFFF) % newCapacity;
		e.next = newMap[index];
		newMap[index] = e;
	    }
	}
    }

    /**
     * Clears this hashtable so that it contains no keys. 
     */
    synchronized void clear() {
	MsEntry tab[] = table;
	for (int index = tab.length; --index >= 0; ) { 
	    for (MsEntry e = tab[index]; e != null; e = e.next) { 
		Object o = e.ref.getReferent();
		if (o instanceof Persistent) { 
		    ((Persistent)o).unget();
		}
	    }
	    tab[index] = null;
	}
	count = 0;
    }

    synchronized void mark(int bitmap[]) {
	MsEntry tab[] = table;
	for (int index = tab.length; --index >= 0; ) { 
	    for (MsEntry e = tab[index]; e != null; e = e.next) { 
		Object o = e.ref.getReferent();
		if (o instanceof Persistent) { 
		    int oid = ((Persistent)o).oid; 
		    bitmap[oid >>> 5] |= 1 << (oid & 31);
		}
	    }
	}
    }

    int size() { 
	return count;
    }
}

class MsEntry { 
    MsEntry       next;
    WeakReference ref;
    int           oid;

    MsEntry(int oid, WeakReference ref, MsEntry chain) { 
	next = chain;
	this.oid = oid;
	this.ref = ref;
    }
}







